#!/usr/bin/env python3
# Copyright (C) 2020 Checkmk GmbH - License: GNU General Public License v2
# This file is part of Checkmk (https://checkmk.com). It is subject to the terms and
# conditions defined in the file COPYING, which is part of this source code package.
"""Service discovery

A service discovery is the automatic and reliable detection of all services to be monitored on
a host.

You can find an introduction to services including service discovery in the
[Checkmk guide](https://docs.checkmk.com/latest/en/wato_services.html).
"""

import enum
from collections.abc import Mapping, Sequence
from typing import Any, assert_never
from urllib.parse import urlparse

from cmk.utils.everythingtype import EVERYTHING

from cmk.checkengine.discovery import (
    CheckPreviewEntry,
    DiscoveryMode,
    DiscoverySettings,
)

from cmk.gui import fields as gui_fields
from cmk.gui.background_job import BackgroundStatusSnapshot
from cmk.gui.fields.utils import BaseSchema
from cmk.gui.http import request, Response
from cmk.gui.logged_in import user
from cmk.gui.openapi.endpoints.host_config.request_schemas import EXISTING_HOST_NAME
from cmk.gui.openapi.endpoints.service_discovery.response_schemas import (
    DiscoveryBackgroundJobStatusObject,
)
from cmk.gui.openapi.restful_objects import (
    constructors,
    Endpoint,
    permissions,
    response_schemas,
)
from cmk.gui.openapi.restful_objects.constructors import (
    domain_object,
    link_rel,
    object_property,
)
from cmk.gui.openapi.restful_objects.parameters import HOST_NAME
from cmk.gui.openapi.restful_objects.registry import EndpointRegistry
from cmk.gui.openapi.restful_objects.type_defs import LinkType
from cmk.gui.openapi.utils import problem, ProblemException, serve_json
from cmk.gui.site_config import site_is_local
from cmk.gui.watolib.automations import (
    fetch_service_discovery_background_job_status,
    MKAutomationException,
)
from cmk.gui.watolib.bulk_discovery import (
    bulk_discovery_job_status,
    BulkDiscoveryBackgroundJob,
    prepare_hosts_for_discovery,
    start_bulk_discovery,
)
from cmk.gui.watolib.hosts_and_folders import Host
from cmk.gui.watolib.services import (
    Discovery,
    DiscoveryAction,
    DiscoveryOptions,
    DiscoveryResult,
    DiscoveryState,
    get_check_table,
    has_discovery_action_specific_permissions,
    perform_fix_all,
    perform_host_label_discovery,
    perform_service_discovery,
    ServiceDiscoveryBackgroundJob,
)

from cmk import fields

DISCOVERY_PERMISSIONS = permissions.AllPerm(
    [
        permissions.Perm("wato.edit"),
        permissions.Optional(permissions.Perm("wato.service_discovery_to_undecided")),
        permissions.Optional(permissions.Perm("wato.service_discovery_to_monitored")),
        permissions.Optional(permissions.Perm("wato.service_discovery_to_ignored")),
        permissions.Optional(permissions.Perm("wato.service_discovery_to_removed")),
        permissions.Optional(permissions.Perm("wato.services")),
        permissions.Undocumented(permissions.Perm("background_jobs.stop_jobs")),
        permissions.Undocumented(permissions.Perm("background_jobs.stop_foreign_jobs")),
        permissions.Undocumented(permissions.Perm("background_jobs.delete_jobs")),
        permissions.Undocumented(permissions.Perm("background_jobs.delete_foreign_jobs")),
        permissions.Undocumented(permissions.Perm("wato.see_all_folders")),
    ]
)

SERVICE_DISCOVERY_PHASES = {
    "undecided": DiscoveryState.UNDECIDED,
    "vanished": DiscoveryState.VANISHED,
    "monitored": DiscoveryState.MONITORED,
    "changed": DiscoveryState.CHANGED,
    "ignored": DiscoveryState.IGNORED,
    "removed": DiscoveryState.REMOVED,
    "manual": DiscoveryState.MANUAL,
    "active": DiscoveryState.ACTIVE,
    "ignored_active": DiscoveryState.ACTIVE_IGNORED,
    "custom": DiscoveryState.CUSTOM,
    "ignored_custom": DiscoveryState.CUSTOM_IGNORED,
    "clustered_monitored": DiscoveryState.CLUSTERED_OLD,
    "clustered_undecided": DiscoveryState.CLUSTERED_NEW,
    "clustered_vanished": DiscoveryState.CLUSTERED_VANISHED,
    "clustered_ignored": DiscoveryState.CLUSTERED_IGNORED,
    "legacy": "legacy",
    "legacy_ignored": "legacy_ignored",
}


class APIDiscoveryAction(enum.Enum):
    new = "new"
    remove = "remove"
    fix_all = "fix_all"
    refresh = "refresh"
    only_host_labels = "only_host_labels"
    only_service_labels = "only_service_labels"
    tabula_rasa = "tabula_rasa"


def _discovery_mode(default_mode: str):  # type: ignore[no-untyped-def]
    # TODO: documentation should be separated for bulk discovery
    return fields.String(
        description="""The mode of the discovery action. The 'refresh' mode starts a new service
        discovery which will contact the host and identify undecided and vanished services and host
        labels. Those services and host labels can be added or removed accordingly with the
        'fix_all' mode. The 'tabula_rasa' mode combines these two procedures. The 'new', 'remove',
        'only_host_labels' and 'only_service_labels' modes give you more granular control. Both the
        'tabula_rasa' and 'refresh' modes will start a background job and the endpoint will return
        a redirect to the 'wait-for-completion' endpoint. All other modes will return an immediate
        result instead. Keep in mind that the non background job modes only work with scanned data,
        so you may need to run "refresh" first. The corresponding user interface option for each
        discovery mode is shown below.

 * `new` - Monitor undecided services
 * `remove` - Remove vanished services
 * `fix_all` - Accept all
 * `tabula_rasa` - Remove all and find new
 * `refresh` - Rescan
 * `only_host_labels` - Update host labels
 * `only_service_labels` - Update service labels
    """,
        enum=[a.value for a in APIDiscoveryAction],
        example="refresh",
        load_default=default_mode,
    )


DISCOVERY_ACTION = {
    "new": DiscoveryAction.BULK_UPDATE,
    "remove": DiscoveryAction.BULK_UPDATE,
    "fix_all": DiscoveryAction.FIX_ALL,
    "refresh": DiscoveryAction.REFRESH,
    "only_host_labels": DiscoveryAction.UPDATE_HOST_LABELS,
    "only_service_labels": DiscoveryAction.UPDATE_SERVICE_LABELS,
    "tabula_rasa": DiscoveryAction.TABULA_RASA,
}


@Endpoint(
    constructors.object_href("service_discovery", "{host_name}"),
    "cmk/list",
    method="get",
    response_schema=response_schemas.DomainObject,
    tag_group="Setup",
    path_params=[
        {
            "host_name": gui_fields.HostField(
                description="The host of the service discovery result",
                example="example.com",
                required=True,
            )
        }
    ],
    additional_status_codes=[400],
)
def show_service_discovery_result(params: Mapping[str, Any]) -> Response:
    """Show the current service discovery result"""
    host = Host.load_host(params["host_name"])

    try:
        discovery_result = get_check_table(host, DiscoveryAction.NONE, raise_errors=False)
    except MKAutomationException:
        return problem(
            status=400,
            title="Error running automation",
            detail="Please run `Show the last service discovery background job on a host` in order to get details about the error",
        )

    return serve_json(serialize_discovery_result(host, discovery_result))


class UpdateDiscoveryPhase(BaseSchema):
    check_type = fields.String(
        description="The name of the check which this service uses.",
        example="df",
        required=True,
    )
    service_item = fields.String(
        description="The value uniquely identifying the service on a given host.",
        example="/home",
        required=True,
        allow_none=True,
    )
    target_phase = fields.String(
        description="The target phase of the service.",
        enum=sorted(SERVICE_DISCOVERY_PHASES.keys()),
        example="monitored",
        required=True,
    )


@Endpoint(
    constructors.object_action_href("host", "{host_name}", "update_discovery_phase"),
    ".../modify",
    method="put",
    output_empty=True,
    tag_group="Setup",
    path_params=[
        {
            "host_name": gui_fields.HostField(
                description="The host of the service which shall be updated.",
                example="example.com",
            ),
        }
    ],
    status_descriptions={
        404: "Host could not be found",
    },
    request_schema=UpdateDiscoveryPhase,
    # TODO: CMK-10911 (permissions)
    permissions_required=permissions.AnyPerm(
        [
            permissions.Optional(
                permissions.AnyPerm(
                    [
                        permissions.Perm("wato.service_discovery_to_monitored"),
                        permissions.Perm("wato.service_discovery_to_ignored"),
                        permissions.Perm("wato.service_discovery_to_undecided"),
                        permissions.Perm("wato.service_discovery_to_removed"),
                        permissions.Undocumented(permissions.Perm("wato.see_all_folders")),
                    ]
                )
            ),
        ]
    ),
)
def update_service_phase(params: Mapping[str, Any]) -> Response:
    """Update the phase of a service"""
    body = params["body"]
    host = Host.load_host(params["host_name"])
    target_phase = body["target_phase"]
    check_type = body["check_type"]
    service_item = body["service_item"]
    _update_single_service_phase(
        SERVICE_DISCOVERY_PHASES[target_phase],
        host,
        check_type,
        service_item,
    )
    return Response(status=204)


def _update_single_service_phase(
    target_phase: str,
    host: Host,
    check_type: str,
    service_item: str | None,
) -> None:
    action = DiscoveryAction.SINGLE_UPDATE
    Discovery(
        host=host,
        action=action,
        update_target=target_phase,
        selected_services=((check_type, service_item),),
    ).do_discovery(get_check_table(host, action, raise_errors=False))


@Endpoint(
    constructors.object_href("service_discovery_run", "{host_name}"),
    "cmk/show",
    method="get",
    tag_group="Setup",
    path_params=[HOST_NAME],
    response_schema=response_schemas.DomainObject,
)
def show_service_discovery_run(params: Mapping[str, Any]) -> Response:
    """Show the last service discovery background job on a host"""
    host = Host.load_host(params["host_name"])
    snapshot = _job_snapshot(host)
    job_id = snapshot.job_id
    job_status = snapshot.status
    return serve_json(
        constructors.domain_object(
            domain_type="service_discovery_run",
            identifier=job_id,
            title=f"Service discovery background job {job_id} is {job_status.state}",
            extensions={
                "active": job_status.is_active,
                "state": job_status.state,
                "logs": {
                    "result": job_status.loginfo["JobResult"],
                    "progress": job_status.loginfo["JobProgressUpdate"],
                },
            },
            deletable=False,
            editable=False,
        )
    )


@Endpoint(
    constructors.object_action_href("service_discovery_run", "{host_name}", "wait-for-completion"),
    "cmk/wait-for-completion",
    method="get",
    status_descriptions={
        204: "The service discovery has been completed.",
        302: "The service discovery is still running. Redirecting to the "
        "'Wait for completion' endpoint.",
        404: "There is no running service discovery",
    },
    path_params=[HOST_NAME],
    additional_status_codes=[302],
    output_empty=True,
)
def service_discovery_run_wait_for_completion(params: Mapping[str, Any]) -> Response:
    """Wait for service discovery completion

    This endpoint will periodically redirect on itself to prevent timeouts.
    """
    host = Host.load_host(params["host_name"])
    snapshot = _job_snapshot(host)
    if not snapshot.exists:
        raise ProblemException(
            status=404,
            title="The requested service discovery job was not found",
            detail=f"Could not find a service discovery for host {host.name()}",
        )
    if snapshot.is_active:
        response = Response(status=302)
        response.location = urlparse(request.url).path
        return response
    return Response(status=204)


class DiscoverServices(BaseSchema):
    host_name = gui_fields.HostField(
        description="The host of the service which shall be updated.",
        example="example.com",
    )
    mode = _discovery_mode(default_mode="fix_all")


@Endpoint(
    constructors.domain_type_action_href("service_discovery_run", "start"),
    ".../update",
    method="post",
    tag_group="Setup",
    status_descriptions={
        302: "The service discovery background job has been initialized. Redirecting to the "
        "'Wait for service discovery completion' endpoint.",
        409: "A service discovery background job is currently running",
    },
    additional_status_codes=[302, 409],
    request_schema=DiscoverServices,
    response_schema=response_schemas.DomainObject,
    permissions_required=DISCOVERY_PERMISSIONS,
)
def execute_service_discovery(params: Mapping[str, Any]) -> Response:
    """Execute a service discovery on a host"""
    user.need_permission("wato.edit")
    body = params["body"]
    host = Host.load_host(body["host_name"])
    discovery_action = APIDiscoveryAction(body["mode"])
    return _execute_service_discovery(discovery_action, host)


def _execute_service_discovery(api_discovery_action: APIDiscoveryAction, host: Host) -> Response:
    job_snapshot = _job_snapshot(host)
    if job_snapshot.is_active:
        return Response(status=409)

    discovery_action = DISCOVERY_ACTION[api_discovery_action.value]
    if not has_discovery_action_specific_permissions(discovery_action, None):
        return problem(
            status=403,
            title="Permission denied",
            detail="You do not have the necessary permissions to execute this action",
        )
    discovery_result = get_check_table(host, discovery_action, raise_errors=False)
    match api_discovery_action:
        case APIDiscoveryAction.new:
            discovery_result = perform_service_discovery(
                action=discovery_action,
                discovery_result=discovery_result,
                update_source="new",
                update_target="unchanged",
                host=host,
                selected_services=EVERYTHING,
                raise_errors=False,
            )
        case APIDiscoveryAction.remove:
            discovery_result = perform_service_discovery(
                action=discovery_action,
                discovery_result=discovery_result,
                update_source="vanished",
                update_target="removed",
                host=host,
                selected_services=EVERYTHING,
                raise_errors=False,
            )
        case APIDiscoveryAction.fix_all:
            discovery_result = perform_fix_all(
                discovery_result=discovery_result,
                host=host,
                raise_errors=False,
            )
        case APIDiscoveryAction.refresh | APIDiscoveryAction.tabula_rasa:
            discovery_run = _discovery_wait_for_completion_link(host.name())
            response = Response(status=302)
            response.location = urlparse(discovery_run["href"]).path
            return response
        case APIDiscoveryAction.only_host_labels:
            discovery_result = perform_host_label_discovery(
                action=discovery_action,
                discovery_result=discovery_result,
                host=host,
                raise_errors=False,
            )
        case APIDiscoveryAction.only_service_labels:
            discovery_result = perform_service_discovery(
                action=discovery_action,
                discovery_result=discovery_result,
                update_source="changed",
                update_target="unchanged",
                host=host,
                selected_services=EVERYTHING,
                raise_errors=False,
            )

        case _:
            assert_never(api_discovery_action)

    return serve_json(serialize_discovery_result(host, discovery_result))


def _discovery_wait_for_completion_link(host_name: str) -> LinkType:
    return constructors.link_endpoint(
        "cmk.gui.openapi.endpoints.service_discovery",
        "cmk/wait-for-completion",
        parameters={"host_name": host_name},
    )


def _in_phase(phase_to_check: str, discovery_phases: list[str]) -> bool:
    for phase in list(discovery_phases):
        if SERVICE_DISCOVERY_PHASES[phase] == phase_to_check:
            return True
    return False


def _lookup_phase_name(internal_phase_name: str) -> str:
    for key, value in SERVICE_DISCOVERY_PHASES.items():
        if value == internal_phase_name:
            return key
    raise ValueError(f"Key {internal_phase_name} not found in dict.")


def serialize_discovery_result(  # type: ignore[no-untyped-def]
    host: Host,
    discovery_result: DiscoveryResult,
):
    services = {}
    host_name = host.name()
    for entry in discovery_result.check_table:
        service_phase = _lookup_phase_name(entry.check_source)
        services[f"{entry.check_plugin_name}-{entry.item}"] = object_property(
            name=entry.description,
            title=f"The service is currently {service_phase!r}",
            value=service_phase,
            prop_format="string",
            linkable=False,
            extensions={
                "host_name": host_name,
                "check_plugin_name": entry.check_plugin_name,
                "service_name": entry.description,
                "service_item": entry.item,
                "service_phase": service_phase,
            },
            base="",
            links=[
                link_rel(
                    rel="cmk/service.move-monitored",
                    href=update_service_phase.path.format(host_name=host_name),
                    body_params={
                        "target_phase": "monitored",
                        "check_type": entry.check_plugin_name,
                        "service_item": entry.item,
                    },
                    method="put",
                    title="Move the service to monitored",
                ),
                link_rel(
                    rel="cmk/service.move-undecided",
                    href=update_service_phase.path.format(host_name=host_name),
                    body_params={
                        "target_phase": "undecided",
                        "check_type": entry.check_plugin_name,
                        "service_item": entry.item,
                    },
                    method="put",
                    title="Move the service to undecided",
                ),
                link_rel(
                    rel="cmk/service.move-ignored",
                    href=update_service_phase.path.format(host_name=host_name),
                    body_params={
                        "target_phase": "ignored",
                        "check_type": entry.check_plugin_name,
                        "service_item": entry.item,
                    },
                    method="put",
                    title="Move the service to ignored",
                ),
            ],
        )

    return domain_object(
        domain_type="service_discovery",
        identifier=f"service_discovery-{host_name}",
        title=f"Service discovery result of host {host_name}",
        editable=False,
        deletable=False,
        extensions={
            "check_table": services,
            "host_labels": discovery_result.host_labels,
            "vanished_labels": discovery_result.vanished_labels,
            "changed_labels": discovery_result.changed_labels,
        },
    )


# Bulk discovery

JOB_ID = {
    "job_id": fields.String(
        description="The unique identifier of the background job executing the bulk discovery",
        example="bulk_discovery",
        required=True,
    ),
}


class BulkDiscoveryOptions(BaseSchema):
    monitor_undecided_services = fields.Boolean(
        required=False,
        description="The option whether to monitor undecided services or not.",
        example=True,
        load_default=False,
    )
    remove_vanished_services = fields.Boolean(
        required=False,
        description="The option whether to remove vanished services or not.",
        example=True,
        load_default=False,
    )
    update_service_labels = fields.Boolean(
        required=False,
        description="The option whether to update service labels or not.",
        example=True,
        load_default=False,
    )
    update_host_labels = fields.Boolean(
        required=False,
        description="The option whether to update host labels or not.",
        example=True,
        load_default=False,
    )


class BulkDiscovery(BaseSchema):
    hostnames = fields.List(
        EXISTING_HOST_NAME,
        required=True,
        example=["example", "sample"],
        description="A list of host names",
    )
    mode = fields.String(
        description="""The mode of the bulk discovery action. The modes 'new', 'remove' and
        'only_host_labels' modes give you granular control to monitor undecided services, add new
        host labels and remove vanished services respectively. The mode 'fix_all' groups those
        individual actions into one single mode. The 'tabula_rasa' and 'refresh' modes both perform
        a complete refresh of the services and adds new host labels. Both modes perform the same
        action and are kept for compatibility reasons. All modes will start a background job. The
        corresponding user interface option for each discovery mode is shown
        below.

* `new` - Monitor unmonitored services and new host labels
* `remove` - Remove vanished services
* `fix_all` - Add unmonitored services and new host labels, remove vanished services
* `tabula_rasa & refresh` - Refresh all services (tabula rasa), add new host labels
* `only_host_labels` - Only discover new host labels
        """,
        enum=[a.value for a in APIDiscoveryAction],
        example="refresh",
        load_default="new",
        metadata={"deprecated": True},
    )
    options = fields.Nested(
        BulkDiscoveryOptions,
        description="The discovery options for the bulk discovery. The options if specified take "
        "precedence over the mode.",
        required=False,
        example={
            "monitor_undecided": True,
            "remove_vanished": True,
            "update_service_labels": True,
        },
    )
    do_full_scan = fields.Boolean(
        required=False,
        description="The option whether to perform a full scan or not.",
        example=True,
        load_default=True,
    )
    bulk_size = fields.Integer(
        required=False,
        description="The number of hosts to be handled at once.",
        example=10,
        load_default=10,
    )
    ignore_errors = fields.Boolean(
        required=False,
        description="The option whether to ignore errors in single check plug-ins.",
        example=True,
        load_default=True,
    )


@Endpoint(
    constructors.domain_type_action_href("discovery_run", "bulk-discovery-start"),
    "cmk/activate",
    method="post",
    status_descriptions={
        409: "A bulk discovery job is already active",
    },
    additional_status_codes=[409],
    request_schema=BulkDiscovery,
    response_schema=DiscoveryBackgroundJobStatusObject,
)
def execute_bulk_discovery(params: Mapping[str, Any]) -> Response:
    """Start a bulk discovery job"""
    # TODO: documentation should be adjusted; initial fix for resolving tests
    body = params["body"]
    job = BulkDiscoveryBackgroundJob()
    if job.is_active():
        return Response(status=409)

    options = body.get("options")
    if options is not None:
        discovery_settings = DiscoverySettings(
            update_host_labels=options["update_host_labels"],
            add_new_services=options["monitor_undecided_services"],
            remove_vanished_services=options["remove_vanished_services"],
            update_changed_service_labels=options["update_service_labels"],
            update_changed_service_parameters=False,
        )
    else:
        mode = body["mode"]
        if mode == "tabula_rasa":
            mode = "refresh"
        elif mode == "only_host_labels":
            mode = "only-host-labels"
        elif mode == "fix_all":
            mode = "fixall"

        discovery_settings = DiscoverySettings.from_discovery_mode(DiscoveryMode.from_str(mode))

    hosts_to_discover = prepare_hosts_for_discovery(body["hostnames"])
    start_bulk_discovery(
        job,
        hosts_to_discover,
        discovery_settings,
        body["do_full_scan"],
        body["ignore_errors"],
        body["bulk_size"],
    )

    return _serve_background_job(job)


@Endpoint(
    constructors.object_href("discovery_run", "{job_id}"),
    "cmk/show",
    method="get",
    path_params=[JOB_ID],
    status_descriptions={
        404: "There is no running background job with this job_id.",
    },
    response_schema=DiscoveryBackgroundJobStatusObject,
)
def show_bulk_discovery_status(params: Mapping[str, Any]) -> Response:
    """Show the status of a bulk discovery job"""

    job_id = params["job_id"]
    job = BulkDiscoveryBackgroundJob()
    if job.get_job_id() != job_id:
        raise ProblemException(
            status=404,
            title="The requested background job_id was not found",
            detail=f"Could not find a background job with id {job_id}.",
        )
    return _serve_background_job(job)


def _job_snapshot(host: Host) -> BackgroundStatusSnapshot:
    if site_is_local(host.site_id()):
        job = ServiceDiscoveryBackgroundJob(host.name())
        return job.get_status_snapshot()

    return fetch_service_discovery_background_job_status(host.site_id(), host.name())


def _discovery_options(action_mode: DiscoveryAction) -> DiscoveryOptions:
    return DiscoveryOptions(
        action=action_mode,
        show_checkboxes=False,
        show_parameters=False,
        show_discovered_labels=False,
        show_plugin_names=False,
        ignore_errors=True,
    )


def _serve_background_job(job: BulkDiscoveryBackgroundJob) -> Response:
    job_id = job.get_job_id()
    status_details = bulk_discovery_job_status(job)
    return serve_json(
        constructors.domain_object(
            domain_type="discovery_run",
            identifier=job_id,
            title=f"Background job {job_id} {'is active' if status_details['is_active'] else 'is finished'}",
            extensions={
                "active": status_details["is_active"],
                "state": status_details["job_state"],
                "logs": status_details["logs"],
            },
            deletable=False,
            editable=False,
        )
    )


def _serve_services(
    host: Host,
    discovered_services: Sequence[CheckPreviewEntry],
    discovery_phases: list[str],
) -> Response:
    members = {}
    host_name = host.name()
    for entry in discovered_services:
        if _in_phase(entry.check_source, discovery_phases):
            service_phase = _lookup_phase_name(entry.check_source)
            members[f"{entry.check_plugin_name}-{entry.item}"] = object_property(
                name=entry.description,
                title=f"The service is currently {service_phase!r}",
                value=service_phase,
                prop_format="string",
                linkable=False,
                extensions={
                    "host_name": host_name,
                    "check_plugin_name": entry.check_plugin_name,
                    "service_name": entry.description,
                    "service_item": entry.item,
                    "service_phase": service_phase,
                },
                base="",
                links=[
                    link_rel(
                        rel="cmk/service.move-monitored",
                        href=update_service_phase.path.format(host_name=host_name),
                        body_params={
                            "target_phase": "monitored",
                            "check_type": entry.check_plugin_name,
                            "service_item": entry.item,
                        },
                        method="put",
                        title="Move the service to monitored",
                    ),
                    link_rel(
                        rel="cmk/service.move-undecided",
                        href=update_service_phase.path.format(host_name=host_name),
                        body_params={
                            "target_phase": "undecided",
                            "check_type": entry.check_plugin_name,
                            "service_item": entry.item,
                        },
                        method="put",
                        title="Move the service to undecided",
                    ),
                    link_rel(
                        rel="cmk/service.move-ignored",
                        href=update_service_phase.path.format(host_name=host_name),
                        body_params={
                            "target_phase": "ignored",
                            "check_type": entry.check_plugin_name,
                            "service_item": entry.item,
                        },
                        method="put",
                        title="Move the service to ignored",
                    ),
                ],
            )

    return serve_json(
        domain_object(
            domain_type="service_discovery",
            identifier=f"{host_name}-services-wato",
            title="Services discovery",
            members=members,
            editable=False,
            deletable=False,
            extensions={},
        )
    )


def register(endpoint_registry: EndpointRegistry) -> None:
    endpoint_registry.register(show_service_discovery_result)
    endpoint_registry.register(update_service_phase)
    endpoint_registry.register(show_service_discovery_run)
    endpoint_registry.register(service_discovery_run_wait_for_completion)
    endpoint_registry.register(execute_service_discovery)
    endpoint_registry.register(execute_bulk_discovery)
    endpoint_registry.register(show_bulk_discovery_status)
