#!/usr/bin/env python3
# Copyright (C) 2019 Checkmk GmbH - License: GNU General Public License v2
# This file is part of Checkmk (https://checkmk.com). It is subject to the terms and
# conditions defined in the file COPYING, which is part of this source code package.

# Auto-formatting is disabled in the next section to get simpler overview of the keys
# present in the dictionary as they are packed together much more
# densely. You are responsible of manually formatting this dictionary, or
# at your choice adding temporarily in the style file the option:
# each_dict_entry_on_separate_line=False

from cmk.gui.i18n import _l
from cmk.gui.inventory.filters import (
    FilterInvtableAdminStatus,
    FilterInvtableAvailable,
    FilterInvtableIDRange,
    FilterInvtableInterfaceType,
    FilterInvtableOperStatus,
    FilterInvtableTimestampAsAge,
    FilterInvtableVersion,
)
from cmk.gui.num_split import cmp_version
from cmk.gui.utils.sort import cmp_natural_sort

from .registry import InventoryHintRegistry


# fmt: off
def register(inventory_displayhints: InventoryHintRegistry) -> None:
    inventory_displayhints.update({
        ".hardware.": {"title": _l("Hardware"), "icon": "hardware"},
        ".hardware.chassis.": {"title": _l("Chassis")},
        ".hardware.cpu.": {
            "title": _l("Processor"),
            "keyorder": [
                "arch",
                "max_speed",
                "model",
                "type",
                "threads",
                "smt_threads",
                "cpu_max_capa",
                "cpus",
                "logical_cpus",
                "cores",
                "cores_per_cpu",
                "threads_per_cpu",
                "cache_size",
                "bus_speed",
                "voltage",
                "sharing_mode",
                "implementation_mode",
                "entitlement",
            ],
        },
        ".hardware.cpu.arch": {"title": _l("CPU architecture")},
        ".hardware.cpu.max_speed": {"title": _l("Maximum speed"), "paint": "hz"},
        ".hardware.cpu.model": {"title": _l("CPU model")},
        ".hardware.cpu.type": {"title": _l("CPU type")},
        ".hardware.cpu.threads": {"title": _l("#Hyperthreads"), "paint": "count"},
        ".hardware.cpu.smt_threads": {"title": _l("SMT threads"), "paint": "count"},
        ".hardware.cpu.sharing_mode": {"title": _l("Shared processor mode")},
        ".hardware.cpu.implementation_mode": {"title": _l("Processor implementation mode")},
        ".hardware.cpu.entitlement": {"title": _l("Processor entitled capacity")},
        ".hardware.cpu.cpu_max_capa": {"title": _l("Processor max. capacity")},
        ".hardware.cpu.cache_size": {"title": _l("Cache size"), "paint": "bytes"},
        ".hardware.cpu.bus_speed": {"title": _l("Bus speed"), "paint": "hz"},
        ".hardware.cpu.voltage": {"title": _l("Voltage"), "paint": "volt"},
        ".hardware.cpu.cores_per_cpu": {"title": _l("Cores per CPU"), "paint": "count"},
        ".hardware.cpu.threads_per_cpu": {"title": _l("Hyperthreads per CPU"), "paint": "count"},
        ".hardware.cpu.cpus": {"title": _l("#Physical CPUs"), "paint": "count"},
        ".hardware.cpu.logical_cpus": {"title": _l("#Logical CPUs"), "paint": "count"},
        ".hardware.cpu.cores": {"title": _l("#Cores"), "paint": "count"},
        ".hardware.cpu.nodes:": {
            "title": _l("Node processor"),
            "keyorder": [
                "node_name",
                "cores",
                "model"
            ]
        },
        ".hardware.cpu.nodes:*.node_name": {"title": _l("Node name")},
        ".hardware.cpu.nodes:*.cores": {"title": _l("#Cores"), "paint": "count"},
        ".hardware.cpu.nodes:*.model": {"title": _l("CPU model")},
        ".hardware.memory.": {
            "title": _l("Memory (RAM)"),
            "keyorder": ["total_ram_usable", "total_swap", "total_vmalloc"],
        },
        ".hardware.memory.total_ram_usable": {
            "title": _l("Total usable RAM"),
            "paint": "bytes_rounded",
        },
        ".hardware.memory.total_swap": {"title": _l("Total swap space"), "paint": "bytes_rounded"},
        ".hardware.memory.total_vmalloc": {
            "title": _l("Virtual addresses for mapping"),
            "paint": "bytes_rounded",
        },
        # FIXME Cleanup: This path has no table.
        ".hardware.memory.arrays:": {
            "title": _l("Arrays (Controllers)"),
            "keyorder": ["maximum_capacity"],
        },
        ".hardware.memory.arrays:*.maximum_capacity": {
            "title": _l("Maximum capacity"),
            "paint": "bytes",
        },
        ".hardware.memory.arrays:*.": {"title": _l("Controller %d")},
        ".hardware.memory.arrays:*.devices:": {
            "title": _l("Devices"),
            "keyorder": [
                "locator",
                "bank_locator",
                "type",
                "form_factor",
                "speed",
                "data_width",
                "total_width",
                "manufacturer",
                "serial",
                "size",
            ],
        },
        ".hardware.memory.arrays:*.devices:*.locator": {"title": _l("Locator")},
        ".hardware.memory.arrays:*.devices:*.bank_locator": {"title": _l("Bank locator")},
        ".hardware.memory.arrays:*.devices:*.type": {"title": _l("Type")},
        ".hardware.memory.arrays:*.devices:*.form_factor": {"title": _l("Form factor")},
        ".hardware.memory.arrays:*.devices:*.speed": {"title": _l("Speed"), "paint": "hz"},
        ".hardware.memory.arrays:*.devices:*.data_width": {"title": _l("Data width")},
        ".hardware.memory.arrays:*.devices:*.total_width": {"title": _l("Total width")},
        ".hardware.memory.arrays:*.devices:*.manufacturer": {"title": _l("Manufacturer")},
        ".hardware.memory.arrays:*.devices:*.serial": {"title": _l("Serial")},
        ".hardware.memory.arrays:*.devices:*.size": {"title": _l("Size"), "paint": "bytes"},
        ".hardware.system.": {
            "title": _l("System"),
            "keyorder": [
                "manufacturer",
                "product",
                "serial",
                "model",
                "node_name",
                "partition_name",
                "expresscode",
                "pki_appliance_version",
                "device_number",
                "description",
                "mac_address",
            ],
        },
        ".hardware.system.product": {"title": _l("Product"), "is_show_more": False},
        ".hardware.system.serial": {"title": _l("Serial number")},
        ".hardware.system.expresscode": {"title": _l("Express servicecode")},
        ".hardware.system.model": {"title": _l("Model name")},
        ".hardware.system.node_name": {"title": _l("Node name")},
        ".hardware.system.manufacturer": {"title": _l("Manufacturer")},
        ".hardware.system.partition_name": {"title": _l("Partition name")},
        ".hardware.system.pki_appliance_version": {"title": _l("Version of PKI appliance")},
        ".hardware.system.device_number": {"title": _l("Device number")},
        ".hardware.system.description": {"title": _l("Description")},
        ".hardware.system.mac_address": {"title": _l("MAC address")},
        # Legacy ones. Kept to not break existing views - DON'T use these values for new plugins
        ".hardware.system.serial_number": {"title": _l("Serial number - LEGACY, don't use")},
        ".hardware.system.model_name": {"title": _l("Model name - LEGACY, don't use")},
        ".hardware.system.nodes:": {
            "title": _l("Node system"),
            "keyorder": [
                "node_name",
                "id",
                "model",
                "product",
                "serial",
            ],
        },
        ".hardware.system.nodes:*.node_name": {"title": _l("Node name")},
        ".hardware.system.nodes:*.id": {"title": _l("ID")},
        ".hardware.system.nodes:*.model": {"title": _l("Model name")},
        ".hardware.system.nodes:*.product": {"title": _l("Product")},
        ".hardware.system.nodes:*.serial": {"title": _l("Serial number")},
        ".hardware.components.": {"title": _l("Physical components")},
        ".hardware.components.others:": {
            "title": _l("Other entities"),
            "keyorder": [
                "index",
                "name",
                "description",
                "software",
                "serial",
                "manufacturer",
                "model",
                "location",
            ],
            "view": "invother",
        },
        ".hardware.components.others:*.index": {"title": _l("Index")},
        ".hardware.components.others:*.name": {"title": _l("Name")},
        ".hardware.components.others:*.description": {"title": _l("Description")},
        ".hardware.components.others:*.software": {"title": _l("Software")},
        ".hardware.components.others:*.serial": {"title": _l("Serial number")},
        ".hardware.components.others:*.manufacturer": {"title": _l("Manufacturer")},
        ".hardware.components.others:*.model": {"title": _l("Model name")},
        ".hardware.components.others:*.location": {"title": _l("Location")},
        ".hardware.components.unknowns:": {
            "title": _l("Unknown entities"),
            "keyorder": [
                "index",
                "name",
                "description",
                "software",
                "serial",
                "manufacturer",
                "model",
                "location",
            ],
            "view": "invunknown",
        },
        ".hardware.components.unknowns:*.index": {"title": _l("Index")},
        ".hardware.components.unknowns:*.name": {"title": _l("Name")},
        ".hardware.components.unknowns:*.description": {"title": _l("Description")},
        ".hardware.components.unknowns:*.software": {"title": _l("Software")},
        ".hardware.components.unknowns:*.serial": {"title": _l("Serial number")},
        ".hardware.components.unknowns:*.manufacturer": {"title": _l("Manufacturer")},
        ".hardware.components.unknowns:*.model": {"title": _l("Model name")},
        ".hardware.components.unknowns:*.location": {"title": _l("Location")},
        ".hardware.components.chassis:": {
            "title": _l("Chassis"),
            "keyorder": [
                "index",
                "name",
                "description",
                "software",
                "serial",
                "manufacturer",
                "model",
                "location",
            ],
            "view": "invchassis",
        },
        ".hardware.components.chassis:*.index": {"title": _l("Index")},
        ".hardware.components.chassis:*.name": {"title": _l("Name")},
        ".hardware.components.chassis:*.description": {"title": _l("Description")},
        ".hardware.components.chassis:*.software": {"title": _l("Software")},
        ".hardware.components.chassis:*.serial": {"title": _l("Serial number")},
        ".hardware.components.chassis:*.manufacturer": {"title": _l("Manufacturer")},
        ".hardware.components.chassis:*.model": {"title": _l("Model name")},
        ".hardware.components.chassis:*.location": {"title": _l("Location")},
        ".hardware.components.backplanes:": {
            "title": _l("Backplanes"),
            "keyorder": [
                "index",
                "name",
                "description",
                "software",
                "serial",
                "manufacturer",
                "model",
                "location",
            ],
            "view": "invbackplane",
        },
        ".hardware.components.backplanes:*.index": {"title": _l("Index")},
        ".hardware.components.backplanes:*.name": {"title": _l("Name")},
        ".hardware.components.backplanes:*.description": {"title": _l("Description")},
        ".hardware.components.backplanes:*.software": {"title": _l("Software")},
        ".hardware.components.backplanes:*.serial": {"title": _l("Serial number")},
        ".hardware.components.backplanes:*.manufacturer": {"title": _l("Manufacturer")},
        ".hardware.components.backplanes:*.model": {"title": _l("Model name")},
        ".hardware.components.backplanes:*.location": {"title": _l("Location")},
        ".hardware.components.containers:": {
            "title": _l("HW containers"),
            "keyorder": [
                "index",
                "name",
                "description",
                "software",
                "serial",
                "manufacturer",
                "model",
                "location",
            ],
            "view": "invcontainer",
        },
        ".hardware.components.containers:*.index": {"title": _l("Index")},
        ".hardware.components.containers:*.name": {"title": _l("Name")},
        ".hardware.components.containers:*.description": {"title": _l("Description")},
        ".hardware.components.containers:*.software": {"title": _l("Software")},
        ".hardware.components.containers:*.serial": {"title": _l("Serial number")},
        ".hardware.components.containers:*.manufacturer": {"title": _l("Manufacturer")},
        ".hardware.components.containers:*.model": {"title": _l("Model name")},
        ".hardware.components.containers:*.location": {"title": _l("Location")},
        ".hardware.components.psus:": {
            "title": _l("Power supplies"),
            "keyorder": [
                "index",
                "name",
                "description",
                "software",
                "serial",
                "manufacturer",
                "model",
                "location",
            ],
            "view": "invpsu",
        },
        ".hardware.components.psus:*.index": {"title": _l("Index")},
        ".hardware.components.psus:*.name": {"title": _l("Name")},
        ".hardware.components.psus:*.description": {"title": _l("Description")},
        ".hardware.components.psus:*.software": {"title": _l("Software")},
        ".hardware.components.psus:*.serial": {"title": _l("Serial number")},
        ".hardware.components.psus:*.manufacturer": {"title": _l("Manufacturer")},
        ".hardware.components.psus:*.model": {"title": _l("Model name")},
        ".hardware.components.psus:*.location": {"title": _l("Location")},
        ".hardware.components.fans:": {
            "title": _l("Fans"),
            "keyorder": [
                "index",
                "name",
                "description",
                "software",
                "serial",
                "manufacturer",
                "model",
                "location",
            ],
            "view": "invfan",
        },
        ".hardware.components.fans:*.index": {"title": _l("Index")},
        ".hardware.components.fans:*.name": {"title": _l("Name")},
        ".hardware.components.fans:*.description": {"title": _l("Description")},
        ".hardware.components.fans:*.software": {"title": _l("Software")},
        ".hardware.components.fans:*.serial": {"title": _l("Serial number")},
        ".hardware.components.fans:*.manufacturer": {"title": _l("Manufacturer")},
        ".hardware.components.fans:*.model": {"title": _l("Model name")},
        ".hardware.components.fans:*.location": {"title": _l("Location")},
        ".hardware.components.sensors:": {
            "title": _l("Sensors"),
            "keyorder": [
                "index",
                "name",
                "description",
                "software",
                "serial",
                "manufacturer",
                "model",
                "location",
            ],
            "view": "invsensor",
        },
        ".hardware.components.sensors:*.index": {"title": _l("Index")},
        ".hardware.components.sensors:*.name": {"title": _l("Name")},
        ".hardware.components.sensors:*.description": {"title": _l("Description")},
        ".hardware.components.sensors:*.software": {"title": _l("Software")},
        ".hardware.components.sensors:*.serial": {"title": _l("Serial number")},
        ".hardware.components.sensors:*.manufacturer": {"title": _l("Manufacturer")},
        ".hardware.components.sensors:*.model": {"title": _l("Model name")},
        ".hardware.components.sensors:*.location": {"title": _l("Location")},
        ".hardware.components.modules:": {
            "title": _l("Modules"),
            "keyorder": [
                "index",
                "name",
                "description",
                "software",
                "serial",
                "model",
                "manufacturer",
                "bootloader",
                "firmware",
                "type",
                "location",
            ],
            "view": "invmodule",
        },
        ".hardware.components.modules:*.index": {"title": _l("Index")},
        ".hardware.components.modules:*.name": {"title": _l("Name")},
        ".hardware.components.modules:*.description": {"title": _l("Description")},
        ".hardware.components.modules:*.software": {"title": _l("Software")},
        ".hardware.components.modules:*.serial": {"title": _l("Serial number")},
        ".hardware.components.modules:*.model": {"title": _l("Model name")},
        ".hardware.components.modules:*.manufacturer": {"title": _l("Manufacturer")},
        ".hardware.components.modules:*.bootloader": {"title": _l("Bootloader")},
        ".hardware.components.modules:*.firmware": {"title": _l("Firmware")},
        ".hardware.components.modules:*.type": {"title": _l("Type")},
        ".hardware.components.modules:*.location": {"title": _l("Location")},
        ".hardware.components.stacks:": {
            "title": _l("Stacks"),
            "keyorder": [
                "index",
                "name",
                "description",
                "software",
                "serial",
                "manufacturer",
                "model",
                "location",
            ],
            "view": "invstack",
        },
        ".hardware.components.stacks:*.index": {"title": _l("Index")},
        ".hardware.components.stacks:*.name": {"title": _l("Name")},
        ".hardware.components.stacks:*.description": {"title": _l("Description")},
        ".hardware.components.stacks:*.software": {"title": _l("Software")},
        ".hardware.components.stacks:*.serial": {"title": _l("Serial number")},
        ".hardware.components.stacks:*.manufacturer": {"title": _l("Manufacturer")},
        ".hardware.components.stacks:*.model": {"title": _l("Model name")},
        ".hardware.components.stacks:*.location": {"title": _l("Location")},
        ".hardware.storage.": {"title": _l("Storage")},
        ".hardware.storage.controller.": {"title": _l("Controller"), "keyorder": ["version"]},
        ".hardware.storage.controller.version": {"title": _l("Version")},
        ".hardware.storage.disks:": {
            "title": _l("Block devices"),
            # FIXME Only use key columns used in inventory_lnx_block_devices, inventory_win_disks
            # hp_proliant_da_phydrv, netapp_api_disk. Cleanup these different entries.
            # See also
            # tests/unit/cmk/gui/plugins/views/test_views_builtin_inventory_plugins.py::_IGNORED_KEYS_BY_PATH
            "keyorder": ["fsnode", "controller", "signature"],
        },
        ".hardware.storage.disks:*.": {"title": _l("Block device %d")},
        ".hardware.storage.disks:*.fsnode": {"title": _l("Filesystem node")},
        ".hardware.storage.disks:*.controller": {"title": _l("Controller")},
        ".hardware.storage.disks:*.drive_index": {"title": _l("Drive")},
        ".hardware.storage.disks:*.signature": {"title": _l("Disk ID")},
        ".hardware.storage.disks:*.vendor": {"title": _l("Vendor")},
        ".hardware.storage.disks:*.local": {"title": _l("Local")},
        ".hardware.storage.disks:*.bus": {"title": _l("Bus")},
        ".hardware.storage.disks:*.product": {"title": _l("Product")},
        ".hardware.storage.disks:*.serial": {"title": _l("Serial number")},
        ".hardware.storage.disks:*.size": {"title": _l("Size"), "paint": "size"},
        ".hardware.storage.disks:*.type": {"title": _l("Type")},
        # this displayhint is for Attributes and NOT for above Table
        ".hardware.storage.disks.size": {"title": _l("Size"), "paint": "size"},
        ".hardware.volumes.": {"title": _l("Volumes")},
        ".hardware.volumes.physical_volumes:": {
            "title": _l("Physical volumes"),
            "keyorder": [
                "volume_group_name",
                "physical_volume_name",
                "physical_volume_status",
                "physical_volume_total_partitions",
                "physical_volume_free_partitions",
            ],
        },
        ".hardware.volumes.physical_volumes:*.volume_group_name": {
            "title": _l("Volume group name")
        },
        ".hardware.volumes.physical_volumes:*.physical_volume_name": {
            "title": _l("Physical volume name")
        },
        ".hardware.volumes.physical_volumes:*.physical_volume_status": {
            "title": _l("Physical volume status")
        },
        ".hardware.volumes.physical_volumes:*.physical_volume_total_partitions": {
            "title": _l("Physical volume total partitions")
        },
        ".hardware.volumes.physical_volumes:*.physical_volume_free_partitions": {
            "title": _l("Physical volume free partitions")
        },
        ".hardware.video:": {
            "title": _l("Graphic cards"),
            "keyorder": [
                "name",
                "subsystem",
                "driver",
                "driver_version",
                "driver_date",
                "graphic_memory",
            ],
        },
        ".hardware.video:*.": {"title": _l("Graphic card %d")},
        ".hardware.video:*.name": {"title": _l("Graphic card name")},
        ".hardware.video:*.subsystem": {"title": _l("Vendor and device ID")},
        ".hardware.video:*.driver": {"title": _l("Driver")},
        ".hardware.video:*.driver_date": {"title": _l("Driver date")},
        ".hardware.video:*.driver_version": {"title": _l("Driver version")},
        ".hardware.video:*.graphic_memory": {"title": _l("Memory"), "paint": "bytes_rounded"},
        ".hardware.nwadapter:": {
            "title": _l("Network adapters"),
            "keyorder": [
                "name",
                "type",
                "macaddress",
                "speed",
                "gateway",
                "ipv4_address",
                "ipv6_address",
                "ipv4_subnet",
                "ipv6_subnet",
            ],
        },
        ".hardware.nwadapter:*.": {"title": _l("Network adapter %d")},
        ".hardware.nwadapter:*.name": {"title": _l("Name")},
        ".hardware.nwadapter:*.type": {"title": _l("Type")},
        ".hardware.nwadapter:*.macaddress": {"title": _l("Physical address (MAC)")},
        ".hardware.nwadapter:*.speed": {"title": _l("Speed"), "paint": "nic_speed"},
        ".hardware.nwadapter:*.ipv4_address": {"title": _l("IPv4 address")},
        ".hardware.nwadapter:*.ipv4_subnet": {"title": _l("IPv4 subnet")},
        ".hardware.nwadapter:*.ipv6_address": {"title": _l("IPv6 address")},
        ".hardware.nwadapter:*.ipv6_subnet": {"title": _l("IPv6 subnet")},
        ".hardware.nwadapter:*.gateway": {"title": _l("Gateway")},
        ".software.": {"title": _l("Software"), "icon": "software"},
        ".software.applications.": {"title": _l("Applications")},
        ".software.applications.check_mk.": {
            "title": _l("Checkmk"),
            "keyorder": ["num_hosts", "num_services"],
        },
        ".software.applications.check_mk.num_hosts": {
            "title": _l("#Hosts"),
            "sort": cmp_natural_sort,
        },
        ".software.applications.check_mk.num_services": {
            "title": _l("#Services"),
            "sort": cmp_natural_sort,
        },
        ".software.applications.check_mk.versions:": {
            "title": _l("Checkmk versions"),
            "keyorder": ["version", "number", "edition", "demo", "num_sites"],
            "view": "invcmkversions",
        },
        ".software.applications.check_mk.versions:*.version": {"title": _l("Version")},
        ".software.applications.check_mk.versions:*.number": {"title": _l("Number")},
        ".software.applications.check_mk.versions:*.edition": {"title": _l("Edition")},
        ".software.applications.check_mk.versions:*.demo": {"title": _l("Demo"), "paint": "bool"},
        ".software.applications.check_mk.versions:*.num_sites": {"title": _l("#Sites")},
        ".software.applications.check_mk.sites:": {
            "title": _l("Checkmk sites"),
            "keyorder": [
                "site",
                "used_version",
                "num_hosts",
                "num_services",
                "check_mk_helper_usage",
                "fetcher_helper_usage",
                "checker_helper_usage",
                "livestatus_usage",
                "check_helper_usage",
                "autostart",
                "apache",
                "cmc",
                "crontab",
                "dcd",
                "liveproxyd",
                "mkeventd",
                "mknotifyd",
                "rrdcached",
                "stunnel",
                "xinetd",
                "nagios",
                "npcd",
            ],
            "view": "invcmksites",
        },
        ".software.applications.check_mk.sites:*.site": {"title": _l("Site")},
        ".software.applications.check_mk.sites:*.used_version": {"title": _l("Version")},
        ".software.applications.check_mk.sites:*.num_hosts": {
            "title": _l("#Hosts"),
            "sort": cmp_natural_sort,
        },
        ".software.applications.check_mk.sites:*.num_services": {
            "title": _l("#Services"),
            "sort": cmp_natural_sort,
        },
        ".software.applications.check_mk.sites:*.check_mk_helper_usage": {
            "title": _l("CMK helper usage"),
            "short": _l("CMK helper")
        },
        ".software.applications.check_mk.sites:*.fetcher_helper_usage": {
            "title": _l("Fetcher helper usage"),
            "short": _l("Fetcher helper")
        },
        ".software.applications.check_mk.sites:*.checker_helper_usage": {
            "title": _l("Checker helper usage"),
            "short": _l("Checker helper")
        },
        ".software.applications.check_mk.sites:*.livestatus_usage": {
            "title": _l("Live helper usage"),
            "short": _l("Live helper")
        },
        ".software.applications.check_mk.sites:*.check_helper_usage": {
            "title": _l("Actual helper usage"),
            "short": _l("Act. helper")
        },
        ".software.applications.check_mk.sites:*.autostart": {
            "title": _l("Autostart"),
            "paint": "bool",
        },
        ".software.applications.check_mk.sites:*.apache": {
            "title": _l("Apache status"),
            "short": _l("Apache"),
            "paint": "service_status",
        },
        ".software.applications.check_mk.sites:*.cmc": {
            "title": _l("CMC status"),
            "short": _l("CMC"),
            "paint": "service_status",
        },
        ".software.applications.check_mk.sites:*.crontab": {
            "title": _l("Crontab status"),
            "short": _l("Crontab"),
            "paint": "service_status",
        },
        ".software.applications.check_mk.sites:*.dcd": {
            "title": _l("DCD status"),
            "short": _l("DCD"),
            "paint": "service_status",
        },
        ".software.applications.check_mk.sites:*.liveproxyd": {
            "title": _l("Liveproxyd status"),
            "short": _l("Liveproxyd"),
            "paint": "service_status",
        },
        ".software.applications.check_mk.sites:*.mkeventd": {
            "title": _l("MKEventd status"),
            "short": _l("MKEventd"),
            "paint": "service_status",
        },
        ".software.applications.check_mk.sites:*.mknotifyd": {
            "title": _l("MKNotifyd status"),
            "short": _l("MKNotifyd"),
            "paint": "service_status",
        },
        ".software.applications.check_mk.sites:*.rrdcached": {
            "title": _l("RRDCached status"),
            "short": _l("RRDCached"),
            "paint": "service_status",
        },
        ".software.applications.check_mk.sites:*.stunnel": {
            "title": _l("STunnel status"),
            "short": _l("STunnel"),
            "paint": "service_status",
        },
        ".software.applications.check_mk.sites:*.xinetd": {
            "title": _l("XInetd status"),
            "short": _l("XInetd"),
            "paint": "service_status",
        },
        ".software.applications.check_mk.sites:*.nagios": {
            "title": _l("Nagios status"),
            "short": _l("Nagios"),
            "paint": "service_status",
        },
        ".software.applications.check_mk.sites:*.npcd": {
            "title": _l("NPCD status"),
            "short": _l("NPCD"),
            "paint": "service_status",
        },
        ".software.applications.check_mk.cluster.": {
            "title": _l("Cluster"),
            "keyorder": ["is_cluster"],
        },
        ".software.applications.check_mk.cluster.is_cluster": {
            "title": _l("Cluster host"),
            "paint": "bool",
        },
        ".software.applications.check_mk.cluster.nodes:": {
            "title": _l("Nodes"),
            "keyorder": ["name"],
        },
        ".software.applications.check_mk.cluster.nodes:*.name": {"title": _l("Node name")},
        ".software.applications.check_mk.host_labels:": {
            "title": _l("Discovered host labels"),
            "keyorder": [
                "label",
                "plugin_name",
            ],
        },
        ".software.applications.check_mk.host_labels:*.label": {
            "title": _l("Label"),
            "paint": "cmk_label",
        },
        ".software.applications.check_mk.host_labels:*.plugin_name": {
            "title": _l("Discovered by plug-in"),
        },
        ".software.applications.checkmk-agent.": {
            "title": _l("Checkmk Agent"),
            "keyorder": [
                "version",
                "agentdirectory",
                "datadirectory",
                "spooldirectory",
                "pluginsdirectory",
                "localdirectory",
                "agentcontroller",
            ],
        },
        ".software.applications.checkmk-agent.version": {"title": _l("Version")},
        ".software.applications.checkmk-agent.agentdirectory": {"title": _l("Agent directory")},
        ".software.applications.checkmk-agent.datadirectory": {"title": _l("Data directory")},
        ".software.applications.checkmk-agent.spooldirectory": {"title": _l("Spool directory")},
        ".software.applications.checkmk-agent.pluginsdirectory": {"title": _l("Plug-ins directory")},
        ".software.applications.checkmk-agent.localdirectory": {"title": _l("Local directory")},
        ".software.applications.checkmk-agent.agentcontroller": {"title": _l("Agent controller")},
        ".software.applications.checkmk-agent.plugins:": {
            "title": _l("Agent plug-ins"),
            "keyorder": ["name", "version", "cache_interval"],
        },
        ".software.applications.checkmk-agent.plugins:*.name": {"title": _l("Name")},
        ".software.applications.checkmk-agent.plugins:*.version": {"title": _l("Version")},
        ".software.applications.checkmk-agent.plugins:*.cache_interval": {
            "title": _l("Cache interval")
        },
        ".software.applications.checkmk-agent.local_checks:": {
            "title": _l("Local checks"),
            "keyorder": ["name", "version", "cache_interval"],
        },
        ".software.applications.checkmk-agent.local_checks:*.name": {"title": _l("Name")},
        ".software.applications.checkmk-agent.local_checks:*.version": {"title": _l("Version")},
        ".software.applications.checkmk-agent.local_checks:*.cache_interval": {
            "title": _l("Cache interval")
        },
        ".software.applications.docker.": {
            "icon": "docker",
            "title": "Docker",
            "keyorder": [
                "version",
                "registry",
                "swarm_state",
                "swarm_node_id",
                "num_containers_total",
                "num_containers_running",
                "num_containers_stopped",
                "num_containers_paused",
                "num_images",
            ],
        },
        ".software.applications.docker.version": {"title": _l("Version")},
        ".software.applications.docker.registry": {"title": _l("Registry")},
        ".software.applications.docker.swarm_state": {"title": _l("Swarm state")},
        ".software.applications.docker.swarm_node_id": {"title": _l("Swarm node ID")},
        ".software.applications.docker.num_containers_total": {"title": _l("#Containers")},
        ".software.applications.docker.num_containers_running": {
            "title": _l("#Containers running")
        },
        ".software.applications.docker.num_containers_stopped": {
            "title": _l("#Containers stopped")
        },
        ".software.applications.docker.num_containers_paused": {"title": _l("#Containers paused")},
        ".software.applications.docker.num_images": {"title": _l("#Images")},
        ".software.applications.docker.node_labels:": {
            "title": _l("Node labels"),
            "keyorder": ["label"],
        },
        ".software.applications.docker.node_labels:*.label": {"title": _l("Label")},
        ".software.applications.docker.swarm_manager:": {
            "title": _l("Swarm managers"),
            "keyorder": ["NodeID", "Addr"],
        },
        ".software.applications.docker.swarm_manager:*.NodeID": {"title": _l("Node ID")},
        ".software.applications.docker.swarm_manager:*.Addr": {"title": _l("Address")},
        ".software.applications.docker.images:": {
            "title": _l("Docker images"),
            "keyorder": [
                "id",
                "creation",
                "size",
                "labels",
                "amount_containers",
                "repotags",
                "repodigests",
            ],
            "view": "invdockerimages",
            "is_show_more": False,
        },
        ".software.applications.docker.images:*.id": {"title": _l("ID")},
        ".software.applications.docker.images:*.creation": {"title": _l("Creation")},
        ".software.applications.docker.images:*.size": {"title": _l("Size"), "paint": "size"},
        ".software.applications.docker.images:*.labels": {
            "title": _l("Labels"),
            "paint": "csv_labels",
        },
        ".software.applications.docker.images:*.amount_containers": {"title": _l("# Containers")},
        ".software.applications.docker.images:*.repotags": {
            "title": _l("Repository/Tag"),
            "paint": "csv_labels",
        },
        ".software.applications.docker.images:*.repodigests": {
            "title": _l("Digests"),
            "paint": "csv_labels",
        },
        # Node containers
        ".software.applications.docker.containers:": {
            "title": _l("Docker containers"),
            "keyorder": ["id", "creation", "name", "labels", "status", "image"],
            "view": "invdockercontainers",
            "is_show_more": False,
        },
        ".software.applications.docker.containers:*.id": {"title": _l("ID")},
        ".software.applications.docker.containers:*.creation": {"title": _l("Creation")},
        ".software.applications.docker.containers:*.name": {"title": _l("Name")},
        ".software.applications.docker.containers:*.labels": {
            "title": _l("Labels"),
            "paint": "csv_labels",
        },
        ".software.applications.docker.containers:*.status": {"title": _l("Status")},
        ".software.applications.docker.containers:*.image": {"title": _l("Image")},
        ".software.applications.docker.networks.": {"title": "Networks"},
        ".software.applications.docker.networks.*.": {
            "title": "Network %s",
            "keyorder": [
                "network_id",
                "name",
                "scope",
                "labels",
            ],
        },
        ".software.applications.docker.networks.*.network_id": {"title": "Network ID"},
        ".software.applications.docker.networks.*.name": {"title": "Name"},
        ".software.applications.docker.networks.*.scope": {"title": "Scope"},
        ".software.applications.docker.networks.*.labels": {
            "title": "Labels",
            "paint": "csv_labels",
        },
        ".software.applications.docker.container.": {
            "title": _l("Container"),
            "keyorder": ["node_name"],
        },
        ".software.applications.docker.container.node_name": {"title": _l("Node name")},
        ".software.applications.docker.container.ports:": {
            "title": _l("Ports"),
            "keyorder": ["port", "protocol", "host_addresses"],
        },
        ".software.applications.docker.container.ports:*.port": {"title": _l("Port")},
        ".software.applications.docker.container.ports:*.protocol": {"title": _l("Protocol")},
        ".software.applications.docker.container.ports:*.host_addresses": {
            "title": _l("Host addresses")
        },
        ".software.applications.docker.container.networks:": {
            "title": _l("Networks"),
            "keyorder": [
                "name",
                "ip_address",
                "ip_prefixlen",
                "gateway",
                "mac_address",
                "network_id",
            ],
        },
        ".software.applications.docker.container.networks:*.name": {"title": _l("Name")},
        ".software.applications.docker.container.networks:*.ip_address": {
            "title": _l("IP address")
        },
        ".software.applications.docker.container.networks:*.ip_prefixlen": {
            "title": _l("IP prefix")
        },
        ".software.applications.docker.container.networks:*.gateway": {"title": _l("Gateway")},
        ".software.applications.docker.container.networks:*.mac_address": {
            "title": _l("MAC address")
        },
        ".software.applications.docker.container.networks:*.network_id": {
            "title": _l("Network ID")
        },
        ".software.applications.docker.networks.*.containers:": {
            "title": _l("Containers"),
            "keyorder": ["name", "id", "ipv4_address", "ipv6_address", "mac_address"],
        },
        ".software.applications.docker.networks.*.containers:*.name": {"title": _l("Name")},
        ".software.applications.docker.networks.*.containers:*.id": {"title": _l("ID")},
        ".software.applications.docker.networks.*.containers:*.ipv4_address": {
            "title": _l("IPv4 address"),
        },
        ".software.applications.docker.networks.*.containers:*.ipv6_address": {
            "title": _l("IPv6 address"),
        },
        ".software.applications.docker.networks.*.containers:*.mac_address": {
            "title": _l("MAC address"),
        },
        ".software.applications.fortinet.": {"title": _l("Fortinet")},
        ".software.applications.fortinet.fortisandbox:": {
            "title": _l("FortiSandbox software"),
            "keyorder": ["name", "version"],
        },
        ".software.applications.fortinet.fortisandbox:*.name": {"title": _l("Name")},
        ".software.applications.fortinet.fortisandbox:*.version": {"title": _l("Version")},
        ".software.applications.fortinet.fortigate_high_availability.": {
            "title": _l("FortiGate HighAvailability"),
        },
        ".software.applications.fritz.": {
            "title": _l("Fritz"),
            "keyorder": [
                "link_type",
                "wan_access_type",
                "auto_disconnect_time",
                "dns_server_1",
                "dns_server_2",
                "voip_dns_server_1",
                "voip_dns_server_2",
                "upnp_config_enabled",
            ],
        },
        ".software.applications.fritz.link_type": {"title": _l("Link type")},
        ".software.applications.fritz.wan_access_type": {"title": _l("WAN access type")},
        ".software.applications.fritz.auto_disconnect_time": {"title": _l("Auto-disconnect time")},
        ".software.applications.fritz.dns_server_1": {"title": _l("DNS server 1")},
        ".software.applications.fritz.dns_server_2": {"title": _l("DNS server 2")},
        ".software.applications.fritz.voip_dns_server_1": {"title": _l("VoIP DNS server 1")},
        ".software.applications.fritz.voip_dns_server_2": {"title": _l("VoIP DNS server 2")},
        ".software.applications.fritz.upnp_config_enabled": {
            "title": _l("uPnP configuration enabled")
        },
        ".software.applications.kube.": {"title": _l("Kubernetes")},
        ".software.applications.kube.labels:": {
            "title": _l("Labels"),
            "keyorder": ["label_name", "label_value"],
        },
        ".software.applications.kube.labels:*.label_name": {
            "title": _l("Name"),
        },
        ".software.applications.kube.labels:*.label_value": {
            "title": _l("Value"),
        },
        ".networking.kube:": {
            "title": _l("Kubernetes"),
            "keyorder": ["ip", "address_type"],
        },
        ".networking.kube:*.ip": {
            "title": _l("IP address"),
        },
        ".networking.kube:*.address_type": {
            "title": _l("Type"),
        },
        ".software.applications.kube.containers:": {
            "title": _l("Containers"),
            "keyorder": [
                "name",
                "ready",
                "restart_count",
                "image",
                "image_pull_policy",
                "image_id",
                "container_id",
            ],
        },
        ".software.applications.kube.metadata.": {
            "title": _l("Metadata"),
            "keyorder": [
                "object",
                "name",
                "namespace",
            ],
        },
        ".software.applications.kube.metadata.object": {"title": _l("Object")},
        ".software.applications.kube.metadata.name": {"title": _l("Name")},
        ".software.applications.kube.metadata.namespace": {"title": _l("Namespace")},
        ".software.applications.kube.containers:*.name": {"title": _l("Name")},
        ".software.applications.kube.containers:*.image": {"title": _l("Image")},
        ".software.applications.kube.containers:*.image_pull_policy": {
            "title": _l("Image pull policy")
        },
        ".software.applications.kube.containers:*.image_id": {"title": _l("Image ID")},
        ".software.applications.kube.containers:*.ready": {
            "title": _l("Ready"),
            "paint": "container_ready",
        },
        ".software.applications.kube.containers:*.restart_count": {"title": _l("Restart count")},
        ".software.applications.kube.containers:*.container_id": {"title": _l("Container ID")},
        ".software.applications.kube.cluster.": {"title": _l("Cluster"), "keyorder": ["version"]},
        ".software.applications.kube.cluster.version": {"title": _l("Version")},
        ".software.applications.kube.deployment.": {
            "title": _l("Deployment"),
            "keyorder": [
                "strategy",
                "match_labels",
                "match_expressions",
            ],
        },
        ".software.applications.kube.deployment.strategy": {
            "title": _l("StrategyType"),
            "keyorder": [
                "strategy",
                "match_labels",
                "match_expressions",
            ],
        },
        ".software.applications.kube.deployment.match_labels": {"title": _l("matchLabels")},
        ".software.applications.kube.deployment.match_expressions": {
            "title": _l("matchExpressions")
        },
        ".software.applications.kube.daemonset.": {
            "title": _l("DaemonSet"),
            "keyorder": [
                "strategy",
                "match_labels",
                "match_expressions",
            ],
        },
        ".software.applications.kube.daemonset.strategy": {"title": _l("StrategyType")},
        ".software.applications.kube.daemonset.match_labels": {"title": _l("matchLabels")},
        ".software.applications.kube.daemonset.match_expressions": {
            "title": _l("matchExpressions")
        },
        ".software.applications.kube.statefulset.": {
            "title": _l("StatefulSet"),
            "keyorder": [
                "strategy",
                "match_labels",
                "match_expressions",
            ],
        },
        ".software.applications.kube.statefulset.strategy": {"title": _l("StrategyType")},
        ".software.applications.kube.statefulset.match_labels": {"title": _l("matchLabels")},
        ".software.applications.kube.statefulset.match_expressions": {
            "title": _l("matchExpressions")
        },
        ".software.applications.kube.node.": {
            "title": _l("Node"),
            "keyorder": [
                "operating_system",
                "os_image",
                "kernel_version",
                "architecture",
                "container_runtime_version",
                "kubelet_version",
                "kube_proxy_version",
            ],
        },
        ".software.applications.kube.node.operating_system": {"title": _l("Operating system")},
        ".software.applications.kube.node.os_image": {"title": _l("OS image")},
        ".software.applications.kube.node.kernel_version": {"title": _l("Kernel version")},
        ".software.applications.kube.node.architecture": {"title": _l("Architecture")},
        ".software.applications.kube.node.container_runtime_version": {
            "title": _l("Container runtime version")
        },
        ".software.applications.kube.node.kubelet_version": {"title": _l("Kubelet version")},
        ".software.applications.kube.node.kube_proxy_version": {"title": _l("Kube-proxy version")},
        ".software.applications.kube.pod.": {
            "title": _l("Pod"),
            "keyorder": [
                "dns_policy",
                "host_ip",
                "host_network",
                "node",
                "pod_ip",
                "qos_class",
            ],
        },
        ".software.applications.kube.pod.dns_policy": {"title": _l("DNS policy")},
        ".software.applications.kube.pod.host_ip": {"title": _l("Host IP")},
        ".software.applications.kube.pod.host_network": {"title": _l("Host network")},
        ".software.applications.kube.pod.node": {"title": _l("Node")},
        ".software.applications.kube.pod.pod_ip": {"title": _l("Pod IP")},
        ".software.applications.kube.pod.qos_class": {"title": _l("QoS class")},
        ".software.applications.synthetic_monitoring.": {"title": _l("Synthetic monitoring")},
        ".software.applications.synthetic_monitoring.plans:": {
            "title": _l("Plans"),
            "keyorder": [
                "application",
                "suite_name",
                "variant",
                "plan_id",
            ],
            "view": "invsyntheticmonitoringplans",
        },
        ".software.applications.synthetic_monitoring.plans:*.application": {"title": _l("Application")},
        ".software.applications.synthetic_monitoring.plans:*.suite_name": {"title": _l("Suite name")},
        ".software.applications.synthetic_monitoring.plans:*.variant": {"title": _l("Variant")},
        ".software.applications.synthetic_monitoring.plans:*.plan_id": {"title": _l("Plan ID")},
        ".software.applications.synthetic_monitoring.tests:": {
            "title": _l("Tests"),
            "keyorder": [
                "application",
                "suite_name",
                "variant",
                "top_level_suite_name",
                "bottom_level_suite_name",
                "test_name",
                "plan_id",
                "test_item",
            ],
            "view": "invsyntheticmonitoringtests",
        },
        ".software.applications.synthetic_monitoring.tests:*.application": {"title": _l("Application")},
        ".software.applications.synthetic_monitoring.tests:*.suite_name": {"title": _l("Suite name")},
        ".software.applications.synthetic_monitoring.tests:*.variant": {"title": _l("Variant")},
        ".software.applications.synthetic_monitoring.tests:*.top_level_suite_name": {"title": _l("Top level suite")},
        ".software.applications.synthetic_monitoring.tests:*.bottom_level_suite_name": {"title": _l("Bottom level suite")},
        ".software.applications.synthetic_monitoring.tests:*.test_name": {"title": _l("Test")},
        ".software.applications.synthetic_monitoring.tests:*.plan_id": {"title": _l("Plan ID")},
        ".software.applications.synthetic_monitoring.tests:*.test_item": {"title": _l("Item")},
        ".software.applications.mobileiron.": {
            "title": _l("Mobileiron"),
            "keyorder": [
                "partition_name",
                "registration_state",
            ],
        },
        ".software.applications.mobileiron.partition_name": {"title": _l("Partition name")},
        ".software.applications.mobileiron.registration_state": {"title": _l("Registration state")},
        ".software.applications.citrix.": {"title": _l("Citrix")},
        ".software.applications.citrix.controller.": {
            "title": _l("Controller"),
            "keyorder": ["controller_version"],
        },
        ".software.applications.citrix.controller.controller_version": {
            "title": _l("Controller version"),
        },
        ".software.applications.citrix.vm.": {
            "title": _l("Virtual machine"),
            "keyorder": [
                "desktop_group_name",
                "catalog",
                "agent_version",
            ],
        },
        ".software.applications.citrix.vm.desktop_group_name": {"title": _l("Desktop group name")},
        ".software.applications.citrix.vm.catalog": {"title": _l("Catalog")},
        ".software.applications.citrix.vm.agent_version": {"title": _l("Agent version")},
        ".software.applications.oracle.": {"title": _l("Oracle DB")},
        ".software.applications.oracle.systemparameter:": {
            "title": _l("Oracle system parameters"),
            "keyorder": [
                "sid",
                "name",
                "value",
                "isdefault",
            ],
            "view": "invorasystemparameter",
        },
        ".software.applications.oracle.systemparameter:*.sid": {"title": _l("SID")},
        ".software.applications.oracle.systemparameter:*.name": {"title": _l("Name")},
        ".software.applications.oracle.systemparameter:*.value": {"title": _l("Value")},
        ".software.applications.oracle.systemparameter:*.isdefault": {"title": _l("Is default")},
        ".software.applications.oracle.instance:": {
            "title": _l("Oracle instances"),
            "keyorder": [
                "sid",
                "pname",
                "version",
                "openmode",
                "logmode",
                "logins",
                "db_uptime",
                "db_creation_time",
            ],
            "view": "invorainstance",
        },
        ".software.applications.oracle.instance:*.sid": {"title": _l("SID")},
        ".software.applications.oracle.instance:*.pname": {"title": _l("Process name")},
        ".software.applications.oracle.instance:*.version": {"title": _l("Version")},
        ".software.applications.oracle.instance:*.openmode": {"title": _l("Open mode")},
        ".software.applications.oracle.instance:*.logmode": {"title": _l("Log mode")},
        ".software.applications.oracle.instance:*.logins": {"title": _l("Logins")},
        ".software.applications.oracle.instance:*.db_uptime": {
            "title": _l("Uptime"),
            "paint": "age",
        },
        ".software.applications.oracle.instance:*.db_creation_time": {
            "title": _l("Creation time"),
        },
        ".software.applications.oracle.dataguard_stats:": {
            "title": _l("Oracle dataguard statistics"),
            "keyorder": ["sid", "db_unique", "role", "switchover"],
            "view": "invoradataguardstats",
        },
        ".software.applications.oracle.dataguard_stats:*.sid": {"title": _l("SID")},
        ".software.applications.oracle.dataguard_stats:*.db_unique": {"title": _l("Name")},
        ".software.applications.oracle.dataguard_stats:*.role": {"title": _l("Role")},
        ".software.applications.oracle.dataguard_stats:*.switchover": {"title": _l("Switchover")},
        ".software.applications.oracle.recovery_area:": {
            "title": _l("Oracle recovery areas"),
            "keyorder": ["sid", "flashback"],
            "view": "invorarecoveryarea",
        },
        ".software.applications.oracle.recovery_area:*.sid": {"title": _l("SID")},
        ".software.applications.oracle.recovery_area:*.flashback": {"title": _l("Flashback")},
        ".software.applications.oracle.sga:": {
            "title": _l("Oracle SGA info"),
            "keyorder": [
                "sid",
                "fixed_size",
                "redo_buffer",
                "buf_cache_size",
                "in_mem_area_size",
                "shared_pool_size",
                "large_pool_size",
                "java_pool_size",
                "streams_pool_size",
                "shared_io_pool_size",
                "data_trans_cache_size",
                "granule_size",
                "max_size",
                "start_oh_shared_pool",
                "free_mem_avail",
            ],
            "view": "invorasga",
        },
        ".software.applications.oracle.sga:*.sid": {"title": _l("SID")},
        ".software.applications.oracle.sga:*.fixed_size": {
            "title": _l("Fixed size"),
            "paint": "size",
        },
        ".software.applications.oracle.sga:*.max_size": {
            "title": _l("Maximum size"),
            "paint": "size",
        },
        ".software.applications.oracle.sga:*.redo_buffer": {
            "title": _l("Redo buffers"),
            "paint": "size",
        },
        ".software.applications.oracle.sga:*.buf_cache_size": {
            "title": _l("Buffer cache size"),
            "paint": "size",
        },
        ".software.applications.oracle.sga:*.in_mem_area_size": {
            "title": _l("In-memory area"),
            "paint": "size",
        },
        ".software.applications.oracle.sga:*.shared_pool_size": {
            "title": _l("Shared pool size"),
            "paint": "size",
        },
        ".software.applications.oracle.sga:*.large_pool_size": {
            "title": _l("Large pool size"),
            "paint": "size",
        },
        ".software.applications.oracle.sga:*.java_pool_size": {
            "title": _l("Java pool size"),
            "paint": "size",
        },
        ".software.applications.oracle.sga:*.streams_pool_size": {
            "title": _l("Streams pool size"),
            "paint": "size",
        },
        ".software.applications.oracle.sga:*.shared_io_pool_size": {
            "title": _l("Shared pool size"),
            "paint": "size",
        },
        ".software.applications.oracle.sga:*.data_trans_cache_size": {
            "title": _l("Data transfer cache size"),
            "paint": "size",
        },
        ".software.applications.oracle.sga:*.granule_size": {
            "title": _l("Granule size"),
            "paint": "size",
        },
        ".software.applications.oracle.sga:*.start_oh_shared_pool": {
            "title": _l("Startup overhead in shared pool"),
            "paint": "size",
        },
        ".software.applications.oracle.sga:*.free_mem_avail": {
            "title": _l("Free SGA memory available"),
            "paint": "size",
        },
        ".software.applications.oracle.pga:": {
            "title": _l("Oracle PGA info"),
            "keyorder": [
                "sid",
                "aggregate_pga_auto_target",
                "aggregate_pga_target_parameter",
                "bytes_processed",
                "extra_bytes_read_written",
                "global_memory_bound",
                "maximum_pga_allocated",
                "maximum_pga_used_for_auto_workareas",
                "maximum_pga_used_for_manual_workareas",
                "total_pga_allocated",
                "total_pga_inuse",
                "total_pga_used_for_auto_workareas",
                "total_pga_used_for_manual_workareas",
                "total_freeable_pga_memory",
            ],
            "view": "invorapga",
        },
        ".software.applications.oracle.pga:*.sid": {"title": _l("SID")},
        ".software.applications.oracle.pga:*.aggregate_pga_auto_target": {
            "title": _l("Aggregate PGA auto target"),
            "paint": "size",
        },
        ".software.applications.oracle.pga:*.aggregate_pga_target_parameter": {
            "title": _l("Aggregate PGA target parameter"),
            "paint": "size",
        },
        ".software.applications.oracle.pga:*.bytes_processed": {
            "title": _l("Bytes processed"),
            "paint": "size",
        },
        ".software.applications.oracle.pga:*.extra_bytes_read_written": {
            "title": _l("Extra bytes read/written"),
            "paint": "size",
        },
        ".software.applications.oracle.pga:*.global_memory_bound": {
            "title": _l("Global memory bound"),
            "paint": "size",
        },
        ".software.applications.oracle.pga:*.maximum_pga_allocated": {
            "title": _l("Maximum PGA allocated"),
            "paint": "size",
        },
        ".software.applications.oracle.pga:*.maximum_pga_used_for_auto_workareas": {
            "title": _l("Maximum PGA used for auto workareas"),
            "paint": "size",
        },
        ".software.applications.oracle.pga:*.maximum_pga_used_for_manual_workareas": {
            "title": _l("Maximum PGA used for manual workareas"),
            "paint": "size",
        },
        ".software.applications.oracle.pga:*.total_pga_allocated": {
            "title": _l("Total PGA allocated"),
            "paint": "size",
        },
        ".software.applications.oracle.pga:*.total_pga_inuse": {
            "title": _l("Total PGA inuse"),
            "paint": "size",
        },
        ".software.applications.oracle.pga:*.total_pga_used_for_auto_workareas": {
            "title": _l("Total PGA used for auto workareas"),
            "paint": "size",
        },
        ".software.applications.oracle.pga:*.total_pga_used_for_manual_workareas": {
            "title": _l("Total PGA used for manual workareas"),
            "paint": "size",
        },
        ".software.applications.oracle.pga:*.total_freeable_pga_memory": {
            "title": _l("Total freeable PGA memory"),
            "paint": "size",
        },
        ".software.applications.oracle.tablespaces:": {
            "title": _l("Oracle tablespaces"),
            "keyorder": [
                "sid",
                "name",
                "version",
                "type",
                "autoextensible",
                "current_size",
                "max_size",
                "used_size",
                "num_increments",
                "increment_size",
                "free_space",
            ],
            "view": "invoratablespace",
        },
        ".software.applications.oracle.tablespaces:*.sid": {"title": _l("SID")},
        ".software.applications.oracle.tablespaces:*.name": {"title": _l("Name")},
        ".software.applications.oracle.tablespaces:*.version": {"title": _l("Version")},
        ".software.applications.oracle.tablespaces:*.type": {"title": _l("Type")},
        ".software.applications.oracle.tablespaces:*.autoextensible": {
            "title": _l("Autoextensible")
        },
        ".software.applications.oracle.tablespaces:*.current_size": {
            "title": _l("Current size"),
            "paint": "size",
        },
        ".software.applications.oracle.tablespaces:*.max_size": {
            "title": _l("Max. size"),
            "paint": "size",
        },
        ".software.applications.oracle.tablespaces:*.used_size": {
            "title": _l("Used size"),
            "paint": "size",
        },
        ".software.applications.oracle.tablespaces:*.num_increments": {
            "title": _l("#Increments"),
        },
        ".software.applications.oracle.tablespaces:*.increment_size": {
            "title": _l("Increment size"),
            "paint": "size",
        },
        ".software.applications.oracle.tablespaces:*.free_space": {
            "title": _l("Free space"),
            "paint": "size",
        },
        ".software.applications.vmwareesx.": {"title": _l("VMware ESX")},
        ".software.applications.vmwareesx:*.": {"title": _l("Datacenter %d")},
        ".software.applications.vmwareesx:*.clusters.": {"title": _l("Clusters")},
        ".software.applications.vmwareesx:*.clusters:*.": {"title": _l("Cluster %d")},
        ".software.applications.mssql.": {"title": _l("MSSQL")},
        ".software.applications.mssql.instances:": {
            "title": _l("Instances"),
            "keyorder": [
                "name",
                "product",
                "edition",
                "version",
                "clustered",
                "cluster_name",
                "active_node",
                "node_names",
            ],
        },
        ".software.applications.mssql.instances:*.name": {"title": _l("Name")},
        ".software.applications.mssql.instances:*.product": {"title": _l("Product")},
        ".software.applications.mssql.instances:*.edition": {"title": _l("Edition")},
        ".software.applications.mssql.instances:*.version": {"title": _l("Version")},
        ".software.applications.mssql.instances:*.clustered": {
            "title": _l("Clustered"),
            "paint": "mssql_is_clustered",
        },
        ".software.applications.mssql.instances:*.cluster_name": {"title": _l("Cluster name")},
        ".software.applications.mssql.instances:*.active_node": {"title": _l("Active node")},
        ".software.applications.mssql.instances:*.node_names": {"title": _l("Node names")},
        ".software.applications.ibm_mq.": {
            "title": _l("IBM MQ"),
            "keyorder": ["managers", "channels", "queues"],
        },
        ".software.applications.ibm_mq.managers": {"title": _l("Managers")},
        ".software.applications.ibm_mq.channels": {"title": _l("Channels")},
        ".software.applications.ibm_mq.queues": {"title": _l("Queues")},
        ".software.applications.ibm_mq.managers:": {
            "title": _l("IBM MQ managers"),
            "keyorder": ["name", "instver", "instname", "status", "standby", "ha"],
            "view": "invibmmqmanagers",
        },
        ".software.applications.ibm_mq.managers:*.name": {"title": _l("Queue manager name")},
        ".software.applications.ibm_mq.managers:*.instver": {"title": _l("Version")},
        ".software.applications.ibm_mq.managers:*.instname": {"title": _l("Installation")},
        ".software.applications.ibm_mq.managers:*.status": {"title": _l("Status")},
        ".software.applications.ibm_mq.managers:*.standby": {"title": _l("Standby")},
        ".software.applications.ibm_mq.managers:*.ha": {"title": _l("HA")},
        ".software.applications.ibm_mq.channels:": {
            "title": _l("IBM MQ channels"),
            "keyorder": ["qmgr", "name", "type", "status", "monchl"],
            "view": "invibmmqchannels",
        },
        ".software.applications.ibm_mq.channels:*.qmgr": {"title": _l("Queue manager name")},
        ".software.applications.ibm_mq.channels:*.name": {"title": _l("Channel")},
        ".software.applications.ibm_mq.channels:*.type": {"title": _l("Type")},
        ".software.applications.ibm_mq.channels:*.status": {"title": _l("Status")},
        ".software.applications.ibm_mq.channels:*.monchl": {"title": _l("Monitoring")},
        ".software.applications.ibm_mq.queues:": {
            "title": _l("IBM MQ queues"),
            "keyorder": ["qmgr", "name", "maxdepth", "maxmsgl", "created", "altered", "monq"],
            "view": "invibmmqqueues",
        },
        ".software.applications.ibm_mq.queues:*.qmgr": {"title": _l("Queue manager name")},
        ".software.applications.ibm_mq.queues:*.name": {"title": _l("Queue")},
        ".software.applications.ibm_mq.queues:*.maxdepth": {"title": _l("Max depth")},
        ".software.applications.ibm_mq.queues:*.maxmsgl": {"title": _l("Max length")},
        ".software.applications.ibm_mq.queues:*.created": {"title": _l("Created")},
        ".software.applications.ibm_mq.queues:*.altered": {"title": _l("Altered")},
        ".software.applications.ibm_mq.queues:*.monq": {"title": _l("Monitoring")},
        ".software.applications.azure.": {"title": _l("Azure")},
        ".software.applications.azure.application_gateways.": {"title": _l("Application gateways")},
        ".software.applications.azure.application_gateways.rules.": {"title": _l("Rules")},
        ".software.applications.azure.application_gateways.rules.listeners:": {
            "title": _l("Listeners"),
            "keyorder": [
                "application_gateway",
                "rule",
                "listener",
                "protocol",
                "port",
                "host_names",
            ],
        },
        ".software.applications.azure.application_gateways.rules.listeners:*.application_gateway": {"title": _l("Application gateway")},
        ".software.applications.azure.application_gateways.rules.listeners:*.rule": {"title": _l("Rule")},
        ".software.applications.azure.application_gateways.rules.listeners:*.listener": {"title": _l("Listener")},
        ".software.applications.azure.application_gateways.rules.listeners:*.protocol": {"title": _l("Protocol")},
        ".software.applications.azure.application_gateways.rules.listeners:*.port": {"title": _l("Port")},
        ".software.applications.azure.application_gateways.rules.listeners:*.host_names": {"title": _l("Hosts")},
        ".software.applications.azure.application_gateways.rules.listeners.private_ips:": {
            "title": _l("Private IPs"),
            "keyorder": [
                "application_gateway",
                "rule",
                "listener",
                "ip_address",
                "allocation_method"
            ],
        },
        ".software.applications.azure.application_gateways.rules.listeners.private_ips:*.application_gateway": {"title": _l("Application gateway")},
        ".software.applications.azure.application_gateways.rules.listeners.private_ips:*.rule": {"title": _l("Rule")},
        ".software.applications.azure.application_gateways.rules.listeners.private_ips:*.listener": {"title": _l("Listener")},
        ".software.applications.azure.application_gateways.rules.listeners.private_ips:*.ip_address": {"title": _l("IP address")},
        ".software.applications.azure.application_gateways.rules.listeners.private_ips:*.allocation_method": {"title": _l("Allocation method")},
        ".software.applications.azure.application_gateways.rules.listeners.public_ips:": {
            "title": _l("Public IPs"),
            "keyorder": [
                "application_gateway",
                "rule",
                "listener",
                "name",
                "location",
                "ip_address",
                "allocation_method",
                "dns_fqdn"
            ],
        },
        ".software.applications.azure.application_gateways.rules.listeners.public_ips:*.application_gateway": {"title": _l("Application gateway")},
        ".software.applications.azure.application_gateways.rules.listeners.public_ips:*.rule": {"title": _l("Rule")},
        ".software.applications.azure.application_gateways.rules.listeners.public_ips:*.listener": {"title": _l("Listener")},
        ".software.applications.azure.application_gateways.rules.listeners.public_ips:*.name": {"title": _l("Name")},
        ".software.applications.azure.application_gateways.rules.listeners.public_ips:*.location": {"title": _l("Location")},
        ".software.applications.azure.application_gateways.rules.listeners.public_ips:*.ip_address": {"title": _l("IP address")},
        ".software.applications.azure.application_gateways.rules.listeners.public_ips:*.allocation_method": {"title": _l("Allocation method")},
        ".software.applications.azure.application_gateways.rules.listeners.public_ips:*.dns_fqdn": {"title": _l("DNS FQDN")},
        ".software.applications.azure.application_gateways.rules.backends:": {
            "title": _l("Backends"),
            "keyorder": [
                "application_gateway",
                "rule",
                "address_pool_name",
                "protocol",
                "port"
            ],
        },
        ".software.applications.azure.application_gateways.rules.backends:*.application_gateway": {"title": _l("Application gateway")},
        ".software.applications.azure.application_gateways.rules.backends:*.rule": {"title": _l("Rule")},
        ".software.applications.azure.application_gateways.rules.backends:*.address_pool_name": {"title": _l("Address pool name")},
        ".software.applications.azure.application_gateways.rules.backends:*.protocol": {"title": _l("Protocol")},
        ".software.applications.azure.application_gateways.rules.backends:*.port": {"title": _l("Port")},
        ".software.applications.azure.load_balancers.": {"title": _l("Load balancers")},
        ".software.applications.azure.load_balancers.inbound_nat_rules:": {
            "title": _l("Inbound NAT rules"),
            "keyorder": [
                "load_balancer",
                "inbound_nat_rule",
                "frontend_port",
                "backend_port",
            ],
        },
        ".software.applications.azure.load_balancers.inbound_nat_rules:*.load_balancer": {"title": _l("Load balancer")},
        ".software.applications.azure.load_balancers.inbound_nat_rules:*.inbound_nat_rule": {"title": _l("Inbound NAT rule")},
        ".software.applications.azure.load_balancers.inbound_nat_rules:*.frontend_port": {"title": _l("Frontend port")},
        ".software.applications.azure.load_balancers.inbound_nat_rules:*.backend_port": {"title": _l("Backend port")},
        ".software.applications.azure.load_balancers.inbound_nat_rules.public_ips:": {
            "title": _l("Public IPs"),
            "keyorder": [
                "load_balancer",
                "inbound_nat_rule",
                "location",
                "public_ip_name",
                "ip_address",
                "ip_allocation_method",
                "dns_fqdn",
            ],
        },
        ".software.applications.azure.load_balancers.inbound_nat_rules.public_ips:*.load_balancer": {"title": _l("Load balancer")},
        ".software.applications.azure.load_balancers.inbound_nat_rules.public_ips:*.inbound_nat_rule": {"title": _l("Inbound NAT rule")},
        ".software.applications.azure.load_balancers.inbound_nat_rules.public_ips:*.location": {"title": _l("Location")},
        ".software.applications.azure.load_balancers.inbound_nat_rules.public_ips:*.public_ip_name": {"title": _l("Name")},
        ".software.applications.azure.load_balancers.inbound_nat_rules.public_ips:*.ip_address": {"title": _l("IP Address")},
        ".software.applications.azure.load_balancers.inbound_nat_rules.public_ips:*.ip_allocation_method": {"title": _l("Allocation method")},
        ".software.applications.azure.load_balancers.inbound_nat_rules.public_ips:*.dns_fqdn": {"title": _l("DNS FQDN")},
        ".software.applications.azure.load_balancers.inbound_nat_rules.private_ips:": {
            "title": _l("Private IPs"),
            "keyorder": [
                "load_balancer",
                "inbound_nat_rule",
                "ip_address",
                "ip_allocation_method",
            ],
        },
        ".software.applications.azure.load_balancers.inbound_nat_rules.private_ips:*.load_balancer": {"title": _l("Load balancer")},
        ".software.applications.azure.load_balancers.inbound_nat_rules.private_ips:*.inbound_nat_rule": {"title": _l("Inbound NAT rule")},
        ".software.applications.azure.load_balancers.inbound_nat_rules.private_ips:*.ip_address": {"title": _l("IP address")},
        ".software.applications.azure.load_balancers.inbound_nat_rules.private_ips:*.ip_allocation_method": {"title": _l("Allocation method")},

        ".software.applications.azure.load_balancers.inbound_nat_rules.backend_ip_configs:": {
            "title": _l("Public IPs"),
            "keyorder": [
                "load_balancer",
                "inbound_nat_rule",
                "backend_ip_config",
                "ip_address",
                "ip_allocation_method",
            ],
        },
        ".software.applications.azure.load_balancers.inbound_nat_rules.backend_ip_configs:*.load_balancer": {"title": _l("Load balancer")},
        ".software.applications.azure.load_balancers.inbound_nat_rules.backend_ip_configs:*.inbound_nat_rule": {"title": _l("Inbound NAT rule")},
        ".software.applications.azure.load_balancers.inbound_nat_rules.backend_ip_configs:*.backend_ip_config": {"title": _l("Backend IP config")},
        ".software.applications.azure.load_balancers.inbound_nat_rules.backend_ip_configs:*.ip_address": {"title": _l("IP address")},
        ".software.applications.azure.load_balancers.inbound_nat_rules.backend_ip_configs:*.ip_allocation_method": {"title": _l("Allocation method")},
        ".software.applications.azure.load_balancers.outbound_rules:": {
            "title": _l("Outbound rules"),
            "keyorder": [
                "load_balancer",
                "outbound_rule",
                "protocol",
                "idle_timeout",
            ],
        },
        ".software.applications.azure.load_balancers.outbound_rules:*.load_balancer": {"title": _l("Load balancer")},
        ".software.applications.azure.load_balancers.outbound_rules:*.outbound_rule": {"title": _l("Outbound rule")},
        ".software.applications.azure.load_balancers.outbound_rules:*.protocol": {"title": _l("Protocol")},
        ".software.applications.azure.load_balancers.outbound_rules:*.idle_timeout": {"title": _l("Idle timeout")},
        ".software.applications.azure.load_balancers.outbound_rules.backend_pools:": {
            "title": _l("Backend pools"),
            "keyorder": [
                "load_balancer",
                "outbound_rule",
                "backend_pool",
            ],
        },
        ".software.applications.azure.load_balancers.outbound_rules.backend_pools:*.load_balancer": {"title": _l("Load balancer")},
        ".software.applications.azure.load_balancers.outbound_rules.backend_pools:*.outbound_rule": {"title": _l("Outbound rule")},
        ".software.applications.azure.load_balancers.outbound_rules.backend_pools:*.backend_pool": {"title": _l("Backend pool")},
        ".software.applications.azure.load_balancers.outbound_rules.backend_pools.addresses:": {
            "title": _l("Addresses"),
            "keyorder": [
                "load_balancer",
                "outbound_rule",
                "backend_pool",
                "address_name",
                "ip_address",
                "ip_allocation_method",
                "primary"
            ],
        },
        ".software.applications.azure.load_balancers.outbound_rules.backend_pools.addresses:*.load_balancer": {"title": _l("Load balancer")},
        ".software.applications.azure.load_balancers.outbound_rules.backend_pools.addresses:*.outbound_rule": {"title": _l("Outbound rule")},
        ".software.applications.azure.load_balancers.outbound_rules.backend_pools.addresses:*.backend_pool": {"title": _l("Backend pool")},
        ".software.applications.azure.load_balancers.outbound_rules.backend_pools.addresses:*.address_name": {"title": _l("Address name")},
        ".software.applications.azure.load_balancers.outbound_rules.backend_pools.addresses:*.ip_address": {"title": _l("IP address")},
        ".software.applications.azure.load_balancers.outbound_rules.backend_pools.addresses:*.ip_allocation_method": {"title": _l("Allocation method")},
        ".software.applications.azure.load_balancers.outbound_rules.backend_pools.addresses:*.primary": {"title": _l("Primary")},
        ".software.bios.": {
            "title": _l("BIOS"),
            "keyorder": [
                "vendor",
                "version",
                "date",
            ],
        },
        ".software.bios.vendor": {"title": _l("Vendor")},
        ".software.bios.version": {"title": _l("Version")},
        ".software.bios.date": {"title": _l("Date"), "paint": "date"},
        ".software.configuration.": {"title": _l("Configuration")},
        ".software.configuration.snmp_info.": {
            "title": _l("SNMP Information"),
            "keyorder": [
                "contact",
                "location",
                "name",
            ],
        },
        ".software.configuration.snmp_info.contact": {"title": _l("Contact")},
        ".software.configuration.snmp_info.location": {
            "title": _l("Location"),
            "is_show_more": False,
        },
        ".software.configuration.snmp_info.name": {"title": _l("System name")},
        ".software.configuration.organisation.": {
            "title": _l("Organisation"),
            "keyorder": [
                "organisation_id",
                "organisation_name",
                "network_id",
                "address",
            ],
        },
        ".software.configuration.organisation.organisation_id": {"title": _l("Organisation ID")},
        ".software.configuration.organisation.organisation_name": {"title": _l("Organisation name")},
        ".software.configuration.organisation.network_id": {"title": _l("Network ID")},
        ".software.configuration.organisation.address": {"title": _l("Address")},
        ".software.firmware.": {
            "title": _l("Firmware"),
            "keyorder": [
                "vendor",
                "version",
                "platform_level",
            ],
        },
        ".software.firmware.vendor": {"title": _l("Vendor")},
        ".software.firmware.version": {"title": _l("Version")},
        ".software.firmware.platform_level": {"title": _l("Platform firmware level")},
        ".software.kernel_config:": {
            "title": _l("Kernel configuration (sysctl)"),
            "keyorder": ["name", "value"],
            "view": "invkernelconfig",
        },
        ".software.kernel_config:*.name": {"title": _l("Parameter name")},
        ".software.kernel_config:*.value": {"title": _l("Value")},
        ".software.os.": {
            "title": _l("Operating system"),
            "keyorder": [
                "name",
                "version",
                "vendor",
                "type",
                "install_date",
                "kernel_version",
                "arch",
                "service_pack",
                "build",
            ],
        },
        ".software.os.name": {"title": _l("Operating system"), "is_show_more": False},
        ".software.os.version": {"title": _l("Version")},
        ".software.os.vendor": {"title": _l("Vendor")},
        ".software.os.type": {"title": _l("Type"), "is_show_more": False},  # e.g. "linux"
        ".software.os.install_date": {"title": _l("Install date"), "paint": "date"},
        ".software.os.kernel_version": {"title": _l("Kernel version")},
        ".software.os.arch": {"title": _l("Kernel Architecture")},
        ".software.os.service_pack": {"title": _l("Latest service pack")},
        ".software.os.build": {"title": _l("Build")},
        ".software.os.service_packs:": {"title": _l("Service packs"), "keyorder": ["name"]},
        ".software.os.service_packs:*.name": {"title": _l("Name")},
        ".software.packages:": {
            "title": _l("Software packages"),
            "icon": "packages",
            "keyorder": [
                "name",
                "arch",
                "package_type",
                "summary",
                "version",
                "vendor",
                "package_version",
                "install_date",
                "size",
                "path",
            ],
            "view": "invswpac",
            "is_show_more": False,
        },
        ".software.packages:*.name": {"title": _l("Name")},
        ".software.packages:*.arch": {"title": _l("Architecture")},
        ".software.packages:*.package_type": {"title": _l("Type")},
        ".software.packages:*.summary": {"title": _l("Description")},
        ".software.packages:*.version": {
            "title": _l("Version"),
            "sort": cmp_version,
            "filter": FilterInvtableVersion,
        },
        ".software.packages:*.vendor": {"title": _l("Publisher")},
        ".software.packages:*.package_version": {
            "title": _l("Package version"),
            "sort": cmp_version,
            "filter": FilterInvtableVersion,
        },
        ".software.packages:*.install_date": {"title": _l("Install date"), "paint": "date"},
        ".software.packages:*.size": {"title": _l("Size"), "paint": "count"},
        ".software.packages:*.path": {"title": _l("Path")},
        ".networking.": {
            "title": _l("Networking"),
            "icon": "networking",
            "keyorder": [
                "hostname",
                "total_interfaces",
                "total_ethernet_ports",
                "available_ethernet_ports",
            ],
        },
        ".networking.total_interfaces": {"title": _l("Total interfaces"), "paint": "count"},
        ".networking.total_ethernet_ports": {"title": _l("Ports"), "paint": "count"},
        ".networking.hostname": {"title": _l("Host name")},
        ".networking.available_ethernet_ports": {"title": _l("Ports available"), "paint": "count"},
        ".networking.addresses:": {
            "title": _l("IP addresses"),
            "keyorder": ["address", "device", "type"],
        },
        ".networking.addresses:*.address": {"title": _l("Address")},
        ".networking.addresses:*.device": {"title": _l("Device")},
        ".networking.addresses:*.type": {"title": _l("Address type"), "paint": "ip_address_type"},
        ".networking.interfaces:": {
            "title": _l("Network interfaces"),
            "keyorder": [
                "index",
                "description",
                "alias",
                "oper_status",
                "admin_status",
                "available",
                "speed",
                "last_change",
                "port_type",
                "phys_address",
                "vlantype",
                "vlans",
            ],
            "view": "invinterface",
            "is_show_more": False,
        },
        ".networking.interfaces:*.index": {
            "title": _l("Index"),
            "paint": "number",
            "filter": FilterInvtableIDRange,
        },
        ".networking.interfaces:*.description": {"title": _l("Description")},
        ".networking.interfaces:*.alias": {"title": _l("Alias")},
        ".networking.interfaces:*.oper_status": {
            "title": _l("Operational status"),
            "paint": "if_oper_status",
            "filter": FilterInvtableOperStatus,
        },
        ".networking.interfaces:*.admin_status": {
            "title": _l("Administrative status"),
            "paint": "if_admin_status",
            "filter": FilterInvtableAdminStatus,
        },
        ".networking.interfaces:*.available": {
            "title": _l("Port usage"),
            "paint": "if_available",
            "filter": FilterInvtableAvailable,
        },
        ".networking.interfaces:*.speed": {"title": _l("Speed"), "paint": "nic_speed"},
        ".networking.interfaces:*.phys_address": {"title": _l("Physical address (MAC)")},
        ".networking.interfaces:*.port_type": {
            "title": _l("Type"),
            "paint": "if_port_type",
            "filter": FilterInvtableInterfaceType,
        },
        ".networking.interfaces:*.last_change": {
            "title": _l("Last change"),
            "paint": "timestamp_as_age_days",
            "filter": FilterInvtableTimestampAsAge,
        },
        ".networking.interfaces:*.vlans": {"title": _l("VLANs")},
        ".networking.interfaces:*.vlantype": {"title": _l("VLAN type")},
        ".networking.routes:": {
            "title": _l("Routes"),
            "keyorder": ["target", "device", "type", "gateway"],
        },
        ".networking.routes:*.target": {"title": _l("Target"), "paint": "ipv4_network"},
        ".networking.routes:*.device": {"title": _l("Device")},
        ".networking.routes:*.type": {"title": _l("Type of route"), "paint": "route_type"},
        ".networking.routes:*.gateway": {"title": _l("Gateway")},
        ".networking.wlan.": {"title": _l("WLAN")},
        ".networking.wlan.controller.": {"title": _l("Controller")},
        ".networking.wlan.controller.accesspoints:": {
            "title": _l("Access points"),
            "keyorder": ["name", "group", "ip_addr", "model", "serial", "sys_location"],
        },
        ".networking.wlan.controller.accesspoints:*.name": {"title": _l("Name")},
        ".networking.wlan.controller.accesspoints:*.group": {"title": _l("Group")},
        ".networking.wlan.controller.accesspoints:*.ip_addr": {"title": _l("IP address")},
        ".networking.wlan.controller.accesspoints:*.model": {"title": _l("Model")},
        ".networking.wlan.controller.accesspoints:*.serial": {"title": _l("Serial number")},
        ".networking.wlan.controller.accesspoints:*.sys_location": {"title": _l("System location")},
        ".networking.tunnels:": {
            "title": _l("Networking tunnels"),
            "keyorder": [
                "peername",
                "index",
                "peerip",
                "sourceip",
                "tunnelinterface",
                "linkpriority",
            ],
            "view": "invtunnels",
        },
        ".networking.tunnels:*.index": {"title": _l("Index")},
        ".networking.tunnels:*.linkpriority": {"title": _l("Link priority")},
        ".networking.tunnels:*.peerip": {"title": _l("Peer IP address")},
        ".networking.tunnels:*.peername": {"title": _l("Peer name")},
        ".networking.tunnels:*.sourceip": {"title": _l("Source IP address")},
        ".networking.tunnels:*.tunnelinterface": {"title": _l("Tunnel interface")},
    })
# fmt: on
