#!/usr/bin/env python3
# Copyright (C) 2019 Checkmk GmbH - License: GNU General Public License v2
# This file is part of Checkmk (https://checkmk.com). It is subject to the terms and
# conditions defined in the file COPYING, which is part of this source code package.

import sys
from contextlib import contextmanager
from typing import Iterator

import urllib3
from opsgenie_sdk.api.alert import AlertApi  # type: ignore[import-untyped]
from opsgenie_sdk.api.alert.acknowledge_alert_payload import (  # type: ignore[import-untyped]
    AcknowledgeAlertPayload,
)
from opsgenie_sdk.api.alert.close_alert_payload import (  # type: ignore[import-untyped]
    CloseAlertPayload,
)
from opsgenie_sdk.api.alert.create_alert_payload import (  # type: ignore[import-untyped]
    CreateAlertPayload,
)
from opsgenie_sdk.api_client import ApiClient  # type: ignore[import-untyped]
from opsgenie_sdk.configuration import Configuration  # type: ignore[import-untyped]
from opsgenie_sdk.exceptions import (  # type: ignore[import-untyped]
    ApiException,
    AuthenticationException,
)
from requests.utils import get_environ_proxies
from tenacity import RetryError
from urllib3.util import parse_url

from cmk.utils.http_proxy_config import (
    deserialize_http_proxy_config,
    EnvironmentProxyConfig,
    ExplicitProxyConfig,
    NoProxyConfig,
)
from cmk.utils.paths import trusted_ca_file

from cmk.notification_plugins import utils
from cmk.notification_plugins.utils import retrieve_from_passwordstore


@contextmanager
def _handle_api_exceptions(api_call_name: str, ignore_retry: bool = False) -> Iterator[None]:
    try:
        yield
    except (ApiException, AuthenticationException) as err:
        sys.stderr.write(f"Exception when calling AlertApi -> {api_call_name}: {err}\n")
        sys.exit(2)
    except Exception as e:
        if not ignore_retry and isinstance(e, RetryError):
            with _handle_api_exceptions(api_call_name, ignore_retry=True):
                e.reraise()
        sys.stderr.write(f"Unhandled exception when calling AlertApi -> {api_call_name}: {e}\n")
        sys.exit(2)


# https://docs.opsgenie.com/docs/opsgenie-python-api-v2-1
class Connector:
    def __init__(
        self, api_key: str, host_url: str | None, proxy_url: str | None, ignore_ssl: bool
    ) -> None:
        conf: "Configuration" = Configuration()
        conf.api_key["Authorization"] = api_key
        if host_url is not None:
            conf.host = "%s" % host_url
        if proxy_url is not None:
            sys.stdout.write(f"Using proxy: {proxy_url}\n")
            conf.proxy = proxy_url

        if ignore_ssl:
            sys.stdout.write("Ignoring SSL certificate verification\n")
            conf.verify_ssl = False
            urllib3.disable_warnings(urllib3.connectionpool.InsecureRequestWarning)
        else:
            sys.stdout.write(f"Using trust store: {trusted_ca_file}\n")
            conf.ssl_ca_cert = trusted_ca_file

        sys.stdout.flush()

        api_client: "ApiClient" = ApiClient(configuration=conf)
        self.alert_api = AlertApi(api_client=api_client)

    def handle_alert_creation(
        self,
        note_created: str,
        actions_list: list[str],
        desc: str,
        msg: str,
        priority: str,
        teams_list: list[dict[str, str] | None],
        tags_list: list[str],
        entity_value: str,
        alert_source: str | None,
        alias: str | None,
        owner: str | None,
    ) -> int:
        body = CreateAlertPayload(
            note=note_created,
            actions=actions_list,
            description=desc,
            message=msg,
            priority=priority,
            responders=teams_list,
            tags=tags_list,
            entity=entity_value,
            source=alert_source,
            alias=alias,
            user=owner,
        )

        with _handle_api_exceptions("create_alert"):
            response = self.alert_api.create_alert(create_alert_payload=body)

        sys.stdout.write(f"Request id: {response.request_id}, successfully created alert.\n")
        return 0

    def handle_alert_deletion(
        self,
        note_closed: str,
        owner: str | None,
        alias: str | None,
        alert_source: str | None,
    ) -> int:
        body = CloseAlertPayload(
            source=alert_source,
            user=owner,
            note=note_closed,
        )

        with _handle_api_exceptions("close_alert"):
            response = self.alert_api.close_alert(
                identifier=alias,
                identifier_type="alias",
                close_alert_payload=body,
            )

        sys.stdout.write(f"Request id: {response.request_id}, successfully closed alert.\n")
        return 0

    def handle_alert_ack(
        self,
        ack_author: str,
        ack_comment: str,
        alias: str | None,
        alert_source: str | None,
    ) -> int:
        body = AcknowledgeAlertPayload(
            source=alert_source,
            user=ack_author,
            note=ack_comment,
        )

        with _handle_api_exceptions("acknowledge_alert"):
            response = self.alert_api.acknowledge_alert(
                identifier=alias,
                identifier_type="alias",
                acknowledge_alert_payload=body,
            )

        sys.stdout.write(
            f"Request id: {response.request_id}, successfully added acknowledgedment.\n"
        )
        return 0


def _get_proxy_url(proxy_setting: str | None, url: str | None) -> str | None:
    proxy = deserialize_http_proxy_config(proxy_setting)
    if isinstance(proxy, EnvironmentProxyConfig):
        parsed_url = parse_url(url or "https://api.opsgenie.com")
        proxies = get_environ_proxies(parsed_url.url)
        return proxies.get(parsed_url.scheme)

    if isinstance(proxy, NoProxyConfig):
        return None

    if isinstance(proxy, ExplicitProxyConfig):
        return proxy_setting

    sys.stderr.write(f"Unsupported proxy setting: {proxy_setting}\n")
    sys.exit(2)


def main() -> int:
    context = utils.collect_context()

    if "PARAMETER_PASSWORD" not in context:
        sys.stderr.write("API key not set\n")
        return 2

    api_key = retrieve_from_passwordstore(context["PARAMETER_PASSWORD"])
    note_created = context.get("PARAMETER_NOTE_CREATED") or "Alert created by Check_MK"
    note_closed = context.get("PARAMETER_NOTE_CLOSED") or "Alert closed by Check_MK"
    priority = context.get("PARAMETER_PRIORITY", "P3")
    entity_value = context.get("PARAMETER_ENTITY", "")
    alert_source: str | None = context.get("PARAMETER_SOURCE")
    owner: str | None = context.get("PARAMETER_OWNER")
    host_url: str | None = context.get("PARAMETER_URL")
    proxy_url: str | None = _get_proxy_url(context.get("PARAMETER_PROXY_URL"), host_url)

    tags_list: list[str] = []
    if context.get("PARAMETER_TAGSS"):
        tags_list = context.get("PARAMETER_TAGSS", "").split(" ")

    actions_list: list[str] = []
    if context.get("PARAMETER_ACTIONSS"):
        actions_list = context.get("PARAMETER_ACTIONSS", "").split(" ")

    teams_list: list[dict[str, str] | None] = []
    if context.get("PARAMETER_TEAMSS"):
        teams_list = [
            {"name": str(context[k]), "type": "team"}
            for k in context
            if k.startswith("PARAMETER_TEAMS_")
        ]

    if context["WHAT"] == "HOST":
        tmpl_host_msg: str = "Check_MK: $HOSTNAME$ - $HOSTSHORTSTATE$"
        tmpl_host_desc: str = """Host: $HOSTNAME$
Event:    $EVENT_TXT$
Output:   $HOSTOUTPUT$
Perfdata: $HOSTPERFDATA$
$LONGHOSTOUTPUT$
"""
        desc = context.get("PARAMETER_HOST_DESC") or tmpl_host_desc
        msg = context.get("PARAMETER_HOST_MSG") or tmpl_host_msg
        alias = (
            f'HOST_PROBLEM_ID: {context["LASTHOSTPROBLEMID"]}'
            if context["HOSTPROBLEMID"] == "0"
            else f'HOST_PROBLEM_ID: {context["HOSTPROBLEMID"]}'
        )
        ack_author = context["HOSTACKAUTHOR"]
        ack_comment = context["HOSTACKCOMMENT"]
    else:
        tmpl_svc_msg = "Check_MK: $HOSTNAME$/$SERVICEDESC$ $SERVICESHORTSTATE$"
        tmpl_svc_desc = """Host: $HOSTNAME$
Service:  $SERVICEDESC$
Event:    $EVENT_TXT$
Output:   $SERVICEOUTPUT$
Perfdata: $SERVICEPERFDATA$
$LONGSERVICEOUTPUT$
"""
        desc = context.get("PARAMETER_SVC_DESC") or tmpl_svc_desc
        msg = context.get("PARAMETER_SVC_MSG") or tmpl_svc_msg
        alias = (
            f'SVC_PROBLEM_ID: {context["LASTSERVICEPROBLEMID"]}'
            if context["SERVICEPROBLEMID"] == "0"
            else f'SVC_PROBLEM_ID: {context["SERVICEPROBLEMID"]}'
        )
        ack_author = context["SERVICEACKAUTHOR"]
        ack_comment = context["SERVICEACKCOMMENT"]

    desc = utils.substitute_context(desc, context)
    msg = utils.substitute_context(msg, context)

    connector = Connector(
        api_key, host_url, proxy_url, ignore_ssl="PARAMETER_IGNORE_SSL" in context
    )

    if context["NOTIFICATIONTYPE"] == "PROBLEM":
        return connector.handle_alert_creation(
            note_created,
            actions_list,
            desc,
            msg,
            priority,
            teams_list,
            tags_list,
            entity_value,
            alert_source,
            alias,
            owner,
        )
    if context["NOTIFICATIONTYPE"] == "RECOVERY":
        return connector.handle_alert_deletion(note_closed, owner, alias, alert_source)
    if context["NOTIFICATIONTYPE"] == "ACKNOWLEDGEMENT":
        return connector.handle_alert_ack(ack_author, ack_comment, alias, alert_source)

    sys.stdout.write("Notification type %s not supported\n" % (context["NOTIFICATIONTYPE"]))
    return 0
