// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef AbstractHelper_h
#define AbstractHelper_h

#include <functional>
#include <string>

#include "livestatus/Logger.h"
class Object;
class Poller;

class AbstractHelper {
public:
    AbstractHelper(std::string name, ContextLogger logger);
    virtual ~AbstractHelper() = default;
    // Note that move would be OK in principle.
    AbstractHelper(const AbstractHelper &) = delete;
    AbstractHelper &operator=(const AbstractHelper &) = delete;
    AbstractHelper(AbstractHelper &&) = delete;
    AbstractHelper &operator=(AbstractHelper &&) = delete;

    std::string name() const;
    Logger *logger() const;

    void start();
    void stop();
    void restart();
    void setUpPoller(Poller &poller) const;
    void communicate(const Poller &poller);

protected:
    // This is currently a bit high in the class hierarchy, but this is the
    // first place where helpers and pools meet. :-/
    template <typename T>
    using reschedule_handler_t = std::function<void(T *)>;

private:
    std::string name_;
    mutable ContextLogger logger_;

    virtual void start_() = 0;
    virtual void stop_() = 0;
    virtual void setUpPoller_(Poller &poller) const = 0;
    virtual void communicate_(const Poller &poller) = 0;
};

inline AbstractHelper::AbstractHelper(std::string name, ContextLogger logger)
    : name_{std::move(name)}, logger_{std::move(logger)} {};

[[nodiscard]] inline std::string AbstractHelper::name() const { return name_; }
[[nodiscard]] inline Logger *AbstractHelper::logger() const { return &logger_; }
inline void AbstractHelper::start() { this->start_(); }
inline void AbstractHelper::stop() { this->stop_(); }
inline void AbstractHelper::restart() {
    this->stop();
    this->start();
}
inline void AbstractHelper::setUpPoller(Poller &poller) const {
    this->setUpPoller_(poller);
}
inline void AbstractHelper::communicate(const Poller &poller) {
    this->communicate_(poller);
}

#endif
