// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef CarbonConnection_h
#define CarbonConnection_h

#include <chrono>
#include <iosfwd>
#include <ostream>
#include <string>
#include <utility>
#include <vector>

#include "cmc/CarbonConfig.h"
#include "livestatus/Logger.h"

enum class Counter;

namespace carbon::protocol {
struct Update;
}  // namespace carbon::protocol

namespace carbon {
std::vector<std::string> serialize(const carbon::protocol::Update &message);
}  // namespace carbon

/** FSM for the connection
    @startuml
    state CarbonConnection {
    state unconnected
    state connection_in_progress
    state connected

    state try_to_initiate_choice <<choice>>

    [*] --> unconnected
    unconnected --> try_to_initiate_choice : tryToInitiate()
    try_to_initiate_choice --> connected : connect() == 0
    try_to_initiate_choice --> connection_in_progress : ELSE
    connection_in_progress --> connected : tryToComplete()
    CarbonConnection -> unconnected : error OR close()
    connected --> connected : send()
    }
    @enduml
*/
class CarbonConnection {
public:
    enum class State { unconnected, connection_in_progress, connected };

    CarbonConnection(const CarbonConnectionConfig &config,
                     Counter bytes_sent_counter);
    ~CarbonConnection() { close(); }
    // We use dynamically allocated resources (a socket) and our ContextLogger
    // captures 'this' => define move constructor
    CarbonConnection(CarbonConnection &&cc) noexcept
        : config_{cc.config_}
        , socket_{cc.socket_}
        , bytes_sent_counter_{cc.bytes_sent_counter_}
        , state_{cc.state_}
        , last_connection_attempt_{cc.last_connection_attempt_}
        , log_prefix_{std::move(cc.log_prefix_)} {
        cc.socket_ = -1;
        cc.state_ = State::unconnected;
    }

    State tryToInitiateConnection(
        std::chrono::steady_clock::time_point time_now);
    State tryToCompleteConnection();
    void send(const carbon::protocol::Update &message);
    void doHousekeeping() const {}
    [[nodiscard]] Logger *logger() const { return &logger_; }
    bool operator==(const CarbonConnection &other) const;
    bool operator!=(const CarbonConnection &other) const {
        return !(*this == other);
    }

private:
    const CarbonConnectionConfig config_;
    int socket_;
    Counter bytes_sent_counter_;
    State state_;
    std::chrono::steady_clock::time_point last_connection_attempt_;
    std::string log_prefix_;
    mutable ContextLogger logger_{
        Logger::getLogger("cmk.carbon"),
        [this](std::ostream &os) { os << log_prefix_; }};

    void send(const std::vector<std::string> &lines);
    void close();
};

#endif  // CarbonConnection_h
