// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef CheckHelper_h
#define CheckHelper_h

#include <unistd.h>

#include <chrono>
#include <filesystem>
#include <functional>
#include <memory>
#include <optional>
#include <string>
#include <utility>
#include <vector>

#include "cmc/MessagePayload.h"  // IWYU pragma: keep
#include "cmc/ProcessHelper.h"
#include "cmc/SerialToken.h"
#include "livestatus/Logger.h"

class CheckResult;
class FetcherReply;
class Object;
class Poller;
class Service;
class World;

// NOTE: This is temporary class for collecting the common code of CheckHelper
// and FetcherHelper. It will vanish again when FetcherHelper is a subclass of
// CheckHelper or merged with it.
class CommonHelper : public ProcessHelper {
public:
    CommonHelper(const std::filesystem::path &path, const args_type &args,
                 Logger *logger);

    bool runCheck(Object *object);
    void killCheck();
    bool migrate(const World &world);

    enum class State { idle, sending, receiving, dead };

    // State machine accessors
    State state() const { return state_; }
    Object *object() const { return object_; }
    SerialToken &&serialToken() { return std::move(*serial_token_); }
    config_serial_t serial() const { return serial_token_->value(); }

    // State machine modification API: To guarantee some invariants, it is
    // essential to avoid fiddling around with the various parts of our state
    // machine by hand.
    void gotoIdle();
    void gotoSending(Object &object, SerialToken &&serial_token);
    void gotoReceiving();
    void gotoDead();

protected:
    void assertState(State state) const;

private:
    ContextLogger logger_;

    // The local state of our state machine consists of these three parts.
    State state_;
    Object *object_;
    std::optional<SerialToken> serial_token_;

    virtual bool sendReloadConfigRequest() = 0;
    virtual void incrementCounter() const = 0;
    virtual bool sendRequest() = 0;
};

/** State Machine for the checker mode
    @startuml
    hide empty description
    [*] --> Idle
    Idle -[#green]> Sending : Reply is consumed
    Idle : Initial State for the Helper
    Sending : Reply to be send to checker
    Receiving: Waiting for CheckResult

    Sending -[dotted]---> Dead : IO failed
    Sending -[#green]> Receiving : Reply is delivered
    Receiving -[dotted]---> Dead : IO Failed
    Receiving -[#green]> Idle : Valid CheckResult received
    Dead -> [*]
    Dead : No more action
    @enduml
 */
class CheckHelper : public CommonHelper {
public:
    using result_handler_t = std::function<void(
        std::chrono::system_clock::time_point, const CheckResult &)>;

    CheckHelper(const std::filesystem::path &path, const args_type &args,
                result_handler_t result_handler, Logger *logger);
    ~CheckHelper() override;

    void consumeReply(std::unique_ptr<FetcherReply> reply);

    virtual std::string makeRequest() = 0;

protected:
    ssize_t send(const std::string &message);

private:
    result_handler_t _result_handler;

    // Only used while state_ is State::sending
    std::optional<MessagePayloadSender> checker_request_sender_;

    // AbstractHelper API
    void start_() override;
    void stop_() override;
    void setUpPoller_(Poller &poller) const override;
    void communicate_(const Poller &poller) override;

    // CommonHelper API
    bool sendRequest() override;

    void handleStdIn();
    void handleStdOut();
    void handleStdErr();

    void getResult(Object *object);

    std::string _pending_error;  // beginning of stderr line from client

    void do_stop();

    ssize_t readWithTimeout(char *buffer, size_t size);

    void fetchHelperErrors(bool need_linefeed);  // reads and logs helper errors

    [[nodiscard]] bool tryReadStatusPart13(char *buffer);
    [[nodiscard]] bool checkStatusIsGood(char *status_part,
                                         const Object *object);
    [[nodiscard]] bool readDataToVector(std::vector<char> &output);

    bool readData(std::vector<char> &output);

    void logPendingErrors(bool need_linefeed);

    virtual void reschedule(Object *object) const = 0;
    virtual bool expectsActiveCheckResult() = 0;
};

class GenericHelper : public CheckHelper {
public:
    GenericHelper(const std::filesystem::path &path, const args_type &args,
                  result_handler_t result_handler, Logger *logger,
                  reschedule_handler_t<Object> reschedule_handler);
    std::string makeRequest() override;

private:
    reschedule_handler_t<Object> reschedule_handler_;

    // CommonHelper API
    bool sendReloadConfigRequest() override;
    void incrementCounter() const override;

    // CheckHelper API
    void reschedule(Object *object) const override;
    bool expectsActiveCheckResult() override;
};

class RealTimeHelper : public CheckHelper {
public:
    using CheckHelper::CheckHelper;
    std::string makeRequest() override;

private:
    // CommonHelper API
    bool sendReloadConfigRequest() override;
    void incrementCounter() const override;

    // CheckHelper API
    void reschedule(Object *object) const override;
    bool expectsActiveCheckResult() override;
};

class CheckerHelper : public CheckHelper {
public:
    CheckerHelper(const std::filesystem::path &path, const args_type &args,
                  result_handler_t result_handler, Logger *logger,
                  reschedule_handler_t<Service> reschedule_handler);
    std::string makeRequest() override;

private:
    reschedule_handler_t<Service> reschedule_handler_;

    // CommonHelper API
    bool sendReloadConfigRequest() override;
    void incrementCounter() const override;

    // CheckHelper API
    void reschedule(Object *object) const override;
    bool expectsActiveCheckResult() override;
};

#endif  // CheckHelper_h
