// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef CheckHelperPool_h
#define CheckHelperPool_h

#include <chrono>
#include <cstddef>
#include <deque>
#include <filesystem>
#include <memory>
#include <optional>
#include <string>
#include <vector>

#include "cmc/AbstractHelper.h"
#include "cmc/CheckHelper.h"
#include "cmc/FetcherHelper.h"
#include "cmc/Schedule.h"
#include "livestatus/Average.h"

class CheckingConfig;
class GeneralConfig;
class Logger;
class Object;
class Poller;
class Service;
class World;

class CheckHelperPool : public AbstractHelper {
public:
    CheckHelperPool(const std::string &name, Logger *logger,
                    Logger *logger_schedule);
    ~CheckHelperPool() override;

    double averageUsage() const;
    double averageLatency() const;
    void updateUsageAverage();

    size_t getNumIdleHelpers() const;
    // TODO(sp) The return type forces the introduction of a downcast to
    // CheckHelper* at one use site. We can fix this by templatizing
    // CheckHelperPool or (probably better) by actually using the Schedule
    // within the checkers instead of Core's ReplyQueue.
    CommonHelper *getIdleHelper() const;
    size_t size() const { return _helpers.size(); }

    [[nodiscard]] std::optional<std::chrono::system_clock::time_point> insert(
        const Schedule<Object *>::Job &job);
    [[nodiscard]] std::optional<Schedule<Object *>::Job> lowestUnder(
        std::chrono::system_clock::time_point deadline) const;
    void clearSchedule();
    size_t runScheduledChecks(std::chrono::system_clock::time_point now);
    size_t runnableCount(std::chrono::system_clock::time_point now) const;

    bool updateConfig(const GeneralConfig &general,
                      const CheckingConfig &checking);
    // TODO(sp) Why do we have updateConfig() *and* migrate()?
    bool migrate(const World &world);
    void reload();
    void restartDeadHelpers();

protected:
    struct HelperConfig {
        size_t quantity{0};
        std::filesystem::path path;
        std::vector<std::string> args;

        bool operator==(const HelperConfig &other) const {
            return quantity == other.quantity &&  //
                   path == other.path &&          //
                   args == other.args;
        }

        bool operator!=(const HelperConfig &rhs) const {
            return !(*this == rhs);
        }
    } config_;

private:
    Average _avg_usage;
    Average _avg_latency;

    bool _config_has_been_propagated{true};

    Schedule<Object *> _schedule;
    std::deque<std::unique_ptr<CommonHelper>> _helpers;

    virtual HelperConfig helperConfig(const GeneralConfig &general,
                                      const CheckingConfig &checking) const = 0;
    virtual std::unique_ptr<CommonHelper> makeHelper() const = 0;

    bool tryToRunCheck(std::chrono::system_clock::time_point now,
                       const Schedule<Object *>::Job &job);

    // implementation of abstract methods from AbstractHelper
    void start_() override;
    void stop_() override;
    void setUpPoller_(Poller &poller) const override;
    void communicate_(const Poller &poller) override;

    void decreaseNumberOfHelpers();
    void increaseNumberOfHelpers();

    void do_stop();
};

class GenericHelperPool : public CheckHelperPool {
public:
    GenericHelperPool(const std::string &name, Logger *logger,
                      Logger *logger_schedule,
                      reschedule_handler_t<Object> reschedule_handler,
                      CheckHelper::result_handler_t result_handler);

private:
    reschedule_handler_t<Object> reschedule_handler_;
    CheckHelper::result_handler_t _result_handler;

    // CheckHelperPool API
    HelperConfig helperConfig(const GeneralConfig &general,
                              const CheckingConfig &checking) const override;
    std::unique_ptr<CommonHelper> makeHelper() const override;
};

// Passive checker helper pool, supplements FetcherPool
class CheckerHelperPool : public CheckHelperPool {
public:
    CheckerHelperPool(const std::string &name, Logger *logger,
                      Logger *logger_schedule,
                      reschedule_handler_t<Service> reschedule_handler,
                      CheckHelper::result_handler_t result_handler);

private:
    reschedule_handler_t<Service> reschedule_handler_;
    CheckHelper::result_handler_t _result_handler;

    // CheckHelperPool API
    HelperConfig helperConfig(const GeneralConfig &general,
                              const CheckingConfig &checking) const override;
    std::unique_ptr<CommonHelper> makeHelper() const override;
};

class RealTimeHelperPool : public CheckHelperPool {
public:
    RealTimeHelperPool(const std::string &name, Logger *logger,
                       Logger *logger_schedule,
                       CheckHelper::result_handler_t result_handler);

private:
    CheckHelper::result_handler_t _result_handler;

    // CheckHelperPool API
    HelperConfig helperConfig(const GeneralConfig &general,
                              const CheckingConfig &checking) const override;
    std::unique_ptr<CommonHelper> makeHelper() const override;
};

class FetcherHelperPool : public CheckHelperPool {
public:
    FetcherHelperPool(const std::string &name, Logger *logger,
                      Logger *logger_schedule,
                      reschedule_handler_t<Service> reschedule_handler,
                      FetcherHelper::reply_handler_t reply_handler);

private:
    reschedule_handler_t<Service> reschedule_handler_;
    FetcherHelper::reply_handler_t reply_handler_;

    // CheckHelperPool API
    HelperConfig helperConfig(const GeneralConfig &general,
                              const CheckingConfig &checking) const override;
    std::unique_ptr<CommonHelper> makeHelper() const override;
};

#endif  // CheckHelperPool_h
