// Copyright (C) 2023 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef CmcCore_h
#define CmcCore_h

#include <chrono>
#include <cstddef>
#include <cstdint>
#include <functional>
#include <memory>
#include <string>
#include <vector>

#include "livestatus/ICore.h"
#include "livestatus/Metric.h"

class Core;
enum class Encoding;
class IComment;
class IContact;
class IContactGroup;
class IDowntime;
class IGlobalFlags;
class IHost;
class IHostGroup;
class IPaths;
class IService;
class IServiceGroup;
class ITimeperiod;
class Logger;
class Triggers;
class User;

class CmcCore : public ICore {
public:
    explicit CmcCore(Core *core) : core_(core) {}

    [[nodiscard]] const IHost *find_host(
        const std::string &name) const override;
    [[nodiscard]] const IHostGroup *find_hostgroup(
        const std::string &name) const override;
    [[nodiscard]] std::unique_ptr<const IHost> getHostByDesignation(
        const std::string &designation) const override;
    bool all_of_hosts(
        const std::function<bool(const IHost &)> &pred) const override;
    bool all_of_services(
        const std::function<bool(const IService &)> &pred) const override;

    [[nodiscard]] const IService *find_service(
        const std::string &host_name,
        const std::string &service_description) const override;

    [[nodiscard]] const IContactGroup *find_contactgroup(
        const std::string &name) const override;

    [[nodiscard]] const IServiceGroup *find_servicegroup(
        const std::string &name) const override;

    [[nodiscard]] const IContact *find_contact(
        const std::string &name) const override;
    bool all_of_contacts(
        const std::function<bool(const IContact &)> &pred) const override;

    [[nodiscard]] std::unique_ptr<const User> find_user(
        const std::string &name) const override;

    [[nodiscard]] std::chrono::system_clock::time_point last_logfile_rotation()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_config_change()
        const override;
    [[nodiscard]] size_t maxLinesPerLogFile() const override;

    [[nodiscard]] Command find_command(const std::string &name) const override;
    [[nodiscard]] std::vector<Command> commands() const override;

    [[nodiscard]] std::vector<std::unique_ptr<const IComment>>
    comments_unlocked(const IHost &host) const override;
    [[nodiscard]] std::vector<std::unique_ptr<const IComment>> comments(
        const IHost &host) const override;
    [[nodiscard]] std::vector<std::unique_ptr<const IComment>>
    comments_unlocked(const IService &service) const override;
    [[nodiscard]] std::vector<std::unique_ptr<const IComment>> comments(
        const IService &service) const override;
    bool all_of_comments(
        const std::function<bool(const IComment &)> &pred) const override;

    [[nodiscard]] std::vector<std::unique_ptr<const IDowntime>>
    downtimes_unlocked(const IHost &host) const override;
    [[nodiscard]] std::vector<std::unique_ptr<const IDowntime>> downtimes(
        const IHost &host) const override;
    [[nodiscard]] std::vector<std::unique_ptr<const IDowntime>>
    downtimes_unlocked(const IService &service) const override;
    [[nodiscard]] std::vector<std::unique_ptr<const IDowntime>> downtimes(
        const IService &service) const override;
    bool all_of_downtimes(
        const std::function<bool(const IDowntime &)> &pred) const override;

    bool all_of_timeperiods(
        const std::function<bool(const ITimeperiod &)> &pred) const override;

    bool all_of_contact_groups(
        const std::function<bool(const IContactGroup &)> &pred) const override;

    bool all_of_host_groups(
        const std::function<bool(const IHostGroup &)> &pred) const override;

    bool all_of_service_groups(
        const std::function<bool(const IServiceGroup &)> &pred) const override;

    [[nodiscard]] bool mkeventdEnabled() const override;

    [[nodiscard]] int32_t pid() const override;
    [[nodiscard]] std::unique_ptr<const IGlobalFlags> globalFlags()
        const override;
    [[nodiscard]] std::unique_ptr<const IPaths> paths() const override;
    [[nodiscard]] std::chrono::system_clock::time_point programStartTime()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point lastCommandCheckTime()
        const override;
    [[nodiscard]] int32_t intervalLength() const override;
    [[nodiscard]] int32_t maxLongOutputSize() const override;
    [[nodiscard]] int32_t numHosts() const override;
    [[nodiscard]] int32_t numServices() const override;
    [[nodiscard]] std::string programVersion() const override;
    [[nodiscard]] std::string edition() const override;

    [[nodiscard]] int32_t externalCommandBufferSlots() const override;
    [[nodiscard]] int32_t externalCommandBufferUsage() const override;
    [[nodiscard]] int32_t externalCommandBufferMax() const override;

    [[nodiscard]] int32_t livestatusActiveConnectionsNum() const override;
    [[nodiscard]] std::string livestatusVersion() const override;
    [[nodiscard]] int32_t livestatusQueuedConnectionsNum() const override;
    [[nodiscard]] int32_t livestatusThreadsNum() const override;
    [[nodiscard]] double livestatusUsage() const override;

    [[nodiscard]] double averageLatencyGeneric() const override;
    [[nodiscard]] double averageLatencyRealTime() const override;
    [[nodiscard]] double averageLatencyFetcher() const override;
    [[nodiscard]] double averageLatencyChecker() const override;

    [[nodiscard]] double helperUsageGeneric() const override;
    [[nodiscard]] double helperUsageRealTime() const override;
    [[nodiscard]] double helperUsageFetcher() const override;
    [[nodiscard]] double helperUsageChecker() const override;

    [[nodiscard]] bool hasEventHandlers() const override;

    [[nodiscard]] double averageRunnableJobsFetcher() const override;
    [[nodiscard]] double averageRunnableJobsChecker() const override;

    [[nodiscard]] std::chrono::system_clock::time_point stateFileCreatedTime()
        const override;

    [[nodiscard]] std::vector<std::string> metrics(
        const IHost &host) const override;
    [[nodiscard]] std::vector<std::string> metrics(
        const IService &service) const override;

    [[nodiscard]] Encoding dataEncoding() const override;
    [[nodiscard]] size_t maxResponseSize() const override;
    [[nodiscard]] size_t maxCachedMessages() const override;

    [[nodiscard]] Logger *loggerCore() const override;
    [[nodiscard]] Logger *loggerLivestatus() const override;
    [[nodiscard]] Logger *loggerRRD() const override;

    Triggers &triggers() override;

    [[nodiscard]] size_t numQueuedNotifications() const override;
    [[nodiscard]] size_t numQueuedAlerts() const override;
    [[nodiscard]] size_t numCachedLogMessages() override;

    [[nodiscard]] bool isPnpGraphPresent(const IHost &host) const override;
    [[nodiscard]] bool isPnpGraphPresent(
        const IService &service) const override;

    [[nodiscard]] MetricLocation metricLocation(
        const std::string &host_name, const std::string &service_description,
        const Metric::Name &var) const override;

    [[nodiscard]] bool pnp4nagiosEnabled() const override;

private:
    Core *core_;

    [[nodiscard]] void *implInternal() const override { return core_; }
};

#endif  // CmcCore_h
