// Copyright (C) 2023 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef CmcHost_h
#define CmcHost_h

#include <chrono>
#include <cstdint>
#include <functional>
#include <string>
#include <vector>

#include "livestatus/Interface.h"

class Host;

class CmcHost : public IHost {
public:
    explicit CmcHost(const Host &host);

    [[nodiscard]] const Host &handle() const { return host_; }

    [[nodiscard]] const void *handleForStateHistory() const override;
    [[nodiscard]] bool hasContact(const IContact &contact) const override;
    [[nodiscard]] std::string notificationPeriodName() const override;
    [[nodiscard]] std::string servicePeriodName() const override;

    [[nodiscard]] std::string name() const override;
    [[nodiscard]] std::string display_name() const override;
    [[nodiscard]] std::string alias() const override;
    [[nodiscard]] std::string ip_address() const override;
    [[nodiscard]] std::string check_command() const override;
    [[nodiscard]] std::string check_command_expanded() const override;
    [[nodiscard]] std::string event_handler() const override;
    [[nodiscard]] std::string notification_period() const override;
    [[nodiscard]] std::string check_period() const override;
    [[nodiscard]] std::string notes() const override;
    [[nodiscard]] std::string notes_expanded() const override;
    [[nodiscard]] std::string notes_url() const override;
    [[nodiscard]] std::string notes_url_expanded() const override;
    [[nodiscard]] std::string action_url() const override;
    [[nodiscard]] std::string action_url_expanded() const override;
    [[nodiscard]] std::string plugin_output() const override;
    [[nodiscard]] std::string perf_data() const override;
    [[nodiscard]] std::string icon_image() const override;
    [[nodiscard]] std::string icon_image_expanded() const override;
    [[nodiscard]] std::string icon_image_alt() const override;
    [[nodiscard]] std::string status_map_image() const override;
    [[nodiscard]] std::string long_plugin_output() const override;
    [[nodiscard]] int32_t initial_state() const override;
    [[nodiscard]] int32_t max_check_attempts() const override;
    [[nodiscard]] bool flap_detection_enabled() const override;
    [[nodiscard]] bool check_freshness() const override;
    [[nodiscard]] bool process_performance_data() const override;
    [[nodiscard]] bool accept_passive_host_checks() const override;
    [[nodiscard]] int32_t event_handler_enabled() const override;
    [[nodiscard]] int32_t acknowledgement_type() const override;
    [[nodiscard]] int32_t check_type() const override;
    [[nodiscard]] int32_t last_state() const override;
    [[nodiscard]] int32_t last_hard_state() const override;
    [[nodiscard]] int32_t current_attempt() const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_notification()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point next_notification()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point next_check()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_hard_state_change()
        const override;
    [[nodiscard]] bool has_been_checked() const override;
    [[nodiscard]] int32_t current_notification_number() const override;
    [[nodiscard]] int32_t pending_flex_downtime() const override;
    [[nodiscard]] int32_t total_services() const override;
    [[nodiscard]] bool notifications_enabled() const override;
    [[nodiscard]] bool problem_has_been_acknowledged() const override;
    [[nodiscard]] int32_t current_state() const override;
    [[nodiscard]] int32_t hard_state() const override;
    [[nodiscard]] int32_t state_type() const override;
    [[nodiscard]] int32_t no_more_notifications() const override;
    [[nodiscard]] int32_t check_flapping_recovery_notification() const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_check()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_state_change()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_time_up()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_time_down()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_time_unreachable()
        const override;

    [[nodiscard]] bool is_flapping() const override;
    [[nodiscard]] int32_t scheduled_downtime_depth() const override;
    [[nodiscard]] bool is_executing() const override;
    [[nodiscard]] bool active_checks_enabled() const override;
    [[nodiscard]] int32_t check_options() const override;
    [[nodiscard]] int32_t obsess_over_host() const override;
    [[nodiscard]] uint32_t modified_attributes() const override;
    [[nodiscard]] double check_interval() const override;
    [[nodiscard]] double retry_interval() const override;
    [[nodiscard]] double notification_interval() const override;
    [[nodiscard]] double first_notification_delay() const override;
    [[nodiscard]] double low_flap_threshold() const override;
    [[nodiscard]] double high_flap_threshold() const override;
    [[nodiscard]] double x_3d() const override;
    [[nodiscard]] double y_3d() const override;
    [[nodiscard]] double z_3d() const override;
    [[nodiscard]] double latency() const override;
    [[nodiscard]] double execution_time() const override;
    [[nodiscard]] double percent_state_change() const override;
    [[nodiscard]] double staleness() const override;
    [[nodiscard]] double flappiness() const override;
    [[nodiscard]] bool in_notification_period() const override;
    [[nodiscard]] bool in_check_period() const override;
    [[nodiscard]] bool in_service_period() const override;
    [[nodiscard]] std::vector<std::string> contacts() const override;
    [[nodiscard]] Attributes attributes(AttributeKind kind) const override;
    [[nodiscard]] std::string filename() const override;
    [[nodiscard]] std::string notification_postponement_reason() const override;
    [[nodiscard]] int32_t previous_hard_state() const override;
    [[nodiscard]] int32_t smartping_timeout() const override;

    bool all_of_services(
        std::function<bool(const IService &)> pred) const override;
    bool all_of_labels(
        const std::function<bool(const Attribute &)> &pred) const override;
    bool all_of_parents(std::function<bool(const IHost &)> pred) const override;
    bool all_of_children(
        std::function<bool(const IHost &)> pred) const override;
    bool all_of_host_groups(
        std::function<bool(const IHostGroup &)> pred) const override;
    bool all_of_contact_groups(
        std::function<bool(const IContactGroup &)> pred) const override;

private:
    const Host &host_;

    [[nodiscard]] double double_attr(const std::string &name) const;
};

#endif  // CmcHost_h
