// Copyright (C) 2023 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef CmcService_h
#define CmcService_h

#include <chrono>
#include <cstdint>
#include <filesystem>
#include <functional>
#include <ratio>
#include <string>
#include <vector>

#include "livestatus/Interface.h"
class Service;

class CmcService : public IService {
public:
    using minutes_d = std::chrono::duration<double, std::ratio<60> >;

    explicit CmcService(const Service &service);

    [[nodiscard]] const Service &handle() const { return service_; }

    [[nodiscard]] const void *handleForStateHistory() const override;
    [[nodiscard]] const IHost &host() const override;
    [[nodiscard]] bool hasContact(const IContact &contact) const override;
    [[nodiscard]] std::string servicePeriodName() const override;
    [[nodiscard]] bool in_custom_time_period() const override;

    [[nodiscard]] std::string name() const override;
    [[nodiscard]] std::string description() const override;
    [[nodiscard]] std::string host_name() const override;
    [[nodiscard]] std::string display_name() const override;
    [[nodiscard]] std::string check_command() const override;
    [[nodiscard]] std::string check_command_expanded() const override;
    [[nodiscard]] std::filesystem::path robotmk_dir() const override;
    [[nodiscard]] std::string event_handler() const override;
    [[nodiscard]] std::string plugin_output() const override;
    [[nodiscard]] std::string long_plugin_output() const override;
    [[nodiscard]] std::string perf_data() const override;
    [[nodiscard]] std::string notificationPeriodName() const override;
    [[nodiscard]] std::string check_period() const override;
    [[nodiscard]] std::string notes() const override;
    [[nodiscard]] std::string notes_expanded() const override;
    [[nodiscard]] std::string notes_url() const override;
    [[nodiscard]] std::string notes_url_expanded() const override;
    [[nodiscard]] std::string action_url() const override;
    [[nodiscard]] std::string action_url_expanded() const override;
    [[nodiscard]] std::string icon_image() const override;
    [[nodiscard]] std::string icon_image_expanded() const override;
    [[nodiscard]] std::string icon_image_alt() const override;
    [[nodiscard]] int32_t initial_state() const override;
    [[nodiscard]] int32_t max_check_attempts() const override;
    [[nodiscard]] int32_t current_attempt() const override;
    [[nodiscard]] int32_t current_state() const override;
    [[nodiscard]] bool has_been_checked() const override;
    [[nodiscard]] int32_t last_state() const override;
    [[nodiscard]] int32_t last_hard_state() const override;
    [[nodiscard]] int32_t state_type() const override;
    [[nodiscard]] int32_t check_type() const override;
    [[nodiscard]] bool problem_has_been_acknowledged() const override;
    [[nodiscard]] int32_t acknowledgement_type() const override;
    [[nodiscard]] bool no_more_notifications() const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_time_ok()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_time_warning()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_time_critical()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_time_unknown()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_check()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point next_check()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_notification()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point next_notification()
        const override;
    [[nodiscard]] int32_t current_notification_number() const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_state_change()
        const override;
    [[nodiscard]] std::chrono::system_clock::time_point last_hard_state_change()
        const override;
    [[nodiscard]] int32_t scheduled_downtime_depth() const override;
    [[nodiscard]] bool is_flapping() const override;
    [[nodiscard]] bool checks_enabled() const override;
    [[nodiscard]] bool accept_passive_checks() const override;
    [[nodiscard]] bool event_handler_enabled() const override;
    [[nodiscard]] bool notifications_enabled() const override;
    [[nodiscard]] bool process_performance_data() const override;
    [[nodiscard]] bool is_executing() const override;
    [[nodiscard]] bool active_checks_enabled() const override;
    [[nodiscard]] int32_t check_options() const override;
    [[nodiscard]] bool flap_detection_enabled() const override;
    [[nodiscard]] bool check_freshness() const override;
    [[nodiscard]] bool obsess_over_service() const override;
    [[nodiscard]] uint32_t modified_attributes() const override;
    [[nodiscard]] int32_t hard_state() const override;
    [[nodiscard]] double staleness() const override;
    [[nodiscard]] double check_interval() const override;
    [[nodiscard]] double retry_interval() const override;
    [[nodiscard]] double notification_interval() const override;
    [[nodiscard]] double first_notification_delay() const override;
    [[nodiscard]] double low_flap_threshold() const override;
    [[nodiscard]] double high_flap_threshold() const override;
    [[nodiscard]] double latency() const override;
    [[nodiscard]] double execution_time() const override;
    [[nodiscard]] double percent_state_change() const override;
    [[nodiscard]] bool in_check_period() const override;
    [[nodiscard]] bool in_service_period() const override;
    [[nodiscard]] bool in_notification_period() const override;
    [[nodiscard]] std::vector<std::string> contacts() const override;
    [[nodiscard]] Attributes attributes(AttributeKind kind) const override;
    [[nodiscard]] int32_t cache_interval() const override;
    [[nodiscard]] bool in_passive_check_period() const override;
    [[nodiscard]] std::string passive_check_period() const override;
    [[nodiscard]] double flappiness() const override;
    [[nodiscard]] std::string notification_postponement_reason() const override;
    [[nodiscard]] int32_t previous_hard_state() const override;
    [[nodiscard]] int32_t pending_flex_downtime() const override;
    [[nodiscard]] bool check_flapping_recovery_notification() const override;
    bool all_of_service_groups(
        std::function<bool(const IServiceGroup &)> pred) const override;
    bool all_of_contact_groups(
        std::function<bool(const IContactGroup &)> pred) const override;
    [[nodiscard]] std::chrono::system_clock::time_point cached_at()
        const override;

    bool all_of_labels(
        const std::function<bool(const Attribute &)> &pred) const override;

private:
    const Service &service_;
};

#endif  // CmcService_h
