// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef ConnectionPool_h
#define ConnectionPool_h

#include <algorithm>
#include <chrono>
#include <cstddef>
#include <iostream>
#include <string>
#include <utility>
#include <vector>

#include "cmc/DataExportThread.h"
#include "livestatus/Logger.h"
#include "livestatus/global_counters.h"

template <typename Update, typename Reload, typename Connection>
class ConnectionPool : public DataExportThread<Update, Reload> {
public:
    ConnectionPool(const std::string &name, Counter overflow_counter,
                   Counter queue_usage, Counter bytes_sent_counter,
                   Logger *logger);

private:
    Counter bytes_sent_counter_;
    std::vector<Connection> connections_;
    void handle(const Update &message) override;
    void handle(const Reload &message) override;
    void close() override {}
    size_t establishConnections();
};

template <typename Update, typename Reload, typename Connection>
ConnectionPool<Update, Reload, Connection>::ConnectionPool(
    const std::string &name, Counter overflow_counter, Counter queue_usage,
    Counter bytes_sent_counter, Logger *logger)
    : DataExportThread<Update, Reload>  //
    {name,
     overflow_counter,
     queue_usage,
     std::chrono::milliseconds{100},
     [this]() {
         std::for_each(connections_.begin(), connections_.end(),
                       [](auto &&conn) { conn.doHousekeeping(); });
     },
     {logger,
      [name](std::ostream &os) { os << "[" + name + " connection pool] "; }}}
    , bytes_sent_counter_{bytes_sent_counter} {}

template <typename Update, typename Reload, typename Connection>
void ConnectionPool<Update, Reload, Connection>::handle(const Reload &message) {
    connections_.clear();
    for (auto &&conf : message.configs) {
        connections_.emplace_back(conf, bytes_sent_counter_);
    }
    establishConnections();
}

template <typename Update, typename Reload, typename Connection>
size_t ConnectionPool<Update, Reload, Connection>::establishConnections() {
    auto now = std::chrono::steady_clock::now();
    size_t num_connected_sockets = 0;
    for (auto &conn : connections_) {
        if (conn.tryToInitiateConnection(now) !=
            Connection::State::unconnected) {
            num_connected_sockets++;
        }
    }
    return num_connected_sockets;
}

template <typename Update, typename Reload, typename Connection>
void ConnectionPool<Update, Reload, Connection>::handle(const Update &message) {
    if (establishConnections() == 0) {
        return;
    }
    for (auto &conn : connections_) {
        conn.send(std::move(message));
    }
}

#endif  // ConnectionPool_h
