// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef Core_h
#define Core_h

#include <atomic>
#include <chrono>
#include <cstdint>
#include <filesystem>
#include <fstream>
#include <memory>
#include <mutex>
#include <ratio>
#include <shared_mutex>
#include <string>
#include <string_view>
#include <unordered_map>
#include <unordered_set>
#include <utility>

#include "asio/io_context.hpp"
#include "cmc/AlertHelper.h"
#include "cmc/CarbonConnection.h"
#include "cmc/CarbonProtocol.h"
#include "cmc/CheckHelperPool.h"
#include "cmc/Comments.h"
#include "cmc/ConfigCleaner.h"
#include "cmc/ConnectionPool.h"
#include "cmc/Downtimes.h"
#include "cmc/EventSpooler.h"
#include "cmc/FetcherReply.h"
#include "cmc/GlobalState.h"
#include "cmc/ICMPReceiver.h"
#include "cmc/ICMPSender.h"
#include "cmc/InfluxDBConnection.h"
#include "cmc/InfluxDBProtocol.h"
#include "cmc/LivestatusExternalCommand.h"
#include "cmc/LivestatusLocalCommand.h"
#include "cmc/MonitoringHistory.h"
#include "cmc/MonitoringLog.h"
#include "cmc/NotifyHelper.h"
#include "cmc/RRDHelper.h"
#include "cmc/RealTimeCheck.h"
#include "cmc/Schedule.h"
#include "cmc/SerialToken.h"
#include "cmc/State.h"
#include "cmc/TableCachedStatehist.h"
#include "cmc/Worker.h"
#include "livestatus/Average.h"
#include "livestatus/Logger.h"
#include "livestatus/Store.h"
#include "livestatus/TrialManager.h"
#include "livestatus/Triggers.h"

class CheckResult;
class Core;
class Downtime;
class GlobalConfig;
class HistoryConfig;
class Host;
class Livestatus;
class ICore;
class Object;
class Service;
class World;
namespace checkmk::cmc ::config ::v1 {
class Core;
}  // namespace checkmk::cmc::config::v1
namespace checkmk::cmc::state ::v1 {
class StateFile;
}  // namespace checkmk::cmc::state::v1

struct StateSaverTask {
    Core &core;
    std::filesystem::path state_path;
    std::filesystem::path renamed_hosts_path;
};

class Core {
private:
    asio::io_context _io_context;

public:
    char **_argv;
    int32_t _pid;

    // lock loggers during threaded I/O
    mutable std::mutex _logging_lock;
    std::ofstream *_log_file;
    Logger *const _logger_alert;
    ThreadNameLogger _carbon_context_logger;
    Logger *const _logger_carbon;
    Logger *const _logger_core;
    Logger *const _logger_downtime;
    Logger *const _logger_helper;
    ThreadNameLogger _livestatus_context_logger;
    Logger *const _logger_livestatus;
    Logger *const _logger_notification;
    ThreadNameLogger _rrd_context_logger;
    Logger *const _logger_rrd;
    ThreadNameLogger _influxdb_context_logger;
    Logger *const _logger_influxdb;
    Logger *const _logger_smart_ping;
    Logger *const _logger_fetcher;

    SerialTokenFactory _serial_token_factory;
    std::string _config_filename;
    FetcherReplyQueue _reply_queue;

    // Configuration data currently in use
    World *_world;
    // lingering old configuration, waiting for deletion
    std::unordered_set<World *> _old_worlds;

    std::unique_ptr<ICore> _mc;

    TrialManager trial_manager_;

    // lock worlds during log cache queries
    mutable std::shared_mutex _world_lock;

    // List of check helpers. They live here.
    GenericHelperPool _helper_pool_generic;
    CheckerHelperPool _helper_pool_checker;
    RealTimeHelperPool _helper_pool_real_time;
    FetcherHelperPool _helper_pool_fetcher;

    // Other dedicated helpers
    EventSpooler notification_spooler_;
    NotifyHelper _notify_helper;
    EventSpooler alert_spooler_;
    AlertHelper _alert_helper;
    ICMPSender _icmp_sender;
    ICMPReceiver _icmp_receiver;

    // Pipe for classical commands (tmp/run/nagios.cmd)
    LivestatusLocalCommand _livestatus_local;

    Schedule<Host *> _smart_ping_schedule;
    Schedule<Host *> _always_up_schedule;
    Schedule<Host *> _service_does_not_exist_schedule;
    Schedule<Host *> _cluster_host_schedule;

    MonitoringHistory monitoring_history_;
    MonitoringLog monitoring_log_;

    // Microcore data for Livestatus
    Average _average_runnable_jobs_checker;
    Average _average_runnable_jobs_fetcher;

    // Livestatus
    std::unique_ptr<Livestatus> _livestatus;
    int _livestatus_threads;  // cache, always == _livestatus->numThreads()
    std::atomic_int32_t _livestatus_active_connections;

    LivestatusExternalCommand _livestatus_ext;

    Comments _comments;
    Downtimes _downtimes;

    // Various, not retained in saved state
    bool _is_shutting_down;  // true in shutdown process
    std::chrono::system_clock::time_point _core_start;
    std::chrono::system_clock::time_point _last_config_change;
    std::chrono::system_clock::time_point _last_housekeeping;
    std::chrono::system_clock::time_point _last_state_retention;
    std::chrono::system_clock::time_point _last_downtimes_check;
    std::chrono::system_clock::time_point _last_comments_check;
    std::chrono::system_clock::time_point _last_smart_ping_check;
    std::chrono::system_clock::time_point _last_restart_dead_helpers;
    std::chrono::system_clock::time_point _last_config_clean;
    std::chrono::system_clock::time_point _schedule_pointer_cmk;
    std::chrono::system_clock::time_point _schedule_pointer_generic;
    std::chrono::system_clock::time_point _last_flapping_check;
    std::chrono::duration<double, std::milli> _state_load_time;
    bool _should_terminate;
    bool _should_reload;
    int _num_host_states_loaded;
    int _num_service_states_loaded;
    int _num_hosts_vanished;
    int _num_services_vanished;
    Average _avg_livestatus_usage;

    Store _store;

    RRDHelper _rrd_helper;
    ConnectionPool<carbon::protocol::Update, carbon::protocol::Reload,
                   CarbonConnection>
        _carbon_connection_pool;
    ConnectionPool<influxdb::protocol::Update, influxdb::protocol::Reload,
                   influxdb::Connection>
        _influxdb_connection_pool;

    Triggers _triggers;
    Worker<ConfigCleaner::Task> config_cleaner_;
    Worker<StateSaverTask, 1> state_saver_;

    Core(char **argv, std::string config_filename,
         std::unique_ptr<checkmk::cmc::config::v1::Core> core_config,
         std::ofstream *log_file, Logger *logger_core,
         TrialManager trial_manager);
    ~Core();
    // NOLINTNEXTLINE(readability-make-member-function-const)
    [[nodiscard]] World *world() { return _world; }
    [[nodiscard]] World *world() const { return _world; }
    const GlobalConfig *config() const;
    Logger *loggerLivestatus() const { return _logger_livestatus; }
    Logger *loggerNotifications() const { return _notify_helper.logger(); }
    void reloadConfig();
    void migrateStates(std::chrono::system_clock::time_point now,
                       World *new_world);
    void createState(
        std::unique_ptr<checkmk::cmc::state::v1::StateFile> state_file);
    void run();
    bool isShuttingDown() const { return _is_shutting_down; }
    void logNotification(const Object &object, const std::string &message);
    void flappingEvent(const Object &object);
    void downtimeEvent(const Object &object, const std::string &state_type,
                       const std::string &comment);
    void processCheckResult(std::chrono::system_clock::time_point now,
                            const CheckResult &cr);
    void objectStateHasBeenChanged(std::chrono::system_clock::time_point now,
                                   Object *object);
    void processAlertHelperResponse(bool started, std::string result);
    Object *getAlertHandlerResponseObject(
        const std::string &host_name,
        const std::string &service_description) const;
    void notify(const NotificationInfo &info);
    void broadcast(const NotificationInfo &info);
    uint32_t nextProblemId() {
        return globalState()->increaseNotificationNextProblemId();
    }
    GlobalState *globalState() { return &_global_state; };
    const GlobalState *globalState() const { return &_global_state; };
    void answerCommandRequest(const std::string &command);
    void spoolCommand(const std::string &command);
    std::filesystem::path renamed_hosts_path() const;
    std::filesystem::path state_path() const;
    static std::filesystem::path state_path(
        const std::filesystem::path &core_directory);
    std::filesystem::path versioned_state_path(uint32_t version) const;
    void saveState();
    void scheduleHostJob(const Schedule<Host *>::Job &job);
    void scheduleServiceJob(const Schedule<Service *>::Job &job);
    std::chrono::system_clock::time_point initialScheduleTime(
        std::chrono::system_clock::time_point now, Object *object);
    std::pair<std::string, std::string> findAcknowledgement(
        const Object &object) const;
    static std::unique_ptr<checkmk::cmc::config::v1::Core> loadConfigFromFile(
        const std::filesystem::path &path);
    static std::unique_ptr<checkmk::cmc::state::v1::StateFile>
    loadStateFromFile(const std::filesystem::path &path);

private:
    GlobalState _global_state;  // settings like notifications_enabled
    RealTimeCheck _real_time_check;
    TableCachedStatehist table_cached_statehist_;

    // We use dynamically allocated memory => disable copy/assignment
    Core(const Core &);
    Core &operator=(const Core &);

    void switchStatehistTable();
    void insertDowntime(std::unique_ptr<Downtime> downtime);

    static void saveStateProtobuf(Logger *logger, const Core &core,
                                  const std::filesystem::path &state_path,
                                  const std::filesystem::path &renamed_hosts);
    void setStateFrom(const checkmk::cmc::state::v1::StateFile &state_file);

    void configureCoreDumps() const;
    static std::unordered_set<const Object *> findVanishedObjects(
        World *old_world, World *new_world);
    void logVanishedObject(const std::unordered_set<const Object *> &vanished);
    void logVanishedObject(const Object &object);
    void cancelDowntimes(const std::unordered_set<const Object *> &objects);
    void handleChangedDowntimeRules(std::chrono::system_clock::time_point now);
    void removeAllHistcachePointers() const;
    void migrateCheckHelpers(World *);
    void scheduleAllChecks();
    static void installSignalHandlers();
    static void installSegfaultHandler();
    static void removeSegfaultHandler();
    void mainLoop();
    uint32_t newNoteId();
    void expireComments(std::chrono::system_clock::time_point now);
    void expireDowntimes(std::chrono::system_clock::time_point now);
    void handleSmartPingTimeouts(std::chrono::system_clock::time_point now);
    void doHousekeeping(std::chrono::system_clock::time_point now);
    void unflap(std::chrono::system_clock::time_point now) const;
    std::chrono::system_clock::time_point initialScheduleTime(
        std::chrono::system_clock::time_point now,
        std::chrono::system_clock::time_point *schedule_pointer,
        std::chrono::nanoseconds scheduling_interval,
        std::chrono::nanoseconds check_interval) const;
    void setLogConfig() const;
    void rotateMonitoringLog(const HistoryConfig &hc);
    void logCurrentDowntimes();
    void deleteOldWorlds();
    void checkNotificationsAndFreshness(
        std::chrono::system_clock::time_point now);
    void updateStats(std::chrono::system_clock::time_point now);

    // external commands (in Core-commands.cc)
    void handleExternalCommands();
    void handleCommand(const std::string &command);
    void handleEventConsoleCommand(const std::string &command) const;
    Object *getCommandObject(std::string_view &args, bool is_host) const;
    bool commandFakeCheck(std::string_view args, bool is_host);
    bool commandProcessPerfdata(std::string_view args, bool is_host);
    bool commandScheduleCheck(std::string_view args, bool is_host);
    bool commandNotification(std::string_view args, bool is_host);
    bool commandAddComment(std::string_view args, bool is_host);
    bool commandDelComment(std::string_view args);
    bool commandDelCrashReport(std::string_view args) const;
    bool commandAcknowledge(std::string_view args, bool is_host);
    bool commandRemoveAck(std::string_view args, bool is_host);
    bool commandScheduleDowntime(std::string_view args, bool is_host);
    bool commandDelDowntime(std::string_view args);
    bool commandModifyDowntime(std::string_view args);
    bool commandEnable(std::string_view args, bool is_host, bool enabled,
                       State::Enabling::Key what);
    bool commandClearModattr(std::string_view args, bool is_host);
    bool commandEnableServices(std::string_view args, bool enabled,
                               State::Enabling::Key what);
    bool commandSaveState();
    bool commandReloadConfig();
    bool commandRotateLogfile();
    bool commandLog(std::string_view args);
    bool commandReopenDaemonLog();
    bool commandSegfault() const;
    bool commandUpdateShadowState(std::string_view args, bool is_host);
    bool commandMkLogwatchAcknowledge(std::string_view args);

    // saving and loading of state
    void createMissingStates(std::chrono::system_clock::time_point now) const;
    std::unordered_map<std::string, std::string> renamedHosts() const;

    void processPerfdata(const Object *object);
    void processFetcherReplyQueue();
};

#endif  // Core_h
