// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef DataExportThread_h
#define DataExportThread_h

#include <deque>
#include <functional>
#include <optional>
#include <thread>
#include <utility>
#include <variant>

#include "livestatus/Logger.h"
#include "livestatus/POSIXUtils.h"
#include "livestatus/Queue.h"
#include "livestatus/global_counters.h"

template <typename Update, typename Reload>
class DataExportThread {
public:
    using timeout_t = std::chrono::nanoseconds;
    using callback_t = std::function<void()>;

    DataExportThread(std::string thread_name, Counter overflow_counter,
                     Counter queue_usage, ContextLogger logger)
        : DataExportThread{thread_name, overflow_counter,
                           queue_usage, std::nullopt,
                           [] {},       logger} {}
    DataExportThread(std::string thread_name, Counter overflow_counter,
                     Counter queue_usage, std::optional<timeout_t> timeout,
                     callback_t on_idle, ContextLogger logger)
        : thread_name_{std::move(thread_name)}
        , overflow_counter_{overflow_counter}
        , queue_usage_{queue_usage}
        , timeout_{timeout}
        , on_idle_{std::move(on_idle)}
        , logger_{std::move(logger)} {}
    DataExportThread(const DataExportThread &) = delete;
    DataExportThread &operator=(const DataExportThread &) = delete;
    DataExportThread(DataExportThread &&) = delete;
    DataExportThread &operator=(DataExportThread &&) = delete;
    virtual ~DataExportThread() = default;
    // accessed from main thread -----------------------------------------------
    void update(Update &&u);
    void reload(Reload &&r);
    void start();
    void stop();

protected:
    // accessed from both threads ----------------------------------------------
    [[nodiscard]] Logger *logger() { return &logger_; }

private:
    std::string thread_name_;
    Counter overflow_counter_;
    Counter queue_usage_;
    std::optional<timeout_t> timeout_;
    callback_t on_idle_;
    mutable ContextLogger logger_;
    std::thread thread_;
    Queue<std::variant<Update, Reload>> queue_{
        queue_join_strategy::shutdown_push_pop, 20000};
    mutable std::atomic<bool> stop_ = false;

    // accessed from both threads ----------------------------------------------
    virtual void close() = 0;

    // called from our own thread ----------------------------------------------
    virtual void handle(const Update &u) = 0;
    virtual void handle(const Reload &r) = 0;

    // accessed from main thread -----------------------------------------------
    void run();
};

template <typename Update, typename Reload>
void DataExportThread<Update, Reload>::start() {
    if (thread_.joinable()) {
        Warning(logger()) << "thread already running";
        return;
    }
    thread_ = std::thread(&DataExportThread::run, this);
}

template <typename Update, typename Reload>
void DataExportThread<Update, Reload>::stop() {
    stop_ = true;
    if (thread_.joinable()) {
        Notice(logger()) << "stopping...";
        queue_.join();
        thread_.join();
    }
}

template <typename Update, typename Reload>
void DataExportThread<Update, Reload>::run() {
    setThreadName(thread_name_);
    Notice(logger()) << "started";
    do {
        while (auto elem = queue_.pop(queue_pop_strategy::blocking, timeout_)) {
            try {
                std::visit([this](auto &&message) { this->handle(message); },
                           elem->first);
            } catch (const generic_error &ge) {
                Warning(logger()) << ge;
                close();
            }
        }
        on_idle_();
    } while (!stop_);
    close();
    Notice(logger()) << "stopped";
}

template <typename Update, typename Reload>
void DataExportThread<Update, Reload>::update(Update &&u) {
    auto [ok, size] =
        queue_.push(std::move(u), queue_overflow_strategy::dont_push);
    switch (ok) {
        case queue_status::ok:
            Debug(logger()) << "pushing update";
            break;
        case queue_status::overflow:
            counterIncrement(overflow_counter_);
            Debug(logger()) << "queue full, didn't push update";
            break;
        case queue_status::joinable:
            Debug(logger()) << "queue shutting down, didn't push update";
            break;
    }
    if (queue_.limit() && *queue_.limit() != 0) {
        counterSet(queue_usage_, static_cast<double>(size) /
                                     static_cast<double>(*queue_.limit()));
    }
}

template <typename Update, typename Reload>
void DataExportThread<Update, Reload>::reload(Reload &&r) {
    auto [ok, size] =
        queue_.push(std::move(r), queue_overflow_strategy::pop_oldest);
    switch (ok) {
        case queue_status::ok:
            Debug(logger()) << "pushing reload";
            break;
        case queue_status::overflow:
            Warning(logger())
                << "queue full when pushing reload, dropped oldest message";
            break;
        case queue_status::joinable:
            Debug(logger()) << "queue shutting down, didn't push reload";
            break;
    }
}
#endif
