// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef Downtime_h
#define Downtime_h

#include <chrono>
#include <cstdint>
#include <iosfwd>
#include <string>

#include "cmc/CycleTime.h"
class Core;
class Logger;
class DowntimeRule;
class Object;
class World;

class Downtime {
public:
    // queried in downtimes table
    enum class Status : uint32_t { active = 0, pending = 1 };

    enum class Treatment { keep, remove };

    Core &core_;
    Logger *const logger_downtime_;
    Logger *const logger_notification_;
    // modified in Downtimes::migrate
    const Object *_object;
    const uint32_t _id;
    const std::chrono::system_clock::time_point _entry_time;
    const std::string _author;
    const std::string _comment;
    // modified in Downtime::start and Downtime::end
    Status _status;
    // modified in Downtime::moveForward
    std::chrono::system_clock::time_point _start_time;
    // modified in Downtime::moveForward
    std::chrono::system_clock::time_point _end_time;
    // modified in Downtime::start
    std::chrono::system_clock::time_point _flex_downtime_start;
    const bool _fixed;
    const CycleTime _cycle_time;
    // NOTE, this is slightly confusing: For fixed downtimes, _duration contains
    // just the boring derived value (_end_time - _start_time), for flexible
    // downtimes it contains the duration after the actual start.
    const std::chrono::nanoseconds _duration;
    // modified in Downtime::examine, Downtime::start and Downtime::end
    bool _hot;  // true for flexible downtimes after _start_time
    const bool _origin_is_rule;
    // Pointer to the rule that originated this downtime, or null.
    // modified in Downtimes::handleChangedDowntimeRules
    const DowntimeRule *_rule;

    Treatment expire(std::chrono::system_clock::time_point now);
    void start(std::chrono::system_clock::time_point now);
    void cancel();
    bool migrate(const World &world);

    [[nodiscard]] bool operator==(const Downtime &other) const {
        return _object == other._object && _id == other._id &&
               _entry_time == other._entry_time && _author == other._author &&
               _comment == other._comment && _status == other._status &&
               _start_time == other._start_time &&
               _end_time == other._end_time &&
               _flex_downtime_start == other._flex_downtime_start &&
               _fixed == other._fixed && _cycle_time == other._cycle_time &&
               _duration == other._duration && _hot == other._hot &&
               _origin_is_rule == other._origin_is_rule && _rule == other._rule;
    }
    [[nodiscard]] bool operator!=(const Downtime &other) const {
        return !(*this == other);
    }

private:
    Treatment examine(std::chrono::system_clock::time_point now);
    void skip(const std::string &type, const std::string &why) const;
    void publish(const std::string &how, const std::string &type,
                 bool in_downtime);
    void end(const std::string &how);
    bool moveForward(std::chrono::system_clock::time_point now);
};

std::ostream &operator<<(std::ostream &os, const Downtime &d);

#endif  // Downtime_h
