// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef Entity_h
#define Entity_h

// Base class for all configuration entities that have attributes, like
// Services, Hosts and Contacts

#include <optional>
#include <string>
#include <unordered_map>

#include "livestatus/Interface.h"

// Things are a bit tricky regarding attributes in Nagios and CMC, so here is
// the full story:
//
// Nagios has quite a few builtin attributes, which are effectively fields in
// the corresponding host/service/contact structs. On the CMC side, we do not
// represent them as fields, but as entries in a map from attribute names to
// attribute values. All those builtin attributes do *not* have names starting
// with an underscore, and all of the are accessible via explicit Livestatus
// columns in the corresponding hosts/services/contacts table. Here is the full
// list of those attribute names:
//
// builtin host/service attributes:
//    * action_url
//    * action_url_expanded
//    * check_command_expanded
//    * display_name
//    * icon_image
//    * icon_image_alt
//    * icon_image_expanded
//    * notes
//    * notes_expanded
//    * notes_url
//    * notes_url_expanded
//
// builtin host attributes:
//    * address
//    * alias
//    * statusmap_image
//    * x_3d
//    * y_3d
//    * z_3d
//
// builtin contact attributes:
//    * address1
//    * address2
//    * address3
//    * address4
//    * address5
//    * address6
//    * alias
//    * email
//    * pager
//
// In addition to those builtin attributes, Nagios offers a mechanism to define
// custom attributes, which are distinguished from builtin ones by a leading
// underscore in the attribute name. In addition, the names of these custom
// attributes are implicitly converted to upper case. One can access these
// custom attributes via a field in the hosts/services/contacts structs which
// points to a linked list of name/value pairs (struct customvariablesmember).
// The names in these lists do *not* contain the leading underscore and are in
// upper case. The CMC gets all attributes, including custom ones, in a single
// bunch, but builtin ones can be distinguished from custom ones by the leading
// underscore. Normally these custom attributes are accessed in a generic way
// via Livestatus, with one exception: The "FILENAME" custom host attribute is
// mapped to the "filename" column in the hosts table.

class Entity {
public:
    explicit Entity(std::string name);
    Entity(std::string name,
           const std::unordered_map<std::string, std::string> &attributes);
    Entity(const Entity &) = default;
    Entity(Entity &&) noexcept = default;
    virtual ~Entity() = default;

    std::string name() const;

    // TODO(sp) Can we get rid of the optional somehow?
    std::string lookup(std::optional<AttributeKind> kind,
                       const std::string &name) const;

    // TODO(sp) Make this a view/range/... to avoid copying
    Attributes attributes(AttributeKind kind) const {
        return attributes_for(kind);
    }

    bool operator==(const Entity &other) const;
    bool operator!=(const Entity &other) const { return !(*this == other); }

private:
    std::string _name;
    Attributes _attributes;
    Attributes _custom_attributes;
    Attributes _tags;
    Attributes _labels;
    Attributes _label_sources;

    const Attributes &attributes_for(std::optional<AttributeKind> kind) const;
};

#endif  // Entity_h
