// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef FetcherHelper_h
#define FetcherHelper_h

#include <chrono>
#include <filesystem>
#include <functional>
#include <memory>
#include <optional>

#include "cmc/CheckHelper.h"
#include "cmc/FetcherHelperChannel.h"
#include "cmc/MessagePayload.h"

class FetcherReply;
class Logger;
class Poller;
class Service;

/** Interaction Between Core(), bin/fetcher and fetcher object.
    @startuml {core_fetcher}

    autonumber

    Core -> fetcher: send to stdout "serial;host;timeout"

    activate fetcher
    fetcher -> fetcher : read json file
    loop for every fetcher in json file
    fetcher -> fetcher_object : create and execute

    activate fetcher_object
    fetcher_object --> fetcher_object: read data from own source
    fetcher_object -> fetcher: send data
    deactivate fetcher_object
    fetcher -> fetcher : process data from the fetcher object
    end


    fetcher --> Core: write Blob to stdout
    fetcher --> Core: write Error Answers to stdout
    fetcher -> Core: write Ending Answer to stdout
    deactivate fetcher

    @enduml
*/
class FetcherHelper : public CommonHelper {
public:
    using reply_handler_t = std::function<void(std::unique_ptr<FetcherReply>)>;

    FetcherHelper(const std::filesystem::path &path, const args_type &args,
                  reply_handler_t reply_handler,
                  reschedule_handler_t<Service> reschedule_handler,
                  Logger *logger);
    ~FetcherHelper() override;

private:
    reply_handler_t reply_handler_;
    reschedule_handler_t<Service> reschedule_handler_;
    FetcherHelperChannel ch_;

    std::chrono::steady_clock::time_point fetcher_started_at_;

    std::optional<MessagePayload> payload_;

    // AbstractHelper/ProcessHelper API
    void start_() override;
    void stop_() override;
    void setUpPoller_(Poller &poller) const override;
    void communicate_(const Poller &poller) override;

    // CommonHelper API
    bool sendRequest() override;
    bool sendReloadConfigRequest() override;
    void incrementCounter() const override;

    void handleStdOut();
    void handleStdErr();

    fetcher::protocol::Request makeRequest();
    void do_stop();
    void enqueueCheckerRequest();

    enum class how_to_proceed { carry_on, stop };
    how_to_proceed handle(const fetcher::protocol::Incomplete &msg) const;
    how_to_proceed handle(const fetcher::protocol::Failure &msg);
    how_to_proceed handle(fetcher::protocol::Result &msg);
    how_to_proceed handle(const fetcher::protocol::Log &msg) const;
    how_to_proceed handle(const fetcher::protocol::EndOfReply &msg);
};

#endif  // FetcherHelper_h
