// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef GlobalConfig_h
#define GlobalConfig_h

#include <chrono>
#include <cstdint>
#include <filesystem>
#include <memory>
#include <optional>
#include <string>
#include <string_view>
#include <vector>

#include "cmc/CarbonConfig.h"
#include "cmc/InfluxDBConfig.h"
#include "cmc/SerialToken.h"

class Formatter;
enum class GroupAuthorization;
enum class LogLevel;
enum class ServiceAuthorization;

constexpr std::string_view helper_config_dir_name{"helper_config"};

class GeneralConfig {
public:
    config_serial_t serial_;
    std::filesystem::path core_directory_;
    std::filesystem::path log_file_;
    std::filesystem::path cmk_executable_;
    std::filesystem::path crash_reports_directory_;
    std::filesystem::path license_usage_history_file_;
    std::filesystem::path inventory_directory_;
    std::filesystem::path structured_status_directory_;
    std::filesystem::path robotmk_html_log_directory_;
    std::filesystem::path logwatch_directory_;
    std::filesystem::path prediction_directory_;
    std::filesystem::path event_console_status_socket_;
    std::filesystem::path state_file_created_file_;
    std::filesystem::path licensed_state_file_;
    LogLevel log_level_core_;
    LogLevel log_level_downtime_;
    bool log_microtime_;
    bool core_dumps_enabled_;
    std::chrono::microseconds houskeeping_interval_;
    std::chrono::seconds state_retention_interval_;
    bool event_console_enabled_;
    std::string edition_;
    uint32_t max_long_output_size_;
};

class SmartPingConfig {
public:
    std::chrono::microseconds check_interval_;
    bool omit_payload_;
    uint32_t num_sockets_;
    bool ignore_rst_;
    std::chrono::nanoseconds wait_duration_;
    uint32_t bulk_size_;
    LogLevel log_level_;
    std::filesystem::path icmp_sender_executable_;
    std::filesystem::path icmp_receiver_executable_;
};

class CheckingConfig {
public:
    std::filesystem::path generic_helper_executable_;
    uint32_t num_generic_helpers_;
    uint32_t num_real_time_helpers_;
    uint16_t real_time_port_;
    std::filesystem::path fetcher_executable_;
    uint32_t num_fetchers_;
    bool fetchers_enabled_;
    uint32_t num_checkers_;
    LogLevel log_level_;
    bool debug_;
    uint32_t initial_scheduling_burst_;
    std::chrono::nanoseconds initial_generic_scheduling_interval_;
    std::chrono::nanoseconds initial_checkmk_scheduling_interval_;
};

class LivestatusConfig {
public:
    std::filesystem::path socket_;
    std::filesystem::path command_pipe_;
    uint32_t max_response_size_;
    uint32_t stack_size_;
    uint32_t num_threads_;
    LogLevel log_level_;
    ServiceAuthorization service_authorization_;
    GroupAuthorization group_authorization_;
};

class NotificationHandlerConfig {
public:
    bool debug_;
    LogLevel log_level_;
};

// bits in AlertHandlerConfig::event_types_
enum class AlertHandlerEventType { statechange = 1, checkresult = 2 };

class AlertHandlerConfig {
public:
    bool enabled_;
    uint32_t event_types_;
    LogLevel log_level_;
};

class InfluxDbConfig {
public:
    std::vector<influxdb::ConnectionConfig> connections_;
    LogLevel log_level_;
    bool operator==(const InfluxDbConfig &other) const;
    bool operator!=(const InfluxDbConfig &other) const {
        return !(*this == other);
    }
};

class CarbonConfig {
public:
    std::vector<CarbonConnectionConfig> connections_;
    LogLevel log_level_;
    bool operator==(const CarbonConfig &other) const;
    bool operator!=(const CarbonConfig &other) const {
        return !(*this == other);
    }
};

enum class LogRotationInterval {
    minute = 0,
    hour = 1,
    day = 2,
    week = 3,
    never = 4,
};

class HistoryConfig {
public:
    uint32_t max_cached_messages_;
    uint32_t max_lines_per_file_;
    std::optional<std::chrono::seconds> cache_horizon_;
    std::chrono::seconds max_core_downtime_;
    std::filesystem::path file_;
    std::filesystem::path archive_directory_;
    LogRotationInterval rotation_interval_;
    uint32_t rotation_size_limit_;
    bool operator==(const HistoryConfig &other) const;
    bool operator!=(const HistoryConfig &other) const {
        return !(*this == other);
    }
};

class RRDConfig {
public:
    std::filesystem::path single_directory_;
    std::filesystem::path multiple_directory_;
    std::filesystem::path rrdcached_socket_;
    LogLevel log_level_;
    bool pnp4nagios_enabled_;
    std::chrono::seconds pnp_update_delay_;
    bool pnp_update_on_restart_;
    bool operator==(const RRDConfig &other) const;
    bool operator!=(const RRDConfig &other) const { return !(*this == other); }
};

class GlobalConfig {
public:
    GeneralConfig general_;
    SmartPingConfig smart_ping_;
    CheckingConfig checking_;
    LivestatusConfig livestatus_;
    NotificationHandlerConfig notification_handler_;
    AlertHandlerConfig alert_handler_;
    CarbonConfig carbon_;
    InfluxDbConfig influxdb_;
    HistoryConfig history_;
    RRDConfig rrd_;

    [[nodiscard]] std::unique_ptr<Formatter> getFormatter() const;
};

#endif  // GlobalConfig_h
