// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef GlobalState_h
#define GlobalState_h

#include <chrono>
#include <cstdint>
#include <mutex>

struct GlobalStateFields {
    // Trick to use structured bindings with ::GlobalState().
    bool _enable_notifications;
    bool _execute_service_checks;
    bool _execute_host_checks;
    bool _enable_flap_detection;
    bool _process_performance_data;
    bool _enable_alert_handlers;
    uint32_t _notification_next_problem_id;
    // Trial installation variable
    std::chrono::system_clock::time_point _state_file_created;
    [[nodiscard]] bool operator==(const GlobalStateFields &other) const;
    [[nodiscard]] bool operator!=(const GlobalStateFields &other) const {
        return !(*this == other);
    }
};

// This structure contains all data that must globally be persistent between
// core restarts and that is not contained the state of an object.
class GlobalState {
public:
    explicit GlobalState(GlobalStateFields fields) : fields_{fields} {}
    GlobalState(const GlobalState &other) {
        std::lock_guard lg_other{other.lock_};
        fields_ = other.fields_;
    }
    GlobalState &operator=(const GlobalState &other) {
        if (this != &other) {
            std::scoped_lock lg{other.lock_, lock_};
            fields_ = other.fields_;
        }
        return *this;
    }

    bool getEnableNotifications() const {
        std::lock_guard lg{lock_};
        return fields_._enable_notifications;
    }
    void setEnableNotifications(bool x) {
        std::lock_guard lg{lock_};
        fields_._enable_notifications = x;
    }

    bool getExecuteServiceChecks() const {
        std::lock_guard lg{lock_};
        return fields_._execute_service_checks;
    }
    void setExecuteServiceChecks(bool x) {
        std::lock_guard lg{lock_};
        fields_._execute_service_checks = x;
    }

    bool getExecuteHostChecks() const {
        std::lock_guard lg{lock_};
        return fields_._execute_host_checks;
    }
    void setExecuteHostChecks(bool x) {
        std::lock_guard lg{lock_};
        fields_._execute_host_checks = x;
    }

    bool getEnableFlapDetection() const {
        std::lock_guard lg{lock_};
        return fields_._enable_flap_detection;
    }
    void setEnableFlapDetection(bool x) {
        std::lock_guard lg{lock_};
        fields_._enable_flap_detection = x;
    }

    bool getProcessPerformanceData() const {
        std::lock_guard lg{lock_};
        return fields_._process_performance_data;
    }
    void setProcessPerformanceData(bool x) {
        std::lock_guard lg{lock_};
        fields_._process_performance_data = x;
    }

    bool getEnableAlertHandlers() const {
        std::lock_guard lg{lock_};
        return fields_._enable_alert_handlers;
    }
    void setEnableAlertHandlers(bool x) {
        std::lock_guard lg{lock_};
        fields_._enable_alert_handlers = x;
    }

    uint32_t getNotificationNextProblemId() const {
        std::lock_guard lg{lock_};
        return fields_._notification_next_problem_id;
    }
    uint32_t increaseNotificationNextProblemId() {
        std::lock_guard lg{lock_};
        return fields_._notification_next_problem_id++;
    }
    void setNotificationNextProblemId(uint32_t x) {
        std::lock_guard lg{lock_};
        fields_._notification_next_problem_id = x;
    }

    std::chrono::system_clock::time_point getStateFileCreated() const {
        std::lock_guard lg{lock_};
        return fields_._state_file_created;
    }
    void setStateFileCreated(std::chrono::system_clock::time_point x) {
        std::lock_guard lg{lock_};
        fields_._state_file_created = x;
    }

    [[nodiscard]] bool operator==(const GlobalState &other) const {
        std::lock_guard lg{lock_};
        return fields_ == other.fields_;
    }
    [[nodiscard]] bool operator!=(const GlobalState &other) const {
        std::lock_guard lg{lock_};
        return !(*this == other);
    }

private:
    GlobalStateFields fields_;
    mutable std::mutex lock_;
};

#endif  // GlobalState_h
