// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

// NOTE: If we only use a forward declaration for Service below, things still
// compile with both GCC and Clang, but IWYU fails with a Clang error. Weird...
#ifndef Host_h
#define Host_h

#include <netinet/in.h>

#include <chrono>
#include <functional>
#include <memory>
#include <optional>
#include <string>
#include <unordered_map>
#include <unordered_set>
#include <utility>

#include "cmc/CmcHost.h"
#include "cmc/MessagePayload.h"
#include "cmc/Object.h"
#include "cmc/Service.h"  // IWYU pragma: keep
#include "cmc/State.h"
#include "livestatus/LogEntry.h"
#include "livestatus/PnpUtils.h"

class Core;
class Entity;
class IHost;
class HostGroup;
class Timeperiod;
class World;

class Host : public Object {
public:
    enum class Reachability { reachable, unreachable, unknown };
    using services_t = std::unordered_set<std::unique_ptr<Service>>;

    Host(Core &core, World &world, const std::string &name,
         const Timeperiod &tp);
    Host(Core &core, World &world, Entity &&entity);
    // disallow copy and /implicitly/move: copy is forbidden, move isn't used
    Host(const Host &) = delete;
    Host &operator=(const Host &) = delete;

    const IHost &ihost() const { return ihost_; }
    const std::unordered_set<HostGroup *> &groups() const { return _groups; };
    HostState remap(HostState state) const;
    Service *getService(const std::string &name);
    void addParent(Host *p) { _parents.insert(p); }
    void addChild(Host *c) { _children.insert(c); }
    bool hasParents() const { return !_parents.empty(); }
    bool isHost() const override { return true; };
    const Host *host() const override { return this; };
    Host *host() override { return this; };
    const Service *service() const override { return nullptr; };
    Service *service() override { return nullptr; };
    std::string serviceDescription() const override {
        return dummy_service_description();
    };
    std::chrono::system_clock::duration staleness_check_interval()
        const override;

    void updateNetworkOutages(std::chrono::system_clock::time_point now,
                              const result_handler_t &result_handler);
    Reachability getReachability() const;
    void setUp(std::chrono::system_clock::time_point now,
               const result_handler_t &result_handler);
    void setServiceDoesNotExist(std::chrono::system_clock::time_point now,
                                const result_handler_t &result_handler);
    void updateReachability(std::chrono::system_clock::time_point now,
                            const result_handler_t &result_handler);
    std::pair<std::string, std::string> serviceFromRouting() const;

    HostState getCurrentState() const {
        return static_cast<HostState>(state()->current_state_);
    }

    HostState getLastState() const {
        return static_cast<HostState>(state()->last_state_);
    }

    bool isCurrentStateOK() const override {
        return getCurrentState() == HostState::up;
    }

    bool isPreviousHardStateOK() const override {
        return static_cast<HostState>(state()->hard_state_.previous_) ==
               HostState::up;
    }

    void insertInGroup(HostGroup &group);

    bool operator==(const Host &other) const;
    bool operator!=(const Host &other) const { return !(*this == other); }

    class Services {
    public:
        explicit Services(const Host &host) : _services(host._services) {}
        auto begin() { return _services.begin(); }
        auto end() { return _services.end(); }
        [[nodiscard]] auto cbegin() const { return _services.cbegin(); }
        [[nodiscard]] auto cend() const { return _services.cend(); }

    private:
        const services_t &_services;
    };

    // TODO(sk): get rid of those pesky public data members.
    std::unordered_map<HostState, HostState> _state_mapping;
    std::chrono::nanoseconds _smart_ping_timeout;
    in_addr_t _ipv4_address;
    services_t _services;

    std::unordered_set<Host *> _children;
    std::unordered_set<Host *> _parents;

    // Speed up for services
    std::unordered_map<std::string, Service *> _services_by_name;

    std::optional<MessagePayload> _real_time_payload;

private:
    CmcHost ihost_;
    std::unordered_set<HostGroup *> _groups;
};

#endif  // Host_h
