// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef InfluxDBConfig_h
#define InfluxDBConfig_h

#include <chrono>
#include <cstddef>
#include <cstdint>
#include <filesystem>
#include <iosfwd>
#include <map>
#include <string>
#include <utility>

namespace influxdb {
struct Token {
    explicit Token(std::string token) : token_{std::move(token)} {}
    [[nodiscard]] explicit operator std::string() const;
    bool operator==(const influxdb::Token &other) const;
    bool operator!=(const influxdb::Token &other) const {
        return !(*this == other);
    }

private:
    const std::string token_;
};

struct Headers {
    Token token;
    std::string org;
    std::multimap<std::string, std::string> extra;
    [[nodiscard]] explicit operator std::string() const;
};

struct Address {
    const std::string protocol;  // one of http or https
    const std::string host;
    const uint16_t port;
    [[nodiscard]] std::string to_string() const;
    bool operator==(const influxdb::Address &other) const;
    bool operator!=(const influxdb::Address &other) const {
        return !(*this == other);
    }
};

struct Auth {
    // We may need other authentication strategies.
    const std::filesystem::path ca_cert_file;
    const bool use_certificate_check;
    bool operator==(const influxdb::Auth &other) const;
    bool operator!=(const influxdb::Auth &other) const {
        return !(*this == other);
    }
};

/// \sa core_cmc::cmc_influxdb_config()
struct ConnectionConfig {
    std::string id;
    bool enabled;
    influxdb::Address address;
    influxdb::Auth auth;
    std::string api_path;
    influxdb::Token token;
    std::size_t max_batch_size;
    std::chrono::seconds max_batch_age;
    std::chrono::nanoseconds retry_interval;
    bool operator==(const influxdb::ConnectionConfig &other) const;
    bool operator!=(const influxdb::ConnectionConfig &other) const {
        return !(*this == other);
    }
};

namespace protocol {
struct Update;
}  // namespace protocol

/// \sa CMCHostConfigBase::_cmc_influxdb_sender_config()
struct SenderMetrics {
    bool metric_value;
    bool metric_hst_state;
    bool metric_svc_state;
    bool add_hst_state;
    bool add_svc_state;
    bool add_downtime;
    bool add_acknowledgement;
    bool add_uom;
    bool add_thresholds;
    bool add_min_max;
    bool tag_hst_name;
    bool tag_svc_name;
    bool tag_hst_alias;
    bool tag_hst_tags;
    bool tag_hst_label;
    bool tag_svc_label;

    [[nodiscard]] bool anyState() const {
        return metric_hst_state || metric_svc_state;
    }

    [[nodiscard]] bool anyMetric() const { return metric_value || anyState(); }

    bool operator==(const SenderMetrics &other) const;
    bool operator!=(const SenderMetrics &other) const {
        return !(*this == other);
    }

private:
    [[nodiscard]] bool anyAdditional() const {
        return add_uom || add_thresholds || add_min_max || add_downtime ||
               add_acknowledgement || add_hst_state || add_svc_state;
    }
    [[nodiscard]] bool anyTag() const {
        return tag_hst_name || tag_svc_name || tag_hst_alias || tag_hst_tags ||
               tag_hst_label || tag_svc_label;
    }
    friend struct protocol::Update;
};

/// \sa CMCHostConfigBase::_cmc_influxdb_sender_config()
struct SenderConfig {
    std::string connection;
    std::string org;
    std::string bucket;
    SenderMetrics parts;
    static SenderConfig make();
    bool operator==(const SenderConfig &other) const;
    bool operator!=(const SenderConfig &other) const {
        return !(*this == other);
    }
};
}  // namespace influxdb

std::ostream &operator<<(std::ostream &, const influxdb::Token &);
std::ostream &operator<<(std::ostream &, const influxdb::Address &);
std::ostream &operator<<(std::ostream &, const influxdb::Headers &);

#endif
