// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef InfluxDBConnection_h
#define InfluxDBConnection_h

#include <chrono>
#include <cstddef>
#include <functional>
#include <iosfwd>
#include <memory>
#include <ostream>
#include <string>
#include <unordered_map>
#include <utility>
#include <vector>

#include "cmc/InfluxDBConfig.h"
#include "cmc/InfluxDBSender.h"
#include "livestatus/Logger.h"

enum class Counter;
namespace influxdb {
struct LineProtocol;
}  // namespace influxdb
namespace influxdb::protocol {
struct Update;
}  // namespace influxdb::protocol

namespace influxdb {
namespace detail {
struct SenderConfigHash {
    std::size_t operator()(
        const influxdb::SenderConfig &sender_config) const noexcept {
        auto h1 = std::hash<std::string>{}(sender_config.connection);
        auto h2 = std::hash<std::string>{}(sender_config.org);
        auto h3 = std::hash<std::string>{}(sender_config.bucket);
        // If SenderMetrics is relevant, then we need a rule ID, these are
        // too many parameters that may change in the future.
        return h1 ^ (h2 << 1) ^ (h3 << 2);
    }
};
}  // namespace detail

class Connection {
public:
    enum class State { unconnected, connection_in_progress, connected };

    explicit Connection(const influxdb::ConnectionConfig &config,
                        Counter bytes_sent_counter);
    ~Connection() { close(); }
    // We use dynamically allocated resources and our
    // ContextLogger captures 'this' => define move constructor
    Connection(Connection &&cc) noexcept
        : _config(cc._config)
        , _bytes_sent_counter(cc._bytes_sent_counter)
        , _state(cc._state)
        , _last_connection_attempt(cc._last_connection_attempt)
        , _log_prefix(std::move(cc._log_prefix)) {
        cc._state = State::unconnected;
    }

    State tryToInitiateConnection(
        std::chrono::steady_clock::time_point time_now);
    State tryToCompleteConnection(
        std::chrono::steady_clock::time_point time_now);
    void doHousekeeping();
    void send(const influxdb::protocol::Update &message);
    [[nodiscard]] Logger *logger() const { return &_logger; }

private:
    const influxdb::ConnectionConfig _config;
    std::unique_ptr<influxdb::Sender> _sender;
    Counter _bytes_sent_counter;
    State _state;
    std::unordered_map<influxdb::SenderConfig,
                       std::pair<std::chrono::steady_clock::time_point,
                                 std::vector<LineProtocol> >,
                       detail::SenderConfigHash>
        _batch;
    std::chrono::steady_clock::time_point _last_connection_attempt;
    std::string _log_prefix;
    mutable ContextLogger _logger{
        Logger::getLogger("cmk.influxdb"),
        [this](std::ostream &os) { os << _log_prefix; }};
    std::size_t batch(const influxdb::SenderConfig &sender_config,
                      const std::vector<LineProtocol> &new_lines,
                      std::chrono::steady_clock::time_point now);
    std::size_t sendBatch(std::chrono::steady_clock::time_point now);
    void close();
};
}  // namespace influxdb

#endif  // InfluxDBConnection_h
