// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef InfluxDBLineProtocol_h
#define InfluxDBLineProtocol_h

#include <chrono>
#include <cstdint>
#include <iosfwd>
#include <string>
#include <type_traits>
#include <utility>
#include <variant>
#include <vector>

namespace influxdb {
namespace detail {
struct KeyString {
    // We do want the converting constructor for this private class:
    // NOLINTNEXTLINE(google-explicit-constructor,hicpp-explicit-conversions)
    KeyString(const std::string &s) : value{sanitize(s)} {}
    // NOLINTNEXTLINE(google-explicit-constructor,hicpp-explicit-conversions)
    operator std::string() const;
    std::string value;

private:
    static constexpr auto escape_sequence = ",= ";
    // TODO(sp) Figure out what we actually want to do here.
    static std::string sanitize(const std::string &s) { return s; }
};

struct TagString {
    // We do want the converting constructor for this private class:
    // NOLINTNEXTLINE(google-explicit-constructor,hicpp-explicit-conversions)
    TagString(const std::string &s) : value{sanitize(s)} {}
    // NOLINTNEXTLINE(google-explicit-constructor,hicpp-explicit-conversions)
    operator std::string() const;
    std::string value;

private:
    static constexpr auto escape_sequence = ",= ";
    // TODO(sp) Figure out what we actually want to do here.
    static std::string sanitize(const std::string &s) { return s; }
};

struct FieldString {
    // We do want the converting constructor for this private class:
    // NOLINTNEXTLINE(google-explicit-constructor,hicpp-explicit-conversions)
    FieldString(const std::string &s) : value{sanitize(s)} {}
    // NOLINTNEXTLINE(google-explicit-constructor,hicpp-explicit-conversions)
    operator std::string() const;
    std::string value;

private:
    static constexpr auto escape_sequence = "\"\\";
    // TODO(sp) Figure out what we actually want to do here.
    static std::string sanitize(const std::string &s) { return s; }
};
}  // namespace detail

struct Measurement {
    explicit Measurement(const std::string &m) : value{sanitize(m)} {}
    explicit operator std::string() const;
    std::string value;

private:
    static constexpr auto escape_sequence = ", ";
    // TODO(sp) Figure out what we actually want to do here.
    static std::string sanitize(const std::string &s) { return s; }
};

struct Tag {
    using key_type = detail::KeyString;
    using value_type = detail::TagString;
    Tag(key_type k, value_type v) : key{std::move(k)}, value{std::move(v)} {}
    key_type key;
    value_type value;
};

struct Field {
    using key_type = detail::KeyString;
    using value_type = std::variant<double, std::int64_t, std::uint64_t,
                                    detail::FieldString, bool>;
    Field(key_type k, value_type v) : key{std::move(k)}, value{std::move(v)} {}
    key_type key;
    value_type value;
};

namespace detail {
template <typename Resolution>
constexpr auto GetTimestampSuffix() noexcept {
    if constexpr (std::is_same<std::chrono::seconds, Resolution>::value) {
        return "s";
    }
    if constexpr (std::is_same<std::chrono::milliseconds, Resolution>::value) {
        return "ms";
    }
    if constexpr (std::is_same<std::chrono::microseconds, Resolution>::value) {
        return "us";
    }
    if constexpr (std::is_same<std::chrono::nanoseconds, Resolution>::value) {
        return "ns";
    }
    // Other resolutions are invalid.
}
}  // namespace detail

/// \brief Line protocol serialization
///
/// \see <a href="Line protocol">
///      https://docs.influxdata.com/influxdb/v2.0/reference/syntax/line-protocol/
///      </a>
struct LineProtocol {
    using resolution = std::chrono::seconds;

    LineProtocol(Measurement measurement, std::vector<Tag> tags,
                 const std::vector<Field> &fields,
                 const std::chrono::system_clock::time_point &timestamp)
        : measurement{std::move(measurement)}
        , tags{std::move(tags)}
        , fields{sanitize(fields)}
        , timestamp{timestamp} {}

    Measurement measurement;
    std::vector<Tag> tags;
    std::vector<Field> fields;
    std::chrono::system_clock::time_point timestamp;

    static constexpr auto TimestampSuffix =
        detail::GetTimestampSuffix<resolution>();

private:
    // TODO(sp) Figure out what we actually want to do here.
    static std::vector<Field> sanitize(const std::vector<Field> &fields) {
        return fields;
    }
};

std::ostream &operator<<(std::ostream &, const detail::KeyString &);
std::ostream &operator<<(std::ostream &, const detail::TagString &);
std::ostream &operator<<(std::ostream &, const detail::FieldString &);
std::ostream &operator<<(std::ostream &, const Measurement &);
std::ostream &operator<<(std::ostream &, const Tag &);
std::ostream &operator<<(std::ostream &, const Field &);
std::ostream &operator<<(std::ostream &, const LineProtocol &);
}  // namespace influxdb

#endif
