// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <cstdint>
#include <optional>
#include <string>
#include <utility>
#include <variant>

#include "livestatus/Metric.h"

namespace influxdb {

class Metric {
public:
    using Name = ::Metric::Name;

    /// The value, min or max field.
    ///
    /// Value in the class [-0-9.] or literal "U" for unknown.
    ///
    /// \sa https://nagios-plugins.org/doc/guidelines.html#AEN200
    class Value {
    public:
        using value_type = std::variant<double, std::int64_t>;

        explicit Value(const value_type &v) : v_{v} {};
        Value() = default;
        [[nodiscard]] static Value make(const std::string &serialized);
        [[nodiscard]] const value_type &operator*() const { return *v_; }
        [[nodiscard]] bool empty() const { return !v_; }

    private:
        std::optional<value_type> v_;
    };

    /// The crit and warn field.
    ///
    /// \sa https://nagios-plugins.org/doc/guidelines.html#THRESHOLDFORMAT
    class Range {
    public:
        using bound_type = std::variant<double, std::int64_t>;
        struct value_type {
            bound_type lower_bound;
            bound_type higher_bound;
        };

        Range(const bound_type &lower_bound, const bound_type &higher_bound,
              bool invert)
            : bounds_{{lower_bound, higher_bound}}, invert_{invert} {}
        Range() : invert_{false} {}
        [[nodiscard]] static Range make(const std::string &serialized);
        [[nodiscard]] std::optional<value_type> bounds() const {
            return bounds_;
        }
        [[nodiscard]] bool inverted() const { return invert_; }
        [[nodiscard]] bool contains(const bound_type &value) const;

    private:
        std::optional<value_type> bounds_;
        bool invert_;
    };
    Metric(Name label, Value value, std::string uom, Range warn, Range crit,
           Value min, Value max)
        : name_{std::move(label)}
        , value_{value}
        , uom_{std::move(uom)}
        , warn_{warn}
        , crit_{crit}
        , min_{min}
        , max_{max} {}
    [[nodiscard]] static Metric make(const ::Metric &);
    [[nodiscard]] Name name() const { return name_; }
    [[nodiscard]] Value value() const { return value_; }
    [[nodiscard]] std::string uom() const { return uom_; }
    [[nodiscard]] Range warn() const { return warn_; }
    [[nodiscard]] Range crit() const { return crit_; }
    [[nodiscard]] Value min() const { return min_; }
    [[nodiscard]] Value max() const { return max_; }

private:
    Name name_;
    Value value_;
    std::string uom_;
    Range warn_;
    Range crit_;
    Value min_;
    Value max_;
};

}  // namespace influxdb
