// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef Object_h
#define Object_h

#include <chrono>
#include <cstdint>
#include <functional>
#include <iosfwd>
#include <memory>
#include <string>
#include <unordered_set>
#include <vector>

#include "cmc/CarbonConfig.h"
#include "cmc/DowntimeRule.h"
#include "cmc/Entity.h"
#include "cmc/InfluxDBConfig.h"
#include "cmc/RRDInfo.h"
#include "cmc/State.h"
#include "livestatus/BitMask.h"
#include "livestatus/Metric.h"

class CheckResult;
enum class CheckResultKind;
class ContactList;
class Core;
class Host;
class Logger;
class RegExp;
class Service;
class Timeperiod;
class World;

// see notification_option_bitmask in enterprise/cmk_base/cee/core_cmc.py
enum class NotificationOptions : uint32_t {
    host_up = 1 << 0,
    host_down = 1 << 1,
    host_unreachable = 1 << 2,
    host_flapping = 1 << 4,
    host_downtime = 1 << 5,
    service_ok = 1 << 8,
    service_warning = 1 << 9,
    service_critical = 1 << 10,
    service_unknown = 1 << 11,
    service_flapping = 1 << 12,
    service_downtime = 1 << 13
};
IS_BIT_MASK(NotificationOptions)

// How to write the state of an object into the monitoring history.
enum class LogStateType { initial, current };

class Replacer {
public:
    Replacer(std::shared_ptr<RegExp> regexp, std::string replacement);
    [[nodiscard]] std::string replace(const std::string &str) const;

private:
    const std::shared_ptr<RegExp> _regex;
    const std::string _replacement;
};

// Base class for monitored objects (hosts and services)
class Object : public Entity {
public:
    using result_handler_t = std::function<void(
        std::chrono::system_clock::time_point, const CheckResult &)>;

    Object(Core &core, World &world, Entity &&entity);
    Object(Core &core, World &world, const std::string &name,
           const Timeperiod &tp);
    ~Object() override;

    DowntimeRule *find_prd(const DowntimeRule &other) const;
    std::string specname() const;
    std::string statename() const;
    std::string lastStatename() const;
    std::string previousHardStatename() const;
    const char *stateTypeName() const;
    const State *state() const { return _state; }
    // TODO(sp) Nuke this! All modifications should go through Object!
    State *stateNonConst() const { return _state; }
    void setState(State *s) { _state = s; }

    bool isEnabled(State::Enabling::Key what) const;
    void setEnabled(State::Enabling::Key what, bool enabled) const;
    void resetEnablings() const;

    bool isCheckEnabled(CheckResultKind kind) const;
    bool canBeScheduled() const {
        return !_is_shadow && !_check_command_line.empty();
    };
    bool isShadow() const { return _is_shadow; };
    void createNewState(std::chrono::system_clock::time_point now,
                        Logger *logger);
    void sanitizeState(Logger *logger);
    std::chrono::system_clock::time_point nextCheckTime() const;
    std::chrono::system_clock::time_point nextImmediateCheckTime() const;
    std::chrono::system_clock::time_point nextCheck() const {
        return state()->next_check_;
    }
    void setNextCheck(std::chrono::system_clock::time_point,
                      Logger *logger) const;
    void checkStarts(std::chrono::system_clock::time_point when) const {
        stateNonConst()->last_check_started_ = when;
    }

    void checkHelperStarts(std::chrono::system_clock::time_point when,
                           std::chrono::duration<double> job_latency) const {
        stateNonConst()->latency_ = job_latency;
        stateNonConst()->last_check_started_ = when;
        stateNonConst()->is_helper_running_ = true;
    }

    void checkHelperStops() const {
        stateNonConst()->is_helper_running_ = false;
    }

    virtual bool isHost() const = 0;
    virtual const Host *host() const = 0;
    virtual Host *host() = 0;
    virtual const Service *service() const = 0;
    virtual Service *service() = 0;
    virtual std::string serviceDescription() const = 0;
    virtual std::chrono::system_clock::duration staleness_check_interval()
        const = 0;
    bool isService() const { return !isHost(); };
    const Timeperiod *notificationPeriod() const {
        return _notification_period;
    };
    const Timeperiod *servicePeriod() const { return _service_period; };
    void notifyHardStateChange(std::chrono::system_clock::time_point now,
                               Logger *logger);
    void checkFreshness(std::chrono::system_clock::time_point now,
                        const result_handler_t &result_handler, Logger *logger);
    void checkPendingNotification(std::chrono::system_clock::time_point now,
                                  Logger *logger);
    bool acknowledged() const;
    AcknowledgementType acknowledgementType() const;
    void setAcknowledgementType(AcknowledgementType value) const;
    bool inDowntime() const;
    bool handled() const;
    bool notificationsEnabled(Logger *logger) const;
    bool inNotificationPeriod() const;
    NotificationOptions notificationBitmask() const;
    bool hasContact(const std::string &contact_name) const;
    void flap(std::chrono::system_clock::time_point now, Logger *logger);
    void unflap(std::chrono::system_clock::time_point now, Logger *logger);
    bool stateChanged() const { return state()->stateChanged(); }

    void clearEnabledPerConfig() { _enabling = 0; }
    bool getEnabledPerConfig(State::Enabling::Key what) const {
        return (_enabling & (1U << static_cast<int>(what))) != 0;
    }
    void setEnabledPerConfig(State::Enabling::Key what, bool enabled) {
        if (enabled) {
            _enabling |= 1U << static_cast<int>(what);
        } else {
            _enabling &= ~(1U << static_cast<int>(what));
        }
    }
    void addOutputMappings(const Replacer &replacer);
    uint32_t maxLongOutputSize() const;

    virtual bool isCurrentStateOK() const = 0;
    virtual bool isPreviousHardStateOK() const = 0;
    std::string remapOutput(std::string output);
    void logDowntimes(Logger *logger) const;

    RRDFormat currentRrdFormat() const;
    RRDInfo rrdInfo() const;
    MetricLocation metricLocation(const Metric::MangledName &name) const;
    void forgetRRDInfo() const;

    bool operator==(const Object &other) const;
    bool operator!=(const Object &other) const { return !(*this == other); }

    World *world() const { return &world_; }

    // -------------------------------------------------------------------------
    // Slightly confusing command handling below, slightly incomatible with
    // Nagios...
    // -------------------------------------------------------------------------
    // _check_command_name field:
    //
    // * CMC basically doesn't care about it, just passes it on:
    //   * Puts it into notification/alert contexts as HOSTCHECKCOMMAND or
    //     SERVICECHECKCOMMAND
    //   * Puts it into PNP XML files for multiple RRDs
    //   * Livestatus: "check_command" column in hosts/services
    // * Forms for:
    //   * Check_MK services: "check_mk-"+checkname, e.g. "Number of threads:
    //    "check_mk-cpu.threads"
    //   * Check_MK service itself: "check-mk"
    //   * active checks: "check_mk_active-" + check_type + "!" + args, e.g.
    //     HTTP check: "check_mk_active-http!'-I' '$_HOSTADDRESS_4$'"
    //   * legacy-checks: command_name + "!" + command_line, e.g.
    //     "check-mk-custom!echo "$SERVICESTATEID$"" although replacements
    //     may have already happened at config-generation time.
    // -------------------------------------------------------------------------
    // check_command_expanded attribute in Entity::_attributes:
    //
    // * _check_command_name w/ a few macros expanded at config generation time
    // * Livestatus: "check_command_expanded" column in hosts/services
    // -------------------------------------------------------------------------
    // _check_command_line field:
    //
    // * empty for passive checks, non-empty for active checks
    // * contains routing info:
    //   * "@cmk " prefix: executed by Check_MK helper pool (doesn't see prefix)
    //   * "@smartping": put into smartping scheduler
    //   * "@up": put into always_up scheduler
    //   * "@parents": put into cluster_host scheduler
    //   * everything not starting with "@": executed by generic helper pool
    // -------------------------------------------------------------------------
    std::string _check_command_name;
    std::string _check_command_line;

    bool _is_shadow;  // shadow objects are created by cmcdump
    bool _write_long_output;
    int _check_type;  // 0: active, 1: passive, 2: shadow make Livestatus easier
    std::chrono::nanoseconds _check_interval;
    std::chrono::nanoseconds _retry_interval;
    std::chrono::nanoseconds _notification_interval;
    std::chrono::nanoseconds _first_notification_delay;
    uint32_t _max_check_attempts;
    const Timeperiod *_check_period;
    const Timeperiod *_service_period;
    const Timeperiod *_notification_period;
    NotificationOptions _notification_options;
    std::chrono::seconds _check_timeout;
    uint32_t _freshness_state;      // state in cased of unfresh check
    std::string _freshness_output;  // faked check output for unfresh check
    double _low_flap_threshold;
    double _high_flap_threshold;
    double _flap_decay;
    RRDFormat _rrd_format;
    CarbonSendMetricParts _carbon_send_metric_parts;
    influxdb::SenderConfig _influxdb_send_config;

    std::vector<std::unique_ptr<DowntimeRule>> _downtime_rules;

    // Information about contacts and contact groups
    ContactList *_contact_list;

    // temporarily shared between old world and new world
    State *_state;

    // used for implementing host checks via service status
    std::unordered_set<Object *> _check_result_receivers;

private:
    void setupNotification(uint32_t problem_id, NotificationKind kind,
                           std::chrono::system_clock::time_point now,
                           Logger *logger);
    void cancelNotification(const std::string &message, Logger *logger) const;
    bool periodicNotificationsEnabled() const;
    void postponeNotification(const std::string &why, Logger *logger) const;
    std::string statenameOf(int stateIndex) const;

    uint32_t _enabling;  // TODO(sp) Use std::bitset
    Core &core_;
    World &world_;
    std::vector<Replacer> _output_mappings;
};

double staleness(const Object &object);
std::ostream &operator<<(std::ostream &os, const Object &object);

#endif  // Object_h
