// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef ObjectGroup_h
#define ObjectGroup_h

#include <cstddef>
#include <string>
#include <unordered_set>
#include <utility>

#include "cmc/CmcHostGroup.h"
#include "cmc/CmcServiceGroup.h"
#include "cmc/Host.h"
#include "cmc/Service.h"

template <typename T>
class ObjectGroup {
public:
    using values_type = std::unordered_set<T *>;

    ObjectGroup(std::string name, std::string alias)
        : _name(std::move(name)), _alias(std::move(alias)) {}

    // disallow copy and /implicitly/move: copy is forbidden, move isn't used
    ObjectGroup(const ObjectGroup &rhs) = delete;
    ObjectGroup &operator=(const ObjectGroup &rhs) = delete;

    void setAlias(std::string alias) { _alias = std::move(alias); }
    [[nodiscard]] std::string name() const { return _name; }
    [[nodiscard]] std::string alias() const { return _alias; }
    void insert(T *object) { _objects.insert(object); }
    [[nodiscard]] size_t size() const { return _objects.size(); }
    auto begin() const { return _objects.cbegin(); }
    auto end() const { return _objects.cend(); }
    bool operator==(const ObjectGroup<T> &other) const {
        return name() == other.name() && alias() == other.alias() &&
               std::is_permutation(begin(), end(), other.begin(), other.end(),
                                   [&](const auto *l, const auto *r) {
                                       return l->name() == r->name();
                                   });
    }
    bool operator!=(const ObjectGroup<T> &other) const {
        return !(*this == other);
    }

private:
    std::string _name;
    std::string _alias;
    values_type _objects;

    friend class TableHostGroups;     // for "members" and "members_with_state"
    friend class TableServiceGroups;  // for "members" and "members_with_state"
};

class HostGroup : public ObjectGroup<Host> {
public:
    HostGroup(std::string name, std::string alias)
        : ObjectGroup<Host>(std::move(name), std::move(alias))
        , ihostgroup_(*this) {}

    const IHostGroup &ihostgroup() const { return ihostgroup_; }

private:
    CmcHostGroup ihostgroup_;
};

class ServiceGroup : public ObjectGroup<Service> {
public:
    ServiceGroup(std::string name, std::string alias)
        : ObjectGroup<Service>(std::move(name), std::move(alias))
        , iservicegroup_(*this) {}

    const IServiceGroup &iservicegroup() const { return iservicegroup_; }

private:
    CmcServiceGroup iservicegroup_;
};

#endif  // ObjectGroup_h
