// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef RRDCacheDThread_h
#define RRDCacheDThread_h

#include <sys/types.h>

#include <chrono>
#include <cstdio>
#include <filesystem>
#include <memory>
#include <mutex>
#include <string>
#include <variant>
#include <vector>

#include "cmc/DataExportThread.h"
#include "livestatus/Metric.h"

enum class Counter;
class Logger;
class RRDUDSSocket;

namespace rrdcached::protocol {

class Update {
public:
    Update(std::string rrd_path, std::chrono::system_clock::time_point,
           std::vector<Metric>);
    [[nodiscard]] std::string string() const;
    static bool isSendable() { return true; }

private:
    std::string rrd_path_;
    std::chrono::system_clock::time_point timestamp_;
    std::vector<Metric> metrics_;
};

class Reload {
public:
    // NOTE: We actually want a std::filesystem::path parameter here, but
    // GCC 10.0 and 10.1 suffer from a specification bug for <iterator>, see
    // https://cplusplus.github.io/LWG/issue3420.
    explicit Reload(const std::string &socket_path)
        : socket_path_{socket_path} {}
    [[nodiscard]] std::filesystem::path socket_path() const {
        return socket_path_;
    }

private:
    std::filesystem::path socket_path_;
};

using Message = std::variant<Update, Reload>;

};  // namespace rrdcached::protocol

class RRDCacheDThread : public DataExportThread<rrdcached::protocol::Update,
                                                rrdcached::protocol::Reload> {
public:
    RRDCacheDThread(std::chrono::microseconds retry_interval,
                    Counter overflow_counter, Counter queue_usage,
                    Counter bytes_sent_counter, Logger *logger);
    ~RRDCacheDThread() override { close_impl(); };

private:
    // This mutex protects socket_path_.
    mutable std::mutex socket_path_mutex_;
    std::filesystem::path socket_path_;

    Counter bytes_sent_counter_;
    const std::chrono::microseconds retry_interval_;
    std::chrono::steady_clock::time_point last_connection_attempt_;
    std::unique_ptr<RRDUDSSocket> socket_{};

    std::filesystem::path socket_path() const;
    void socket_path(const std::filesystem::path &value);

    void handle(const rrdcached::protocol::Update &message) override;
    void handle(const rrdcached::protocol::Reload &message) override;
    void close() override { close_impl(); }
    void close_impl() { socket_.reset(); }
    void tryReconnect();
    void receiveReply();
};

class RRDUDSSocket {
public:
    RRDUDSSocket(std::filesystem::path path, Logger *logger);
    ~RRDUDSSocket();
    void connect();
    [[nodiscard]] Logger *logger() const { return logger_; }
    [[nodiscard]] std::string readLine() const;
    [[nodiscard]] ssize_t write(const std::string &text,
                                std::chrono::milliseconds timeout) const;
    [[nodiscard]] bool isConnected() const { return socket_ != -1; }
    void close();

private:
    std::filesystem::path path_;
    mutable Logger *logger_;
    int socket_{-1};
    FILE *file_{nullptr};
};

#endif  // RRDCacheDThread_h
