// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef Schedule_h
#define Schedule_h

#include <algorithm>
#include <cassert>
#include <chrono>
#include <iosfwd>
#include <map>
#include <optional>
#include <ostream>
#include <string>
#include <utility>

#include "livestatus/Logger.h"

// TODO(sp) We need all this template Kung Fu *only* for testing, the CMC itself
// always uses it with Object*. Reconsider this templatization.
template <typename T>
class Schedule {
    // We keep maps in both directions to make rescheduling efficient.
    using kv_type =
        typename std::multimap<std::chrono::system_clock::time_point, T>;
    using vk_type = std::multimap<typename kv_type::mapped_type,
                                  typename kv_type::key_type>;

public:
    using key_type = typename kv_type::key_type;
    using mapped_type = typename kv_type::mapped_type;
    using value_type = typename kv_type::value_type;
    using size_type = typename kv_type::size_type;

    struct const_iterator : public kv_type::const_iterator {
        const_iterator() : Schedule<T>::kv_type::const_iterator() {}
        explicit const_iterator(typename kv_type::const_iterator it)
            : Schedule<T>::kv_type::const_iterator{std::move(it)} {}
        mapped_type *operator->() const;
        mapped_type operator*() const;
    };
    using iterator = const_iterator;

    struct Job : private kv_type::value_type {
        Job(mapped_type m, key_type k) : kv_type::value_type{k, m} {};
        [[nodiscard]] key_type timestamp() const { return this->first; }
        [[nodiscard]] mapped_type value() const { return this->second; }
    };

    Schedule(Logger *logger, std::string name);

    std::string name() const;
    Logger *logger() const;

    bool empty() const;
    size_type size() const;
    void clear();
    std::optional<key_type> insert(const Job &job);
    [[nodiscard]] std::optional<Job> lowestUnder(key_type when) const;
    std::optional<Job> extractLowestUnder(key_type when);
    std::optional<Job> extractFirst();
    [[nodiscard]] size_t runnableCount(key_type when) const;

    const_iterator begin() const;
    const_iterator end() const;

private:
    mutable ContextLogger logger_;
    std::string name_;
    kv_type primary_;
    vk_type secondary_;

    void erase(const Job &job);
};

template <typename T>
inline typename Schedule<T>::mapped_type *
Schedule<T>::const_iterator::operator->() const {
    return static_cast<Schedule<T>::mapped_type *const>(
        &kv_type::const_iterator::operator->()->second);
}

template <typename T>
inline typename Schedule<T>::mapped_type
Schedule<T>::const_iterator::operator*() const {
    return Schedule<T>::kv_type::const_iterator::operator*().second;
}

template <typename T>
inline Schedule<T>::Schedule(Logger *logger, std::string name)
    : logger_{logger,
              [this](std::ostream &os) {
                  os << "[" << this->name() << " scheduler] ";
              }}
    , name_{std::move(name)} {}

template <typename T>
[[nodiscard]] inline std::string Schedule<T>::name() const {
    return name_;
}

template <typename T>
[[nodiscard]] inline Logger *Schedule<T>::logger() const {
    return &logger_;
}

template <typename T>
[[nodiscard]] inline bool Schedule<T>::empty() const {
    return primary_.empty();
}

template <typename T>
[[nodiscard]] inline typename Schedule<T>::size_type Schedule<T>::size() const {
    return primary_.size();
}

template <typename T>
[[maybe_unused]] inline std::optional<typename Schedule<T>::key_type>
Schedule<T>::insert(const Job &job) {
    if (job.timestamp() == key_type{}) {
        Warning(logger()) << "refuse to schedule " << job
                          << " with an invalid timestamp";
        return std::nullopt;
    }
    erase(job);
    if (logger()->isLoggable(LogLevel::debug)) {
        Debug(logger()) << "scheduling " << job;
    }
    primary_.emplace(job.timestamp(), job.value());
    secondary_.emplace(job.value(), job.timestamp());
    assert(primary_.size() == secondary_.size());
    return job.timestamp();
}

template <typename T>
inline void Schedule<T>::erase(const Job &job) {
    // Has the job been scheduled before?
    const auto it2 = secondary_.find(job.value());
    if (it2 == secondary_.end()) {
        return;
    }
    // If yes, we have to remove the old entry. This can be done quite
    // efficiently, because we know the old scheduled time. This way, we can
    // restrict our search to the small range of jobs with exactly the same
    // time.
    const auto [begin, end] = primary_.equal_range(it2->second);
    const auto it1 = std::find_if(begin, end, [&job](const auto &elem) {
        return elem.second == job.value();
    });
    if (it1 == end) {
        return;
    }
    Debug(logger()) << "removing old scheduling entry for " << job;
    primary_.erase(it1);
    secondary_.erase(it2);
    assert(primary_.size() == secondary_.size());
}

template <typename T>
inline std::optional<typename Schedule<T>::Job> Schedule<T>::extractLowestUnder(
    key_type when) {
    auto job = lowestUnder(when);
    if (job) {
        if (logger()->isLoggable(LogLevel::debug)) {
            Debug(logger()) << "unscheduling " << *job;
        }
        primary_.erase(primary_.begin());
        secondary_.erase(job->value());
    }
    assert(primary_.size() == secondary_.size());
    return job;
}

template <typename T>
inline std::optional<typename Schedule<T>::Job> Schedule<T>::extractFirst() {
    auto it = primary_.begin();
    if (it == primary_.end()) {
        return std::nullopt;
    }

    Job job{it->second, it->first};
    primary_.erase(primary_.begin());
    secondary_.erase(job.value());
    return job;
}

template <typename T>
inline std::optional<typename Schedule<T>::Job> Schedule<T>::lowestUnder(
    key_type when) const {
    auto it = primary_.begin();
    if (it == primary_.end()) {
        return std::nullopt;
    }
    if (it->first > when) {
        return std::nullopt;
    }
    return Job{it->second, it->first};
}

template <typename T>
inline size_t Schedule<T>::runnableCount(key_type when) const {
    return std::distance(primary_.begin(), primary_.upper_bound(when));
}

template <typename T>
inline void Schedule<T>::clear() {
    Debug(logger()) << "unscheduling all objects";
    primary_.clear();
    secondary_.clear();
}

template <typename T>
inline typename Schedule<T>::const_iterator Schedule<T>::begin() const {
    return const_iterator{primary_.begin()};
}

template <typename T>
inline typename Schedule<T>::const_iterator Schedule<T>::end() const {
    return const_iterator{primary_.end()};
}

#endif  // Schedule_h
