// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef State_h
#define State_h

#include <array>
#include <chrono>
#include <cstdint>
#include <functional>
#include <string>

class CheckResult;
class Histcache;
enum class StateType : uint32_t;

// queried in hosts/services tables
enum class AcknowledgementType : uint32_t { none = 0, normal = 1, sticky = 2 };

enum class NotificationKind { none, normal, flapping };

struct State {
    AcknowledgementType acknowledgement_type_;

    struct StateCheckResult {
        // TODO(ml): Why would we need this *and* CheckResult?
        std::string output_;
        std::string long_output_;
        std::string perf_data_;
        std::chrono::system_clock::time_point cached_at_;
        std::chrono::seconds cache_interval_;

        bool operator==(const StateCheckResult &other) const {
            return output_ == other.output_ &&
                   long_output_ == other.long_output_ &&
                   perf_data_ == other.perf_data_ &&
                   cached_at_ == other.cached_at_ &&
                   cache_interval_ == other.cache_interval_;
        }
        bool operator!=(const StateCheckResult &other) const {
            return !(*this == other);
        }
    } check_result_;

    struct Downtime {
        uint32_t pending_flex_;  // no of dt with _hot == true
        uint32_t scheduled_depth_;

        bool operator==(const Downtime &other) const {
            return pending_flex_ == other.pending_flex_ &&
                   scheduled_depth_ == other.scheduled_depth_;
        }
        bool operator!=(const Downtime &other) const {
            return !(*this == other);
        }
    } downtime_;

    struct Enabling {
        // Bit positions for modified (and unmodified) attributes (enablings)
        // We have to keep the same values as Nagios to avoid translation
        // in Livestatus.
        // For a full set see nagios/common.h: MODATTR_*, which are
        // representing corresponding bit values.
        enum class Key {
            notifications = 0,
            active_checks = 1,
            passive_checks = 2,
            flap_detection = 4,
            performance_data = 6,
            freshness_checks = 13
        };

        // NOTE: To keep our mind sane, the type of the two members below (which
        // are actually bit fields) should be in sync with Nagios'
        // modified_attributes field in the host/service/contact structs. But
        // what we *really* want is a std::bitset for both...
        unsigned long value_;     // actual settings
        unsigned long modified_;  // flags with explicit setting
        bool operator==(const Enabling &other) const {
            return value_ == other.value_ && modified_ == other.modified_;
        }
        bool operator!=(const Enabling &other) const {
            return !(*this == other);
        }

        [[nodiscard]] bool isModified(Key what, bool &enabled) const {
            bool modified = isEnabled(modified_, what);
            if (modified) {
                enabled = isEnabled(value_, what);
            }
            return modified;
        }

        void set(Key what, bool enabled, bool enabled_per_config) {
            if (enabled_per_config != enabled) {
                setEnabled(modified_, what, true);
            }
            setEnabled(value_, what, enabled);
        }

        void clear() {
            modified_ = 0;
            value_ = 0;
        }

    private:
        [[nodiscard]] static bool isEnabled(unsigned long x, Key what) {
            return (x & (1UL << static_cast<int>(what))) != 0;
        }

        static void setEnabled(unsigned long &x, Key what, bool enabled) {
            if (enabled) {
                x |= 1UL << static_cast<int>(what);
            } else {
                x &= ~(1UL << static_cast<int>(what));
            }
        }
    } enabling_;

    struct Flappiness {
        bool is_flapping_;
        double value_;

        bool operator==(const Flappiness &other) const {
            return is_flapping_ == other.is_flapping_ && value_ == other.value_;
        }
        bool operator!=(const Flappiness &other) const {
            return !(*this == other);
        }
        void reset();
        bool flap(double high_flap_threshold);
        bool unflap(double low_flap_threshold, double flap_decay);
    } flappiness_;

    struct HardState {
        uint32_t last_;
        uint32_t previous_;
        bool changed_;  // a hard state change just happened
        std::chrono::system_clock::time_point last_time_;

        bool operator==(const HardState &other) const {
            return last_ == other.last_ && previous_ == other.previous_ &&
                   changed_ == other.changed_ && last_time_ == other.last_time_;
        }
        bool operator!=(const HardState &other) const {
            return !(*this == other);
        }
    } hard_state_;

    struct Notification {
        uint32_t current_number_;
        std::chrono::system_clock::time_point last_time_;
        std::chrono::system_clock::time_point next_time_;
        NotificationKind pending_kind_;
        uint32_t problem_id_;
        bool problem_sent_;
        std::string postponement_reason_;

        bool operator==(const Notification &other) const {
            return current_number_ == other.current_number_ &&
                   last_time_ == other.last_time_ &&
                   next_time_ == other.next_time_ &&
                   pending_kind_ == other.pending_kind_ &&
                   problem_id_ == other.problem_id_ &&
                   problem_sent_ == other.problem_sent_ &&
                   postponement_reason_ == other.postponement_reason_;
        }
        bool operator!=(const Notification &other) const {
            return !(*this == other);
        }
    } notification_;

    uint32_t current_attempt_;
    bool has_been_checked_;
    std::chrono::duration<double> execution_time_;
    std::chrono::duration<double> latency_;
    std::chrono::system_clock::time_point last_check_;
    std::chrono::system_clock::time_point last_check_started_;
    std::chrono::system_clock::time_point last_state_change_;
    std::array<std::chrono::system_clock::time_point, 4> last_time_state_;
    std::chrono::system_clock::time_point next_check_;
    // NOTE: current_state_ is the internal state of the object, i.e.
    // OK/WARN/CRIT/UNKNOWN for services resp. UP/DOWN/UNREACHABLE for
    // hosts, same for _last_state.
    uint32_t current_state_;
    uint32_t initial_state_;  // TODO(sp) What's this? It's never touched...
    uint32_t last_state_;
    StateType state_type_;

    // Volatile settings, not saved
    bool is_helper_running_ = false;
    unsigned _shared = 0;  // wether this state is shared by more worlds
    Histcache *_histcache = nullptr;

    bool operator==(const State &other) const;
    bool operator!=(const State &other) const { return !(*this == other); }

    void update(std::chrono::system_clock::time_point tp_now,
                const CheckResult &cr,
                const std::function<void(int state, const std::string &output,
                                         const std::string &long_output)>
                    &register_alert,
                bool enable_flap_detection);
    void checkForHardStateChange();
    Histcache *histcache();

    [[nodiscard]] bool stateChanged() const {
        return last_state_ != current_state_;
    }
};

#endif  // State_h
