// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef StateHistoryThread_h
#define StateHistoryThread_h

#include <chrono>
#include <cstddef>
#include <filesystem>
#include <mutex>
#include <string>
#include <string_view>
#include <unordered_map>
#include <unordered_set>

class Core;
class Histcache;
class Object;
class TableCachedStatehist;

enum class EventType { alert, flapping, downtime, vanished };

enum class StateHistoryObject { host, service };

enum class StateHistoryState { initial, current };

struct StateHistoryEvent {
    EventType type;
    StateHistoryObject object;
    StateHistoryState state;
};

class StateHistoryThread {
public:
    explicit StateHistoryThread(TableCachedStatehist *table);
    void run(Core *core);

    bool shouldTerminate();
    void setShouldTerminate(bool value);

    bool threadFinished();
    void setThreadFinished(bool value);

    Histcache *getOrCreateObjectHistcache(Object *object);
    void flushCache();

    std::string_view hashedCopyOf(const std::string &s);

private:
    TableCachedStatehist *const _table;

    // The mutex protects _should_terminate.
    std::mutex _should_terminate_mutex;
    bool _should_terminate;

    // The mutex protects _thread_finished.
    std::mutex _thread_finished_mutex;
    bool _thread_finished;

    // Cache of all histories of all existing and non-existing objects.
    // Key is a namespec, e.g. 'localhost' or 'localhost;CPU load'
    // TODO(sp) Not all accesses to _histcaches are protected by this mutex!!!
    std::mutex _histcaches_mutex;
    std::unordered_map<std::string, Histcache *> _histcaches;

    // String hash for keeping only one copy of each unique string
    std::unordered_set<std::string> _strings;
    int _cached_strings;

    Core *_core{};
    int _parsed_messages;
    size_t _parsed_bytes_total;

    void logStatistics(std::chrono::nanoseconds processing_time,
                       std::chrono::nanoseconds history_extent,
                       size_t num_log_files);
    void processLogfile(const std::filesystem::path &path);
    void processLogfileBuffer(const std::filesystem::path &path,
                              std::string_view content);
    void processLogfileLine(const std::filesystem::path &path,
                            std::string_view line);
    void processStateLine(const std::filesystem::path &path,
                          std::chrono::system_clock::time_point timestamp,
                          StateHistoryEvent event, std::string_view line);
    void processProgramLine(std::chrono::system_clock::time_point timestamp,
                            std::string_view line);
    void handleObjectLogEntry(std::chrono::system_clock::time_point timestamp,
                              Object *object, const std::string &host_name,
                              const std::string &svc_desc, EventType event_type,
                              StateHistoryState state_history_state, int state,
                              const std::string &plugin_output,
                              const std::string &long_plugin_output);
    void setAllUnmonitored(std::chrono::system_clock::time_point when);
    void finalizeCaches();
    void registerCoreStartstop(bool start_and_not_stop,
                               std::chrono::system_clock::time_point when);
    Histcache *getOrCreateHistcache(const std::string &specname);
};

#endif  // StateHistoryThread_h
