// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef Timeperiod_h
#define Timeperiod_h

#include <chrono>
#include <cstdint>
#include <string>
#include <vector>

class MonitoringLog;

class Timeperiod {
public:
    using value_type = std::chrono::system_clock::time_point;
    using const_iterator = std::vector<value_type>::const_iterator;
    using iterator = std::vector<value_type>::iterator;

    Timeperiod(std::string name, std::string alias,
               std::vector<value_type> transitions);

    void update(MonitoringLog *monitoring_log, value_type now);

    // Directly exposed via e.g. TableTimeperiods (both cores)
    [[nodiscard]] std::string name() const { return _name; }
    [[nodiscard]] std::string alias() const { return _alias; }
    [[nodiscard]] bool isActive() const { return (_next_transition % 2) != 0U; }

    // Directly exposed via e.g. TableTimeperiods (CMC only)
    [[nodiscard]] value_type nextTransitionTime() const {
        return _next_transition < _transitions.size()
                   ? _transitions[_next_transition]
                   : value_type();
    }
    [[nodiscard]] int nextTransition() const { return _next_transition; }
    [[nodiscard]] int numTransitions() const { return _transitions.size(); }
    [[nodiscard]] const_iterator begin() const;
    [[nodiscard]] const_iterator end() const;

    // Indirectly exposed via staleness in CMC's TableHosts/TableServices
    [[nodiscard]] value_type lastStateChange() const;
    [[nodiscard]] value_type previousLastStateChange() const;

    // For TableCachedStatehist only
    [[nodiscard]] value_type transition(int i) const { return _transitions[i]; }
    // find index of transition following a time
    [[nodiscard]] int findTransition(value_type when) const;

    // CMC-scheduling-related stuff
    [[nodiscard]] value_type nextActiveTime(value_type from) const;
    [[nodiscard]] bool isActiveTime(value_type t) const;

    bool operator==(const Timeperiod &other) const;
    bool operator!=(const Timeperiod &other) const { return !(*this == other); }

private:
    std::string _name;
    std::string _alias;
    std::vector<value_type> _transitions;
    // TODO(ml): It would probably be less cumbersome to use
    //           iterators instead of indices.
    uint32_t _next_transition;  // Index of the next transition
                                // helper variable for livestatus
};

#endif  // Timeperiod_h
