// Copyright (C) 2021 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef Worker_h
#define Worker_h

#include <deque>
#include <functional>
#include <thread>
#include <utility>

#include "livestatus/Logger.h"
#include "livestatus/POSIXUtils.h"
#include "livestatus/Queue.h"

template <typename T, std::size_t QueueSize = 10>
class Worker {
public:
    Worker(Logger *logger, std::string name, queue_join_strategy join_strategy,
           std::function<void(Logger *, T &&)> handler)
        : logger_{logger}
        , name_{std::move(name)}
        , handler_{std::move(handler)}
        , queue_{join_strategy, QueueSize} {}
    ~Worker() { stop(); }

    void start() {
        thread_ = std::thread{[this]() {
            setThreadName(name_);
            Notice{&logger_} << "started";
            while (auto &&task = queue_.pop(queue_pop_strategy::blocking, {})) {
                auto &&[t, size] = *task;
                handler_(&logger_, std::move(t));
            }
            Notice{&logger_} << "stopped";
        }};
    }

    void stop() {
        if (thread_.joinable()) {
            Notice{&logger_} << "stopping " << name_ << "...";
            queue_.join();
            thread_.join();
        }
    }

    [[nodiscard]] std::pair<queue_status, typename Queue<T>::size_type> submit(
        T task, queue_overflow_strategy strategy) {
        return queue_.push(std::move(task), strategy);
    }

private:
    ThreadNameLogger logger_;
    std::string name_;
    std::function<void(Logger *, T &&)> handler_;
    Queue<T> queue_;
    std::thread thread_;
};

#endif
