// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef World_h
#define World_h

#include <netinet/in.h>

#include <chrono>
#include <cstddef>
#include <cstdint>
#include <functional>
#include <iosfwd>
#include <memory>
#include <optional>
#include <string>
#include <unordered_map>
#include <vector>

#include "cmc/ContactList.h"
#include "cmc/GlobalConfig.h"
#include "cmc/Host.h"
#include "cmc/SerialToken.h"
#include "cmc/Timeperiod.h"  // IWYU pragma: keep

class Contact;
class ContactGroup;
class Core;
class HostGroup;
class Logger;
enum class LogStateType;
class MonitoringLog;
class Object;
class RRDFileLocator;
class RRDInfoCache;
class Service;
class ServiceGroup;

namespace checkmk::cmc::config::v1 {
class Core;
}  // namespace checkmk::cmc::config::v1

class CoreConfig {
public:
    GlobalConfig global_config_;
    std::unordered_map<std::string, std::unique_ptr<Timeperiod>> timeperiods_;
    std::unordered_map<std::string, std::unique_ptr<ContactGroup>>
        contact_groups_;
    std::unordered_map<std::string, std::unique_ptr<Contact>> contacts_;
    std::vector<std::unique_ptr<ContactList>> contact_lists_;
    std::vector<std::unique_ptr<Host>> hosts_;
    std::unordered_map<std::string, std::unique_ptr<HostGroup>> host_groups_;
    std::unordered_map<std::string, std::unique_ptr<ServiceGroup>>
        service_groups_;
    std::vector<std::string> commands_;  // for "commands" table only
};

class World {
public:
    Logger *const logger_;
    CoreConfig config_;

    // derived data to speedup access to hosts
    std::unordered_map<std::string, Host *> hosts_by_name_;
    std::unordered_multimap<in_addr_t, Host *> hosts_by_ipv4address_;
    std::unordered_map<std::string, Host *> hosts_by_designation_;

    // derived data, can probably be removed
    size_t num_hosts_;
    size_t num_services_;

    double config_load_time_;  // in ms
    std::chrono::system_clock::time_point creation_time_;

    std::unique_ptr<RRDInfoCache> info_cache_;
    std::unique_ptr<RRDFileLocator> rrd_file_locator_single_;
    std::unique_ptr<RRDFileLocator> rrd_file_locator_multiple_;

    World(Core &core,
          std::unique_ptr<checkmk::cmc::config::v1::Core> core_config,
          Logger *logger_core, Logger *logger_rrd,
          SerialTokenFactory &serial_token_factory);
    ~World();
    void logLoadStatisics() const;
    [[nodiscard]] Host *getHost(const std::string &name) const;
    [[nodiscard]] std::vector<Host *> getHostsByIPv4Address(in_addr_t a) const;
    [[nodiscard]] Host *getHostByDesignation(
        const std::string &designation) const;
    [[nodiscard]] Service *getService(const std::string &host_name,
                                      const std::string &description) const;
    [[nodiscard]] Object *correspondingObject(const Object *object) const;
    [[nodiscard]] HostGroup *getHostGroup(const std::string &name) const;
    [[nodiscard]] HostGroup &ensureHostGroup(const std::string &name);
    [[nodiscard]] ServiceGroup *getServiceGroup(const std::string &name) const;
    [[nodiscard]] ServiceGroup &ensureServiceGroup(const std::string &name);
    [[nodiscard]] ContactGroup *getContactGroup(const std::string &name) const;
    [[nodiscard]] Contact *getContact(const std::string &name) const;
    [[nodiscard]] ContactList *getContactList(uint32_t id) const;
    [[nodiscard]] const Timeperiod &getTimeperiod(
        const std::string &description, const std::string &name);
    void updateTimeperiods(MonitoringLog &monitoring_log,
                           std::chrono::system_clock::time_point now);
    void logCurrentStates(MonitoringLog &monitoring_log,
                          LogStateType log_state_type);
    void logTimeperiods(MonitoringLog &monitoring_log,
                        std::chrono::system_clock::time_point now);
    [[nodiscard]] RRDInfoCache *infoCache() const { return info_cache_.get(); };
    [[nodiscard]] RRDFileLocator *rrdFileLocatorSingle() const {
        return rrd_file_locator_single_.get();
    }
    [[nodiscard]] RRDFileLocator *rrdFileLocatorMultiple() const {
        return rrd_file_locator_multiple_.get();
    }

    // We use dynamically allocated memory => disable copy/assignment
    World(const World &) = delete;
    World &operator=(const World &) = delete;

    // special API
    [[nodiscard]] SerialToken createSerialToken(
        const std::string &creator) const;
    SerialTokenFactory &serial_token_factory_;

private:
    void createConfig(Core &core,
                      const checkmk::cmc::config::v1::Core &core_config,
                      Logger *logger_rrd);

    // NOTE: We do not need the token itself at all, we just need to keep the
    // configuration files alive.
    std::optional<SerialToken> serial_token_;

    friend std::ostream &operator<<(std::ostream &os, const World &w);
};

#endif  // World_h
