// Copyright (C) 2021 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#ifndef finally_h
#define finally_h

#include <type_traits>
#include <utility>

namespace mk {

// final_action and finally are taken from the C++ Guideline Support Library
// (GSL). If you look closely enough, they are just dead simple wrappers around
// the well-known RAII idiom plus some machinery to prevent silly usage and to
// allow move semantics. Taking away all that stuff, it's just a few obvious
// lines:
//
// template <typename F>
// struct final_action {
//     final_action(F f) : f_(f) {}
//     ~final_action() { f_(); }
//     F f_;
// };

// final_action allows you to ensure something gets run at the end of a scope
template <typename F>
class final_action {
public:
    static_assert(!std::is_reference<F>::value && !std::is_const<F>::value &&
                      !std::is_volatile<F>::value,
                  "final_action should store its callable by value");

    explicit final_action(F f) noexcept : f_(std::move(f)) {}

    final_action(final_action &&other) noexcept
        : f_(std::move(other.f_))
        , invoke_(std::exchange(other.invoke_, false)) {}

    final_action(const final_action &) = delete;
    final_action &operator=(const final_action &) = delete;
    final_action &operator=(final_action &&) = delete;

    ~final_action() noexcept {
        if (invoke_) {
            f_();
        }
    }

private:
    F f_;
    bool invoke_{true};
};

// finally() - convenience function to generate a final_action
template <typename F>
[[nodiscard]] final_action<
    typename std::remove_cv<typename std::remove_reference<F>::type>::type>
finally(F &&f) noexcept {
    return final_action<
        typename std::remove_cv<typename std::remove_reference<F>::type>::type>(
        std::forward<F>(f));
}

}  // namespace mk

#endif
