// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/AlertHelper.h"

#include <algorithm>
#include <cerrno>
#include <chrono>
#include <filesystem>
#include <map>
#include <optional>
#include <sstream>
#include <stdexcept>
#include <utility>

#include "cmc/EventSpooler.h"
#include "cmc/GlobalConfig.h"
#include "cmc/Macro.h"
#include "cmc/MonitoringLog.h"
#include "cmc/NotifyHelper.h"
#include "cmc/Object.h"
#include "cmc/Process.h"
#include "livestatus/Logger.h"
#include "livestatus/POSIXUtils.h"
#include "livestatus/Poller.h"

using namespace std::chrono_literals;
using namespace std::string_literals;
using namespace std::string_view_literals;

AlertHelper::AlertHelper(
    std::string name, const std::function<bool()> &enabled_via_global_state,
    NotifyHelper *notify_helper,
    const std::function<Object *(const std::string &, const std::string &)>
        &get_response_object,
    MonitoringLog &monitoring_log, Logger *logger, EventSpooler &spooler)
    : ProcessHelper{std::move(name),
                    "",
                    {"--handle-alerts"s, "--keepalive"s},
                    logger}
    , enabled_via_global_state_{enabled_via_global_state}
    , _notify_helper(notify_helper)
    , _get_response_object(get_response_object)
    , monitoring_log_{monitoring_log}
    , _stdout_consumer{LineProducer{
          [this](auto response) { processResponse(response); }}}
    , _helper_ready{false}
    , spooler_{spooler} {}

AlertHelper::~AlertHelper() { this->do_stop(); }

void AlertHelper::start_() { spawnConnectedSubprocess(); }

void AlertHelper::stop_() { this->do_stop(); }

void AlertHelper::setUpPoller_(Poller &poller) const {
    if (stdOut() >= 0) {
        poller.addFileDescriptor(stdOut(), PollEvents::in);
    }
}

void AlertHelper::communicate_(const Poller &poller) {
    if (stdOut() >= 0 && poller.isFileDescriptorSet(stdOut(), PollEvents::in)) {
        handleStdOut();
    }
}

void AlertHelper::handleStdOut() {
    try {
        _stdout_consumer(stdOut());
    } catch (const std::runtime_error &e) {
        Error(logger()) << e.what();
        restart();
        return;
    }
}

void AlertHelper::do_stop() {
    _helper_ready = false;
    try {
        if (stdOut() >= 0) {
            _stdout_consumer(stdOut());
        }
    } catch (const std::runtime_error &e) {
        Error(logger()) << e.what();
    }
    despawn();
}

bool AlertHelper::updateConfig(const GeneralConfig &general,
                               const AlertHandlerConfig &alert_handler) {
    bool changed = updatePath(general.cmk_executable_);
    if (enabled_via_config_ != alert_handler.enabled_) {
        enabled_via_config_ = alert_handler.enabled_;
        changed = true;
    }
    if (event_types_ != alert_handler.event_types_) {
        event_types_ = alert_handler.event_types_;
        changed = true;
    }
    return changed;
}

void AlertHelper::handle(const Object &object,
                         AlertHandlerEventType event_type) {
    std::string type_name = event_type == AlertHandlerEventType::statechange
                                ? "STATECHANGE"
                                : "CHECKRESULT";
    auto log_not_sending = [&](auto reason) {
        Debug(logger()) << "not sending alert of type " << type_name
                        << " about " << object << ": " << reason;
    };
    if (!enabled_via_config_) {
        log_not_sending("there are no alert handlers defined");
        return;
    }
    if (!enabled_via_global_state_()) {
        log_not_sending("alert handlers are disabled globally");
        return;
    }
    if ((static_cast<uint32_t>(event_type) & event_types_) == 0) {
        // Yes, we log skipping of statechange but not checkresult. This
        // is introduced in 0fdbb143240fd439295aacd70fdbdcf2e483440c (MK)
        if (event_type == AlertHandlerEventType::statechange) {
            log_not_sending("this event type is disabled");
        }
        return;
    }
    Debug(logger()) << "sending alert of type " << type_name << " about "
                    << object << " for being handled";
    std::ostringstream os;
    os << Macro("ALERTTYPE", type_name) << ObjectMacros(object) << "\n";
    spooler_.spool(os.str());
}

void AlertHelper::tryToSendEvent() {
    if (!_helper_ready || spooler_.empty()) {
        return;
    }
    auto event = spooler_.unspool();
    if (writeWithTimeout(stdIn(), event, 3s) == -1) {
        auto saved_errno = errno;
        spooler_.spool(event);      // write back!
        if (auto &p = process()) {  // TODO(sp) Why do we do this *here*?
            p->interrupt();
        }
        throw generic_error{saved_errno, "cannot send alert data"};
    }
    _helper_ready = false;
    Debug(logger()) << "successfully sent event";
}

// TODO(sp): Shotgun parsing ahead! Clearly separate message parsing/dispatching
// from actual business logic.
void AlertHelper::processResponse(std::string_view response) {
    if (response == "*"sv) {
        return processResponseStar();
    }
    constexpr auto started = "STARTED "sv;
    if (response.starts_with(started)) {
        return processResponseSTARTED(response.substr(started.size()));
    }
    constexpr auto stopped = "STOPPED "sv;
    if (response.starts_with(stopped)) {
        return processResponseSTOPPED(response.substr(stopped.size()));
    }
    throw generic_error{EINVAL,
                        "invalid response '" + std::string{response} + "'"};
}

void AlertHelper::processResponseStar() {
    _helper_ready = true;
    tryToSendEvent();
}

namespace {
std::string_view next_field(std::string_view &str) {
    auto field = str.substr(0, str.find(';'));
    str.remove_prefix(std::min(str.size(), field.size() + 1));
    return field;
}

std::string next_terminated_field(std::string_view &str,
                                  std::string_view full) {
    auto old_response = str;
    auto field = next_field(str);
    if (field == old_response) {
        throw generic_error{"incomplete result from alert helper: '" +
                            std::string{full} + "'"};
    }
    return std::string{field};
}
}  // namespace

// Form of the STARTED response (service can be empty for host alerts):
// host;service;handler
void AlertHelper::processResponseSTARTED(std::string_view response) {
    auto full = response;
    auto host = next_terminated_field(response, full);
    auto service = next_terminated_field(response, full);
    auto handler = std::string{response};

    const auto *object = _get_response_object(host, service);
    if (object == nullptr) {
        return;  // Object could have vanished in the meantime
    }
    monitoring_log_.alertHandlerStarted(*object, handler);
}

// Form of the STOPPED response (service can be empty for host alerts, status
// could be OK/WARNING/CRITICAL/UNKNOWN):
// host;service;handler;status;output
void AlertHelper::processResponseSTOPPED(std::string_view response) {
    auto full = response;
    auto host = next_terminated_field(response, full);
    auto service = next_terminated_field(response, full);
    auto handler = next_terminated_field(response, full);
    auto status = next_terminated_field(response, full);
    auto output = std::string{response};

    const auto *object = _get_response_object(host, service);
    if (object == nullptr) {
        return;  // Object could have vanished in the meantime
    }
    monitoring_log_.alertHandlerStopped(*object, handler, status, output);
    // Context: This check is required, because we do not know yet whether our
    // _notify_helper nullabale may be nullable. Reference is also possible. And
    // in this case we will disable this check. Also refcatoring here is
    // mandatory. And code below and a bit *above* may be moved to separate
    // notification function.
    if (_notify_helper != nullptr && object->notificationsEnabled(logger()) &&
        object->inNotificationPeriod()) {
        _notify_helper->notify({*object,
                                "ALERTHANDLER (" + std::string{status} + ")",
                                "",
                                "",
                                {{"ALERTHANDLERNAME", handler},
                                 {"ALERTHANDLERSTATE", status},
                                 {"ALERTHANDLEROUTPUT", output}}});
    }
}
