// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/CarbonConnection.h"

#include <arpa/inet.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>

#include <cerrno>
#include <compare>
#include <cstddef>
#include <numeric>

#include "cmc/CarbonSerializer.h"
#include "livestatus/BitMask.h"
#include "livestatus/POSIXUtils.h"
#include "livestatus/Poller.h"
#include "livestatus/StringUtils.h"
#include "livestatus/global_counters.h"

using namespace std::chrono_literals;

CarbonConnection::CarbonConnection(const CarbonConnectionConfig &config,
                                   Counter bytes_sent_counter)
    : config_{config}
    , socket_{-1}
    , bytes_sent_counter_{bytes_sent_counter}
    , state_{State::unconnected}
    , log_prefix_{"[carbon " + mk::ipv4ToString(config.ip_) + ":" +
                  std::to_string(config.port_) + "] "} {}

CarbonConnection::State CarbonConnection::tryToInitiateConnection(
    std::chrono::steady_clock::time_point time_now) {
    if (state_ != State::unconnected ||
        time_now - last_connection_attempt_ <= config_.retry_interval_) {
        return state_;
    }

    last_connection_attempt_ = time_now;

    const int sock = ::socket(PF_INET, SOCK_STREAM, 0);
    if (sock == -1) {
        const generic_error ge{"Cannot create socket"};
        Warning(logger()) << ge;
        socket_ = -1;
        state_ = State::unconnected;
        return state_;
    }

    // Set Carbon socket to non-blocking. If carbon is too slow for handling
    // our data then we rather lose data then throttle the monitoring (or hang
    // for ages if the target host is down)
    if (::fcntl(sock, F_SETFL, O_NONBLOCK) == -1) {
        const generic_error ge{"cannot make connection non-blocking"};
        Alert(logger()) << ge;
        ::close(sock);
        socket_ = -1;
        state_ = State::unconnected;
        return state_;
    }

    sockaddr_in sa{};
    sa.sin_family = AF_INET;
    sa.sin_port = htons(config_.port_);
    sa.sin_addr.s_addr = config_.ip_;

    const int status =
        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
        ::connect(sock, reinterpret_cast<const struct sockaddr *>(&sa),
                  sizeof(sockaddr_in));
    if (status == -1 && errno != EINPROGRESS) {
        const generic_error ge{"Cannot connect"};
        Alert(logger()) << ge;
        ::close(sock);
        socket_ = -1;
        state_ = State::unconnected;
        return state_;
    }

    if (status == 0) {
        Notice(logger()) << "Successfully connected";
        state_ = State::connected;
    } else {
        Notice(logger()) << "Successfully initiated connection";
        state_ = State::connection_in_progress;
    }
    socket_ = sock;
    return state_;
}

CarbonConnection::State CarbonConnection::tryToCompleteConnection() {
    if (state_ != State::connection_in_progress) {
        return state_;
    }

    if (!Poller{}.wait(0ms, socket_, PollEvents::in | PollEvents::out,
                       logger())) {
        if (errno != ETIMEDOUT) {
            close();
        }
        return state_;
    }

    // something happened, so let's check if we were successful
    int error = 0;
    auto len = static_cast<socklen_t>(sizeof error);
    // In case of a socket error on some OSes (e.g. Solaris), the getsockopt
    // call below fails and sets errno directly.
    if (getsockopt(socket_, SOL_SOCKET, SO_ERROR, &error, &len) == 0) {
        errno = error;
    }

    if (errno != 0) {
        const generic_error ge{"Connection failed"};
        Warning(logger()) << ge;
        close();
        return state_;
    }

    // finally: we're connected!
    Notice(logger()) << "Successfully connected";
    state_ = State::connected;
    return state_;
}

bool CarbonConnection::operator==(const CarbonConnection &other) const {
    return config_ == other.config_;
}

void CarbonConnection::send(const carbon::protocol::Update &message) {
    if (tryToCompleteConnection() != State::connected) {
        return;
    }
    auto serialized = carbon::serialize(message, config_);
    send(serialized);
    counterIncrementBy(
        bytes_sent_counter_,
        std::accumulate(serialized.cbegin(), serialized.cend(), std::size_t{0},
                        [](const auto &acc, const std::string &elem) {
                            return acc + elem.size();
                        }));
}

void CarbonConnection::send(const std::vector<std::string> &lines) {
    if (state_ != State::connected) {
        return;  // no need to spam warnings to the logfile
    }
    for (auto &&line : lines) {
        if (writeWithTimeout(socket_, line, 100ms) == -1) {
            const generic_error ge{errno == ETIMEDOUT
                                       ? "too slow, skipping update"
                                       : "cannot send update"};
            Warning(logger()) << ge;
            close();
            return;
        }
    }
}

void CarbonConnection::close() {
    if (state_ == State::unconnected) {
        return;
    }
    Notice(logger()) << "Closing connection";
    ::close(socket_);
    socket_ = -1;
    state_ = State::unconnected;
}
