// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/CarbonSerializer.h"

#include <algorithm>
#include <utility>

#include "cmc/CarbonConfig.h"
#include "cmc/CarbonProtocol.h"
#include "livestatus/Metric.h"
#include "livestatus/PerformanceData.h"

namespace {
class CarbonSerializer {
public:
    CarbonSerializer(std::string host_name, std::string service_description,
                     std::string timestamp, CarbonConnectionConfig config)
        : host_name_{std::move(host_name)}
        , service_description_{std::move(service_description)}
        , timestamp_{std::move(timestamp)}
        , config_{std::move(config)} {}

    void append(bool enabled, const std::string &suffix,
                const std::string &value) {
        if (!enabled || value.empty()) {
            return;
        }
        return append(suffix, value,
                      config_.path(host_name_, service_description_));
    }
    void append(bool enabled, const std::string &suffix, const Metric &metric,
                const std::string &value) {
        if (!enabled || value.empty()) {
            return;
        }
        return append(suffix, value,
                      config_.path(host_name_, service_description_, metric));
    }

    [[nodiscard]] std::vector<std::string> dump() {
        auto out = serialized_;
        serialized_.clear();
        return out;
    }

private:
    std::string host_name_;
    std::string service_description_;
    std::string timestamp_;
    CarbonConnectionConfig config_;
    std::vector<std::string> serialized_;

    void append(const std::string &suffix, const std::string &value,
                const std::string &path) {
        serialized_.emplace_back(path + suffix + " " + value + timestamp_);
    }
};
}  // namespace

namespace carbon {
std::vector<std::string> serialize(const protocol::Update &message,
                                   const CarbonConnectionConfig &config) {
    CarbonSerializer serializer(message.host_name, message.service_description,
                                message.timestamp, config);
    // NOTE: Nagios performance data labels (a.k.a. metric names) do not
    // contain an equals sign or a single quote, so there can't be any
    // naming collision below.
    const auto &parts = message.carbon_send_metric_parts;
    serializer.append(parts.state, "=state", message.current_state);
    serializer.append(parts.scheduled_downtime_depth,
                      "=scheduled_downtime_depth",
                      message.scheduled_downtime_depth);
    serializer.append(parts.acknowledged, "=acknowledged",
                      message.acknowledged);
    for (const auto &metric : message.perf_data) {
        serializer.append(parts.value, "", metric, metric.value());
        serializer.append(parts.uom, ".uom", metric, metric.uom());
        serializer.append(parts.warn, ".warn", metric, metric.warn());
        serializer.append(parts.crit, ".crit", metric, metric.crit());
        serializer.append(parts.min, ".min", metric, metric.min());
        serializer.append(parts.max, ".max", metric, metric.max());
    }
    return serializer.dump();
}
}  // namespace carbon
