// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/CheckHelper.h"

#include <algorithm>
#include <cctype>
#include <cerrno>
#include <chrono>
#include <cstdlib>
#include <initializer_list>
#include <sstream>
#include <stdexcept>
#include <string_view>

#include "cmc/CheckResult.h"
#include "cmc/FetcherReply.h"
#include "cmc/Host.h"
#include "cmc/Object.h"
#include "cmc/Process.h"
#include "cmc/World.h"
#include "cmc/cmc.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/POSIXUtils.h"
#include "livestatus/Poller.h"
#include "livestatus/StringUtils.h"
#include "livestatus/global_counters.h"

using namespace std::chrono_literals;
using namespace std::string_literals;

CommonHelper::CommonHelper(const std::filesystem::path &path,
                           const args_type &args, Logger *logger)
    : ProcessHelper{"helper", path, args, &logger_}
    , logger_{logger, [this](std::ostream &os) {
                  if (auto *object = this->object()) {
                      os << "[" << *object << "] ";
                  }
              }} {
    gotoIdle();
}

bool CommonHelper::runCheck(Object *object) {
    assertState(CommonHelper::State::idle);
    gotoSending(*object, object->world()->createSerialToken(
                             "Request[" + object->host()->name() + "]"));
    if (!sendRequest()) {
        gotoDead();
        return false;
    }
    incrementCounter();
    gotoReceiving();
    return true;
}

void CommonHelper::killCheck() {
    Notice(logger()) << "aborting running check";
    closeSockets();  // TODO(sp) Necessary?
    if (auto &p = process()) {
        p->terminate();
    }
    gotoDead();
}

void CommonHelper::assertState(State state) const {
    if (this->state() != state) {
        throw std::logic_error{"invalid helper state " +
                               std::to_string(static_cast<int>(state))};
    }
}

bool CommonHelper::migrate(const World &world) {
    if (object_ != nullptr) {
        if (auto *migrated_object = world.correspondingObject(object_)) {
            object_ = migrated_object;
        } else {
            Notice(logger())
                << *object_
                << " vanished from configuration while helper was running";
            killCheck();
        }
    }
    if (!sendReloadConfigRequest()) {
        gotoDead();
        return false;
    }
    return true;
}

void CommonHelper::gotoIdle() {
    state_ = State::idle;
    object_ = nullptr;
    serial_token_.reset();
}

void CommonHelper::gotoSending(Object &object, SerialToken &&serial_token) {
    assertState(State::idle);
    state_ = State::sending;
    object_ = &object;
    serial_token_ = std::move(serial_token);
}

void CommonHelper::gotoReceiving() {
    assertState(State::sending);
    state_ = State::receiving;
}

void CommonHelper::gotoDead() {
    state_ = State::dead;
    object_ = nullptr;
    serial_token_.reset();
}

// -----------------------------------------------------------------------------

CheckHelper::CheckHelper(const std::filesystem::path &path,
                         const args_type &args, result_handler_t result_handler,
                         Logger *logger)
    : CommonHelper{path, args, logger}
    , _result_handler{std::move(result_handler)} {}

CheckHelper::~CheckHelper() { do_stop(); }

void CheckHelper::start_() {
    if (spawnConnectedSubprocess()) {
        gotoIdle();
    } else {
        gotoDead();
    }
}

void CheckHelper::stop_() { do_stop(); }

void CheckHelper::setUpPoller_(Poller &poller) const {
    if (state() == State::sending && stdIn() >= 0) {
        poller.addFileDescriptor(stdIn(), PollEvents::out);
    }
    if (state() == State::receiving && stdOut() >= 0) {
        poller.addFileDescriptor(stdOut(), PollEvents::in);
    }
    if (stdErr() >= 0) {
        poller.addFileDescriptor(stdErr(), PollEvents::in);
    }
}

void CheckHelper::communicate_(const Poller &poller) {
    if (state() == State::sending && stdIn() >= 0 &&
        poller.isFileDescriptorSet(stdIn(), PollEvents::out)) {
        handleStdIn();
    }
    if (state() == State::receiving && stdOut() >= 0 &&
        poller.isFileDescriptorSet(stdOut(), PollEvents::in)) {
        handleStdOut();
    }
    if (stdErr() >= 0 && poller.isFileDescriptorSet(stdErr(), PollEvents::in)) {
        handleStdErr();
    }
}

bool CheckHelper::sendRequest() { return send(makeRequest()) != -1; }

void CheckHelper::do_stop() {
    // Assume that process is not busy
    fetchHelperErrors(false /* no linefeed necessary */);
    gotoIdle();
    despawn();
}

namespace {
/// \brief returns string with options known for checker
std::string GetSupportedCommandLineOptions(const std::string &cmdline) {
    std::string ret;
    for (auto const &option : {"--check-discovery"}) {
        // TODO(sp) The search for options is broken! (consider start/end!)
        auto result_string = option + " "s;
        if (cmdline.find(" "s + result_string) != std::string::npos) {
            ret += result_string;
        }
    }
    return ret;
}

MessagePayload MakeHeaderFromReply(const FetcherReply &reply) {
    return MessagePayload{
        GetSupportedCommandLineOptions(reply.object_->_check_command_line) +
        std::to_string(reply.serial_token_.value()) + " "s +  //
        reply.object_->host()->name() + " "s +                //
        std::to_string(std::string_view{reply.payload_}.size()) + "\n"s +
        std::to_string(mk::ticks<std::chrono::seconds>(reply.remaining_time_)) +
        "\n"s};
}
}  // namespace

void CheckHelper::handleStdIn() {
    assertState(State::sending);
    if (!checker_request_sender_) {
        throw std::logic_error{"internal helper error, no request sender"};
    }
    auto err = checker_request_sender_->send(stdOut());
    if (err != 0) {
        if (err != EAGAIN) {
            Warning(logger()) << generic_error{err, "cannot send request"};
            gotoDead();
        }
        return;
    }
    Debug(logger()) << "request sent";
    checker_request_sender_.reset();
    gotoReceiving();
}

void CheckHelper::handleStdOut() {
    // getResult() goes to "idle" or "dead", both have no associated object.
    auto *object = this->object();
    getResult(object);
    reschedule(object);
}

void CheckHelper::handleStdErr() { fetchHelperErrors(true); }

// Does non-blocking read from the stderr pipe until
// EOF of WOULDBLOCK. Unfinished lines are stored
// in _pending_error. Note: the helper can close
// the pipe because it died meanwhile.
void CheckHelper::fetchHelperErrors(bool need_linefeed) {
    if (stdErr() == -1) {  // Helper has already closed error pipe
        // TODO(sp): Set _dead = true?
        return;
    }

    // To improve the logging output, we delay the transition to "dead".
    auto dead{false};
    char line[512];
    while (true) {
        const ssize_t r = ::read(stdErr(), line, sizeof(line) - 1);
        if (r == 0) {  // EOF
            if (need_linefeed) {
                Warning(logger()) << "closed error pipe";
            }
            dead = true;
            break;
        }
        if (r < 0) {
            if (errno != EAGAIN) {
                const generic_error ge("failed to read error message");
                Warning(logger()) << ge;
                dead = true;
            }
            break;
        }
        line[r] = 0;
        _pending_error += line;
    }

    logPendingErrors(need_linefeed);
    if (dead) {
        closeSockets();
        gotoDead();
    }
}

// TODO (sk): refactoring is required
void CheckHelper::logPendingErrors(bool need_linefeed) {
    // output complete lines to our logfile
    const char *start = _pending_error.c_str();
    while (true) {
        const char *scanpos = start;
        while (*scanpos != 0 && *scanpos != '\n') {
            scanpos++;
        }

        // now scanpos points to either \0 or \n.
        if (*scanpos == '\n' || !need_linefeed) {
            const std::string part(start, scanpos - start);
            if (!part.empty()) {
                Warning(logger()) << "[log] " << part;
            }
        }
        if (*scanpos == 0) {
            break;
        }
        start = scanpos + 1;
    }
    _pending_error = start;
}

namespace {
bool looksLikeText(const std::string &str) {  // quick-n-dirty approximation
    auto prefix{std::string_view{str}.substr(0, 100)};
    return std::all_of(prefix.cbegin(), prefix.cend(), [](auto ch) {
        return ch >= ' ' || ch == '\t' || ch == '\n' || ch == '\r';
    });
}
}  // namespace

ssize_t CheckHelper::send(const std::string &message) {
    Informational(logger()) << "sending " << message.size() << "-byte message";
    if (logger()->isLoggable(LogLevel::debug)) {
        if (looksLikeText(message)) {
            Debug(logger()) << "message content ["
                            << mk::escape_nonprintable{message} << "]";
        } else {
            Debug(logger()) << "message content is binary data";
        }
    }
    auto ret = writeWithTimeout(stdIn(), message, 3s);
    if (ret == -1) {
        const generic_error ge{"cannot send request"};
        Warning(logger()) << ge;
    }
    return ret;
}

bool CheckHelper::tryReadStatusPart13(char *buffer) {
    int r = 0;
    while (r < 13) {
        const ssize_t num = readWithTimeout(buffer + r, 13 - r);
        if (num <= 0) {
            const generic_error ge("cannot read response");
            Critical(logger()) << ge;
            return false;
        }
        r += num;
        if (isdigit(buffer[0]) == 0) {
            buffer[r] = 0;
            Critical(logger()) << "invalid response '" << buffer
                               << "': Expecting number (status code)";
            return false;
        }
    }
    return true;
}

bool CheckHelper::checkStatusIsGood(char *status_part, const Object *object) {
    status_part[13] = 0;
    if ((isdigit(status_part[0]) == 0) || (isdigit(status_part[4]) == 0)) {
        char rest[512];
        const ssize_t num = readWithTimeout(rest, sizeof(rest) - 1);
        if (num >= 0) {
            rest[num] = 0;
        } else {
            rest[0] = 0;
        }
        if (object == nullptr) {
            Critical(logger())
                << "garbled response while checking nothing: " << status_part
                << rest;

        } else {
            Critical(logger()) << "garbled response while checking " << *object
                               << ": " << status_part << rest;
        }
        return false;
    }

    return true;
}

bool CheckHelper::readDataToVector(std::vector<char> &output) {
    char *write_pointer = output.data();
    size_t bytes_left = output.size();

    // Read response. Beware: this can be blocked. It comes from a socket.
    // This is no pipe.
    while (bytes_left != 0U) {
        const ssize_t chunksize = readWithTimeout(write_pointer, bytes_left);
        if (chunksize == 0) {
            Emergency(logger()) << "read zero bytes, exited?";
            return false;
        }
        if (chunksize == -1) {
            const generic_error ge("error reading response");
            Emergency(logger()) << ge;
            return false;
        }
        bytes_left -= chunksize;
        write_pointer += chunksize;
    }
    return true;
}

void CheckHelper::getResult(Object *object) {
    assertState(State::receiving);
    object->checkHelperStops();

    // Current state: Poller::poll() has seen, that the check helper process
    // has sent us some data. Under normal, bug-free circumstances, this
    // means, that a check command has been executed. We now get the
    // answer as follows:
    // 1. The return code in ascii in format "%03d\n"
    // 2. The number of bytes of the output of the command in format
    // "%08ld\n"
    // 3. The output of the the command, followed by nothing
    char status_part[14];
    if (!tryReadStatusPart13(status_part)) {
        gotoDead();
        return;
    }
    if (!checkStatusIsGood(status_part, object)) {
        gotoDead();
        return;
    }
    status_part[3] = 0;
    const int return_code = atoi(status_part);
    const int output_size = atoi(status_part + 4);
    std::vector<char> output(output_size);
    if (!readDataToVector(output)) {
        gotoDead();
        return;
    }

    const std::string output_str(output.begin(), output.end());
    Informational(logger())
        << "received " << output_size << " bytes of check results, status "
        << return_code;
    Debug(logger()) << "check results: '" << output_str << "'";
    if (!mk::is_utf8(output_str)) {
        Warning(logger()) << generic_error{EINVAL,
                                           "invalid UTF-8 in check result"};
        // Note that we don't want to kill the helper here, it's probably just
        // an active check using an invalid encoding.
        gotoIdle();
        return;
    }

    auto now = std::chrono::system_clock::now();
    try {
        std::istringstream is(output_str);
        while (is.peek() == '\t') {
            auto cr{CheckResult::createPassive(is, object)};
            if (cr._object->isCheckEnabled(CheckResultKind::passive)) {
                cr._object->checkStarts(now);
                _result_handler(now, cr);
            }
        }
        if (expectsActiveCheckResult()) {
            // NOTE: CheckHelperPool::tryToRunCheck already made sure we
            // want to process the result.
            _result_handler(now,
                            CheckResult::createActive(is, object, return_code));
        }
        gotoIdle();
    } catch (const parse_error &error) {
        Critical(logger()) << error.what();
        // CONTEXT:
        // This check is dead and will be killed in next doHouseKeeping
        // call. But the kill leads to unwanted race condition: we will
        // close sockets in do_stop before cmk.utils.log.set_stream performs
        // stream flush. As result python code generates BrokenPipe
        // exception and corresponding crash report. This case must be
        // handled correctly on python side TOO.
        //
        // Condition: -vv and broken check(to make check dead)
        fetchHelperErrors(false /* no linefeed necessary */);
        gotoDead();
    }
}
ssize_t CheckHelper::readWithTimeout(char *buffer, size_t size) {
    // Socket is set to non-blocking. In order to avoid an extra
    // system call in normal situations, we first try a simple
    // read. If that does not return data yet, we wait for some
    // with Poller::poll()
    const ssize_t r = ::read(stdOut(), buffer, size);
    if (r >= 0 || errno != EAGAIN) {
        return r;
    }

    if (!Poller{}.wait(2s, stdOut(), PollEvents::in, logger())) {
        return -1;
    }
    return ::read(stdOut(), buffer, size);
}

void CheckHelper::consumeReply(std::unique_ptr<FetcherReply> reply) {
    assertState(State::idle);
    // Alas, there is currently no support for move-only objects in
    // std::initializer_list, only a proposal:  :-/
    // http://open-std.org/jtc1/sc22/wg21/docs/papers/2014/n4166.pdf
    std::vector<MessagePayload> request;
    request.push_back(MakeHeaderFromReply(*reply));
    request.push_back(std::move(reply->payload_));
    checker_request_sender_ = MessagePayloadSender{std::move(request)};
    // NOTE: We do not need the token itself at all, we just need to keep
    // the configuration files alive during sending/receiving.
    gotoSending(*reply->object_, std::move(reply->serial_token_));
}

// -----------------------------------------------------------------------------

GenericHelper::GenericHelper(const std::filesystem::path &path,
                             const args_type &args,
                             result_handler_t result_handler, Logger *logger,
                             reschedule_handler_t<Object> reschedule_handler)
    : CheckHelper{path, args, std::move(result_handler), logger}
    , reschedule_handler_{std::move(reschedule_handler)} {}

std::string GenericHelper::makeRequest() {
    return object()->host()->name() + "\n" + object()->_check_command_line +
           "\n" +
           std::to_string(
               mk::ticks<std::chrono::seconds>(object()->_check_timeout)) +
           "\n";
}

bool GenericHelper::sendReloadConfigRequest() { return true; }

void GenericHelper::incrementCounter() const {
    counterIncrement(Counter::livechecks);
}

void GenericHelper::reschedule(Object *object) const {
    reschedule_handler_(object);
}

bool GenericHelper::expectsActiveCheckResult() { return true; }

// -----------------------------------------------------------------------------

std::string RealTimeHelper::makeRequest() {
    auto *host = static_cast<Host *>(object());
    // TODO(sp) Transporting the real-time payload via the host object is a
    // real hack: We should better handle this like a reply coming from a
    // (special) fetcher. This would remove the need for a separate
    // real-time pool, too.
    auto &payload = host->_real_time_payload;
    if (!payload) {
        throw std::logic_error{"missing real-time payload"};
    }
    const std::string_view sv{*payload};
    auto message = std::to_string(sv.size()) + " " + host->name() + " " +
                   mk::ipv4ToString(host->_ipv4_address) + "\n1\n" +
                   std::string{sv};
    payload.reset();
    return message;
}

bool RealTimeHelper::sendReloadConfigRequest() { return send("*\n") != -1; }

void RealTimeHelper::incrementCounter() const {
    counterIncrement(Counter::livechecks);
}

void RealTimeHelper::reschedule(Object * /*object*/) const {}

bool RealTimeHelper::expectsActiveCheckResult() { return false; }

// -----------------------------------------------------------------------------

CheckerHelper::CheckerHelper(const std::filesystem::path &path,
                             const args_type &args,
                             result_handler_t result_handler, Logger *logger,
                             reschedule_handler_t<Service> reschedule_handler)
    : CheckHelper{path, args, std::move(result_handler), logger}
    , reschedule_handler_{std::move(reschedule_handler)} {}

std::string CheckerHelper::makeRequest() {
    // TODO(sp): Currently the checker schedule is always empty, so even
    // if CheckHelperPool::runScheduledChecks is called, we never reach
    // the blocking CheckHelper::runCheck. Sending requests to checkers
    // is non-blocking and handled via CheckHelper::consumeReply and
    // CheckHelper::communicate_. We should really unify this chaos by
    // always using the schedule and making everything non-blocking!
    throw std::logic_error("a checker cannot run a scheduled check");
}

bool CheckerHelper::sendReloadConfigRequest() { return send("*\n") != -1; }

void CheckerHelper::incrementCounter() const {
    counterIncrement(Counter::livechecks);
}

void CheckerHelper::reschedule(Object *object) const {
    reschedule_handler_(object->service());
}

bool CheckerHelper::expectsActiveCheckResult() { return true; }
