// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

// NOTE: Some IYWU versions are a bit confused about the header for std::abs.
#include "cmc/CheckHelperPool.h"

#include <algorithm>
#include <functional>
#include <ostream>
#include <utility>

#include "cmc/GlobalConfig.h"
#include "cmc/IOHelper.h"
#include "cmc/Object.h"
#include "cmc/PoolSizeLogger.h"
#include "cmc/cmc.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/Logger.h"

using namespace std::chrono_literals;

CheckHelperPool::CheckHelperPool(const std::string &name, Logger *logger,
                                 Logger *logger_schedule)
    : AbstractHelper{name, ContextLogger{logger,
                                         [this](std::ostream &os) {
                                             os << "[" << this->name() << "] ";
                                         }}}
    , _schedule{logger_schedule, this->name()} {}

CheckHelperPool::~CheckHelperPool() { do_stop(); }

double CheckHelperPool::averageUsage() const { return _avg_usage.get(); }

double CheckHelperPool::averageLatency() const { return _avg_latency.get(); }

void CheckHelperPool::updateUsageAverage() {
    auto num_idle = getNumIdleHelpers();
    const double usage =
        size() == 0 ? 0.0 : 1.0 - (static_cast<double>(num_idle) / size());
    if (usage < 0.0 || usage > 1.0) {
        Warning(logger()) << name() << " usage is " << usage
                          << ", pool size: " << size()
                          << ", num idle: " << num_idle;
    } else {
        _avg_usage.update(usage);
    }
}

size_t CheckHelperPool::getNumIdleHelpers() const {
    size_t num_idle = 0;
    for (const auto &ch : _helpers) {
        if (ch->state() == CommonHelper::State::idle) {
            num_idle++;
        }
    }
    return num_idle;
}

CommonHelper *CheckHelperPool::getIdleHelper() const {
    for (const auto &ch : _helpers) {
        if (ch->state() == CommonHelper::State::idle) {
            return ch.get();
        }
    }
    return nullptr;
}

std::optional<std::chrono::system_clock::time_point> CheckHelperPool::insert(
    const Schedule<Object *>::Job &job) {
    return _schedule.insert(job);
}

std::optional<Schedule<Object *>::Job> CheckHelperPool::lowestUnder(
    std::chrono::system_clock::time_point deadline) const {
    if (getIdleHelper() != nullptr) {
        return _schedule.lowestUnder(deadline);
    }
    return {};
}

void CheckHelperPool::clearSchedule() { _schedule.clear(); }

bool CheckHelperPool::migrate(const World &world) {
    for (auto &ch : _helpers) {
        ch->migrate(world);
    }
    return true;
}

void CheckHelperPool::reload() {
    // Optimization only
    if (_config_has_been_propagated) {
        return;
    }
    const PoolSizeLogger psl{[this]() { return size(); }, logger()};
    decreaseNumberOfHelpers();
    increaseNumberOfHelpers();
    _config_has_been_propagated =
        size() == config_.quantity &&
        std::all_of(_helpers.begin(), _helpers.end(), [&](const auto &ch) {
            return ch->hasSameConfig(config_.path, config_.args);
        });
}

void CheckHelperPool::restartDeadHelpers() {
    for (auto &ch : _helpers) {
        if (ch->state() == CommonHelper::State::dead) {
            ch->restart();
        }
    }
}

bool CheckHelperPool::tryToRunCheck(std::chrono::system_clock::time_point now,
                                    const Schedule<Object *>::Job &job) {
    auto *object = job.value();
    if (!object->isCheckEnabled(CheckResultKind::active)) {
        if (auto t = insert({object, object->nextCheckTime()})) {
            object->setNextCheck(*t, _schedule.logger());
        }
        return false;
    }
    while (auto *ch = getIdleHelper()) {
        if (ch->runCheck(object)) {
            const std::chrono::duration<double> latency = now - job.timestamp();
            _avg_latency.update(
                mk::ticks<std::chrono::duration<double>>(latency));
            object->checkHelperStarts(now, latency);
            return true;
        }
    }
    // Lack of free working helper requires special management. We can't
    // shift the check in time on few hours later just because of a temporary
    // system overload.
    if (auto t = insert({object, now + 2s})) {
        object->setNextCheck(*t, _schedule.logger());
    }
    return false;
}

size_t CheckHelperPool::runScheduledChecks(
    std::chrono::system_clock::time_point now) {
    reload();
    size_t num_checks_started = 0;
    while (auto job = _schedule.extractLowestUnder(now)) {
        // clang-tidy-16 seems to be very confused here, the optional accesses
        // *is* guarded by a while.
        // NOLINTNEXTLINE(bugprone-unchecked-optional-access)
        num_checks_started += tryToRunCheck(now, *job) ? 1 : 0;
    }
    return num_checks_started;
}

size_t CheckHelperPool::runnableCount(
    std::chrono::system_clock::time_point now) const {
    return _schedule.runnableCount(now);
}

bool CheckHelperPool::updateConfig(const GeneralConfig &general,
                                   const CheckingConfig &checking) {
    auto new_config = helperConfig(general, checking);
    auto changed = config_ != new_config;
    config_ = new_config;
    _config_has_been_propagated = false;
    reload();
    return changed;
}

// -----------------------------------------------------------------------------
// implementation of abstract methods from AbstractHelper

void CheckHelperPool::start_() {
    _config_has_been_propagated = false;
    reload();
}

void CheckHelperPool::stop_() { do_stop(); }

void CheckHelperPool::setUpPoller_(Poller &poller) const {
    for (const auto &ch : _helpers) {
        ch->setUpPoller(poller);
    }
}

void CheckHelperPool::communicate_(const Poller &poller) {
    for (auto &ch : _helpers) {
        ch->communicate(poller);
    }
}

// -----------------------------------------------------------------------------
// other stuff

void CheckHelperPool::decreaseNumberOfHelpers() {
    // We have to be careful here: Erasing an element from a sequence container
    // invalidates the iterator pointing to it. The solution is to use the
    // return value of erase, see Scott Meyers' "Effective STL", item 9 ("Choose
    // carefully among erasing options.").
    for (auto it = _helpers.begin(); it != _helpers.end();) {
        if ((*it)->state() == CommonHelper::State::idle &&
            (size() > config_.quantity ||
             !(*it)->hasSameConfig(config_.path, config_.args))) {
            it = _helpers.erase(it);
        } else {
            ++it;
        }
    }
}

void CheckHelperPool::increaseNumberOfHelpers() {
    while (size() < config_.quantity) {
        auto helper = this->makeHelper();
        helper->start();
        _helpers.push_back(std::move(helper));
    }
}

void CheckHelperPool::do_stop() {
    for (auto &ch : _helpers) {
        if (ch->state() != CommonHelper::State::idle) {
            ch->killCheck();
        }
    }
    _helpers.clear();
}

// -----------------------------------------------------------------------------
// other stuff

namespace {
std::vector<std::string> buildHelperArgs(
    const CheckingConfig &config, const std::vector<std::string> &mode_args) {
    std::vector<std::string> args;
    switch (config.log_level_) {
        case LogLevel::emergency:
        case LogLevel::alert:
        case LogLevel::critical:
        case LogLevel::error:
        case LogLevel::warning:
        case LogLevel::notice:
            break;
        case LogLevel::informational:
            args.emplace_back("-v");
            break;
        case LogLevel::debug:
            args.emplace_back("-vv");
            break;
    }
    if (config.debug_) {
        args.emplace_back("--debug");
    }
    args.insert(args.end(), mode_args.begin(), mode_args.end());
    return args;
}
}  // namespace

// -----------------------------------------------------------------------------

GenericHelperPool::GenericHelperPool(
    const std::string &name, Logger *logger, Logger *logger_schedule,
    reschedule_handler_t<Object> reschedule_handler,
    CheckHelper::result_handler_t result_handler)
    : CheckHelperPool{name, logger, logger_schedule}
    , reschedule_handler_{std::move(reschedule_handler)}
    , _result_handler{std::move(result_handler)} {}

CheckHelperPool::HelperConfig GenericHelperPool::helperConfig(
    const GeneralConfig & /*general*/, const CheckingConfig &checking) const {
    return {
        checking.num_generic_helpers_, checking.generic_helper_executable_, {}};
}

std::unique_ptr<CommonHelper> GenericHelperPool::makeHelper() const {
    // NOTE: Things are a bit tricky here regarding scheduling, so simply
    // calling our own insert() method is not correct! Consider the following
    // sequence of events, containing a race condition:
    //    * An object is configured to be checked via active checks.
    //    * The active check is triggered, but it has not returned a result yet.
    //    * A config reload happens.
    //    * The object is now configured to be checked via e.g.smartping.
    //    * The check result of the active check returns.
    //    * We should better reschedule into the smartping queue now!
    return std::make_unique<GenericHelper>(config_.path, config_.args,
                                           _result_handler, logger(),
                                           reschedule_handler_);
}

// -----------------------------------------------------------------------------

CheckerHelperPool::CheckerHelperPool(
    const std::string &name, Logger *logger, Logger *logger_schedule,
    reschedule_handler_t<Service> reschedule_handler,
    CheckHelper::result_handler_t result_handler)
    : CheckHelperPool{name, logger, logger_schedule}
    , reschedule_handler_{std::move(reschedule_handler)}
    , _result_handler{std::move(result_handler)} {}

CheckHelperPool::HelperConfig CheckerHelperPool::helperConfig(
    const GeneralConfig &general, const CheckingConfig &checking) const {
    return {checking.num_checkers_, general.cmk_executable_,
            buildHelperArgs(checking, {"--checker"})};
}

std::unique_ptr<CommonHelper> CheckerHelperPool::makeHelper() const {
    return std::make_unique<CheckerHelper>(config_.path, config_.args,
                                           _result_handler, logger(),
                                           reschedule_handler_);
}

// -----------------------------------------------------------------------------

RealTimeHelperPool::RealTimeHelperPool(
    const std::string &name, Logger *logger, Logger *logger_schedule,
    CheckHelper::result_handler_t result_handler)
    : CheckHelperPool{name, logger, logger_schedule}
    , _result_handler{std::move(result_handler)} {}

CheckHelperPool::HelperConfig RealTimeHelperPool::helperConfig(
    const GeneralConfig &general, const CheckingConfig &checking) const {
    return {checking.num_real_time_helpers_, general.cmk_executable_,
            buildHelperArgs(checking, {"--keepalive", "--real-time-checks"})};
}

std::unique_ptr<CommonHelper> RealTimeHelperPool::makeHelper() const {
    return std::make_unique<RealTimeHelper>(config_.path, config_.args,
                                            _result_handler, logger());
}

// -----------------------------------------------------------------------------

FetcherHelperPool::FetcherHelperPool(
    const std::string &name, Logger *logger, Logger *logger_schedule,
    reschedule_handler_t<Service> reschedule_handler,
    FetcherHelper::reply_handler_t reply_handler)
    : CheckHelperPool{name, logger, logger_schedule}
    , reschedule_handler_{std::move(reschedule_handler)}
    , reply_handler_{std::move(reply_handler)} {}

CheckHelperPool::HelperConfig FetcherHelperPool::helperConfig(
    const GeneralConfig & /*general*/, const CheckingConfig &checking) const {
    return {checking.num_fetchers_, checking.fetcher_executable_, {}};
}

std::unique_ptr<CommonHelper> FetcherHelperPool::makeHelper() const {
    return std::make_unique<FetcherHelper>(config_.path, config_.args,
                                           reply_handler_, reschedule_handler_,
                                           logger());
}
