// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/CheckResult.h"

#include <algorithm>
#include <cstddef>
#include <cstdlib>
#include <iostream>
#include <iterator>
#include <utility>

#include "cmc/Host.h"
#include "cmc/Object.h"
#include "cmc/Service.h"
#include "cmc/cmc.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/Interface.h"
#include "livestatus/LogEntry.h"

using namespace std::chrono_literals;

namespace {
void strip_nl(std::string &str) {
    str.erase(std::remove(str.begin(), str.end(), '\n'), str.cend());
}
}  // namespace

CheckResult::CheckResult(Host *host, HostState state, std::string output)
    : CheckResult(host, static_cast<int>(state), std::move(output), "", "",
                  std::chrono::system_clock::time_point(), 0s) {}

CheckResult::CheckResult(Object *object, int state, std::string output,
                         std::string long_output, std::string perf_data,
                         std::chrono::system_clock::time_point cached_at,
                         std::chrono::seconds cache_interval)
    : _object(object)
    , _state(state)
    , _output(std::move(output))
    , _long_output(std::move(long_output))
    , _perf_data(std::move(perf_data))
    , _cached_at(cached_at)
    , _cache_interval(cache_interval) {}

namespace {
void extractNextField(std::string &field, size_t &pos, const std::string &line,
                      char sep) {
    const size_t sep_pos = line.find(sep, pos);
    if (sep_pos == std::string::npos) {
        field = line.substr(pos);
        pos = line.size();
    } else {
        field = line.substr(pos, sep_pos - pos);
        pos = sep_pos + 1;
    }
}

void verifyHostName(const std::string &kind, Object *object,
                    const std::string &host_name) {
    if (host_name != object->host()->name()) {
        throw parse_error("host name mismatch in " + kind +
                          " check result: expected '" + object->host()->name() +
                          "', got '" + host_name + "'");
    }
}
}  // namespace

// Almost Nagios-compatible format, i.e. the first line looks like
// 'host_name\tcheck output|perf_data' (where |perf_data is optional), and all
// further lines contain the long output, with optional trailing performance
// data. Example:
//    foobar\tOK - Das ist der normale Output|a=17
//    and this is line 2
//    and this line 3|b=18
//    c=19
//    'Foo Bar'=20;0;0;10;50

// static
CheckResult CheckResult::createActive(std::istream &is, Object *object,
                                      int return_code) {
    std::string first_line;
    if (!getline(is, first_line)) {
        throw parse_error("could not read check result");
    }

    size_t pos = 0;
    std::string host_name;
    extractNextField(host_name, pos, first_line, '\t');
    verifyHostName("active", object, host_name);

    std::string output;
    extractNextField(output, pos, first_line, '|');
    std::string perf_data = first_line.substr(pos);

    const std::string rest_lines((std::istreambuf_iterator<char>(is)),
                                 std::istreambuf_iterator<char>());
    pos = 0;
    std::string long_output;
    extractNextField(long_output, pos, rest_lines, '|');
    // Note: Nagios currently doesn't add any space here.
    if (!perf_data.empty() && pos < rest_lines.size()) {
        perf_data += " ";
    }
    perf_data += rest_lines.substr(pos);

    strip_nl(perf_data);
    return {
        object,    returnCodeToState(object, return_code),  output, long_output,
        perf_data, std::chrono::system_clock::time_point(), 0s};
}

std::ostream &operator<<(std::ostream &os, const CheckResult &c) {
    auto ca = std::chrono::system_clock::to_time_t(c._cached_at);
    return os << "_state=" << c._state << "\n"                //
              << "_output=[" << c._output << "]\n"            //
              << "_long_output=[" << c._long_output << "]\n"  //
              << "_perf_data=[" << c._perf_data << "]\n"      //
              << "_cached_at=" << ca << "\n"                  //
              << "_cache_interval="
              << mk::ticks<std::chrono::seconds>(c._cache_interval);
}

// A passive check result consists of a leading TAB, some optional meta
// information (currently only cached_at or cache_interval, TAB-separated), the
// state (TAB-separated), the service description (TAB-separated), check output
// (long check output separated by '\001') and optional performance data (starts
// after '|'). Example:
// "\tcache_interval=10\t0\tCPU load\tOK - 15min load 0.28 at 12
// CPUs|load1=0.87;60;120;0;12 load5=0.5;60;120;0;12 load15=0.28;60;120;0;12"

// static
CheckResult CheckResult::createPassive(std::istream &is, Object *object) {
    // Normally one would do things in a stream-based manner, but to recover
    // from parsing errors in a sensible way and to have good error messages, we
    // retrieve the whole line first.
    std::string line;
    if (!getline(is, line)) {
        throw parse_error("could not read passive check result");
    }
    size_t pos = 1;  // skip initial TAB

    int return_code = 0;
    std::chrono::system_clock::time_point cached_at =
        std::chrono::system_clock::time_point();
    std::chrono::seconds cache_interval{0};
    while (true) {
        std::string field;
        extractNextField(field, pos, line, '\t');
        if (field.empty()) {
            throw parse_error("Missing state in passive check result: " + line);
        }
        const size_t equal_pos = field.find('=');
        if (equal_pos == std::string::npos) {
            return_code = static_cast<int>(strtol(field.c_str(), nullptr, 10));
            break;
        }
        const std::string key = field.substr(0, equal_pos);
        const long value = strtol(field.c_str() + equal_pos + 1, nullptr, 10);
        if (key == "cached_at") {
            cached_at = std::chrono::system_clock::from_time_t(value);
        } else if (key == "cache_interval") {
            cache_interval = std::chrono::seconds(value);
        } else {
            throw parse_error("Invalid meta data in passive check result: " +
                              line);
        }
    }

    std::string host_name;
    extractNextField(host_name, pos, line, '\t');
    verifyHostName("passive", object, host_name);

    std::string service_description;
    extractNextField(service_description, pos, line, '\t');

    std::string all_output;
    extractNextField(all_output, pos, line, '|');
    std::string perf_data = line.substr(pos);
    strip_nl(perf_data);

    std::string output;
    size_t out_pos = 0;
    extractNextField(output, out_pos, all_output, '\001');
    const std::string long_output = all_output.substr(out_pos);

    Service *service = object->host()->getService(service_description);
    if (service == nullptr) {
        throw parse_error("ignoring check result for " +
                          object->host()->name() + " / " + service_description +
                          ": no such service");
    }

    return {service,       returnCodeToState(service, return_code),
            output,        long_output,
            perf_data,     cached_at,
            cache_interval};
}

namespace {
ServiceState toServiceState(HostState state) {
    switch (state) {
        case HostState::up:
            return ServiceState::ok;
        case HostState::down:
            return ServiceState::critical;
        case HostState::unreachable:
            return ServiceState::unknown;
    }
    return ServiceState::ok;  // unreachable, make GCC happy
}

HostState toHostState(ServiceState state, Host *host) {
    switch (state) {
        case ServiceState::ok:
        case ServiceState::warning:
            return HostState::up;
        case ServiceState::critical:
        case ServiceState::unknown:
            return host->getReachability() == Host::Reachability::unreachable
                       ? HostState::unreachable
                       : HostState::down;
    }
    return HostState::up;  // unreachable, make GCC happy
}

int convert(const Object *from, int state, Object *to) {
    // TODO(sp) Remove ugly double dispatch and casts.
    return from->isHost()
               ? (to->isHost() ? state
                               : static_cast<int>(toServiceState(
                                     static_cast<HostState>(state))))
               : (to->isHost()
                      ? static_cast<int>(toHostState(
                            static_cast<ServiceState>(state), to->host()))
                      : state);
}
}  // namespace

CheckResult CheckResult::createCopyFor(Object *object) const {
    return {object,         convert(_object, _state, object),
            _output,        _long_output,
            _perf_data,     _cached_at,
            _cache_interval};
}
