// Copyright (C) 2023 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/CmcCore.h"

#include <algorithm>
#include <iterator>
#include <unordered_map>

#include "cmc/CheckHelperPool.h"
#include "cmc/CmcComment.h"
#include "cmc/CmcDowntime.h"
#include "cmc/CmcGlobalFlags.h"
#include "cmc/CmcHost.h"
#include "cmc/CmcHostGroup.h"
#include "cmc/CmcPaths.h"
#include "cmc/CmcService.h"
#include "cmc/CmcServiceGroup.h"
#include "cmc/CmcTimeperiod.h"
#include "cmc/Comment.h"
#include "cmc/Comments.h"
#include "cmc/Contact.h"
#include "cmc/ContactGroup.h"
#include "cmc/Core.h"
#include "cmc/Downtime.h"
#include "cmc/Downtimes.h"
#include "cmc/EventSpooler.h"
#include "cmc/GlobalConfig.h"
#include "cmc/GlobalState.h"
#include "cmc/Host.h"
#include "cmc/LivestatusExternalCommand.h"
#include "cmc/MonitoringHistory.h"
#include "cmc/ObjectGroup.h"
#include "cmc/RRDInfo.h"
#include "cmc/Service.h"
#include "cmc/State.h"
#include "cmc/World.h"
#include "cmc/cmc.h"
#include "livestatus/Average.h"
#include "livestatus/Interface.h"
#include "livestatus/PnpUtils.h"
#include "livestatus/Store.h"
#include "livestatus/User.h"
#include "livestatus/data_encoding.h"

namespace {
const World *world() { return g_live_world; }

const GlobalConfig *config() { return &world()->config_.global_config_; }
}  // namespace

const IHost *CmcCore::find_host(const std::string &name) const {
    const auto *host = world()->getHost(name);
    return host == nullptr ? nullptr : &host->ihost();
}

const IHostGroup *CmcCore::find_hostgroup(const std::string &name) const {
    const auto *hg = world()->getHostGroup(name);
    return hg == nullptr ? nullptr : &hg->ihostgroup();
}

std::unique_ptr<const IHost> CmcCore::getHostByDesignation(
    const std::string &designation) const {
    const auto *host = world()->getHostByDesignation(designation);
    return host == nullptr ? nullptr : std::make_unique<CmcHost>(*host);
}

bool CmcCore::all_of_hosts(
    const std::function<bool(const IHost &)> &pred) const {
    const auto &hosts = world()->config_.hosts_;
    return std::all_of(hosts.cbegin(), hosts.cend(), [&pred](const auto &host) {
        return pred(CmcHost{*host});
    });
}

bool CmcCore::all_of_services(
    const std::function<bool(const IService &)> &pred) const {
    return all_of_hosts([&pred](const auto &host) {
        return host.all_of_services(
            [&pred](const auto &svc) { return pred(svc); });
    });
}

const IService *CmcCore::find_service(
    const std::string &host_name,
    const std::string &service_description) const {
    const auto *svc = world()->getService(host_name, service_description);
    return svc == nullptr ? nullptr : &svc->iservice();
}

const IContactGroup *CmcCore::find_contactgroup(const std::string &name) const {
    const auto *cg = world()->getContactGroup(name);
    return cg == nullptr ? nullptr : &cg->icontactgroup();
}

const IServiceGroup *CmcCore::find_servicegroup(const std::string &name) const {
    const auto *sg = world()->getServiceGroup(name);
    return sg == nullptr ? nullptr : &sg->iservicegroup();
}

const IContact *CmcCore::find_contact(const std::string &name) const {
    const auto *c = world()->getContact(name);
    return c == nullptr ? nullptr : &c->icontact();
}

bool CmcCore::all_of_contacts(
    const std::function<bool(const IContact &)> &pred) const {
    const auto &contacts = world()->config_.contacts_;
    return std::all_of(
        contacts.cbegin(), contacts.cend(),
        [&pred](const auto &entry) { return pred(entry.second->icontact()); });
}

std::unique_ptr<const User> CmcCore::find_user(const std::string &name) const {
    if (const auto *ctc = find_contact(name)) {
        return std::make_unique<AuthUser>(
            *ctc, config()->livestatus_.service_authorization_,
            config()->livestatus_.group_authorization_,
            [this](const auto &n) { return this->find_contactgroup(n); });
    }
    return std::make_unique<UnknownUser>();
}

std::chrono::system_clock::time_point CmcCore::last_logfile_rotation() const {
    return core_->monitoring_history_.last_rotation_;
}

std::chrono::system_clock::time_point CmcCore::last_config_change() const {
    return core_->_last_config_change;
}

size_t CmcCore::maxLinesPerLogFile() const {
    return config()->history_.max_lines_per_file_;
}

Command CmcCore::find_command(const std::string &name) const {
    return Command{._name = name, ._command_line = ""};
}

std::vector<Command> CmcCore::commands() const {
    const auto &cmds = world()->config_.commands_;
    std::vector<Command> commands{cmds.size()};
    std::transform(cmds.cbegin(), cmds.cend(), commands.begin(),
                   [](const auto &cmd) {
                       return Command{._name = cmd, ._command_line = ""};
                   });
    return commands;
}

std::vector<std::unique_ptr<const IComment>> CmcCore::comments_unlocked(
    const IHost &host) const {
    const auto &h = static_cast<const CmcHost &>(host).handle();
    std::vector<std::unique_ptr<const IComment>> result;
    core_->_comments.all_of_unlocked([&result, &h](const Comment &c) {
        if (c._object == &h) {
            result.emplace_back(std::make_unique<CmcComment>(c));
        }
        return true;
    });
    return result;
}

std::vector<std::unique_ptr<const IComment>> CmcCore::comments(
    const IHost &host) const {
    const auto &h = static_cast<const CmcHost &>(host).handle();
    std::vector<std::unique_ptr<const IComment>> result;
    core_->_comments.all_of([&result, &h](const Comment &c) {
        if (c._object == &h) {
            result.emplace_back(std::make_unique<CmcComment>(c));
        }
        return true;
    });
    return result;
}

std::vector<std::unique_ptr<const IComment>> CmcCore::comments_unlocked(
    const IService &service) const {
    const auto &s = static_cast<const CmcService &>(service).handle();
    std::vector<std::unique_ptr<const IComment>> result;
    core_->_comments.all_of_unlocked([&result, &s](const Comment &c) {
        if (c._object == &s) {
            result.emplace_back(std::make_unique<CmcComment>(c));
        }
        return true;
    });
    return result;
}

std::vector<std::unique_ptr<const IComment>> CmcCore::comments(
    const IService &service) const {
    const auto &s = static_cast<const CmcService &>(service).handle();
    std::vector<std::unique_ptr<const IComment>> result;
    core_->_comments.all_of([&result, &s](const Comment &c) {
        if (c._object == &s) {
            result.emplace_back(std::make_unique<CmcComment>(c));
        }
        return true;
    });
    return result;
}

bool CmcCore::all_of_comments(
    const std::function<bool(const IComment &)> &pred) const {
    return core_->_comments.all_of(
        [&pred](const Comment &c) { return pred(CmcComment{c}); });
}

std::vector<std::unique_ptr<const IDowntime>> CmcCore::downtimes_unlocked(
    const IHost &host) const {
    const auto &h = static_cast<const CmcHost &>(host).handle();
    std::vector<std::unique_ptr<const IDowntime>> result;
    core_->_downtimes.all_of_unlocked([&result, &h](const Downtime &d) {
        if (d._object == &h) {
            result.emplace_back(std::make_unique<CmcDowntime>(d));
        }
        return true;
    });
    return result;
}

std::vector<std::unique_ptr<const IDowntime>> CmcCore::downtimes(
    const IHost &host) const {
    const auto &h = static_cast<const CmcHost &>(host).handle();
    std::vector<std::unique_ptr<const IDowntime>> result;
    core_->_downtimes.all_of([&result, &h](const Downtime &d) {
        if (d._object == &h) {
            result.emplace_back(std::make_unique<CmcDowntime>(d));
        }
        return true;
    });
    return result;
}

std::vector<std::unique_ptr<const IDowntime>> CmcCore::downtimes_unlocked(
    const IService &service) const {
    const auto &s = static_cast<const CmcService &>(service).handle();
    std::vector<std::unique_ptr<const IDowntime>> result;
    core_->_downtimes.all_of_unlocked([&result, &s](const Downtime &d) {
        if (d._object == &s) {
            result.emplace_back(std::make_unique<CmcDowntime>(d));
        }
        return true;
    });
    return result;
}

std::vector<std::unique_ptr<const IDowntime>> CmcCore::downtimes(
    const IService &service) const {
    const auto &s = static_cast<const CmcService &>(service).handle();
    std::vector<std::unique_ptr<const IDowntime>> result;
    core_->_downtimes.all_of([&result, &s](const Downtime &d) {
        if (d._object == &s) {
            result.emplace_back(std::make_unique<CmcDowntime>(d));
        }
        return true;
    });
    return result;
}

bool CmcCore::all_of_downtimes(
    const std::function<bool(const IDowntime &)> &pred) const {
    return core_->_downtimes.all_of(
        [&pred](const Downtime &d) { return pred(CmcDowntime{d}); });
}

bool CmcCore::all_of_timeperiods(
    const std::function<bool(const ITimeperiod &)> &pred) const {
    const auto &tps = world()->config_.timeperiods_;
    return std::all_of(tps.cbegin(), tps.cend(), [&pred](const auto &tp) {
        return pred(CmcTimeperiod{*tp.second});
    });
}

bool CmcCore::all_of_contact_groups(
    const std::function<bool(const IContactGroup &)> &pred) const {
    const auto &cgs = world()->config_.contact_groups_;
    return std::all_of(cgs.cbegin(), cgs.cend(), [&pred](const auto &entry) {
        return pred(entry.second->icontactgroup());
    });
}

bool CmcCore::all_of_host_groups(
    const std::function<bool(const IHostGroup &)> &pred) const {
    const auto &hgs = world()->config_.host_groups_;
    return std::all_of(hgs.cbegin(), hgs.cend(), [&pred](const auto &hg) {
        return pred(CmcHostGroup{*hg.second});
    });
}

bool CmcCore::all_of_service_groups(
    const std::function<bool(const IServiceGroup &)> &pred) const {
    const auto &sgs = world()->config_.service_groups_;
    return std::all_of(sgs.cbegin(), sgs.cend(), [&pred](const auto &sg) {
        return pred(CmcServiceGroup{*sg.second});
    });
}

bool CmcCore::mkeventdEnabled() const {
    return config()->general_.event_console_enabled_;
}

int32_t CmcCore::pid() const { return core_->_pid; }

std::unique_ptr<const IGlobalFlags> CmcCore::globalFlags() const {
    return std::make_unique<CmcGlobalFlags>(*core_->globalState());
}

std::unique_ptr<const IPaths> CmcCore::paths() const {
    return std::make_unique<CmcPaths>(*config());
}

std::chrono::system_clock::time_point CmcCore::programStartTime() const {
    return last_config_change();
}
std::chrono::system_clock::time_point CmcCore::lastCommandCheckTime() const {
    return std::chrono::system_clock::from_time_t(0);
}
int32_t CmcCore::intervalLength() const { return 60; }

int32_t CmcCore::maxLongOutputSize() const {
    return world()->config_.global_config_.general_.max_long_output_size_;
}

int32_t CmcCore::numHosts() const { return world()->num_hosts_; }

int32_t CmcCore::numServices() const { return world()->num_services_; }

std::string CmcCore::programVersion() const { return "Check_MK " VERSION; }

std::string CmcCore::edition() const { return config()->general_.edition_; }

int32_t CmcCore::externalCommandBufferSlots() const { return 0; }

int32_t CmcCore::externalCommandBufferUsage() const {
    return core_->_livestatus_ext.queueSize();
}

int32_t CmcCore::externalCommandBufferMax() const {
    return core_->_livestatus_ext.commandCount();
}

int32_t CmcCore::livestatusActiveConnectionsNum() const {
    return core_->_livestatus_active_connections.load();
    ;
}

std::string CmcCore::livestatusVersion() const { return VERSION; }

int32_t CmcCore::livestatusQueuedConnectionsNum() const { return 0; }

int32_t CmcCore::livestatusThreadsNum() const {
    return core_->_livestatus_threads;
}

double CmcCore::livestatusUsage() const {
    return core_->_avg_livestatus_usage.get();
}

double CmcCore::averageLatencyGeneric() const {
    return core_->_helper_pool_generic.averageLatency();
}

double CmcCore::averageLatencyRealTime() const {
    return core_->_helper_pool_real_time.averageLatency();
}

double CmcCore::averageLatencyFetcher() const {
    return core_->_helper_pool_fetcher.averageLatency();
}

double CmcCore::averageLatencyChecker() const {
    return core_->_helper_pool_checker.averageLatency();
}

double CmcCore::helperUsageGeneric() const {
    return core_->_helper_pool_generic.averageUsage();
}

double CmcCore::helperUsageRealTime() const {
    return core_->_helper_pool_real_time.averageUsage();
}

double CmcCore::helperUsageFetcher() const {
    return core_->_helper_pool_fetcher.averageUsage();
}

double CmcCore::helperUsageChecker() const {
    return core_->_helper_pool_checker.averageUsage();
}

bool CmcCore::hasEventHandlers() const {
    return config()->alert_handler_.enabled_;
}

double CmcCore::averageRunnableJobsFetcher() const {
    return core_->_average_runnable_jobs_fetcher.get();
}

double CmcCore::averageRunnableJobsChecker() const {
    return core_->_average_runnable_jobs_checker.get();
}

std::chrono::system_clock::time_point CmcCore::stateFileCreatedTime() const {
    return core_->globalState()->getStateFileCreated();
}

Encoding CmcCore::dataEncoding() const { return Encoding::utf8; }

size_t CmcCore::maxResponseSize() const {
    return config()->livestatus_.max_response_size_;
}

size_t CmcCore::maxCachedMessages() const {
    return config()->history_.max_cached_messages_;
}

Logger *CmcCore::loggerCore() const { return core_->_logger_core; }

Logger *CmcCore::loggerLivestatus() const { return core_->loggerLivestatus(); }

Logger *CmcCore::loggerRRD() const { return core_->_logger_rrd; }

Triggers &CmcCore::triggers() { return core_->_triggers; }

size_t CmcCore::numQueuedNotifications() const {
    return core_->notification_spooler_.size();
}

size_t CmcCore::numQueuedAlerts() const { return core_->alert_spooler_.size(); }

size_t CmcCore::numCachedLogMessages() {
    return core_->_store.numCachedLogMessages();
}

MetricLocation CmcCore::metricLocation(const std::string &host_name,
                                       const std::string &service_description,
                                       const Metric::Name &var) const {
    const Metric::MangledName v{var};
    return service_description == dummy_service_description()
               ? world()->getHost(host_name)->metricLocation(v)
               : world()
                     ->getService(host_name, service_description)
                     ->metricLocation(v);
}

bool CmcCore::isPnpGraphPresent(const IHost &host) const {
    const auto &h = static_cast<const CmcHost &>(host).handle();
    return !h.rrdInfo().names.empty();
}

bool CmcCore::isPnpGraphPresent(const IService &service) const {
    const auto &s = static_cast<const CmcService &>(service).handle();
    return !s.rrdInfo().names.empty();
}

namespace {
std::vector<std::string> toMetrics(const RRDInfo &rrd_info) {
    std::vector<std::string> metrics;
    const auto &names = rrd_info.names;
    std::transform(
        names.cbegin(), names.cend(), std::back_inserter(metrics),
        [](const Metric::MangledName &name) { return name.string(); });
    return metrics;
}
}  // namespace

std::vector<std::string> CmcCore::metrics(const IHost &host) const {
    const auto &h = static_cast<const CmcHost &>(host).handle();
    if (h.isEnabled(State::Enabling::Key::performance_data)) {
        return toMetrics(h.rrdInfo());
    }
    return {};
}

std::vector<std::string> CmcCore::metrics(const IService &service) const {
    const auto &s = static_cast<const CmcService &>(service).handle();
    if (s.isEnabled(State::Enabling::Key::performance_data)) {
        return toMetrics(s.rrdInfo());
    }
    return {};
}

bool CmcCore::pnp4nagiosEnabled() const {
    return config()->rrd_.pnp4nagios_enabled_;
}
