// Copyright (C) 2023 Checkmk GmbH - License: GNU General Public License v2
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/CmcHost.h"

#include <algorithm>
#include <array>
#include <charconv>
#include <memory>
#include <optional>
#include <ratio>
#include <unordered_set>
#include <utility>

#include "cmc/CmcContactGroup.h"
#include "cmc/CmcHostGroup.h"
#include "cmc/CmcService.h"
#include "cmc/ContactList.h"
#include "cmc/Host.h"
#include "cmc/Object.h"
#include "cmc/State.h"
#include "cmc/Timeperiod.h"
#include "cmc/cmc.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/LogEntry.h"

class Service;
class HostGroup;

using namespace std::string_literals;
using minutes_d = std::chrono::duration<double, std::ratio<60>>;

CmcHost::CmcHost(const Host &host) : host_{host} {}

[[nodiscard]] bool CmcHost::hasContact(const IContact &contact) const {
    return host_.hasContact(contact.name());
}

[[nodiscard]] const void *CmcHost::handleForStateHistory() const {
    return &host_;
}

[[nodiscard]] std::string CmcHost::notificationPeriodName() const {
    return host_.notificationPeriod()->name();
}

[[nodiscard]] std::string CmcHost::servicePeriodName() const {
    return host_.servicePeriod()->name();
}

[[nodiscard]] std::string CmcHost::name() const { return host_.name(); }

[[nodiscard]] std::string CmcHost::display_name() const {
    return host_.lookup({}, "display_name");
}

[[nodiscard]] std::string CmcHost::alias() const {
    return host_.lookup({}, "alias");
}

[[nodiscard]] std::string CmcHost::ip_address() const {
    return host_.lookup({}, "address");
}

[[nodiscard]] std::string CmcHost::check_command() const {
    return host_._check_command_name;
}

[[nodiscard]] std::string CmcHost::check_command_expanded() const {
    return host_.lookup({}, "check_command_expanded");
}

[[nodiscard]] std::string CmcHost::event_handler() const { return {}; }

[[nodiscard]] std::string CmcHost::notification_period() const {
    return host_._notification_period->name();
}

[[nodiscard]] std::string CmcHost::check_period() const {
    return host_._check_period->name();
}

[[nodiscard]] std::string CmcHost::notes() const {
    return host_.lookup({}, "notes");
}

[[nodiscard]] std::string CmcHost::notes_expanded() const {
    return host_.lookup({}, "notes_expanded");
}

[[nodiscard]] std::string CmcHost::notes_url() const {
    return host_.lookup({}, "notes_url");
}

[[nodiscard]] std::string CmcHost::notes_url_expanded() const {
    return host_.lookup({}, "notes_url_expanded");
}

[[nodiscard]] std::string CmcHost::action_url() const {
    return host_.lookup({}, "action_url");
}

[[nodiscard]] std::string CmcHost::action_url_expanded() const {
    return host_.lookup({}, "action_url_expanded");
}

[[nodiscard]] std::string CmcHost::plugin_output() const {
    return host_.state()->check_result_.output_;
}

[[nodiscard]] std::string CmcHost::perf_data() const {
    return host_.state()->check_result_.perf_data_;
}

[[nodiscard]] std::string CmcHost::icon_image() const {
    return host_.lookup({}, "icon_image");
}

[[nodiscard]] std::string CmcHost::icon_image_expanded() const {
    return host_.lookup({}, "icon_image_expanded");
}

[[nodiscard]] std::string CmcHost::icon_image_alt() const {
    return host_.lookup({}, "icon_image_alt");
}

[[nodiscard]] std::string CmcHost::status_map_image() const {
    return host_.lookup({}, "statusmap_image");
}

[[nodiscard]] std::string CmcHost::long_plugin_output() const {
    return host_.state()->check_result_.long_output_;
}

[[nodiscard]] int32_t CmcHost::initial_state() const {
    return host_.state()->initial_state_;
}

[[nodiscard]] int32_t CmcHost::max_check_attempts() const {
    return host_._max_check_attempts;
}

[[nodiscard]] bool CmcHost::flap_detection_enabled() const {
    return host_.isEnabled(State::Enabling::Key::flap_detection);
}

[[nodiscard]] bool CmcHost::check_freshness() const {
    return host_.isEnabled(State::Enabling::Key::freshness_checks);
}

[[nodiscard]] bool CmcHost::process_performance_data() const {
    return host_.isEnabled(State::Enabling::Key::performance_data);
}

[[nodiscard]] bool CmcHost::accept_passive_host_checks() const {
    return host_.isEnabled(State::Enabling::Key::passive_checks);
}

[[nodiscard]] int32_t CmcHost::event_handler_enabled() const { return 0; }

[[nodiscard]] int32_t CmcHost::acknowledgement_type() const {
    return static_cast<int32_t>(host_.acknowledgementType());
}

[[nodiscard]] int32_t CmcHost::check_type() const { return host_._check_type; }

[[nodiscard]] int32_t CmcHost::last_state() const {
    return host_.state()->last_state_;
}

[[nodiscard]] int32_t CmcHost::last_hard_state() const {
    return host_.state()->hard_state_.last_;
}
[[nodiscard]] int32_t CmcHost::current_attempt() const {
    return host_.state()->current_attempt_;
}

[[nodiscard]] std::chrono::system_clock::time_point CmcHost::last_notification()
    const {
    return host_.state()->notification_.last_time_;
}

[[nodiscard]] std::chrono::system_clock::time_point CmcHost::next_notification()
    const {
    return host_.state()->notification_.next_time_;
}

[[nodiscard]] std::chrono::system_clock::time_point CmcHost::next_check()
    const {
    return host_.state()->next_check_;
}

[[nodiscard]] std::chrono::system_clock::time_point
CmcHost::last_hard_state_change() const {
    return host_.state()->hard_state_.last_time_;
}

[[nodiscard]] bool CmcHost::has_been_checked() const {
    return host_.state()->has_been_checked_;
}

[[nodiscard]] int32_t CmcHost::current_notification_number() const {
    return host_.state()->notification_.current_number_;
}

[[nodiscard]] int32_t CmcHost::pending_flex_downtime() const {
    return host_.state()->downtime_.pending_flex_;
}

[[nodiscard]] int32_t CmcHost::total_services() const {
    return host_._services.size();
}

[[nodiscard]] bool CmcHost::notifications_enabled() const {
    return host_.isEnabled(State::Enabling::Key::notifications);
}

[[nodiscard]] bool CmcHost::problem_has_been_acknowledged() const {
    return host_.acknowledged();
}

[[nodiscard]] int32_t CmcHost::current_state() const {
    return host_.state()->current_state_;
}

[[nodiscard]] int32_t CmcHost::hard_state() const {
    if (host_.isCurrentStateOK()) {
        return 0;
    }
    const auto *const state = host_.state();
    return state->state_type_ == StateType::hard ? state->current_state_
                                                 : state->hard_state_.last_;
}

[[nodiscard]] int32_t CmcHost::state_type() const {
    return static_cast<int32_t>(host_.state()->state_type_);
}

[[nodiscard]] int32_t CmcHost::no_more_notifications() const { return 0; }

[[nodiscard]] int32_t CmcHost::check_flapping_recovery_notification() const {
    return 1;
}

[[nodiscard]] std::chrono::system_clock::time_point CmcHost::last_check()
    const {
    return host_.state()->last_check_;
}

[[nodiscard]] std::chrono::system_clock::time_point CmcHost::last_state_change()
    const {
    return host_.state()->last_state_change_;
}

[[nodiscard]] std::chrono::system_clock::time_point CmcHost::last_time_up()
    const {
    return host_.state()->last_time_state_[static_cast<int>(HostState::up)];
}

[[nodiscard]] std::chrono::system_clock::time_point CmcHost::last_time_down()
    const {
    return host_.state()->last_time_state_[static_cast<int>(HostState::down)];
}

[[nodiscard]] std::chrono::system_clock::time_point
CmcHost::last_time_unreachable() const {
    return host_.state()
        ->last_time_state_[static_cast<int>(HostState::unreachable)];
}

[[nodiscard]] bool CmcHost::is_flapping() const {
    return host_.state()->flappiness_.is_flapping_;
}

[[nodiscard]] int32_t CmcHost::scheduled_downtime_depth() const {
    return host_.state()->downtime_.scheduled_depth_;
}

[[nodiscard]] bool CmcHost::is_executing() const {
    return host_.state()->is_helper_running_;
}

[[nodiscard]] bool CmcHost::active_checks_enabled() const {
    return host_.isEnabled(State::Enabling::Key::active_checks);
}

[[nodiscard]] int32_t CmcHost::check_options() const { return 0; }

[[nodiscard]] int32_t CmcHost::obsess_over_host() const { return 0; }

[[nodiscard]] uint32_t CmcHost::modified_attributes() const {
    return host_._state->enabling_.modified_;
}

[[nodiscard]] double CmcHost::check_interval() const {
    return mk::ticks<minutes_d>(host_._check_interval);
}

[[nodiscard]] double CmcHost::retry_interval() const {
    return mk::ticks<minutes_d>(host_._retry_interval);
}

[[nodiscard]] double CmcHost::notification_interval() const {
    return mk::ticks<minutes_d>(host_._notification_interval);
}

[[nodiscard]] double CmcHost::first_notification_delay() const {
    return mk::ticks<minutes_d>(host_._first_notification_delay);
}

[[nodiscard]] double CmcHost::low_flap_threshold() const { return 0.0; }

[[nodiscard]] double CmcHost::high_flap_threshold() const { return 0.0; }

[[nodiscard]] double CmcHost::x_3d() const { return double_attr("x_3d"); }

[[nodiscard]] double CmcHost::y_3d() const { return double_attr("y_3d"); }

[[nodiscard]] double CmcHost::z_3d() const { return double_attr("z_3d"); }

[[nodiscard]] double CmcHost::latency() const {
    return mk::ticks<std::chrono::duration<double>>(host_.state()->latency_);
}

[[nodiscard]] double CmcHost::execution_time() const {
    return mk::ticks<std::chrono::duration<double>>(
        host_.state()->execution_time_);
}

[[nodiscard]] double CmcHost::percent_state_change() const { return 0.0; }

[[nodiscard]] double CmcHost::staleness() const { return ::staleness(host_); }

[[nodiscard]] double CmcHost::flappiness() const {
    return host_.state()->flappiness_.value_;
}

[[nodiscard]] bool CmcHost::in_notification_period() const {
    return host_._notification_period->isActive();
}

[[nodiscard]] bool CmcHost::in_check_period() const {
    return host_._check_period->isActive();
}

[[nodiscard]] bool CmcHost::in_service_period() const {
    return host_._service_period->isActive();
}

[[nodiscard]] std::vector<std::string> CmcHost::contacts() const {
    return host_._contact_list->contactNames();
}

[[nodiscard]] Attributes CmcHost::attributes(AttributeKind kind) const {
    return host_.attributes(kind);
}

[[nodiscard]] std::string CmcHost::filename() const {
    return host_.lookup(AttributeKind::custom_variables, "FILENAME");
}

[[nodiscard]] std::string CmcHost::notification_postponement_reason() const {
    return host_.state()->notification_.postponement_reason_;
}

[[nodiscard]] int32_t CmcHost::previous_hard_state() const {
    return host_.state()->hard_state_.previous_;
}

[[nodiscard]] int32_t CmcHost::smartping_timeout() const {
    return mk::ticks<std::chrono::milliseconds>(host_._smart_ping_timeout);
}

bool CmcHost::all_of_services(
    std::function<bool(const IService &)> pred) const {
    return std::all_of(host_._services.cbegin(), host_._services.cend(),
                       [&pred](const std::unique_ptr<Service> &s) {
                           return pred(CmcService{*s});
                       });
}

bool CmcHost::all_of_labels(
    const std::function<bool(const Attribute &)> &pred) const {
    // TODO(sp) Avoid construction of temporary map
    auto labels = host_.attributes(AttributeKind::labels);
    return std::all_of(
        labels.cbegin(), labels.cend(),
        [&pred](const std::pair<std::string, std::string> &label) {
            return pred(Attribute{label.first, label.second});
        });
}

bool CmcHost::all_of_parents(std::function<bool(const IHost &)> pred) const {
    return std::all_of(host_._parents.cbegin(), host_._parents.cend(),
                       [&pred](const Host *h) { return pred(CmcHost{*h}); });
}

bool CmcHost::all_of_children(std::function<bool(const IHost &)> pred) const {
    return std::all_of(host_._children.cbegin(), host_._children.cend(),
                       [&pred](const Host *h) { return pred(CmcHost{*h}); });
}

bool CmcHost::all_of_host_groups(
    std::function<bool(const IHostGroup &)> pred) const {
    return std::all_of(
        host_.groups().cbegin(), host_.groups().cend(),
        [&pred](const HostGroup *hg) { return pred(CmcHostGroup{*hg}); });
}

bool CmcHost::all_of_contact_groups(
    std::function<bool(const IContactGroup &)> pred) const {
    const auto groups = host_._contact_list->groups();
    return std::all_of(groups.cbegin(), groups.cend(), [&pred](const auto *cg) {
        return pred(CmcContactGroup{*cg});
    });
}

[[nodiscard]] double CmcHost::double_attr(const std::string &name) const {
    const auto value = host_.lookup({}, name);
    double result{0.0};
    std::from_chars(value.data(), value.data() + value.size(), result);
    return result;
}
