// Copyright (C) 2023 Checkmk GmbH - License: GNU General Public License v2
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/CmcService.h"

#include <algorithm>
#include <array>
#include <optional>
#include <unordered_set>
#include <utility>

#include "cmc/CmcContactGroup.h"
#include "cmc/CmcServiceGroup.h"
#include "cmc/ContactList.h"
#include "cmc/Host.h"
#include "cmc/Object.h"
#include "cmc/Service.h"
#include "cmc/State.h"
#include "cmc/Timeperiod.h"
#include "cmc/cmc.h"
#include "livestatus/ChronoUtils.h"

CmcService::CmcService(const Service &service) : service_{service} {}

const void *CmcService::handleForStateHistory() const { return &service_; }
const IHost &CmcService::host() const { return service_.host_.ihost(); }
std::string CmcService::host_name() const { return service_.host_.name(); }
bool CmcService::hasContact(const IContact &contact) const {
    return service_.hasContact(contact.name());
}
std::string CmcService::servicePeriodName() const {
    return service_.servicePeriod()->name();
}
bool CmcService::in_custom_time_period() const {
    return service_._service_period->isActive();
}

std::string CmcService::name() const { return service_.name(); }
std::string CmcService::description() const { return name(); }
std::string CmcService::display_name() const {
    return service_.lookup({}, "display_name");
}
std::string CmcService::check_command() const {
    return service_._check_command_name;
}
std::string CmcService::check_command_expanded() const {
    return service_.lookup({}, "check_command_expanded");
}
std::filesystem::path CmcService::robotmk_dir() const {
    return std::filesystem::path{host_name()} / description();
}
std::string CmcService::event_handler() const { return ""; }
std::string CmcService::plugin_output() const {
    return service_.state()->check_result_.output_;
}
std::string CmcService::long_plugin_output() const {
    return service_.state()->check_result_.long_output_;
}
std::string CmcService::perf_data() const {
    return service_.state()->check_result_.perf_data_;
}
std::string CmcService::notificationPeriodName() const {
    return service_.notificationPeriod()->name();
}
std::string CmcService::check_period() const {
    return service_._check_period->name();
}
std::string CmcService::notes() const { return service_.lookup({}, "notes"); }
std::string CmcService::notes_expanded() const {
    return service_.lookup({}, "notes_expanded");
}
std::string CmcService::notes_url() const {
    return service_.lookup({}, "notes_url");
}
std::string CmcService::notes_url_expanded() const {
    return service_.lookup({}, "notes_url_expanded");
}
std::string CmcService::action_url() const {
    return service_.lookup({}, "action_url");
}
std::string CmcService::action_url_expanded() const {
    return service_.lookup({}, "action_url_expanded");
}
std::string CmcService::icon_image() const {
    return service_.lookup({}, "icon_image");
}
std::string CmcService::icon_image_expanded() const {
    return service_.lookup({}, "icon_image_expanded");
}
std::string CmcService::icon_image_alt() const {
    return service_.lookup({}, "icon_image_alt");
}
int32_t CmcService::initial_state() const {
    return service_.state()->initial_state_;
}
int32_t CmcService::max_check_attempts() const {
    return service_._max_check_attempts;
}
int32_t CmcService::current_attempt() const {
    return service_.state()->current_attempt_;
}
int32_t CmcService::current_state() const {
    return service_.state()->current_state_;
}
bool CmcService::has_been_checked() const {
    return service_.state()->has_been_checked_;
}
int32_t CmcService::last_state() const {
    return service_.state()->last_state_;
}  // last_state
int32_t CmcService::last_hard_state() const {
    return service_.state()->hard_state_.last_;
}
int32_t CmcService::state_type() const {
    return static_cast<int32_t>(service_.state()->state_type_);
}
int32_t CmcService::check_type() const { return service_._check_type; }
bool CmcService::problem_has_been_acknowledged() const {
    return service_.acknowledged();
}
int32_t CmcService::acknowledgement_type() const {
    return static_cast<int32_t>(service_.acknowledgementType());
}
bool CmcService::no_more_notifications() const { return false; }
std::chrono::system_clock::time_point CmcService::last_time_ok() const {
    return service_.state()
        ->last_time_state_[static_cast<int>(ServiceState::ok)];
}
std::chrono::system_clock::time_point CmcService::last_time_warning() const {
    return service_.state()
        ->last_time_state_[static_cast<int>(ServiceState::warning)];
}
std::chrono::system_clock::time_point CmcService::last_time_critical() const {
    return service_.state()
        ->last_time_state_[static_cast<int>(ServiceState::critical)];
}
std::chrono::system_clock::time_point CmcService::last_time_unknown() const {
    return service_.state()
        ->last_time_state_[static_cast<int>(ServiceState::unknown)];
}
std::chrono::system_clock::time_point CmcService::last_check() const {
    return service_.state()->last_check_;
}
std::chrono::system_clock::time_point CmcService::next_check() const {
    return service_.state()->next_check_;
}
std::chrono::system_clock::time_point CmcService::last_notification() const {
    return service_.state()->notification_.last_time_;
}
std::chrono::system_clock::time_point CmcService::next_notification() const {
    return service_.state()->notification_.next_time_;
}
int32_t CmcService::current_notification_number() const {
    return service_.state()->notification_.current_number_;
}
std::chrono::system_clock::time_point CmcService::last_state_change() const {
    return service_.state()->last_state_change_;
}
std::chrono::system_clock::time_point CmcService::last_hard_state_change()
    const {
    return service_.state()->hard_state_.last_time_;
}
int32_t CmcService::scheduled_downtime_depth() const {
    return service_.state()->downtime_.scheduled_depth_;
}
bool CmcService::is_flapping() const {
    return service_.state()->flappiness_.is_flapping_;
}
bool CmcService::checks_enabled() const {
    return service_.isEnabled(State::Enabling::Key::active_checks);
}
bool CmcService::accept_passive_checks() const {
    return service_.isEnabled(State::Enabling::Key::passive_checks);
}
bool CmcService::event_handler_enabled() const { return false; }
bool CmcService::notifications_enabled() const {
    return service_.isEnabled(State::Enabling::Key::notifications);
}
bool CmcService::process_performance_data() const {
    return service_.isEnabled(State::Enabling::Key::performance_data);
}
bool CmcService::is_executing() const {
    return service_.state()->is_helper_running_;
}
bool CmcService::active_checks_enabled() const { return checks_enabled(); }
int32_t CmcService::check_options() const { return 0; }
bool CmcService::flap_detection_enabled() const {
    return service_.isEnabled(State::Enabling::Key::flap_detection);
}
bool CmcService::check_freshness() const {
    return service_.isEnabled(State::Enabling::Key::freshness_checks);
}
bool CmcService::obsess_over_service() const { return false; }
uint32_t CmcService::modified_attributes() const {
    return service_._state->enabling_.modified_;
}
int32_t CmcService::hard_state() const {
    if (service_.isCurrentStateOK()) {
        return 0;
    }
    const auto *const state = service_.state();
    return state->state_type_ == StateType::hard ? state->current_state_
                                                 : state->hard_state_.last_;
}
double CmcService::staleness() const { return ::staleness(service_); }
double CmcService::check_interval() const {
    return mk::ticks<minutes_d>(service_._check_interval);
}
double CmcService::retry_interval() const {
    return mk::ticks<minutes_d>(service_._retry_interval);
}
double CmcService::notification_interval() const {
    return mk::ticks<minutes_d>(service_._notification_interval);
}
double CmcService::first_notification_delay() const {
    return mk::ticks<minutes_d>(service_._first_notification_delay);
}
double CmcService::low_flap_threshold() const { return 0.0; }
double CmcService::high_flap_threshold() const { return 0.0; }
double CmcService::latency() const {
    return mk::ticks<std::chrono::duration<double>>(service_.state()->latency_);
}
double CmcService::execution_time() const {
    return mk::ticks<std::chrono::duration<double>>(
        service_.state()->execution_time_);
}
double CmcService::percent_state_change() const { return 0.0; }
bool CmcService::in_check_period() const {
    return service_._check_period->isActive();
}
bool CmcService::in_service_period() const {
    return service_._service_period->isActive();
}
bool CmcService::in_notification_period() const {
    return service_._notification_period->isActive();
}
std::vector<std::string> CmcService::contacts() const {
    return service_._contact_list->contactNames();
}
Attributes CmcService::attributes(AttributeKind kind) const {
    return service_.attributes(kind);
}

std::chrono::system_clock::time_point CmcService::cached_at() const {
    return service_.state()->check_result_.cached_at_;
}
int32_t CmcService::cache_interval() const {
    return mk::ticks<std::chrono::seconds>(
        service_.state()->check_result_.cache_interval_);
}
bool CmcService::in_passive_check_period() const {
    return service_.passive_check_period()->isActive();
}
std::string CmcService::passive_check_period() const {
    return service_.passive_check_period()->name();
}
double CmcService::flappiness() const {
    return service_.state()->flappiness_.value_;
}
std::string CmcService::notification_postponement_reason() const {
    return service_.state()->notification_.postponement_reason_;
}
int32_t CmcService::previous_hard_state() const {
    return service_.state()->hard_state_.previous_;
}
int32_t CmcService::pending_flex_downtime() const {
    return service_.state()->downtime_.pending_flex_;
}
bool CmcService::check_flapping_recovery_notification() const { return true; }

bool CmcService::all_of_service_groups(
    std::function<bool(const IServiceGroup &)> pred) const {
    return std::all_of(
        service_.groups().cbegin(), service_.groups().cend(),
        [&pred](const auto *sg) { return pred(CmcServiceGroup{*sg}); });
}

bool CmcService::all_of_contact_groups(
    std::function<bool(const IContactGroup &)> pred) const {
    const auto groups = service_._contact_list->groups();
    return std::all_of(groups.cbegin(), groups.cend(), [&pred](const auto *cg) {
        return pred(CmcContactGroup{*cg});
    });
}

bool CmcService::all_of_labels(
    const std::function<bool(const Attribute &)> &pred) const {
    // TODO(sp) Avoid construction of temporary map
    auto labels = service_.attributes(AttributeKind::labels);
    return std::all_of(
        labels.cbegin(), labels.cend(),
        [&pred](const std::pair<std::string, std::string> &label) {
            return pred(Attribute{label.first, label.second});
        });
}
