// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/Comments.h"

#include <algorithm>

#include "cmc/MonitoringLog.h"
#include "cmc/Object.h"
#include "cmc/State.h"
#include "livestatus/Logger.h"
#include "livestatus/Triggers.h"

void Comments::expire(std::chrono::time_point<std::chrono::system_clock> now) {
    // See also `Downtimes::expire()`.
    const std::lock_guard lg(lock_);
    std::unordered_set<key_type> expired_comment_ids;
    std::unordered_set<const Object *> expired_ack_objects;
    for (const auto &[id, comment] : comments_) {
        if (comment->isExpired(now)) {
            Notice(logger_) << "expired: " << *comment;
            switch (comment->_entry_type) {
                case Comment::Type::user:
                case Comment::Type::downtime:
                case Comment::Type::flapping:
                    expired_comment_ids.insert(id);
                    break;
                case Comment::Type::acknowledgement:
                    expired_ack_objects.insert(comment->_object);
                    break;
            }
        }
    }
    unlocked_remove(expired_comment_ids);
    unlocked_removeAcknowledgements(expired_ack_objects, "EXPIRED");
}

void Comments::removeAcknowledgements(
    const std::unordered_set<const Object *> &objects, const std::string &how) {
    const std::lock_guard lg(lock_);
    unlocked_removeAcknowledgements(objects, how);
}

void Comments::unlocked_removeAcknowledgements(
    const std::unordered_set<const Object *> &objects, const std::string &how) {
    for (const auto *object : objects) {
        object->setAcknowledgementType(AcknowledgementType::none);
    }
    std::unordered_set<Comments::key_type> comment_ids_to_remove;
    for (const auto &[id, comment] : comments_) {
        switch (comment->_entry_type) {
            case Comment::Type::user:
            case Comment::Type::downtime:
            case Comment::Type::flapping:
                break;
            case Comment::Type::acknowledgement:
                if (objects.contains(comment->_object)) {
                    monitoring_log_.acknowledgeAlert(*comment->_object, how,
                                                     comment->_author,
                                                     comment->_comment);
                    if (comment->_persistent) {
                        comment->_expire_time =
                            std::chrono::system_clock::time_point{};
                    } else {
                        comment_ids_to_remove.insert(id);
                    }
                }
                break;
        }
    }
    unlocked_remove(comment_ids_to_remove);
}

std::pair<std::string, std::string> Comments::findAcknowledgement(
    const Object &object) const {
    if (!object.acknowledged()) {
        return {};
    }

    const std::lock_guard lg(lock_);
    for (const auto &co : comments_) {
        const auto &cmnt = co.second;
        if (cmnt->_object == &object &&
            cmnt->_entry_type == Comment::Type::acknowledgement &&
            !cmnt->_persistent) {
            return {cmnt->_author, cmnt->_comment};
        }
    }

    return {};
}

bool Comments::all_of_unlocked(
    const std::function<bool(const Comment &)> &pred) const {
    return std::all_of(
        comments_.cbegin(), comments_.cend(),
        [&pred](const auto &comment) { return pred(*comment.second); });
}

bool Comments::all_of(const std::function<bool(const Comment &)> &pred) const {
    const std::lock_guard lg{lock_};
    return all_of_unlocked(pred);
}

Comments::size_type Comments::size() const {
    const std::lock_guard lg{lock_};
    return comments_.size();
}

bool Comments::migrate(const World &world) {
    const std::lock_guard lg(lock_);
    // see Scott Meyers' "Effective STL", item 9 ("Choose carefully among
    // erasing options.")
    for (auto it = comments_.begin(); it != comments_.end();) {
        if (it->second->migrate(world)) {
            ++it;
        } else {
            comments_.erase(it++);
        }
    }
    return true;
}

void Comments::insert(std::unique_ptr<Comment> comment) {
    // See also `Downtimes::insert()`.
    Informational(logger_) << "inserting " << *comment;
    const std::lock_guard lg{lock_};
    switch (comment->_entry_type) {
        case Comment::Type::user:
        case Comment::Type::downtime:
        case Comment::Type::flapping:
            break;
        case Comment::Type::acknowledgement:
            monitoring_log_.acknowledgeAlert(*comment->_object, "STARTED",
                                             comment->_author,
                                             comment->_comment);
            break;
    }
    comments_.emplace(comment->_id, std::move(comment));
    triggers_.notify_all(Triggers::Kind::comment);
}

size_t Comments::remove(const std::unordered_set<key_type> &ids) {
    // See also `Downtimes::remove()`.
    const std::lock_guard lg(lock_);
    return unlocked_remove(ids);
}

size_t Comments::unlocked_remove(const std::unordered_set<key_type> &ids) {
    size_t num_removed{0};
    for (const auto &id : ids) {
        auto it = comments_.find(id);
        if (it == comments_.end()) {
            Warning(logger_)
                << "cannot delete comment " << id << ": no such id";
        } else {
            Informational(logger_) << "removing " << *it->second;
            comments_.erase(it);
            num_removed++;
        }
    }
    if (num_removed > 0) {
        triggers_.notify_all(Triggers::Kind::comment);
    }
    return num_removed;
}

void Comments::clear() {
    const std::lock_guard lg(lock_);
    comments_.clear();
}
