// Copyright (C) 2021 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/ConfigCleaner.h"

#include <charconv>
#include <string>
#include <system_error>
#include <utility>

#include "livestatus/Logger.h"

namespace fs = std::filesystem;

ConfigCleaner::ConfigCleaner(Logger *logger, Task task)
    : logger_{logger}, task_{std::move(task)} {}

void ConfigCleaner::run() const {
    for (const auto &entry : config_root_entries()) {
        if (should_be_removed(entry)) {
            remove_config(entry);
        }
    }
}

std::vector<fs::directory_entry> ConfigCleaner::config_root_entries() const {
    std::error_code ec;
    std::vector<fs::directory_entry> entries;
    for (fs::directory_iterator it{task_.config_root, ec}; !ec && it != end(it);
         it.increment(ec)) {
        entries.push_back(*it);
    }
    if (ec) {
        Warning(logger_) << "failed to scan " << task_.config_root << ": "
                         << ec.message();
        return {};
    }
    return entries;
}

bool ConfigCleaner::should_be_removed(const fs::directory_entry &entry) const {
    std::error_code ec;
    auto is_dir = entry.is_directory(ec);
    if (ec) {
        Warning(logger_) << "failed to determine type of " << entry << ": "
                         << ec.message();
        return false;
    }
    auto filename = entry.path().filename().string();
    if (!is_dir || filename == "latest") {
        return false;
    }
    config_serial_t serial{};
    const auto *end = filename.data() + filename.size();
    if (auto [ptr, ec] = std::from_chars(filename.data(), end, serial);
        ec != std::errc{} || ptr != end) {
        Warning(logger_) << "invalid config directory name " << entry;
        return false;
    }
    // NOTE: There can be newer configurations on disk which have not been
    // activated yet, these must survive.
    return serial < task_.current_serial &&
           !task_.serials_used.contains(serial);
}

void ConfigCleaner::remove_config(const fs::directory_entry &entry) const {
    std::error_code ec;
    fs::remove_all(entry, ec);
    if (ec) {
        Warning(logger_) << "failed to remove dead configuration directory "
                         << entry << ": " << ec.message();
    } else {
        Notice(logger_) << "removed dead configuration directory " << entry;
    }
}
