// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <algorithm>
#include <array>
#include <charconv>
#include <chrono>
#include <compare>
#include <cstdint>
#include <cstdlib>
#include <ctime>
#include <filesystem>
#include <functional>
#include <map>
#include <memory>
#include <optional>
#include <ratio>
#include <sstream>
#include <stdexcept>
#include <string>
#include <string_view>
#include <system_error>
#include <unordered_set>
#include <utility>

#include "cmc/CheckResult.h"
#include "cmc/Comment.h"
#include "cmc/Comments.h"
#include "cmc/Core.h"
#include "cmc/CycleTime.h"
#include "cmc/Downtime.h"
#include "cmc/Downtimes.h"
#include "cmc/GlobalConfig.h"
#include "cmc/GlobalState.h"
#include "cmc/Host.h"
#include "cmc/LivestatusExternalCommand.h"
#include "cmc/LivestatusLocalCommand.h"
#include "cmc/MonitoringLog.h"
#include "cmc/NotifyHelper.h"
#include "cmc/Object.h"
#include "cmc/Service.h"
#include "cmc/State.h"
#include "cmc/TableCachedStatehist.h"
#include "cmc/World.h"
#include "cmc/cmc.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/CrashReport.h"
#include "livestatus/EventConsoleConnection.h"
#include "livestatus/ICore.h"
#include "livestatus/Interface.h"
#include "livestatus/Logger.h"
#include "livestatus/StringUtils.h"
#include "livestatus/Triggers.h"
#include "livestatus/global_counters.h"
#include "livestatus/mk_logwatch.h"

using namespace std::chrono_literals;
using namespace std::string_view_literals;

namespace {
std::string_view next(std::string_view &args) {
    auto field = args.substr(0, args.find(';'));
    args.remove_prefix(std::min(args.size(), field.size() + 1));
    return field;
}

template <typename T>
T parse_number(std::string_view str) {
    T value{};
    auto [ptr, ec] = std::from_chars(str.begin(), str.end(), value);
    // TODO(sp) Error handling
    return ec != std::errc{} || ptr != str.end() ? T{} : value;
}

std::chrono::system_clock::time_point parse_timestamp(std::string_view str) {
    return std::chrono::system_clock::from_time_t(parse_number<time_t>(str));
}

std::chrono::seconds parse_seconds(std::string_view str) {
    return std::chrono::seconds{parse_number<std::chrono::seconds::rep>(str)};
}
}  // namespace

void Core::answerCommandRequest(const std::string &command) {
    _livestatus_local.answerCommandRequest(command);
}

void Core::spoolCommand(const std::string &command) {
    _livestatus_ext.spoolCommand(command);
}

// TODO(ml): This and all the command* functions do not belong here.
void Core::handleCommand(const std::string &command) {
    // Do not create log entries for shadow updates. Prevent log spamming!
    if (!command.starts_with("UPDATE_SHADOW_")) {
        monitoring_log_.extCommand(command);
        Notice(_logger_core) << "Executing external command: " << command;
    }
    if (command.starts_with("EC_")) {
        handleEventConsoleCommand("COMMAND " + command.substr(3));
        return;
    }
    std::string_view args{command};
    auto tag = next(args);
    bool success = false;
    if (tag == "PROCESS_HOST_CHECK_RESULT"sv) {
        success = commandFakeCheck(args, true);
    } else if (tag == "PROCESS_SERVICE_CHECK_RESULT"sv) {
        success = commandFakeCheck(args, false);
    } else if (tag == "PROCESS_HOST_PERFDATA"sv) {
        success = commandProcessPerfdata(args, true);
    } else if (tag == "PROCESS_SVC_PERFDATA"sv) {
        success = commandProcessPerfdata(args, false);
    } else if (tag == "SCHEDULE_FORCED_HOST_CHECK"sv) {
        success = commandScheduleCheck(args, true);
    } else if (tag == "SCHEDULE_FORCED_SVC_CHECK"sv) {
        success = commandScheduleCheck(args, false);
    } else if (tag == "SEND_CUSTOM_HOST_NOTIFICATION"sv) {
        success = commandNotification(args, true);
    } else if (tag == "SEND_CUSTOM_SVC_NOTIFICATION"sv) {
        success = commandNotification(args, false);
    } else if (tag == "ADD_HOST_COMMENT"sv) {
        success = commandAddComment(args, true);
    } else if (tag == "ADD_SVC_COMMENT"sv) {
        success = commandAddComment(args, false);
    } else if (tag == "DEL_HOST_COMMENT"sv || tag == "DEL_SVC_COMMENT"sv) {
        success = commandDelComment(args);
    } else if (tag == "ACKNOWLEDGE_HOST_PROBLEM"sv) {
        success = commandAcknowledge(args, true);
    } else if (tag == "ACKNOWLEDGE_SVC_PROBLEM"sv) {
        success = commandAcknowledge(args, false);
    } else if (tag == "REMOVE_HOST_ACKNOWLEDGEMENT"sv) {
        success = commandRemoveAck(args, true);
    } else if (tag == "REMOVE_SVC_ACKNOWLEDGEMENT"sv) {
        success = commandRemoveAck(args, false);
    } else if (tag == "SCHEDULE_HOST_DOWNTIME"sv) {
        success = commandScheduleDowntime(args, true);
    } else if (tag == "SCHEDULE_SVC_DOWNTIME"sv) {
        success = commandScheduleDowntime(args, false);
    } else if (tag == "DEL_HOST_DOWNTIME"sv || tag == "DEL_SVC_DOWNTIME"sv) {
        success = commandDelDowntime(args);
    } else if (tag == "MODIFY_HOST_DOWNTIME"sv ||
               tag == "MODIFY_SVC_DOWNTIME"sv) {
        success = commandModifyDowntime(args);
        // enabling/disabling of object settings
    } else if (tag == "ENABLE_HOST_CHECK"sv) {
        success = commandEnable(args, true, true,
                                State::Enabling::Key::active_checks);
    } else if (tag == "ENABLE_SVC_CHECK"sv) {
        success = commandEnable(args, false, true,
                                State::Enabling::Key::active_checks);
    } else if (tag == "ENABLE_PASSIVE_HOST_CHECKS"sv) {
        success = commandEnable(args, true, true,
                                State::Enabling::Key::passive_checks);
    } else if (tag == "ENABLE_PASSIVE_SVC_CHECKS"sv) {
        success = commandEnable(args, false, true,
                                State::Enabling::Key::passive_checks);
    } else if (tag == "ENABLE_HOST_NOTIFICATIONS"sv) {
        success = commandEnable(args, true, true,
                                State::Enabling::Key::notifications);
    } else if (tag == "ENABLE_SVC_NOTIFICATIONS"sv) {
        success = commandEnable(args, false, true,
                                State::Enabling::Key::notifications);
    } else if (tag == "DISABLE_HOST_CHECK"sv) {
        success = commandEnable(args, true, false,
                                State::Enabling::Key::active_checks);
    } else if (tag == "DISABLE_SVC_CHECK"sv) {
        success = commandEnable(args, false, false,
                                State::Enabling::Key::active_checks);
    } else if (tag == "DISABLE_PASSIVE_HOST_CHECKS"sv) {
        success = commandEnable(args, true, false,
                                State::Enabling::Key::passive_checks);
    } else if (tag == "DISABLE_PASSIVE_SVC_CHECKS"sv) {
        success = commandEnable(args, false, false,
                                State::Enabling::Key::passive_checks);
    } else if (tag == "DISABLE_HOST_NOTIFICATIONS"sv) {
        success = commandEnable(args, true, false,
                                State::Enabling::Key::notifications);
    } else if (tag == "DISABLE_SVC_NOTIFICATIONS"sv) {
        success = commandEnable(args, false, false,
                                State::Enabling::Key::notifications);
    } else if (tag == "CHANGE_HOST_MODATTR"sv) {
        success = commandClearModattr(args, true);
    } else if (tag == "CHANGE_SVC_MODATTR"sv) {
        success = commandClearModattr(args, false);
    } else if (tag == "ENABLE_HOST_SVC_NOTIFICATIONS"sv) {
        success = commandEnableServices(args, true,
                                        State::Enabling::Key::notifications);
    } else if (tag == "DISABLE_HOST_SVC_NOTIFICATIONS"sv) {
        success = commandEnableServices(args, false,
                                        State::Enabling::Key::notifications);

        // global settings - a lot of commands
    } else if (tag == "ENABLE_EVENT_HANDLERS"sv) {
        globalState()->setEnableAlertHandlers(true);
        _triggers.notify_all(Triggers::Kind::program);
        success = true;
    } else if (tag == "DISABLE_EVENT_HANDLERS"sv) {
        globalState()->setEnableAlertHandlers(false);
        _triggers.notify_all(Triggers::Kind::program);
        success = true;
    } else if (tag == "ENABLE_NOTIFICATIONS"sv) {
        globalState()->setEnableNotifications(true);
        _triggers.notify_all(Triggers::Kind::program);
        success = true;
    } else if (tag == "DISABLE_NOTIFICATIONS"sv) {
        globalState()->setEnableNotifications(false);
        _triggers.notify_all(Triggers::Kind::program);
        success = true;
    } else if (tag == "START_EXECUTING_SVC_CHECKS"sv) {
        globalState()->setExecuteServiceChecks(true);
        _triggers.notify_all(Triggers::Kind::program);
        success = true;
    } else if (tag == "STOP_EXECUTING_SVC_CHECKS"sv) {
        globalState()->setExecuteServiceChecks(false);
        _triggers.notify_all(Triggers::Kind::program);
        success = true;
    } else if (tag == "START_EXECUTING_HOST_CHECKS"sv) {
        globalState()->setExecuteHostChecks(true);
        _triggers.notify_all(Triggers::Kind::program);
        success = true;
    } else if (tag == "STOP_EXECUTING_HOST_CHECKS"sv) {
        globalState()->setExecuteHostChecks(false);
        _triggers.notify_all(Triggers::Kind::program);
        success = true;
    } else if (tag == "ENABLE_FLAP_DETECTION"sv) {
        globalState()->setEnableFlapDetection(true);
        _triggers.notify_all(Triggers::Kind::program);
        success = true;
    } else if (tag == "DISABLE_FLAP_DETECTION"sv) {
        globalState()->setEnableFlapDetection(false);
        _triggers.notify_all(Triggers::Kind::program);
        success = true;
    } else if (tag == "ENABLE_PERFORMANCE_DATA"sv) {
        globalState()->setProcessPerformanceData(true);
        _triggers.notify_all(Triggers::Kind::program);
        success = true;
    } else if (tag == "DISABLE_PERFORMANCE_DATA"sv) {
        globalState()->setProcessPerformanceData(false);
        _triggers.notify_all(Triggers::Kind::program);
        success = true;

        // Various special commands, only known by CMC
    } else if (tag == "DEL_CRASH_REPORT"sv) {
        success = commandDelCrashReport(args);
    } else if (tag == "SAVE_STATE_INFORMATION"sv) {
        success = commandSaveState();
    } else if (tag == "RELOAD_CONFIG"sv) {
        success = commandReloadConfig();
    } else if (tag == "ROTATE_LOGFILE"sv) {
        success = commandRotateLogfile();
    } else if (tag == "LOG"sv) {
        success = commandLog(args);
    } else if (tag == "REOPEN_DAEMONLOG"sv) {
        success = commandReopenDaemonLog();
    } else if (tag == "SEGFAULT"sv) {
        success = commandSegfault();
    } else if (tag == "UPDATE_SHADOW_HOST_STATE"sv) {
        success = commandUpdateShadowState(args, true);
    } else if (tag == "UPDATE_SHADOW_SERVICE_STATE"sv) {
        success = commandUpdateShadowState(args, false);
    } else if (tag == "MK_LOGWATCH_ACKNOWLEDGE"sv) {
        success = commandMkLogwatchAcknowledge(args);

    } else {
        Warning(_logger_core)
            << "Ignoring non-implemented command '" << tag << "'";
        return;
    }
    if (success) {
        counterIncrement(Counter::commands);
        _triggers.notify_all(Triggers::Kind::command);
    } else {
        Warning(_logger_core)
            << "Ignoring command " << command << ": invalid parameters";
    }
}

namespace {
class ECTableConnection : public EventConsoleConnection {
public:
    ECTableConnection(Logger *logger, std::string path, std::string command)
        : EventConsoleConnection(logger, std::move(path))
        , command_(std::move(command)) {}

private:
    void sendRequest(std::ostream &os) override { os << command_; }
    void receiveReply(std::istream & /*is*/) override {}
    std::string command_;
};
}  // namespace

void Core::handleEventConsoleCommand(const std::string &command) const {
    if (!_mc->mkeventdEnabled()) {
        Notice(_logger_core)
            << "event console disabled, ignoring command '" << command << "'";
        return;
    }

    try {
        ECTableConnection(_mc->loggerLivestatus(),
                          _mc->paths()->event_console_status_socket(), command)
            .run();
    } catch (const std::runtime_error &err) {
        Alert(_mc->loggerLivestatus()) << err.what();
    }
}

Object *Core::getCommandObject(std::string_view &args, bool is_host) const {
    auto host = std::string{next(args)};
    if (is_host) {
        return _world->getHost(host);
    }
    return _world->getService(host, std::string{next(args)});
}

bool Core::commandFakeCheck(std::string_view args, bool is_host) {
    Object *object = getCommandObject(args, is_host);
    if (object == nullptr) {
        return false;
    }
    if (object->state()->is_helper_running_) {
        Warning(_logger_core) << "Won't fake check result of " << *object
                              << ": check is currently running";
        return false;
    }

    int return_code = parse_number<int>(next(args));
    // NOTE: The Nagios documentation for PROCESS_HOST_CHECK_RESULT
    // explicitly states that the return code is already the host state, not
    // the result code of a check result. This is somehow inconsistent with
    // the actual command name, and we previously interpreted this as a
    // result code. To avoid compatibility hell, we deviate slightly from
    // Nagios and map the codes 1 and 2 to DOWN or UNREACHABLE.
    if (is_host && return_code == 1) {
        return_code = 2;
    }

    // HACK: To be compatible with Nagios' behavior, we replace just the first
    // \n (i.e. 2 characters!) with a newline and create an active check result
    // out of it. This leaves any following \n combinations unchanged in the
    // long plugin output. Ugly...
    std::stringstream ss;
    ss << object->host()->name() << '\t'
       << mk::replace_first(std::string{args}, R"(\n)", "\n");
    auto now = std::chrono::system_clock::now();
    object->checkStarts(now);
    processCheckResult(now, CheckResult::createActive(ss, object, return_code));
    return true;
}

// This command inserts performance data from external sources
// into the graphing engine(s) that are attached to the CMC.
// Note: the host/service does *not* need to be known to the
// monitoring!
// PROCESS_HOST_PERFDATA;foohost;fooservice;foo=17;8;8;2 bar=238.7;;;100.0
bool Core::commandProcessPerfdata(std::string_view args, bool is_host) {
    auto host_name = std::string{next(args)};
    if (host_name.empty()) {
        return false;
    }
    std::string service_desc;
    if (!is_host) {
        service_desc = next(args);
        if (service_desc.empty()) {
            return false;
        }
    }
    // The rest of the command line is the performance data. Beware: that may
    // contain semicolons!
    auto *fakestate =
        new State{.acknowledgement_type_ = AcknowledgementType::none,
                  .check_result_{
                      .output_ = {},
                      .long_output_ = {},
                      .perf_data_ = std::string{args},
                      .cached_at_ = {},
                      .cache_interval_ = 0s,
                  },
                  .downtime_{.pending_flex_ = 0, .scheduled_depth_ = 0},
                  .enabling_ = {.value_ = 0, .modified_ = 0},
                  .flappiness_ = {.is_flapping_ = false, .value_ = 0.0},
                  .hard_state_ = {.last_ = 0,
                                  .previous_ = 0,
                                  .changed_ = false,
                                  .last_time_{}},
                  .notification_ = {.current_number_ = 0,
                                    .last_time_ = {},
                                    .next_time_ = {},
                                    .pending_kind_ = NotificationKind::none,
                                    .problem_id_ = 0,
                                    .problem_sent_ = false,
                                    .postponement_reason_ = {}},
                  .current_attempt_ = 1,
                  .has_been_checked_ = false,
                  .execution_time_ = std::chrono::duration<double>{0},
                  .latency_ = std::chrono::duration<double>{0},
                  .last_check_ = std::chrono::system_clock::now(),
                  .last_check_started_ = {},
                  .last_state_change_ = {},
                  .last_time_state_ = {},
                  .next_check_ = {},
                  .current_state_ = 0,
                  .initial_state_ = 0,
                  .last_state_ = 0,
                  .state_type_ = StateType::hard};

    // Create a fake object
    const auto &always = _world->getTimeperiod("fake", "24X7");
    Host fakehost{*this, *_world, host_name, always};
    if (is_host) {
        fakehost.setState(fakestate);
        fakehost._check_command_name = "no-command";
        processPerfdata(&fakehost);
    } else {
        Service fakeservice{*this, *_world, fakehost, service_desc, always};
        fakeservice.setState(fakestate);
        fakeservice._check_command_name = "no-command";
        processPerfdata(&fakeservice);
    }

    return true;
}

bool Core::commandScheduleCheck(std::string_view args, bool is_host) {
    Object *object = getCommandObject(args, is_host);
    auto schedule_time_string = std::string{next(args)};
    auto option_try = next(args);

    if (object == nullptr) {
        if (option_try == "TRY"sv) {
            return true;  // TRY means: Ignore warning
        }
        Warning(_logger_core)
            << "Invalid host/service specification or missing host/service";
        return false;
    }

    if (object->isShadow()) {
        return false;
    }

    // TODO(sp) Do we *really* want to ignore modified settings here?
    // Furthermore, do we *really* need this test at all? We are here for
    // *forced* checks only, and Nagios seems to schedule those regardless of
    // any setting.
    if (!object->getEnabledPerConfig(State::Enabling::Key::active_checks)) {
        Warning(_logger_core)
            << "Cannot schedule check for " << *object << ": it's passive";
        return false;
    }
    Notice(_logger_core) << "Manually scheduling check for " << *object;
    auto schedule_time = parse_time_t(schedule_time_string);
    if (is_host) {
        scheduleHostJob({object->host(), schedule_time});
    } else {
        scheduleServiceJob({object->service(), schedule_time});
    }
    return true;
}

bool Core::commandNotification(std::string_view args, bool is_host) {
    Object *object = getCommandObject(args, is_host);
    if (object == nullptr || object->isShadow()) {
        return false;
    }
    auto options = parse_number<int>(next(args));
    const bool broadcast_flag = (options & 1) != 0;
    const bool forced = (options & 2) != 0;
    // (increment current notification is 4, and not supported by CMC)
    auto author = std::string{next(args)};
    auto comment = std::string{next(args)};

    if (forced || (object->notificationsEnabled(loggerNotifications()) &&
                   object->inNotificationPeriod())) {
        const NotificationInfo info{*object, "CUSTOM", author, comment, {}};
        if (broadcast_flag) {
            broadcast(info);
        } else {
            notify(info);
        }
    }

    return true;
}

bool Core::commandAddComment(std::string_view args, bool is_host) {
    Object *object = getCommandObject(args, is_host);
    if (object == nullptr) {
        return false;
    }
    auto persistent = parse_number<int>(next(args)) != 0;
    auto author = std::string{next(args)};
    if (args.empty()) {
        return false;
    }
    auto text = std::string{next(args)};

    _comments.insert(std::make_unique<Comment>(
        Comment{._object = object,
                ._id = _comments.next_id_++,
                ._entry_time = std::chrono::system_clock::now(),
                ._author = author,
                ._comment = text,
                ._entry_type = Comment::Type::user,
                ._expire_time = {},
                ._persistent = persistent,
                ._sticky = false,
                ._source = Comment::Source::external}));
    return true;
}

bool Core::commandDelComment(std::string_view args) {
    return _comments.remove({parse_number<Comments::key_type>(next(args))}) > 0;
}

bool Core::commandDelCrashReport(std::string_view args) const {
    return mk::crash_report::delete_id(_mc->paths()->crash_reports_directory(),
                                       std::string{args}, _logger_core);
}

bool Core::commandAcknowledge(std::string_view args, bool is_host) {
    Object *object = getCommandObject(args, is_host);
    if (object == nullptr || object->isShadow()) {
        return false;
    }

    auto type = static_cast<AcknowledgementType>(parse_number<int>(next(args)));
    auto do_notify = parse_number<int>(next(args)) != 0;
    auto persistent = parse_number<int>(next(args)) != 0;
    auto author = std::string{next(args)};
    if (args.empty()) {
        return false;
    }
    auto text = std::string{next(args)};

    // New versions of Multisite now allow expiration of acknowledgements
    // They are optional and shipped in the comment. That way the new
    // command is compatible with old core...
    auto expiry = next(args);
    auto expiry_time = expiry.empty() ? std::chrono::system_clock::time_point{}
                                      : parse_timestamp(expiry);

    if (object->isCurrentStateOK()) {
        return true;  // silently ignore ack of non-problem
    }

    // Bug in Nagios logic: If the object already has an acknowledgement
    // then the sticky-property of that will be overwritten here and thus
    // not be consistent with the setting in the comment
    object->setAcknowledgementType(type == AcknowledgementType::sticky
                                       ? AcknowledgementType::sticky
                                       : AcknowledgementType::normal);

    _comments.insert(std::make_unique<Comment>(
        Comment{._object = object,
                ._id = _comments.next_id_++,
                ._entry_time = std::chrono::system_clock::now(),
                ._author = author,
                ._comment = text,
                ._entry_type = Comment::Type::acknowledgement,
                ._expire_time = expiry_time,
                ._persistent = persistent,
                ._sticky = type == AcknowledgementType::sticky,
                ._source = Comment::Source::internal}));

    if (do_notify && object->notificationsEnabled(loggerNotifications()) &&
        object->inNotificationPeriod()) {
        notify({*object, "ACKNOWLEDGEMENT", author, text, {}});
    }

    return true;
}

bool Core::commandRemoveAck(std::string_view args, bool is_host) {
    Object *object = getCommandObject(args, is_host);
    if (object == nullptr || object->isShadow()) {
        return false;
    }
    _comments.removeAcknowledgements({object}, "CANCELLED");
    return true;
}

namespace {
std::unique_ptr<Downtime> parseDowntimeFields(std::string_view args,
                                              const Downtime &downtime,
                                              bool existing_downtime,
                                              Logger *logger) {
    auto start_time_str = next(args);
    std::chrono::system_clock::time_point start_time;
    if (existing_downtime) {
        if (start_time_str.empty()) {
            start_time = downtime._start_time;
        } else if (start_time_str.starts_with('+')) {
            start_time =
                downtime._start_time + parse_seconds(start_time_str.substr(1));
        } else if (start_time_str.starts_with('-')) {
            start_time =
                downtime._start_time - parse_seconds(start_time_str.substr(1));
        } else {
            start_time = parse_timestamp(start_time_str);
        }
    } else {
        if (start_time_str.empty()) {
            Warning(logger) << "empty start time field";
            return {};
        }
        start_time = parse_timestamp(start_time_str);
    }

    auto end_time_str = next(args);
    std::chrono::system_clock::time_point end_time;
    if (existing_downtime) {
        if (end_time_str.empty()) {
            end_time = downtime._end_time;
        } else if (end_time_str.starts_with('+')) {
            end_time =
                downtime._end_time + parse_seconds(end_time_str.substr(1));
        } else if (end_time_str.starts_with('-')) {
            end_time =
                downtime._end_time - parse_seconds(end_time_str.substr(1));
        } else {
            end_time = parse_timestamp(end_time_str);
        }
    } else {
        if (end_time_str.empty()) {
            Warning(logger) << "empty end time field";
            return {};
        }
        end_time = parse_timestamp(end_time_str);
    }

    auto fixed_and_recurring_str = std::string{next(args)};
    if (fixed_and_recurring_str.empty() && existing_downtime) {
        fixed_and_recurring_str =
            std::to_string((downtime._fixed ? 1 : 0) |
                           (downtime._cycle_time.getKindCode() << 1));
    }
    if (fixed_and_recurring_str.empty()) {
        Warning(logger) << "empty fixed/recurring field";
        return {};
    }
    auto fixed_and_recurring = parse_number<int>(fixed_and_recurring_str);
    // The CMC allows recurring downtimes. The recurring interval is an enum
    // value (see Downtime.h) that is put into the same integer as fixed, so
    // that we keep compatible.
    const bool fixed = (fixed_and_recurring & 0x01) != 0;
    const int recurring_type = fixed_and_recurring >> 1;

    auto trigger_id = parse_number<int>(next(args));
    if (trigger_id != 0) {
        Warning(logger) << "triggering downtimes is not supported";
        return {};
    }

    auto duration_str = std::string{next(args)};
    if (duration_str.empty() && !existing_downtime) {
        Warning(logger) << "empty duration field";
        return {};
    }
    auto duration = duration_str.empty()
                        ? std::chrono::duration_cast<std::chrono::seconds>(
                              downtime._duration)
                        : parse_seconds(duration_str);

    auto author = std::string{next(args)};

    if (args.empty() && !existing_downtime) {
        Warning(logger) << "empty comment field";
        return {};
    }
    auto comment = args.empty() ? downtime._comment : std::string{next(args)};

    if (!fixed && duration == 0s) {
        Warning(logger) << "invalid flexible downtime with duration 0";
        return {};
    }

    // Disallow invalid time range or downtime that ends in the past
    auto entry_time = std::chrono::system_clock::now();
    if (start_time >= end_time) {
        Warning(logger) << "start of downtime must be before end";
        return {};
    }
    if (end_time <= entry_time) {
        Warning(logger) << "downtime end cannot be in the past";
        return {};
    }

    if (fixed) {
        duration = std::chrono::duration_cast<decltype(duration)>(end_time -
                                                                  start_time);
    }

    if (existing_downtime && downtime._status == Downtime::Status::active &&
        downtime._start_time != start_time) {
        Warning(logger) << "cannot change start of active downtime";
        return {};
    }

    return std::make_unique<Downtime>(Downtime{
        .core_ = downtime.core_,
        .logger_downtime_ = downtime.logger_downtime_,
        .logger_notification_ = downtime.logger_notification_,
        ._object = downtime._object,
        ._id = downtime._id,
        ._entry_time = entry_time,
        ._author = author,
        ._comment = comment,
        ._status = downtime._status,
        ._start_time = start_time,
        ._end_time = end_time,
        ._flex_downtime_start = downtime._flex_downtime_start,
        ._fixed = fixed,
        ._cycle_time =
            CycleTime{static_cast<std::uint16_t>(recurring_type), logger},
        ._duration = duration,
        ._hot = downtime._hot,
        ._origin_is_rule = false,
        ._rule = nullptr});
}
}  // namespace

bool Core::commandScheduleDowntime(std::string_view args, bool is_host) {
    Object *object = getCommandObject(args, is_host);
    if (object == nullptr || object->isShadow()) {
        return false;
    }
    auto downtime = Downtime{.core_ = *this,
                             .logger_downtime_ = _logger_downtime,
                             .logger_notification_ = loggerNotifications(),
                             ._object = object,
                             ._id = _downtimes.next_id_++,
                             ._entry_time = std::chrono::system_clock::now(),
                             ._author = "",
                             ._comment = "",
                             ._status = Downtime::Status::pending,
                             ._start_time = {},
                             ._end_time = {},
                             ._flex_downtime_start = {},
                             ._fixed = false,
                             ._cycle_time = {0, _logger_core},
                             ._duration = 0s,
                             ._hot = false,
                             ._origin_is_rule = false,
                             ._rule = nullptr};
    auto new_downtime =
        parseDowntimeFields(args, downtime, false, _logger_core);
    if (new_downtime == nullptr) {
        return false;
    }

    _downtimes.insert(std::move(new_downtime));
    return true;
}

bool Core::commandDelDowntime(std::string_view args) {
    return _downtimes.remove({parse_number<Downtimes::key_type>(next(args))}) >
           0;
}

bool Core::commandModifyDowntime(std::string_view args) {
    auto id = parse_number<Downtimes::key_type>(next(args));
    return _downtimes.modify(id, [this, args](const Downtime &downtime) {
        return parseDowntimeFields(args, downtime, true, _logger_core);
    });
}

bool Core::commandEnable(std::string_view args, bool is_host, bool enabled,
                         State::Enabling::Key what) {
    Object *object = getCommandObject(args, is_host);
    if (object == nullptr || object->isShadow()) {
        return false;
    }
    object->setEnabled(what, enabled);
    return true;
}

bool Core::commandClearModattr(std::string_view args, bool is_host) {
    Object *object = getCommandObject(args, is_host);
    // Note: Nagios accepts and additional value: the actual new value
    // of the modified attributes.
    if (object == nullptr || object->isShadow()) {
        return false;
    }
    object->resetEnablings();
    return true;
}

bool Core::commandEnableServices(std::string_view args, bool enabled,
                                 State::Enabling::Key what) {
    Object *object = getCommandObject(args, true);
    if (object == nullptr || object->isShadow()) {
        return false;
    }
    for (const auto &service : Host::Services(*object->host())) {
        service->setEnabled(what, enabled);
    }
    return true;
}

bool Core::commandSaveState() {
    Notice(_logger_core) << "Saving state due to external command.";
    saveState();
    return true;
}

bool Core::commandReloadConfig() {
    Notice(_logger_core) << "Reloading configuration due to external command.";
    reloadConfig();
    return true;
}

bool Core::commandRotateLogfile() {
    Notice(_logger_core)
        << "rotating monitoring history due to external command";
    rotateMonitoringLog(config()->history_);
    return true;
}

bool Core::commandLog(std::string_view args) {
    monitoring_log_.info(std::string{args});
    return true;
}

bool Core::commandReopenDaemonLog() {
    Notice(_logger_core) << "going to reopen daemon log file";
    setLogConfig();
    return true;
}

bool Core::commandSegfault() const {
    Notice(_logger_core) << "Going to segfault myself. Watch out...";
    abort();
    return false;
}

bool Core::commandUpdateShadowState(std::string_view args, bool is_host) {
    Object *object = getCommandObject(args, is_host);
    if (object == nullptr || !object->isShadow()) {
        return false;
    }

    // For the order of fields please have a look into bin/cmcdump
    auto *state = object->stateNonConst();

    const bool was_flapping = state->flappiness_.is_flapping_;
    const bool had_been_checked = state->has_been_checked_;
    const bool was_in_downtime = object->inDowntime();
    state->last_state_ = state->current_state_;

    state->has_been_checked_ = parse_number<int>(next(args)) != 0;
    // NOTE: This is the internal state of the object, i.e. OK/WARN/CRIT/UNKNOWN
    // for services resp. UP/DOWN/UNREACHABLE for hosts.
    state->current_state_ = parse_number<int>(next(args));
    state->flappiness_.is_flapping_ = parse_number<int>(next(args)) != 0;
    // The type is not specified in the command, so we just use "normal" here.
    object->setAcknowledgementType(parse_number<int>(next(args)) == 0
                                       ? AcknowledgementType::none
                                       : AcknowledgementType::normal);
    state->downtime_.scheduled_depth_ = parse_number<int>(next(args));
    state->last_check_ = parse_time_t(std::string{next(args)});
    state->last_state_change_ = parse_time_t(std::string{next(args)});
    state->current_attempt_ = parse_number<int>(next(args));
    state->state_type_ = static_cast<StateType>(parse_number<int>(next(args)));
    state->latency_ =
        std::chrono::duration<double>(parse_number<double>(next(args)));
    state->execution_time_ =
        std::chrono::duration<double>(parse_number<double>(next(args)));
    state->flappiness_.value_ = parse_number<double>(next(args));
    auto unescape = [](auto &&str) {
        // See also `bin/cmcdump::cmcdump_object_state()`
        str = mk::replace_all(str, "$3B", ";");
        str = mk::replace_all(str, "$0A", "\n");
        str = mk::replace_all(str, "$36", "$");  // must be last
        return str;
    };
    state->check_result_.output_ = unescape(std::string{next(args)});
    if (args.empty()) {
        return false;
    }
    state->check_result_.long_output_ = unescape(std::string{next(args)});

    // needs to be last because of ';' in performance data
    state->check_result_.perf_data_ = args;

    // Some derived things that are normally set in State::update()
    state->last_time_state_[state->current_state_] = state->last_check_;

    if (!had_been_checked &&
        state->has_been_checked_) {  // First check result. Log initial state
        if (config()->history_.cache_horizon_) {
            table_cached_statehist_.registerAlert(
                *object, state->current_state_, state->check_result_.output_,
                state->check_result_.long_output_);
        }
    }

    if (was_flapping != state->flappiness_.is_flapping_) {
        flappingEvent(*object);
    }

    const bool now_in_downtime = object->inDowntime();
    if (was_in_downtime && !now_in_downtime) {
        downtimeEvent(
            *object, "STOPPED",
            "Object went out of scheduled downtime on remote monitoring site");
    } else if (!was_in_downtime && now_in_downtime) {
        downtimeEvent(
            *object, "STARTED",
            "Object went into scheduled downtime on remote monitoring site");
    }

    // Do many things that need to be done after a check execution. isShadow()
    // will prevent notifications and flapping handling and stuff like that.
    state->checkForHardStateChange();
    objectStateHasBeenChanged(std::chrono::system_clock::now(), object);
    return true;
}

bool Core::commandMkLogwatchAcknowledge(std::string_view args) {
    Object *object = getCommandObject(args, true);
    if (object == nullptr) {
        return false;
    }
    if (args.empty()) {
        Warning(_logger_core)
            << "Missing file name argument for MK_LOGWATCH_ACKNOWLEDGE";
        return false;
    }
    mk_logwatch_acknowledge(_logger_core, _mc->paths()->logwatch_directory(),
                            object->name(), std::string{args});
    return true;
}
