// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/Core.h"

#include <execinfo.h>
#include <sys/resource.h>
#include <unistd.h>

#include <algorithm>
#include <compare>
#include <csignal>
#include <cstdlib>
#include <functional>
#include <initializer_list>
#include <iomanip>
#include <map>
#include <mutex>
#include <optional>
#include <sstream>
#include <stdexcept>
#include <system_error>
#include <thread>
#include <vector>

#include "cmc/CarbonConfig.h"
#include "cmc/CheckHelper.h"
#include "cmc/CheckResult.h"
#include "cmc/CmcCore.h"
#include "cmc/Downtime.h"
#include "cmc/Downtimes.h"
#include "cmc/GlobalConfig.h"
#include "cmc/Host.h"
#include "cmc/IOHelper.h"
#include "cmc/InfluxDBConfig.h"
#include "cmc/InfluxDBLineProtocol.h"
#include "cmc/InfluxDBSerializer.h"
#include "cmc/Livestatus.h"
#include "cmc/Object.h"
#include "cmc/ObjectResolver.h"
#include "cmc/PbStateLoader.h"
#include "cmc/PbStateSaver.h"
#include "cmc/RRDInfoCache.h"
#include "cmc/Service.h"
#include "cmc/TableCachedStatehist.h"
#include "cmc/Timeperiod.h"
#include "cmc/World.h"
#include "cmc/cmc.h"
#include "cmc/daemonize.h"
#include "config/v1/types.pb.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/ICore.h"
#include "livestatus/PerformanceData.h"
#include "livestatus/Poller.h"
#include "livestatus/Queue.h"
#include "livestatus/RegExp.h"
#include "livestatus/TableStateHistory.h"
#include "livestatus/global_counters.h"
#include "state/v1/types.pb.h"

using namespace std::chrono_literals;
using namespace std::string_literals;
namespace fs = std::filesystem;

// .-Init/Halt-------------------------------------------------------------.
// |               ___       _ _      ___   _       _ _                    |
// |              |_ _|_ __ (_) |_   / / | | | __ _| | |_                  |
// |               | || '_ \| | __| / /| |_| |/ _` | | __|                 |
// |               | || | | | | |_ / / |  _  | (_| | | |_                  |
// |              |___|_| |_|_|\__/_/  |_| |_|\__,_|_|\__|                 |
// |                                                                       |
// +-----------------------------------------------------------------------+
// |  Startup and Shutdown, Constructors and Destructors                   |
// '-----------------------------------------------------------------------'

namespace {
// for signal handlers
Core *fl_core;  // NOLINT(cppcoreguidelines-avoid-non-const-global-variables)
}  // namespace

namespace {
class Tempfile {
public:
    explicit Tempfile(std::filesystem::path path) noexcept
        : path_{std::move(path)} {}
    ~Tempfile() noexcept {
        // We do not care whether the file still exists or not.
        std::error_code ec;
        std::filesystem::remove(path_, ec);
    }
    Tempfile(Tempfile &&) = default;
    Tempfile &operator=(Tempfile &&) = default;
    Tempfile(const Tempfile &) = delete;
    Tempfile &operator=(const Tempfile &) = delete;

    [[nodiscard]] std::filesystem::path path() const { return path_; }
    [[nodiscard]] std::string string() const { return path_.string(); }

private:
    std::filesystem::path path_;
};
}  // namespace

// TODO(sp) Disentangle Core, World, Config and ConfigParser. Remove the
// mega-cruel hack below to set the thread-local variable g_live_world early
// enough.
Core::Core(char **argv, std::string config_filename,
           std::unique_ptr<checkmk::cmc::config::v1::Core> core_config,
           std::ofstream *log_file, Logger *logger_core,
           TrialManager trial_manager)
    : _argv(argv)
    , _pid(static_cast<int32_t>(getpid()))
    , _log_file(log_file)
    , _logger_alert(Logger::getLogger("cmk.alert"))
    , _carbon_context_logger{"cmk.carbon"}
    , _logger_carbon(&_carbon_context_logger)
    , _logger_core(logger_core)
    , _logger_downtime(Logger::getLogger("cmk.downtime"))
    , _logger_helper(Logger::getLogger("cmk.helper"))
    , _livestatus_context_logger{"cmk.livestatus"}
    , _logger_livestatus(&_livestatus_context_logger)
    , _logger_notification(Logger::getLogger("cmk.notification"))
    , _rrd_context_logger{"cmk.rrd"}
    , _logger_rrd(&_rrd_context_logger)
    , _influxdb_context_logger{"cmk.influxdb"}
    , _logger_influxdb(&_influxdb_context_logger)
    , _logger_smart_ping(Logger::getLogger("cmk.smartping"))
    , _logger_fetcher(Logger::getLogger("cmk.fetcher"))
    , _serial_token_factory(logger_core)
    , _config_filename(std::move(config_filename))
    , _reply_queue{_logger_core}

    , _world(g_live_world =
                 new World(*this, std::move(core_config), _logger_core,
                           _logger_rrd, _serial_token_factory))
    , _mc(std::make_unique<CmcCore>(this))
    , trial_manager_(std::move(trial_manager))
    , _helper_pool_generic{
          "generic pool"s, _logger_helper, _logger_core,
          [this](Object *obj) {
              if (obj->isHost()) {
                  this->scheduleHostJob({obj->host(), obj->nextCheckTime()});
              } else {
                  this->scheduleServiceJob(
                      {obj->service(), obj->nextCheckTime()});
              }
          },
          [this](std::chrono::system_clock::time_point now,
                 const CheckResult &cr) { this->processCheckResult(now, cr); }}
    , _helper_pool_checker{
          "checker pool"s, _logger_helper, _logger_core,
          [this](Service *svc) {
              this->scheduleServiceJob({svc, svc->nextCheckTime()});
          },
          [this](std::chrono::system_clock::time_point now,
                 const CheckResult &cr) { this->processCheckResult(now, cr); }}
    , _helper_pool_real_time{"real-time pool"s, _logger_helper, _logger_core,
                             [this](std::chrono::system_clock::time_point now,
                                    const CheckResult &cr) {
                                 this->processCheckResult(now, cr);
                             }}
    , _helper_pool_fetcher{
          "fetcher pool"s, _logger_fetcher, _logger_core,
          [this](Service *svc) {
              this->scheduleServiceJob({svc, svc->nextCheckTime()});
          },
          [this](std::unique_ptr<FetcherReply> reply) {
              _reply_queue.push(std::move(reply));
          }}
    , notification_spooler_{_logger_notification}
    , _notify_helper{"notification helper"s, _logger_notification,
                     [this](const Object &object) {
                         return _comments.findAcknowledgement(object);
                     },
                     monitoring_log_, notification_spooler_}
    , alert_spooler_{_logger_alert}
    , _alert_helper{
          "alert helper"s,
          [this]() { return this->globalState()->getEnableAlertHandlers(); },
          &this->_notify_helper,
          [this](const std::string &host_name,
                 const std::string &service_description) {
              return this->getAlertHandlerResponseObject(host_name,
                                                         service_description);
          },
          monitoring_log_,
          _logger_alert,
          alert_spooler_}
    , _icmp_sender("icmpsender"s, _logger_smart_ping)
    , _icmp_receiver(
          "icmpreceiver"s, _logger_smart_ping,
          [this](std::chrono::system_clock::time_point now,
                 const CheckResult &cr) { processCheckResult(now, cr); })
    , _livestatus_local{[this](const std::string &cmd) {
                            this->spoolCommand(cmd);
                        },
                        _logger_core}
    , _smart_ping_schedule(_logger_core, "smartping")
    , _always_up_schedule(_logger_core, "always-up")
    , _service_does_not_exist_schedule(_logger_core, "service-does-not-exist")
    , _cluster_host_schedule(_logger_core, "cluster host")
    , monitoring_history_{_logger_core,
                          [this]() {
                              counterIncrement(Counter::log_messages);
                              _triggers.notify_all(Triggers::Kind::log);
                          }}
    , monitoring_log_{[this](const std::string &message) {
        monitoring_history_.log(message);
    }}
    , _livestatus_threads(0)
    , _livestatus_active_connections(0)
    , _livestatus_ext{_io_context,
                      [this](const std::string &cmd) {
                          this->handleCommand(cmd);
                      },
                      _logger_core}
    , _comments{_triggers, _logger_core, monitoring_log_}
    , _downtimes{_triggers, _logger_downtime}
    , _is_shutting_down(false)
    , _should_terminate(false)
    , _should_reload(false)
    , _store(_mc.get())
    , _rrd_helper("RRD helper"s, Counter::rrdcached_overflows,
                  Counter::rrdcached_queue_usage, Counter::rrdcached_bytes_sent,
                  _logger_rrd)
    , _carbon_connection_pool("carbon", Counter::carbon_overflows,
                              Counter::carbon_queue_usage,
                              Counter::carbon_bytes_sent, _logger_carbon)
    , _influxdb_connection_pool("influxdb", Counter::influxdb_overflows,
                                Counter::influxdb_queue_usage,
                                Counter::influxdb_bytes_sent, _logger_carbon)
    , config_cleaner_{_logger_core, "config cleaner",
                      queue_join_strategy::shutdown_push_pop,
                      [](auto logger, auto &&task) {
                          ConfigCleaner(logger,
                                        std::forward<ConfigCleaner::Task>(task))
                              .run();
                      }}
    , state_saver_{_logger_core, "state saver",
                   queue_join_strategy::shutdown_push,
                   [](auto logger, auto &&task) {
                       Core::saveStateProtobuf(logger, task.core,
                                               task.state_path,
                                               task.renamed_hosts_path);
                   }}
    , _global_state{GlobalStateFields{._enable_notifications = true,
                                      ._execute_service_checks = true,
                                      ._execute_host_checks = true,
                                      ._enable_flap_detection = true,
                                      ._process_performance_data = true,
                                      ._enable_alert_handlers = true,
                                      ._notification_next_problem_id = 1,
                                      ._state_file_created = {}}}
    , _real_time_check("real-time check"s, _helper_pool_real_time,
                       _logger_helper)
    , table_cached_statehist_{_mc.get()} {
    fl_core = this;
    // Pipe for receiving signals about commands from Livestatus threads.
    _livestatus_ext.start();
    if (!_livestatus_ext.active()) {
        Emergency(_logger_core) << "livestatus is not active, exiting";
        ::exit(1);
    }
    setLogConfig();
    configureCoreDumps();

    _last_config_change = _last_state_retention = _last_flapping_check =
        std::chrono::system_clock::now();
}

Core::~Core() {
    _livestatus_ext.stop();

    monitoring_log_.coreStopping();

    _real_time_check.stop();
    _rrd_helper.stop();
    _carbon_connection_pool.stop();
    _influxdb_connection_pool.stop();
    config_cleaner_.stop();
    state_saver_.stop();

    // Check helper beenden
    Informational(_logger_core) << "Terminating check helpers...";
    _helper_pool_generic.stop();
    _helper_pool_checker.stop();
    _helper_pool_real_time.stop();
    _helper_pool_fetcher.stop();

    // NOTE: We have to stop all Livestatus threads *before* the world is
    // gone.
    _livestatus = nullptr;

    delete _world;
    deleteOldWorlds();

    Notice(_logger_core) << "Shutdown of core successful.";
    Notice(_logger_core)
        << "This is the end. Good bye and thanks for running me.";
}

// Depending on wether the state history cache is enabled or not a different
// implementation of the statehist table is being used. This can be switched
// dynamically.
void Core::switchStatehistTable() {
    if (auto cache_horizon = config()->history_.cache_horizon_) {
        _store.addTable(table_cached_statehist_);
        Notice(_logger_core)
            << "using state history cache with a horizon of "
            << (mk::ticks<std::chrono::hours>(*cache_horizon) / 24) << " days";
        table_cached_statehist_.startBuildingCache(this);
    } else {
        _store.addTable(_store.getTableStateHistory());
        Notice(_logger_core) << "state history cache will not be used";
    }
}

const GlobalConfig *Core::config() const {
    return &_world->config_.global_config_;
}

void Core::setLogConfig() const {
    const std::lock_guard lg{_logging_lock};
    Logger *logger_cmk = Logger::getLogger("cmk");
    logger_cmk->setUseParentHandlers(false);
    if (_log_file != nullptr) {
        if (_log_file->is_open()) {
            _log_file->close();
        }
        auto filename = config()->general_.log_file_;
        _log_file->open(filename, std::ofstream::app);
        if (!*_log_file) {
            const generic_error ge("could not open logfile " +
                                   filename.string());
            Warning(_logger_core) << ge;
        } else {
            logger_cmk->setHandler(std::make_unique<StreamHandler>(*_log_file));
        }
    }
    if (Handler *handler = logger_cmk->getHandler()) {
        handler->setFormatter(config()->getFormatter());
    }

    _logger_alert->setHandler(nullptr);
    _logger_alert->setLevel(config()->alert_handler_.log_level_);

    _logger_carbon->setHandler(nullptr);
    _logger_carbon->setLevel(config()->carbon_.log_level_);

    _logger_core->setHandler(nullptr);
    _logger_core->setLevel(config()->general_.log_level_core_);

    _logger_downtime->setHandler(nullptr);
    _logger_downtime->setLevel(config()->general_.log_level_downtime_);

    _logger_helper->setHandler(nullptr);
    _logger_helper->setLevel(config()->checking_.log_level_);

    _logger_fetcher->setHandler(nullptr);
    _logger_fetcher->setLevel(config()->checking_.log_level_);

    _logger_livestatus->setHandler(nullptr);
    _logger_livestatus->setLevel(config()->livestatus_.log_level_);

    _logger_notification->setHandler(nullptr);
    _logger_notification->setLevel(config()->notification_handler_.log_level_);

    _logger_rrd->setHandler(nullptr);
    _logger_rrd->setLevel(config()->rrd_.log_level_);

    _logger_influxdb->setHandler(nullptr);
    _logger_influxdb->setLevel(config()->influxdb_.log_level_);

    _logger_smart_ping->setHandler(nullptr);
    _logger_smart_ping->setLevel(config()->smart_ping_.log_level_);
}

void Core::configureCoreDumps() const {
    struct rlimit rlim;
    if (0 != getrlimit(RLIMIT_CORE, &rlim)) {
        const generic_error ge("Cannot read core dump settings");
        Warning(_logger_core) << ge;
        return;
    }
    const rlim_t target_setting =
        config()->general_.core_dumps_enabled_ ? rlim.rlim_max : 0;
    if (rlim.rlim_cur != target_setting) {
        rlim.rlim_cur = target_setting;
        if (0 != setrlimit(RLIMIT_CORE, &rlim)) {
            const generic_error ge("Cannot change core dump settings");
            Warning(_logger_core) << ge;
        }
    }

    if (config()->general_.core_dumps_enabled_) {
        removeSegfaultHandler();
    } else {
        installSegfaultHandler();
    }
}

void sig_exit(int /*unused*/) { fl_core->_should_terminate = true; }

void sig_reload(int /*unused*/) { fl_core->_should_reload = true; }

namespace {
void dump_backtrace(Logger *logger) {
    static void *callstack[64];  // static avoid stack allocation
    const int num_frames = backtrace(callstack, 64);
    if (num_frames <= 0) {
        Emergency(logger) << "Cannot get stack trace";
        return;
    }
    char **syms = backtrace_symbols(callstack, num_frames);
    if (syms == nullptr) {
        Emergency(logger) << "backtrace_symbols failed";
        return;
    }

    Emergency e(logger);
    e << "stack trace with " << num_frames << " frames follows:";
    for (int frame = 0; frame < num_frames; ++frame) {
        e << "\n" << syms[frame];
    }
    // backtrace_symbols uses malloc, so we *have* to use free.
    free(syms);  // NOLINT
}
}  // namespace

void sig_survive(int signum) {
    Logger *logger = fl_core->_logger_core;
    std::this_thread::sleep_for(500ms);  // avoid busy wait
    Emergency(logger) << "Terminated due to unexpected signal " << signum
                      << ". Restarting.";
    dump_backtrace(logger);
    mk::closefrom(3);

    // Kill all subprocesses
    kill(0, SIGTERM);

    static char exe_path[512];  // avoid stack allocation
    const ssize_t l =
        readlink("/proc/self/exe", exe_path, sizeof(exe_path) - 1);
    if (l > 0) {
        exe_path[l] = 0;
        // unblock all signals. We're still in the sighandler. And
        // the block mask will survive an execv()
        sigset_t ss;
        sigfillset(&ss);
        sigprocmask(SIG_UNBLOCK, &ss, nullptr);
        ::execv(exe_path, fl_core->_argv);
    }
    const generic_error ge("Restart failed");
    Emergency(logger) << ge;
    Emergency(logger) << "That was it. Sorry for the mess...";
    exit(66);
}

// static
void Core::installSignalHandlers() {
    struct sigaction sa;
    sigfillset(&sa.sa_mask);
    sa.sa_flags = 0;

    // normal termination
    sa.sa_handler = sig_exit;
    sigaction(SIGINT, &sa, nullptr);
    sigaction(SIGQUIT, &sa, nullptr);
    sigaction(SIGTERM, &sa, nullptr);

    // configuration reload
    sa.sa_handler = sig_reload;
    sigaction(SIGHUP, &sa, nullptr);

    // ignore pipe stuff
    // https://llvm.org/bugs/show_bug.cgi?id=29089
    sa.sa_handler = SIG_IGN;  // NOLINT
    sigaction(SIGPIPE, &sa, nullptr);
}

// static
void Core::installSegfaultHandler() {
    struct sigaction sa;
    sigfillset(&sa.sa_mask);
    sa.sa_flags = 0;

    // crash due to non-existent bugs
    sa.sa_handler = sig_survive;
    sigaction(SIGSEGV, &sa, nullptr);
    sigaction(SIGABRT, &sa, nullptr);
    sigaction(SIGBUS, &sa, nullptr);
    sigaction(SIGILL, &sa, nullptr);
    sigaction(SIGFPE, &sa, nullptr);
}

// static
void Core::removeSegfaultHandler() {
    struct sigaction sa;
    sigfillset(&sa.sa_mask);
    sa.sa_flags = 0;

    // dump core again
    sa.sa_handler = SIG_DFL;
    sigaction(SIGSEGV, &sa, nullptr);
    sigaction(SIGABRT, &sa, nullptr);
    sigaction(SIGBUS, &sa, nullptr);
    sigaction(SIGILL, &sa, nullptr);
    sigaction(SIGFPE, &sa, nullptr);
}

// .-Scheduling------------------------------------------------------------.
// |          ____       _              _       _ _                        |
// |         / ___|  ___| |__   ___  __| |_   _| (_)_ __   __ _            |
// |         \___ \ / __| '_ \ / _ \/ _` | | | | | | '_ \ / _` |           |
// |          ___) | (__| | | |  __/ (_| | |_| | | | | | | (_| |           |
// |         |____/ \___|_| |_|\___|\__,_|\__,_|_|_|_| |_|\__, |           |
// |                                                      |___/            |
// +-----------------------------------------------------------------------+
// |  Scheduling of active checks (also in State.cc/Object.cc)             |
// '-----------------------------------------------------------------------'

void Core::scheduleAllChecks() {
    Debug(_logger_core) << "rescheduling from scratch...";
    _helper_pool_generic.clearSchedule();
    _helper_pool_checker.clearSchedule();
    _helper_pool_real_time.clearSchedule();
    _helper_pool_fetcher.clearSchedule();
    _smart_ping_schedule.clear();
    _always_up_schedule.clear();
    _service_does_not_exist_schedule.clear();
    _cluster_host_schedule.clear();

    for (const auto &host : _world->config_.hosts_) {
        if (host->canBeScheduled()) {
            scheduleHostJob(
                {host.get(),
                 // check if host is currently passive
                 host->nextCheck() == std::chrono::system_clock::time_point{}
                     ? host->nextImmediateCheckTime()
                     : host->nextCheck()});
        }
        for (const auto &service : Host::Services(*host)) {
            if (service->canBeScheduled()) {
                scheduleServiceJob({service.get(), service->nextCheck()});
            }
        }
    }
    Debug(_logger_core) << "rescheduling from scratch done";
}

// Put job in correct schedule, assumes that _next_check has already been
// set!
void Core::scheduleHostJob(const Schedule<Host *>::Job &job) {
    auto *host = job.value();
    if (_logger_core->isLoggable(LogLevel::debug)) {
        Debug(_logger_core) << "scheduling " << job << " with commandline ["
                            << host->_check_command_line << "]";
    }
    if (host->_check_command_line == "@smartping") {
        if (auto t = _smart_ping_schedule.insert(job)) {
            host->setNextCheck(*t, _smart_ping_schedule.logger());
        }
    } else if (host->_check_command_line == "@up") {
        if (auto t = _always_up_schedule.insert(job)) {
            host->setNextCheck(*t, _always_up_schedule.logger());
        }
    } else if (host->_check_command_line.starts_with("@service ")) {
        const auto [host_name, description] = host->serviceFromRouting();
        if (auto *service = _world->getService(host_name, description)) {
            Debug(_logger_core) << "check results of " << *service
                                << " will be sent to " << *host;
            service->_check_result_receivers.insert(host);
            host->setNextCheck(
                {}, _logger_core);  // It's effectively a passive check now.

        } else {
            auto svc = R"(service ")" + host_name + ";" + description + R"(")";
            Warning(_logger_core) << "cannot send results of non-existing "
                                  << svc << " to " << *host;
            // NOTE: Tricky stuff below! Previously we might have made the
            // check effectively passive (see above), so we have to
            // reschedule it now, otherwise it will never run.
            if (auto t = _service_does_not_exist_schedule.insert(
                    {host, host->nextImmediateCheckTime()})) {
                host->setNextCheck(*t,
                                   _service_does_not_exist_schedule.logger());
            }
        }
    } else if (host->_check_command_line == "@parents") {
        if (auto t = _cluster_host_schedule.insert(job)) {
            job.value()->setNextCheck(*t, _cluster_host_schedule.logger());
        }
    } else if (!host->_check_command_line.starts_with("@")) {
        if (auto t = _helper_pool_generic.insert({host, job.timestamp()})) {
            job.value()->setNextCheck(*t, _helper_pool_generic.logger());
        }
    } else {
        Error(_logger_core)
            << "Invalid checkhelper routing prefix for " << *host << " in ["
            << host->_check_command_line << "]";
    }
}

// Put job in correct schedule, assumes that _next_check has already been
// set!
void Core::scheduleServiceJob(const Schedule<Service *>::Job &job) {
    auto *service = job.value();
    const Schedule<Object *>::Job ojob{service,
                                       job.timestamp()};  // TODO(sp) Nuke this!
    if (_logger_core->isLoggable(LogLevel::debug)) {
        Debug(_logger_core) << "scheduling " << job << " with commandline ["
                            << service->_check_command_line << "]";
    }
    if (service->_check_command_line.starts_with("@cmk ")) {
        if (world()->config_.global_config_.checking_.fetchers_enabled_) {
            if (auto t = _helper_pool_fetcher.insert(ojob)) {
                service->setNextCheck(*t, _helper_pool_fetcher.logger());
            }
        }
    } else if (!service->_check_command_line.starts_with("@")) {
        if (auto t = _helper_pool_generic.insert(ojob)) {
            job.value()->setNextCheck(*t, _helper_pool_generic.logger());
        }
    } else {
        Error(_logger_core)
            << "Invalid checkhelper routing prefix for " << *service << " in ["
            << service->_check_command_line << "]";
    }
}

std::chrono::system_clock::time_point Core::initialScheduleTime(
    std::chrono::system_clock::time_point now, Object *object) {
    if (object->_check_command_line.starts_with("@cmk ")) {
        return initialScheduleTime(
            now, &_schedule_pointer_cmk,
            config()->checking_.initial_checkmk_scheduling_interval_,
            object->_check_interval);
    }
    if (!object->_check_command_line.starts_with("@")) {
        return initialScheduleTime(
            now, &_schedule_pointer_generic,
            config()->checking_.initial_generic_scheduling_interval_,
            object->_check_interval);
    }
    // Make sure the other jobs start immediately by putting them into the
    // past.
    return std::chrono::system_clock::now() - object->_check_interval;
}

std::chrono::system_clock::time_point Core::initialScheduleTime(
    std::chrono::system_clock::time_point now,
    std::chrono::system_clock::time_point *schedule_pointer,
    std::chrono::nanoseconds scheduling_interval,
    std::chrono::nanoseconds check_interval) const {
    const int burst = config()->checking_.initial_scheduling_burst_;
    const std::chrono::system_clock::time_point limit =
        now - (burst * scheduling_interval);
    if (*schedule_pointer < limit) {
        *schedule_pointer = limit;
    }
    std::chrono::system_clock::time_point result = *schedule_pointer;

    // Advance schedule pointer. But Do not schedule more than 5 times
    // the check_interval into the future. Wrap around to the present in
    // that case
    *schedule_pointer += scheduling_interval;
    auto horizon = 5 * check_interval;
    if (*schedule_pointer > now + horizon) {
        *schedule_pointer = now + ((*schedule_pointer - now) % horizon);
        ;
    }
    return result;
}

// .-Checkresult-----------------------------------------------------------.
// |         ____ _               _                        _ _             |
// |        / ___| |__   ___  ___| | ___ __ ___  ___ _   _| | |_           |
// |       | |   | '_ \ / _ \/ __| |/ / '__/ _ \/ __| | | | | __|          |
// |       | |___| | | |  __/ (__|   <| | |  __/\__ \ |_| | | |_           |
// |        \____|_| |_|\___|\___|_|\_\_|  \___||___/\__,_|_|\__|          |
// |                                                                       |
// +-----------------------------------------------------------------------+
// |  Processing of check results                                          |
// '-----------------------------------------------------------------------'

// This function is called whenever new monitoring data has arrived (active
// or passive).
// NOLINTNEXTLINE(misc-no-recursion)
void Core::processCheckResult(std::chrono::system_clock::time_point now,
                              const CheckResult &cr) {
    auto *object = cr._object;
    auto register_alert = [object, this](int state, const std::string &output,
                                         const std::string &long_output) {
        if (config()->history_.cache_horizon_) {
            table_cached_statehist_.registerAlert(*object, state, output,
                                                  long_output);
        }
    };
    object->stateNonConst()->update(now, cr, register_alert,
                                    globalState()->getEnableFlapDetection());
    objectStateHasBeenChanged(now, object);

    for (const auto &receiver : object->_check_result_receivers) {
        Debug(_logger_core)
            << "sending check result from " << *object << " to " << *receiver;
        // NOTE: We only send service check results to hosts currently, so
        // there will be no loop. If we ever generalize this, we need some
        // kind of loop detection here to avoid endless recursion.
        processCheckResult(now, cr.createCopyFor(receiver));
    }
}

// TODO(sp): Use ref!
void Core::objectStateHasBeenChanged(std::chrono::system_clock::time_point now,
                                     Object *object) {
    if (object->isHost()) {
        counterIncrement(Counter::host_checks);
    } else {
        counterIncrement(Counter::service_checks);
    }

    _alert_helper.handle(*object, AlertHandlerEventType::checkresult);

    auto *state = object->stateNonConst();

    // Log any real state change into the monitoring history - regardless if
    // soft or hard. Note: there might be hard state changes even without
    // soft state changes (if the max attempts have been reached while the
    // status is still non_OK)!
    if (object->stateChanged() ||
        // log retries
        (!object->isCurrentStateOK() &&
         state->current_attempt_ < object->_max_check_attempts) ||
        // log last try
        state->hard_state_.changed_) {
        monitoring_log_.alert(*object);
        if (config()->history_.cache_horizon_) {
            table_cached_statehist_.registerAlert(
                *object, state->current_state_, state->check_result_.output_,
                state->check_result_.long_output_);
        }
        _alert_helper.handle(*object, AlertHandlerEventType::statechange);
        _triggers.notify_all(Triggers::Kind::state);  // for livestatus waiting

        // handle DOWN/UNREACHABLE logic
        if (object->isHost()) {
            object->host()->updateNetworkOutages(
                now, [this](std::chrono::system_clock::time_point now,
                            const CheckResult &cr) {
                    this->processCheckResult(now, cr);
                });
        }

        if (!object->isShadow()) {
            switch (object->acknowledgementType()) {
                case AcknowledgementType::none:
                    break;
                case AcknowledgementType::normal:
                    if (object->stateChanged() ||
                        !state->hard_state_.changed_) {
                        _comments.removeAcknowledgements({object}, "END");
                    }
                    break;
                case AcknowledgementType::sticky:
                    if (object->isCurrentStateOK()) {
                        _comments.removeAcknowledgements({object}, "END");
                    }
                    break;
            }
        }
    }

    object->flap(now, loggerNotifications());

    // Handle hard state changes
    if (!object->isShadow() && state->hard_state_.changed_) {
        Debug(loggerNotifications())
            << *object << ": hard state change to " + object->statename();
        if (!object->stateChanged()) {
            _triggers.notify_all(Triggers::Kind::state);  // not triggered yet
        }

        // If a flexible downtime is waiting for a non-OK state, start it
        if (state->downtime_.pending_flex_ != 0U &&
            !object->isCurrentStateOK()) {
            _downtimes.all_of([object, now](Downtime &dt) {
                if (dt._object == object && dt._hot) {
                    dt.start(now);
                }
                return true;
            });
        }

        object->notifyHardStateChange(now, loggerNotifications());
    }

    // Store performance data into RRDs
    processPerfdata(object);

    // wake up livestatus clients from wait
    _triggers.notify_all(Triggers::Kind::check);
}

void Core::logNotification(const Object &object, const std::string &message) {
    Debug(_notify_helper.logger()) << object << ": " << message;
}

void Core::flappingEvent(const Object &object) {
    monitoring_log_.flappingAlert(object);
    if (config()->history_.cache_horizon_) {
        table_cached_statehist_.registerFlapping(
            object, object.state()->flappiness_.is_flapping_);
    }
}

void Core::downtimeEvent(const Object &object, const std::string &state_type,
                         const std::string &comment) {
    monitoring_log_.downtimeAlert(object, state_type, comment);
    if (config()->history_.cache_horizon_) {
        table_cached_statehist_.registerDowntime(object,
                                                 state_type == "STARTED");
    }
}

// .-Downtimes-------------------------------------------------------------.
// |         ____                      _   _                               |
// |        |  _ \  _____      ___ __ | |_(_)_ __ ___   ___  ___           |
// |        | | | |/ _ \ \ /\ / / '_ \| __| | '_ ` _ \ / _ \/ __|          |
// |        | |_| | (_) \ V  V /| | | | |_| | | | | | |  __/\__ \          |
// |        |____/ \___/ \_/\_/ |_| |_|\__|_|_| |_| |_|\___||___/          |
// |                                                                       |
// +-----------------------------------------------------------------------+
// |  Handling of scheduled downtimes                                      |
// '-----------------------------------------------------------------------'

void Core::expireComments(std::chrono::system_clock::time_point now) {
    // not more often than once a second
    if (now <= _last_comments_check + 1s) {
        return;
    }
    _comments.expire(now);
    _last_comments_check = now;
}

// Check all downtimes for begin and end
void Core::expireDowntimes(std::chrono::system_clock::time_point now) {
    // not more often than once a second
    if (now <= _last_downtimes_check + 1s) {
        return;
    }

    _downtimes.expire(now);
    _last_downtimes_check = now;
}

// .-Miscellaneous---------------------------------------------------------.
// |     __  __ _              _ _                                         |
// |    |  \/  (_)___  ___ ___| | | __ _ _ __   ___  ___  _   _ ___        |
// |    | |\/| | / __|/ __/ _ \ | |/ _` | '_ \ / _ \/ _ \| | | / __|       |
// |    | |  | | \__ \ (_|  __/ | | (_| | | | |  __/ (_) | |_| \__ \       |
// |    |_|  |_|_|___/\___\___|_|_|\__,_|_| |_|\___|\___/ \__,_|___/       |
// |                                                                       |
// +-----------------------------------------------------------------------+
// |  Stuff that fitted nowhere else                                       |
// '-----------------------------------------------------------------------'

Object *Core::getAlertHandlerResponseObject(
    const std::string &host_name,
    const std::string &service_description) const {
    Host *host = _world->getHost(host_name);
    if (host == nullptr) {
        Warning(_logger_core)
            << "Alert handler existed for non-existing host " << host_name;
        return nullptr;
    }
    if (service_description.empty()) {
        return host;
    }
    Object *object = host->getService(service_description);
    if (object == nullptr) {
        Warning(_logger_core)
            << "Alert handler existed for non-existing service " << host_name
            << ";" << service_description;
    }
    return object;
}

void Core::notify(const NotificationInfo &info) { _notify_helper.notify(info); }

void Core::broadcast(const NotificationInfo &info) {
    _notify_helper.broadcast(info);
}

// .-Smart Ping------------------------------------------------------------.
// |        ____                       _     ____  _                       |
// |       / ___| _ __ ___   __ _ _ __| |_  |  _ \(_)_ __   __ _           |
// |       \___ \| '_ ` _ \ / _` | '__| __| | |_) | | '_ \ / _` |          |
// |        ___) | | | | | | (_| | |  | |_  |  __/| | | | | (_| |          |
// |       |____/|_| |_| |_|\__,_|_|   \__| |_|   |_|_| |_|\__, |          |
// |                                                       |___/           |
// +-----------------------------------------------------------------------+
// |  Sending of smart pings and also timeout handling                     |
// '-----------------------------------------------------------------------'

void Core::handleSmartPingTimeouts(std::chrono::system_clock::time_point now) {
    if (now - _last_smart_ping_check <= config()->smart_ping_.check_interval_) {
        return;
    }

    for (auto *host : _smart_ping_schedule) {
        const auto *check_period = host->_check_period;
        if (!check_period->isActive()) {
            continue;
        }

        if (!host->isCheckEnabled(CheckResultKind::active) ||
            host->_ipv4_address == 0U) {
            continue;  // host checks are not enabled. skip them
        }

        // Avoid ping timeouts due to the downtime of the core itself! Also
        // omit ping timeouts if the check period has just begun.
        auto dead_time =
            _last_smart_ping_check -
            std::max(std::max(_core_start, host->state()->last_check_),
                     check_period->lastStateChange());

        if (dead_time - config()->smart_ping_.check_interval_ >
            host->_smart_ping_timeout) {
            // Fake a critical result.
            const std::string output =
                "No IP packet received for " +
                std::to_string(
                    mk::ticks<std::chrono::duration<double>>(dead_time)) +
                " s (deadline is " +
                std::to_string(mk::ticks<std::chrono::duration<double>>(
                    host->_smart_ping_timeout)) +
                " s)";
            if (_logger_smart_ping->isLoggable(LogLevel::debug)) {
                Debug(_logger_smart_ping)
                    << "smartping timeout for " << *host << ": " << output;
            }
            processCheckResult(
                now, CheckResult(host, returnCodeToHostState(host, 2), output));
        }
    }
    _last_smart_ping_check = now;
}

// .-Housekeeping----------------------------------------------------------.
// |    _   _                      _                   _                   |
// |   | | | | ___  _   _ ___  ___| | _____  ___ _ __ (_)_ __   __ _       |
// |   | |_| |/ _ \| | | / __|/ _ \ |/ / _ \/ _ \ '_ \| | '_ \ / _` |      |
// |   |  _  | (_) | |_| \__ \  __/   <  __/  __/ |_) | | | | | (_| |      |
// |   |_| |_|\___/ \__,_|___/\___|_|\_\___|\___| .__/|_|_| |_|\__, |      |
// |                                            |_|            |___/       |
// +-----------------------------------------------------------------------+
// |  Various regular tasks such as logfile rotation, end of flapping      |
// |  end of downtimes and other stuff that is not done via schedules.     |
// '-----------------------------------------------------------------------'

// All actions that should be done on a regular base, but
// not too often, since they might be time consuming (relatively)

// clang-tidy-16 seems to be very confused here, all optional accesses *are*
// guarded by if/while.
// NOLINTBEGIN(bugprone-unchecked-optional-access)
void Core::doHousekeeping(std::chrono::system_clock::time_point now) {
    _alert_helper.tryToSendEvent();
    _notify_helper.tryToSendEvent();

    auto result_handler = [this](std::chrono::system_clock::time_point now,
                                 const CheckResult &cr) {
        this->processCheckResult(now, cr);
    };
    while (auto job = _cluster_host_schedule.extractLowestUnder(now)) {
        auto *host = job->value();
        host->updateReachability(now, result_handler);
        if (auto t =
                _cluster_host_schedule.insert({host, host->nextCheckTime()})) {
            host->setNextCheck(*t, _cluster_host_schedule.logger());
        }
    }

    while (auto job = _always_up_schedule.extractLowestUnder(now)) {
        auto *host = job->value();
        host->setUp(now, result_handler);
        if (auto t =
                _always_up_schedule.insert({host, host->nextCheckTime()})) {
            host->setNextCheck(*t, _always_up_schedule.logger());
        }
    }

    while (auto job =
               _service_does_not_exist_schedule.extractLowestUnder(now)) {
        auto *host = job->value();
        host->setServiceDoesNotExist(now, result_handler);
        if (auto t = _service_does_not_exist_schedule.insert(
                {host, host->nextCheckTime()})) {
            job->value()->setNextCheck(
                *t, _service_does_not_exist_schedule.logger());
        }
    }

    if (now - _last_housekeeping >= config()->general_.houskeeping_interval_) {
        checkNotificationsAndFreshness(now);
        _last_housekeeping = now;
        deleteOldWorlds();
        _livestatus->adaptThreadCount(config()->livestatus_.num_threads_);

        _avg_livestatus_usage.update(
            static_cast<double>(_livestatus_active_connections.load()) /
            _livestatus->numThreads());
        _helper_pool_generic.updateUsageAverage();
        _helper_pool_checker.updateUsageAverage();
        _helper_pool_real_time.updateUsageAverage();
        _helper_pool_fetcher.updateUsageAverage();

        updateStats(now);

        if (monitoring_history_.needLogfileRotation(config()->history_)) {
            rotateMonitoringLog(config()->history_);
        }
    }

    if (now - _last_state_retention >=
        config()->general_.state_retention_interval_) {
        saveState();
        _last_state_retention = now;
    }

    // Do unflap objects exactly once per *calendary* minute
    if (std::chrono::duration_cast<std::chrono::minutes>(
            now.time_since_epoch()) !=
        std::chrono::duration_cast<std::chrono::minutes>(
            _last_flapping_check.time_since_epoch())) {
        unflap(now);
        _last_flapping_check = now;
    }

    // Restart dead check helpers every 3 seconds
    if (now - _last_restart_dead_helpers >= 3s) {
        _helper_pool_generic.restartDeadHelpers();
        _helper_pool_checker.restartDeadHelpers();
        _helper_pool_real_time.restartDeadHelpers();
        _helper_pool_fetcher.restartDeadHelpers();
        _last_restart_dead_helpers = now;
    }

    if (now - _last_config_clean >= 60s) {
        (void)config_cleaner_.submit(
            {_serial_token_factory.used(),
             world()->config_.global_config_.general_.serial_,
             config()->general_.core_directory_ / helper_config_dir_name},
            queue_overflow_strategy::pop_oldest);
        _last_config_clean = now;
    }

    // Check if thread for asynchronously parsing logfile has finished.
    table_cached_statehist_.finishBuildingCache();
}
// NOLINTEND(bugprone-unchecked-optional-access)

void Core::updateStats(std::chrono::system_clock::time_point now) {
    auto checker_tasks = static_cast<double>(_reply_queue.size());
    _average_runnable_jobs_checker.update(checker_tasks);

    auto fetcher_tasks =
        static_cast<double>(_helper_pool_fetcher.runnableCount(now));
    _average_runnable_jobs_fetcher.update(fetcher_tasks);
}

// Check for delayed and periodic notifications
void Core::checkNotificationsAndFreshness(
    std::chrono::system_clock::time_point now) {
    // Check *all* objects for pending notifications
    auto result_handler = [this](std::chrono::system_clock::time_point now,
                                 const CheckResult &cr) {
        processCheckResult(now, cr);
    };
    for (const auto &host : _world->config_.hosts_) {
        host->checkFreshness(now, result_handler, _logger_core);
        host->checkPendingNotification(now, loggerNotifications());
        for (const auto &service : Host::Services(*host)) {
            service->checkFreshness(now, result_handler, _logger_core);
            service->checkPendingNotification(now, loggerNotifications());
        }
    }
}

void Core::unflap(std::chrono::system_clock::time_point now) const {
    for (const auto &host : _world->config_.hosts_) {
        host->unflap(now, loggerNotifications());
        for (const auto &service : Host::Services(*host)) {
            service->unflap(now, loggerNotifications());
        }
    }
}

void Core::rotateMonitoringLog(const HistoryConfig &hc) {
    auto start = std::chrono::system_clock::now();
    auto start_steady = std::chrono::steady_clock::now();
    monitoring_history_.close();
    std::error_code ec;
    fs::create_directories(hc.archive_directory_, ec);
    if (ec) {
        const generic_error ge{"cannot create archive directory " +
                               hc.archive_directory_.string()};
        Error(monitoring_history_.logger_) << ge;
        return;
    }
    auto destination =
        hc.archive_directory_ /
        ("history-" +
         std::to_string(std::chrono::system_clock::to_time_t(start)));
    fs::rename(hc.file_, destination, ec);
    if (ec) {
        const generic_error ge{"Failed to move logfile to archive"};
        Error(monitoring_history_.logger_) << ge;
        return;
    }
    monitoring_history_.open(hc);

    monitoring_log_.logVersion();
    _world->logTimeperiods(monitoring_log_, start);
    _world->logCurrentStates(monitoring_log_, LogStateType::current);
    logCurrentDowntimes();

    Notice(monitoring_history_.logger_)
        << "monitoring history rotated to " << destination << " in "
        << elapsed_ms_since(start_steady) << " ms";
    monitoring_history_.last_rotation_ = start;
}

void Core::logCurrentDowntimes() {
    // It's a bit unclear if we need to log downtimes of hosts *before*
    // their corresponding service downtimes, so let's play safe...
    for (const bool log_hosts : {true, false}) {
        _downtimes.all_of([this, log_hosts](const Downtime &dt) {
            if (dt._status == Downtime::Status::active &&
                dt._object->isHost() == log_hosts) {
                monitoring_log_.downtimeAlert(*dt._object, "STARTED",
                                              dt._comment);
            }
            return true;
        });
    }
}

// .-Main------------------------------------------------------------------.
// |                        __  __       _                                 |
// |                       |  \/  | __ _(_)_ __                            |
// |                       | |\/| |/ _` | | '_ \                           |
// |                       | |  | | (_| | | | | |                          |
// |                       |_|  |_|\__,_|_|_| |_|                          |
// |                                                                       |
// +-----------------------------------------------------------------------+
// |                                                                       |
// '-----------------------------------------------------------------------'

void Core::run() {
    auto now = std::chrono::system_clock::now();
    const auto &general = config()->general_;

    Notice(_logger_core)
        << "-----------------------------------------------------------------";
    Notice(_logger_core) << "Check_MK Micro Core (" << trial_manager_.state(now)
                         << ") started with PID " << getpid();
#ifndef __TIMESTAMP__
#define __TIMESTAMP__ (__DATE__ " " __TIME__)
#endif

#ifdef __GNUC__
#define BUILD_CXX ("g++ " __VERSION__)
#elif defined(__VERSION)
#define BUILD_CXX __VERSION__
#else
#define BUILD_CXX "unknown C++ compiler"
#endif
    Notice(_logger_core) << "version " << VERSION << " compiled "
                         << __TIMESTAMP__ << " with " << BUILD_CXX << ", using "
                         << RegExp::engine() << " regex engine";
    _mc->dumpPaths(_logger_core);

    _world->logLoadStatisics();
    if (_state_load_time > std::chrono::seconds::zero()) {
        Notice(_logger_core)
            << "loaded saved program state with " << _num_host_states_loaded
            << " hosts, " << _num_service_states_loaded << " services, "
            << _comments.size() << " comments, and " << _downtimes.size()
            << " downtimes in "
            << mk::ticks<std::chrono::duration<double, std::milli>>(
                   _state_load_time)
            << " ms";
        ;
        if (_num_hosts_vanished != 0 || _num_services_vanished != 0) {
            Notice(_logger_core)
                << _num_hosts_vanished << " hosts and "
                << _num_services_vanished << " services vanished.";
        }
    } else {
        Notice(_logger_core) << "No state file, created new state.";
    }

    monitoring_history_.open(config()->history_);

    monitoring_log_.logVersion();
    monitoring_log_.coreStarting();
    _world->logTimeperiods(monitoring_log_, now);
    _world->logCurrentStates(monitoring_log_, LogStateType::initial);

    installSignalHandlers();
    if (!general.core_dumps_enabled_) {
        installSegfaultHandler();
    }

    _livestatus = std::make_unique<Livestatus>(this);
    _livestatus->start();
    switchStatehistTable();
    if (_livestatus_local.updateConfig(_world)) {
        _livestatus_local.restart();
    }
    if (_rrd_helper.updateConfig(_world)) {
        _rrd_helper.restart();
    }
    _carbon_connection_pool.start();
    _carbon_connection_pool.reload(
        carbon::protocol::Reload{config()->carbon_.connections_});
    if (_alert_helper.updateConfig(
            _world->config_.global_config_.general_,
            _world->config_.global_config_.alert_handler_)) {
        _alert_helper.restart();
    }
    _influxdb_connection_pool.start();
    _influxdb_connection_pool.reload(
        influxdb::protocol::Reload{config()->influxdb_.connections_});
    config_cleaner_.start();
    state_saver_.start();

    // pool::updateConfig handles reload internally, even more
    // the restart of any pool is highly undesirable action
    _helper_pool_generic.updateConfig(general, config()->checking_);
    _helper_pool_checker.updateConfig(general, config()->checking_);
    _helper_pool_real_time.updateConfig(general, config()->checking_);
    _helper_pool_fetcher.updateConfig(general, config()->checking_);

    if (_notify_helper.updateConfig(_world)) {
        _notify_helper.restart();
    }
    if (_icmp_sender.updateConfig(_world)) {
        _icmp_sender.restart();
    }
    if (_icmp_receiver.updateConfig(_world)) {
        _icmp_receiver.restart();
    }

    scheduleAllChecks();
    _core_start = now;

    if (config()->alert_handler_.enabled_) {
        Notice(_logger_core) << "Alert handlers are activated.";
    }

    if (_real_time_check.updateConfig(_world)) {
        _real_time_check.restart();
    }

    mainLoop();
    _is_shutting_down = true;  // signals threads to leave wait loops, etc.

    // Do never restart ourselves when we are just stopping
    removeSegfaultHandler();
    saveState();

    _icmp_receiver.stop();
    _icmp_sender.stop();
    _notify_helper.stop();
    _alert_helper.stop();
    _livestatus_local.stop();
}

// TODO(sp) Move housekeeping stuff here. Use a real Schedule instead of
// simulating it.
class Housekeeper {
    using HousekeeperSchedule = Schedule<Core *>;

public:
    explicit Housekeeper(Core *core) : core_(core) {}

    [[nodiscard]] std::optional<HousekeeperSchedule::Job> lowestUnder(
        HousekeeperSchedule::key_type when) const {
        auto next = core_->_last_housekeeping +
                    core_->config()->general_.houskeeping_interval_;
        if (next > when) {
            return {};
        }
        return {{core_, next}};
    }

private:
    Core *core_;
};

void Core::mainLoop() {
    while (!_should_terminate) {
        if (_should_reload) {
            reloadConfig();
            // TODO(sp) Cruel hack to ensure that we do various tasks before
            // actually look at object states, check results, etc.  The
            // *real* fix would be using some kind of task queue for all
            // things in the main loop (splitting up doHousekeeping on the
            // way) and re-prioritize the tasks below.
            _last_comments_check = _last_downtimes_check = _last_housekeeping =
                std::chrono::system_clock::time_point();
        }

        // Never wait longer than 0.5 seconds.
        auto loop_start = std::chrono::system_clock::now();
        auto next = loop_start + 500ms;

        auto earliest = [](const auto &sched, auto timestamp) {
            auto job = sched.lowestUnder(timestamp);
            if (job) {
                return std::min(timestamp, job->timestamp());
            }
            return timestamp;
        };
        next = std::min({earliest(_helper_pool_generic, next),
                         earliest(_helper_pool_fetcher, next),
                         earliest(_helper_pool_real_time, next),
                         earliest(_smart_ping_schedule, next),
                         earliest(_always_up_schedule, next),
                         earliest(_service_does_not_exist_schedule, next),
                         earliest(_cluster_host_schedule, next),
                         earliest(Housekeeper(this), next)});
        // we could be late...
        auto polling_timeout = std::max({}, next - loop_start);

        // We have to distribute queue data between free checkers before
        // Poller is set.
        processFetcherReplyQueue();

        Poller poller;
        // Wait for responses from check helpers
        _helper_pool_generic.setUpPoller(poller);
        _helper_pool_checker.setUpPoller(poller);
        _helper_pool_real_time.setUpPoller(poller);
        _helper_pool_fetcher.setUpPoller(poller);

        // Wait for new livestatus connections
        _livestatus->setUpPoller(poller);

        // Wait for commands on the command pipe
        _livestatus_local.setUpPoller(poller);

        // Wait for incoming pongs and TCP packages
        _icmp_receiver.setUpPoller(poller);

        // Error messages of RRD creator
        _rrd_helper.setUpPoller(poller);

        // Wait for internal commands from livestatus threads
        _livestatus_ext.setUpPoller(poller);

        // Wait for alert helper to be free
        _alert_helper.setUpPoller(poller);

        // Wait for notify helper to be free
        _notify_helper.setUpPoller(poller);

        // Wait for PING queue to be writable
        // Wait for diagnostic messages from icmp_sender and _receiver
        _icmp_sender.setUpPoller(poller);

        // real time check port
        _real_time_check.setUpPoller(poller);

        // Now wait for messages, signals and timeout
        if (poller.poll(polling_timeout) == -1) {
            // Do not handle timeout.
            const generic_error ge{"Polling failed"};
            Error(_logger_core) << ge;
            break;
        }
        if (_should_terminate) {
            break;
        }

        auto now = std::chrono::system_clock::now();
        do_statistics();
        expireComments(now);
        expireDowntimes(now);
        doHousekeeping(now);
        _world->updateTimeperiods(monitoring_log_, now);

        _helper_pool_generic.communicate(poller);
        _helper_pool_generic.runScheduledChecks(now);

        _helper_pool_checker.communicate(poller);
        _helper_pool_checker.runScheduledChecks(now);

        _helper_pool_real_time.communicate(poller);
        _helper_pool_real_time.runScheduledChecks(now);

        _helper_pool_fetcher.communicate(poller);
        _helper_pool_fetcher.runScheduledChecks(now);

        _icmp_sender.schedule(_smart_ping_schedule, *this);
        handleSmartPingTimeouts(now);

        _livestatus_local.communicate(poller);
        _livestatus_ext.communicate(poller);
        _icmp_receiver.communicate(poller);
        _icmp_sender.communicate(poller);
        _alert_helper.communicate(poller);
        _notify_helper.communicate(poller);
        _rrd_helper.communicate(poller);
        _real_time_check.communicate(poller);
        _livestatus->communicate(poller);
    }
    Informational(_logger_core) << "stopped monitoring, shutting down...";
}

// .-Load / Save-----------------------------------------------------------.
// |        _                    _      __  ____                           |
// |       | |    ___   __ _  __| |    / / / ___|  __ ___   _____          |
// |       | |   / _ \ / _` |/ _` |   / /  \___ \ / _` \ \ / / _ \         |
// |       | |__| (_) | (_| | (_| |  / /    ___) | (_| |\ V /  __/         |
// |       |_____\___/ \__,_|\__,_| /_/    |____/ \__,_| \_/ \___|         |
// |                                                                       |
// +-----------------------------------------------------------------------+
// |  Loading and saving the current state.                                |
// '-----------------------------------------------------------------------'

fs::path Core::renamed_hosts_path() const {
    return config()->general_.core_directory_ / "renamed_hosts";
}

fs::path Core::versioned_state_path(uint32_t version) const {
    return config()->general_.core_directory_ /
           ("state-version-" + std::to_string(version));
}

fs::path Core::state_path() const {
    return state_path(config()->general_.core_directory_);
}

// static
fs::path Core::state_path(const fs::path &core_directory) {
    return core_directory / "state.pb";
}

void Core::saveState() {
    (void)state_saver_.submit({*this, state_path(), renamed_hosts_path()},
                              queue_overflow_strategy::dont_push);
}

void Core::saveStateProtobuf(Logger *logger, const Core &core,
                             const std::filesystem::path &state_path,
                             const std::filesystem::path &renamed_hosts_path) {
    // Make sure the World does not go away because we do not lock the states.
    const std::shared_lock lock{core._world_lock};
    const std::lock_guard lg{core._logging_lock};
    auto start = std::chrono::steady_clock::now();
    auto new_path = Tempfile{
        std::filesystem::path{state_path}.replace_extension(".pb.new")};
    {
        auto ofs = std::ofstream{new_path.path()};
        if (!PbStateSaver::saveStateFile(core).SerializeToOstream(&ofs)) {
            const generic_error ge("cannot create new state file for " +
                                   new_path.string());
            Alert(logger) << ge;
            return;
        }
    }
    std::error_code ec;
    fs::rename(new_path.path(), state_path, ec);
    if (ec) {
        Alert(logger) << "failed to save state to " << state_path << ": "
                      << ec.message();
        return;
    }
    Informational(logger) << "saved state to " << state_path << " in "
                          << elapsed_ms_since(start) << " ms";
    // After the state has successfully been loaded and *saved*, we need to
    // delete the host rename information.
    fs::remove(renamed_hosts_path, ec);
}

// static
std::unique_ptr<checkmk::cmc::state::v1::StateFile> Core::loadStateFromFile(
    const std::filesystem::path &path) {
    std::ifstream ifs{path, std::ios::binary};
    if (!ifs.is_open()) {
        if (generic_error{}.code() == std::errc::no_such_file_or_directory) {
            return {};
        }
        throw generic_error{"cannot open state file \"" + path.string() + "\""};
    }
    auto state_file = std::make_unique<checkmk::cmc::state::v1::StateFile>();
    if (!state_file->ParseFromIstream(&ifs)) {
        throw std::runtime_error{"cannot parse state file \"" + path.string() +
                                 "\""};
    }
    return state_file;
}

void Core::createState(
    std::unique_ptr<checkmk::cmc::state::v1::StateFile> state_file) {
    auto start = std::chrono::system_clock::now();  // TODO(sp) Fix calculation
    if (state_file) {
        setStateFrom(*state_file);
        // FIXME: Use optional
        _state_load_time = std::chrono::system_clock::now() - start + 1ns;
    } else {
        _state_load_time = 0s;
    }

    // Create states for all those objects that are new!
    createMissingStates(start);
    handleChangedDowntimeRules(start);
}

void Core::setStateFrom(const checkmk::cmc::state::v1::StateFile &state_file) {
    const ObjectResolver resolver{_world, renamedHosts()};
    const std::shared_lock lock{_world_lock};
    PbStateLoader::loadStateFile(
        *this, resolver, state_file,
        {.num_host_states_loaded = _num_host_states_loaded,
         .num_hosts_vanished = _num_hosts_vanished,
         .num_service_states_loaded = _num_service_states_loaded,
         .num_services_vanished = _num_services_vanished},
        _logger_core);
}

void Core::createMissingStates(
    std::chrono::system_clock::time_point now) const {
    for (const auto &host : _world->config_.hosts_) {
        if (host->state() == nullptr) {
            host->createNewState(now, _logger_core);
        }

        for (const auto &service : Host::Services(*host)) {
            if (service->state() == nullptr) {
                service->createNewState(now, _logger_core);
            }
        }
    }
}

std::unordered_map<std::string, std::string> Core::renamedHosts() const {
    const auto path = renamed_hosts_path();
    auto renamed_hosts = std::unordered_map<std::string, std::string>{};
    std::ostringstream os;
    os << std::quoted(path.string());
    auto quoted_path = "host mapping file " + os.str();

    std::ifstream f(path);
    if (!f.is_open()) {
        if (generic_error().code() != std::errc::no_such_file_or_directory) {
            throw generic_error("cannot open " + quoted_path);
        }
        return {};
    }
    Notice(_world->logger_) << "reading " + quoted_path;
    std::string old_name;
    std::string new_name;
    while (getline(f, old_name) && getline(f, new_name)) {
        renamed_hosts[old_name] = new_name;
        old_name.clear();
    }
    if (f.bad()) {
        throw generic_error("cannot read " + quoted_path);
    }
    f.close();
    if (!old_name.empty()) {
        throw std::runtime_error("odd number of lines in " + quoted_path);
    }
    Notice(_world->logger_) << "successfully read " << renamed_hosts.size()
                            << " mappings from " << quoted_path;
    return renamed_hosts;
}

// .-Reload Config---------------------------------------------------------.
// |    ____      _                 _    ____             __ _             |
// |   |  _ \ ___| | ___   __ _  __| |  / ___|___  _ __  / _(_) __ _       |
// |   | |_) / _ \ |/ _ \ / _` |/ _` | | |   / _ \| '_ \| |_| |/ _` |      |
// |   |  _ <  __/ | (_) | (_| | (_| | | |__| (_) | | | |  _| | (_| |      |
// |   |_| \_\___|_|\___/ \__,_|\__,_|  \____\___/|_| |_|_| |_|\__, |      |
// |                                                           |___/       |
// +-----------------------------------------------------------------------+
// |  Functions for on-the-fly reloading the configuration.                |
// '-----------------------------------------------------------------------'
void Core::reloadConfig() {
    Notice(_logger_core) << "Reloading configuration on-the-fly...";
    monitoring_log_.info("Reloading configuration");
    saveState();

    _should_reload = false;

    auto *new_world =
        new World(*this, loadConfigFromFile(_config_filename), _logger_core,
                  _logger_rrd, _serial_token_factory);
    new_world->infoCache()->migrate(*_world->infoCache());
    const auto now = std::chrono::system_clock::now();
    migrateStates(now, new_world);
    auto vanished = findVanishedObjects(_world, new_world);
    logVanishedObject(vanished);
    cancelDowntimes(vanished);

    // Update pointers to objects in Comments and Downtimes
    _comments.migrate(*new_world);
    _downtimes.migrate(*new_world);
    _helper_pool_generic.migrate(*new_world);
    _helper_pool_checker.migrate(*new_world);
    _helper_pool_real_time.migrate(*new_world);
    _helper_pool_fetcher.migrate(*new_world);
    _reply_queue.migrate(*new_world);

    // Switch to the new world. The old one will be deleted later.
    World *old_world = _world;
    _old_worlds.insert(old_world);

    // thread-local variable (histcache calls Livestatus code from main
    // thread)
    g_live_world = new_world;

    _world = new_world;
    setLogConfig();
    _last_config_change = now;

    if (_alert_helper.updateConfig(
            _world->config_.global_config_.general_,
            _world->config_.global_config_.alert_handler_)) {
        _alert_helper.restart();
    }

    // pool::updateConfig handles everything(reload, for example) internally
    _helper_pool_generic.updateConfig(config()->general_, config()->checking_);
    _helper_pool_checker.updateConfig(config()->general_, config()->checking_);
    _helper_pool_real_time.updateConfig(config()->general_,
                                        config()->checking_);
    _helper_pool_fetcher.updateConfig(config()->general_, config()->checking_);

    configureCoreDumps();
    if (old_world->config_.global_config_.history_.cache_horizon_ !=
            config()->history_.cache_horizon_ ||
        old_world->config_.global_config_.history_.max_core_downtime_ !=
            config()->history_.max_core_downtime_) {
        Notice(_logger_core) << "state history cache settings have changed";
        table_cached_statehist_
            .flushCache();  // terminates thread if still running
        removeAllHistcachePointers();
        switchStatehistTable();
    }
    _livestatus->adaptThreadCount(config()->livestatus_.num_threads_);
    if (_notify_helper.updateConfig(_world)) {
        _notify_helper.restart();
    }
    if (_icmp_receiver.updateConfig(_world)) {
        _icmp_receiver.restart();
    }
    if (_icmp_sender.updateConfig(_world)) {
        _icmp_sender.restart();
    }
    handleChangedDowntimeRules(now);
    if (old_world->config_.global_config_.alert_handler_.enabled_ !=
        config()->alert_handler_.enabled_) {
        Notice(_logger_core)
            << "Alert handlers have been switched "
            << (config()->alert_handler_.enabled_ ? "on" : "off");
    }

    if (_real_time_check.updateConfig(_world)) {
        _real_time_check.restart();
    }
    if (_livestatus_local.updateConfig(_world)) {
        _livestatus_local.restart();
    }

    // Make sure (new) timeperiod definitions are updated and logged
    _world->logTimeperiods(monitoring_log_, now);

    // Recreate the scheduling queue. It needs to point to the new objects.
    // We need to save the old queue. The information is redunant since it's
    // kept in object->_next_check;
    scheduleAllChecks();

    _world->logLoadStatisics();
    if (_rrd_helper.updateConfig(_world)) {
        _rrd_helper.reload();
    }
    _carbon_connection_pool.reload(
        carbon::protocol::Reload{config()->carbon_.connections_});
    _influxdb_connection_pool.reload(
        influxdb::protocol::Reload{config()->influxdb_.connections_});
}

// Take over states from existing objects, create new ones for new hosts and
// services. Mark shared state so that we won't delete them...
void Core::migrateStates(std::chrono::system_clock::time_point now,
                         World *new_world) {
    size_t migrated_hosts = 0;
    size_t new_hosts = 0;
    size_t migrated_services = 0;
    size_t new_services = 0;
    for (const auto &host : new_world->config_.hosts_) {
        if (auto *old_host = _world->correspondingObject(host.get())) {
            host->setState(old_host->stateNonConst());
            host->stateNonConst()->_shared++;
            host->sanitizeState(_logger_core);
            migrated_hosts++;
        } else {
            host->createNewState(now, _logger_core);
            new_hosts++;
            if (config()->history_.cache_horizon_) {
                table_cached_statehist_.getOrCreateObjectHistcache(host.get());
            }
        }

        for (const auto &service : Host::Services(*host)) {
            if (auto *old_service =
                    _world->correspondingObject(service.get())) {
                service->setState(old_service->stateNonConst());
                service->stateNonConst()->_shared++;
                service->sanitizeState(_logger_core);
                migrated_services++;
            } else {
                service->createNewState(now, _logger_core);
                new_services++;
                if (config()->history_.cache_horizon_) {
                    table_cached_statehist_.getOrCreateObjectHistcache(
                        service.get());
                }
            }
        }
    }

    Notice(_logger_core) << "migrated " << migrated_hosts << " hosts and "
                         << migrated_services << " services, created "
                         << new_hosts << " hosts and " << new_services
                         << " services";
}

std::unordered_set<const Object *> Core::findVanishedObjects(World *old_world,
                                                             World *new_world) {
    std::unordered_set<const Object *> vanished{};
    for (const auto &old_host : old_world->config_.hosts_) {
        if (new_world->correspondingObject(old_host.get()) == nullptr) {
            vanished.insert(old_host.get());
        } else {
            for (const auto &old_service : Host::Services(*old_host)) {
                if (new_world->correspondingObject(old_service.get()) ==
                    nullptr) {
                    vanished.insert(old_service.get());
                }
            }
        }
    }
    return vanished;
}

void Core::logVanishedObject(
    const std::unordered_set<const Object *> &vanished) {
    if (vanished.empty()) {
        return;
    }
    std::size_t vanished_hosts{0};
    for (const auto *obj : vanished) {
        logVanishedObject(*obj);
        vanished_hosts += obj->isHost() ? 1 : 0;
    }
    auto vanished_services = vanished.size() - vanished_hosts;
    Notice(_logger_core) << vanished_hosts << " hosts and " << vanished_services
                         << " services vanished";
}

void Core::logVanishedObject(const Object &object) {
    monitoring_log_.vanished(object);
    if (config()->history_.cache_horizon_) {
        table_cached_statehist_.registerAlert(
            object, -1, "Object has been removed from the monitoring", "");
    }
}

void Core::cancelDowntimes(const std::unordered_set<const Object *> &objects) {
    std::unordered_set<Downtimes::key_type> remove{};
    _downtimes.all_of([&objects, &remove](Downtime &dt) {
        if (objects.contains(dt._object)) {
            // Explicitly cancel here because Downtimes::remove won't cancel
            // rule-based downtimes.
            dt.cancel();
            remove.insert(dt._id);
        }
        return true;
    });
    if (_downtimes.remove(remove) == 0 && !remove.empty()) {
        _triggers.notify_all(Triggers::Kind::downtime);
    }
}

void Core::handleChangedDowntimeRules(
    std::chrono::system_clock::time_point now) {
    // TODO(sp): Find a better place for this, but this is a bit tricky due
    // to the fact that we can't really log objects during their
    // construction phase.
    for (const auto &host : _world->config_.hosts_) {
        host->logDowntimes(_logger_downtime);
        for (const auto &service : Host::Services(*host)) {
            service->logDowntimes(_logger_downtime);
        }
    }
    _downtimes.handleChangedDowntimeRules(now, *this);
}

// This function is being called when due to a configuration change the
// status histcache has been disabled. In that case the cache itself is
// deleted and all pointers of State->objects into that cache are invalid
// and must therefore be zeroed.
void Core::removeAllHistcachePointers() const {
    for (const auto &host : _world->config_.hosts_) {
        host->stateNonConst()->_histcache = nullptr;
        for (const auto &service : Host::Services(*host)) {
            service->stateNonConst()->_histcache = nullptr;
        }
    }
}

void Core::deleteOldWorlds() {
    if (_old_worlds.empty()) {  // TODO(sp) Special case really needed?
        return;
    }

    // We can delete the old world as soon as nobody currently accesses it.
    // Livestatus threads lock the worlds during requests. So if we obtain
    // the lock - no request is currently running. Each request
    // automatically switches to the new world when it begins.
    const std::unique_lock<std::shared_mutex> ul(_world_lock, std::try_to_lock);
    if (ul) {
        for (auto *old_world : _old_worlds) {
            delete old_world;
        }
        _old_worlds.clear();
    }

    // There is still one problem: If all Livestatus-threads are never idle,
    // then we never get a chance for deleting the old world. Currently this
    // is handle at the second config reload: It forces a write lock and
    // waits until all requests are finished. If this still is a problem,
    // then we could keep a list of all previous not-yet-deleted worlds. And
    // create some mechanism for detecting, if one of them is still in use.
}

// TODO(ml): This function belongs in a helper.
void Core::processPerfdata(const Object *object) {
    if (!globalState()->getProcessPerformanceData() ||
        !object->isEnabled(State::Enabling::Key::performance_data)) {
        return;
    }
    const PerformanceData perf_data(object->state()->check_result_.perf_data_,
                                    object->_check_command_name);
    counterIncrementBy(Counter::perf_data, perf_data.empty() ? 0 : 1);
    counterIncrementBy(Counter::metrics, perf_data.size());
    _rrd_helper.processPerfData(perf_data, object);
    if (carbon::protocol::Update::isSendable(object, perf_data)) {
        _carbon_connection_pool.update(carbon::protocol::Update{
            .perf_data = perf_data,
            .host_name = object->host()->name(),
            .service_description = object->serviceDescription(),
            .timestamp =
                carbon::protocol::Update::make_timestamp(*object->state()),
            .current_state = std::to_string(object->state()->current_state_),
            .scheduled_downtime_depth =
                std::to_string(object->state()->downtime_.scheduled_depth_),
            .acknowledged = object->state()->acknowledgement_type_ ==
                                    AcknowledgementType::none
                                ? "0"
                                : "1",
            .carbon_send_metric_parts = object->_carbon_send_metric_parts});
    }
    if (influxdb::protocol::Update::isSendable(object, perf_data)) {
        _influxdb_connection_pool.update(influxdb::protocol::Update{
            .sender_config = object->_influxdb_send_config,
            .message = influxdb::serialize(perf_data, object)});
    }
}

// WARNING: not testable by current design: Core must be constant.
// TODO (sk): write unit test, if unit-test is allowed to modify Core.
void Core::processFetcherReplyQueue() {
    while (!_reply_queue.empty()) {
        // TODO(sp) Remove the downcast below, see comment in
        // CheckHelperPool.
        auto *helper =
            static_cast<CheckHelper *>(_helper_pool_checker.getIdleHelper());
        if (helper == nullptr) {
            Debug(_logger_fetcher)
                << "There is no free checker to process reply from fetcher";
            break;
        }

        helper->consumeReply(std::move(_reply_queue.front()));
        _reply_queue.pop();
    }
}

// static
std::unique_ptr<checkmk::cmc::config::v1::Core> Core::loadConfigFromFile(
    const std::filesystem::path &path) {
    std::ifstream ifs(path, std::ios::binary);
    if (!ifs) {
        throw generic_error{"cannot open config file \"" + path.string() +
                            "\""};
    }
    auto core_config = std::make_unique<checkmk::cmc::config::v1::Core>();
    if (!core_config->ParseFromIstream(&ifs)) {
        throw std::runtime_error{"cannot parse config file \"" + path.string() +
                                 "\""};
    }
    return core_config;
}
