// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/CycleTime.h"

#include <compare>
#include <ctime>
#include <ostream>
#include <ratio>

#include "livestatus/ChronoUtils.h"
#include "livestatus/Logger.h"

using namespace std::chrono_literals;

// static
CycleTime::Kind CycleTime::toKind(uint16_t i, Logger *logger) {
    switch (i) {
        case fromKind(Kind::none):
        case fromKind(Kind::hourly):
        case fromKind(Kind::daily):
        case fromKind(Kind::weekly):
        case fromKind(Kind::biweekly):
        case fromKind(Kind::every_4weeks):
        case fromKind(Kind::nth_weekday):
        case fromKind(Kind::nth_weekday_from_end):
        case fromKind(Kind::day_of_month):
        case fromKind(Kind::every_5min):
            return static_cast<Kind>(i);
        default: {
            Warning(logger) << "Invalid kind for recurring downtime: " << i;
            return Kind::none;
        }
    }
}

namespace {
tm recompute(tm t) { return to_tm(from_tm(t)); }

std::chrono::seconds nextRecurringCycleDayStartOfMonth(
    std::chrono::system_clock::time_point from) {
    tm local = to_tm(from);
    local.tm_mon++;
    if (local.tm_mon >= 12) {
        local.tm_mon = 0;
        local.tm_year++;
    }

    // TODO(sp) This is a bug! Consider the case when we set the downtime start
    // at e.g. Sep 30th: The first downtime will start on that day, but all the
    // followin ones will silently start on the 28th of the month in case. We
    // should probably warn the user in the GUI for 29th/30th and do nothing in
    // February. Consider leap years, too!
    if (local.tm_mday > 28) {
        local.tm_mday = 28;
    }

    return std::chrono::duration_cast<std::chrono::seconds>(from_tm(local) -
                                                            from);
}

std::chrono::seconds nextRecurringCycleNthWeekday(
    std::chrono::system_clock::time_point from) {
    // determine the which'th weekday we started
    tm local = to_tm(from);
    const int our_weekday = local.tm_wday;  // 0...6
    int our_nth = 1;
    while (local.tm_mday > 7) {
        our_nth += 1;
        local.tm_mday -= 7;
    }

    // Go to the 1st of the next month
    local.tm_mday = 1;
    local.tm_mon += 1;
    if (local.tm_mon >= 12) {
        local.tm_mon = 0;
        local.tm_year++;
    }

    // Recompute the wday and move forward to correct week day
    local = recompute(local);
    while (local.tm_wday != our_weekday) {
        local.tm_mday += 1;
        local.tm_wday = (local.tm_wday + 1) % 7;
    }

    // Jump forward to the nth of that week day
    while (local.tm_mday <= 28 && our_nth > 1) {
        local.tm_mday += 7;
        our_nth--;
    }

    return std::chrono::duration_cast<std::chrono::seconds>(from_tm(local) -
                                                            from);
}

std::chrono::seconds nextRecurringCycleNthWeekdayFromEnd(
    std::chrono::system_clock::time_point from, Logger *logger) {
    // determine the which'th weekday we started
    tm local = to_tm(from);
    const int our_weekday = local.tm_wday;  // 0...6
    const int our_month = local.tm_mon;

    // The n'th from end is determined by counting how many of
    // the same day are still to come in this month. Since months
    // have different lengths we do this by trial and error
    int our_nth = 1;
    while (true) {
        local.tm_mday += 7;
        local = recompute(local);
        if (local.tm_mon != our_month) {
            break;
        }
        our_nth++;
    }

    // Now reset the time stamp and add one month
    local = to_tm(from);
    local.tm_mon += 1;
    if (local.tm_mon >= 12) {
        local.tm_mon = 0;
        local.tm_year++;
    }
    local = recompute(local);

    // Go to last day of that month
    int last_of_month = 28;
    const int that_year = local.tm_year;
    const int that_month = local.tm_mon;
    while (true) {
        local.tm_mday = last_of_month + 1;
        local = recompute(local);
        if (local.tm_mon != that_month) {
            break;
        }
        last_of_month++;
    }

    // Go back until we find the correct weekday
    local.tm_year = that_year;
    local.tm_mon = that_month;
    local.tm_mday = last_of_month;

    // Recompute day of week
    local = recompute(local);
    while (local.tm_wday != our_weekday) {
        local.tm_mday--;
        if (local.tm_wday > 0) {
            local.tm_wday--;
        } else {
            local.tm_wday = 6;
        }
    }

    // No jump weekwise until correct week found, but not < 1
    while (our_nth > 1 && local.tm_mday > 7) {
        local.tm_mday -= 7;
        our_nth--;
    }

    if (our_nth > 1) {
        Warning(logger) << "Cannot shift recurring downtime to nth weekday "
                        << our_weekday << ": month is too short";
    }

    return std::chrono::duration_cast<std::chrono::seconds>(from_tm(local) -
                                                            from);
}
}  // namespace

namespace detail {
// Compute how much the timeshift due to DST changes between two timestamps.  If
// both are DST or both are not DST the result is 0. If the first is not DST and
// the second is DST, then the result for Germany is 3600 - one hour more to be
// added on UTC.
std::chrono::seconds dst_shift_between(
    std::chrono::system_clock::time_point a,
    std::chrono::system_clock::time_point b) {
    return std::chrono::seconds(to_tm(b).tm_gmtoff - to_tm(a).tm_gmtoff);
}
}  // namespace detail

std::chrono::seconds CycleTime::next(
    std::chrono::system_clock::time_point from) const {
    std::chrono::seconds cycle;

    switch (_kind) {
        case Kind::hourly:
            cycle = 1h;
            break;

        case Kind::daily:
            cycle = std::chrono::days{1};
            break;

        case Kind::weekly:
            cycle = std::chrono::weeks{1};
            break;

        case Kind::biweekly:
            cycle = std::chrono::weeks{1} * 2;
            break;

        case Kind::every_4weeks:
            cycle = std::chrono::weeks{1} * 4;
            break;

        case Kind::day_of_month:
            cycle = nextRecurringCycleDayStartOfMonth(from);
            break;

        case Kind::nth_weekday:
            cycle = nextRecurringCycleNthWeekday(from);
            break;

        case Kind::nth_weekday_from_end:
            cycle = nextRecurringCycleNthWeekdayFromEnd(from, _logger);
            break;

        case Kind::every_5min:
            cycle = 5min;  // just for testing
            break;

        default:
            cycle = 1h;
            break;
    }

    // check out if the new end time is more east of UTC due to daylight saving
    // time (positive) or more west (negative).  This happens if the cycle skips
    // a DST-barrier back or forth.  We must be sure - however - that we do not
    // shift back a full cycle or more in order to avoid an endless loop!
    auto result = cycle - detail::dst_shift_between(from, from + cycle);
    while (result <= 0s) {
        result += cycle;
    }
    return result;
}

std::ostream &operator<<(std::ostream &os, const CycleTime &c) {
    switch (c._kind) {
        case CycleTime::Kind::none:
            return os << "non-recurring";
        case CycleTime::Kind::hourly:
            return os << "hourly";
        case CycleTime::Kind::daily:
            return os << "daily";
        case CycleTime::Kind::weekly:
            return os << "weekly";
        case CycleTime::Kind::biweekly:
            return os << "biweekly";
        case CycleTime::Kind::every_4weeks:
            return os << "every 4 weeks";
        case CycleTime::Kind::nth_weekday:
            return os << "every n-th weekday";
        case CycleTime::Kind::nth_weekday_from_end:
            return os << "every n-th weekday from end";
        case CycleTime::Kind::day_of_month:
            return os << "every same day of month";
        case CycleTime::Kind::every_5min:
            return os << "every 5 minutes";
    }
    return os;  // never reached
}
