// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/Downtime.h"

#include <compare>
#include <map>
#include <ostream>
#include <ratio>

#include "cmc/Core.h"
#include "cmc/DowntimeRule.h"
#include "cmc/NotifyHelper.h"
#include "cmc/Object.h"
#include "cmc/State.h"
#include "cmc/World.h"
#include "livestatus/BitMask.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/Logger.h"

using namespace std::chrono_literals;

Downtime::Treatment Downtime::expire(
    std::chrono::system_clock::time_point now) {
    const Treatment treatment = examine(now);
    return treatment == Treatment::keep && moveForward(now) ? examine(now)
                                                            : treatment;
}

Downtime::Treatment Downtime::examine(
    std::chrono::system_clock::time_point now) {
    if (_status == Status::active) {
        auto real_end = _fixed ? _end_time : (_flex_downtime_start + _duration);
        if (now < real_end) {
            return Treatment::keep;  // still active
        }
        end("END");
        return _cycle_time.isRecurring() ? Treatment::keep : Treatment::remove;
    }

    if (_rule != nullptr && _rule->_end_time <= now) {
        cancel();
        return Treatment::remove;
    }

    if (_end_time <= now) {  // Is the whole downtime is in the past?
        cancel();            // it never became active, nuke it
        return _cycle_time.isRecurring() ? Treatment::keep : Treatment::remove;
    }

    if (now < _start_time) {
        return Treatment::keep;  // not yet active
    }

    if (_fixed || _object->state()->hard_state_.last_ != 0) {
        start(now);
        return Treatment::keep;
    }

    if (!_hot) {
        _object->stateNonConst()->downtime_.pending_flex_++;
        _hot = true;  // only hot for now, starts later if necessary
    }
    return Treatment::keep;
}

void Downtime::skip(const std::string &type, const std::string &why) const {
    if (logger_notification_->isLoggable(LogLevel::debug)) {
        Debug(logger_notification_)
            << *_object << ": skipping " << type << " notification, " << why;
    }
}

void Downtime::publish(const std::string &how, const std::string &type,
                       bool in_downtime) {
    core_.downtimeEvent(*_object, how, _comment);
    if (in_downtime) {
        skip(type, "in (another) downtime");
        return;
    }
    if (!_object->notificationsEnabled(logger_notification_)) {
        skip(type, "notifications are disabled for this object");
        return;
    }
    if (!_object->inNotificationPeriod()) {
        skip(type, "object ist outside of notification period");
        return;
    }
    if (is_empty_bit_mask(_object->_notification_options &
                          (NotificationOptions::host_downtime |
                           NotificationOptions::service_downtime))) {
        skip(type, "downtime notifications are disabled for this object");
        return;
    }
    core_.notify({*_object, type, _author, _comment, {}});
}

// Is being called when the downtimes become effective. For
// flexible downtimes this is the case when the hard state
// goes to non-OK while they are _hot.
void Downtime::start(std::chrono::system_clock::time_point now) {
    if (_status == Status::active) {
        Warning(logger_downtime_) << "already active " << *this;
        return;
    }
    Informational(logger_downtime_) << "starting " << *this;

    const bool was_in_downtime = _object->inDowntime();
    _status = Status::active;
    _flex_downtime_start = now;
    _object->stateNonConst()->downtime_.scheduled_depth_++;

    // If this was a flexible downtime that just has started, it's not pending
    // any longer.
    if (_hot) {
        _object->stateNonConst()->downtime_.pending_flex_--;
        _hot = false;
    }
    publish("STARTED", "DOWNTIMESTART", was_in_downtime);
}

// Called when a downtime is either manually cancelled
// are ended by the core because its end was reached. The
// downtime will be deleted after this function, so it's
// not necessary to update its state.
void Downtime::end(const std::string &how) {
    Informational(logger_downtime_) << "ending (via " << how << ") " << *this;
    // Was it a flexible downtime hot, but not yet triggered?
    if (_hot) {
        _object->stateNonConst()->downtime_.pending_flex_--;
        _hot = false;
        return;
    }

    // Was the downtime already active? Then maybe notify...
    if (_status != Status::active) {
        return;
    }
    _status = Status::pending;
    _object->stateNonConst()->downtime_.scheduled_depth_--;
    publish(how, "DOWNTIME" + how, _object->inDowntime());
}

void Downtime::cancel() { end("CANCELLED"); }

bool Downtime::moveForward(std::chrono::system_clock::time_point now) {
    if (!_cycle_time.isRecurring()) {
        return false;
    }
    // Move forward. One cycle might not be sufficient. The core might have been
    // stopped for a couple of hours or days. So we move forward until the new
    // end time is in the future. This can mean that the new start time is in
    // the past, but that does not matter. It usually means overlapping
    // downtimes, e.g.  a daily downtime of length > 24h. While this is
    // nonsense, it is still valid for the user to do so.
    std::chrono::seconds total{0};
    while (_end_time < now) {
        // The relative cycle definition (e.g. 16:00 each first monday) is
        // always applied on the start time. The distance of the end time to the
        // start time (e.g. the duration) must stay unchanged even in cases of
        // DST barriers
        auto dur = _cycle_time.next(_start_time);
        _start_time += dur;
        _end_time += dur;
        total += dur;
    }
    if (total == 0s) {
        return false;
    }
    Informational(logger_downtime_)
        << "moved forward by " << mk::ticks<std::chrono::seconds>(total)
        << " seconds to " << *this;
    return true;
}

bool Downtime::migrate(const World &world) {
    if (auto *migrated_object = world.correspondingObject(_object)) {
        _object = migrated_object;
        return true;
    }
    return false;
}

std::ostream &operator<<(std::ostream &os, const Downtime &d) {
    os << (d._status == Downtime::Status::active ? "active" : "pending")
       << (d._fixed ? " fixed"
                    : (" flexible" + std::string(d._hot ? " hot" : "")))
       << " downtime " << d._id  //
       << " for " << *d._object  //
       << " entered at " << FormattedTimePoint(d._entry_time)
       << (d._origin_is_rule ? " by a rule" : "")
       << (d._author.empty() ? "" : (" by " + d._author))
       << (d._comment.empty() ? "" : (" (" + d._comment + ")"))
       << ", starts at " << FormattedTimePoint(d._start_time)  //
       << ", ends at " << FormattedTimePoint(d._end_time);
    if (d._cycle_time.isRecurring()) {
        os << " (recurring " << d._cycle_time << ")";
    }
    if (!d._fixed) {
        os << ", flexible duration is "
           << mk::ticks<std::chrono::seconds>(d._duration) << " seconds";
    }
    return os;
}
