// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/Downtimes.h"

#include <algorithm>
#include <utility>
#include <vector>

#include "cmc/Core.h"
#include "cmc/DowntimeRule.h"
#include "cmc/Host.h"
#include "cmc/Object.h"
#include "cmc/Service.h"
#include "cmc/State.h"
#include "cmc/World.h"
#include "livestatus/Logger.h"
#include "livestatus/Triggers.h"

namespace {
void validateDowntimeDepth(std::unordered_map<const Object *, uint32_t> &active,
                           Object *object, Logger *logger) {
    if (object->state() != nullptr &&
        object->state()->downtime_.scheduled_depth_ != active[object]) {
        Warning(logger) << *object << " has downtime depth "
                        << object->state()->downtime_.scheduled_depth_
                        << ", but " << active[object]
                        << " active downtimes. This has been fixed.";
        object->stateNonConst()->downtime_.scheduled_depth_ = active[object];
    }
}
}  // namespace

Downtimes::Downtimes(Triggers &triggers, Logger *logger)
    : triggers_{triggers}, logger_{logger} {}

bool Downtimes::migrate(const World &world) {
    // see Scott Meyers' "Effective STL", item 9 ("Choose carefully among
    // erasing options.")
    const std::lock_guard lg(lock_);
    for (auto it = downtimes_.begin(); it != downtimes_.end();) {
        if (it->second->migrate(world)) {
            ++it;
        } else {
            downtimes_.erase(it++);
        }
    }
    return true;
}

void Downtimes::expire(std::chrono::time_point<std::chrono::system_clock> now) {
    // see Scott Meyers' "Effective STL", item 9 ("Choose carefully among
    // erasing options.")
    const std::lock_guard lg{lock_};
    for (auto it = downtimes_.begin(); it != downtimes_.end();) {
        if (it->second->expire(now) == Downtime::Treatment::remove) {
            Informational(logger_) << "removing " << *it->second;
            downtimes_.erase(it++);
            triggers_.notify_all(Triggers::Kind::downtime);
        } else {
            ++it;
        }
    }
}

void Downtimes::insert(std::unique_ptr<Downtime> downtime) {
    // See also `Comments::insert()`.
    const std::lock_guard lg{lock_};
    Informational(logger_) << "inserting " << *downtime;
    downtimes_.emplace(downtime->_id, std::move(downtime));
    triggers_.notify_all(Triggers::Kind::downtime);
}

bool Downtimes::modify(
    const key_type &id,
    const std::function<std::unique_ptr<Downtime>(const Downtime &)> &f) {
    const std::lock_guard lg{lock_};
    auto it = downtimes_.find(id);
    if (it == downtimes_.end()) {
        Warning(logger_) << "cannot find downtime " << id;
        return false;
    }
    if (it->second->_origin_is_rule) {
        Warning(logger_) << "cannot change rule-based downtime " << id;
        return false;
    }

    auto downtime = f(*(it->second));
    if (downtime == nullptr) {
        return false;
    }

    Informational(logger_) << "removing " << *it->second;
    downtimes_.erase(it);
    Informational(logger_) << "inserting " << *downtime;
    downtimes_.emplace(downtime->_id, std::move(downtime));
    triggers_.notify_all(Triggers::Kind::downtime);
    return true;
}

size_t Downtimes::remove(const std::unordered_set<key_type> &ids) {
    // See also `Comments::remove()`.
    size_t num_removed{0};
    const std::lock_guard lg{lock_};
    for (const auto &id : ids) {
        auto it = downtimes_.find(id);
        if (it == downtimes_.end()) {
            Warning(logger_)
                << "cannot delete downtime " << id << ": no such id";
        } else if (it->second->_origin_is_rule) {
            Warning(logger_) << "cannot change rule-based downtime " << id;
        } else {
            Informational(logger_) << "removing " << *it->second;
            it->second->cancel();
            downtimes_.erase(it);
            num_removed++;
        }
    }
    if (num_removed > 0) {
        triggers_.notify_all(Triggers::Kind::downtime);
    }
    return num_removed;
}

void Downtimes::clear() {
    const std::lock_guard lg{lock_};
    downtimes_.clear();
}

bool Downtimes::all_of_unlocked(
    const std::function<bool(const Downtime &)> &pred) const {
    return std::all_of(downtimes_.cbegin(), downtimes_.cend(),
                       [&pred](const auto &d) { return pred(*d.second); });
}

bool Downtimes::all_of(
    const std::function<bool(const Downtime &)> &pred) const {
    const std::lock_guard lg{lock_};
    return all_of_unlocked(pred);
}

bool Downtimes::all_of(const std::function<bool(Downtime &)> &pred) {
    const std::lock_guard lg{lock_};
    return std::all_of(downtimes_.cbegin(), downtimes_.cend(),
                       [&pred](auto &d) { return pred(*d.second); });
}

void Downtimes::handleChangedDowntimeRules(
    std::chrono::system_clock::time_point now, Core &core) {
    const std::lock_guard lg{lock_};
    std::unordered_set<DowntimeRule *> used;

    // Remove currently existing recurring downtimes if their WATO rules
    // have vanished. See comment above in expireComments
    for (auto it = downtimes_.begin(); it != downtimes_.end();) {
        auto &downtime = it->second;
        if (!downtime->_origin_is_rule) {
            Informational(core._logger_downtime)
                << "keeping ad hoc " << *downtime;
            ++it;
            continue;
        }
        auto *prd =
            downtime->_origin_is_rule
                ? downtime->_object->find_prd(DowntimeRule{
                      ._comment = downtime->_comment,
                      ._start_time = downtime->_start_time,
                      ._end_time = downtime->_end_time,
                      ._duration = downtime->_end_time - downtime->_start_time,
                      ._flex_duration = downtime->_duration,
                      ._cycle_time = downtime->_cycle_time})
                : nullptr;
        downtime->_rule = prd;
        if (prd != nullptr && used.insert(prd).second) {
            Informational(core._logger_downtime)
                << "keeping " << *prd << " for " << *downtime->_object;
            ++it;
            continue;
        }
        downtime->cancel();
        Informational(core._logger_downtime) << "removing " << *it->second;
        downtimes_.erase(it++);
    }

    // Now create downtimes for rules that are new
    std::vector<std::unique_ptr<Downtime>> new_downtimes;
    for (const auto &host : core._world->config_.hosts_) {
        for (const auto &prd : host->_downtime_rules) {
            if (!used.contains(prd.get())) {
                new_downtimes.push_back(std::make_unique<Downtime>(
                    Downtime{.core_ = core,
                             .logger_downtime_ = core._logger_downtime,
                             .logger_notification_ = core.loggerNotifications(),
                             ._object = host.get(),
                             ._id = next_id_++,
                             ._entry_time = now,
                             ._author = "",
                             ._comment = prd->_comment,
                             ._status = Downtime::Status::pending,
                             ._start_time = prd->_start_time,
                             ._end_time = prd->_start_time + prd->_duration,
                             ._flex_downtime_start = {},
                             ._fixed = prd->isFixed(),
                             ._cycle_time = prd->_cycle_time,
                             ._duration = prd->_flex_duration,
                             ._hot = false,
                             ._origin_is_rule = true,
                             ._rule = prd.get()}));
            }
        }
        for (const auto &service : Host::Services(*host)) {
            for (const auto &prd : service->_downtime_rules) {
                if (!used.contains(prd.get())) {
                    new_downtimes.push_back(std::make_unique<Downtime>(Downtime{
                        .core_ = core,
                        .logger_downtime_ = core._logger_downtime,
                        .logger_notification_ = core.loggerNotifications(),
                        ._object = service.get(),
                        ._id = next_id_++,
                        ._entry_time = now,
                        ._author = "",
                        ._comment = prd->_comment,
                        ._status = Downtime::Status::pending,
                        ._start_time = prd->_start_time,
                        ._end_time = prd->_start_time + prd->_duration,
                        ._flex_downtime_start = {},
                        ._fixed = prd->isFixed(),
                        ._cycle_time = prd->_cycle_time,
                        ._duration = prd->_flex_duration,
                        ._hot = false,
                        ._origin_is_rule = true,
                        ._rule = prd.get()}));
                }
            }
        }
    }
    for (auto &&downtime : new_downtimes) {
        Informational(core._logger_downtime) << "inserting " << *downtime;
        downtimes_.emplace(downtime->_id, std::move(downtime));
    }
    triggers_.notify_all(Triggers::Kind::downtime);
}

Downtimes::size_type Downtimes::size() const {
    const std::lock_guard lg{lock_};
    return downtimes_.size();
}

void Downtimes::validateDepths(const World *world) {
    std::unordered_map<const Object *, uint32_t> active;
    for (const auto &d : downtimes_) {
        if (d.second->_status == Downtime::Status::active) {
            active[d.second->_object]++;
        }
    }

    for (const auto &host : world->config_.hosts_) {
        validateDowntimeDepth(active, host.get(), logger_);
        for (const auto &service : Host::Services(*host)) {
            validateDowntimeDepth(active, service.get(), logger_);
        }
    }
}
