// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/Entity.h"

#include <utility>

#include "livestatus/Attributes.h"

Entity::Entity(std::string name) : _name(std::move(name)) {}

Entity::Entity(std::string name,
               const std::unordered_map<std::string, std::string> &attributes)
    : _name(std::move(name)) {
    for (const auto &[key, value] : attributes) {
        if (key.starts_with("_")) {
            auto [kind, name] = to_attribute_kind(key.substr(1));
            switch (kind) {
                case AttributeKind::custom_variables:
                    _custom_attributes[name] = value;
                    break;
                case AttributeKind::tags:
                    _tags[name] = value;
                    break;
                case AttributeKind::labels:
                    _labels[name] = value;
                    break;
                case AttributeKind::label_sources:
                    _label_sources[name] = value;
                    break;
            }
        } else {
            _attributes[key] = value;
        }
    }
}

std::string Entity::name() const { return _name; }

std::string Entity::lookup(std::optional<AttributeKind> kind,
                           const std::string &name) const {
    const auto &attrs = attributes_for(kind);
    auto it = attrs.find(name);
    return it == attrs.end() ? "" : it->second;
}

bool Entity::operator==(const Entity &other) const {
    return name() == other.name() && _attributes == other._attributes &&
           _custom_attributes == other._custom_attributes &&
           _tags == other._tags && _labels == other._labels &&
           _label_sources == other._label_sources;
}

const Attributes &Entity::attributes_for(
    std::optional<AttributeKind> kind) const {
    if (!kind) {
        return _attributes;
    }
    switch (*kind) {
        case AttributeKind::custom_variables:
            return _custom_attributes;
        case AttributeKind::tags:
            return _tags;
        case AttributeKind::labels:
            return _labels;
            break;
        case AttributeKind::label_sources:
            return _label_sources;
    }
    return _attributes;  // unreachable, make GCC happy
}
