// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/FetcherHelper.h"

#include <unistd.h>

#include <algorithm>
#include <compare>
#include <ratio>
#include <string>
#include <string_view>
#include <utility>
#include <variant>

#include "cmc/FetcherReply.h"
#include "cmc/Host.h"
#include "cmc/MessagePayload.h"
#include "cmc/Object.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/Logger.h"
#include "livestatus/Poller.h"

using namespace std::chrono_literals;
using namespace std::string_literals;
namespace fp = fetcher::protocol;

FetcherHelper::FetcherHelper(const std::filesystem::path &path,
                             const args_type &args,
                             reply_handler_t reply_handler,
                             reschedule_handler_t<Service> reschedule_handler,
                             Logger *logger)
    : CommonHelper{path, args, logger}
    , reply_handler_{std::move(reply_handler)}
    , reschedule_handler_{std::move(reschedule_handler)}
    , ch_{this->logger()} {}

FetcherHelper::~FetcherHelper() { do_stop(); }

void FetcherHelper::start_() {
    if (spawnConnectedSubprocess()) {
        gotoIdle();
    } else {
        gotoDead();
    }
}

void FetcherHelper::stop_() { do_stop(); }

void FetcherHelper::setUpPoller_(Poller &poller) const {
    if (state() == State::receiving && stdOut() >= 0) {
        poller.addFileDescriptor(stdOut(), PollEvents::in);
    }
    if (stdErr() >= 0) {
        poller.addFileDescriptor(stdErr(), PollEvents::in);
    }
}

void FetcherHelper::communicate_(const Poller &poller) {
    if (state() == State::receiving && stdOut() >= 0 &&
        poller.isFileDescriptorSet(stdOut(), PollEvents::in)) {
        handleStdOut();
    }
    if (stdErr() >= 0 && poller.isFileDescriptorSet(stdErr(), PollEvents::in)) {
        handleStdErr();
    }
}

bool FetcherHelper::sendReloadConfigRequest() { return true; }

void FetcherHelper::incrementCounter() const {}

fp::Request FetcherHelper::makeRequest() {
    fetcher_started_at_ = std::chrono::steady_clock::now();
    return {serial(), object()->host()->name(),
            (object()->_check_command_line.find(" --check-discovery"s) !=
             std::string::npos)
                ? "discovery"s
                : "checking"s,
            object()->_check_timeout};
}

bool FetcherHelper::sendRequest() {
    return ch_.send(stdIn(), makeRequest()) != -1;
}

void FetcherHelper::do_stop() {
    ch_.reset();
    gotoIdle();
    despawn();
}

void FetcherHelper::enqueueCheckerRequest() {
    auto time_used = std::chrono::steady_clock::now() - fetcher_started_at_;
    auto remaining_time = object()->_check_timeout - time_used;
    if (remaining_time < 0s) {
        auto delay = -mk::ticks<std::chrono::seconds>(remaining_time);
        Debug(logger()) << "fetcher reply came too late by " << delay << " s";
        remaining_time = 0s;
    }
    reply_handler_(std::make_unique<FetcherReply>(
        payload_ ? std::move(*payload_) : MessagePayload{0}, remaining_time,
        serialToken(), object()));
}

void FetcherHelper::handleStdOut() {
    assertState(State::receiving);
    while (true) {
        auto message = ch_.receive(stdOut());
        if (std::visit([this](auto &msg) { return handle(msg); }, message) ==
            how_to_proceed::stop) {
            break;
        }
    }
}

// We need a consistent interface for std::visit above.
// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
FetcherHelper::how_to_proceed FetcherHelper::handle(
    const fp::Incomplete & /*msg*/) const {
    return how_to_proceed::stop;
}

FetcherHelper::how_to_proceed FetcherHelper::handle(
    const fp::Failure & /*msg*/) {
    reschedule_handler_(object()->service());
    payload_.reset();
    gotoDead();
    return how_to_proceed::stop;
}

FetcherHelper::how_to_proceed FetcherHelper::handle(fp::Result &msg) {
    payload_ = std::move(msg.payload);
    return how_to_proceed::carry_on;
}

FetcherHelper::how_to_proceed FetcherHelper::handle(const fp::Log &msg) const {
    if (logger()->isLoggable(msg.log_level)) {  // optimization only
        LogStream(logger(), msg.log_level)
            << "[log] " << std::string_view{msg.message};
    }
    return how_to_proceed::carry_on;
}

FetcherHelper::how_to_proceed FetcherHelper::handle(
    const fp::EndOfReply & /*msg*/) {
    enqueueCheckerRequest();
    gotoIdle();
    return how_to_proceed::stop;
}

void FetcherHelper::handleStdErr() {
    // Read any lines from stderr (error messages) and simply log them
    char buffer[4096];
    auto bytes_read = ::read(stdErr(), buffer, sizeof(buffer));
    if (bytes_read == -1) {
        return;
    }
    std::string_view response{buffer, size_t(bytes_read)};
    while (!response.empty()) {
        auto line = response.substr(0, response.find('\n'));
        response.remove_prefix(std::min(response.size(), line.size() + 1));
        Warning(logger()) << "unexpected message on stderr: " << line;
    }
}
