// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/FetcherHelperChannel.h"

#include <algorithm>
#include <array>
#include <cassert>
#include <cerrno>
#include <iterator>
#include <sstream>
#include <stdexcept>
#include <string_view>
#include <system_error>
#include <utility>
#include <vector>

#include "livestatus/ChronoUtils.h"
#include "livestatus/Logger.h"
#include "livestatus/POSIXUtils.h"
#include "livestatus/StringUtils.h"

using namespace std::chrono_literals;
using namespace std::string_view_literals;
namespace fp = fetcher::protocol;

namespace {
// This is our "parsing monad": The state is represented by the istream and
// failures are represented by throwing a generic_error.

[[noreturn]] void parseFailure(int err, std::string_view what) {
    throw generic_error{err, std::string{what} + " in fetcher header"};
};

std::string parseFixedString(std::istream &is, size_t size) {
    std::vector<char> buf(size);
    if (is.read(buf.data(), size)) {
        return {buf.data(), size};
    }
    parseFailure(errno,
                 "could not read " + std::to_string(size) + " characters");
}

template <typename T, size_t N>
using one_of = std::array<std::pair<std::string_view, T>, N>;

// As complicated and inefficient as it looks, the function below is completely
// unfolded in code: It basically results in very fast if-then-else cascades,
// guarded by the lengths, see: https://www.youtube.com/watch?v=INn3xa4pMfg
template <typename T, size_t N>
constexpr T parseOneOf(std::string_view str, const one_of<T, N> &table,
                       std::string_view what) {
    auto it = std::find_if(begin(table), end(table),
                           [&](const auto &v) { return v.first == str; });
    if (it != end(table)) {
        return it->second;
    }
    parseFailure(EBADMSG, what);
}

void parseProtocolName(std::istream &is) {
    static constexpr one_of<int, 1> protocol_names{{{"fetch"sv, 42}}};
    parseOneOf(parseFixedString(is, 5), protocol_names, "protocol name");
}

void parseSeparator(std::istream &is) {
    static constexpr one_of<int, 1> separators{{{":"sv, 42}}};
    parseOneOf(parseFixedString(is, 1), separators, "separator");
}

FetcherMessageHeader::Kind parseKind(std::istream &is) {
    static constexpr one_of<FetcherMessageHeader::Kind, 3> kinds{
        {{"RESULT "sv, FetcherMessageHeader::Kind::result},
         {"LOG    "sv, FetcherMessageHeader::Kind::log},
         {"ENDREPL"sv, FetcherMessageHeader::Kind::end_of_reply}}};
    return parseOneOf(parseFixedString(is, 7), kinds, "kind");
}

std::optional<LogLevel> parseLogLevel(std::istream &is,
                                      FetcherMessageHeader::Kind kind) {
    static constexpr one_of<LogLevel, 8> log_levels{
        {{"emergenc"sv, LogLevel::emergency},
         {"alert   "sv, LogLevel::alert},
         {"critical"sv, LogLevel::critical},
         {"error   "sv, LogLevel::error},
         {"warning "sv, LogLevel::warning},
         {"notice  "sv, LogLevel::notice},
         {"info    "sv, LogLevel::informational},
         {"debug   "sv, LogLevel::debug}}};
    static constexpr one_of<std::optional<LogLevel>, 1> empty{
        {{"        "sv, {}}}};
    auto log_level = parseFixedString(is, 8);
    switch (kind) {
        case FetcherMessageHeader::Kind::log:
            return parseOneOf(log_level, log_levels, "log level");
        case FetcherMessageHeader::Kind::result:
        case FetcherMessageHeader::Kind::end_of_reply:
            return parseOneOf(log_level, empty, "log level");
    }
    return {};  // unreachable, make GCC happy
}

size_t convertToLength(const std::string &str) {
    try {
        return std::stoul(str);
    } catch (const std::invalid_argument &e) {
        parseFailure(EBADMSG, "bad payload length");
    }
}

size_t parsePayloadLength(std::istream &is, FetcherMessageHeader::Kind kind) {
    auto payload_length = convertToLength(parseFixedString(is, 16));
    switch (kind) {
        case FetcherMessageHeader::Kind::result:
        case FetcherMessageHeader::Kind::log:
            break;
        case FetcherMessageHeader::Kind::end_of_reply:
            if (payload_length != 0) {
                parseFailure(EBADMSG, "non-zero payload length");
            }
    }
    return payload_length;
}
}  // namespace

FetcherMessageHeader::FetcherMessageHeader(std::istream &is) {
    parseProtocolName(is);
    parseSeparator(is);
    kind_ = parseKind(is);
    parseSeparator(is);
    log_level_ = parseLogLevel(is, kind_);
    parseSeparator(is);
    payload_length_ = parsePayloadLength(is, kind_);
    parseSeparator(is);
}

namespace {
std::string makeRequestString(const fp::Request &request) {
    return std::to_string(request.serial_) + ";" +  //
           request.host_ + ";" +                    //
           request.mode_ + ";" +
           std::to_string(mk::ticks<std::chrono::seconds>(request.timeout_)) +
           "\n";
}
}  // namespace

std::ostream &operator<<(std::ostream &os,
                         const fetcher::protocol::Request &r) {
    return os << R"(request for host ")" << r.host_
              << R"(" with configuration )" << r.serial_ << " in " << r.mode_
              << " mode, " << mk::ticks<std::chrono::seconds>(r.timeout_)
              << " s timeout";
}

void FetcherHelperChannel::reset() {
    header_.reset();
    parsed_header_.reset();
    payload_.reset();
}

ssize_t FetcherHelperChannel::send(int std_in,
                                   const fp::Request &request) const {
    Informational(logger_) << "sending " << request;
    auto ret = writeWithTimeout(std_in, makeRequestString(request), 3s);
    if (ret == -1) {
        const generic_error ge{"cannot send request"};
        Warning(logger_) << ge;
    }
    return ret;
}

fp::Message FetcherHelperChannel::receive(int std_out) {
    // NOTE: Things would be much easier and less stateful if we had a notion of
    // "lazy stream of characters coming from a non-blocking file descriptor".
    // This would simplify our state machine below quite a bit and remove some
    // type conversion dance.
    try {
        if (!header_) {
            header_ = MessagePayloadReceiver{FetcherMessageHeader::size};
        }
        if (header_->receive(std_out) ==
            MessagePayloadReceiver::status::partial) {
            return fp::Incomplete{};
        }
        if (!parsed_header_) {
            const std::string_view hc{header_->content};
            Informational(logger_) << "received header [" << hc << "]";
            std::istringstream is{std::string{hc}};
            parsed_header_ = FetcherMessageHeader(is);
        }
        if (!payload_) {
            payload_ = MessagePayloadReceiver{parsed_header_->payload_length()};
        }
        if (payload_->receive(std_out) ==
            MessagePayloadReceiver::status::partial) {
            return fp::Incomplete{};
        }
        return buildMessage();
    } catch (generic_error &e) {
        Critical(logger_) << e;
        return fp::Failure{e.code().value()};
    }
}

// TODO(sp) Remove the asserts below, we have way too many implicit assumptions.
fp::Message FetcherHelperChannel::buildMessage() {
    assert(payload_);  // always set in caller, i.e. receive()
    const std::string_view pc{payload_->content};
    Informational(logger_) << "received " << pc.size() << "-byte payload";
    Debug(logger_) << "payload content [" << mk::escape_nonprintable{pc} << "]";

    assert(parsed_header_);  // always set in caller, i.e. receive()
    auto kind = parsed_header_->kind();
    auto log_level = parsed_header_->log_level();
    auto payload = std::move(payload_->content);

    header_.reset();
    parsed_header_.reset();
    payload_.reset();

    switch (kind) {
        case FetcherMessageHeader::Kind::result:
            return fp::Result{std::move(payload)};
        case FetcherMessageHeader::Kind::log:
            // only log messages have a log level, see parseLogLevel()
            assert(log_level);
            return fp::Log{*log_level, std::move(payload)};
        case FetcherMessageHeader::Kind::end_of_reply:
            return fp::EndOfReply{};
    }
    return fp::Incomplete{};  // unreachable, make GCC happy
}
