// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/Histcache.h"

#include <algorithm>
#include <compare>
#include <list>
#include <string_view>

#include "livestatus/Logger.h"

HistcacheEntry *Histcache::oldestEntry() {
    return _cache.empty() ? nullptr : &_cache.front();
}

HistcacheEntry Histcache::youngestEntry() {
    return _cache.empty() ? HistcacheEntry() : _cache.back();
}

// Return the index of that entry that starts just before since. If since is
// earlier than our first entry, then -1 is being returned.
int Histcache::findEntryBefore(std::chrono::system_clock::time_point since) {
    // We might use binary search, if we need to increase performance
    int i = -1;
    for (auto &hce : _cache) {
        if (hce._timestamp > since) {  // This entry is one too far
            break;  // returns index of previous entry, since i++ comes later
        }
        i++;
    }
    return i;
}

void Histcache::prependEntry(const HistcacheEntry &hce) {
    _cache.push_front(hce);
}

void Histcache::appendEntry(const HistcacheEntry &hce) {
    _cache.push_back(hce);
}

void Histcache::mergeOldStates() {
    while (_cache.size() >= 2) {
        const HistcacheEntry &older = _cache[0];
        HistcacheEntry &newer = _cache[1];
        if (older._bits == newer._bits) {
            newer._timestamp = older._timestamp;
            newer._plugin_output = older._plugin_output;
            newer._long_plugin_output = older._long_plugin_output;
            _cache.pop_front();
        } else {
            break;
        }
    }
}

void Histcache::finalize() {
    // First add "last known plugin output", where missing
    std::string_view last_known_output;
    std::string_view last_known_long_output;
    for (auto &hce : _cache) {
        if (!hce._plugin_output.empty()) {
            last_known_output = hce._plugin_output;
            last_known_long_output = hce._long_plugin_output;
        } else if (!last_known_output.empty()) {
            hce._plugin_output = last_known_output;
            hce._long_plugin_output = last_known_long_output;
        }
    }

    // Next: Remove empty periods and periods with the same state. NOTE: We have
    // to be careful here to avoid quadratic time complexity, as the _cache can
    // be quite big. To do this, we construct a doubly linked list from the
    // dequeue, do any removals there, and finally reconstruct the dequeue from
    // the list. A bit complicated, but linear...
    std::list<HistcacheEntry> entries{_cache.begin(), _cache.end()};
    auto it_old = entries.begin();
    if (it_old != entries.end()) {
        auto it_new = it_old;
        ++it_new;
        while (it_new != entries.end()) {
            if (it_old->_timestamp == it_new->_timestamp) {
                // empty period? => drop old entry
                it_old = entries.erase(it_old);
                ++it_new;
            } else if (it_old->_bits == it_new->_bits) {
                // same state? => drop new entry
                it_new = entries.erase(it_new);
            } else {
                it_old = it_new;
                ++it_new;
            }
        }
    }
    _cache.clear();
    _cache.insert(_cache.end(), entries.begin(), entries.end());
    _cache.shrink_to_fit();
}

void Histcache::dump(Logger *logger) {
    Debug(logger)
        << "Nr.  Timestamp of cache entry        Duration STA DTIME  FLAP  Output";
    Debug(logger)
        << "--------------------------------------------------------------------------------";
    int i = 0;
    for (auto it = _cache.begin(); it != _cache.end(); ++it, ++i) {
        auto next = (it + 1 == _cache.end()) ? std::chrono::system_clock::now()
                                             : (it + 1)->_timestamp;
        it->dump(i, next - it->_timestamp, logger);
    }
}
