// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/Host.h"

#include <algorithm>

#include "cmc/CheckResult.h"
#include "cmc/Entity.h"
#include "cmc/ObjectGroup.h"
#include "cmc/cmc.h"
#include "livestatus/Interface.h"
#include "livestatus/StringUtils.h"
class Timeperiod;

using namespace std::chrono_literals;

Host::Host(Core &core, World &world, Entity &&entity)
    : Object{core, world, std::move(entity)}, ihost_{*this} {}

// Used for creation of temporary fake hosts for PROCESS_HOST_PERFDATA
Host::Host(Core &core, World &world, const std::string &name,
           const Timeperiod &tp)
    : Object{core, world, name, tp}
    , _smart_ping_timeout{0ms}
    , _ipv4_address{0}
    , ihost_{*this} {}

HostState Host::remap(HostState state) const {
    return _state_mapping.at(state);
}

Service *Host::getService(const std::string &name) {
    auto it = _services_by_name.find(name);
    return it == _services_by_name.end() ? nullptr : it->second;
}

// Called whenever the state of the host changes
void Host::updateNetworkOutages(std::chrono::system_clock::time_point now,
                                const result_handler_t &result_handler) {
    // assert(this->_has_been_checked)
    // Only do something if we switch between OK/NON-OK
    if (getLastState() != HostState::up && getCurrentState() != HostState::up) {
        return;
    }

    for (auto *child : _children) {
        if (child->_check_command_line == "@parents") {
            child->updateReachability(now, result_handler);
        } else if (child->getCurrentState() != HostState::up) {
            // child is currently down or unreachable
            switch (child->getReachability()) {
                case Reachability::unreachable:
                    result_handler(
                        now,
                        CheckResult{
                            child, HostState::unreachable,
                            "child becomes unreachable due to state change of parent host"});
                    break;
                case Reachability::reachable:
                case Reachability::unknown:
                    break;
            }
        }
    }
}

// A host is considered reachable if either it has no parents or if at least one
// parent is UP.
Host::Reachability Host::getReachability() const {
    if (_parents.empty()) {
        return Reachability::reachable;
    }

    bool all_parents_have_been_checked = true;
    for (auto *parent : _parents) {
        if (parent->state()->has_been_checked_) {
            if (parent->getCurrentState() == HostState::up) {
                return Reachability::reachable;
            }
        } else {
            all_parents_have_been_checked = false;
        }
    }
    return all_parents_have_been_checked ? Reachability::unreachable
                                         : Reachability::unknown;
}

void Host::setUp(std::chrono::system_clock::time_point now,
                 const result_handler_t &result_handler) {
    checkStarts(now);
    result_handler(now, {this, HostState::up, "Always assumed to be up"});
}

void Host::setServiceDoesNotExist(std::chrono::system_clock::time_point now,
                                  const result_handler_t &result_handler) {
    const auto [host_name, description] = serviceFromRouting();
    checkStarts(now);
    result_handler(now, {this,
                         returnCodeToHostState(
                             this, static_cast<int>(ServiceState::unknown)),
                         R"(The configured service ")" + host_name + ";" +
                             description + R"(" does not exist.)"});
}

// Used for hosts that have check method "@parents".
void Host::updateReachability(std::chrono::system_clock::time_point now,
                              const result_handler_t &result_handler) {
    switch (getReachability()) {
        case Reachability::reachable:
            checkStarts(now);
            result_handler(now,
                           {this, HostState::up,
                            "Assumed up, because at least one parent is up"});
            break;
        case Reachability::unreachable:
            checkStarts(now);
            result_handler(
                now, {this, HostState::unreachable,
                      "Assumed unreachable, because all parents are down"});
            break;
        case Reachability::unknown:
            break;
    }
}

std::pair<std::string, std::string> Host::serviceFromRouting() const {
    // skip "@service" prefix and split the remaining part
    auto host_name_and_description =
        mk::nextField(mk::nextField(_check_command_line).second, ";");
    return {host_name_and_description.first,
            mk::lstrip(host_name_and_description.second)};
}

std::chrono::system_clock::duration Host::staleness_check_interval() const {
    return _check_command_line == "@smartping" ? _smart_ping_timeout
                                               : _check_interval;
};

void Host::insertInGroup(HostGroup &group) {
    group.insert(this);
    _groups.insert(&group);
}

bool Host::operator==(const Host &other) const {
    return Object::operator==(other) &&
           std::all_of(_state_mapping.begin(), _state_mapping.end(),
                       [&](const auto &l) {
                           return l.second == other._state_mapping.at(l.first);
                       }) &&
           _smart_ping_timeout == other._smart_ping_timeout &&
           _ipv4_address == other._ipv4_address &&
           std::is_permutation(
               _services.begin(), _services.end(), other._services.begin(),
               other._services.end(),
               [&](const auto &l, const auto &r) { return *l == *r; }) &&
           std::is_permutation(_children.begin(), _children.end(),
                               other._children.begin(), other._children.end(),
                               [&](const auto *l, const auto *r) {
                                   return l->name() == r->name();
                               }) &&
           std::is_permutation(_parents.begin(), _parents.end(),
                               other._parents.begin(), other._parents.end(),
                               [&](const auto *l, const auto *r) {
                                   return l->name() == r->name();
                               });
}
