// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/ICMPReceiver.h"

#include <unistd.h>

#include <algorithm>
#include <chrono>
#include <filesystem>
#include <optional>
#include <string_view>
#include <utility>

#include "cmc/CheckResult.h"
#include "cmc/GlobalConfig.h"
#include "cmc/Host.h"
#include "cmc/Object.h"
#include "cmc/Process.h"
#include "cmc/World.h"
#include "cmc/cmc.h"
#include "livestatus/LogEntry.h"
#include "livestatus/Logger.h"
#include "livestatus/Poller.h"
#include "livestatus/StringUtils.h"

using namespace std::string_literals;

ICMPReceiver::ICMPReceiver(std::string name, Logger *logger,
                           result_handler_t result_handler)
    : ProcessHelper{std::move(name), "", {}, logger}
    , result_handler_{std::move(result_handler)} {
    _buffer.reserve(4096);
}

ICMPReceiver::~ICMPReceiver() { do_stop(); }

void ICMPReceiver::start_() { spawnConnectedSubprocess(); }

void ICMPReceiver::stop_() { do_stop(); }

void ICMPReceiver::setUpPoller_(Poller &poller) const {
    if (stdOut() >= 0) {
        poller.addFileDescriptor(stdOut(), PollEvents::in);
    }
    if (stdErr() >= 0) {
        poller.addFileDescriptor(stdErr(), PollEvents::in);
    }
}

void ICMPReceiver::communicate_(const Poller &poller) {
    if (stdOut() >= 0 && poller.isFileDescriptorSet(stdOut(), PollEvents::in)) {
        handleStdOut();
    }
    if (stdErr() >= 0 && poller.isFileDescriptorSet(stdErr(), PollEvents::in)) {
        handleStdErr();
    }
}

void ICMPReceiver::handleStdErr() {
    // Read any lines from stderr (error messages) and simply log them
    char buffer[4096];
    auto bytes_read = ::read(stdErr(), buffer, sizeof(buffer));
    if (bytes_read == -1) {
        return;
    }
    std::string_view response{buffer, size_t(bytes_read)};
    while (!response.empty()) {
        auto line = response.substr(0, response.find('\n'));
        response.remove_prefix(std::min(response.size(), line.size() + 1));
        Warning(logger()) << line;
    }
}

[[nodiscard]] bool ICMPReceiver::updateConfig(const World *world) {
    _world = world;
    return updateConfig(world->config_.global_config_.smart_ping_);
}

[[nodiscard]] bool ICMPReceiver::updateConfig(
    const SmartPingConfig &smart_ping) {
    bool changed = false;
    changed |= updatePath(smart_ping.icmp_receiver_executable_);
    args_type args{};
    if (smart_ping.ignore_rst_) {
        args.emplace_back("--ignore-rst");
    }
    changed |= updateArgs(args);
    return changed;
}

void ICMPReceiver::do_stop() {
    _buffer.clear();
    _start = 0;
    // icmpreceiver needs to be killed. It's stuck in the PCAP loop
    if (auto &p = process()) {
        p->terminate();
    }
    despawn();
}

void ICMPReceiver::handleStdOut() {
    readResults();
    auto now = std::chrono::system_clock::now();
    while (const in_addr_t a = nextResult()) {
        for (Host *host : _world->getHostsByIPv4Address(a)) {
            if (host->isCheckEnabled(CheckResultKind::active) &&
                host->_check_command_line == "@smartping") {
                if (logger()->isLoggable(LogLevel::debug)) {
                    Debug(logger())
                        << *host << " is UP because we received a packet from "
                        << mk::ipv4ToString(a);
                }
                host->checkStarts(now);
                result_handler_(now,
                                CheckResult{host, HostState::up,
                                            "Packet received via smart PING"});
            }
        }
    }
}

void ICMPReceiver::readResults() {
    // shift buffered data to the beginning of the buffer
    _buffer.erase(_buffer.begin(), _buffer.begin() + _start);
    _start = 0;

    const size_t old_size = _buffer.size();
    const size_t bytes_to_read = _buffer.capacity() - old_size;
    _buffer.resize(_buffer.capacity());
    const ssize_t r = ::read(stdOut(), &_buffer[old_size], bytes_to_read);
    if (logger()->isLoggable(LogLevel::debug)) {
        Debug(logger()) << "received " << r << " of " << bytes_to_read
                        << " bytes from icmpreceiver";
    }
    if (r == -1) {
        const generic_error ge("Cannot read response from icmpreceiver");
        Warning(logger()) << ge;
        restart();
    } else if (r == 0) {
        Warning(logger()) << "closed connection";
        restart();
    } else {
        _buffer.resize(old_size + r);
    }
}

in_addr_t ICMPReceiver::nextResult() {
    if (_start + sizeof(in_addr_t) > _buffer.size()) {
        return 0;
    }
    const in_addr_t a = *reinterpret_cast<in_addr_t *>(&_buffer[_start]);
    if (logger()->isLoggable(LogLevel::debug)) {
        Debug(logger()) << "received packet from " << mk::ipv4ToString(a);
    }
    _start += sizeof(in_addr_t);
    return a;
}
