// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/ICMPSender.h"

#include <unistd.h>

#include <algorithm>
#include <cerrno>
#include <chrono>
#include <filesystem>
#include <optional>
#include <string_view>
#include <utility>

#include "cmc/CheckResult.h"
#include "cmc/Core.h"
#include "cmc/GlobalConfig.h"
#include "cmc/Host.h"
#include "cmc/IOHelper.h"
#include "cmc/Object.h"
#include "cmc/Schedule.h"
#include "cmc/World.h"
#include "cmc/cmc.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/Logger.h"
#include "livestatus/Poller.h"
#include "livestatus/StringUtils.h"

using namespace std::literals::chrono_literals;
using namespace std::string_literals;

ICMPSender::ICMPSender(std::string name, Logger *logger)
    : ProcessHelper{std::move(name), "", {}, logger} {
    _queue.reserve(65536);
}

ICMPSender::~ICMPSender() { do_stop(); }

void ICMPSender::start_() { spawnConnectedSubprocess(); }

void ICMPSender::stop_() { do_stop(); }

void ICMPSender::setUpPoller_(Poller &poller) const {
    if (!queueEmpty() && stdIn() >= 0) {
        poller.addFileDescriptor(stdIn(), PollEvents::out);
    }
    if (stdErr() >= 0) {
        poller.addFileDescriptor(stdErr(), PollEvents::in);
    }
}

void ICMPSender::communicate_(const Poller &poller) {
    if (!queueEmpty() && stdIn() >= 0 &&
        poller.isFileDescriptorSet(stdIn(), PollEvents::out)) {
        handleStdIn();
    }
    if (stdErr() >= 0 && poller.isFileDescriptorSet(stdErr(), PollEvents::in)) {
        handleStdErr();
    }
}

void ICMPSender::handleStdErr() {
    // Read any lines from stderr (error messages) and simply log them
    char buffer[4096];
    size_t bytes_read{0};
    while (bytes_read < sizeof(buffer)) {
        if (!Poller{}.wait(3s, stdErr(), PollEvents::in, logger())) {
            break;
        }
        auto ret =
            ::read(stdErr(), buffer + bytes_read, sizeof(buffer) - bytes_read);
        if (ret > 0) {
            bytes_read += ret;
        } else if (ret == 0) {
            break;
        } else if (ret == -1 && (errno == EAGAIN || errno == EWOULDBLOCK)) {
            continue;
        } else {
            return;
        }
    }
    std::string_view response{buffer, bytes_read};
    while (!response.empty()) {
        auto line = response.substr(0, response.find('\n'));
        response.remove_prefix(std::min(response.size(), line.size() + 1));
        Warning(logger()) << line;
    }
}

bool ICMPSender::updateConfig(const World *world) {
    return updateConfig(world->config_.global_config_.smart_ping_);
}

bool ICMPSender::updateConfig(const SmartPingConfig &smart_ping) {
    bool changed = false;
    changed |= updatePath(smart_ping.icmp_sender_executable_);
    args_type args{std::to_string(smart_ping.num_sockets_),
                   std::to_string(mk::ticks<std::chrono::microseconds>(
                       smart_ping.wait_duration_)),
                   std::to_string(smart_ping.bulk_size_)};
    if (smart_ping.omit_payload_) {
        args.emplace_back("--omit-payload");
    }
    changed |= updateArgs(args);
    return changed;
}

void ICMPSender::do_stop() {
    _queue.clear();
    despawn();
}

void ICMPSender::schedule(Schedule<Host *> &schedule, Core &core) {
    auto start = std::chrono::steady_clock::now();
    while (!queueFull()) {
        auto now = std::chrono::system_clock::now();
        if (auto job = schedule.extractLowestUnder(now)) {
            // send ping if checks are enabled
            auto *host = job->value();
            if (host->isCheckEnabled(CheckResultKind::active)) {
                if (host->_ipv4_address != 0U) {
                    host->checkStarts(now);
                    queuePing(host->_ipv4_address);
                } else {
                    if (logger()->isLoggable(LogLevel::debug)) {
                        Debug(logger())
                            << "no IP address, not queueing packet for "
                            << *host;
                    }
                    core.processCheckResult(
                        now,
                        CheckResult(
                            host, returnCodeToHostState(host, 2),
                            "Failed to lookup IP address and no explicit IP address configured"));
                }
            }
            if (auto t = schedule.insert({host, host->nextCheckTime()})) {
                host->setNextCheck(*t, schedule.logger());
            }
        } else {
            break;
        }
    }
    const double e = elapsed_ms_since(start);
    if (e > 100.0) {
        Warning(logger()) << "WARNING: ping-queueing has lasted " << e
                          << " ms!";
    }
}

bool ICMPSender::queueFull() const {
    return _queue.size() > _queue.capacity() - sizeof(in_addr_t);
}

bool ICMPSender::queueEmpty() const { return _queue.empty(); }

void ICMPSender::queuePing(in_addr_t a) {
    if (logger()->isLoggable(LogLevel::debug)) {
        Debug(logger()) << "queueing packet for " << mk::ipv4ToString(a);
    }
    // ugly!
    const auto *p = reinterpret_cast<const char *>(&a);
    for (size_t i = 0; i < sizeof(in_addr_t); ++i) {
        _queue.push_back(*p++);
    }
}

void ICMPSender::handleStdIn() {
    auto written = ::write(stdIn(), _queue.data(), _queue.size());
    if (written == -1) {
        const generic_error ge("Cannot send IP addresses to icmpsender");
        Warning(logger()) << ge;
        restart();
        return;
    }
    if (logger()->isLoggable(LogLevel::debug)) {
        Debug(logger()) << "sent " << written << " of " << _queue.size()
                        << " bytes to icmpsender";
    }
    _queue.erase(_queue.begin(), _queue.begin() + written);
}
