// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/InfluxDBConnection.h"

#include <algorithm>
#include <compare>
#include <iterator>
#include <sstream>

#include "cmc/InfluxDBLineProtocol.h"
#include "cmc/InfluxDBProtocol.h"
#include "livestatus/global_counters.h"

using namespace std::chrono_literals;
using namespace std::string_literals;

namespace influxdb {

Connection::Connection(const influxdb::ConnectionConfig &config,
                       Counter bytes_sent_counter)
    : _config(config)
    , _bytes_sent_counter{bytes_sent_counter}
    , _state(State::unconnected)
    , _log_prefix([](const influxdb::ConnectionConfig &config) {
        auto out = std::ostringstream{};
        out << "[influxdb " << config.address.to_string() << "] ";
        return out.str();
    }(config)) {}

Connection::State Connection::tryToInitiateConnection(
    std::chrono::steady_clock::time_point time_now) {
    if (_state != State::unconnected || !_config.enabled ||
        time_now - _last_connection_attempt <= _config.retry_interval) {
        return _state;
    }

    _sender = std::make_unique<influxdb::Sender>(_config, logger());
    _state = State::connection_in_progress;
    Notice(logger()) << "Successfully initiated connection";

    return _state;
}

Connection::State Connection::tryToCompleteConnection(
    std::chrono::steady_clock::time_point time_now) {
    if (_state != State::connection_in_progress) {
        return _state;
    }
    if (!_sender) {
        return _state;
    }
    _last_connection_attempt = time_now;
    if (!_sender->ping()) {
        close();
        return _state;
    }
    if (_sender->connected()) {
        Notice(logger()) << "Successfully connected";
        _state = State::connected;
    }
    return _state;
}

void Connection::doHousekeeping() {
    auto now = std::chrono::steady_clock::now();
    if (tryToCompleteConnection(now) != State::connected) {
        return;  // no need to spam warnings to the logfile
    }
    auto lines_sent = sendBatch(now);
    if (lines_sent != 0) {
        Debug(logger()) << "Sent " << lines_sent << " lines";
    }
}

void Connection::send(const protocol::Update &message) {
    auto now = std::chrono::steady_clock::now();
    if (tryToCompleteConnection(now) != State::connected) {
        return;  // no need to spam warnings to the logfile
    }
    auto batched = batch(message.sender_config, message.message, now);
    if (batched != 0) {
        Debug(logger()) << "Batched " << batched << " lines";
    }
    auto lines_sent = sendBatch(now);
    if (lines_sent != 0) {
        Debug(logger()) << "Sent " << lines_sent << " lines";
    }
}

std::size_t Connection::batch(const influxdb::SenderConfig &sender_config,
                              const std::vector<LineProtocol> &new_lines,
                              std::chrono::steady_clock::time_point now) {
    if (new_lines.empty() || sender_config.connection != _config.id) {
        return 0;
    }
    auto it = _batch.find(sender_config);
    if (it == _batch.end()) {
        _batch.emplace(sender_config, std::make_pair(now, new_lines));
    } else {
        it->second.second.insert(it->second.second.end(),
                                 std::make_move_iterator(new_lines.begin()),
                                 std::make_move_iterator(new_lines.end()));
    }
    return new_lines.size();
}

std::size_t Connection::sendBatch(std::chrono::steady_clock::time_point now) {
    if (!_sender) {
        return 0;
    }
    std::size_t lines_sent = 0;
    std::size_t bytes_sent = 0;
    auto it = _batch.begin();
    while (it != _batch.end()) {
        auto &[sender_config, batch_value] = *it;
        if (now - batch_value.first < _config.max_batch_age &&
            batch_value.second.size() < _config.max_batch_size) {
            it++;
            continue;
        }
        auto [ok, sent] = _sender->post(sender_config, batch_value.second);
        if (!ok) {
            close();
            break;
        }
        lines_sent += batch_value.second.size();
        bytes_sent += sent;
        it = _batch.erase(it);
    }
    counterIncrementBy(_bytes_sent_counter, bytes_sent);
    return lines_sent;
}

void Connection::close() {
    if (_state == State::unconnected) {
        return;
    }
    Notice(logger()) << "Closing connection";
    _sender.reset();
    _state = State::unconnected;
}
}  // namespace influxdb
