// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/InfluxDBLineProtocol.h"

#include <algorithm>
#include <cstddef>
#include <iterator>
#include <sstream>
#include <string_view>

// overload.h is necessary for mk::overload but IWYU wants it gone.
#include "livestatus/ChronoUtils.h"
#include "livestatus/overload.h"

namespace influxdb {
namespace {
std::string quote(const std::string &str, const char *chr) {
    return chr + str + chr;
}

std::string escape(std::string_view str, const char *seq) {
    std::stringstream line{};
    std::size_t pos = 0;
    std::size_t start = 0;
    while ((pos = str.find_first_of(seq, start)) != std::string::npos) {
        line << str.substr(start, pos - start) << "\\" << str[pos];
        start = ++pos;
    }
    line << str.substr(start, str.length() - start);
    return line.str();
}

std::ostream &operator<<(std::ostream &out, const Field::value_type &key) {
    std::visit(
        mk::overload{
            [&out](const double &k) -> std::ostream & { return out << k; },
            [&out](const std::int64_t &k) -> std::ostream & {
                return out << k << "i";
            },
            [&out](const std::uint64_t &k) -> std::ostream & {
                return out << k << "u";
            },
            [&out](const detail::FieldString &k) -> std::ostream & {
                return out << std::string{k};
            },
            [&out](const bool &k) -> std::ostream & {
                return out << (k ? "true" : "false");
            }},
        key);
    return out;
}

std::ostream &operator<<(
    std::ostream &out, const std::chrono::system_clock::time_point &timestamp) {
    return out << mk::ticks<LineProtocol::resolution>(
               timestamp.time_since_epoch());
}
}  // namespace

detail::KeyString::operator std::string() const {
    return escape(value, detail::KeyString::escape_sequence);
}

detail::TagString::operator std::string() const {
    return escape(value, detail::TagString::escape_sequence);
}

detail::FieldString::operator std::string() const {
    return quote(escape(value, detail::FieldString::escape_sequence), "\"");
}

Measurement::operator std::string() const {
    return escape(value, Measurement::escape_sequence);
}

std::ostream &operator<<(std::ostream &out, const detail::KeyString &x) {
    return out << std::string{x};
}

std::ostream &operator<<(std::ostream &out, const detail::TagString &x) {
    return out << std::string{x};
}

std::ostream &operator<<(std::ostream &out, const detail::FieldString &x) {
    return out << std::string{x};
}

std::ostream &operator<<(std::ostream &out, const Measurement &x) {
    return out << std::string{x};
}

std::ostream &operator<<(std::ostream &out, const Tag &tag) {
    return out << tag.key << "=" << tag.value;
}

std::ostream &operator<<(std::ostream &out, const Field &field) {
    return out << field.key << "=" << field.value;
}

std::ostream &operator<<(std::ostream &out, const LineProtocol &line) {
    out << line.measurement;
    for (auto &&tag : line.tags) {
        out << "," << tag;
    }
    out << " ";
    std::copy(std::begin(line.fields), std::prev(std::end(line.fields)),
              std::ostream_iterator<Field>(out, ","));
    out << line.fields.back();
    out << " " << line.timestamp;
    out << "\n";
    return out;
}
}  // namespace influxdb
