// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/InfluxDBMetric.h"

#include <compare>
#include <iterator>
#include <limits>
#include <stdexcept>

influxdb::Metric::Value influxdb::Metric::Value::make(
    const std::string &serialized) {
    if (serialized.empty() || serialized == "U" || serialized == "u") {
        return {};
    }
    try {
        return Value{std::stod(serialized)};
    } catch (const std::invalid_argument &) {
        return Value{};
    } catch (const std::out_of_range &) {
        return Value{};
    }
}

namespace {
template <typename T>
T stonum(const std::string &x);

template <>
double stonum(const std::string &x) {
    return std::stod(x);
}

template <typename T>
influxdb::Metric::Range make_range(const std::string &serialized) {
    // Symbols
    //
    // [] : optional
    // (|) : alternative
    // BB : lower bound
    // EE : higher bound
    auto begin = std::begin(serialized);
    auto invert = serialized.starts_with("@");
    if (invert) {
        begin += 1;
    }

    // "[@]EE"
    auto sep = serialized.find_first_of(':');
    if (sep == std::string::npos) {
        return influxdb::Metric::Range{
            T{0},
            stonum<T>(serialized.substr(
                std::distance(std::begin(serialized), begin))),
            invert};
    }

    // "[@](BB|~):[EE]"
    auto lower_bound =
        serialized.find_first_of('~') == std::string::npos
            ? stonum<T>(serialized.substr(
                  std::distance(std::begin(serialized), begin), sep))
            : std::numeric_limits<T>::lowest();

    // "[@](BB|~):"
    if (sep == serialized.size() - 1) {
        return influxdb::Metric::Range{lower_bound,
                                       std::numeric_limits<T>::max(), invert};
    }

    // "[@](BB|~):EE"
    return influxdb::Metric::Range{
        lower_bound, stonum<T>(serialized.substr(sep + 1)), invert};
}
}  // namespace

influxdb::Metric::Range influxdb::Metric::Range::make(
    const std::string &serialized) {
    if (serialized.empty()) {
        return Range{};
    }

    try {
        return make_range<double>(serialized);
    } catch (const std::invalid_argument &) {
        return Range{};
    } catch (const std::out_of_range &) {
        return Range{};
    }
}

bool influxdb::Metric::Range::contains(const bound_type &value) const {
    // No alarm on an empty range.
    const bool in_range = !bounds_ || (bounds_->lower_bound <= value &&
                                       value <= bounds_->higher_bound);
    return inverted() ? !in_range : in_range;
}

influxdb::Metric influxdb::Metric::make(const ::Metric &m) {
    return Metric{m.name(),
                  Value::make(m.value()),
                  m.uom(),
                  Range::make(m.warn()),
                  Range::make(m.crit()),
                  Value::make(m.min()),
                  Value::make(m.max())};
}
