// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/InfluxDBSender.h"

#include <algorithm>
#include <filesystem>
#include <iterator>
#include <map>
#include <sstream>

#include "cmc/InfluxDBLineProtocol.h"
#include "httplib.h"
#include "livestatus/Logger.h"
#include "livestatus/StringUtils.h"

using namespace std::string_literals;

namespace {
[[nodiscard]] influxdb::Headers makeHeaders(
    const influxdb::ConnectionConfig &config,
    const influxdb::SenderConfig &sender_config,
    const std::multimap<std::string, std::string> &extra) {
    return influxdb::Headers{config.token, sender_config.org, extra};
}

class HttplibSender : public influxdb::SenderImpl {
public:
    HttplibSender(const influxdb::ConnectionConfig &config, Logger *logger)
        : cli_{config.address.to_string()}, logger_{logger} {
        cli_.set_keep_alive(true);
        // https://github.com/yhirose/cpp-httplib#compress-request-body-on-client
        cli_.set_compress(true);
        cli_.enable_server_certificate_verification(
            config.auth.use_certificate_check);
        cli_.set_ca_cert_path(config.auth.ca_cert_file);
    }

    bool get(const std::string &uri,
             const influxdb::Headers &headers) override {
        Debug(logger_) << "GET " << uri << " " << headers;
        return handleResult(cli_.Get(uri, toHttplibHeaders(headers)), "GET");
    }

    bool get_noauth(const std::string &uri) override {
        Debug(logger_) << "GET " << uri;
        return handleResult(cli_.Get(uri), "GET");
    }

    std::pair<bool, std::size_t> post(
        const std::string &uri, const influxdb::Headers &headers,
        const std::string &content, const std::string &content_type) override {
        if (logger_->isLoggable(LogLevel::debug)) {
            Debug(logger_) << "POST " << uri << " " << headers << " "
                           << " [" << content.size() << "B]\n"
                           << content;
        }
        auto ok = handleResult(
            cli_.Post(uri, toHttplibHeaders(headers), content, content_type),
            "POST");
        return std::make_pair(ok, ok ? content.size() : 0);
    }

    [[nodiscard]] bool connected() const override {
        return cli_.is_socket_open() != 0;
    }

private:
    httplib::Client cli_;
    Logger *logger_;

    static httplib::Headers toHttplibHeaders(const influxdb::Headers &h) {
        httplib::Headers output{{"Authorization", std::string{h.token}},
                                {"org", h.org}};
        output.insert(h.extra.begin(), h.extra.end());
        return output;
    }

    bool handleResult(const httplib::Result &result, const std::string &label) {
        if (!result) {
            Error(logger_) << label << " failed [" << result.error() << "]";
            return false;
        }
        if (result->status < 200 or result->status >= 300) {
            Error(logger_) << label << " [" << result->status << "] "
                           << result->body;
            return false;
        }
        Debug(logger_) << label << " [" << result->status << "] "
                       << result->body;
        return true;
    }
};
}  // namespace

namespace influxdb {
Sender::Sender(const ConnectionConfig &config, Logger *logger)
    : config_{config}, impl_{std::make_unique<HttplibSender>(config, logger)} {}

bool Sender::ping() const {
    const auto uri = std::string{ping_uri_};
    return impl_->get_noauth(uri);
}

std::pair<bool, std::size_t> Sender::post(
    const SenderConfig &sender_config,
    const std::vector<LineProtocol> &lines) const {
    if (lines.empty()) {
        return std::make_pair(true, 0);
    }
    // https://docs.influxdata.com/influxdb/v2.0/write-data/best-practices/optimize-writes/#write-multiple-data-points-in-one-request
    std::ostringstream out;
    std::copy(std::begin(lines), std::end(lines),
              std::ostream_iterator<LineProtocol>(out, "\n"));
    return post(sender_config, out.str());
}

std::pair<bool, std::size_t> Sender::post(const SenderConfig &sender_config,
                                          const std::string &content) const {
    const auto uri =
        mk::join({config_.api_path, post_uri_, "&org=", sender_config.org,
                  "&bucket=", sender_config.bucket,
                  "&precision=", LineProtocol::TimestampSuffix},
                 "");
    const auto headers =
        makeHeaders(config_, sender_config, {{"Accept", "application/json"}});
    return impl_->post(uri, headers, content, content_type_);
}

bool Sender::connected() const { return impl_->connected(); }
}  // namespace influxdb
