// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/InfluxDBSerializer.h"

#include <algorithm>
#include <chrono>
#include <cstdint>
#include <iterator>
#include <optional>
#include <string>
#include <unordered_map>
#include <utility>
#include <variant>

#include "cmc/Host.h"
#include "cmc/InfluxDBConfig.h"
#include "cmc/InfluxDBLineProtocol.h"
#include "cmc/InfluxDBMetric.h"
#include "cmc/Object.h"
#include "cmc/Service.h"
#include "cmc/State.h"
#include "livestatus/Interface.h"
#include "livestatus/Metric.h"
#include "livestatus/PerformanceData.h"

using namespace std::string_literals;

namespace {
// Converting a variant to a superset is safe but C++...  :(
// https://stackoverflow.com/questions/47203255/convert-stdvariant-to-another-stdvariant-with-super-set-of-types
template <typename... Args>
struct variant_cast_proxy {
    template <typename... ToArgs>
    // NOLINTNEXTLINE(hicpp-explicit-conversions,google-explicit-constructor)
    [[nodiscard]] constexpr operator std::variant<ToArgs...>() const {
        return std::visit(
            [](auto &&arg) -> std::variant<ToArgs...> { return arg; }, v);
    }
    std::variant<Args...> v;
};

template <typename... Args>
constexpr variant_cast_proxy<Args...> variant_cast(
    const std::variant<Args...> &v) {
    return {v};
}
}  // namespace

namespace {
template <typename T>
void emplace_if(std::vector<T> &vec, bool enabled,
                const typename T::key_type &k,
                const typename T::value_type &v) {
    if (!enabled) {
        return;
    }
    vec.emplace_back(k, v);
}
}  // namespace

namespace influxdb {
class Serializer {
public:
    explicit Serializer(PerformanceData perf_data, const Object *object)
        : perf_data_{std::move(perf_data)}, object_{object} {}
    [[nodiscard]] std::vector<LineProtocol> serialize() const {
        auto lines = metric_value();
        auto hst_state = metric_hst_state();
        if (hst_state) {
            lines.push_back(*hst_state);
        }
        auto svc_state = metric_svc_state();
        if (svc_state) {
            lines.push_back(*svc_state);
        }
        return lines;
    }

private:
    [[nodiscard]] std::vector<Tag> tags() const;
    [[nodiscard]] std::vector<LineProtocol> metric_value() const;
    [[nodiscard]] std::optional<LineProtocol> metric_hst_state() const;
    [[nodiscard]] std::optional<LineProtocol> metric_svc_state() const;

    PerformanceData perf_data_;
    const Object *object_;
    [[nodiscard]] influxdb::SenderConfig sender_config() const {
        return object_->_influxdb_send_config;
    }
    [[nodiscard]] std::chrono::system_clock::time_point timestamp() const {
        return object_->state()->last_check_;
    }
};

std::vector<Tag> Serializer::tags() const {
    const auto &parts = sender_config().parts;
    auto t = std::vector<Tag>{};
    emplace_if(t, parts.tag_hst_name, "host_name"s, object_->host()->name());
    emplace_if(t, parts.tag_svc_name && object_->isService(),
               "service_description"s, object_->serviceDescription());
    emplace_if(t, parts.tag_hst_alias, "host_alias"s,
               object_->host()->lookup({}, "alias"));
    if (parts.tag_hst_tags) {
        auto attrs = object_->host()->attributes(AttributeKind::tags);
        std::transform(std::begin(attrs), std::end(attrs),
                       std::back_inserter(t), [](const auto &kv) -> Tag {
                           return {kv.first, kv.second};
                       });
    }
    if (parts.tag_hst_label) {
        auto attrs = object_->host()->attributes(AttributeKind::labels);
        std::transform(std::begin(attrs), std::end(attrs),
                       std::back_inserter(t), [](const auto &kv) -> Tag {
                           return {kv.first, kv.second};
                       });
    }
    if (parts.tag_svc_label) {
        auto attrs = object_->service()->attributes(AttributeKind::labels);
        std::transform(std::begin(attrs), std::end(attrs),
                       std::back_inserter(t), [](const auto &kv) -> Tag {
                           return {kv.first, kv.second};
                       });
    }
    return t;
}

std::vector<LineProtocol> Serializer::metric_value() const {
    if (!sender_config().parts.metric_value || perf_data_.empty()) {
        return {};
    }

    const auto t = tags();
    const auto &parts = sender_config().parts;
    auto fields_common = std::vector<Field>{};
    emplace_if(fields_common, parts.add_hst_state && object_->isHost(),
               "host_state"s, std::uint64_t{object_->state()->current_state_});
    emplace_if(fields_common, parts.add_svc_state && object_->isService(),
               "service_state"s,
               std::uint64_t{object_->state()->current_state_});
    emplace_if(fields_common, parts.add_downtime, "in_downtime"s,
               object_->state()->downtime_.scheduled_depth_ > 0);
    emplace_if(
        fields_common, parts.add_acknowledgement, "is_acknowledged"s,
        object_->state()->acknowledgement_type_ != AcknowledgementType::none);

    std::vector<LineProtocol> lines{};
    lines.reserve(perf_data_.size());
    for (const auto &m : perf_data_) {
        const auto metric = influxdb::Metric::make(m);
        const auto measurement = Measurement{metric.name().string()};
        auto fields = fields_common;
        emplace_if(fields, parts.add_uom, "uom"s, metric.uom());
        if (parts.add_thresholds) {
            if (const auto bounds = metric.warn().bounds()) {
                fields.emplace_back("warn_low"s,
                                    variant_cast(bounds->lower_bound));
                fields.emplace_back("warn_high"s,
                                    variant_cast(bounds->higher_bound));
                fields.emplace_back(
                    "warn"s, !metric.value().empty() &&
                                 metric.warn().contains(*metric.value()));
            }
            if (const auto bounds = metric.crit().bounds()) {
                fields.emplace_back("crit_low"s,
                                    variant_cast(bounds->lower_bound));
                fields.emplace_back("crit_high"s,
                                    variant_cast(bounds->higher_bound));
                fields.emplace_back(
                    "crit"s, !metric.value().empty() &&
                                 metric.crit().contains(*metric.value()));
            }
        }
        if (parts.add_min_max && !metric.max().empty() &&
            !metric.min().empty()) {
            fields.emplace_back("min"s, variant_cast(*metric.min()));
            fields.emplace_back("max"s, variant_cast(*metric.max()));
        }
        if (parts.metric_value && !metric.value().empty()) {
            fields.emplace_back("value"s, variant_cast(*metric.value()));
        }
        if (!fields.empty()) {
            lines.emplace_back(measurement, t, fields, timestamp());
        }
    };
    return lines;
}

std::optional<LineProtocol> Serializer::metric_hst_state() const {
    if (!sender_config().parts.metric_hst_state || !object_->isHost()) {
        return {};
    }

    const auto &parts = sender_config().parts;
    auto fields = std::vector<Field>{};
    emplace_if(fields, parts.metric_hst_state, "host_state"s,
               std::uint64_t{object_->state()->current_state_});
    emplace_if(fields, parts.add_downtime, "in_downtime"s,
               object_->state()->downtime_.scheduled_depth_ > 0);
    emplace_if(
        fields, parts.add_acknowledgement, "is_acknowledged"s,
        object_->state()->acknowledgement_type_ != AcknowledgementType::none);
    if (fields.empty()) {
        return {};
    }
    return LineProtocol{Measurement{object_->host()->name()}, tags(), fields,
                        timestamp()};
}

std::optional<LineProtocol> Serializer::metric_svc_state() const {
    if (!sender_config().parts.metric_svc_state || !object_->isService()) {
        return {};
    }

    const auto &parts = sender_config().parts;
    auto fields = std::vector<Field>{};
    emplace_if(fields, parts.metric_svc_state, "service_state"s,
               std::uint64_t{object_->state()->current_state_});
    emplace_if(fields, parts.add_downtime, "in_downtime"s,
               object_->state()->downtime_.scheduled_depth_ > 0);
    emplace_if(
        fields, parts.add_acknowledgement, "is_acknowledged"s,
        object_->state()->acknowledgement_type_ != AcknowledgementType::none);
    if (fields.empty()) {
        return {};
    }
    return LineProtocol{Measurement{object_->serviceDescription()}, tags(),
                        fields, timestamp()};
}

std::vector<LineProtocol> serialize(const PerformanceData &perf_data,
                                    const Object *object) {
    return Serializer{perf_data, object}.serialize();
}
}  // namespace influxdb
