// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/Livestatus.h"

#include <fcntl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <unistd.h>

#include <cerrno>
#include <chrono>
#include <cstdlib>
#include <filesystem>
#include <ostream>

#include "cmc/Core.h"
#include "cmc/GlobalConfig.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/Logger.h"
#include "livestatus/Poller.h"
#include "livestatus/Store.h"
#include "livestatus/global_counters.h"

namespace {
constexpr int socket_backlog = 10;
constexpr uint32_t max_allowed_threads = 2000;
}  // namespace

Livestatus::Livestatus(Core *core)
    : AbstractHelper{"livestatus manager",
                     ContextLogger{core->_store.logger(),
                                   [this](std::ostream &os) {
                                       os << "[" << this->name() << "] ";
                                   }}}
    , _core(core)
    , _sock(-1)
    , _socket_path(_core->config()->livestatus_.socket_) {}

Livestatus::~Livestatus() { this->do_stop(); }

void Livestatus::start_() {
    Notice(logger()) << "starting";
    createUnixSocket();

    auto start = std::chrono::steady_clock::now();
    // Pre-create a couple of client threads
    uint32_t n = _core->config()->livestatus_.num_threads_;
    if (n > max_allowed_threads) {
        Notice(logger()) << "sorry, I won't start more than "
                         << max_allowed_threads
                         << " Livestatus threads (you wanted " << n << ")";
        n = max_allowed_threads;
    }
    for (uint32_t i = 0; i < n; i++) {
        createThread();
    }
    Notice(logger()) << "created " << n
                     << " Livestatus threads with stack size "
                     << _core->config()->livestatus_.stack_size_
                     << " and a maximum response size of "
                     << _core->config()->livestatus_.max_response_size_ / 1024 /
                            1024
                     << " MiB in " << elapsed_ms_since(start) << " ms";
}

void Livestatus::stop_() { this->do_stop(); }

void Livestatus::setUpPoller_(Poller &poller) const {
    if (this->sock() >= 0) {
        poller.addFileDescriptor(this->sock(), PollEvents::in);
    }
}

void Livestatus::communicate_(const Poller &poller) {
    if (this->sock() >= 0 &&
        poller.isFileDescriptorSet(this->sock(), PollEvents::in)) {
        this->acceptClient();
    }
}

void Livestatus::do_stop() {
    Notice(logger()) << "terminating " << _threads.size()
                     << " client threads...";
    // The usual two-phase shutdown of a thread pool: First, tell all workers to
    // stop accepting new work...
    for (auto &thread : _threads) {
        thread->prepareForShutdown();
    }
    // ... then wait for each worker to terminate.
    _threads.clear();
    Notice(logger()) << "all threads have terminated";

    if (_sock >= 0) {
        removeUnixSocket();
    }
}

void Livestatus::createUnixSocket() {
    _sock = -1;
    struct stat st;
    if (0 == stat(_socket_path.c_str(), &st)) {
        if (0 == ::unlink(_socket_path.c_str())) {
            Notice(logger()) << "removed stale socket file " << _socket_path;
        } else {
            const generic_error ge("cannot remove stale socket file " +
                                   _socket_path);
            Alert(logger()) << ge;
            return;
        }
    }

    _sock = ::socket(PF_UNIX, SOCK_STREAM, 0);
    if (_sock < 0) {
        const generic_error ge("cannot create UNIX socket");
        Alert(logger()) << ge;
        return;
    }

    // Imortant: close on exec -> forked processes must not inherit it!
    if (::fcntl(_sock, F_SETFD, FD_CLOEXEC) == -1) {
        const generic_error ge("cannot set close-exec bit on socket");
        Alert(logger()) << ge;
        ::close(_sock);
        _sock = -1;
        return;
    }

    // Bind it to its address. This creates the socket file.
    struct sockaddr_un sockaddr {
        .sun_family = AF_UNIX, .sun_path = ""
    };
    _socket_path.copy(&sockaddr.sun_path[0], sizeof(sockaddr.sun_path) - 1);
    sockaddr.sun_path[sizeof(sockaddr.sun_path) - 1] = '\0';
    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
    if (::bind(_sock, reinterpret_cast<struct sockaddr *>(&sockaddr),
               sizeof(sockaddr)) < 0) {
        const generic_error ge("cannot bind UNIX socket to address " +
                               _socket_path);
        Alert(logger()) << ge;
        ::close(_sock);
        _sock = -1;
        return;
    }

    if (0 != chmod(_socket_path.c_str(), 0660)) {
        const generic_error ge("cannot chown UNIX socket at " + _socket_path +
                               " to 0660");
        Error(logger()) << ge;
        ::close(_sock);
        _sock = -1;
        return;
    }

    if (0 != listen(_sock, socket_backlog)) {
        const generic_error ge("cannot listen to unix socket at " +
                               _socket_path);
        Alert(logger()) << ge;
        ::close(_sock);
        _sock = -1;
        return;
    }

    Notice(logger()) << "listening on " << _socket_path;
}

void Livestatus::removeUnixSocket() {
    ::unlink(_socket_path.c_str());
    if (_sock >= 0) {
        ::close(_sock);
        _sock = -1;
    }
    Notice(logger()) << "removed socket " << _socket_path;
}

// If no thread is currently free, we simply do not even wait for a connection,
// relying on the TCP queue.
int Livestatus::sock() const {
    return getFreeLivestatusThread() == _threads.end() ? -1 : _sock;
};

void Livestatus::acceptClient() {
    // Note: To avoid busy waiting, we *always* have to accept() here!
    const int client_fd = ::accept4(_sock, nullptr, nullptr, SOCK_CLOEXEC);
    if (client_fd == -1) {
        const generic_error ge("cannot accept client connections");
        switch (errno) {
            case EMFILE:   // too many files open
            case ENFILE:   // too many files open in system
            case ENOBUFS:  // no buffer space
            case ENOMEM:   // not enough memory
                Emergency(logger()) << "irrecoverable error" << ge;
                std::abort();
            default:
                Warning(logger()) << ge;
                return;
        }
    }
    auto it = getFreeLivestatusThread();
    if (it == _threads.end()) {
        // This should never happen, see sock(), but anyway...
        counterIncrement(Counter::livestatus_overflows);
        ::close(client_fd);
        return;
    }
    counterIncrement(Counter::connections);
    (*it)->acceptClient(client_fd);
}

Livestatus::threads::iterator Livestatus::getFreeLivestatusThread() {
    for (auto it = _threads.begin(); it != _threads.end(); ++it) {
        if (!(*it)->hasClient()) {
            return it;
        }
    }
    return _threads.end();
}

Livestatus::threads::const_iterator Livestatus::getFreeLivestatusThread()
    const {
    for (auto it = _threads.cbegin(); it != _threads.cend(); ++it) {
        if (!(*it)->hasClient()) {
            return it;
        }
    }
    return _threads.cend();
}

// Called by house keeping of Core
void Livestatus::adaptThreadCount(uint32_t target) {
    size_t have = _threads.size();
    size_t terminated = 0;
    while (have > target) {
        auto it = getFreeLivestatusThread();
        if (it == _threads.end()) {
            break;
        }
        removeThread(it);
        have--;
        terminated++;
    }
    if (terminated != 0U) {
        Notice(logger()) << "Terminated " << terminated
                         << " idle livestatus threads";
    }

    if (have < target) {
        Notice(logger()) << "Creating " << (target - have)
                         << " additional Livestatus threads";
    }
    while (have < target) {
        createThread();
        have++;
    }
}

std::string Livestatus::getThreadID() {
    return "client " + std::to_string(_threads.size());
}

void Livestatus::createThread() {
    _threads.push_back(
        std::make_unique<LivestatusThread>(_core, getThreadID()));
    _core->_livestatus_threads++;
}

void Livestatus::removeThread(threads::iterator it) {
    _threads.erase(it);
    _core->_livestatus_threads--;
}
