// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/LivestatusExternalCommand.h"

#include <cstdlib>
#include <optional>
#include <ostream>
#include <stdexcept>
#include <system_error>
#include <utility>

#include "asio/buffer.hpp"
#include "asio/io_context.hpp"
#include "asio/local/connect_pair.hpp"
#include "livestatus/Logger.h"
#include "livestatus/Poller.h"
#include "livestatus/global_counters.h"

LivestatusExternalCommand::LivestatusExternalCommand(
    asio::io_context &io_context, continuation_type cont, Logger *logger)
    : AbstractHelper{"livestatus external",
                     ContextLogger{logger,
                                   [this](std::ostream &os) {
                                       os << "[" << this->name() << "] ";
                                   }}}
    , _command_pop{io_context}
    , _command_push{io_context}
    , _continuation{std::move(cont)} {};

LivestatusExternalCommand::~LivestatusExternalCommand() { this->do_stop(); };

void LivestatusExternalCommand::start_() {
    std::error_code ec;
    asio::local::connect_pair(_command_pop, _command_push, ec);
    if (ec) {
        Emergency(logger()) << ec;
    }
}

void LivestatusExternalCommand::stop_() { this->do_stop(); }

void LivestatusExternalCommand::setUpPoller_(Poller &poller) const {
    if (_command_pop.is_open()) {
        poller.addFileDescriptor(command_pop_fd(), PollEvents::in);
    }
}

void LivestatusExternalCommand::communicate_(const Poller &poller) {
    if (_command_pop.is_open() &&
        poller.isFileDescriptorSet(command_pop_fd(), PollEvents::in)) {
        this->handleExternalCommands();
    }
}

int LivestatusExternalCommand::command_pop_fd() const {
    // Alas, asio's native_handle() is not const-correct.
    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-const-cast)
    return const_cast<socket_type &>(_command_pop).native_handle();
}

void LivestatusExternalCommand::spoolCommand(const std::string &command) {
    auto [ok, size] =
        _command_spool.push(command, queue_overflow_strategy::dont_push);
    switch (ok) {
        case queue_status::ok:
            Debug(logger()) << "spooled command '" << command << "'";
            break;
        case queue_status::overflow:
            Warning(logger())
                << "command queue full, didn't push '" << command << "'";
            return;
        case queue_status::joinable:
            Warning(logger()) << "command queue shutting down, didn't push '"
                              << command << "'";
            return;
    }

    _num_queued_commands_max++;
    counterIncrement(Counter::commands);
    auto r = _command_push.write_some(
        asio::buffer("*", 1));  // send signal to main thread
    if (r != 1) {
        const generic_error ge("Cannot signal master thread about command");
        Critical(logger()) << ge;
    }
}

size_t LivestatusExternalCommand::queueSize() const {
    return _command_spool.approx_size();
}

size_t LivestatusExternalCommand::commandCount() const {
    return _num_queued_commands_max;
}

bool LivestatusExternalCommand::active() const {
    return _command_pop.is_open() && _command_push.is_open();
}

void LivestatusExternalCommand::do_stop() {
    std::error_code ec;
    _command_pop.close(ec);
    if (ec) {
        Error(logger()) << ec;
    }
    _command_push.close(ec);
    if (ec) {
        Error(logger()) << ec;
    }
}

void LivestatusExternalCommand::handleExternalCommands() {
    char drain[64];

    std::error_code ec;
    auto num = _command_pop.read_some(asio::buffer(drain), ec);
    if (ec) {
        Critical(logger()) << ec;
    }
    if (num == 0) {
        // This should never ever happen, but somehow it does... :-/
        Emergency(logger()) << "command signal pipe was closed, terminating...";
        ::abort();
    }
    while (num-- != 0) {
        auto &&elem = _command_spool.pop(queue_pop_strategy::nonblocking, {});
        if (!elem) {
            Critical(logger()) << "command was signalled but queue is empty!";
            return;
        }
        try {
            _continuation(elem->first);
        } catch (const std::invalid_argument &e) {
            // External commands can theoretically supply malformed UTF-8
            Error(logger()) << "malformed results when handle command";
        }
    }
}
