// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/LivestatusLocalCommand.h"

#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>

#include <algorithm>
#include <cerrno>
#include <cstdlib>
#include <sstream>
#include <system_error>
#include <utility>

#include "cmc/GlobalConfig.h"
#include "cmc/World.h"
#include "livestatus/BitMask.h"
#include "livestatus/Logger.h"
#include "livestatus/Poller.h"

LivestatusLocalCommand::LivestatusLocalCommand(continuation_type cont,
                                               std::filesystem::path endpoint,
                                               Logger *logger)
    : AbstractHelper{"livestatus local",
                     ContextLogger{logger,
                                   [this](std::ostream &os) {
                                       os << "[" << this->name() << "] ";
                                   }}}
    , continuation_{std::move(cont)}
    , endpoint_{std::move(endpoint)} {
    buffer_[0] = 0;
}

LivestatusLocalCommand::LivestatusLocalCommand(continuation_type cont,
                                               Logger *logger)
    : LivestatusLocalCommand{std::move(cont), "", logger} {}

LivestatusLocalCommand::~LivestatusLocalCommand() { do_stop(); }

void LivestatusLocalCommand::start_() { do_start(); }

void LivestatusLocalCommand::stop_() { do_stop(); }

void LivestatusLocalCommand::setUpPoller_(Poller &poller) const {
    if (this->socket() >= 0) {
        poller.addFileDescriptor(this->socket(), PollEvents::in);
    }
}

void LivestatusLocalCommand::communicate_(const Poller &poller) {
    if (socket() < 0) {
        restart();
    }
    if (socket() >= 0 && poller.isFileDescriptorSet(
                             socket(), (PollEvents::in | PollEvents::hup))) {
        handleCommandPipe();
    }
}

int LivestatusLocalCommand::socket() const { return pipe_; }

bool LivestatusLocalCommand::updateConfig(const World *world) {
    return updateConfig(world->config_.global_config_.livestatus_);
}

bool LivestatusLocalCommand::updateConfig(const LivestatusConfig &livestatus) {
    const bool update = endpoint_ != livestatus.command_pipe_;
    if (update) {
        endpoint_ = livestatus.command_pipe_;
    }
    return update;
}

void LivestatusLocalCommand::do_start() {
    buffer_size_ = 0;
    if (std::error_code ec; std::filesystem::exists(endpoint_, ec) &&
                            std::filesystem::is_fifo(endpoint_, ec)) {
        return;
    }
    try {
        std::filesystem::remove_all(endpoint_);
        std::filesystem::create_directories(endpoint_.parent_path());
    } catch (const std::system_error &ec) {
        Emergency(logger()) << "Could not create command pipe at " << endpoint_
                            << ": " << ec.what();
    }
    // NOTE: We have to do the open first and then use the fd afterwards,
    // otherwise we will have race conditions! This is a general principle, but
    // there are various other places where we get this wrong, too.
    do {  // NOLINT(cppcoreguidelines-avoid-do-while)
        pipe_ = ::open(endpoint_.c_str(), O_RDONLY | O_NONBLOCK);
        if (pipe_ == -1) {
            {
                const generic_error ge{"could not open command pipe at \"" +
                                       endpoint_.string() + "\""};
                switch (errno) {
                    case ENOENT:
                        break;
                    case EMFILE:  // too many files open
                    case ENFILE:  // too many files open in system
                    case ENOMEM:  // not enough memory
                        Emergency(logger()) << "irrecoverable error" << ge;
                        std::abort();
                    default:
                        Emergency(logger()) << ge;
                        return;
                }
            }
            if (::mkfifo(endpoint_.c_str(), 0660) == -1) {
                const generic_error ge{"could not create command pipe at \"" +
                                       endpoint_.string() + "\""};
                Emergency(logger()) << ge;
                return;
            }
            Notice(logger()) << "Successfully created new command pipe at "
                             << endpoint_ << ".";
        } else {
            try {
                if (std::filesystem::is_fifo(endpoint_)) {
                    break;
                }
                Alert(logger()) << endpoint_ << " is not a FIFO";
            } catch (const std::system_error &ec) {
                Alert(logger()) << "Could not determine type of " << endpoint_
                                << ": " << ec.what();
            }
            // TODO(ml): The next few lines C&P `do_stop()`.
            ::close(pipe_);
            pipe_ = -1;
            if (::unlink(endpoint_.c_str()) == -1) {
                const generic_error ge{"could not unlink \"" +
                                       endpoint_.string() + "\""};
                Emergency(logger()) << ge;
                return;
            }
            Alert(logger()) << endpoint_ << " has been unlinked";
        }
    } while (pipe_ == -1);
    Notice(logger()) << "Successfully opened command pipe at " << endpoint_
                     << ".";
}

void LivestatusLocalCommand::do_stop() {
    if (pipe_ >= 0) {
        Informational(logger()) << "Closing command pipe " << endpoint_;
        ::close(pipe_);
        pipe_ = -1;
    }
    if (std::error_code ec; std::filesystem::exists(endpoint_, ec) &&
                            !std::filesystem::remove(endpoint_, ec)) {
        Emergency(logger()) << "Could not unlink " << endpoint_ << " " << ec;
        return;
    }
}

void LivestatusLocalCommand::handleCommandPipe() {
    const size_t bytes_left = sizeof(buffer_) - buffer_size_ - 1;
    if (bytes_left == 0) {
        Warning(logger())
            << "Command buffer full while command still incomplete. "
               "Closing pipe.";
        restart();
        return;
    }

    const ssize_t r = ::read(pipe_, buffer_ + buffer_size_, bytes_left);
    if (r > 0) {
        buffer_size_ += r;
        buffer_[buffer_size_] = '\0';
        queuePipedCommands(false);
    } else if (r == 0) {
        Debug(logger()) << "Pipe closed by client. Going to reopen.";
        queuePipedCommands(true);
        restart();
    } else {
        const generic_error ge("Cannot read data from command pipe");
        Warning(logger()) << ge;
        restart();
    }
}

void LivestatusLocalCommand::queuePipedCommands(bool pipe_closed) {
    char *begin = buffer_;
    while (true) {
        const auto cmd =
            findCommand(begin, buffer_ + buffer_size_, pipe_closed);
        if (cmd.empty()) {
            break;
        }
        answerCommandRequest(cmd);
        begin += cmd.size() + 1;
    }
    if (begin != buffer_) {
        std::move(begin, buffer_ + buffer_size_, buffer_);
    }
    buffer_size_ -= begin - buffer_;
}

// static
std::string LivestatusLocalCommand::findCommand(const char *begin,
                                                const char *const end,
                                                bool pipe_closed) {
    const char *b = begin;
    while (*b != '\n' && b < end) {
        b++;
    }
    if (b == begin || b == end) {
        return {};
    }
    if (*b != '\n' && !pipe_closed) {
        return {};
    }
    return {begin, static_cast<size_t>(b - begin)};
}

void LivestatusLocalCommand::answerCommandRequest(const std::string &command) {
    if (command.size() < 14 || command[0] != '[' || command[11] != ']' ||
        command[12] != ' ') {
        Warning(logger()) << "Ignoring malformed command '" << command << "'";
        return;
    }
    continuation_(command.substr(13));
}
