// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/LivestatusThread.h"

#include <unistd.h>

#include <chrono>
#include <functional>
#include <shared_mutex>
#include <sstream>
#include <utility>

#include "cmc/Core.h"
#include "cmc/GlobalConfig.h"
#include "livestatus/InputBuffer.h"
#include "livestatus/Logger.h"
#include "livestatus/OutputBuffer.h"
#include "livestatus/Store.h"
#include "livestatus/StringUtils.h"
#include "livestatus/global_counters.h"
class World;

using namespace std::chrono_literals;

// Thread local pointer to currently active world. By making this thread
// local we are sure that the world cannot be switched during a query.
// Switching of world happends after a config reload, but the old config
// is still available for all running queries.
// NOLINTNEXTLINE(cppcoreguidelines-avoid-non-const-global-variables)
__thread World *g_live_world;

// maximum idle time for connection in keep alive state
constexpr std::chrono::milliseconds idle_timeout = 5min;

// maximum time for reading a query
constexpr std::chrono::milliseconds query_timeout = 10s;

void *thread_run(void *data) {
    auto *lt = static_cast<LivestatusThread *>(data);
    setThreadName(lt->getThreadID());
    lt->run();
    return nullptr;
}

LivestatusThread::LivestatusThread(Core *core, std::string thread_id)
    : _core(core)
    , _thread_id(std::move(thread_id))
    , _client_fd(-1)
    , _logger(core->_store.logger()) {
    shouldTerminate(false);
    pthread_attr_t attr;
    if (auto result = pthread_attr_init(&attr); result != 0) {
        Warning(_logger) << generic_error{
            result, "cannot create livestatus thread attributes"};
    }
    auto stack_size = core->config()->livestatus_.stack_size_;
    if (auto result = pthread_attr_setstacksize(&attr, stack_size);
        result != 0) {
        Warning(_logger) << generic_error{
            result, "cannot set livestatus thread stack size to " +
                        std::to_string(stack_size)};
    }
    if (auto result = pthread_create(&_thread, &attr, thread_run, this);
        result != 0) {
        Warning(_logger) << generic_error{result,
                                          "cannot create livestatus thread"};
    }
    if (auto result = pthread_attr_destroy(&attr); result != 0) {
        Warning(_logger) << generic_error{
            result, "cannot destroy livestatus thread attributes"};
    }
}

LivestatusThread::~LivestatusThread() {
    prepareForShutdown();
    auto fd = client_fd();
    if (fd >= 0) {
        ::close(fd);  // In case of a race condition
    }

    Notice(_logger) << "[" << getThreadID() << "] waiting for termination...";
    if (auto result = pthread_join(_thread, nullptr); result != 0) {
        Warning(_logger) << generic_error{result,
                                          "cannot join livestatus thread"};
    }
    Notice(_logger) << "[" << getThreadID() << "] terminated";
}

void LivestatusThread::prepareForShutdown() {
    if (!shouldTerminate()) {
        shouldTerminate(true);
        _semaphore.post();
    }
}

void LivestatusThread::acceptClient(int fd) {
    client_fd(fd);
    _semaphore.post();  // called by main thread
}

bool LivestatusThread::hasClient() const { return client_fd() >= 0; };

int LivestatusThread::client_fd() const {
    const std::lock_guard<std::mutex> lg(_client_fd_mutex);
    return _client_fd;
}

void LivestatusThread::client_fd(int value) {
    const std::lock_guard<std::mutex> lg(_client_fd_mutex);
    _client_fd = value;
}

bool LivestatusThread::shouldTerminate() const {
    return _should_terminate.load();
}

void LivestatusThread::shouldTerminate(bool value) {
    _should_terminate = value;
}

// All functions below run in the context of the thread

void LivestatusThread::run() {
    while (!shouldTerminate()) {
        _semaphore.wait();
        if (!shouldTerminate()) {
            handleClient();
        }
    }
}

void LivestatusThread::handleClient() {
    const int fd = client_fd();
    _core->_livestatus_active_connections++;
    auto should_terminate = [this] { return shouldTerminate(); };
    InputBuffer input_buffer{fd, should_terminate, _logger, query_timeout,
                             idle_timeout};
    bool keepalive = true;
    while (keepalive && !shouldTerminate()) {
        counterIncrement(Counter::requests);
        OutputBuffer output{fd, should_terminate, _logger};
        keepalive = answerRequest(input_buffer, output);
    }
    ::close(fd);
    client_fd(-1);
    _core->_livestatus_active_connections--;
}

void LivestatusThread::logRequest(const std::string &line,
                                  const std::vector<std::string> &lines) {
    Informational log(_logger);
    log << "request: " << line;
    if (_logger->isLoggable(LogLevel::debug)) {
        for (const auto &l : lines) {
            log << R"(\n)" << l;
        }
    } else {
        const size_t s = lines.size();
        if (s > 0) {
            log << R"(\n{)" << s << (s == 1 ? " line follows" : " lines follow")
                << "...}";
        }
    }
}

// We have our own version of answerRequest and do not use the
// one implemented in Store. That way we can detect a world
// switch that happened while we read the request.
bool LivestatusThread::answerRequest(InputBuffer &input, OutputBuffer &output) {
    // Precondition: output has been reset
    const InputBuffer::Result res = input.readRequest();
    if (res != InputBuffer::Result::request_read) {
        if (res != InputBuffer::Result::eof) {
            std::ostringstream os;
            os << "terminating client connection: " << res;
            output.setError(OutputBuffer::ResponseCode::incomplete_request,
                            os.str());
        }
        return false;
    }
    const std::string line = input.nextLine();
    if (line.starts_with("GET ")) {
        return handleGet(input, output, line, mk::lstrip(line.substr(4)));
    }
    if (line.starts_with("GET")) {
        return handleGet(input, output, line, "");  // only to get error message
    }
    if (line.starts_with("COMMAND ")) {
        logRequest(line, {});
        _core->answerCommandRequest(mk::lstrip(line.substr(8)));
        return true;
    }
    logRequest(line, {});
    output.setError(
        OutputBuffer::ResponseCode::invalid_request,
        "terminating client connection: invalid request '" + line + "'");
    return false;
}

bool LivestatusThread::handleGet(InputBuffer &input, OutputBuffer &output,
                                 const std::string &line,
                                 const std::string &table_name) {
    auto lines = input.getLines();
    logRequest(line, lines);
    const std::shared_lock<std::shared_mutex> sl(_core->_world_lock);
    // make sure that we are using the current world
    g_live_world = _core->_world;
    return _core->_store.answerGetRequest(lines, output, table_name);
}
