// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

// NOTE: The pragma below is needed, without the header we get compilation
// errors.
#include "cmc/Macro.h"

#include <algorithm>
#include <array>
#include <iterator>
#include <optional>
#include <unordered_set>
#include <utility>

#include "cmc/ContactList.h"
#include "cmc/Entity.h"
#include "cmc/Host.h"
#include "cmc/Object.h"
#include "cmc/ObjectGroup.h"  // IWYU pragma: keep
#include "cmc/Service.h"
#include "cmc/State.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/Interface.h"
#include "livestatus/StringUtils.h"

Macro::Macro(std::string name, std::string value)
    : _name(std::move(name)), _value(std::move(value)) {}

Macro::Macro(std::string name, uint32_t value)
    : Macro(std::move(name), std::to_string(value)) {}

Macro::Macro(std::string name, std::chrono::system_clock::time_point value)
    : Macro(std::move(name),
            std::to_string(mk::ticks<std::chrono::microseconds>(
                value.time_since_epoch()))) {}

Macro::Macro(std::string name, const std::vector<std::string> &value)
    : Macro(std::move(name), mk::join(value, ",")) {}

namespace {
template <typename T>
std::vector<std::string> groupNames(const std::unordered_set<T *> &value) {
    std::vector<std::string> names;
    std::transform(value.begin(), value.end(), std::back_inserter(names),
                   [](const T *group) { return group->name(); });
    return names;
}
}  // namespace

std::ostream &operator<<(std::ostream &os, const Macro &m) {
    os << m._name << "=";
    for (auto it = m._value.begin(); it != m._value.end(); ++it) {
        if (it[0] == '\\') {
            os << R"(\\)";
        } else if (it[0] == '\n') {
            if (it[1] != 0) {  // do not send final newline
                os << R"(\n)";
            }
        } else {
            os << it[0];
        }
    }
    return os << "\n";
}

CustomMacros::CustomMacros(std::string prefix, const Entity &entity)
    : _prefix(std::move(prefix)), _entity(entity) {}

std::ostream &operator<<(std::ostream &os, const CustomMacros &cm) {
    for (const auto &[key, value] :
         cm._entity.attributes(AttributeKind::custom_variables)) {
        os << Macro(cm._prefix + mk::unsafe_toupper("_" + key), value);
    }
    return os;
}

ObjectMacros::ObjectMacros(const Object &object) : _object(object) {}

std::ostream &operator<<(std::ostream &os, const ObjectMacros &om) {
    const Object &object = om._object;

    // General information
    os << Macro("MICROTIME", std::chrono::system_clock::now());

    // Host information
    const Host *host = object.host();
    os << Macro("HOSTNOTIFICATIONNUMBER",
                host->state()->notification_.current_number_)
       << Macro("LASTHOSTPROBLEMID", host->state()->notification_.problem_id_)
       << Macro("HOSTPROBLEMID", host->state()->notification_.problem_id_)
       << Macro("HOSTATTEMPT", host->state()->current_attempt_)
       << Macro("MAXHOSTATTEMPTS", host->_max_check_attempts)
       << Macro("HOSTNAME", host->name())
       << Macro("HOSTALIAS", host->lookup({}, "alias"))
       << Macro("HOSTADDRESS", host->lookup({}, "address"))
       << Macro("HOSTOUTPUT", host->state()->check_result_.output_)
       << Macro("HOSTPERFDATA", host->state()->check_result_.perf_data_)
       << Macro("HOSTSTATE", host->statename())
       << Macro("HOSTSTATEID", host->state()->current_state_)
       << Macro("LASTHOSTSTATE", host->lastStatename())
       << Macro("LASTHOSTSTATEID", host->state()->last_state_)
       << Macro("PREVIOUSHOSTHARDSTATE", host->previousHardStatename())
       << Macro("PREVIOUSHOSTHARDSTATEID", host->state()->hard_state_.previous_)
       << Macro("LONGHOSTOUTPUT", host->state()->check_result_.long_output_)
       << Macro("HOSTCHECKCOMMAND", host->_check_command_name)
       << Macro("LASTHOSTSTATECHANGE", std::chrono::system_clock::to_time_t(
                                           host->state()->last_state_change_))
       << Macro("LASTHOSTUP", std::chrono::system_clock::to_time_t(
                                  host->state()->last_time_state_[0]))
       << Macro("HOSTDOWNTIME", host->state()->downtime_.scheduled_depth_)
       << Macro("HOSTGROUPNAMES", groupNames(host->groups()))
       << Macro("HOSTCONTACTGROUPNAMES", host->_contact_list->groupNames())
       << Macro("HOSTTAGS",
                host->lookup(AttributeKind::custom_variables, "TAGS"))
       << Macro("HOST_SL",
                host->lookup(AttributeKind::custom_variables, "EC_SL"))
       << Macro("HOST_EC_CONTACT",
                host->lookup(AttributeKind::custom_variables, "EC_CONTACT"))
       << Macro("HOSTNOTESURL", host->lookup({}, "notes_url"))
       << Macro("HOSTNOTES", host->lookup({}, "notes_expanded"))
       << CustomMacros("HOST", *host);

    if (object.isService()) {
        const Service &service = *object.service();
        const auto *state = service.state();
        os << Macro("SERVICEDESC", service.name())
           << Macro("SERVICEOUTPUT", state->check_result_.output_)
           << Macro("SERVICEPERFDATA", state->check_result_.perf_data_)
           << Macro("SERVICENOTIFICATIONNUMBER",
                    state->notification_.current_number_)
           << Macro("SERVICEPROBLEMID", state->notification_.problem_id_)
           << Macro("LASTSERVICEPROBLEMID", state->notification_.problem_id_)
           << Macro("SERVICEATTEMPT", state->current_attempt_)
           << Macro("MAXSERVICEATTEMPTS", service._max_check_attempts)
           << Macro("SERVICESTATE", service.statename())
           << Macro("SERVICESTATEID", state->current_state_)
           << Macro("LASTSERVICESTATE", service.lastStatename())
           << Macro("LASTSERVICESTATEID", state->last_state_)
           << Macro("PREVIOUSSERVICEHARDSTATE", service.previousHardStatename())
           << Macro("PREVIOUSSERVICEHARDSTATEID", state->hard_state_.previous_)
           << Macro("LASTSERVICESTATECHANGE",
                    std::chrono::system_clock::to_time_t(
                        state->last_state_change_))
           << Macro("LASTSERVICEOK", std::chrono::system_clock::to_time_t(
                                         state->last_time_state_[0]))
           << Macro("SERVICEDOWNTIME", state->downtime_.scheduled_depth_)
           << Macro("LONGSERVICEOUTPUT", state->check_result_.long_output_)
           << Macro("SERVICECHECKCOMMAND", service._check_command_name)
           << Macro("SERVICEGROUPNAMES", groupNames(service.groups()))
           << Macro("SERVICECONTACTGROUPNAMES",
                    service._contact_list->groupNames())
           << Macro("SVC_SL",
                    service.lookup(AttributeKind::custom_variables, "EC_SL"))
           << Macro("SERVICE_SL",
                    service.lookup(AttributeKind::custom_variables, "EC_SL"))
           << Macro(
                  "SERVICE_EC_CONTACT",
                  service.lookup(AttributeKind::custom_variables, "EC_CONTACT"))
           << Macro("SERVICENOTESURL", service.lookup({}, "notes_url"))
           << Macro("SERVICENOTES", service.lookup({}, "notes_expanded"))
           << Macro("SERVICEDISPLAYNAME", service.lookup({}, "display_name"))
           << CustomMacros("SERVICE", object);
    }
    return os;
}
