// Copyright (C) 2020 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/MessagePayload.h"

#include <sys/uio.h>
#include <unistd.h>

#include <algorithm>
#include <cerrno>
#include <numeric>
#include <utility>

#include "livestatus/Logger.h"

MessagePayloadSender::MessagePayloadSender(std::vector<MessagePayload> payloads)
    : content_(std::move(payloads))
    , total_size_{std::accumulate(content_.begin(), content_.end(), size_t{0},
                                  [](size_t sum, const auto &c) {
                                      return sum + std::string_view{c}.size();
                                  })}
    , bytes_sent_{0} {}

int MessagePayloadSender::send(int fd) {
    auto iovs = buildIovs();  // TODO(sp) Would building this once be good?
    if (iovs.empty()) {
        return 0;  // *not* an optimization: writev() expects a positive count!
    }
    while (true) {
        auto ret = ::writev(fd, iovs.data(), iovs.size());
        if (ret != -1) {
            bytes_sent_ += ret;
            return bytes_sent_ == total_size_ ? 0 : EAGAIN;
        }
        if (errno != EINTR) {
            return errno;
        }
    }
}

std::vector<iovec> MessagePayloadSender::buildIovs() const {
    std::vector<iovec> iovs;
    auto pos = bytes_sent_;
    for (const auto &payload : content_) {
        const std::string_view buf{payload};
        if (pos < buf.size()) {
            // We use the iovec only for ::write, so the const_cast below is OK.
            iovs.push_back(
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-const-cast)
                iovec{.iov_base = const_cast<char *>(buf.data() + pos),
                      .iov_len = buf.size() - pos});
            pos = 0;
        } else {
            pos -= buf.size();
        }
    }
    return iovs;
}

MessagePayloadReceiver::status MessagePayloadReceiver::receive(int fd) {
    auto bytes_required{std::string_view{content}.size()};
    while (bytes_received_ < bytes_required) {
        auto num_bytes_read = ::read(fd, content.data() + bytes_received_,
                                     bytes_required - bytes_received_);
        if (num_bytes_read == -1) {
            if (errno == EAGAIN) {
                return status::partial;
            }
            throw generic_error{"failed to read message payload"};
        }

        if (num_bytes_read == 0) {  // EOF
            // Throw error because communication is broken(stopped) at peer
            // side without fulfilling payload. Data are invalid and so on.
            throw generic_error{EPIPE, "peer closed pipe"};
        }

        bytes_received_ += num_bytes_read;
    }
    return status::complete;
}
