// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/MonitoringHistory.h"

#include <ctime>
#include <filesystem>
#include <system_error>
#include <utility>

#include "cmc/GlobalConfig.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/Logger.h"

MonitoringHistory::MonitoringHistory(Logger *logger,
                                     std::function<void()> after_publish)
    : logger_{logger}
    , after_publish_{std::move(after_publish)}
    , last_rotation_{std::chrono::system_clock::now()} {}

MonitoringHistory::~MonitoringHistory() { close(); }

void MonitoringHistory::open(const HistoryConfig &hc) {
    stream_.open(hc.file_, std::ios_base::app);
    if (!stream_) {
        const generic_error ge{"cannot open monitoring history " +
                               hc.file_.string() + " for writing"};
        Alert(logger_) << ge;
    }
}

void MonitoringHistory::close() { stream_.close(); }

void MonitoringHistory::log(const std::string &message) {
    if (stream_.is_open()) {
        stream_ << "["
                << std::chrono::system_clock::to_time_t(
                       std::chrono::system_clock::now())
                << "] " << message << "\n"
                << std::flush;
    }
    after_publish_();
}

bool MonitoringHistory::needLogfileRotation(const HistoryConfig &hc) const {
    return needLogfileRotationTime(hc) || needLogfileRotationSize(hc);
}

bool MonitoringHistory::needLogfileRotationTime(const HistoryConfig &hc) const {
    auto last{to_tm(last_rotation_)};
    auto now{to_tm(std::chrono::system_clock::now())};
    auto should_rotate{false};
    switch (hc.rotation_interval_) {
        case LogRotationInterval::minute: {
            should_rotate = last.tm_min != now.tm_min;
            break;
        }
        case LogRotationInterval::hour: {
            should_rotate = last.tm_hour != now.tm_hour;
            break;
        }
        case LogRotationInterval::day: {
            should_rotate = last.tm_yday != now.tm_yday;
            break;
        }
        case LogRotationInterval::week: {
            should_rotate = last.tm_wday != now.tm_wday && now.tm_wday == 0;
            break;
        }
        case LogRotationInterval::never:
            break;
    }
    if (should_rotate) {
        Notice(logger_) << "last monitoring history rotation at "
                        << FormattedTimePoint(last_rotation_)
                        << " outside rotation interval, rotating";
    }
    return should_rotate;
}

bool MonitoringHistory::needLogfileRotationSize(const HistoryConfig &hc) const {
    std::error_code ec;
    auto size = std::filesystem::file_size(hc.file_, ec);
    if (ec) {
        const generic_error ge{"cannot determine size of " + hc.file_.string()};
        Warning(logger_) << ge;
        size = 0;
    }
    auto should_rotate{size > hc.rotation_size_limit_};
    if (should_rotate) {
        Notice(logger_) << "monitoring history size " << size
                        << " exceeds limit of " << hc.rotation_size_limit_
                        << ", rotating";
    }
    return should_rotate;
}
