// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/MonitoringLog.h"

#include <cstddef>
#include <map>
#include <sstream>
#include <tuple>
#include <utility>

#include "cmc/NotifyHelper.h"
#include "cmc/Object.h"
#include "cmc/State.h"
#include "livestatus/StringUtils.h"

// Keep in sync with values in notify.py
constexpr size_t max_comment_length = 2000;
constexpr size_t max_plugin_output_length = 1000;

// See also cmk.utils.notify.
const std::string SEMICOLON = "%3B";

std::string MonitoringLog::decode(const std::string &str) {
    return mk::replace_all(str, "\n", R"(\n)");
}

namespace {
struct OptionalLongOutput {
    explicit OptionalLongOutput(const Object *object) : _object(object) {}
    const Object *_object;
};

std::ostream &operator<<(std::ostream &os, const OptionalLongOutput &o) {
    return o._object != nullptr && o._object->_write_long_output
               ? os << ";"
                    << mk::replace_all(
                           MonitoringLog::decode(
                               o._object->state()
                                   ->check_result_.long_output_.substr(
                                       0, o._object->maxLongOutputSize())),
                           ";", SEMICOLON)
               : os;
}
}  // namespace

MonitoringLog::MonitoringLog(std::function<void(const std::string &)> sink)
    : sink_{std::move(sink)} {}

void MonitoringLog::acknowledgeAlert(const Object &object,
                                     const std::string &state_type,
                                     const std::string &author,
                                     const std::string &comment) {
    std::ostringstream os;
    os << (object.isHost() ? "HOST ACKNOWLEDGE ALERT"
                           : "SERVICE ACKNOWLEDGE ALERT")  //
       << ": "                                             //
       << object.specname()                                //
       << ";"                                              //
       << state_type                                       //
       << ";"                                              //
       << author                                           //
       << ";"                                              //
       << mk::replace_all(comment.substr(0, max_comment_length), ";",
                          SEMICOLON);
    sink_(os.str());
}

void MonitoringLog::alert(const Object &object) {
    const auto *state = object.state();
    std::ostringstream os;
    os << (object.isHost() ? "HOST ALERT" : "SERVICE ALERT")  //
       << ": "                                                //
       << object.specname()                                   //
       << ";"                                                 //
       << object.statename()                                  //
       << ";"                                                 //
       << object.stateTypeName()                              //
       << ";"                                                 //
       << state->current_attempt_                             //
       << ";"                                                 //
       << mk::replace_all(
              state->check_result_.output_.substr(0, max_plugin_output_length),
              ";", SEMICOLON)  //
       << OptionalLongOutput(&object);
    sink_(os.str());
}

void MonitoringLog::downtimeAlert(const Object &object,
                                  const std::string &state_type,
                                  const std::string &comment) {
    // Note: Nagios just logs 'Host has entered a period of scheduled downtime'
    // here. That is not very useful. We rather log the downtime comment.
    std::ostringstream os;
    os << (object.isHost() ? "HOST DOWNTIME ALERT"
                           : "SERVICE DOWNTIME ALERT")  //
       << ": "                                          //
       << object.specname()                             //
       << ";"                                           //
       << state_type                                    //
       << ";"                                           //
       << mk::replace_all(comment.substr(0, max_comment_length), ";",
                          SEMICOLON);
    sink_(os.str());
}

void MonitoringLog::flappingAlert(const Object &object) {
    const bool is_flapping = object.state()->flappiness_.is_flapping_;
    std::ostringstream os;
    os << (object.isHost() ? "HOST FLAPPING ALERT"
                           : "SERVICE FLAPPING ALERT")  //
       << ": "                                          //
       << object.specname()                             //
       << ";"                                           //
       << (is_flapping ? "STARTED" : "STOPPED")         //
       << ";"                                           //
       << (is_flapping ? "Started flapping" : "Stopped flapping");
    sink_(os.str());
}

void MonitoringLog::state(const Object &object, LogStateType log_state_type) {
    const bool is_host = object.isHost();
    const auto *state = object.state();
    const bool checked = state->has_been_checked_;
    std::ostringstream os;
    os << (log_state_type == LogStateType::initial
               ? (is_host ? "INITIAL HOST STATE"  //
                          : "INITIAL SERVICE STATE")
               : (is_host ? "CURRENT HOST STATE"       //
                          : "CURRENT SERVICE STATE"))  //
       << ": "                                         //
       << object.specname()                            //
       << ";"                                          //
       << (checked   ? object.statename()              //
           : is_host ? "UP"                            //
                     : "OK")                           //
       << ";"                                          //
       << (checked ? object.stateTypeName() : "HARD")  //
       << ";"                                          //
       << (checked ? state->current_attempt_ : 1)      //
       << ";"                                          //
       << (checked ? mk::replace_all(state->check_result_.output_.substr(
                                         0, max_plugin_output_length),
                                     ";", SEMICOLON)
                   : "")  //
       << OptionalLongOutput(&object);
    sink_(os.str());
}

void MonitoringLog::vanished(const Object &object) {
    std::ostringstream os;
    os << (object.isHost() ? "VANISHED HOST" : "VANISHED SERVICE")  //
       << ": "                                                      //
       << object.specname();
    sink_(os.str());
}

void MonitoringLog::timeperiodTransition(const std::string &name, int from,
                                         int to) {
    std::ostringstream os;
    os << "TIMEPERIOD TRANSITION"  //
       << ": "                     //
       << name                     //
       << ";"                      //
       << from                     //
       << ";"                      //
       << to;
    sink_(os.str());
}

void MonitoringLog::coreStarting() {
    sink_("Check_MK Micro Core " VERSION " starting...");
}

void MonitoringLog::coreStopping() {
    sink_("Check_MK Micro Core shutting down...");
}

void MonitoringLog::logVersion() { sink_("LOG VERSION: 2.0"); }

void MonitoringLog::logInitialStates() { sink_("logging initial states"); }

void MonitoringLog::info(const std::string &line) { sink_(line); }

void MonitoringLog::extCommand(const std::string &command) {
    std::ostringstream os;
    os << "EXTERNAL COMMAND"  //
       << ": "                //
       << command;
    sink_(os.str());
}

namespace {
std::tuple<std::string, std::string, const Object *> ParseInfo(
    const NotificationInfo &info) {
    if (info._type.starts_with("ALERTHANDLER")) {
        return {info._type,                                             // state
                info._extra_macros.find("ALERTHANDLEROUTPUT")->second,  // outp
                nullptr};                                               // obj
    }

    auto state = (info._type == "PROBLEM" || info._type == "RECOVERY")
                     ? info._object.statename()
                     : info._type + " (" + info._object.statename() + ")";

    return {state,                                        // state
            info._object.state()->check_result_.output_,  // outp
            &info._object};                               // obj
}
}  // namespace

void MonitoringLog::notification(const std::string &contact_name,
                                 const NotificationInfo &info) {
    auto [state, output, object_for_long_output] = ParseInfo(info);
    constexpr auto command_name = "check-mk-notify";
    std::ostringstream os;
    os << (info._object.isHost() ? "HOST NOTIFICATION"
                                 : "SERVICE NOTIFICATION")  //
       << ": "                                              //
       << contact_name                                      //
       << ";"                                               //
       << info._object.specname()                           //
       << ";"                                               //
       << state                                             //
       << ";"                                               //
       << command_name                                      //
       << ";"                                               //
       << mk::replace_all(output.substr(0, max_plugin_output_length), ";",
                          SEMICOLON)  //
       << ";"                         //
       << info._author                //
       << ";"                         //
       << mk::replace_all(info._comment.substr(0, max_comment_length), ";",
                          SEMICOLON)
       << OptionalLongOutput(object_for_long_output);
    sink_(os.str());
}

void MonitoringLog::alertHandlerStarted(const Object &object,
                                        const std::string &handler_name) {
    std::ostringstream os;
    os << (object.isHost() ? "HOST ALERT HANDLER STARTED"
                           : "SERVICE ALERT HANDLER STARTED")
       << ": "               //
       << object.specname()  //
       << ";"                //
       << handler_name;
    sink_(os.str());
}

void MonitoringLog::alertHandlerStopped(const Object &object,
                                        const std::string &handler_name,
                                        const std::string &exit_code,
                                        const std::string &output) {
    std::ostringstream os;
    os << (object.isHost() ? "HOST ALERT HANDLER STOPPED"
                           : "SERVICE ALERT HANDLER STOPPED")
       << ": "               //
       << object.specname()  //
       << ";"                //
       << handler_name       //
       << ";"                //
       << exit_code          //
       << ";"                //
       << mk::replace_all(output.substr(0, max_plugin_output_length), ";",
                          SEMICOLON);
    sink_(os.str());
}
