// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/NotifyHelper.h"

#include <algorithm>
#include <cerrno>
#include <chrono>
#include <filesystem>
#include <iterator>
#include <memory>
#include <optional>
#include <sstream>
#include <stdexcept>
#include <unordered_map>

#include "cmc/Contact.h"
#include "cmc/ContactList.h"
#include "cmc/EventSpooler.h"
#include "cmc/GlobalConfig.h"
#include "cmc/Host.h"
#include "cmc/Macro.h"
#include "cmc/MonitoringLog.h"
#include "cmc/Object.h"
#include "cmc/Process.h"
#include "cmc/World.h"
#include "livestatus/Logger.h"
#include "livestatus/POSIXUtils.h"
#include "livestatus/Poller.h"

using namespace std::chrono_literals;
using namespace std::string_literals;

NotifyHelper::NotifyHelper(std::string name, Logger *logger,
                           CommentsFinder comments_finder,
                           MonitoringLog &monitoring_log, EventSpooler &spooler)
    : ProcessHelper{std::move(name), "", {"--notify"s, "--keepalive"s}, logger}
    , _comments_finder{std::move(comments_finder)}
    , monitoring_log_{monitoring_log}
    , _stdout_consumer{LineProducer{[this](auto response) {
        if (response != "*") {
            throw generic_error{
                EINVAL, "invalid response '" + std::string{response} + "'"};
        }
        _helper_ready = true;
        tryToSendEvent();
    }}}
    , _helper_ready{false}
    , spooler_{spooler} {}

NotifyHelper::~NotifyHelper() { do_stop(); }

void NotifyHelper::start_() { spawnConnectedSubprocess(); }

void NotifyHelper::stop_() { do_stop(); }

void NotifyHelper::setUpPoller_(Poller &poller) const {
    if (stdOut() >= 0) {
        poller.addFileDescriptor(stdOut(), PollEvents::in);
    }
}

void NotifyHelper::communicate_(const Poller &poller) {
    if (stdOut() >= 0 && poller.isFileDescriptorSet(stdOut(), PollEvents::in)) {
        handleStdOut();
    }
}

void NotifyHelper::handleStdOut() {
    try {
        _stdout_consumer(stdOut());
    } catch (const std::runtime_error &e) {
        Error(logger()) << e.what();
        restart();
    }
}

bool NotifyHelper::updateConfig(const World *world) {
    auto changed = false;
    std::unordered_set<const Contact *> contacts;
    contacts.reserve(world->config_.contacts_.size());
    for (const auto &entry : world->config_.contacts_) {
        contacts.insert(entry.second.get());
    }
    if (contacts != _contacts) {
        _contacts = contacts;
        changed = true;
    }
    if (updatePath(world->config_.global_config_.general_.cmk_executable_)) {
        changed = true;
    }
    return changed;
}

void NotifyHelper::do_stop() {
    _helper_ready = false;
    despawn();
}
// type can be one of:
// "ACKNOWLEDGEMENT (statename)"
// "FLAPPINGSTART"
// "FLAPPINGSTOP"
// "DOWNTIMESTART"
// "DOWNTIMEEND"
// "DOWNTIMECANCELLED"
// "CUSTOM"
// "RECOVERY"
// "PROBLEM"
// "ALERTHANDLER (exitcode)"

// Find out contacts to notify. Note: the plain list of contacts is not stored
// in the object. This is due to excessive memory consumption. Guess we have
// 100.000k services and 1000 users. That would make up to 100Mio user/service
// relations. In an 64 Bit systems with 3-5 pointers per list entry this could
// sum up to 3.8GB of data!

void NotifyHelper::notify(const NotificationInfo &info) {
    Debug(logger()) << info._object << ": sending " << info._type
                    << " notification to its contacts";
    do_notify(info._object._contact_list->contactNames(), info);
}

namespace {
std::vector<std::string> contactNames(
    const std::unordered_set<const Contact *> &value) {
    std::vector<std::string> names;
    std::transform(value.begin(), value.end(), std::back_inserter(names),
                   [](const Contact *ctc) { return ctc->name(); });
    return names;
}
}  // namespace

void NotifyHelper::broadcast(const NotificationInfo &info) {
    Debug(logger()) << info._object << ": sending " << info._type
                    << " notification to all contacts";
    do_notify(contactNames(_contacts), info);
}

void NotifyHelper::do_notify(const std::vector<std::string> &contact_names,
                             const NotificationInfo &info) {
    monitoring_log_.notification("check-mk-notify", info);
    std::ostringstream os;
    os << Macro("CONTACTS", contact_names);
    addMacros(os, info);
    Debug(logger()) << info._object
                    << ": spooling notification to rule based notifications";
    os << "\n";
    spooler_.spool(os.str());
}

namespace {
class AcknowledgeMacros {
public:
    AcknowledgeMacros(const Object &object, CommentsFinder comments_finder)
        : object_{&object}, comments_finder_{std::move(comments_finder)} {}

    friend std::ostream &operator<<(std::ostream &os,
                                    const AcknowledgeMacros &am) {
        auto [hauthor, hcomment] = am.comments_finder_(*am.object_->host());
        os << Macro("HOSTACKAUTHOR", hauthor)
           << Macro("HOSTACKCOMMENT", hcomment);
        if (am.object_->isService()) {
            auto [sauthor, scomment] = am.comments_finder_(*am.object_);
            return os << Macro("SERVICEACKAUTHOR", sauthor)
                      << Macro("SERVICEACKCOMMENT", scomment);
        }
        return os;
    }

private:
    const Object *object_;
    CommentsFinder comments_finder_;
};
}  // namespace

void NotifyHelper::addMacros(std::ostream &os, const NotificationInfo &info) {
    os << Macro("NOTIFICATIONTYPE", info._type)
       << Macro("NOTIFICATIONCOMMENT", info._comment)
       << Macro("NOTIFICATIONAUTHOR", info._author)
       << Macro("NOTIFICATIONAUTHORNAME", info._author)
       << Macro("NOTIFICATIONAUTHORALIAS", info._author)
       << AcknowledgeMacros(info._object, _comments_finder)
       << ObjectMacros(info._object);
    for (const auto &[key, value] : info._extra_macros) {
        os << Macro(key, value);
    }
}

void NotifyHelper::tryToSendEvent() {
    if (!_helper_ready || spooler_.empty()) {
        return;
    }
    auto event = spooler_.unspool();
    if (writeWithTimeout(stdIn(), event, 3s) == -1) {
        auto saved_errno = errno;
        spooler_.spool(event);      // write back!
        if (auto &p = process()) {  // TODO(sp) Why do we do this *here*?
            p->interrupt();
        }
        throw generic_error{saved_errno, "cannot send notification data"};
    }
    _helper_ready = false;
    Debug(logger()) << "successfully sent event";
}
