// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/Object.h"

#include <algorithm>
#include <compare>
#include <filesystem>
#include <map>
#include <ratio>
#include <sstream>
#include <system_error>
#include <utility>

#include "cmc/CheckResult.h"
#include "cmc/ContactList.h"
#include "cmc/Core.h"
#include "cmc/DowntimeRule.h"
#include "cmc/GlobalConfig.h"
#include "cmc/GlobalState.h"
#include "cmc/Host.h"
#include "cmc/MonitoringLog.h"
#include "cmc/NotifyHelper.h"
#include "cmc/RRDFileLocator.h"
#include "cmc/RRDInfoCache.h"
#include "cmc/Service.h"
#include "cmc/Timeperiod.h"
#include "cmc/World.h"
#include "cmc/cmc.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/Interface.h"
#include "livestatus/LogEntry.h"
#include "livestatus/Logger.h"
#include "livestatus/RegExp.h"

using namespace std::chrono_literals;
using minutes_d = std::chrono::duration<double, std::ratio<60>>;

Replacer::Replacer(std::shared_ptr<RegExp> regexp, std::string replacement)
    : _regex(std::move(regexp)), _replacement(std::move(replacement)) {}

std::string Replacer::replace(const std::string &str) const {
    return _regex->replace(str, _replacement);
}

double staleness(const Object &object) {
    // TODO(ml): Make that a method to Object with a `now` parameter so
    //           that we can test it.
    const auto *state = object.state();
    std::chrono::system_clock::duration check_result_age;
    const Timeperiod *check_period = object._check_period;
    const std::chrono::system_clock::time_point last_period_change =
        check_period->lastStateChange();
    const std::chrono::system_clock::time_point last_check = state->last_check_;

    // Compute the age of the check result. When the check is currently in its
    // check period then...
    auto m_now = std::chrono::system_clock::now();
    if (check_period->isActive()) {
        // Has a check happened since the beginning of the current phase? Then
        // simply compare last check with current time. This should be the 99%
        // case.
        if (last_check >= last_period_change) {
            check_result_age = m_now - last_check;

        } else {
            // otherwise the active phase has just begun. Take the time since
            // the beginning of the phase.
            check_result_age = m_now - last_period_change;

            // Add time at the end of the pre-last transition
            const std::chrono::system_clock::time_point prelast_period_change =
                check_period->previousLastStateChange();
            if (prelast_period_change !=
                std::chrono::system_clock::time_point()) {
                if (last_check < prelast_period_change) {
                    check_result_age += prelast_period_change - last_check;
                }
                // else: a check happend out of the check period. Ignore this
            }
            // else: no information about past. Ignore this
        }
    } else {
        // Check is currently out of its check period? Then use the beginning of
        // the inactive phase as reference for computing the check age. This
        // effectively freezes the staleness value when a goes goes into its
        // inactive phase.
        if (last_period_change != std::chrono::system_clock::time_point()) {
            check_result_age = last_period_change - last_check;
        } else {
            // e.g. for timeperiod "never"
            check_result_age = 0s;
        }
    }

    // Is the checks' result based on cached agent data? Then use the age of
    // that data as check result age
    std::chrono::duration<double> interval{object.staleness_check_interval()};
    if (state->check_result_.cached_at_ !=
        std::chrono::system_clock::time_point()) {
        // Cache interval and check interval can add up in the worst case.
        interval += state->check_result_.cache_interval_;
        const std::chrono::system_clock::duration cached_age =
            m_now - state->check_result_.cached_at_;
        if (cached_age > check_result_age) {
            check_result_age = cached_age;
        }
    }

    // Check_MK configures the interval for its passive checks correctly. Just
    // make sure that we do not fail if it is set to 0 by some error.
    return mk::ticks<std::chrono::seconds>(check_result_age) /
           (interval == 0s
                ? 1
                : mk::ticks<std::chrono::duration<double>>(interval));
}

// Used for creation of temporary fake hosts for PROCESS_HOST_PERFDATA and
// temporary fake services for PROCESS_SVC_PERFDATA.
Object::Object(Core &core, World &world, const std::string &name,
               const Timeperiod &tp)
    : Entity{name}
    , _is_shadow(false)
    , _write_long_output(false)
    , _check_type(0)
    , _max_check_attempts(0)
    , _check_period(&tp)
    , _service_period(&tp)
    , _notification_period(&tp)
    , _notification_options(static_cast<NotificationOptions>(0))
    , _check_timeout(0s)
    , _freshness_state(0)
    , _low_flap_threshold(0)
    , _high_flap_threshold(0)
    , _flap_decay(0)
    , _rrd_format(RRDFormat::single)
    , _contact_list(nullptr)
    , _state(nullptr)
    , _enabling(0)
    , core_{core}
    , world_{world} {}

Object::Object(Core &core, World &world, Entity &&entity)
    : Entity{std::move(entity)}, core_{core}, world_{world} {}

Object::~Object() {
    if (auto *state = stateNonConst()) {
        if (state->_shared != 0U) {
            state->_shared--;
        } else {
            delete state;
        }
    }
}

void Object::addOutputMappings(const Replacer &replacer) {
    _output_mappings.emplace_back(replacer);
}

uint32_t Object::maxLongOutputSize() const {
    return world()->config_.global_config_.general_.max_long_output_size_;
}

std::string Object::specname() const {
    return isHost() ? name() : (host()->name() + ";" + name());
}

std::string Object::statenameOf(int stateIndex) const {
    static std::vector<std::string> host_states{
        "UP",
        "DOWN",
        "UNREACHABLE",
    };
    static std::vector<std::string> service_states{
        "OK",
        "WARNING",
        "CRITICAL",
        "UNKNOWN",
    };
    return state()->has_been_checked_ ? (isHost() ? host_states[stateIndex]
                                                  : service_states[stateIndex])
                                      : "PENDING";
}

std::string Object::lastStatename() const {
    return statenameOf(state()->last_state_);
}

std::string Object::previousHardStatename() const {
    return statenameOf(state()->hard_state_.previous_);
}

std::string Object::statename() const {
    return statenameOf(state()->current_state_);
}

const char *Object::stateTypeName() const {
    return state()->state_type_ == StateType::hard ? "HARD" : "SOFT";
}

void Object::createNewState(std::chrono::system_clock::time_point now,
                            Logger *logger) {
    setState(new State{
        .acknowledgement_type_ = AcknowledgementType::none,
        .check_result_ =
            {
                .output_ = {},
                .long_output_ = {},
                .perf_data_ = {},
                .cached_at_ = {},
                .cache_interval_ = 0s,
            },
        .downtime_ = {.pending_flex_ = 0, .scheduled_depth_ = 0},
        .enabling_ = {.value_ = 0, .modified_ = 0},
        .flappiness_ =
            {
                .is_flapping_ = false,
                .value_ = 0.0,
            },
        .hard_state_ =
            {.last_ = 0, .previous_ = 0, .changed_ = false, .last_time_{}},
        .notification_ = {.current_number_ = 0,
                          .last_time_ = {},
                          .next_time_ = {},
                          .pending_kind_ = NotificationKind::none,
                          .problem_id_ = 0,
                          .problem_sent_ = false,
                          .postponement_reason_ = {}},
        .current_attempt_ = 1,
        .has_been_checked_ = false,
        .execution_time_ = std::chrono::duration<double>{0},
        .latency_ = std::chrono::duration<double>{0},
        .last_check_ = now,
        .last_check_started_ = {},
        .last_state_change_ = {},
        .last_time_state_ = {},
        .next_check_ = {},
        .current_state_ = 0,
        .initial_state_ = 0,
        .last_state_ = 0,
        .state_type_ = StateType::hard});
    if (canBeScheduled()) {
        setNextCheck(core_.initialScheduleTime(now, this), logger);
    }
    core_.monitoring_log_.state(*this, LogStateType::initial);
}

// Make sure that the state is in a sane state after a configuration change.
void Object::sanitizeState(Logger *logger) {
    // Fix the current attempt. Otherwise current hard problems do not show the
    // correct number of attempts after a change of max_check_attempts.
    if ((state()->state_type_ == StateType::hard && !isCurrentStateOK()) ||
        (state()->current_attempt_ > _max_check_attempts)) {
        stateNonConst()->current_attempt_ = _max_check_attempts;
    }

    // Make sure that if the check interval has been *reduced* that the next
    // scheduled check reflects the new and shorter period.
    if (canBeScheduled()) {
        auto interval_min = state()->state_type_ == StateType::hard
                                ? _check_interval
                                : _retry_interval;
        // allow for 10 exceeding secs
        auto max_next = std::chrono::system_clock::now() + interval_min + 10s;

        // If due to a configuration change the next scheduled check is in an
        // active timeperiod but beyond the normal interval and it could be
        // scheduled earlier, then reschedule it. This is helpful if e.g. you
        // set a timeperiod "nine_to_five" for a service and do a reload at -
        // say 00:00. The next check will then be scheduled for 9:00. But if you
        // then change that back to "24X7", the schedule would stay at 9:00 -
        // while the user expects it to be at lastest 00:00 plus one check
        // period!
        if (nextCheck() > max_next) {
            Debug(logger) << "next check of " << *this
                          << " would be too late by "
                          << mk::ticks<std::chrono::duration<double>>(
                                 nextCheck() - max_next)
                          << "s, rescheduling it";
            setNextCheck(nextImmediateCheckTime(), logger);
        }

        // If the check_period of a check has changed and it's next check now
        // lies without the period, we do a reschedule.
        else if (!_check_period->isActiveTime(nextCheck())) {
            Debug(logger) << "next check of " << *this
                          << " would be out of check period, rescheduling it";
            setNextCheck(nextImmediateCheckTime(), logger);
        }
    }
}

bool Object::isEnabled(State::Enabling::Key what) const {
    bool enabled = false;
    return state()->enabling_.isModified(what, enabled)
               ? enabled
               : getEnabledPerConfig(what);
}

void Object::setEnabled(State::Enabling::Key what, bool enabled) const {
    stateNonConst()->enabling_.set(what, enabled, getEnabledPerConfig(what));
}

void Object::resetEnablings() const { stateNonConst()->enabling_.clear(); }

bool Object::isCheckEnabled(CheckResultKind kind) const {
    return isEnabled(kind == CheckResultKind::active
                         ? State::Enabling::Key::active_checks
                         : State::Enabling::Key::passive_checks) &&
           (isHost() ? core_.globalState()->getExecuteHostChecks()
                     : core_.globalState()->getExecuteServiceChecks());
}

std::chrono::system_clock::time_point Object::nextCheckTime() const {
    // If the check is currently in a soft state it is retrying.  We use an
    // alternative interval in that case
    auto interval = state()->state_type_ == StateType::hard ? _check_interval
                                                            : _retry_interval;
    auto next = std::chrono::system_clock::now() +
                std::chrono::duration_cast<std::chrono::system_clock::duration>(
                    interval);
    // If the check period is currently off, then postpone check until beginning
    // of next one time. If there is no next active time then next_check will be
    // set to 0: meaning: do not schedule.
    return _check_period->nextActiveTime(next);
}

std::chrono::system_clock::time_point Object::nextImmediateCheckTime() const {
    // If the check period is currently off, then postpone check until beginning
    // of next one time. If there is no next active time then next_check will be
    // set to 0: meaning: do not schedule.
    return _check_period->nextActiveTime(std::chrono::system_clock::now());
}

void Object::setNextCheck(std::chrono::system_clock::time_point t,
                          Logger *logger) const {
    if (t == std::chrono::system_clock::time_point{}) {
        Informational(logger) << *this << " is passive";
    } else if (nextCheck() == std::chrono::system_clock::time_point{}) {
        Informational(logger)
            << *this << " is active at " << FormattedTimePoint(t);
    }
    stateNonConst()->next_check_ = t;
}

void Object::notifyHardStateChange(std::chrono::system_clock::time_point now,
                                   Logger *logger) {
    if (isShadow()) {
        Debug(logger) << *this
                      << ": suppressing notification for shadow object";
        return;
    }
    if (state()->flappiness_.is_flapping_ &&
        isEnabled(State::Enabling::Key::flap_detection) &&
        core_.globalState()->getEnableFlapDetection()) {
        Debug(logger) << *this
                      << ": suppressing notification because of flapping";
        return;
    }
    if (state()->notification_.pending_kind_ == NotificationKind::none) {
        // Create a new notification, but choose a new problem ID only, if the
        // state transition was from OK to a non-OK state.
        const int problem_id = isPreviousHardStateOK() && !isCurrentStateOK()
                                   ? core_.nextProblemId()
                                   : state()->notification_.problem_id_;
        setupNotification(problem_id, NotificationKind::normal, now, logger);
        return;
    }
    Debug(logger) << *this << ": notification number "
                  << state()->notification_.current_number_
                  << " is already pending";
    if (isCurrentStateOK()) {
        if (state()->notification_.problem_sent_) {
            Debug(logger)
                << *this
                << ": recovery for an already sent notification, cancelling delay";
            stateNonConst()->notification_.next_time_ = now;
        } else {
            cancelNotification("state changed back to OK", logger);
        }
        return;
    }
    if (state()->notification_.current_number_ > 0) {
        // When periodic notifications are in use and we are still within the
        // first period (due to a notification delay), then the delay must not
        // be restarted! Otherwise we would hanlde this as a new notifications
        // with a new delay.
        Debug(logger) << *this
                      << ": state change after first periodic notification";
        // Note: first_notification_delay will *not* be applied.
        stateNonConst()->notification_.next_time_ = now;
    }
}

// Mark the object (i.e. its state) that it should be notified, regardless of
// its notification settings. These are checked later in checkNotification().
// Doing it that way ensures that if the notification settings change while a
// notification is pending, those changed settings will be honored.
void Object::setupNotification(uint32_t problem_id, NotificationKind kind,
                               std::chrono::system_clock::time_point now,
                               Logger *logger) {
    stateNonConst()->notification_.pending_kind_ = kind;

    // Honor first notification delay for non-OK notifications
    auto delay = (  // start flapping notification
                     (kind == NotificationKind::flapping &&
                      state()->flappiness_.is_flapping_) ||
                     // problem notification
                     (kind == NotificationKind::normal && !isCurrentStateOK()))
                     ? _first_notification_delay
                     : 0s;  // no delay for recoveries
    stateNonConst()->notification_.next_time_ = now + delay;
    stateNonConst()->notification_.current_number_ = 0;
    stateNonConst()->notification_.problem_id_ = problem_id;
    Debug(logger) << *this << ": setting up, problem ID " << problem_id
                  << ", delay " << mk::ticks<std::chrono::seconds>(delay)
                  << "sec";
    checkPendingNotification(now, logger);
}

void Object::cancelNotification(const std::string &message,
                                Logger *logger) const {
    Debug(logger) << *this << ": cancelled: " << message;

    stateNonConst()->notification_.pending_kind_ = NotificationKind::none;
    stateNonConst()->notification_.current_number_ = 0;
    stateNonConst()->notification_.postponement_reason_.clear();
    // NOTE: We do *not* reset the problem id to 0 here. We still need this
    // information when we send a recovery notification. That notification needs
    // to bear the same problem id is the initial notification. This is
    // necessary for attaching ticket systems and the like. They need to match
    // a notification and a later recovery.
}

// The CMC has no extra field for the freshness check interval. We simply use
// the normal check interval. That has the additional advantage that the
// estimated number of checks per time can be is more precise.
void Object::checkFreshness(std::chrono::system_clock::time_point now,
                            const result_handler_t &result_handler,
                            Logger *logger) {
    if (!isEnabled(
            State::Enabling::Key::freshness_checks) ||  // currently disabled?
        (_freshness_output[0] == '\0') ||  // Huh? Contradicting config?
        !_check_period->isActive()) {
        return;
    }

    auto age = now - state()->last_check_;
    if (age <= _check_interval) {
        return;
    }

    Notice(logger) << "(" << *this << " not fresh after "
                   << mk::ticks<std::chrono::seconds>(age)
                   << " seconds have elapsed (limit is "
                   << mk::ticks<std::chrono::seconds>(_check_interval)
                   << " secs). Setting to " << _freshness_state  //
                   << R"( ")" << _freshness_output << R"("))";   // )
    std::stringstream ss;
    ss << host()->name() << "\t" << _freshness_output;
    result_handler(now, CheckResult::createActive(ss, this, _freshness_state));
}

// Check, if a (normal) notification should be sent out for this object. There
// are three possible outcomes:
// 1. Do not send the notification *yet*
// 2. Send out the notification
// 3. Cancel the notification
void Object::checkPendingNotification(std::chrono::system_clock::time_point now,
                                      Logger *logger) {
    if (state()->notification_.pending_kind_ == NotificationKind::none) {
        return;  // nothing pending, nothing to do
    }

    auto postpone = [logger, this](const auto &why) {
        return postponeNotification(why, logger);
    };
    if (now < state()->notification_.next_time_) {
        return postpone(state()->notification_.current_number_ == 0
                            ? "delayed notification"
                            : "periodic notification");
    }

    // Currently in a downtime? Do *not* cancel the notification, postpone it.
    // That way the problem will immediately be alerted when the downtime stops
    // while still being non-OK.
    if (inDowntime()) {
        return postpone("currently in downtime");
    }

    // Same if a service's host in in downtime
    if (isService() && host()->inDowntime()) {
        return postpone("host of this service is currently in downtime");
    }

    auto cancel = [logger, this](const auto &why) {
        return cancelNotification(why, logger);
    };
    if (acknowledged()) {
        if (periodicNotificationsEnabled()) {
            return postpone(
                "problem acknowledged and periodic notifications are enabled");
        }
        return cancel("problem is acknowledged");
    }

    // Are the notifications enabled for this object?
    if (!notificationsEnabled(logger)) {
        if (periodicNotificationsEnabled()) {
            return postpone(
                "notifications are disabled, but periodic notifications are enabled");
        }
        return cancel("no periodic notifications are enabled");
    }

    if (!inNotificationPeriod()) {
        return postpone("not in notification period");
    }

    // Now we need to take into account the dependencies of services from their
    // host and of hosts from their parents. Note: this must *not* be done for
    // recovery notifications.

    if (state()->notification_.pending_kind_ == NotificationKind::flapping ||
        !isCurrentStateOK()) {
        Host *hst = host();

        // Services depend on the state of their host
        if (isService()) {
            // Case 1: the host is DOWN/UNREACH If the host is down, do not send
            // a notification now.  But do not cancel it. If the host comes up
            // again and the problem still persists, send it out anyway!
            if (hst->getCurrentState() != HostState::up) {
                return postpone("host of this service is not up");
            }

            // Case 2: the host is UP

            // Case 2a) This host is still shown as up, but might be in fact
            // down. This situation arises when the service check is the first
            // one to be executed after the host goes down. We detect this
            // situation by comparing the times of the last check of the host
            // and of the service. For hosts that are assumed to be up always,
            // we skip this check.
            if (hst->state()->last_check_ < state()->last_check_ &&
                hst->_check_command_line != "@up") {
                return postpone("last host check not recent enough");
            }

            // Case 2b) The host was down and just came up again. The service
            // has a pending notification since it is critical. If it was just
            // critical because of the host being down we do not want to send a
            // notification. So we wait for one further service check before
            // notifying. So here we compare the time of the hosts last state
            // change and the last check time of the service.
            if (hst->state()->last_state_change_ > state()->last_check_) {
                return postpone("last service check not recent enough");
            }

        } else if (hst->getCurrentState() == HostState::unreachable &&
                   !is_empty_bit_mask(hst->_notification_options &
                                      NotificationOptions::host_unreachable)) {
            Debug(logger) << *this
                          << ": unreachable, but we should send it anyway";

        } else if (hst->hasParents()) {
            bool one_recent_is_up = false;
            bool one_is_up = false;
            for (auto *parent : hst->_parents) {
                if (parent->getCurrentState() == HostState::up) {
                    one_is_up = true;
                    // Again: The date of the host check might not be recent
                    // enough. In that case we need to postpone our decision
                    if (parent->state()->last_check_ > state()->last_check_) {
                        one_recent_is_up = true;
                        break;
                    }
                }
            }
            if (!one_is_up) {
                return postpone("all parents are down");
            }
            if (!one_recent_is_up) {
                return postpone(
                    "at least one parent is up, but no check is recent enough");
            }
        }
    }

    if (isCurrentStateOK() &&
        state()->notification_.pending_kind_ != NotificationKind::flapping &&
        !state()->notification_.problem_sent_) {
        return cancel("no problem notification previously sent");
    }

    // Now finally really send out a notification.

    // Increase the current notification number, which is slightly tricky: We
    // must *not* increase the number in case of a recovery. Otherwise the
    // recovery would escalate if an escalation-like notification configurations
    // is in use: Assume that notification #1 is going to the normal contact and
    // #2...#n is going to an escalated contact. In that case, the recovery of a
    // non-escalated notification must not be 2, but 1. Another special case: We
    // never want to send out a notification with number 0 (e.g. non-periodic OK
    // notification).
    if (!isCurrentStateOK() || state()->notification_.current_number_ == 0) {
        stateNonConst()->notification_.current_number_++;
    }

    const char *type = "SHOULDNOTHAPPEN";
    switch (state()->notification_.pending_kind_) {
        case NotificationKind::none:
            break;
        case NotificationKind::normal:
            type = isCurrentStateOK() ? "RECOVERY" : "PROBLEM";
            break;
        case NotificationKind::flapping:
            type = state()->flappiness_.is_flapping_ ? "FLAPPINGSTART"
                                                     : "FLAPPINGSTOP";
            break;
    }
    core_.notify({*this, type, "", "", {}});
    stateNonConst()->notification_.problem_sent_ = !isCurrentStateOK();

    // Perioding notifications enabled?
    if (periodicNotificationsEnabled() &&
        state()->notification_.pending_kind_ == NotificationKind::normal &&
        !isCurrentStateOK()) {
        stateNonConst()->notification_.next_time_ =
            std::chrono::system_clock::now() + _notification_interval;
        Debug(logger) << *this << ": next notification in "
                      << mk::ticks<minutes_d>(_notification_interval)
                      << " minutes";
    } else {
        cancel("no more notifications necessary");
    }
}

void Object::postponeNotification(const std::string &why,
                                  Logger *logger) const {
    Debug(logger) << *this << ": postponing, " << why;
    stateNonConst()->notification_.postponement_reason_ = why;
}

bool Object::periodicNotificationsEnabled() const {
    return _notification_interval != 0s;
}

bool Object::acknowledged() const {
    return acknowledgementType() != AcknowledgementType::none;
}

AcknowledgementType Object::acknowledgementType() const {
    return state()->acknowledgement_type_;
}

void Object::setAcknowledgementType(AcknowledgementType value) const {
    stateNonConst()->acknowledgement_type_ = value;
}

bool Object::inDowntime() const {
    return state()->downtime_.scheduled_depth_ > 0;
}

bool Object::handled() const { return acknowledged() || inDowntime(); }

bool Object::notificationsEnabled(Logger *logger) const {
    if (!core_.globalState()->getEnableNotifications()) {
        Debug(logger) << *this
                      << ": skipping, notifications are disabled globally";
        return false;
    }
    if (is_empty_bit_mask(_notification_options & notificationBitmask())) {
        Debug(logger) << *this
                      << ": skipping, this type of notification is disabled";
        return false;
    }
    if (!isEnabled(State::Enabling::Key::notifications)) {
        Debug(logger) << *this << ": skipping, notifications are disabled";
        return false;
    }
    return true;
}

bool Object::inNotificationPeriod() const {
    return _notification_period->isActive();
}

NotificationOptions Object::notificationBitmask() const {
    if (isHost()) {
        if (state()->notification_.pending_kind_ ==
            NotificationKind::flapping) {
            return NotificationOptions::host_flapping;
        }
        switch (host()->getCurrentState()) {
            case HostState::up:
                return NotificationOptions::host_up;
            case HostState::down:
                return NotificationOptions::host_down;
            case HostState::unreachable:
                return NotificationOptions::host_unreachable;
        }
    } else {
        if (state()->notification_.pending_kind_ ==
            NotificationKind::flapping) {
            return NotificationOptions::service_flapping;
        }
        switch (service()->getCurrentState()) {
            case ServiceState::ok:
                return NotificationOptions::service_ok;
            case ServiceState::warning:
                return NotificationOptions::service_warning;
            case ServiceState::critical:
                return NotificationOptions::service_critical;
            case ServiceState::unknown:
                return NotificationOptions::service_unknown;
        }
    }
    return NotificationOptions::service_unknown;  // unreachable
}

bool Object::hasContact(const std::string &contact_name) const {
    return _contact_list->isMember(contact_name);
}

void Object::flap(std::chrono::system_clock::time_point now, Logger *logger) {
    if (!isShadow() && stateChanged() &&
        isEnabled(State::Enabling::Key::flap_detection) &&
        core_.globalState()->getEnableFlapDetection() &&
        stateNonConst()->flappiness_.flap(_high_flap_threshold)) {
        core_.flappingEvent(*this);
        Debug(logger) << *this << ": started flapping, preparing notification";
        // Notify about this flapping event. If a notification still pending
        // from a previous state change (due to a notification delay), we will
        // replace that.
        setupNotification(core_.nextProblemId(), NotificationKind::flapping,
                          now, logger);
    }
}

void Object::unflap(std::chrono::system_clock::time_point now, Logger *logger) {
    if (!isShadow() &&
        stateNonConst()->flappiness_.unflap(_low_flap_threshold, _flap_decay)) {
        Debug(logger) << *this << ": stopped flapping";
        core_.flappingEvent(*this);
        if (state()->notification_.pending_kind_ !=
            NotificationKind::flapping) {
            // prepare unflap notification
            setupNotification(core_.nextProblemId(), NotificationKind::flapping,
                              now, logger);
            // restart periodic notifications
            stateNonConst()->notification_.pending_kind_ =
                NotificationKind::normal;
            return;
        }
        // If a flapping notification is still pending, silently cancel it if
        // the state is OK. Convert it into a normal notification otherwise.
        if (state()->hard_state_.last_ == 0) {
            return cancelNotification("stopped flapping while in OK state",
                                      logger);
        }
        cancelNotification("stopped flapping while in non-OK state", logger);
        setupNotification(core_.nextProblemId(), NotificationKind::normal, now,
                          logger);
    }
}

// This method is needed after a config reload. We need to identify persistent
// recurring downtimes that are not changed in the configuration during a
// reload. The argument comes from the state file and the return value is the
// corresponding rule from the config.
DowntimeRule *Object::find_prd(const DowntimeRule &other) const {
    for (const auto &prd : _downtime_rules) {
        if (*prd == other) {
            return prd.get();
        }
    }
    return nullptr;
}

std::string Object::remapOutput(std::string output) {
    for (const auto &replacer : _output_mappings) {
        output = replacer.replace(output);
    }
    return output;
}

void Object::logDowntimes(Logger *logger) const {
    if (logger->isLoggable(LogLevel::informational)) {
        for (const auto &prd : _downtime_rules) {
            Informational(logger) << *this << " has " << *prd;
        }
    }
}

RRDFormat Object::currentRrdFormat() const {
    std::error_code ec;
    if (_rrd_format == RRDFormat::single) {
        if (std::filesystem::exists(
                world()->rrdFileLocatorSingle()->metadata(this), ec)) {
            return RRDFormat::single;
        }
        if (std::filesystem::exists(
                world()->rrdFileLocatorMultiple()->metadata(this), ec)) {
            return RRDFormat::multiple;
        }
    } else {
        if (std::filesystem::exists(
                world()->rrdFileLocatorMultiple()->metadata(this), ec)) {
            return RRDFormat::multiple;
        }
        if (std::filesystem::exists(
                world()->rrdFileLocatorSingle()->metadata(this), ec)) {
            return RRDFormat::single;
        }
    }
    return _rrd_format;
}

RRDInfo Object::rrdInfo() const { return world()->infoCache()->rrdInfo(this); }

MetricLocation Object::metricLocation(const Metric::MangledName &name) const {
    return world()->infoCache()->metricLocation(this, name);
}

void Object::forgetRRDInfo() const {
    world()->infoCache()->forgetRRDInfo(this);
}

std::ostream &operator<<(std::ostream &os, const Object &object) {
    if (object.isHost()) {
        os << R"(host ")" << object.name() << R"(")";
    } else {
        os << R"(service ")" << object.host()->name() << ";" << object.name()
           << R"(")";
    }
    return os;
}

bool Object::operator==(const Object &other) const {
    return Entity::operator==(other) &&
           _check_command_name == other._check_command_name &&
           _check_command_line == other._check_command_line &&
           _is_shadow == other._is_shadow &&
           _write_long_output == other._write_long_output &&
           _check_type == other._check_type &&
           _check_interval == other._check_interval &&
           _retry_interval == other._retry_interval &&
           _notification_interval == other._notification_interval &&
           _first_notification_delay == other._first_notification_delay &&
           _max_check_attempts == other._max_check_attempts &&
           _check_period->name() == other._check_period->name() &&
           _service_period->name() == other._service_period->name() &&
           _notification_period->name() == other._notification_period->name() &&
           _notification_options == other._notification_options &&
           _check_timeout == other._check_timeout &&
           _freshness_state == other._freshness_state &&
           _freshness_output == other._freshness_output &&
           _low_flap_threshold == other._low_flap_threshold &&
           _high_flap_threshold == other._high_flap_threshold &&
           _flap_decay == other._flap_decay &&
           _rrd_format == other._rrd_format &&
           _carbon_send_metric_parts == other._carbon_send_metric_parts &&
           _influxdb_send_config == other._influxdb_send_config &&
           std::is_permutation(
               _downtime_rules.begin(), _downtime_rules.end(),
               other._downtime_rules.begin(), other._downtime_rules.end(),
               [&](const auto &l, const auto &r) { return *l == *r; }) &&
           *_contact_list == *other._contact_list;
}
