// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/PNPXMLWriter.h"

#include <compare>
#include <ratio>
#include <system_error>
#include <utility>
#include <vector>

#include "cmc/GlobalConfig.h"
#include "cmc/Host.h"
#include "cmc/Object.h"
#include "cmc/RRDFileLocator.h"
#include "cmc/State.h"
#include "cmc/World.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/Logger.h"
#include "livestatus/Metric.h"
#include "livestatus/PerformanceData.h"
#include "livestatus/PnpUtils.h"

using namespace std::chrono_literals;

void PNPXMLWriter::WriteTimeCache::clear() {
    const std::lock_guard<std::mutex> lg(_mutex);
    _write_times.clear();
}

std::filesystem::file_time_type PNPXMLWriter::WriteTimeCache::last_write_time(
    const std::filesystem::path &path) {
    const std::lock_guard<std::mutex> lg(_mutex);
    auto it = _write_times.find(path);
    if (it != _write_times.end()) {
        return it->second;
    }
    // NOTE: If something goes wrong, last_write_time() returns the oldest
    // possible point in time, which is what we want.
    std::error_code ec;
    auto write_time = std::filesystem::last_write_time(path, ec);
    if (!ec) {
        _write_times.emplace(path, write_time);
    }
    return write_time;
}

void PNPXMLWriter::WriteTimeCache::forget(const std::filesystem::path &path) {
    const std::lock_guard<std::mutex> lg(_mutex);
    _write_times.erase(path);
}

// A predictable sequence of delays for spreading is totally fine.
// NOLINTNEXTLINE(cert-msc32-c,cert-msc51-cpp)
PNPXMLWriter::PNPXMLWriter(Logger *logger) : _logger(logger) {}

bool PNPXMLWriter::updateConfig(const RRDConfig &rrd) {
    bool changed = false;
    if (_update_delay != rrd.pnp_update_delay_) {
        _update_delay = rrd.pnp_update_delay_;
        changed = true;
    }
    if (_update_on_restart != rrd.pnp_update_on_restart_) {
        _update_on_restart = rrd.pnp_update_on_restart_;
        changed = true;
    }
    if (changed) {
        _write_time_cache.clear();
        _last_world_change = std::filesystem::file_time_type::clock::now();
    }
    return changed;
}

void PNPXMLWriter::start() { _write_time_cache.clear(); }

void PNPXMLWriter::stop() {
    _write_time_cache.clear();
    _update_delay = 0s;
    _update_on_restart = false;
}

void PNPXMLWriter::process(const PerformanceData &perf_data,
                           const Object *object) {
    if (isUpToDate(object)) {
        return;
    }
    std::ofstream os;
    open(os, object);
    if (!os) {
        return;  // in case of disk full or permission error
    }

    // Remove everyting after the ! from the check command in order to supply
    // the correct template information for PNP.
    const auto &name = perf_data.checkCommandName();
    auto pnp_template = name.substr(0, name.find('!'));
    for (const auto &metric : perf_data) {
        add(os, object, metric, pnp_template);
    }
    close(os, perf_data, object);
}

bool PNPXMLWriter::isUpToDate(const Object *object) {
    // NOTE: If the XML file is not there yet, it is considered extremely old by
    // the cache, so we will definitely not consider it up-to-date here. This is
    // exactly what we want: We will write the new XML file immediately, so
    // pnp4nagios can do its job without waiting for any delay.
    auto write_time = _write_time_cache.last_write_time(pathFor(object));
    const bool use_time_from_file =
        !_update_on_restart || write_time > _last_world_change;
    auto now = std::filesystem::file_time_type::clock::now();
    const bool up_to_date = use_time_from_file
                                ? (now <= write_time + _update_delay)
                                : (now <= _last_world_change + spreadDelay());
    if (logger()->isLoggable(LogLevel::debug)) {
        Debug(logger()) << (up_to_date ? "delaying upate of" : "updating")
                        << " PNP XML file for " << *object;
    }
    return up_to_date;
}

// When an update is needed, don't force the update, this could lead to too much
// load e.g. after reloads. Better try to spread it over 10 minutes.
std::chrono::nanoseconds PNPXMLWriter::spreadDelay() {
    using unit = std::chrono::microseconds;
    std::uniform_int_distribution<long> dist(0, mk::ticks<unit>(10min));
    return unit{dist(_random_engine)};
}

void PNPXMLWriter::open(std::ofstream &os, const Object *object) {
    const std::filesystem::path xml_path_new = newPathFor(object);
    try {
        const std::filesystem::path directory = xml_path_new.parent_path();
        Debug(logger()) << "creating directory " << directory
                        << " if necessary";
        std::filesystem::create_directory(directory, directory.parent_path());
    } catch (const std::filesystem::filesystem_error &e) {
        Error(logger()) << "cannot create directory for PNP XML: " << e.what();
        return;
    }

    os.open(xml_path_new);
    if (!os) {
        const generic_error ge("cannot create XML file " +
                               xml_path_new.string());
        Warning(logger()) << ge;
        return;
    }
    os << "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n"
          "<NAGIOS>\n";
}

namespace {
std::string xmlEncode(const std::string &text) {
    std::string encoded;
    for (const char *scan = text.c_str(); *scan != 0; scan++) {
        if (*scan == '&') {
            encoded += "&amp;";
        } else if (*scan == '>') {
            encoded += "&gt;";
        } else if (*scan == '<') {
            encoded += "&lt;";
        } else {
            encoded += *scan;
        }
    }
    return encoded;
}
}  // namespace

// static
void PNPXMLWriter::add(std::ostream &os, const Object *object,
                       const Metric &metric, const std::string &pnp_template) {
    // If the unit is '%' we need to convert it to '%%'. This is only due to
    // some PNP templates and not valid for future other graphing tools.
    std::string unit_pnp = metric.uom();
    if (unit_pnp == "%") {
        unit_pnp = "%%";
    }

    auto varname = metric.mangled_name();
    const std::filesystem::path rrd_file_path =
        object->world()->rrdFileLocatorMultiple()->path(object, varname);

    os << "   <DATASOURCE>\n"                                             //
       << "     <TEMPLATE>" << pnp_template << "</TEMPLATE>\n"            //
       << "     <RRDFILE>" << xmlEncode(rrd_file_path) << "</RRDFILE>\n"  //
       << "     <RRD_STORAGE_TYPE>MULTIPLE</RRD_STORAGE_TYPE>\n"          //
       << "     <RRD_HEARTBEAT>8460</RRD_HEARTBEAT>\n"                    //
       << "     <IS_MULTI>0</IS_MULTI>\n"                                 //
       << "     <DS>1</DS>\n"                                             //
       << "     <NAME>" << varname.string() << "</NAME>\n"                //
       << "     <LABEL>" << xmlEncode(varname.string()) << "</LABEL>\n"   //
       << "     <UNIT>" << xmlEncode(unit_pnp) << "</UNIT>\n"             //
       << "     <ACT>" << metric.value() << "</ACT>\n"                    //
       << "     <WARN>" << metric.warn() << "</WARN>\n"                   //
       << "     <WARN_MIN></WARN_MIN>\n"                                  //
       << "     <WARN_MAX></WARN_MAX>\n"                                  //
       << "     <WARN_RANGE_TYPE></WARN_RANGE_TYPE>\n"                    //
       << "     <CRIT>" << metric.crit() << "</CRIT>\n"                   //
       << "     <CRIT_MIN></CRIT_MIN>\n"                                  //
       << "     <CRIT_MAX></CRIT_MAX>\n"                                  //
       << "     <CRIT_RANGE_TYPE></CRIT_RANGE_TYPE>\n"                    //
       << "     <MIN>" << metric.min() << "</MIN>\n"                      //
       << "     <MAX>" << metric.max() << "</MAX>\n"                      //
       << "   </DATASOURCE>\n";
}

void PNPXMLWriter::close(std::ofstream &os, const PerformanceData &perf_data,
                         const Object *object) {
    const std::filesystem::path xml_path = pathFor(object);
    if (object->isService()) {
        os << "   <RRD>\n"                            //
           << "     <RC>0</RC>\n"                     //
           << "     <TXT>successful updated</TXT>\n"  //
           << "   </RRD>\n"                           //
           << "   <NAGIOS_AUTH_HOSTNAME>" << xmlEncode(object->host()->name())
           << "</NAGIOS_AUTH_HOSTNAME>\n"  //
           << "   <NAGIOS_AUTH_SERVICEDESC>" << xmlEncode(object->name())
           << "</NAGIOS_AUTH_SERVICEDESC>\n"  //
           << "   <NAGIOS_CHECK_COMMAND>"
           << xmlEncode(perf_data.checkCommandName())
           << "</NAGIOS_CHECK_COMMAND>\n"                                //
           << "   <NAGIOS_DATATYPE>SERVICEPERFDATA</NAGIOS_DATATYPE>\n"  //
           << "   <NAGIOS_DISP_HOSTNAME>" << xmlEncode(object->host()->name())
           << "</NAGIOS_DISP_HOSTNAME>\n"  //
           << "   <NAGIOS_DISP_SERVICEDESC>" << xmlEncode(object->name())
           << "</NAGIOS_DISP_SERVICEDESC>\n"  //
           << "   <NAGIOS_HOSTNAME>"
           << xmlEncode(pnp_cleanup(object->host()->name()))
           << "</NAGIOS_HOSTNAME>\n"                              //
           << "   <NAGIOS_MULTI_PARENT></NAGIOS_MULTI_PARENT>\n"  //
           << "   <NAGIOS_PERFDATA>"
           << xmlEncode(object->state()->check_result_.perf_data_)
           << "</NAGIOS_PERFDATA>\n"                    //
           << "   <NAGIOS_RRDFILE></NAGIOS_RRDFILE>\n"  //
           << "   <NAGIOS_SERVICECHECKCOMMAND>"
           << xmlEncode(perf_data.checkCommandName())
           << "</NAGIOS_SERVICECHECKCOMMAND>\n"  //
           << "   <NAGIOS_SERVICEDESC>"
           << xmlEncode(pnp_cleanup(object->name()))
           << "</NAGIOS_SERVICEDESC>\n"  //
           << "   <NAGIOS_SERVICEPERFDATA>"
           << xmlEncode(object->state()->check_result_.perf_data_)
           << "</NAGIOS_SERVICEPERFDATA>\n"  //
           << "   <NAGIOS_SERVICESTATETYPE>"
           << static_cast<int>(object->state()->state_type_)
           << "</NAGIOS_SERVICESTATETYPE>\n";
    } else {
        os << "   <NAGIOS_AUTH_HOSTNAME>" << xmlEncode(object->name())
           << "</NAGIOS_AUTH_HOSTNAME>\n"  //
           << "   <NAGIOS_AUTH_SERVICEDESC>Host Perfdata</NAGIOS_AUTH_SERVICEDESC>\n"  //
           << "   <NAGIOS_CHECK_COMMAND>"
           << xmlEncode(perf_data.checkCommandName())
           << "</NAGIOS_CHECK_COMMAND>\n"                             //
           << "   <NAGIOS_DATATYPE>HOSTPERFDATA</NAGIOS_DATATYPE>\n"  //
           << "   <NAGIOS_DISP_HOSTNAME>" << xmlEncode(object->name())
           << "</NAGIOS_DISP_HOSTNAME>\n"  //
           << "   <NAGIOS_DISP_SERVICEDESC>Host Perfdata</NAGIOS_DISP_SERVICEDESC>\n"  //
           << "   <NAGIOS_HOSTCHECKCOMMAND>"
           << xmlEncode(perf_data.checkCommandName())
           << "</NAGIOS_HOSTCHECKCOMMAND>\n"  //
           << "   <NAGIOS_HOSTNAME>" << xmlEncode(pnp_cleanup(object->name()))
           << "</NAGIOS_HOSTNAME>\n"  //
           << "   <NAGIOS_HOSTPERFDATA>"
           << xmlEncode(object->state()->check_result_.perf_data_)
           << "</NAGIOS_HOSTPERFDATA>\n"  //
           << "   <NAGIOS_HOSTSTATETYPE>"
           << static_cast<int>(object->state()->state_type_)
           << "</NAGIOS_HOSTSTATETYPE>\n"                         //
           << "   <NAGIOS_MULTI_PARENT></NAGIOS_MULTI_PARENT>\n"  //
           << "   <NAGIOS_PERFDATA>"
           << xmlEncode(object->state()->check_result_.perf_data_)
           << "</NAGIOS_PERFDATA>\n"                    //
           << "   <NAGIOS_RRDFILE></NAGIOS_RRDFILE>\n"  //
           << "   <NAGIOS_SERVICEDESC>_HOST_</NAGIOS_SERVICEDESC>\n";
    }
    os << "   <NAGIOS_TIMET>"
       << std::chrono::system_clock::to_time_t(object->state()->last_check_)
       << "</NAGIOS_TIMET>\n"  //
       << "   <NAGIOS_XMLFILE>" << xmlEncode(xml_path.string())
       << "</NAGIOS_XMLFILE>\n"
       << "  <XML>\n"                  //
       << "   <VERSION>4</VERSION>\n"  //
       << "  </XML>\n"                 //
       << "</NAGIOS>\n";
    os.close();

    const std::filesystem::path xml_path_new = newPathFor(object);
    try {
        std::filesystem::rename(xml_path_new, xml_path);
    } catch (const std::filesystem::filesystem_error &e) {
        Error(logger()) << "cannot rename " << xml_path_new << " to "
                        << xml_path << ": " << e.what();
    }
    _write_time_cache.forget(xml_path);
}

// static
std::filesystem::path PNPXMLWriter::pathFor(const Object *object) {
    return object->world()->rrdFileLocatorMultiple()->metadata(object);
}

// static
std::filesystem::path PNPXMLWriter::newPathFor(const Object *object) {
    return pathFor(object) += ".new";
}
