// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/PbConfigLoader.h"

#include <arpa/inet.h>
#include <netinet/in.h>

#include <algorithm>
#include <chrono>
#include <functional>
#include <optional>
#include <stdexcept>
#include <unordered_map>
#include <unordered_set>
#include <vector>

#include "cmc/CarbonConfig.h"
#include "cmc/Contact.h"
#include "cmc/ContactGroup.h"
#include "cmc/ContactList.h"
#include "cmc/CycleTime.h"
#include "cmc/DowntimeRule.h"
#include "cmc/Entity.h"
#include "cmc/GlobalConfig.h"
#include "cmc/Host.h"
#include "cmc/InfluxDBConfig.h"
#include "cmc/Object.h"
#include "cmc/PbCMC.h"
#include "cmc/PbDuration.h"
#include "cmc/PbTimestamp.h"
#include "cmc/RRDInfo.h"
#include "cmc/Service.h"
#include "cmc/State.h"
#include "cmc/Timeperiod.h"
#include "cmc/World.h"
#include "config/v1/types.pb.h"
#include "google/protobuf/timestamp.pb.h"
#include "livestatus/BitMask.h"
#include "livestatus/Interface.h"
#include "livestatus/LogEntry.h"
#include "livestatus/Logger.h"
#include "livestatus/RegExp.h"
#include "livestatus/User.h"

namespace cfg = checkmk::cmc::config::v1;
namespace fs = std::filesystem;
using namespace std::chrono_literals;

namespace {
in_addr_t ToInAddr(const cfg::IPv4Address &address) {
    return ::ntohl(address.number());
}

LogRotationInterval ToLogRotationInterval(
    cfg::LogRotationInterval log_rotation_interval) {
    switch (log_rotation_interval) {
        case cfg::LOG_ROTATION_INTERVAL_MINUTE:
            return LogRotationInterval::minute;
        case cfg::LOG_ROTATION_INTERVAL_HOUR:
            return LogRotationInterval::hour;
        case cfg::LOG_ROTATION_INTERVAL_DAY:
            return LogRotationInterval::day;
        case cfg::LOG_ROTATION_INTERVAL_WEEK:
            return LogRotationInterval::week;
        case cfg::LOG_ROTATION_INTERVAL_NEVER:
            return LogRotationInterval::never;
        default:
            // not possible, but now we have no unified policy what to do
            // with bad config: abort or exception may be better
            return LogRotationInterval::minute;
    }
}

// TODO (sk): not consistent API - estimate possibility makeMangling()
CarbonConnectionConfig::Mangling ToMangling(cfg::CarbonConnectionMangling m) {
    switch (m) {
        case cfg::CARBON_CONNECTION_MANGLING_C_STYLE:
            return CarbonConnectionConfig::Mangling::c_style;
        case cfg::CARBON_CONNECTION_MANGLING_UNDERSCORE:
            return CarbonConnectionConfig::Mangling::underscore;
        default:
            // not possible, but now we have no unified policy what to do
            // with bad config: abort or exception may be better
            return CarbonConnectionConfig::Mangling::c_style;
    }
}
}  // namespace

PbConfigLoader::PbConfigLoader(const cfg::Core &config)
    : string_pool_{config.string_pool()}
    , string_pool_offsets_(config.string_pool_offsets())
    , contact_list_pool_offsets_(config.contact_list_pool_offsets()) {}

std::string PbConfigLoader::makeString(uint32_t ref) const {
    if (ref >= uint32_t(string_pool_.entries_size())) {
        return {};
    }
    return string_pool_.entries(ref);
}

std::string PbConfigLoader::makeString(uint32_t ref,
                                       PoolOffset pool_offset) const {
    if (pool_offset.offset >= uint32_t(string_pool_offsets_.entries_size())) {
        return {};
    }
    return makeString(string_pool_offsets_.entries(pool_offset.offset) + ref);
}

fs::path PbConfigLoader::makePath(const cfg::Path &path) const {
    return makeString(path.str_ref());
}

// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
LogLevel PbConfigLoader::makeLogLevel(cfg::LogLevel log_level) const {
    switch (log_level) {
        case cfg::LOG_LEVEL_EMERGENCY:
            return LogLevel::emergency;
        case cfg::LOG_LEVEL_ALERT:
            return LogLevel::alert;
        case cfg::LOG_LEVEL_CRITICAL:
            return LogLevel::critical;
        case cfg::LOG_LEVEL_ERROR:
            return LogLevel::error;
        case cfg::LOG_LEVEL_WARNING:
            return LogLevel::warning;
        case cfg::LOG_LEVEL_NOTICE:
            return LogLevel::notice;
        case cfg::LOG_LEVEL_INFORMATIONAL:
            return LogLevel::informational;
        case cfg::LOG_LEVEL_DEBUG:
            return LogLevel::debug;
        default:
            // not possible, but now we have no unified policy what to do
            // with bad config: abort or exception may be better
            return LogLevel::emergency;
    }
}

// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
ServiceAuthorization PbConfigLoader::makeServiceAuthorization(
    cfg::ServiceAuthorization sa) const {
    switch (sa) {
        case cfg::SERVICE_AUTHORIZATION_STRICT:
            return ServiceAuthorization::strict;
        case cfg::SERVICE_AUTHORIZATION_LOOSE:
            return ServiceAuthorization::loose;
        default:
            // not possible, but now we have no unified policy what to do
            // with bad config: abort or exception may be better
            return ServiceAuthorization::strict;
    }
}

// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
GroupAuthorization PbConfigLoader::makeGroupAuthorization(
    cfg::GroupAuthorization ga) const {
    switch (ga) {
        case cfg::GROUP_AUTHORIZATION_STRICT:
            return GroupAuthorization::strict;
        case cfg::GROUP_AUTHORIZATION_LOOSE:
            return GroupAuthorization::loose;
        default:
            // not possible, but now we have no unified policy what to do
            // with bad config: abort or exception may be better
            return GroupAuthorization::strict;
    }
}

GlobalConfig PbConfigLoader::makeGlobalConfig(const cfg::Global &global) const {
    GlobalConfig g;
    g.general_ = makeGeneral(global.general());
    g.smart_ping_ = makeSmartPingConfig(global.smart_ping());
    g.checking_ = makeCheckingConfig(global.checking());
    g.livestatus_ = makeLivestatusConfig(global.livestatus());
    g.notification_handler_ =
        makeNotificationHandlerConfig(global.notification_handler());
    g.alert_handler_ = makeAlertHandlerConfig(global.alert_handler());
    g.carbon_ = makeCarbonConfig(global.carbon());
    g.influxdb_ = makeInfluxDbConfig(global.influx_db());
    g.history_ = makeHistoryConfig(global.history());
    g.rrd_ = makeRRDConfig(global.rrd());
    return g;
}

GeneralConfig PbConfigLoader::makeGeneral(const cfg::General &general) const {
    return {
        .serial_ = general.serial(),
        .core_directory_ = makePath(general.core_directory()),
        .log_file_ = makePath(general.log_file()),
        .cmk_executable_ = makePath(general.cmk_executable()),
        .crash_reports_directory_ = makePath(general.crash_reports_directory()),
        .license_usage_history_file_ =
            makePath(general.license_usage_history_file()),
        .inventory_directory_ = makePath(general.inventory_directory()),
        .structured_status_directory_ =
            makePath(general.structured_status_directory()),
        .robotmk_html_log_directory_ =
            makePath(general.robotmk_html_log_directory()),
        .logwatch_directory_ = makePath(general.logwatch_directory()),
        .prediction_directory_ = makePath(general.prediction_directory()),
        .event_console_status_socket_ =
            makePath(general.event_console_status_socket()),
        .state_file_created_file_ = makePath(general.state_file_created_file()),
        .licensed_state_file_ = makePath(general.licensed_state_file()),
        .log_level_core_ = makeLogLevel(general.log_level_core()),
        .log_level_downtime_ = makeLogLevel(general.log_level_downtime()),
        .log_microtime_ = general.log_microtime(),
        .core_dumps_enabled_ = general.core_dumps_enabled(),
        .houskeeping_interval_ =
            mk::ToChronoMicros(general.housekeeping_interval()),
        .state_retention_interval_ =
            mk::ToChronoSeconds(general.state_retention_interval()),
        .event_console_enabled_ = general.event_console_enabled(),
        .edition_ = makeString(general.edition_ref()),
        .max_long_output_size_ = general.max_long_output_size(),
    };
}

// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
AlertHandlerConfig PbConfigLoader::makeAlertHandlerConfig(
    const cfg::AlertHandler &alert_handler) const {
    return {
        .enabled_ = alert_handler.enabled(),
        .event_types_ = alert_handler.event_types(),
        .log_level_ = makeLogLevel(alert_handler.log_level()),
    };
}

SmartPingConfig PbConfigLoader::makeSmartPingConfig(
    const cfg::SmartPing &smartping) const {
    return {
        .check_interval_ = mk::ToChronoMicros(smartping.check_interval()),
        .omit_payload_ = smartping.omit_payload(),
        .num_sockets_ = smartping.num_sockets(),
        .ignore_rst_ = smartping.ignore_rst(),
        .wait_duration_ = mk::ToChronoNanos(smartping.wait_duration()),
        .bulk_size_ = smartping.bulk_size(),
        .log_level_ = makeLogLevel(smartping.log_level()),
        .icmp_sender_executable_ = makePath(smartping.icmp_sender_executable()),
        .icmp_receiver_executable_ =
            makePath(smartping.icmp_receiver_executable()),

    };
}

CheckingConfig PbConfigLoader::makeCheckingConfig(
    const cfg::Checking &checking) const {
    return {
        .generic_helper_executable_ =
            makePath(checking.generic_helper_executable()),
        .num_generic_helpers_ = checking.num_generic_helpers(),
        .num_real_time_helpers_ = checking.num_real_time_helpers(),
        .real_time_port_ =
            static_cast<uint16_t>(checking.real_time_port().number()),
        .fetcher_executable_ = makePath(checking.fetcher_executable()),
        .num_fetchers_ = checking.num_fetchers(),
        .fetchers_enabled_ = checking.fetchers_enabled(),
        .num_checkers_ = checking.num_checkers(),
        .log_level_ = makeLogLevel(checking.log_level()),
        .debug_ = checking.debug(),
        .initial_scheduling_burst_ = checking.initial_scheduling_burst(),
        .initial_generic_scheduling_interval_ =
            mk::ToChronoNanos(checking.initial_generic_scheduling_interval()),
        .initial_checkmk_scheduling_interval_ =
            mk::ToChronoNanos(checking.initial_checkmk_scheduling_interval()),
    };
}

LivestatusConfig PbConfigLoader::makeLivestatusConfig(
    const cfg::Livestatus &livestatus) const {
    return {
        .socket_ = makePath(livestatus.socket()),
        .command_pipe_ = makePath(livestatus.command_pipe()),
        .max_response_size_ = livestatus.max_response_size(),
        .stack_size_ = livestatus.stack_size(),
        .num_threads_ = livestatus.num_threads(),
        .log_level_ = makeLogLevel(livestatus.log_level()),
        .service_authorization_ =
            makeServiceAuthorization(livestatus.service_authorization()),
        .group_authorization_ =
            makeGroupAuthorization(livestatus.group_authorization()),
    };
}

// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
NotificationHandlerConfig PbConfigLoader::makeNotificationHandlerConfig(
    const cfg::NotificationHandler &notification_handler) const {
    return {
        .debug_ = notification_handler.debug(),
        .log_level_ = makeLogLevel(notification_handler.log_level()),
    };
}

HistoryConfig PbConfigLoader::makeHistoryConfig(
    const cfg::History &history) const {
    return {
        .max_cached_messages_ = history.max_cached_messages(),
        .max_lines_per_file_ = history.max_lines_per_file(),
        .cache_horizon_ = history.has_cache_horizon()
                              ? mk::ToChronoSeconds(history.cache_horizon())
                              : std::optional<std::chrono::seconds>{},
        .max_core_downtime_ = mk::ToChronoSeconds(history.max_core_downtime()),
        .file_ = makePath(history.file()),
        .archive_directory_ = makePath(history.archive_directory()),
        .rotation_interval_ =
            ToLogRotationInterval(history.rotation_interval()),
        .rotation_size_limit_ = history.rotation_size_limit(),
    };
}

RRDConfig PbConfigLoader::makeRRDConfig(const cfg::RRD &rrd) const {
    return {.single_directory_ = makePath(rrd.single_directory()),
            .multiple_directory_ = makePath(rrd.multiple_directory()),
            .rrdcached_socket_ = makePath(rrd.rrdcached_socket()),
            .log_level_ = makeLogLevel(rrd.log_level()),
            .pnp4nagios_enabled_ = rrd.pnp4nagios_enabled(),
            .pnp_update_delay_ = mk::ToChronoSeconds(rrd.pnp_update_delay()),
            .pnp_update_on_restart_ = rrd.pnp_update_on_restart()};
}

influxdb::ConnectionConfig PbConfigLoader::makeInfluxDbConnection(
    const cfg::InfluxDbConnection &connection) const {
    return {
        .id = makeString(connection.id_ref()),
        .enabled = connection.enabled(),
        .address =
            {
                .protocol = makeString(connection.protocol_ref()),
                .host = makeString(connection.host_name_ref()),
                .port = static_cast<uint16_t>(connection.port().number()),
            },
        .auth =
            {
                .ca_cert_file = makePath(connection.ca_cert_file()),
                .use_certificate_check = connection.use_certificate_check(),
            },
        .api_path = makeString(connection.api_path_ref()),
        .token = influxdb::Token{makeString(connection.token_ref())},
        .max_batch_size = connection.max_batch_size(),
        .max_batch_age = mk::ToChronoSeconds(connection.max_batch_age()),
        .retry_interval = mk::ToChronoMicros(connection.retry_interval()),
    };
}

InfluxDbConfig PbConfigLoader::makeInfluxDbConfig(
    const cfg::InfluxDb &influx_db) const {
    std::vector<influxdb::ConnectionConfig> connections;
    for (const auto &c : influx_db.connections()) {
        connections.emplace_back(makeInfluxDbConnection(c));
    }

    return {
        .connections_ = std::move(connections),
        .log_level_ = makeLogLevel(influx_db.log_level()),
    };
}

CarbonConnectionConfig PbConfigLoader::makeCarbonConnection(
    const cfg::CarbonConnection &connection) const {
    return {ToInAddr(connection.ip()),
            static_cast<uint16_t>(connection.port().number()), 15s,
            makeString(connection.connection_prefix_ref()),
            ToMangling(connection.mangling())};
}

CarbonConfig PbConfigLoader::makeCarbonConfig(const cfg::Carbon &carbon) const {
    std::vector<CarbonConnectionConfig> connections;
    for (const auto &c : carbon.connections()) {
        connections.emplace_back(makeCarbonConnection(c));
    }

    return {
        .connections_ = std::move(connections),
        .log_level_ = makeLogLevel(carbon.log_level()),
    };
}

std::unique_ptr<ContactGroup> PbConfigLoader::makeContactGroup(
    const cfg::ContactGroup &contact_group) const {
    return std::make_unique<ContactGroup>(
        makeString(contact_group.name_ref()),
        makeString(contact_group.alias_ref()));
}

ContactList PbConfigLoader::makeContactList(
    const cfg::ContactList &contact_list, const World *world) const {
    std::unordered_set<std::string> contact_names;
    contact_names.reserve(contact_list.contact_refs().size());
    for (auto ref : contact_list.contact_refs()) {
        contact_names.insert(makeString(ref));
    }
    std::vector<std::string> group_names;
    group_names.reserve(contact_list.group_refs().size());
    for (auto ref : contact_list.group_refs()) {
        group_names.emplace_back(makeString(ref));
    }
    return {std::move(contact_names), group_names, world};
}

Timeperiod PbConfigLoader::makeTimeperiod(
    const cfg::Timeperiod &timeperiod) const {
    std::vector<Timeperiod::value_type> transitions;
    for (const auto &transition : timeperiod.transitions()) {
        transitions.emplace_back(mk::ToChronoTimePoint(transition));
    }
    return Timeperiod{makeString(timeperiod.name_ref()),
                      makeString(timeperiod.alias_ref()),
                      std::move(transitions)};
}

Entity PbConfigLoader::makeEntity(const cfg::Entity &entity,
                                  PoolOffset pool_offset) const {
    std::unordered_map<std::string, std::string> attributes;
    for (const auto &attribute : entity.attributes()) {
        attributes.emplace(makeString(attribute.key_ref(), pool_offset),
                           makeString(attribute.value_ref(), pool_offset));
    }
    return {makeString(entity.name_ref(), pool_offset), attributes};
}

Entity PbConfigLoader::makeEntity(const cfg::Entity &entity) const {
    std::unordered_map<std::string, std::string> attributes;
    for (const auto &attribute : entity.attributes()) {
        attributes.emplace(makeString(attribute.key_ref()),
                           makeString(attribute.value_ref()));
    }
    return {makeString(entity.name_ref()), attributes};
}

// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
NotificationOptions PbConfigLoader::makeNotificationOptions(
    const cfg::NotificationOptions &options) const {
    NotificationOptions o{0};
    if (options.host_up()) {
        o |= NotificationOptions::host_up;
    }
    if (options.host_down()) {
        o |= NotificationOptions::host_down;
    }
    if (options.host_unreachable()) {
        o |= NotificationOptions::host_unreachable;
    }
    if (options.host_flapping()) {
        o |= NotificationOptions::host_flapping;
    }
    if (options.host_downtime()) {
        o |= NotificationOptions::host_downtime;
    }
    if (options.service_ok()) {
        o |= NotificationOptions::service_ok;
    }
    if (options.service_warning()) {
        o |= NotificationOptions::service_warning;
    }
    if (options.service_critical()) {
        o |= NotificationOptions::service_critical;
    }
    if (options.service_unknown()) {
        o |= NotificationOptions::service_unknown;
    }
    if (options.service_flapping()) {
        o |= NotificationOptions::service_flapping;
    }
    if (options.service_downtime()) {
        o |= NotificationOptions::service_downtime;
    }

    return o;
}

std::unique_ptr<Contact> PbConfigLoader::makeContact(
    const cfg::Contact &contact, World *world) const {
    auto c = std::make_unique<Contact>(makeEntity(contact.entity()));
    c->_notifications_enabled = contact.notifications_enabled();
    c->_notification_period = &world->getTimeperiod(
        "contact notification", makeString(contact.notification_period_ref()));
    c->_notification_options =
        makeNotificationOptions(contact.notification_options());
    return c;
}

// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
HostState PbConfigLoader::makeHostState(uint32_t host_state) const {
    switch (host_state) {
        case 0:
            return HostState::up;
        case 1:
            return HostState::down;
        case 2:
            return HostState::unreachable;
        default:
            // not possible, but now we have no unified policy what to do
            // with bad config: abort or exception may be better
            return HostState::up;
    }
}

std::unique_ptr<Host> PbConfigLoader::makeHost(Core &core, World &world,
                                               const cfg::Host &host) const {
    const PoolOffset pool_offset{host.pool_index()};
    auto h = std::make_unique<Host>(
        core, world, makeEntity(host.object().entity(), pool_offset));
    fillObject(*h, host.object(), pool_offset, world);
    h->_state_mapping[HostState::up] = makeHostState(host.state_up());
    h->_state_mapping[HostState::down] = makeHostState(host.state_down());
    h->_state_mapping[HostState::unreachable] =
        makeHostState(host.state_unreachable());
    for (const auto group_ref : host.group_refs()) {
        h->insertInGroup(
            world.ensureHostGroup(makeString(group_ref, pool_offset)));
    }
    h->_smart_ping_timeout = mk::ToChronoNanos(host.smart_ping_timeout());
    h->_ipv4_address = ToInAddr(host.ip());
    for (const auto &s : host.services()) {
        auto service = makeService(core, world, s, *h, pool_offset);
        h->_services_by_name[service->name()] = service.get();
        h->_services.insert(std::move(service));
    }

    return h;
}

void PbConfigLoader::fillObject(Object &o, const cfg::Object &object,
                                PoolOffset pool_offset, World &world) const {
    o.setState(nullptr);
    o._is_shadow = object.is_shadow();
    o._write_long_output = object.write_long_output();
    o._check_command_name =
        makeString(object.check_command_name_ref(), pool_offset);
    o._check_command_line =
        makeString(object.check_command_line_ref(), pool_offset);
    o._check_type = o._is_shadow ? 2 : (o._check_command_line.empty() ? 1 : 0);
    o.clearEnabledPerConfig();
    o.setEnabledPerConfig(State::Enabling::Key::active_checks,
                          !o._check_command_line.empty());
    o.setEnabledPerConfig(State::Enabling::Key::passive_checks,
                          o._check_command_line.empty());
    o.setEnabledPerConfig(State::Enabling::Key::notifications,
                          object.notifications_enabled());
    o.setEnabledPerConfig(State::Enabling::Key::flap_detection,
                          object.flap_detection_enabled());
    o.setEnabledPerConfig(State::Enabling::Key::performance_data,
                          object.process_performance_info());
    o._check_interval = mk::ToChronoNanos(object.check_interval());
    o._retry_interval = mk::ToChronoNanos(object.retry_interval());
    o._notification_interval =
        mk::ToChronoNanos(object.notification_interval());
    o._first_notification_delay =
        mk::ToChronoNanos(object.first_notification_delay());
    o._max_check_attempts = object.max_check_attempts();
    o._check_period = &world.getTimeperiod(
        "check", makeString(object.check_period_ref(), pool_offset));
    o._service_period = &world.getTimeperiod(
        "service", makeString(object.service_period_ref(), pool_offset));
    o._notification_period = &world.getTimeperiod(
        "notification",
        makeString(object.notification_period_ref(), pool_offset));
    o._notification_options =
        makeNotificationOptions(object.notification_options());
    o._check_timeout = mk::ToChronoSeconds(object.check_timeout());
    o._freshness_state = object.freshness_state();
    o._freshness_output =
        makeString(object.freshness_output_ref(), pool_offset);
    o._low_flap_threshold = object.low_flap_threshold();
    o._high_flap_threshold = object.high_flap_threshold();
    o._flap_decay = object.flap_decay();
    for (const auto &mapping : object.output_mappings()) {
        o.addOutputMappings(
            Replacer{makeRegex(mapping.regex_ref(), pool_offset, world.logger_),
                     makeString(mapping.replacement_ref(), pool_offset)});
    }
    o._rrd_format = makeRRDFormat(object.rrd_format());
    o._carbon_send_metric_parts =
        makeCarbonSendMetricParts(object.carbon_send_metric_parts());
    o._influxdb_send_config =
        object.has_influx_db_sender()
            ? makeInfluxDbSenderConfig(object.influx_db_sender(), pool_offset)
            : influxdb::SenderConfig::make();

    o.setEnabledPerConfig(State::Enabling::Key::freshness_checks,
                          o._freshness_output[0] != '\0');
    o._contact_list = world.getContactList(
        contact_list_pool_offsets_.entries(pool_offset.offset) +
        object.contact_list_index());
    for (const auto &downtime : object.recurring_downtimes()) {
        o._downtime_rules.push_back(std::make_unique<DowntimeRule>(
            makeDowntimeRule(downtime, pool_offset, world.logger_)));
    }
}

// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
ServiceState PbConfigLoader::makeServiceState(uint32_t service_state) const {
    switch (service_state) {
        case 0:
            return ServiceState::ok;
        case 1:
            return ServiceState::warning;
        case 2:
            return ServiceState::critical;
        case 3:
            return ServiceState::unknown;
        default:
            // not possible, but now we have no unified policy what to do
            // with bad config: abort or exception may be better
            return ServiceState::ok;
    }
}

std::unique_ptr<Service> PbConfigLoader::makeService(
    Core &core, World &world, const cfg::Service &service, Host &host,
    PoolOffset pool_offset) const {
    auto s = std::make_unique<Service>(
        core, world, host, makeEntity(service.object().entity(), pool_offset));
    fillObject(*s, service.object(), pool_offset, world);
    s->_state_mapping[ServiceState::ok] = makeServiceState(service.state_ok());
    s->_state_mapping[ServiceState::warning] =
        makeServiceState(service.state_warning());
    s->_state_mapping[ServiceState::critical] =
        makeServiceState(service.state_critical());
    s->_state_mapping[ServiceState::unknown] =
        makeServiceState(service.state_unknown());

    for (const auto group_ref : service.group_refs()) {
        s->insertInGroup(
            world.ensureServiceGroup(makeString(group_ref, pool_offset)));
    }

    s->passive_check_period_ = &world.getTimeperiod(
        "passive_check",
        makeString(service.passive_check_period_ref(), pool_offset));
    return s;
}

// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
RRDFormat PbConfigLoader::makeRRDFormat(cfg::RRDFormat rrd_format) const {
    switch (rrd_format) {
        case cfg::RRD_FORMAT_MULTIPLE:
            return RRDFormat::multiple;
        case cfg::RRD_FORMAT_SINGLE:
            return RRDFormat::single;
        default:
            // not possible, but now we have no unified policy what to do
            // with bad config: abort or exception may be better
            return RRDFormat::none;
    }
}

// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
CarbonSendMetricParts PbConfigLoader::makeCarbonSendMetricParts(
    const cfg::CarbonSendMetricParts &metrics) const {
    return {
        .value = metrics.value(),
        .uom = metrics.uom(),
        .warn = metrics.warn(),
        .crit = metrics.crit(),
        .min = metrics.min(),
        .max = metrics.max(),
        .state = metrics.state(),
        .scheduled_downtime_depth = metrics.scheduled_downtime_depth(),
        .acknowledged = metrics.acknowledged(),
    };
}

influxdb::SenderConfig PbConfigLoader::makeInfluxDbSenderConfig(
    const cfg::InfluxDbSender &sender, PoolOffset pool_offset) const {
    return {
        .connection = makeString(sender.connection_ref(), pool_offset),
        .org = makeString(sender.org_ref(), pool_offset),
        .bucket = makeString(sender.bucket_ref(), pool_offset),
        .parts =
            {
                .metric_value = sender.metric_value(),
                .metric_hst_state = sender.metric_hst_state(),
                .metric_svc_state = sender.metric_svc_state(),
                .add_hst_state = sender.add_hst_state(),
                .add_svc_state = sender.add_svc_state(),
                .add_downtime = sender.add_downtime(),
                .add_acknowledgement = sender.add_acknowledgement(),
                .add_uom = sender.add_uom(),
                .add_thresholds = sender.add_thresholds(),
                .add_min_max = sender.add_min_max(),
                .tag_hst_name = sender.tag_hst_name(),
                .tag_svc_name = sender.tag_svc_name(),
                .tag_hst_alias = sender.tag_hst_alias(),
                .tag_hst_tags = sender.tag_hst_tags(),
                .tag_hst_label = sender.tag_hst_label(),
                .tag_svc_label = sender.tag_svc_label(),
            },
    };
}

DowntimeRule PbConfigLoader::makeDowntimeRule(
    const cfg::PersistentRecurringDowntime &downtime, PoolOffset pool_offset,
    Logger *logger) const {
    return {
        ._comment = makeString(downtime.comment_ref(), pool_offset),
        ._start_time = mk::ToChronoTimePoint(downtime.start_time()),
        ._end_time = mk::ToChronoTimePoint(downtime.end_time()),
        ._duration = mk::ToChronoNanos(downtime.duration()),
        ._flex_duration = mk::ToChronoNanos(downtime.flex_duration()),
        ._cycle_time =
            CycleTime(mk::ToCMCCycleTime(downtime.cycle_time()), logger),
    };
}

std::shared_ptr<RegExp> PbConfigLoader::makeRegex(uint32_t ref,
                                                  PoolOffset pool_offset,
                                                  Logger *logger) const {
    const auto re = makeString(ref, pool_offset);
    try {
        return regex_cache_[re] = std::make_shared<RegExp>(
                   re, RegExp::Case::respect, RegExp::Syntax::pattern);
    } catch (const std::runtime_error &e) {
        Warning(logger) << "invalid regular expression '" << re
                        << "': " << e.what();
        return nullptr;
    }
}

[[nodiscard]] std::pair<std::string, std::string>
PbConfigLoader::makeObjectGroup(const cfg::ObjectGroup &group) const {
    return {makeString(group.name_ref()), makeString(group.alias_ref())};
}
