// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/PbStateLoader.h"

#include <algorithm>
#include <array>
#include <chrono>
#include <cstddef>
#include <iterator>
#include <memory>
#include <stdexcept>
#include <string>

#include "cmc/Comment.h"
#include "cmc/Comments.h"
#include "cmc/Core.h"
#include "cmc/Downtime.h"
#include "cmc/Downtimes.h"
#include "cmc/EventSpooler.h"
#include "cmc/GlobalState.h"
#include "cmc/Host.h"
#include "cmc/MonitoringHistory.h"
#include "cmc/Object.h"
#include "cmc/ObjectResolver.h"
#include "cmc/PbCMC.h"
#include "cmc/PbDuration.h"
#include "cmc/PbTimestamp.h"
#include "cmc/State.h"
#include "cmc/cmc.h"
#include "livestatus/ChronoUtils.h"
#include "state/v1/types.pb.h"

namespace pb = checkmk::cmc::state::v1;

namespace {
Comment::Type loadCommentType(const pb::CommentType &ct) {
    switch (ct) {
        case pb::CommentType::COMMENT_TYPE_USER:
            return Comment::Type::user;
        case pb::CommentType::COMMENT_TYPE_DOWNTIME:
            return Comment::Type::downtime;
        case pb::CommentType::COMMENT_TYPE_FLAPPING:
            return Comment::Type::flapping;
        case pb::CommentType::COMMENT_TYPE_ACKNOWLEDGEMENT:
            return Comment::Type::acknowledgement;
        default:
            throw std::runtime_error{"unknown Comment::Type: " +
                                     std::to_string(static_cast<int>(ct))};
    }
}

bool keepOnLoad(const Object &object,
                const checkmk::cmc::state::v1::Comment &comment) {
    switch (loadCommentType(comment.type())) {
        case Comment::Type::user:
        case Comment::Type::downtime:
        case Comment::Type::flapping:
            return comment.persistent();
        case Comment::Type::acknowledgement:
            // NOTE: We should never see a non-persistent acknowledgement where
            // the original problem is no longer acknowledged, but just to be
            // sure... Furthermore, we want to keep comments for objects which
            // have no state yet, e.g. objects after a migration from Nagios.
            return comment.persistent() ||       //
                   object.state() == nullptr ||  //
                   object.acknowledged();
    }
    return true;  // unreachable, make GCC happy
}

void loadComments(Comments &comments, const ObjectResolver &resolver,
                  const pb::Comments &x) {
    comments.next_id_ = x.next_id();
    comments.clear();
    for (auto &&comment : x.entries()) {
        // The associated object could have vanished in a new configuration.
        if (const auto *object = resolver.objectRef(comment.host_name(),
                                                    comment.service_name())) {
            if (keepOnLoad(*object, comment)) {
                comments.insert(std::make_unique<Comment>(
                    PbStateLoader::loadComment(*object, comment)));
            }
        }
    }
}

void loadDowntimes(Downtimes &downtimes, Core &core,
                   const ObjectResolver &resolver, const pb::Downtimes &x) {
    downtimes.next_id_ = x.next_id();
    downtimes.clear();
    for (auto &&downtime : x.entries()) {
        // The associated object could have vanished in a new configuration.
        if (const auto *object = resolver.objectRef(downtime.host_name(),
                                                    downtime.service_name())) {
            downtimes.insert(std::make_unique<Downtime>(
                PbStateLoader::loadDowntime(*object, core, downtime)));
        }
    }
    // TODO(ml): Is that still required?
    downtimes.validateDepths(core.world());
}

void loadSpooledEvents(
    EventSpooler &spooler,
    const google::protobuf::RepeatedPtrField<pb::SpooledEvent> &events) {
    spooler.clear();
    for (const auto &event : events) {
        spooler.spool(event.payload());
    }
}
}  // namespace

void PbStateLoader::loadStateFile(Core &core, const ObjectResolver &resolver,
                                  const pb::StateFile &state_file, Count count,
                                  Logger *logger) {
    *core.globalState() =
        PbStateLoader::loadGlobalState(state_file.global_state());
    PbStateLoader::loadHosts(resolver, state_file.hosts(), count, logger);
    loadComments(core._comments, resolver, state_file.comments());
    loadDowntimes(core._downtimes, core, resolver, state_file.downtimes());
    loadSpooledEvents(core.alert_spooler_, state_file.spooled_alerts());
    loadSpooledEvents(core.notification_spooler_,
                      state_file.spooled_notifications());
    core.monitoring_history_.last_rotation_ = mk::ToChronoTimePoint(
        state_file.global_state().monitoring_history_last_rotation_time());
}

void PbStateLoader::loadHosts(const ObjectResolver &resolver,
                              const checkmk::cmc::state::v1::Hosts &pb_hsts,
                              Count count, Logger *logger) {
    count.reset();
    for (auto &&[pb_hst_name, pb_hst] : pb_hsts.hosts()) {
        if (auto *host = resolver.hostRef(pb_hst_name)) {
            host->setState(std::make_unique<State>(
                               PbStateLoader::loadState(pb_hst.state()))
                               .release());
            host->sanitizeState(logger);
            count.num_host_states_loaded++;
            for (auto &&[pb_svc_name, pb_svc] : pb_hst.services()) {
                if (auto *svc = resolver.objectRef(pb_hst_name, pb_svc_name)) {
                    svc->setState(std::make_unique<State>(
                                      PbStateLoader::loadState(pb_svc))
                                      .release());
                    svc->sanitizeState(logger);
                    count.num_service_states_loaded++;
                } else {
                    count.num_services_vanished++;
                }
            }
        } else {
            count.num_hosts_vanished++;
        }
    }
}

GlobalState PbStateLoader::loadGlobalState(
    const pb::GlobalState &global_state) {
    return GlobalState{GlobalStateFields{
        ._enable_notifications = global_state.enable_notifications(),
        ._execute_service_checks = global_state.execute_service_checks(),
        ._execute_host_checks = global_state.execute_host_checks(),
        ._enable_flap_detection = global_state.enable_flap_detection(),
        ._process_performance_data = global_state.process_performance_info(),
        ._enable_alert_handlers = global_state.enable_alert_handlers(),
        ._notification_next_problem_id =
            global_state.notification_next_problem_id(),
        ._state_file_created = mk::demangleTimePoint(global_state.creation())}};
}

namespace {
Comment::Source loadCommentSource(const pb::CommentSource &cs) {
    switch (cs) {
        case pb::CommentSource::COMMENT_SOURCE_INTERNAL:
            return Comment::Source::internal;
        case pb::CommentSource::COMMENT_SOURCE_EXTERNAL:
            return Comment::Source::external;
        default:
            throw std::runtime_error{"unknown Comment::Source: " +
                                     std::to_string(static_cast<int>(cs))};
    }
}
}  // namespace

Comment PbStateLoader::loadComment(
    const Object &object, const checkmk::cmc::state::v1::Comment &comment) {
    return Comment{._object = &object,
                   ._id = comment.id(),
                   ._entry_time = mk::ToChronoTimePoint(comment.entry_time()),
                   ._author = comment.author(),
                   ._comment = comment.comment(),
                   ._entry_type = loadCommentType(comment.type()),
                   ._expire_time = mk::ToChronoTimePoint(comment.expire_time()),
                   ._persistent = comment.persistent(),
                   ._sticky = comment.sticky(),
                   ._source = loadCommentSource(comment.source())};
}

namespace {
Downtime::Status loadDowntimeStatus(const pb::DowntimeStatus &ds) {
    switch (ds) {
        case pb::DowntimeStatus::DOWNTIME_STATUS_ACTIVE:
            return Downtime::Status::active;
        case pb::DowntimeStatus::DOWNTIME_STATUS_PENDING:
            return Downtime::Status::pending;
        default:
            throw std::runtime_error{"unknown Downtime::Status: " +
                                     std::to_string(static_cast<int>(ds))};
    }
}
}  // namespace

Downtime PbStateLoader::loadDowntime(const Object &object, Core &core,
                                     const pb::Downtime &downtime) {
    return {.core_ = core,
            .logger_downtime_ = core._logger_downtime,
            .logger_notification_ = core._logger_notification,
            ._object = &object,
            ._id = downtime.id(),
            ._entry_time = mk::ToChronoTimePoint(downtime.entry_time()),
            ._author = downtime.author(),
            ._comment = downtime.comment(),
            ._status = loadDowntimeStatus(downtime.downtime_status()),
            ._start_time = mk::ToChronoTimePoint(downtime.start_time()),
            ._end_time = mk::ToChronoTimePoint(downtime.end_time()),
            ._flex_downtime_start =
                mk::ToChronoTimePoint(downtime.flex_downtime_start_time()),
            ._fixed = downtime.fixed(),
            ._cycle_time = {mk::ToCMCCycleTime(downtime.cycle_time()),
                            core._logger_downtime},
            ._duration = mk::ToChronoNanos(downtime.duration()),
            ._hot = downtime.hot(),
            ._origin_is_rule = downtime.origin_is_rule(),
            // handleChangedDowntimeRules() sets _rule correctly.
            ._rule = nullptr};
}

namespace {
AcknowledgementType loadAcknowledgementType(const pb::AcknowledgementType &at) {
    switch (at) {
        case pb::AcknowledgementType::ACKNOWLEDGEMENT_TYPE_NONE:
            return AcknowledgementType::none;
        case pb::AcknowledgementType::ACKNOWLEDGEMENT_TYPE_NORMAL:
            return AcknowledgementType::normal;
        case pb::AcknowledgementType::ACKNOWLEDGEMENT_TYPE_STICKY:
            return AcknowledgementType::sticky;
        default:
            throw std::runtime_error{"unknown AcknowledgementType: " +
                                     std::to_string(static_cast<int>(at))};
    }
}

NotificationKind loadNotificationKind(const pb::NotificationKind &nk) {
    switch (nk) {
        case pb::NotificationKind::NOTIFICATION_KIND_NONE:
            return NotificationKind::none;
        case pb::NotificationKind::NOTIFICATION_KIND_NORMAL:
            return NotificationKind::normal;
        case pb::NotificationKind::NOTIFICATION_KIND_FLAPPING:
            return NotificationKind::flapping;
        default:
            throw std::runtime_error{"unknown NotificationKind: " +
                                     std::to_string(static_cast<int>(nk))};
    }
}

StateType loadStateType(const pb::StateType &st) {
    switch (st) {
        case pb::StateType::STATE_TYPE_SOFT:
            return StateType::soft;
        case pb::StateType::STATE_TYPE_HARD:
            return StateType::hard;
        default:
            throw std::runtime_error{"unknown StateType: " +
                                     std::to_string(static_cast<int>(st))};
    }
}

State::StateCheckResult loadStateCheckResult(const pb::StateCheckResult &cr) {
    return {.output_ = cr.output(),
            .long_output_ = cr.long_output(),
            .perf_data_ = cr.perf_info(),
            .cached_at_ = mk::ToChronoTimePoint(cr.cached_at_time()),
            .cache_interval_ = mk::ToChronoSeconds(cr.cache_interval())};
}

State::Downtime loadStateDowntime(const pb::StateDowntime &dt) {
    return {.pending_flex_ = dt.pending_flex(),
            .scheduled_depth_ = dt.scheduled_depth()};
}

State::Enabling loadStateEnabling(const pb::StateEnabling &es) {
    return {.value_ = es.value(), .modified_ = es.modified()};
}

State::Flappiness loadStateFlappiness(const pb::StateFlappiness &fs) {
    return {fs.is_flapping(), fs.value()};
}

State::HardState loadStateHardState(const pb::StateHardState &hs) {
    return {.last_ = hs.last(),
            .previous_ = hs.previous(),
            .changed_ = hs.changed(),
            .last_time_ = mk::ToChronoTimePoint(hs.last_time())};
}

State::Notification loadStateNotification(const pb::StateNotification &ns) {
    return {.current_number_ = ns.current_number(),
            .last_time_ = mk::ToChronoTimePoint(ns.last_time()),
            .next_time_ = mk::ToChronoTimePoint(ns.next_time()),
            .pending_kind_ = loadNotificationKind(ns.pending_kind()),
            .problem_id_ = ns.problem_id(),
            .problem_sent_ = ns.problem_sent(),
            .postponement_reason_ = ns.postponement_reason()};
}
}  // namespace

State PbStateLoader::loadState(const pb::State &state) {
    auto out =
        State{.acknowledgement_type_ =
                  loadAcknowledgementType(state.acknowledgement_type()),
              .check_result_ = loadStateCheckResult(state.check_result()),
              .downtime_ = loadStateDowntime(state.downtime()),
              .enabling_ = loadStateEnabling(state.enabling()),
              .flappiness_ = loadStateFlappiness(state.flappiness()),
              .hard_state_ = loadStateHardState(state.hard_state()),
              .notification_ = loadStateNotification(state.notification()),
              .current_attempt_ = state.current_attempt(),
              .has_been_checked_ = state.has_been_checked(),
              .execution_time_ = mk::ToChronoNanos(state.execution_time()),
              .latency_ = mk::ToChronoNanos(state.latency()),
              .last_check_ = mk::ToChronoTimePoint(state.last_check_time()),
              .last_check_started_ =
                  mk::ToChronoTimePoint(state.last_check_started_time()),
              .last_state_change_ =
                  mk::ToChronoTimePoint(state.last_state_change_time()),
              // Filled bellow because it is difficult to fill
              // an array with a one liner.
              .last_time_state_ = {},
              .next_check_ = mk::ToChronoTimePoint(state.next_check_time()),
              .current_state_ = state.current_state(),
              .initial_state_ = state.initial_state(),
              .last_state_ = state.last_state(),
              .state_type_ = loadStateType(state.state_type())};
    if (static_cast<std::size_t>(state.last_time_state_time_size()) ==
        out.last_time_state_.size()) {
        // Silently ignore size mismatch.
        std::transform(std::begin(state.last_time_state_time()),
                       std::end(state.last_time_state_time()),
                       std::begin(out.last_time_state_), mk::ToChronoTimePoint);
    }
    return out;
}
