// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/PbStateSaver.h"

#include <array>
#include <chrono>
#include <functional>
#include <string>
#include <unordered_set>

#include "cmc/Comment.h"
#include "cmc/Comments.h"
#include "cmc/Core.h"
#include "cmc/CycleTime.h"
#include "cmc/Downtime.h"
#include "cmc/Downtimes.h"
#include "cmc/EventSpooler.h"
#include "cmc/GlobalState.h"
#include "cmc/Host.h"
#include "cmc/MonitoringHistory.h"
#include "cmc/Object.h"
#include "cmc/PbCMC.h"
#include "cmc/PbDuration.h"
#include "cmc/PbTimestamp.h"
#include "cmc/Service.h"
#include "cmc/State.h"
#include "cmc/World.h"
#include "cmc/cmc.h"
#include "google/protobuf/duration.pb.h"
#include "google/protobuf/timestamp.pb.h"
#include "livestatus/ChronoUtils.h"
#include "state/v1/types.pb.h"

namespace pb = checkmk::cmc::state::v1;

namespace {
pb::Comments saveComments(const Comments &x) {
    auto out = pb::Comments{};
    out.set_next_id(x.next_id_);
    x.all_of([&out](const Comment &comment) {
        *out.add_entries() = PbStateSaver::saveComment(comment);
        return true;
    });
    return out;
}

pb::Downtimes saveDowntimes(const Downtimes &x) {
    auto out = pb::Downtimes{};
    out.set_next_id(x.next_id_);
    x.all_of([&out](const Downtime &downtime) {
        *out.add_entries() = PbStateSaver::saveDowntime(downtime);
        return true;
    });
    return out;
}

template <typename CreateEvent>
void saveSpooledEvents(const EventSpooler &spooler, CreateEvent create_event) {
    spooler.for_each([&create_event](const std::string &payload) {
        create_event()->set_payload(payload);
    });
}
}  // namespace

pb::StateFile PbStateSaver::saveStateFile(const Core &core) {
    auto out = pb::StateFile{};
    *out.mutable_global_state() =
        PbStateSaver::saveGlobalState(*core.globalState());
    *out.mutable_global_state()
         ->mutable_monitoring_history_last_rotation_time() =
        mk::ToProtoTimestamp(core.monitoring_history_.last_rotation_);
    *out.mutable_hosts() =
        PbStateSaver::saveHosts(core.world()->config_.hosts_);
    *out.mutable_comments() = saveComments(core._comments);
    *out.mutable_downtimes() = saveDowntimes(core._downtimes);
    saveSpooledEvents(core.alert_spooler_,
                      [&out]() { return out.add_spooled_alerts(); });
    saveSpooledEvents(core.notification_spooler_,
                      [&out]() { return out.add_spooled_notifications(); });
    return out;
}

pb::Hosts PbStateSaver::saveHosts(
    std::vector<std::unique_ptr<Host>> &cmc_hsts) {
    auto out = pb::Hosts{};
    for (const auto &cmc_hst : cmc_hsts) {
        auto [it, ok] = out.mutable_hosts()->insert({cmc_hst->name(), {}});
        if (!ok) {
            continue;
        }
        *it->second.mutable_state() = PbStateSaver::saveState(*cmc_hst->_state);
        for (const auto &cmc_svc : cmc_hst->_services) {
            it->second.mutable_services()->insert(
                {cmc_svc->name(), PbStateSaver::saveState(*cmc_svc->_state)});
        }
    }
    return out;
}

pb::GlobalState PbStateSaver::saveGlobalState(const GlobalState &global_state) {
    auto out = pb::GlobalState{};
    out.set_enable_notifications(global_state.getEnableNotifications());
    out.set_execute_service_checks(global_state.getExecuteServiceChecks());
    out.set_execute_host_checks(global_state.getExecuteHostChecks());
    out.set_enable_flap_detection(global_state.getEnableFlapDetection());
    out.set_process_performance_info(global_state.getProcessPerformanceData());
    out.set_enable_alert_handlers(global_state.getEnableAlertHandlers());
    out.set_notification_next_problem_id(
        global_state.getNotificationNextProblemId());
    out.set_creation(mk::mangleTimePoint(global_state.getStateFileCreated()));
    return out;
}

namespace {
pb::CommentType saveCommentType(const Comment::Type &t) {
    switch (t) {
        case Comment::Type::user:
            return pb::CommentType::COMMENT_TYPE_USER;
        case Comment::Type::downtime:
            return pb::CommentType::COMMENT_TYPE_DOWNTIME;
        case Comment::Type::flapping:
            return pb::CommentType::COMMENT_TYPE_FLAPPING;
        case Comment::Type::acknowledgement:
            return pb::CommentType::COMMENT_TYPE_ACKNOWLEDGEMENT;
    }
    // unreachable
    return {};
};

pb::CommentSource saveCommentSource(const Comment::Source &s) {
    switch (s) {
        case Comment::Source::internal:
            return pb::CommentSource::COMMENT_SOURCE_INTERNAL;
        case Comment::Source::external:
            return pb::CommentSource::COMMENT_SOURCE_EXTERNAL;
    }
    // unreachable
    return {};
};
}  // namespace

pb::Comment PbStateSaver::saveComment(const Comment &comment) {
    auto out = pb::Comment{};
    if (comment._object != nullptr && comment._object->host() != nullptr) {
        out.set_host_name(comment._object->host()->name());
    }
    if (comment._object != nullptr && comment._object->service() != nullptr) {
        out.set_service_name(comment._object->service()->name());
    }
    out.set_id(comment._id);
    *out.mutable_entry_time() = mk::ToProtoTimestamp(comment._entry_time);
    out.set_author(comment._author);
    out.set_comment(comment._comment);
    out.set_type(saveCommentType(comment._entry_type));
    *out.mutable_expire_time() = mk::ToProtoTimestamp(comment._expire_time);
    out.set_persistent(comment._persistent);
    out.set_sticky(comment._sticky);
    out.set_source(saveCommentSource(comment._source));
    return out;
}

namespace {
pb::DowntimeStatus saveDowntimeStatus(const Downtime::Status &ds) {
    switch (ds) {
        case Downtime::Status::active:
            return pb::DowntimeStatus::DOWNTIME_STATUS_ACTIVE;
        case Downtime::Status::pending:
            return pb::DowntimeStatus::DOWNTIME_STATUS_PENDING;
    }
    return {};
}
}  // namespace

pb::Downtime PbStateSaver::saveDowntime(const Downtime &downtime) {
    auto out = pb::Downtime{};
    if (downtime._object != nullptr && downtime._object->host() != nullptr) {
        out.set_host_name(downtime._object->host()->name());
    }
    if (downtime._object != nullptr && downtime._object->service() != nullptr) {
        out.set_service_name(downtime._object->service()->name());
    }
    out.set_id(downtime._id);
    *out.mutable_entry_time() = mk::ToProtoTimestamp(downtime._entry_time);
    out.set_author(downtime._author);
    out.set_comment(downtime._comment);
    out.set_downtime_status(saveDowntimeStatus(downtime._status));
    *out.mutable_start_time() = mk::ToProtoTimestamp(downtime._start_time);
    *out.mutable_end_time() = mk::ToProtoTimestamp(downtime._end_time);
    *out.mutable_flex_downtime_start_time() =
        mk::ToProtoTimestamp(downtime._flex_downtime_start);
    out.set_fixed(downtime._fixed);
    out.set_cycle_time(mk::ToProtoCycleTime(downtime._cycle_time.getKind()));
    *out.mutable_duration() = mk::ToProtoNanos(downtime._duration);
    out.set_hot(downtime._hot);
    out.set_origin_is_rule(downtime._origin_is_rule);
    return out;
}

namespace {
pb::AcknowledgementType saveAcknowledgementType(const AcknowledgementType &at) {
    switch (at) {
        case AcknowledgementType::none:
            return pb::AcknowledgementType::ACKNOWLEDGEMENT_TYPE_NONE;
        case AcknowledgementType::normal:
            return pb::AcknowledgementType::ACKNOWLEDGEMENT_TYPE_NORMAL;
        case AcknowledgementType::sticky:
            return pb::AcknowledgementType::ACKNOWLEDGEMENT_TYPE_STICKY;
    }
    // unreachable
    return {};
}

pb::NotificationKind saveNotificationKind(const NotificationKind &nk) {
    switch (nk) {
        case NotificationKind::none:
            return pb::NotificationKind::NOTIFICATION_KIND_NONE;
        case NotificationKind::normal:
            return pb::NotificationKind::NOTIFICATION_KIND_NORMAL;
        case NotificationKind::flapping:
            return pb::NotificationKind::NOTIFICATION_KIND_FLAPPING;
    }
    // unreachable
    return {};
}

pb::StateType saveStateType(const StateType &st) {
    switch (st) {
        case StateType::soft:
            return pb::StateType::STATE_TYPE_SOFT;
        case StateType::hard:
            return pb::StateType::STATE_TYPE_HARD;
    }
    // unreachable
    return {};
}

pb::StateCheckResult saveStateCheckResult(const State::StateCheckResult &cr) {
    auto out = pb::StateCheckResult{};
    out.set_output(cr.output_);
    out.set_long_output(cr.long_output_);
    out.set_perf_info(cr.perf_data_);
    *out.mutable_cached_at_time() = mk::ToProtoTimestamp(cr.cached_at_);
    *out.mutable_cache_interval() = mk::ToProtoSeconds(cr.cache_interval_);
    return out;
}

pb::StateDowntime saveStateDowntime(const State::Downtime &dt) {
    auto out = pb::StateDowntime{};
    out.set_pending_flex(dt.pending_flex_);
    out.set_scheduled_depth(dt.scheduled_depth_);
    return out;
}

pb::StateEnabling saveStateEnabling(const State::Enabling &e) {
    auto out = pb::StateEnabling{};
    out.set_value(e.value_);
    out.set_modified(e.modified_);
    return out;
}

pb::StateFlappiness saveStateFlappiness(const State::Flappiness &sf) {
    auto out = pb::StateFlappiness{};
    out.set_is_flapping(sf.is_flapping_);
    out.set_value(sf.value_);
    return out;
}

pb::StateHardState saveStateHardState(const State::HardState &hs) {
    auto out = pb::StateHardState{};
    out.set_last(hs.last_);
    out.set_previous(hs.previous_);
    out.set_changed(hs.changed_);
    *out.mutable_last_time() = mk::ToProtoTimestamp(hs.last_time_);
    return out;
}

pb::StateNotification saveStateNotification(const State::Notification &ns) {
    auto out = pb::StateNotification{};
    *out.mutable_last_time() = mk::ToProtoTimestamp(ns.last_time_);
    *out.mutable_next_time() = mk::ToProtoTimestamp(ns.next_time_);
    out.set_current_number(ns.current_number_);
    out.set_pending_kind(saveNotificationKind(ns.pending_kind_));
    out.set_problem_id(ns.problem_id_);
    out.set_problem_sent(ns.problem_sent_);
    out.set_postponement_reason(ns.postponement_reason_);
    return out;
}
}  // namespace

pb::State PbStateSaver::saveState(const State &state) {
    auto out = pb::State{};
    out.set_acknowledgement_type(
        saveAcknowledgementType(state.acknowledgement_type_));
    *out.mutable_check_result() = saveStateCheckResult(state.check_result_);
    *out.mutable_downtime() = saveStateDowntime(state.downtime_);
    *out.mutable_enabling() = saveStateEnabling(state.enabling_);
    *out.mutable_hard_state() = saveStateHardState(state.hard_state_);
    *out.mutable_flappiness() = saveStateFlappiness(state.flappiness_);
    *out.mutable_notification() = saveStateNotification(state.notification_);
    out.set_current_attempt(state.current_attempt_);
    out.set_has_been_checked(state.has_been_checked_);
    *out.mutable_execution_time() =
        mk::ToProtoNanos(std::chrono::duration_cast<std::chrono::nanoseconds>(
            state.execution_time_));
    *out.mutable_latency() = mk::ToProtoNanos(
        std::chrono::duration_cast<std::chrono::nanoseconds>(state.latency_));
    *out.mutable_last_check_time() = mk::ToProtoTimestamp(state.last_check_);
    *out.mutable_last_check_started_time() =
        mk::ToProtoTimestamp(state.last_check_started_);
    *out.mutable_last_state_change_time() =
        mk::ToProtoTimestamp(state.last_state_change_);
    for (auto &&x : state.last_time_state_) {
        *out.add_last_time_state_time() = mk::ToProtoTimestamp(x);
    }
    *out.mutable_next_check_time() = mk::ToProtoTimestamp(state.next_check_);
    out.set_current_state(state.current_state_);
    out.set_initial_state(state.initial_state_);
    out.set_last_state(state.last_state_);
    out.set_state_type(saveStateType(state.state_type_));
    return out;
}
