// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/Process.h"

#include <fcntl.h>
#include <sys/wait.h>
#include <unistd.h>

#include <algorithm>
#include <cerrno>
#include <chrono>
#include <compare>
#include <csignal>
#include <cstdlib>
#include <ostream>
#include <thread>
#include <utility>

#include "cmc/daemonize.h"
#include "livestatus/Logger.h"

using namespace std::chrono_literals;

namespace details {
void close(int &fd) {
    if (fd == -1) {
        return;
    }
    ::close(fd);
    fd = -1;
}
}  // namespace details

namespace {
template <typename Rep, typename Period>
pid_t waitpid_with_timeout(
    pid_t pid, int *wstatus,
    const std::chrono::duration<Rep, Period> &timeout_duration) {
    auto deadline = std::chrono::steady_clock::now() + timeout_duration;
    do {  // NOLINT(cppcoreguidelines-avoid-do-while)
        const pid_t wpid = waitpid(pid, wstatus, WNOHANG);
        if (wpid > 0 || (wpid == -1 && errno != EINTR)) {
            return wpid;
        }
        if (wpid == 0) {
            std::this_thread::sleep_for(5ms);
        }
    } while (std::chrono::steady_clock::now() <= deadline);
    errno = ETIMEDOUT;
    return -1;
}

void waitForPID(Logger *logger, pid_t pid, int &status) {
    Debug(logger) << "waiting for termination";
    pid_t wpid = waitpid_with_timeout(pid, &status, 3s);
    if (wpid == -1 && errno == ETIMEDOUT) {
        Warning(logger) << "forcibly terminating child process";
        errno = 0;
        kill(pid, SIGKILL);
        wpid = waitpid_with_timeout(pid, &status, 3s);
    }
    if (wpid == -1) {
        const generic_error ge("error while waiting");
        Error(logger) << ge;
        return;
    }

    if (WIFSIGNALED(status)) {
        if (WTERMSIG(status) == SIGTERM) {
            Notice(logger) << "terminated ";
        } else {
            Error(logger) << "killed by signal " << WTERMSIG(status);
        }
    } else if (WIFSTOPPED(status)) {
        Error(logger) << "stopped by signal " << WSTOPSIG(status);
    } else if (WIFCONTINUED(status)) {
        Error(logger) << "continued";
    } else if (WIFEXITED(status)) {
        if (WEXITSTATUS(status) == 0) {
            Notice(logger) << "exited normally";
        } else {
            Error(logger) << "exited with status " << WEXITSTATUS(status);
        }
    }
}

}  // namespace

Process::Process(std::filesystem::path path, std::vector<std::string> args,
                 int fd_in, int fd_out, int fd_err, Logger *logger)
    : path_{std::move(path)}
    , args_{std::move(args)}
    , fd_in_{fd_in}
    , fd_out_{fd_out}
    , fd_err_{fd_err}
    , logger_{logger} {}

Process::~Process() { this->closeSockets(); }

void Process::closeSockets() {
    details::close(fd_in_);
    details::close(fd_out_);
    details::close(fd_err_);
}

bool Process::detach() {
    const pid_t new_pid = ::fork();
    switch (new_pid) {
        case -1: {
            return false;
        }
        case 0: {
            this->run();
            return false;
        }
        default: {
            pid_ = new_pid;
            closeSockets();
            return true;
        }
    }
}

namespace {
int check_fd(int fd, Logger *logger) {
    if (fd != -1) {
        return fd;
    }
    switch (errno) {
        case EMFILE:  // too many files open
        case ENFILE:  // too many files open in system
        case ENOMEM:  // not enough memory
        {
            const generic_error ge{"cannot create file descriptor"};
            Emergency(logger) << "irrecoverable error" << ge;
            std::abort();
        }
        default:
            break;
    }
    return fd;
}
}  // namespace

void Process::run() {
    check_fd(::dup2(fd_in_ == -1 ? ::open("/dev/null", O_RDONLY) : fd_in_, 0),
             logger());
    check_fd(::dup2(fd_out_ == -1 ? ::open("/dev/null", O_WRONLY) : fd_out_, 1),
             logger());
    check_fd(::dup2(fd_err_ == -1 ? ::open("/dev/null", O_WRONLY) : fd_err_, 2),
             logger());
    mk::closefrom(3);

    std::vector<const char *> argv;
    argv.push_back(path_.c_str());
    for (auto &arg : args_) {
        argv.push_back(arg.c_str());
    }
    argv.push_back(nullptr);

    // http://pubs.opengroup.org/onlinepubs/9699919799/functions/exec.html
    // explains in its rationale why the const_cast is actually OK and only
    // caused by a histortic artifact.
    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-const-cast)
    ::execv(path_.c_str(), const_cast<char *const *>(argv.data()));

    const generic_error ge("cannot start " + path_.string());
    Alert(logger()) << ge;

    std::this_thread::sleep_for(3s);
    exit(1);
}

void Process::terminate() {
    if (!this->valid()) {
        return;
    }
    ::kill(this->id(), SIGTERM);
    terminated_ = true;
}

void Process::interrupt() {
    if (!this->valid()) {
        return;
    }
    ::kill(this->id(), SIGINT);
    terminated_ = true;
}

int Process::exitCode() const { return WEXITSTATUS(status_); }

void Process::wait() {
    if (this->valid()) {
        waitForPID(logger(), this->id(), status_);
        this->closeSockets();
        pid_ = -1;
    }
}

void Process::join() { this->wait(); }

std::ostream &operator<<(std::ostream &os, const Process &p) {
    os << p.path_.c_str();
    for (const auto &arg : p.args_) {
        os << " " << arg;
    }
    return os;
}
