// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/ProcessHelper.h"

#include <unistd.h>

#include <chrono>
#include <ostream>
#include <utility>

#include "livestatus/Logger.h"
#include "livestatus/POSIXUtils.h"

using namespace std::chrono_literals;
using namespace std::string_literals;

ProcessHelper::ProcessHelper(std::string name, std::filesystem::path path,
                             args_type args, Logger *logger)
    : AbstractHelper{std::move(name),
                     ContextLogger{logger,
                                   [this](std::ostream &os) {
                                       os << "[" << this->name() << " "
                                          << this->pidStr() << "] ";
                                   }}}
    , path_{std::move(path)}
    , args_{std::move(args)} {}

bool ProcessHelper::hasSameConfig(const std::filesystem::path &path,
                                  const args_type &args) const {
    return path == this->path() && args == this->args();
}

// static
void ProcessHelper::close(int &fd) {
    if (fd == -1) {
        return;
    }
    ::close(fd);
    fd = -1;
}

void ProcessHelper::closeSockets() {
    this->close(std_in_);
    this->close(std_out_);
    this->close(std_err_);
}

[[nodiscard]] std::string ProcessHelper::pidStr() const {
    const auto &p = process();
    return (p && p->valid()) ? std::to_string(p->id()) : "<no PID yet>"s;
}

bool ProcessHelper::spawnConnectedSubprocess() {
    auto std_in =
        SocketPair::make(SocketPair::Mode::local_non_blocking,
                         SocketPair::Direction::bidirectional, logger());
    if (!std_in) {
        return false;
    }
    auto std_err =
        SocketPair::make(SocketPair::Mode::local_non_blocking,
                         SocketPair::Direction::remote_to_local, logger());
    if (!std_err) {
        std_in->close();
        return false;
    }
    return spawn(*std_in, *std_in, *std_err);
}

bool ProcessHelper::spawn(const SocketPair &std_in, const SocketPair &std_out,
                          const SocketPair &std_err) {
    setFds(std_in.local(), std_out.local(), std_err.local());
    process().emplace(path(), args(), std_in.remote(), std_out.remote(),
                      std_err.remote(), logger());
    return detach();
}

void ProcessHelper::despawn() {
    closeSockets();
    if (auto &p = process()) {
        p->join();
    }
    process().reset();
}

bool ProcessHelper::detach() {
    auto &p = process();
    if (p && p->detach()) {
        Notice(logger()) << "started, commandline: " << *p;
        return true;
    }
    const generic_error ge("cannot create subprocess");
    Alert(logger()) << ge;
    closeSockets();
    process().reset();
    return false;
}

void ProcessHelper::setFds(int std_in, int std_out, int std_err) {
    std_in_ = std_in;
    std_out_ = std_out;
    std_err_ = std_err;
}
