// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/RRDCacheDThread.h"

#include <sys/socket.h>
#include <sys/un.h>

#include <compare>
#include <cstdlib>
#include <sstream>
#include <utility>

#include "livestatus/Logger.h"
#include "livestatus/POSIXUtils.h"
#include "livestatus/StringUtils.h"
#include "livestatus/global_counters.h"

using namespace std::chrono_literals;
namespace rp = rrdcached::protocol;

rp::Update::Update(std::string rrd_path,
                   std::chrono::system_clock::time_point timestamp,
                   std::vector<Metric> metrics)
    : rrd_path_{std::move(rrd_path)}
    , timestamp_{timestamp}
    , metrics_{std::move(metrics)} {}

std::string rp::Update::string() const {
    std::string command =
        "UPDATE " + rrd_path_ + " " +
        std::to_string(std::chrono::system_clock::to_time_t(timestamp_));
    for (const auto &metric : metrics_) {
        command += ":" + (metric.value().empty() ? "U" : metric.value());
    }
    command += "\n";
    return command;
}

RRDCacheDThread::RRDCacheDThread(std::chrono::microseconds retry_interval,
                                 Counter overflow_counter, Counter queue_usage,
                                 Counter bytes_sent_counter, Logger *logger)
    : DataExportThread{"rrdcached",
                       overflow_counter,
                       queue_usage,
                       {logger,
                        [this](std::ostream &os) {
                            auto sp = socket_path();
                            if (!sp.empty()) {
                                os << "[rrdcached at " << sp << "] ";
                            }
                        }}}
    , bytes_sent_counter_{bytes_sent_counter}
    , retry_interval_(retry_interval) {}

std::filesystem::path RRDCacheDThread::socket_path() const {
    const std::lock_guard lg(socket_path_mutex_);
    return socket_path_;
}

void RRDCacheDThread::socket_path(const std::filesystem::path &value) {
    const std::lock_guard lg(socket_path_mutex_);
    socket_path_ = value;
}

void RRDCacheDThread::handle(const rp::Update &message) {
    if (!socket_ || !socket_->isConnected()) {
        tryReconnect();
    }
    if (!socket_ || !socket_->isConnected()) {
        return;
    }
    const std::string serialized = message.string();
    if (socket_->write(serialized, 200ms) == -1) {
        throw generic_error("cannot write update");
    }
    counterIncrementBy(bytes_sent_counter_, serialized.size());
    receiveReply();
}

void RRDCacheDThread::handle(const rp::Reload &message) {
    auto path = message.socket_path();
    if (socket_path() != path) {
        close();
        socket_path(path);
    } else {
        Notice(logger()) << "settings unchanged";
    }
}

void RRDCacheDThread::tryReconnect() {
    auto now = std::chrono::steady_clock::now();
    if (now - last_connection_attempt_ <= retry_interval_) {
        return;
    }
    last_connection_attempt_ = now;

    if (!socket_) {
        socket_ = std::make_unique<RRDUDSSocket>(socket_path(), logger());
    }
    socket_->connect();
}

void RRDCacheDThread::receiveReply() {
    // rrdcached replys as described in
    // http://rrdtool.mirror.graemef.net/doc/rrdcached.en.html
    const std::string reply = socket_->readLine();
    const int retcode = atoi(reply.c_str());
    // Ignore an rrd message that appears if values are updated too often.
    // This can e.g. happen, if a check is rescheduled twice within the same
    // second. CMC is faster than allowed by the police ;-)
    if (retcode != 0 &&
        !reply.starts_with("-1 illegal attempt to update using time")) {
        Warning(logger()) << "[log] " << reply;
        const int continuation_lines = retcode > 0 ? retcode : 0;
        for (int i = 0; i < continuation_lines; ++i) {
            Warning(logger()) << "[log] " << socket_->readLine();
        }
    }
}

RRDUDSSocket::RRDUDSSocket(std::filesystem::path path, Logger *logger)
    : path_{std::move(path)}, logger_{logger} {}

RRDUDSSocket::~RRDUDSSocket() { close(); }

void RRDUDSSocket::connect() {
    const int sock = ::socket(PF_UNIX, SOCK_STREAM, 0);
    if (sock == -1) {
        throw generic_error{"cannot create socket"};
    }

    struct sockaddr_un sockaddr {
        .sun_family = AF_UNIX, .sun_path = ""
    };
    path_.string().copy(&sockaddr.sun_path[0], sizeof(sockaddr.sun_path) - 1);
    sockaddr.sun_path[sizeof(sockaddr.sun_path) - 1] = '\0';
    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
    if (::connect(sock, reinterpret_cast<const struct sockaddr *>(&sockaddr),
                  sizeof(sockaddr)) == -1) {
        throw generic_error{"cannot connect"};
    }

    Notice(logger()) << "successfully connected";
    socket_ = sock;
    file_ = fdopen(sock, "r");
}

std::string RRDUDSSocket::readLine() const {
    char answer[512];
    if (fgets(answer, sizeof(answer), file_) == nullptr) {
        throw generic_error("cannot read reply");
    }
    return mk::rstrip(answer);
}

ssize_t RRDUDSSocket::write(const std::string &text,
                            std::chrono::milliseconds timeout) const {
    if (!isConnected()) {
        return 0;
    }
    return writeWithTimeout(socket_, text, timeout);
}

void RRDUDSSocket::close() {
    if (!isConnected()) {
        return;
    }
    Notice(logger()) << "closing connection";
    (void)::fclose(file_);
    socket_ = -1;
    file_ = nullptr;
}
