// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/RRDHelper.h"

#include <unistd.h>

#include <algorithm>
#include <cerrno>
#include <compare>
#include <filesystem>
#include <functional>
#include <optional>
#include <unordered_map>
#include <unordered_set>
#include <utility>

#include "cmc/GlobalConfig.h"
#include "cmc/Host.h"
#include "cmc/Object.h"
#include "cmc/Process.h"
#include "cmc/RRDFileLocator.h"
#include "cmc/RRDInfo.h"
#include "cmc/Service.h"
#include "cmc/State.h"
#include "cmc/World.h"
#include "livestatus/Logger.h"
#include "livestatus/Metric.h"
#include "livestatus/POSIXUtils.h"
#include "livestatus/PerformanceData.h"
#include "livestatus/PnpUtils.h"
#include "livestatus/Poller.h"

using namespace std::chrono_literals;
using namespace std::string_literals;
using namespace std::string_view_literals;

// TODO(sp): Just replace all this manual fiddling with pointers,
// offsets, etc. with vector.

RRDHelper::RRDHelper(std::string name, Counter overflow_counter,
                     Counter queue_usage, Counter bytes_sent_counter,
                     Logger *logger)
    : ProcessHelper{std::move(name),
                    "",
                    {"--create-rrd"s, "--keepalive"s},
                    logger}
    , _xml(logger)
    , _rrdcached(15s, overflow_counter, queue_usage, bytes_sent_counter,
                 logger) {
    _buffer[0] = 0;
}

RRDHelper::~RRDHelper() { do_stop(); }

void RRDHelper::start_() {
    _rrdcached.start();
    _xml.start();
    if (spawnConnectedSubprocess()) {
        _start = std::chrono::steady_clock::now();
    }
}

void RRDHelper::stop_() { do_stop(); }

void RRDHelper::setUpPoller_(Poller &poller) const {
    if (stdOut() >= 0) {
        poller.addFileDescriptor(stdOut(), PollEvents::in);
    }
    if (stdErr() >= 0) {
        poller.addFileDescriptor(stdErr(), PollEvents::in);
    }
}

void RRDHelper::communicate_(const Poller &poller) {
    if (stdOut() >= 0 && poller.isFileDescriptorSet(stdOut(), PollEvents::in)) {
        handleStdOut();
    }
    if (stdErr() >= 0 && poller.isFileDescriptorSet(stdErr(), PollEvents::in)) {
        handleStdErr();
    }
}

bool RRDHelper::updateConfig(const World *world) {
    _world = world;
    const auto global_config = world->config_.global_config_;
    // Config changes in _rrdcached and _xml may be safely ignored.
    _rrdcached.reload(
        rrdcached::protocol::Reload{global_config.rrd_.rrdcached_socket_});
    _xml.updateConfig(global_config.rrd_);
    updateConfig(global_config.general_, global_config.rrd_);

    // reload always due to impossibility to predict
    // which part of config affects RRD functionality
    return true;
}

bool RRDHelper::updateConfig(const GeneralConfig &general,
                             const RRDConfig &rrd) {
    // NOTE: We could even avoid having a helper + a thread when we
    // definitely have no rrdcached running, but we keep things simple
    // and do not optimize for this edge case.
    pnp4nagios_enabled_ = rrd.pnp4nagios_enabled_;
    return updatePath(general.cmk_executable_);
}

void RRDHelper::reload() {
    Notice(logger()) << "reload configuration";
    if (writeWithTimeout(stdIn(), "*\n", 3s) == -1) {
        // NOTE: Silently ignoring the error is not an option here: A helper
        // running with an outdated configuration can actually harm the RRD
        // data!
        const generic_error ge{"cannot send reload command to RRD creator"};
        Warning(logger()) << ge;
        restart();
    }
}

void RRDHelper::handleStdOut() { handleIO(stdOut()); }

void RRDHelper::handleStdErr() { handleIO(stdErr()); }

// Note: The history metrics for ALL objects must be unlocked on entry.
// TODO (sk): remove C-style processing
// NOTE: The protocol between the CMC and the RRD helper is a total chaos:
//   * The helper receives commands via stdin (OK)
//   * The helper sends CREATED replies via stdout (OK)
//   * The helper logs some stuff via stdout (not OK)
//   * The helper logs other stuff via stderr (not OK)
// In a nutshell: This should really be disentangled on both sides.
void RRDHelper::handleIO(int fd) {
    if (_buffer_size == sizeof(_buffer)) {
        Error(logger()) << "too many bytes without newline, clearing buffer.";
        _buffer_size = 0;
    }
    auto bytes_read =
        ::read(fd, _buffer + _buffer_size, sizeof(_buffer) - _buffer_size);
    if (bytes_read > 0) {
        _buffer_size += bytes_read;
        std::string_view content{_buffer, _buffer_size};
        for (size_t newline_pos{}; (newline_pos = content.find_first_of(
                                        "\n"sv)) != std::string_view::npos;) {
            const std::string_view line{content.data(), newline_pos};
            // The helper sends us one line starting with "CREATED " for each
            // RRD it has created. That way we can keep track of existing RRDs
            // of that object without rescanning the directory again and again.
            constexpr auto created = "CREATED "sv;
            if (line.starts_with(created)) {
                updateMetrics(line.substr(created.size()));
            } else {
                Warning(logger()) << "[log] " << line;
            }
            content = content.substr(newline_pos + 1);
        }
        if (content != std::string_view{_buffer, _buffer_size}) {
            std::move(content.data(), content.data() + content.size(), _buffer);
            _buffer_size = content.size();
        }
    } else {
        Warning(logger()) << "closed error pipe, terminating";
        stop();
    }
}

void RRDHelper::do_stop() {
    _rrdcached.stop();
    _xml.stop();
    despawn();
}

namespace {
// TODO (sk): Move to the std::string RRDHelperChannel::encodeMetric(....);
std::string BuildCommand(std::string_view specname, RRDFormat rrd_format,
                         const std::vector<Metric> &metrics) {
    // Examples for helper command:
    //    "cmc_single;serv0123;CPU_load;load1;12.3;load5;11.8;load15;88.9"
    // In multiple format there is exactly one variable definition:
    //    "pnp_multiple;serv0123;CPU_load;load1;12.3"
    std::string command;
    command.reserve(512);
    command +=
        rrd_format == RRDFormat::multiple ? "pnp_multiple;" : "cmc_single;";
    command += specname;

    // NOTE: Probably we should find effective method to build the string here.
    for (const auto &metric : metrics) {
        command += ";";
        command += metric.mangled_name().string();
        command += ";";
        command += metric.value();
    }
    return command;
}
}  // namespace

// Note: object's history metrics are locked on entry.
void RRDHelper::create(const Object *object, RRDFormat rrd_format,
                       const std::vector<Metric> &metrics) {
    // Restart helper if it has previously died and some time has elapsed
    const auto &p1 = process();
    if (p1 && !p1->valid() && std::chrono::steady_clock::now() - _start > 15s) {
        start();
    }
    const auto &p2 = process();
    if (p2 && p2->valid()) {
        auto command = BuildCommand(object->specname(), rrd_format, metrics);
        Informational(logger()) << "creating new RRD file for " << *object
                                << ", command: " << command;
        command += "\n";
        // TODO(sp): The synchronous write is horrible, and the RRD helper is
        // buggy, too: It doesn't queue the commands, it just executes them
        // synchronosly, which can starve the CMC.
        if (writeWithTimeout(stdIn(), command, 20s) == -1) {
            if (errno == ETIMEDOUT) {
                const generic_error ge{
                    "RRD helper is too slow, dropping creation/rewrite"};
                Warning(logger()) << *object << ": " << ge;
            } else {
                const generic_error ge{"cannot send command to RRD helper"};
                Warning(logger()) << ge;
                Warning(logger()) << *object << ": " << ge;
                restart();
            }
        }
    }
}

void RRDHelper::processPerfData(const PerformanceData &perf_data,
                                const Object *object) {
    if (perf_data.empty() || !pnp4nagios_enabled_) {
        // NOTE: If the user explicitly switched off the rrdcached via "omd
        // config set PNP4NAGIOS off", we avoid creating any RRDs.
        return;
    }
    auto info = object->rrdInfo();
    if (info.format == RRDFormat::single) {
        processSingle(perf_data, object, info);
    } else {
        processMultiple(perf_data, object, info);
    }
}

void RRDHelper::processSingle(const PerformanceData &perf_data,
                              const Object *object, const RRDInfo &info) {
    // If the RRD does not yet exist or if not yet all metrics exist then we
    // need to (re-) create the RRD. For this operation we need to look into
    // the new metrics if there is one that we haven't seen yet. Note that
    // we need to re-order the metrics into the order given by the existing
    // RRD.
    std::unordered_map<Metric::MangledName, Metric> metric_for_varname;
    for (const auto &metric : perf_data) {
        metric_for_varname.emplace(metric.mangled_name(), metric);
    }

    std::vector<Metric> ordered_metrics;
    for (const auto &mangled_varname : info.names) {
        auto it = metric_for_varname.find(mangled_varname);
        std::string value;
        // Note: A metric we've seen before could be missing now.
        if (it != metric_for_varname.end()) {
            value = it->second.value();
            metric_for_varname.erase(it);
        }
        // NOTE: Going back from a mangled name to an umangled one is a
        // hack, it assumes that mangling is idempotent (which it is for our
        // case, but nevertheless...)
        ordered_metrics.emplace_back(mangled_varname.string(), value, "", "",
                                     "", "", "");
    }

    for (const auto &entry : metric_for_varname) {
        ordered_metrics.push_back(entry.second);
    }

    if (!metric_for_varname.empty()) {
        // If we have at least 1 new metric, then we need to (re-)create the
        // RRD
        create(object, RRDFormat::single, ordered_metrics);
    } else {
        // TODO(sp): This is a bit ugly, because our underlying protocol shines
        // through, so we have to pass a dummy metric name. :-/
        _rrdcached.update(rrdcached::protocol::Update{
            _world->rrdFileLocatorSingle()->path(object,
                                                 Metric::MangledName("")),
            object->state()->last_check_, ordered_metrics});
    }
}

void RRDHelper::processMultiple(const PerformanceData &perf_data,
                                const Object *object, const RRDInfo &info) {
    _xml.process(perf_data, object);
    const std::unordered_set<Metric::MangledName> known{info.names.begin(),
                                                        info.names.end()};
    for (const auto &metric : perf_data) {
        if (known.contains(metric.mangled_name())) {
            _rrdcached.update(rrdcached::protocol::Update{
                _world->rrdFileLocatorMultiple()->path(object,
                                                       metric.mangled_name()),
                object->state()->last_check_,
                {metric}});
        } else {
            create(object, RRDFormat::multiple, {metric});
        }
    }
}

// Note: The history metrics for ALL objects must be unlocked on entry.
// Examples for response:
//    cmc_single myhost;myservice;var1;var2;var3
//    pnp_multiple myhost;_HOST_;var
void RRDHelper::updateMetrics(std::string_view response) {
    auto next = [&response](char sep) {
        auto field = response.substr(0, response.find(sep));
        response.remove_prefix(std::min(response.size(), field.size() + 1));
        return std::string{field};
    };

    next(' ');  // format
    auto host_name = next(';');
    auto service_description = next(';');
    auto is_host = service_description == dummy_service_description();
    auto *object = is_host ? static_cast<Object *>(_world->getHost(host_name))
                           : static_cast<Object *>(_world->getService(
                                 host_name, service_description));
    if (object == nullptr) {
        Notice(logger()) << "ignoring reply for object "
                         << (is_host ? host_name
                                     : (host_name + ";" + service_description));
        return;
    }
    object->forgetRRDInfo();
    while (!response.empty()) {
        Informational(logger())
            << "created RRD for " << *object << ", metric " << next(';');
    }
}
