// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/RRDInfoCache.h"

#include <cstddef>
#include <filesystem>
#include <fstream>
#include <string>
#include <utility>
#include <vector>

#include "cmc/Object.h"
#include "cmc/RRDFileLocator.h"
#include "cmc/World.h"
#include "livestatus/Logger.h"
#include "livestatus/StringUtils.h"

// Note: We need to determine the physical path to RRD files to avoid a bug in
// rrdcached when both /omd and /opt/omd are being used.
RRDInfoCache::RRDInfoCache(const World &world, Logger *logger)
    : _world(world), logger_{logger} {}

RRDInfo RRDInfoCache::rrdInfo(const Object *object) {
    const std::lock_guard<std::mutex> lg(_mutex);
    const auto &[it, cache_miss] = _cache.insert({object, RRDInfo()});
    RRDInfo &info = it->second;
    if (cache_miss) {
        info.format = object->currentRrdFormat();
        info.names = info.format == RRDFormat::single
                         ? getNamesFromInfo(object)
                         : getNamesFromRRDs(object);
        if (logger_->isLoggable(LogLevel::debug)) {
            Debug(logger_) << *object << " uses " << info;
        }
    }
    return info;
}

void RRDInfoCache::forgetRRDInfo(const Object *object) {
    const std::lock_guard<std::mutex> lg(_mutex);
    Informational(logger_) << "dropping cached information about " << *object;
    _cache.erase(object);
}

MetricLocation RRDInfoCache::metricLocation(const Object *object,
                                            const Metric::MangledName &name) {
    auto info = rrdInfo(object);
    if (info.format == RRDFormat::single) {
        size_t i = 0;
        for (const auto &n : info.names) {
            ++i;
            if (n == name) {
                return {_world.rrdFileLocatorSingle()->path(object, n),
                        std::to_string(i)};
            }
        }
        return {};
    }
    return info.names.empty()
               ? MetricLocation{}
               : MetricLocation{
                     _world.rrdFileLocatorMultiple()->path(object, name), "1"};
}

Metric::Names RRDInfoCache::getNamesFromInfo(const Object *object) {
    Metric::Names names;
    const std::filesystem::path path =
        _world.rrdFileLocatorSingle()->metadata(object);
    Informational(logger_) << "reading metrics of " << *object << " from "
                           << path;
    std::ifstream is(path);
    if (!is) {
        const generic_error ge("cannot open info file " + path.string());
        Informational(logger_) << ge;
        return names;
    }

    // Example for info file:
    // ==> Filesystem__boot_efi.info <==
    // HOST heute
    // SERVICE Filesystem /boot/efi
    // METRICS trend;growth;fs_size;_boot_efi
    // Additional keywords may be introduced later.

    std::string line;
    while (std::getline(is, line)) {
        if (line.starts_with("METRICS ")) {
            for (const auto &varname :
                 mk::split(mk::rstrip(line.substr(8)), ';')) {
                // NOTE: The .info file contains already mangled names.
                names.emplace_back(Metric::Name(varname));
            }
            break;
        }
    }
    is.close();

    if (names.empty()) {
        Warning(logger_) << "no metrics in " << path;
    }
    return names;
}

Metric::Names RRDInfoCache::getNamesFromRRDs(const Object *object) {
    // TODO(ml): Here again a strong case of feature envy!
    //           At least now, the BL is testable.
    return scan_rrd(_world.rrdFileLocatorMultiple()->directory(object),
                    object->serviceDescription(), logger_);
}

void RRDInfoCache::migrate(const RRDInfoCache &from) {
    const std::lock_guard<std::mutex> lg(_mutex);
    size_t vanished = 0;
    size_t changed = 0;
    size_t migrated = 0;
    for (const auto &entry : from._cache) {
        const auto *object = entry.first;
        Object *new_object = _world.correspondingObject(object);
        if (new_object == nullptr) {
            Informational(logger_)
                << *object << " vanished, dropping cached RRD info";
            vanished++;
        } else if (object->_rrd_format != new_object->_rrd_format) {
            Informational(logger_) << "RRD format for " << *object
                                   << " changed, dropping cached RRD info";
            changed++;
        } else {
            _cache[new_object] = entry.second;
            migrated++;
        }
    }
    Notice(logger_) << migrated << " cached RRD infos migrated, "  //
                    << vanished << " objects vanished, "           //
                    << changed << " objects changed their RRD format";
}

std::ostream &operator<<(std::ostream &os, const RRDInfo &info) {
    os << (info.format == RRDFormat::single ? "single RRD" : "multiple RRDS");
    if (info.names.empty()) {
        os << ", no metrics";
    } else {
        os << ", metrics:";
        for (const auto &name : info.names) {
            os << " " << name.string();
        }
    }
    return os;
}
