// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/RealTimeCheck.h"

#include <arpa/inet.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>

#include <chrono>
#include <optional>
#include <sstream>
#include <string_view>
#include <utility>
#include <vector>

#include "cmc/CheckHelperPool.h"
#include "cmc/GlobalConfig.h"
#include "cmc/Host.h"
#include "cmc/MessagePayload.h"
#include "cmc/World.h"
#include "livestatus/Logger.h"
#include "livestatus/Poller.h"

RealTimeCheck::RealTimeCheck(std::string name, CheckHelperPool &pool,
                             Logger *logger)
    : AbstractHelper{std::move(name),
                     ContextLogger{logger,
                                   [this](std::ostream &os) {
                                       os << "[real-time socket on port "
                                          << _port << "] ";
                                   }}}
    , _pool{pool} {}

RealTimeCheck::~RealTimeCheck() { do_stop(); }

void RealTimeCheck::start_() {
    if (_port == 0) {
        return;
    }

    // TODO (ml): Refactor to function
    // legacy/system API better to place in separate function/wrappers
    // and close in namespace. Nobody wants to know how we make socket
    // cmc::tools::CreateSocket(Sock::Type::datagram, Sock::Mode::non_blocking)
    // or with class wrapper
    // cmc::tools::Sock::Create(Sock::Type::datagram, Sock::Mode::non_blocking)
    // This is light to test, virtually null errors possible, and easy to read
    const int sock = ::socket(AF_INET, SOCK_DGRAM, 0);
    if (sock == -1) {
        const generic_error ge("cannot create socket");
        Warning(logger()) << ge;
        return;
    }

    sockaddr_in sa{};
    sa.sin_family = AF_INET;
    sa.sin_port = htons(_port);
    sa.sin_addr.s_addr = INADDR_ANY;

    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
    if (::bind(sock, reinterpret_cast<const struct sockaddr *>(&sa),
               sizeof(sa)) == -1) {
        const generic_error ge("cannot bind socket");
        Warning(logger()) << ge;
        ::close(sock);
        return;
    }

    if (::fcntl(sock, F_SETFL, O_NONBLOCK) == -1) {
        const generic_error ge("cannot make socket non-blocking");
        Alert(logger()) << ge;
        ::close(sock);
        return;
    }

    _socket = sock;
    Notice(logger()) << "successfully opened";
}

void RealTimeCheck::stop_() { do_stop(); }

void RealTimeCheck::setUpPoller_(Poller &poller) const {
    if (_socket >= 0) {
        poller.addFileDescriptor(_socket, PollEvents::in);
    }
}

void RealTimeCheck::communicate_(const Poller &poller) {
    if (_socket >= 0 && poller.isFileDescriptorSet(_socket, PollEvents::in)) {
        receive();
    }
}

void RealTimeCheck::do_stop() {
    if (_socket == -1) {
        return;
    }
    Notice(logger()) << "closing";
    ::close(_socket);
    _socket = -1;
}

bool RealTimeCheck::updateConfig(const World *world) {
    _world = world;  // for World::getHostsByIPv4Address()
    return updateConfig(world->config_.global_config_.checking_);
}

bool RealTimeCheck::updateConfig(const CheckingConfig &checking) {
    bool changed = false;
    if (_port != checking.real_time_port_) {
        _port = checking.real_time_port_;
        changed = true;
    }
    return changed;
}

void RealTimeCheck::receive() {
    if (_world == nullptr) {
        return;
    }
    std::vector<char> buffer(65536);
    struct sockaddr_in sa;
    socklen_t len = sizeof(sa);
    auto bytes_received = ::recvfrom(
        _socket, buffer.data(), buffer.size(), 0,
        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
        reinterpret_cast<struct sockaddr *>(&sa), &len);
    if (bytes_received == -1) {
        const generic_error ge{"could not receive message"};
        Warning(logger()) << ge;
        return;
    }

    auto now = std::chrono::system_clock::now();
    const std::string_view sv{buffer.data(),
                              static_cast<size_t>(bytes_received)};
    for (auto *host : _world->getHostsByIPv4Address(sa.sin_addr.s_addr)) {
        host->_real_time_payload = MessagePayload{sv};
        if (auto t = _pool.insert({host, now})) {
            host->setNextCheck(*t, _pool.logger());
        }
    }
}
