// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/SerialToken.h"

#include <ostream>
#include <utility>

#include "livestatus/Logger.h"

SerialToken::SerialToken(config_serial_t value, SerialTokenFactory *owner,
                         std::string creator)
    : value_{value}, owner_{owner}, creator_{std::move(creator)} {}

SerialToken::~SerialToken() {
    if (owner_ != nullptr) {
        owner_->release(this);
    }
}

SerialToken::SerialToken(SerialToken &&rhs) noexcept
    : value_{rhs.value_}
    , owner_{rhs.owner_}
    , creator_{std::move(rhs.creator_)} {
    rhs.value_ = 0;
    rhs.owner_ = nullptr;
}

SerialToken &SerialToken::operator=(SerialToken &&rhs) noexcept {
    if (this != &rhs) {
        if (owner_ != nullptr) {
            owner_->release(this);
        }
        value_ = rhs.value_;
        owner_ = rhs.owner_;
        creator_ = rhs.creator_;
        rhs.value_ = 0;
        rhs.owner_ = nullptr;
        rhs.creator_ = "";
    }
    return *this;
}

config_serial_t SerialToken::value() const { return value_; }

std::ostream &operator<<(std::ostream &os, const SerialToken &s) {
    return os << "SerialToken{" << s.creator_ << "," << s.value_ << "}";
}

SerialToken SerialTokenFactory::create(const std::string &creator,
                                       config_serial_t serial) {
    const std::scoped_lock l(lock_);
    const auto &[it, initial] = tokens_.insert({serial, 0});
    ++it->second;
    SerialToken token{serial, this, creator};
    if (logger_->isLoggable(LogLevel::debug)) {
        Debug(logger_) << "obtained " << (initial ? "initial " : "") << token
                       << " => " << *this;
    }
    return token;
}

void SerialTokenFactory::release(SerialToken *token) {
    const std::scoped_lock l(lock_);
    auto it = tokens_.find(token->value());
    if (it == tokens_.end()) {
        Alert(logger_) << "unknown reference " << *token << " into " << *this;
        return;
    }
    auto last = --it->second == 0;
    if (last) {
        tokens_.erase(it);
    }
    if (logger_->isLoggable(LogLevel::debug)) {
        Debug(logger_) << "released " << (last ? "last " : "") << *token
                       << " => " << *this;
    }
}

std::unordered_set<config_serial_t> SerialTokenFactory::used() const {
    const std::scoped_lock l(lock_);
    std::unordered_set<config_serial_t> result;
    for (const auto &[serial, count] : tokens_) {
        result.insert(serial);
    }
    return result;
}

// NOTE: We avoid recursive locks, so this should be called with s.lock_ locked.
std::ostream &operator<<(std::ostream &os, const SerialTokenFactory &s) {
    os << "SerialTokenFactory{";
    auto emit_separator{false};
    for (const auto &[k, v] : s.tokens_) {
        os << (emit_separator ? ", " : "") << k << ":" << v;
        emit_separator = true;
    }
    return os << "}";
}
