// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/State.h"

#include "cmc/CheckResult.h"
#include "cmc/Histcache.h"
#include "cmc/Host.h"
#include "cmc/Object.h"
#include "cmc/Service.h"
#include "cmc/cmc.h"

enum class HostState;
enum class ServiceState;

bool State::operator==(const State &other) const {
    return acknowledgement_type_ == other.acknowledgement_type_ &&
           check_result_ == other.check_result_ &&
           downtime_ == other.downtime_ && enabling_ == other.enabling_ &&
           flappiness_ == other.flappiness_ &&
           hard_state_ == other.hard_state_ &&
           notification_ == other.notification_ &&
           current_attempt_ == other.current_attempt_ &&
           has_been_checked_ == other.has_been_checked_ &&
           execution_time_ == other.execution_time_ &&
           latency_ == other.latency_ && last_check_ == other.last_check_ &&
           last_check_started_ == other.last_check_started_ &&
           last_state_change_ == other.last_state_change_ &&
           last_time_state_ == other.last_time_state_ &&
           next_check_ == other.next_check_ &&
           current_state_ == other.current_state_ &&
           initial_state_ == other.initial_state_ &&
           last_state_ == other.last_state_ && state_type_ == other.state_type_;
}

// NOTE: cr._object->state() == this
void State::update(
    std::chrono::system_clock::time_point tp_now, const CheckResult &cr,
    const std::function<void(int state, const std::string &output,
                             const std::string &long_output)> &register_alert,
    bool enable_flap_detection) {
    Object *object = cr._object;
    const int state = cr._object->isHost()
                          ? static_cast<int>(object->host()->remap(
                                static_cast<HostState>(cr._state)))
                          : static_cast<int>(object->service()->remap(
                                static_cast<ServiceState>(cr._state)));
    const std::string output = object->remapOutput(cr._output);
    const std::string long_output = cr._long_output;

    last_check_ = tp_now;
    last_time_state_[state] = tp_now;
    execution_time_ = tp_now - last_check_started_;
    check_result_.output_ = output;
    check_result_.long_output_ = long_output;
    check_result_.perf_data_ = cr._perf_data;
    is_helper_running_ = false;
    check_result_.cached_at_ = cr._cached_at;
    check_result_.cache_interval_ = cr._cache_interval;

    if (!has_been_checked_) {  // First check result. Log initial state
        has_been_checked_ = true;
        register_alert(state, output, long_output);
    }

    if (state != static_cast<int>(current_state_) ||
        (last_state_change_ == std::chrono::system_clock::time_point())) {
        last_state_change_ = tp_now;
    }

    const bool is_last_state_ok = object->isCurrentStateOK();
    last_state_ = current_state_;
    current_state_ = state;

    if (object->isCurrentStateOK()) {
        current_attempt_ = 1;
        state_type_ = StateType::hard;
    } else {
        if (is_last_state_ok) {
            // We just switched from OK to non-OK.
            current_attempt_ = 1;
        } else {
            current_attempt_++;
            // Stay on the last attempt for any  exceeding non-OK check result.
            if (current_attempt_ > object->_max_check_attempts) {
                current_attempt_ = object->_max_check_attempts;
            }
        }
        state_type_ = (current_attempt_ >= object->_max_check_attempts)
                          ? StateType::hard
                          : StateType::soft;
    }

    // Now we know if the state is hard or soft. Now let's see if we have
    // a hard state *change*, i.e. one that leads to an alert.
    checkForHardStateChange();

    // If the flap detection has been turned off meanwhile, remove all flapping
    // information from that object
    if (!object->isEnabled(Enabling::Key::flap_detection) ||
        !enable_flap_detection) {
        flappiness_.reset();
    }
}

void State::checkForHardStateChange() {
    hard_state_.changed_ =
        state_type_ == StateType::hard && hard_state_.last_ != current_state_;
    if (hard_state_.changed_) {
        hard_state_.previous_ = hard_state_.last_;  // for RBN
        hard_state_.last_ = current_state_;
        hard_state_.last_time_ = last_check_;
    }
}

void State::Flappiness::reset() {
    value_ = 0.0;
    is_flapping_ = false;
}

bool State::Flappiness::flap(double high_flap_threshold) {
    value_ += 1.0;
    if (!is_flapping_ && value_ >= high_flap_threshold) {
        is_flapping_ = true;
        return true;
    }
    return false;
}

bool State::Flappiness::unflap(double low_flap_threshold, double flap_decay) {
    value_ *= (1.0 - flap_decay);
    if (is_flapping_ && value_ < low_flap_threshold) {
        is_flapping_ = false;
        return true;
    }
    return false;
}

Histcache *State::histcache() {
    if (_histcache == nullptr) {
        _histcache = new Histcache();
    }
    return _histcache;
}
