// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/StateHistoryThread.h"

#include <algorithm>
#include <cctype>
#include <charconv>
#include <compare>
#include <deque>
#include <fstream>
#include <functional>
#include <iomanip>
#include <memory>
#include <new>
#include <optional>
#include <ratio>
#include <shared_mutex>
#include <string_view>
#include <system_error>
#include <utility>
#include <vector>

#include "cmc/CmcCore.h"
#include "cmc/Core.h"
#include "cmc/GlobalConfig.h"
#include "cmc/Histcache.h"
#include "cmc/HistcacheEntry.h"
#include "cmc/Host.h"
#include "cmc/Object.h"
#include "cmc/Service.h"
#include "cmc/State.h"
#include "cmc/TableCachedStatehist.h"
#include "cmc/World.h"
#include "cmc/cmc.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/LogCache.h"
#include "livestatus/LogEntry.h"
#include "livestatus/Logger.h"

using namespace std::chrono_literals;
using namespace std::string_view_literals;

StateHistoryThread::StateHistoryThread(TableCachedStatehist *table)
    : _table(table)
    , _should_terminate(false)
    , _thread_finished(false)
    , _cached_strings(0)
    , _parsed_messages(0)
    , _parsed_bytes_total(0) {}

void StateHistoryThread::run(Core *core) {
    _core = core;
    _parsed_messages = 0;
    _parsed_bytes_total = 0;
    _table->_first_startstop_entry_is_running = false;
    _table->_parsed_since = std::chrono::system_clock::now();

    // Make sure that our world is not being deleted by a config reload!
    const std::shared_lock<std::shared_mutex> sl(_core->_world_lock);
    // set thread-local pointer to world being used
    g_live_world = _core->_world;

    auto horizon = _core->config()->history_.cache_horizon_;
    if (!horizon) {
        return;
    }
    auto before = std::chrono::steady_clock::now();
    auto until = std::chrono::system_clock::now();
    auto since = until - *horizon;
    Notice(_core->_logger_core)
        << "building state history cache for the time period from "
        << FormattedTimePoint(since) << " to " << FormattedTimePoint(until)
        << " (" << mk::ticks<std::chrono::days>(*horizon) << " days)";
    _table->_start_of_cache = since;  // remember the starting time of our cache

    CmcCore icore{_core};
    auto [paths, first_skipped] = LogCache{&icore}.pathsSince(since);
    if (first_skipped) {
        // pragma because of a clang-tidy bug
        Notice(_core->_logger_core)
            << "skipping "
            << *first_skipped  // NOLINT(bugprone-unchecked-optional-access)
            << " and older log files, they are beyond the caching horizon";
    }
    if (paths.empty()) {
        return;  // No logfiles found
    }

    // Now process backwards until horizon is being reached.
    for (const auto &path : paths) {
        if (shouldTerminate()) {
            Notice(_core->_logger_core)
                << "stopping construction of state history cache";
            break;
        }
        processLogfile(path);
    }
    finalizeCaches();
    logStatistics(std::chrono::steady_clock::now() - before,
                  until - _table->_parsed_since, paths.size());
    _table->_parsed_since = _table->_start_of_cache;  // everying parsed
    setThreadFinished(true);
}

void StateHistoryThread::logStatistics(std::chrono::nanoseconds processing_time,
                                       std::chrono::nanoseconds history_extent,
                                       size_t num_log_files) {
    if (!shouldTerminate()) {
        auto elapsed_secs =
            mk::ticks<std::chrono::duration<double>>(processing_time);
        auto days = mk::ticks<std::chrono::duration<double, std::ratio<86400>>>(
            history_extent);
        size_t num_entries{0};
        for (const auto &[specname, histcache] : _histcaches) {
            num_entries += histcache->size();
        }

        Notice(_core->_logger_core)
            << ",-Cache for statehist------------------------------------------------------------.";

        Notice(_core->_logger_core)
            << "|                                                                                |";

        Notice(_core->_logger_core)
            << "|         parsed                speed                    cached                  |";

        Notice(_core->_logger_core)
            << "|         -----------           -------------            ----------------------  |";

        Notice(_core->_logger_core)
            << std::fixed << std::setprecision(2) << "|" << std::setw(7)
            << num_log_files << "  Logfiles" << std::setw(12)
            << (num_log_files / elapsed_secs) << "  Logfiles/s" << std::setw(13)
            << _histcaches.size() << "  hosts/services          |";

        Notice(_core->_logger_core)
            << std::fixed << std::setprecision(3) << "|" << std::setw(7)
            << (_parsed_bytes_total / (1024.0 * 1024.0 * 1024.0))
            << "  GB of data" << std::setw(10)
            << (_parsed_bytes_total / (1024.0 * 1024.0) / elapsed_secs)
            << "  MB/s" << std::setw(19)
            << (num_entries + _table->_core_startstops.size())
            << "  host/service events     |";

        Notice(_core->_logger_core)
            << std::fixed << std::setprecision(3) << "|" << std::setw(7)
            << (_parsed_messages / 1000000.0) << "  Mio messages"
            << std::setw(8) << (_parsed_messages / 1000000.0 / elapsed_secs)
            << "  Mio messages/s" << std::setw(9)
            << _table->_core_startstops.size() << "  core starts/stops       |";

        Notice(_core->_logger_core)
            << std::fixed << std::setprecision(1) << "|" << std::setw(7) << days
            << "  days of history" << std::setprecision(2) << std::setw(30)
            << (static_cast<double>(num_entries) / _histcaches.size())
            << "  entries per host/serv.  |";

        Notice(_core->_logger_core)
            << std::fixed << std::setprecision(2) << "|" << std::setw(54)
            << (static_cast<double>(num_entries) / std::max(1.0, days))
            << "  entries per day         |";

        Notice(_core->_logger_core) << "|" << std::setw(54) << _cached_strings
                                    << "  strings                 |";

        Notice(_core->_logger_core)
            << std::fixed << std::setprecision(1) << "|" << std::setw(54)
            << _strings.size() << "  unique strings (" << std::setw(4)
            << (_strings.size() * 100.0 / _cached_strings) << "%)  |";

        Notice(_core->_logger_core)
            << std::fixed << std::setfill('0') << "|  " << std::setw(2)
            << (static_cast<int>(elapsed_secs) / 60) << ":" << std::setw(2)
            << (static_cast<int>(elapsed_secs) % 60)
            << " parsing time                                                            |";

        Notice(_core->_logger_core)
            << "'--------------------------------------------------------------------------------'";
    } else {
        Notice(_core->_logger_core) << "Aborted building cache for statehist.";
    }
}

namespace {
std::unique_ptr<std::vector<char>> readIntoBuffer(
    const std::filesystem::path &path, Logger *logger) {
    std::unique_ptr<std::vector<char>> result;
    std::ifstream is(path, std::ios::binary | std::ios::ate);
    if (!is) {
        const generic_error ge("cannot open logfile " + path.string());
        Warning(logger) << ge;
        return result;
    }

    auto end = is.tellg();
    is.seekg(0, std::ios::beg);
    if (!is) {
        const generic_error ge("cannot determine size of " + path.string());
        Warning(logger) << ge;
        return result;
    }
    auto size = end - is.tellg();

    try {
        result = std::make_unique<std::vector<char>>(size);
    } catch (std::bad_alloc &) {
        Warning(logger) << "cannot allocate " << size << " byte buffer for "
                        << path;
        return result;
    }

    is.read(result->data(), size);
    if (!is) {
        const generic_error ge("cannot open read " + std::to_string(size) +
                               " bytes from " + path.string());
        Warning(logger) << ge;
    }

    return result;
}
}  // namespace

void StateHistoryThread::processLogfile(const std::filesystem::path &path) {
    auto start_reading = std::chrono::steady_clock::now();
    auto buffer = readIntoBuffer(path, _core->_logger_core);
    if (!buffer) {
        return;
    }
    const std::string_view content{buffer->data(), buffer->size()};
    _parsed_bytes_total += content.size();
    const double reading = elapsed_ms_since(start_reading);

    auto start_proc = std::chrono::steady_clock::now();
    processLogfileBuffer(path, content);
    const double proc = elapsed_ms_since(start_proc);

    const double total = reading + proc;
    auto s = content.size() * 1000.0 / (1024 * 1024);
    Informational(_core->_logger_core)
        << std::fixed << std::setprecision(1)  //
        << "parsed log file " << path << ": " << content.size() << " bytes, "
        << "read in " << reading << " ms (" << s / reading << " MB/s), "
        << "processed in " << proc << " ms, (" << s / proc << " MB/s), "
        << "total " << total << " ms, (" << s / total << " MB/s)";
}

void StateHistoryThread::processLogfileBuffer(const std::filesystem::path &path,
                                              std::string_view content) {
    // go backwards through the content, line by line
    auto next = [&content]() {
        auto newline = content.rfind('\n');
        auto start = newline == std::string_view::npos ? 0 : newline + 1;
        auto line = content.substr(start);
        content.remove_suffix(content.size() - start);
        return line;
    };

    if (!content.ends_with('\n')) {
        Warning(_core->_logger_core)
            << path << " is missing trailing newline, skipping last line";
        next();
    }

    while (!content.empty() && !shouldTerminate()) {
        content.remove_suffix(1);
        processLogfileLine(path, next());
    }
}

namespace {
std::optional<StateHistoryEvent> parseEvent(std::string_view event) {
    using enum EventType;
    using enum StateHistoryObject;
    using enum StateHistoryState;
    if (event == "VANISHED HOST"sv) {
        return StateHistoryEvent{vanished, host, current};
    }
    if (event == "VANISHED SERVICE"sv) {
        return StateHistoryEvent{vanished, service, current};
    }
    if (event == "HOST ALERT"sv) {
        return StateHistoryEvent{alert, host, current};
    }
    if (event == "SERVICE ALERT"sv) {
        return StateHistoryEvent{alert, service, current};
    }
    if (event == "HOST DOWNTIME ALERT"sv) {
        return StateHistoryEvent{downtime, host, current};
    }
    if (event == "SERVICE DOWNTIME ALERT"sv) {
        return StateHistoryEvent{downtime, service, current};
    }
    if (event == "HOST FLAPPING ALERT"sv) {
        return StateHistoryEvent{flapping, host, current};
    }
    if (event == "SERVICE FLAPPING ALERT"sv) {
        return StateHistoryEvent{flapping, service, current};
    }
    if (event == "INITIAL HOST STATE"sv) {
        return StateHistoryEvent{alert, host, initial};
    }
    if (event == "INITIAL SERVICE STATE"sv) {
        return StateHistoryEvent{alert, service, initial};
    }
    if (event == "CURRENT HOST STATE"sv) {
        return StateHistoryEvent{alert, host, current};
    }
    if (event == "CURRENT SERVICE STATE"sv) {
        return StateHistoryEvent{alert, service, current};
    }
    return {};  // not interested in that type of event
}
}  // namespace

void StateHistoryThread::processLogfileLine(const std::filesystem::path &path,
                                            std::string_view line) {
    _parsed_messages++;
    auto invalid = [this, path, line](const std::string &message) {
        Warning(_core->_logger_core)
            << path << " contains invalid line '" << line << ": " << message;
    };

    if (!line.starts_with('[')) {
        return invalid("expected '['");
    }
    line.remove_prefix(1);

    if (line.size() < 10) {
        return invalid("timestamp too short");
    }
    auto timestamp_str = line.substr(0, 10);
    line.remove_prefix(10);

    int timestamp_int{};
    auto [ptr, ec] = std::from_chars(timestamp_str.begin(), timestamp_str.end(),
                                     timestamp_int);
    if (ec != std::errc{} || ptr != timestamp_str.end()) {
        return invalid("timestamp is not a number");
    }
    auto timestamp = std::chrono::system_clock::from_time_t(timestamp_int);

    if (!line.starts_with("] "sv)) {
        return invalid("expected '] '");
    }
    line.remove_prefix(2);

    // TODO(sp) What shoud this cryptic comment mean? Why +1?
    // add one, since more entries per second
    _table->_parsed_since = timestamp + 1s;

    auto event = line.substr(0, line.find(':'));
    line.remove_prefix(std::min(line.size(), event.size() + 1));

    if (line.starts_with(' ') && !event.empty() &&
        std::isupper(event[0]) != 0 && event[0] != 'L') {  // LOG VERSION:...
        if (auto parsed_event = parseEvent(event)) {
            processStateLine(path, timestamp, *parsed_event, line.substr(1));
        }
    } else {
        processProgramLine(timestamp, event);
    }
}

void StateHistoryThread::processStateLine(
    const std::filesystem::path &path,
    std::chrono::system_clock::time_point timestamp, StateHistoryEvent event,
    std::string_view line) {
    auto next_field = [&line]() {
        auto field = line.substr(0, line.find(';'));
        line.remove_prefix(std::min(line.size(), field.size() + 1));
        return field;
    };
    auto invalid = [this, timestamp, path](const std::string &message) {
        Warning(_core->_logger_core)
            << "Ignoring invalid line for timestamp "
            << std::chrono::system_clock::to_time_t(timestamp) << " in " << path
            << ": " << message;
    };

    // Extract host and/or service
    if (line.empty()) {
        return invalid("missing host field");
    }
    const std::string host_name{next_field()};

    std::string svc_desc;
    const Object *object = nullptr;
    if (event.object == StateHistoryObject::host) {
        if (const auto *h = _core->world()->getHost(host_name)) {
            object = h;
        }
    } else {
        if (line.empty()) {
            return invalid("missing service field");
        }
        svc_desc = next_field();
        if (const auto *s = _core->world()->getService(host_name, svc_desc)) {
            object = s;
        }
    }

    std::string plugin_output;
    std::string long_plugin_output;
    int state = 0;
    switch (event.type) {
        case EventType::alert: {
            // INITIAL HOST STATE / CURRENT HOST STATE / HOST ALERT /
            // INITIAL SERVICE STATE/CURRENT SERVICE STATE/SERVICE ALERT
            if (line.empty()) {
                return invalid("missing state field");
            }
            auto statename = next_field();
            state =
                event.object == StateHistoryObject::host
                    ? static_cast<int>(LogEntry::parseHostState(statename))
                    : static_cast<int>(LogEntry::parseServiceState(statename));
            next_field();  // hard_soft
            next_field();  // attempt
            plugin_output = std::string{next_field()};
            long_plugin_output = LogEntry::encode(std::string{next_field()});
            break;
        }
        case EventType::flapping:
        case EventType::downtime:
            // flapping: line is one of "STARTED", "STOPPED"
            // downtime: line is one of "STARTED", "STOPPED", "END", "CANCELLED"
            // TODO(sp) Why does the CMC have both "STOPPED" and "END"? Nagios
            // has only "STOPPED".
            state = line.starts_with("STARTED"sv) ? 1 : 0;
            break;
        case EventType::vanished: {
            plugin_output = "Object has been removed from the monitoring";
            state = -1;
            event.type = EventType::alert;
            break;
        }
    }

    // TODO(sp) handleObjectLogEntry() can call getOrCreateObjectHistcache()
    // which might modify object->_state->_histcache. We should really cleanup
    // this cache handling.
    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-const-cast)
    handleObjectLogEntry(timestamp, const_cast<Object *>(object), host_name,
                         svc_desc, event.type, event.state, state,
                         plugin_output, long_plugin_output);
}

void StateHistoryThread::processProgramLine(
    std::chrono::system_clock::time_point timestamp, std::string_view line) {
    // TODO(sp) Use std::string_view::contains() with C++23 below.
    auto contains = [](std::string_view haystack, std::string_view needle) {
        return haystack.find(needle) != std::string_view::npos;
    };
    if (line == "logging initial states"sv ||
        line == "logging intitial states"sv) {
        // In order to detect hosts/services that vanish from the monitoring and
        // appear later again, we handle the beginning of the logging of the
        // initial states in a special way. In all histories we enter an
        // "unmonitored" event. That one will be removed by any INITIAL STATE
        // entries the come immediately later.
        setAllUnmonitored(timestamp);
    } else if (contains(line, "starting..."sv) ||
               contains(line, "active mode..."sv) ||
               contains(line, "LOG VERSION"sv)) {
        registerCoreStartstop(true, timestamp);
    } else if (contains(line, "shutting down..."sv) ||
               contains(line, "shutdown..."sv) ||
               contains(line, "Bailing out"sv) ||
               contains(line, "standby mode..."sv)) {
        registerCoreStartstop(false, timestamp);
    }
}

void StateHistoryThread::handleObjectLogEntry(
    std::chrono::system_clock::time_point timestamp, Object *object,
    const std::string &host_name, const std::string &svc_desc,
    EventType event_type, StateHistoryState state_history_state, int state,
    const std::string &plugin_output, const std::string &long_plugin_output) {
    // Do not rely on _host and _service entries! The object might not exist. Or
    // just the host might exist and not the service. The entry will then have
    // set _service to 0. So if there is no service known we rather need to work
    // with the specification string!
    Histcache *histcache =
        object == nullptr
            ? getOrCreateHistcache(
                  svc_desc.empty() ? host_name : host_name + ";" + svc_desc)
            : getOrCreateObjectHistcache(object);

    // Initialize the new oldest before that as a copy (or create a new one)
    HistcacheEntry preceeding_hce;
    preceeding_hce._timestamp = timestamp;
    preceeding_hce._plugin_output = "";
    preceeding_hce._long_plugin_output = "";

    switch (event_type) {
        case EventType::alert: {
            // Nagios logs INITIAL ___ STATEs even for unmonitored new items,
            // *but* these entries have the state OK/UP. We can detect these
            // entries just because there is no plugin output. Older Nagios
            // versions have "(null)" here, new ones and the CMC an empty
            // string.
            if (state_history_state == StateHistoryState::initial &&
                state == 0 &&
                (plugin_output.empty() || plugin_output == "(null)")) {
                preceeding_hce._plugin_output = "";
                preceeding_hce._long_plugin_output = "";
                preceeding_hce._bits._state = -1;
            } else {
                preceeding_hce._plugin_output = hashedCopyOf(plugin_output);
                preceeding_hce._long_plugin_output =
                    hashedCopyOf(long_plugin_output);
                preceeding_hce._bits._state = state;
            }
            preceeding_hce._bits._know_state = true;

            // Now that we know the state, update the state in all future
            // entries where we haven't known the state (e.g. in DOWNTIME
            // entries there is no state information. The oldest entry has
            // always the index 0.
            for (size_t i = 0; i < histcache->size(); i++) {
                HistcacheEntry *future_hce = histcache->entry(i);
                if (future_hce->_bits._know_state) {
                    break;
                }
                future_hce->_bits._state = preceeding_hce._bits._state;
                future_hce->_bits._know_state = true;
            }
            break;
        }

        // Start and end of downtimes
        case EventType::downtime: {
            preceeding_hce._bits._in_downtime = state == 1;
            preceeding_hce._bits._know_downtime = true;

            // Now that we know the downtime status, update the state in all
            // future entries where we haven't known the downtime status.
            for (size_t i = 0; i < histcache->size(); i++) {
                HistcacheEntry *future_hce = histcache->entry(i);
                if (future_hce->_bits._know_downtime) {
                    break;
                }
                future_hce->_bits._in_downtime =
                    preceeding_hce._bits._in_downtime;
                future_hce->_bits._know_downtime = true;
            }
            break;
        }

        // Start and stop of flapping
        case EventType::flapping: {
            preceeding_hce._bits._flapping = state == 1;
            preceeding_hce._bits._know_flapping = true;

            // Same as for downtime and status -> propagate the flapping
            // information into the future.
            for (size_t i = 0; i < histcache->size(); i++) {
                HistcacheEntry *future_hce = histcache->entry(i);
                if (future_hce->_bits._know_flapping) {
                    break;
                }
                future_hce->_bits._flapping = preceeding_hce._bits._flapping;
                future_hce->_bits._know_flapping = true;
            }
            break;
        }
        case EventType::vanished:
            break;
    }

    // Merge current oldest states
    histcache->mergeOldStates();

    // Add new oldest entry.
    histcache->prependEntry(preceeding_hce);
}

// When we see a logfile line where the core starts and all initial states will
// be logged, we set all objects to unmonitored. Each object that is still in
// the monitoring will shortly afterwards (same second) be set to monitored
// again. All vanished objects will be classified correctly as unmonitored.
void StateHistoryThread::setAllUnmonitored(
    std::chrono::system_clock::time_point when) {
    const std::lock_guard<std::mutex> lghc(_histcaches_mutex);
    for (auto &histcache : _histcaches) {
        Histcache *hc = histcache.second;
        const std::lock_guard<std::mutex> lg(hc->_lock);
        HistcacheEntry *oldest_hce = hc->oldestEntry();
        // If the history is not empty and the next (more recent) entry follows
        // shortly after, there is no point in setting it to unmonitored - just
        // because the core took a few seconds for starting. Remember always: we
        // parse more recently logged messages first (those at the end of a
        // file).
        if ((oldest_hce == nullptr) ||
            oldest_hce->_timestamp - when >=
                _core->config()->history_.max_core_downtime_) {
            // If the oldest entry is "unmonitored" anyway, we just update its
            // timestamp.
            if ((oldest_hce != nullptr) && oldest_hce->_bits._state == -1) {
                oldest_hce->_timestamp = when;
                oldest_hce->_plugin_output = "";
                oldest_hce->_long_plugin_output = "";
            }
            // Create a new entry of type unmonitored
            else {
                HistcacheEntry hce;
                hce._bits._state = -1;
                hce._timestamp = when;
                hce._plugin_output = "";
                hce._long_plugin_output = "";
                hc->prependEntry(hce);
            }
        }
    }
}

void StateHistoryThread::finalizeCaches() {
    auto start = std::chrono::steady_clock::now();
    for (auto &histcache : _histcaches) {
        Debug(_core->_logger_core)
            << "finalizing history cache for '" << histcache.first << "'";
        histcache.second->finalize();
    }
    Notice(_core->_logger_core)
        << "finalized " << _histcaches.size() << " history caches in "
        << elapsed_ms_since(start) << " ms";
}

// Is called whenever we see a logfile line that tells us that
// the core been started or stopped.
void StateHistoryThread::registerCoreStartstop(
    bool start_and_not_stop, std::chrono::system_clock::time_point when) {
    const std::lock_guard<std::mutex> lg(_table->_startstop_mutex);
    // Case A) This is the first (relevant) core start/stop event we process?
    // Then simply add it to our history and remember wether it's a start or
    // stop.
    // NOTE: We should probably massage the if/then/else cascade below a bit to
    // make the pragma unnecessary.
    // NOLINTNEXTLINE(bugprone-branch-clone)
    if (_table->_core_startstops.empty()) {
        _table->_core_startstops.push_front(when);
        _table->_first_startstop_entry_is_running = start_and_not_stop;
    }

    // Case B) We found an older entry with the same type as last time (e.g.
    // two sequential starting entries). Simply update the timestamp of the
    // oldest entry
    else if (start_and_not_stop == _table->_first_startstop_entry_is_running) {
        _table->_core_startstops.front() = when;
    }

    // Case C) A STOP while the following START is too soon.  We ignore the STOP
    // then. We need to keep the START in order not to lose information. There
    // might be two successive stop events!
    else if (!start_and_not_stop && _table->_first_startstop_entry_is_running &&
             (_table->_core_startstops.front() - when <
              _core->config()->history_.max_core_downtime_)) {
        // We register a STOP, previous event type was START, and it was too
        // short.
    }

    // Case D) Normal STOP <-> START transition
    else {
        _table->_core_startstops.push_front(when);
        _table->_first_startstop_entry_is_running = start_and_not_stop;
    }
}

bool StateHistoryThread::shouldTerminate() {
    const std::lock_guard<std::mutex> lg(_should_terminate_mutex);
    return _should_terminate;
}

void StateHistoryThread::setShouldTerminate(bool value) {
    const std::lock_guard<std::mutex> lg(_should_terminate_mutex);
    _should_terminate = value;
}

bool StateHistoryThread::threadFinished() {
    const std::lock_guard<std::mutex> lg(_thread_finished_mutex);
    return _thread_finished;
}

void StateHistoryThread::setThreadFinished(bool value) {
    const std::lock_guard<std::mutex> lg(_thread_finished_mutex);
    _thread_finished = value;
}

Histcache *StateHistoryThread::getOrCreateObjectHistcache(Object *object) {
    if (object->state()->_histcache == nullptr) {
        object->stateNonConst()->_histcache =
            getOrCreateHistcache(object->specname());
    }
    return object->state()->_histcache;
}

Histcache *StateHistoryThread::getOrCreateHistcache(
    const std::string &specname) {
    const std::lock_guard<std::mutex> lg(_histcaches_mutex);
    // If a host/service has been revived then an (old) hist cache might still
    // be existing. Return that if that is the case.
    auto it = _histcaches.find(specname);
    if (it != _histcaches.end()) {
        return it->second;
    }

    // Create new hist cache
    auto *histcache = new Histcache();
    _histcaches[specname] = histcache;
    return histcache;
}

void StateHistoryThread::flushCache() {
    _strings.clear();
    _cached_strings = 0;
    for (auto &histcache : _histcaches) {
        delete histcache.second;
    }
    _histcaches.clear();
}

// This method makes sure that only one copy of identical strings is kept in
// memory. That is crucial for the memory consumption, since checkout outputs
// are identitical in many cases.  If s is already contained in our string cache
// then a pointer to that copy is being returned. The callee may store that
// pointer until the end of the world but must never modify or free it. If s it
// not yet contained a copy of s is being stored and returned - with the same
// rules.
std::string_view StateHistoryThread::hashedCopyOf(const std::string &s) {
    _cached_strings++;
    return *_strings.insert(s).first;
}
