// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/TableCachedStatehist.h"

#include <algorithm>
#include <compare>
#include <iomanip>
#include <ratio>
#include <stdexcept>
#include <thread>
#include <unordered_map>
#include <vector>

#include "cmc/Core.h"
#include "cmc/Histcache.h"
#include "cmc/HistcacheEntry.h"
#include "cmc/Host.h"
#include "cmc/Object.h"
#include "cmc/Service.h"
#include "cmc/State.h"
#include "cmc/Timeperiod.h"
#include "cmc/World.h"
#include "cmc/cmc.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/Column.h"
#include "livestatus/Filter.h"
#include "livestatus/HostServiceState.h"
#include "livestatus/ICore.h"
#include "livestatus/Logger.h"
#include "livestatus/POSIXUtils.h"
#include "livestatus/Query.h"
#include "livestatus/Row.h"
#include "livestatus/TableStateHistory.h"
#include "livestatus/User.h"

using namespace std::chrono_literals;

namespace {
constexpr auto time_infinity = std::chrono::system_clock::time_point::max();
}  // namespace

void *histcache_thread_run(void *data) {
    setThreadName("state history builder");
    auto *t = static_cast<TableCachedStatehist *>(data);
    t->_state_history_thread.run(t->the_core_);
    return nullptr;
}

TableCachedStatehist::TableCachedStatehist(ICore *mc)
    : _first_startstop_entry_is_running{false}
    , _parsed_since{time_infinity}
    , _state_history_thread{this}
    , logger_core_{mc->loggerCore()} {
    _state_history_thread.setShouldTerminate(false);
    _state_history_thread.setThreadFinished(false);
    TableStateHistory::addColumns(this, *mc, "", ColumnOffsets{});
}

TableCachedStatehist::~TableCachedStatehist() { flushCache(); }

std::string TableCachedStatehist::name() const { return "statehist"; }

std::string TableCachedStatehist::namePrefix() const { return "statehist_"; }

// .-Build Cache-----------------------------------------------------------.
// |         ____        _ _     _    ____           _                     |
// |        | __ ) _   _(_) | __| |  / ___|__ _  ___| |__   ___            |
// |        |  _ \| | | | | |/ _` | | |   / _` |/ __| '_ \ / _ \           |
// |        | |_) | |_| | | | (_| | | |__| (_| | (__| | | |  __/           |
// |        |____/ \__,_|_|_|\__,_|  \____\__,_|\___|_| |_|\___|           |
// |                                                                       |
// +-----------------------------------------------------------------------+
// |  Code for building up the cache                                       |
// '-----------------------------------------------------------------------'
void TableCachedStatehist::startBuildingCache(Core *core) {
    if (_thread) {
        Emergency(logger_core_) << "state history thread already created";
        return;
    }
    _state_history_thread.setShouldTerminate(false);
    _state_history_thread.setThreadFinished(false);
    the_core_ = core;
    pthread_t new_thread{};
    if (auto result =
            pthread_create(&new_thread, nullptr, histcache_thread_run, this);
        result != 0) {
        Emergency(logger_core_)
            << generic_error{result, "cannot create state history thread"};
        return;
    }
    _thread = new_thread;
}

void TableCachedStatehist::stopBuildingCache() {
    if (!_thread) {
        return;
    }
    _state_history_thread.setShouldTerminate(true);
    if (auto result = pthread_join(*_thread, nullptr); result != 0) {
        Alert(logger_core_)
            << generic_error{result, "cannot join state history thread"};
        return;
    }
    _thread.reset();
    the_core_ = nullptr;
    Informational(logger_core_) << "state history thread successfully stopped.";
}

void TableCachedStatehist::finishBuildingCache() {
    if (_state_history_thread.threadFinished()) {
        stopBuildingCache();
    }
}

void TableCachedStatehist::dumpCoreStartstops() const {
    Debug(logger_core_) << "Timestamp of core start/stop     Duration WHAT";
    Debug(logger_core_)
        << "--------------------------------------------------------------------------------";
    int i = 0;
    for (auto it = _core_startstops.begin(); it != _core_startstops.end();
         ++it) {
        i++;
        auto when = *it;
        auto end = (it + 1 == _core_startstops.end())
                       ? std::chrono::system_clock::now()
                       : *(it + 1);
        auto duration_secs = mk::ticks<std::chrono::seconds>(end - when);
        const std::string what =
            (((i % 2) == 0) != _first_startstop_entry_is_running) ? "START"
                                                                  : "STOP";
        Debug(logger_core_) << std::chrono::system_clock::to_time_t(when)
                            << " (" << FormattedTimePoint(when) << ") "
                            << std::setw(7) << duration_secs << "s " << what;
    }
}

template <typename Init>
void TableCachedStatehist::updateCache(const Object &object, Init f) {
    Histcache *shc = object.stateNonConst()->histcache();
    const std::lock_guard<std::mutex> lg(shc->_lock);
    HistcacheEntry new_hce = shc->youngestEntry();
    new_hce._timestamp = std::chrono::system_clock::now();
    f(new_hce);
    shc->appendEntry(new_hce);
}

void TableCachedStatehist::registerAlert(const Object &object, int st,
                                         const std::string &output,
                                         const std::string &long_output) {
    updateCache(object, [&](HistcacheEntry &hce) {
        hce._plugin_output = _state_history_thread.hashedCopyOf(output);
        hce._long_plugin_output =
            _state_history_thread.hashedCopyOf(long_output);
        hce._bits._state = st;
    });
}

void TableCachedStatehist::registerDowntime(const Object &object,
                                            bool started) {
    updateCache(object,
                [&](HistcacheEntry &hce) { hce._bits._in_downtime = started; });
}

void TableCachedStatehist::registerFlapping(const Object &object,
                                            bool started) {
    updateCache(object,
                [&](HistcacheEntry &hce) { hce._bits._flapping = started; });
}

void TableCachedStatehist::flushCache() {
    Notice(logger_core_) << "flushing state history cache";
    stopBuildingCache();
    _state_history_thread.flushCache();
    _core_startstops.clear();
    // TODO(sp): What is this for???
    _state_history_thread.setThreadFinished(false);
}

// .-Query-----------------------------------------------------------------.
// |                      ___                                              |
// |                     / _ \ _   _  ___ _ __ _   _                       |
// |                    | | | | | | |/ _ \ '__| | | |                      |
// |                    | |_| | |_| |  __/ |  | |_| |                      |
// |                     \__\_\\__,_|\___|_|   \__, |                      |
// |                                           |___/                       |
// +-----------------------------------------------------------------------+
// | Code for answering actual queries.                                    |
// '-----------------------------------------------------------------------'

void TableCachedStatehist::answerQuery(Query &query, const User &user,
                                       const ICore &core) {
    auto *logger = core.loggerLivestatus();
    // This flag might be set to true by the return value of
    // processDataset(...)
    _abort_query = false;

    // Optimize time interval for the query. Due to our caching we do not need
    // to impose a limit for the past. If that is missing we simply use our
    // cache horizon as past.
    auto glb = query.greatestLowerBoundFor("time");
    auto since =
        glb ? std::chrono::system_clock::from_time_t(*glb) : _start_of_cache;

    auto lub = query.leastUpperBoundFor("time");
    auto until = (lub ? std::chrono::system_clock::from_time_t(*lub)
                      : std::chrono::system_clock::now()) +
                 1s;

    if (until - since <= 1s) {
        query.invalidRequest("Query timeframe is 0 seconds");
        return;
    }

    // We cannot go further in the past then our _start_of_cache. But our cache
    // logic allows this nevertheless and marks everything beyond as
    // "unmonitored".

    // Check if our cache has already processed the time range that we need for
    // our query. If not, then we need to wait.
    std::chrono::system_clock::duration missing_seconds = 0s;
    while (0s < (missing_seconds =
                     _parsed_since - std::max(since, _start_of_cache))) {
        if (core.impl<Core>()->isShuttingDown()) {
            Informational(logger)
                << "Aborting long-running query: core is shutting down.";
            return;
        }
        Informational(logger)
            << "Missing " << mk::ticks<std::chrono::seconds>(missing_seconds)
            << " seconds in statehist cache. Waiting...";
        std::this_thread::sleep_for(1s);
        if (query.timelimitReached()) {
            Informational(logger)
                << "state history cache query not finished within time limit, aborting";
            return;
        }
    }

    // Now find all required hosts and services. Create a partial filter, that
    // contains only such filters that check attributes of current hosts and
    // services. That way we select all necessary objects.
    auto object_filter = TableStateHistory::createPartialFilter(query);

    // Output object of our table. We need that because our filters and columns
    // are relative to that type of class.
    HostServiceState hss;

    // Iterate over all hosts
    for (const auto &host : g_live_world->config_.hosts_) {
        if (_abort_query) {
            break;
        }
        hss._is_host = true;
        hss._host = &host->ihost();
        hss._service = nullptr;
        hss._host_name = host->name();
        hss._service_description = "";

        if (object_filter->accepts(Row{&hss}, user, query.timezoneOffset()) &&
            user.is_authorized_for_host(host->ihost())) {
            outputStateHistory(query, core, host.get(), since, until);
            if (_abort_query) {
                break;
            }
        }

        // Iterate over services, even if the host filter does not match. The
        // query might just want to get the services.
        for (const auto &service : Host::Services(*host)) {
            if (!user.is_authorized_for_service(service->iservice())) {
                continue;
            }
            hss._is_host = false;
            hss._service = &service->iservice();
            hss._service_description = service->name();
            if (object_filter->accepts(Row{&hss}, user,
                                       query.timezoneOffset())) {
                outputStateHistory(query, core, service.get(), since, until);
                if (_abort_query) {
                    break;
                }
            }
        }
    }
}

void TableCachedStatehist::outputStateHistory(
    Query &query, const ICore &core, Object *object,
    std::chrono::system_clock::time_point since,
    std::chrono::system_clock::time_point until) {
    // We need to zip together several different time lines:
    // [1] The time line of the object itself
    // [2] The time line of the host, if the object is a service (for host up
    //     and downtime state)
    // [3] The time line of the object's notification period
    // [4] The time line of the object's service period
    // [5] The time line of the monitoring core (when it was running and
    //     stopped)
    //
    // For each of the timelines we have a cursor that points to the next event
    // in that timeline. It is crucial that this next event is always in the
    // future of the current phase that we are looking at. So at the beginning
    // all five cursors must point to an event >= since. At timeline may end and
    // have no further events.  The time of the next event is then set to
    // time_infinity so that is never being considered (well, until
    // 2262. Yurks).

    // [1] Time line of the object itself
    Histcache *shc = _state_history_thread.getOrCreateObjectHistcache(object);
    // might be updated during monitoring when alerts happen
    const std::lock_guard<std::mutex> lg(shc->_lock);
    if (logger_core_->isLoggable(LogLevel::debug)) {
        Debug(logger_core_)
            << "================================================================================";
        shc->dump(logger_core_);
        dumpCoreStartstops();
    }

    HistcacheEntry hce_pending;
    hce_pending._plugin_output =
        "No information about that period of time available";
    hce_pending._long_plugin_output = "";
    hce_pending._bits._state = -1;
    hce_pending._timestamp = since;

    // Now find youngest entry that lies *before* our since. If there is no such
    // entry then we create an artificial entry and assume "unmonitored".
    HistcacheEntry *hce = nullptr;
    const size_t num_entries = shc->size();
    int i = shc->findEntryBefore(since);
    if (i == -1) {
        hce = &hce_pending;
        Debug(core.loggerLivestatus())
            << "Query timerange out of cache range. Creating artificial "
               "unmonitored time range.";
    } else {
        hce = shc->entry(i);
    }

    // [2] Time line of the host of a service If we deal with a service, then we
    // also need the history of its host.  From that we derive host_down and
    // in_host_downtime
    HistcacheEntry
        host_hce_pending;  // all set to zero fits our needs. Assume host UP
    Host *host = nullptr;
    Histcache *host_shc = nullptr;
    size_t host_num_entries = 0;
    int hi = -1;
    HistcacheEntry *host_hce = nullptr;
    // TODO(sp): This whole conditional locking contstruct below is ugly and
    // fragile.
    std::unique_ptr<std::lock_guard<std::mutex>> lgp;
    if (object->isService()) {
        host = object->host();
        host_shc = _state_history_thread.getOrCreateObjectHistcache(host);
        if (logger_core_->isLoggable(LogLevel::debug)) {
            host_shc->dump(logger_core_);
        }
        // We do *not* want to read the variable, only its scope matters.
        lgp = std::make_unique<std::lock_guard<std::mutex>>(host_shc->_lock);
        host_num_entries = host_shc->size();
        hi = host_shc->findEntryBefore(since);
        if (hi == -1) {
            host_hce = &host_hce_pending;
        } else {
            host_hce = host_shc->entry(hi);
        }
    }

    // [3] notification period
    const auto *notification_period = object->_notification_period;
    int notification_period_index = notification_period->findTransition(since);

    // [4] service period
    const auto *service_period = object->_service_period;
    int service_period_index = service_period->findTransition(since);

    // [5] Monitoring core: This is in _core_startstops;

    // TODO(sp) We can get a -1 here, which is confusing, and it is unclear if
    // the code below always handles this correctly.
    ssize_t core_startstop_index = findLastCoreStartstopBefore(since);
    const bool is_unmonitored =
        _first_startstop_entry_is_running == (core_startstop_index % 2 == 1);

    // Prepare struct to be consumed by query and add static values
    HostServiceState hss;
    hss._is_host = object->isHost();
    hss._notification_period = notification_period->name();
    hss._service_period = service_period->name();
    hss._host = &object->host()->ihost();
    hss._service =
        object->service() == nullptr ? nullptr : &object->service()->iservice();
    hss._host_name = object->host()->name();
    hss._service_description =
        object->isService() ? object->service()->name() : "";

    // Start values for dynamic entries
    hss._time = since;
    hss._from = since;
    hss._lineno = 0;  // Needed for sorting, always increased.

    // Initialise struct with situation at the beginning of the interval
    // Remember state during phases of core stops
    //
    // NOTE:clangd somehow doesn't respect
    // bugprone-signed-char-misuse.CharTypdefsToIgnore here
    // NOLINTNEXTLINE(cert-str34-c)
    int previous_state = hce->_bits._state;
    hss._state = is_unmonitored ? -1 : previous_state;
    hss._in_downtime = static_cast<int>(hce->_bits._in_downtime);
    hss._is_flapping = static_cast<int>(hce->_bits._flapping);
    hss._log_output = hce->_plugin_output;
    hss._long_log_output = hce->_long_plugin_output;
    hss._in_notification_period =
        static_cast<int>(notification_period_index % 2 == 0);
    hss._in_service_period = static_cast<int>(service_period_index % 2 == 0);
    hss._host_down =
        static_cast<int>((host_hce != nullptr) && host_hce->_bits._state > 0);
    hss._in_host_downtime =
        static_cast<int>((host_hce != nullptr) && host_hce->_bits._in_downtime);

    // Now comes our big loop. We look at all five (or maybe four) time lines at
    // the same time. Each time one of those has a new entry, we update the
    // states in hss accordingly. And before that we output the previous phase,
    // if its duration was > 0.

    while (true) {
        // Find out, which of our five timelines comes next.
        std::chrono::system_clock::time_point next_object_event;
        std::chrono::system_clock::time_point next_host_event;
        std::chrono::system_clock::time_point next_notification_period;
        std::chrono::system_clock::time_point next_service_period;
        std::chrono::system_clock::time_point next_core_startstop;

        // [1] Event in object's own time line
        if (i + 1 < static_cast<int>(num_entries)) {
            // TODO(sp): Dubious cast, fix type of i instead
            next_object_event =
                shc->entry(static_cast<size_t>(i) + 1)->_timestamp;
        } else {
            next_object_event = time_infinity;
        }

        // [2] Event in host's timeline (if object is a service)
        if (host != nullptr && hi + 1 < static_cast<int>(host_num_entries)) {
            // TODO(sp): Dubious cast, fix type of hi instead
            next_host_event =
                host_shc->entry(static_cast<size_t>(hi) + 1)->_timestamp;
        } else {
            next_host_event = time_infinity;
        }

        // [3] Notification period switch
        if (notification_period_index + 1 <
            notification_period->numTransitions()) {
            next_notification_period =
                notification_period->transition(notification_period_index + 1);
        } else {
            next_notification_period = time_infinity;
        }

        // [4] Service period switch
        if (service_period_index + 1 < service_period->numTransitions()) {
            next_service_period =
                service_period->transition(service_period_index + 1);
        } else {
            next_service_period = time_infinity;
        }

        // [5] history of core
        {
            const std::lock_guard<std::mutex> lgss(_startstop_mutex);
            next_core_startstop =
                (static_cast<size_t>(core_startstop_index + 1) <
                 _core_startstops.size())
                    ? _core_startstops[core_startstop_index + 1]
                    : time_infinity;
        }

        // Now that we know the five next time stamps, find out which one is
        // first.  Also it may be the case that our until time is reached and is
        // less then all five timestamps. Note: in case of two timestamps being
        // identical nothing serious happens. One randomly will be processed
        // first.  The other in a later turn. The resulting period with duration
        // 0 will be dropped.

        auto next_event = std::min(
            until, std::min(next_object_event,
                            std::min(next_host_event,
                                     std::min(next_notification_period,
                                              std::min(next_service_period,
                                                       next_core_startstop)))));

        // Now pump out one period. The query can be aborted due to a line or
        // time limit.  But create a phase only when the duration is
        // non-zero. Remember: statehist is not about log entries, but about
        // time periods of certain states.
        hss._until = next_event;
        hss._duration = hss._until - hss._from;

        if (hss._duration != 0s) {
            hss.computePerStateDurations(until - since);
            if (!query.processDataset(Row{&hss})) {
                break;
            }

            // Switch for for next round
            hss._lineno++;  // must be unique and increasing, needed for sorting
                            // purposes
            hss._from = hss._until;
            hss._time = hss._until;
        }

        if (until == next_event) {
            break;  // We are finished. Hooray...
        }

        // Now make hss reflect the next period, depending on which thing has
        // changed.
        if (next_event == next_object_event) {
            hce = shc->entry(++i);
            // NOTE:clangd somehow doesn't respect
            // bugprone-signed-char-misuse.CharTypdefsToIgnore here
            // NOLINTNEXTLINE(cert-str34-c)
            hss._state = hce->_bits._state;
            hss._in_downtime = static_cast<int>(hce->_bits._in_downtime);
            hss._is_flapping = static_cast<int>(hce->_bits._flapping);
            hss._log_output = hce->_plugin_output;
            hss._long_log_output = hce->_long_plugin_output;
            previous_state = hss._state;
        }

        else if (next_event == next_host_event) {
            // Only output a chunk if relevant things change in the host state.
            HistcacheEntry *next_host_hce = host_shc->entry(++hi);
            if (host_hce->_bits._state != next_host_hce->_bits._state ||
                host_hce->_bits._in_downtime !=
                    next_host_hce->_bits._in_downtime) {
                host_hce = next_host_hce;
                hss._host_down = static_cast<int>(host_hce->_bits._state > 0);
                hss._in_host_downtime =
                    static_cast<int>(host_hce->_bits._in_downtime);
            }
        }

        else if (next_event == next_notification_period) {
            notification_period_index++;
            hss._in_notification_period =
                static_cast<int>(notification_period_index % 2 == 0);
        }

        else if (next_event == next_service_period) {
            service_period_index++;
            hss._in_service_period =
                static_cast<int>(service_period_index % 2 == 0);
        }

        else  // must be core event
        {
            core_startstop_index++;
            // TODO(sp): Same as outer is_unmonitored?
            const bool is_unmonitored2 = _first_startstop_entry_is_running ==
                                         (core_startstop_index % 2 == 1);
            if (is_unmonitored2) {
                hss._state = -1;
            } else {
                // In case the core starts again we assume that the state of the
                // object is unchanged (due to state retention). We need this
                // because we optimized away the INITIAL STATE entries in the
                // object history.
                hss._state = previous_state;
            }
        }
    }  // main loop for outputting phases
}

ssize_t TableCachedStatehist::findLastCoreStartstopBefore(
    std::chrono::system_clock::time_point when) {
    const std::lock_guard<std::mutex> lg(_startstop_mutex);
    auto it = _core_startstops.end();
    size_t n = _core_startstops.size();
    while (n > 0) {
        --n;
        --it;
        if (*it < when) {
            return n;
        }
    }
    return -1;
}

std::shared_ptr<Column> TableCachedStatehist::column(
    std::string colname) const {
    try {
        // First try to find column in the usual way
        return Table::column(colname);
    } catch (const std::runtime_error &e) {
        // Now try with prefix "current_", since our joined tables have this
        // prefix in order to make clear that we access current and not historic
        // data and in order to prevent mixing up historic and current fields
        // with the same name.
        return Table::column("current_" + colname);
    }
}
