// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/Timeperiod.h"

#include <algorithm>
#include <compare>
#include <cstddef>
#include <iterator>
#include <utility>

#include "cmc/MonitoringLog.h"

Timeperiod::Timeperiod(std::string name, std::string alias,
                       std::vector<value_type> transitions)
    : _name{std::move(name)}
    , _alias{std::move(alias)}
    , _transitions{std::move(transitions)}
    , _next_transition{0} {}

void Timeperiod::update(MonitoringLog *monitoring_log, value_type now) {
    while (_next_transition < _transitions.size() &&
           now >= _transitions[_next_transition]) {
        _next_transition++;
        if (monitoring_log != nullptr) {
            monitoring_log->timeperiodTransition(_name,
                                                 static_cast<int>(!isActive()),
                                                 static_cast<int>(isActive()));
        }
    }
}

Timeperiod::value_type Timeperiod::lastStateChange() const {
    return (_next_transition != 0U) ? _transitions[_next_transition - 1]
                                    : value_type();
}

Timeperiod::value_type Timeperiod::previousLastStateChange() const {
    return (_next_transition > 1) ? _transitions[_next_transition - 2]
                                  : value_type();
}

Timeperiod::const_iterator Timeperiod::begin() const {
    return std::next(
        std::begin(_transitions),
        std::min(std::size_t{_next_transition}, _transitions.size()));
}

Timeperiod::const_iterator Timeperiod::end() const {
    return std::end(_transitions);
}

Timeperiod::value_type Timeperiod::nextActiveTime(value_type from) const {
    // find interval that from is in and the state of the interval.
    unsigned n = _next_transition;
    while (n < _transitions.size() && _transitions[n] <= from) {
        n++;
    }

    if ((n % 2) != 0U) {
        return from;  // from lies in an active interval
    }
    if (n < _transitions.size()) {
        return _transitions[n];  // next active interval
    }
    return {};  // no next active interval
}

// Find the index of the transition that was just before or at when. This
// allows to know wether 'when' is currently active or not and then to jump
// forward. A return value of -1 indicates the the earliest entry of the
// time period is older then the queried time.
int Timeperiod::findTransition(value_type when) const {
    // TODO(mk): it might be faster to implement a binary search here
    for (uint32_t n = _transitions.size(); n >= 1; n--) {
        if (_transitions[n - 1] <= when) {
            return static_cast<int>(n - 1);
        }
    }
    return -1;
}

bool Timeperiod::isActiveTime(value_type t) const {
    return findTransition(t) % 2 == 0;
}

bool Timeperiod::operator==(const Timeperiod &other) const {
    return name() == other.name() && alias() == other.alias() &&
           nextTransition() == other.nextTransition() &&
           _transitions == other._transitions;
}
