// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/World.h"

#include <algorithm>
#include <cstdlib>
#include <iterator>
#include <ostream>
#include <utility>

#include "cmc/Contact.h"
#include "cmc/ContactGroup.h"
#include "cmc/MonitoringLog.h"
#include "cmc/Object.h"
#include "cmc/ObjectGroup.h"
#include "cmc/PbConfigLoader.h"
#include "cmc/RRDFileLocator.h"
#include "cmc/RRDInfoCache.h"
#include "cmc/Service.h"
#include "config/v1/types.pb.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/Logger.h"
#include "livestatus/StringUtils.h"

World::World(Core &core,
             std::unique_ptr<checkmk::cmc::config::v1::Core> core_config,
             Logger *logger_core, Logger *logger_rrd,
             SerialTokenFactory &serial_token_factory)
    : logger_(logger_core)
    , num_hosts_(0)
    , num_services_(0)
    , creation_time_(std::chrono::system_clock::now())
    , serial_token_factory_{serial_token_factory} {
    createConfig(core, *core_config, logger_rrd);
}

World::~World() { Notice(logger_) << "deleting " << *this; }

void World::createConfig(Core &core,
                         const checkmk::cmc::config::v1::Core &core_config,
                         Logger *logger_rrd) {
    auto start = std::chrono::steady_clock::now();  // TODO(sp) Fix calculation
    const PbConfigLoader pl{core_config};
    config_.global_config_ = pl.makeGlobalConfig(core_config.globals());
    info_cache_ = std::make_unique<RRDInfoCache>(*this, logger_rrd);
    rrd_file_locator_single_ = std::make_unique<RRDFileLocatorSingle>(
        logger_rrd, config_.global_config_.rrd_.single_directory_);
    rrd_file_locator_multiple_ = std::make_unique<RRDFileLocatorMultiple>(
        logger_rrd, config_.global_config_.rrd_.multiple_directory_);
    for (const auto &timeperiod : core_config.timeperiods()) {
        auto tp = std::make_unique<Timeperiod>(pl.makeTimeperiod(timeperiod));
        config_.timeperiods_[tp->name()] = std::move(tp);
    }
    for (const auto &contact_group : core_config.contact_groups()) {
        auto cg = pl.makeContactGroup(contact_group);
        config_.contact_groups_[cg->name()] = std::move(cg);
    }
    for (const auto &contact : core_config.contacts()) {
        auto c = pl.makeContact(contact, this);
        // REVIEW: code below can't be easy moved into makeContact: 'this' is
        // required. Still code is ugly and should be revamped, e.g. keep group
        // list in Contact or vector<string> makeContactGroupList(contact);
        for (auto group_ref : contact.group_name_refs()) {
            getContactGroup(pl.makeString(group_ref))->insert(c->name());
        }
        config_.contacts_[c->name()] = std::move(c);
    }
    for (const auto &contact_list : core_config.contact_list_pool().entries()) {
        config_.contact_lists_.emplace_back(std::make_unique<ContactList>(
            pl.makeContactList(contact_list, this)));
    }
    for (const auto &host : core_config.hosts()) {
        auto h = pl.makeHost(core, *this, host);
        if (hosts_by_name_[h->name()] != nullptr) {
            // Something is wrong with our config, the host name must be unique.
            // This should have been handled at config generation time already,
            // but to save us future debugging fun, we check here, too.
            Alert(logger_) << "duplicate " << *h << ", ignoring...";
            continue;
        }

        auto address = h->lookup({}, "address");
        if (!address.empty()) {
            hosts_by_designation_[mk::unsafe_tolower(address)] = h.get();
        }
        auto alias = h->lookup({}, "alias");
        if (!alias.empty()) {
            hosts_by_designation_[mk::unsafe_tolower(alias)] = h.get();
        }
        hosts_by_designation_[mk::unsafe_tolower(h->name())] = h.get();

        if (h->_ipv4_address != 0U) {
            hosts_by_ipv4address_.emplace(h->_ipv4_address, h.get());
        }

        num_hosts_++;
        num_services_ += h->_services.size();
        hosts_by_name_[h->name()] = h.get();

        config_.hosts_.push_back(std::move(h));
    }
    for (const auto &entry : core_config.parents_of_hosts()) {
        Host *host = getHost(pl.makeString(entry.host_ref()));
        for (auto parent_ref : entry.parent_refs()) {
            auto parent_name = pl.makeString(parent_ref);
            Host *parent = getHost(parent_name);
            if (parent == nullptr) {
                // Something is wrong with our config. This should have been
                // handled at config generation time already, but to save us
                // future debugging fun, we check here, too.
                Alert(logger_) << "unknown parent " << parent_name << " for "
                               << *host << ", ignoring...";
                continue;
            }
            host->addParent(parent);
            parent->addChild(host);
        }
    }
    for (const auto &group : core_config.host_groups()) {
        const auto [name, alias] = pl.makeObjectGroup(group);
        if (auto *group = getHostGroup(name)) {
            group->setAlias(alias);
        } else {
            Informational(logger_) << "host group " << name << " is orphaned.";
        }
    }
    for (const auto &group : core_config.service_groups()) {
        const auto [name, alias] = pl.makeObjectGroup(group);
        if (auto *group = getServiceGroup(name)) {
            group->setAlias(alias);
        } else {
            Informational(logger_)
                << "service group " << name << " is orphaned.";
        }
    }
    for (const auto &command_ref : core_config.command_refs()) {
        config_.commands_.emplace_back(pl.makeString(command_ref));
    }

    serial_token_ = createSerialToken("World");
    config_load_time_ = elapsed_ms_since(start);
}

void World::logLoadStatisics() const {
    Notice(logger_) << "loaded " << *this << " in " << config_load_time_
                    << " ms";
}

Host *World::getHost(const std::string &name) const {
    auto it = hosts_by_name_.find(name);
    return it == hosts_by_name_.end() ? nullptr : it->second;
}

std::vector<Host *> World::getHostsByIPv4Address(in_addr_t a) const {
    auto r = hosts_by_ipv4address_.equal_range(a);
    std::vector<Host *> hosts;
    std::transform(r.first, r.second, std::back_inserter(hosts),
                   [](auto p) { return p.second; });
    return hosts;
}

Host *World::getHostByDesignation(const std::string &designation) const {
    auto it = hosts_by_designation_.find(mk::unsafe_tolower(designation));
    return it == hosts_by_designation_.end() ? nullptr : it->second;
}
Service *World::getService(const std::string &host_name,
                           const std::string &description) const {
    Host *host = getHost(host_name);
    return host != nullptr ? host->getService(description) : nullptr;
}

SerialToken World::createSerialToken(const std::string &creator) const {
    return serial_token_factory_.create(
        creator, config_.global_config_.general_.serial_);
}

Object *World::correspondingObject(const Object *object) const {
    if (object->isHost()) {
        return getHost(object->host()->name());
    }
    return getService(object->host()->name(), object->name());
}

HostGroup *World::getHostGroup(const std::string &name) const {
    auto it = config_.host_groups_.find(name);
    return it == config_.host_groups_.end() ? nullptr : it->second.get();
}

ServiceGroup *World::getServiceGroup(const std::string &name) const {
    auto it = config_.service_groups_.find(name);
    return it == config_.service_groups_.end() ? nullptr : it->second.get();
}

HostGroup &World::ensureHostGroup(const std::string &name) {
    auto *group = getHostGroup(name);
    if (group == nullptr) {
        // create host group on the fly, if necessary
        group = config_.host_groups_
                    .emplace(name, std::make_unique<HostGroup>(name, name))
                    .first->second.get();
    }
    return *group;
}

ServiceGroup &World::ensureServiceGroup(const std::string &name) {
    auto *group = getServiceGroup(name);
    if (group == nullptr) {
        // create service group on the fly, if necessary
        group = config_.service_groups_
                    .emplace(name, std::make_unique<ServiceGroup>(name, name))
                    .first->second.get();
    }
    return *group;
}

ContactGroup *World::getContactGroup(const std::string &name) const {
    auto it = config_.contact_groups_.find(name);
    return it == config_.contact_groups_.end() ? nullptr : it->second.get();
}

Contact *World::getContact(const std::string &name) const {
    auto it = config_.contacts_.find(name);
    return it == config_.contacts_.end() ? nullptr : it->second.get();
}

ContactList *World::getContactList(uint32_t id) const {
    return config_.contact_lists_[id].get();
}

const Timeperiod &World::getTimeperiod(const std::string &description,
                                       const std::string &name) {
    auto it = config_.timeperiods_.find(name);
    if (it == config_.timeperiods_.end()) {
        Emergency(logger_) << "unknown " << description << " period '" << name
                           << "', aborting";
        abort();
    }
    return *it->second;
}

void World::updateTimeperiods(MonitoringLog &monitoring_log,
                              std::chrono::system_clock::time_point now) {
    for (auto &[name, timeperiod] : config_.timeperiods_) {
        timeperiod->update(&monitoring_log, now);
    }
}

void World::logCurrentStates(MonitoringLog &monitoring_log,
                             LogStateType log_state_type) {
    monitoring_log.logInitialStates();
    for (const auto &host : config_.hosts_) {
        monitoring_log.state(*host, log_state_type);
        for (const auto &service : Host::Services(*host)) {
            monitoring_log.state(*service, log_state_type);
        }
    }
}

void World::logTimeperiods(MonitoringLog &monitoring_log,
                           std::chrono::system_clock::time_point now) {
    // Advance all timeperiods to their active state first. This avoid duplicate
    // logging.
    for (const auto &[name, timeperiod] : config_.timeperiods_) {
        timeperiod->update(nullptr, now);
        monitoring_log.timeperiodTransition(
            timeperiod->name(), -1, static_cast<int>(timeperiod->isActive()));
    }
}

std::ostream &operator<<(std::ostream &os, const World &w) {
    return os << "configuration " << w.config_.global_config_.general_.serial_
              << " (" << static_cast<const void *>(&w) << ") from "
              << FormattedTimePoint(w.creation_time_) << " with "
              << w.num_hosts_ << " hosts and " << w.num_services_
              << " services";
}
