// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

// Needed for kill
// NOLINTNEXTLINE(cppcoreguidelines-macro-usage,bugprone-reserved-identifier,cert-dcl37-c,cert-dcl51-cpp)
#define _XOPEN_SOURCE 500

// NOTE: The pragma below is needed, without the header we get compilation
// errors.
#include <fcntl.h>
#include <features.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <unistd.h>

#include <cerrno>
#include <csignal>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <iterator>
#include <sstream>  // IWYU pragma: keep
#include <string>
#include <system_error>
#include <vector>

pid_t g_pid;  // NOLINT(cppcoreguidelines-avoid-non-const-global-variables)

namespace {
// Propagate signal to child, if we are killed
[[noreturn]] void term_handler(int signum) {
    if (g_pid != 0) {
        kill(g_pid, signum);
    }
    exit(0);
}

// handle check timeout
void alarm_handler(int /*signum*/) {
    if (g_pid != 0) {
        kill(-g_pid, SIGKILL);
    }
}
}  // namespace

// TODO(sp): Somehow re-use the same function from daemonize.cc
namespace mk {
void closefrom(int fd) {
#if __GLIBC_PREREQ(2, 34)
    // BSDs, Solaris, and Glibc 2.34+ have this already.
    ::closefrom(fd);
#else
    // There are tons of more or less portable and sensible ways to retrieve
    // the maximum FD (sysconf, scanning /proc/self/fd, etc.), but we try
    // only the simplest one here (getrlimit), which seems to be good
    // enough.
    rlimit limit;
    auto ret = getrlimit(RLIMIT_NOFILE, &limit);
    auto max_fd = ret == -1 ? 1024 : static_cast<int>(limit.rlim_cur);
    while (fd < max_fd) {
        ::close(fd++);
    }
#endif
}
}  // namespace mk

namespace {
int check_fd(int fd) {
    if (fd != -1) {
        return fd;
    }
    switch (errno) {
        case EMFILE:   // too many files open
        case ENFILE:   // too many files open in system
        case ENOBUFS:  // no buffer space
        case ENOMEM:   // not enough memory
            exit(127);
        default:
            break;
    }
    return fd;
}

void prepare_child_file_descriptors(int *fd) {
    ::close(fd[0]);              // close read end
    check_fd(::dup2(fd[1], 1));  // point stdout into pipe
    check_fd(::dup2(fd[1], 2));  // also point stderr into pipe
    ::close(fd[1]);              // lives forth in 1 and 2
    const int f = check_fd(::open("/dev/null", O_RDONLY));
    check_fd(::dup2(f, 0));
    ::close(f);
}

void adjust_stack_limit(rlim_t limit) {
    struct rlimit rl;
    getrlimit(RLIMIT_STACK, &rl);
    rl.rlim_cur = limit;
    rl.rlim_max = limit;
    setrlimit(RLIMIT_STACK, &rl);
}

[[noreturn]] void execute(const std::string &command) {
    // Optimization: If we have an absolute path to the executable and no
    // special characters on the command line, call execv directly, saving 2
    // forks and 1 shell.
    if (command.starts_with("/") &&
        command.find_first_of(R"("'<>;&`|)") == std::string::npos) {
        std::istringstream buffer{command};
        const std::vector<std::string> args{
            std::istream_iterator<std::string>{buffer},
            std::istream_iterator<std::string>{}};
        std::vector<const char *> cmd_argv;
        cmd_argv.reserve(args.size());
        for (const auto &arg : args) {
            cmd_argv.push_back(arg.c_str());
        }
        cmd_argv.push_back(nullptr);
        // http://pubs.opengroup.org/onlinepubs/9699919799/functions/exec.html
        // explains in its rationale why the const_cast is
        // actually OK and only caused by a histortic artifact.
        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-const-cast)
        ::execv(cmd_argv[0], const_cast<char *const *>(cmd_argv.data()));
    } else {
        const int ret = system(command.c_str());
        if (WIFEXITED(ret)) {
            exit(WEXITSTATUS(ret));
        }
    }
    exit(127);
}

std::string error_string() {
    return std::error_code{errno, std::generic_category()}.message();
}

}  // namespace

/* This program is called by the Check_MK Micro Core without any arguments. */
int main(int /*argc*/, char **argv) {
    char output[65536];
    char *end_of_output = nullptr;

    (void)signal(SIGALRM, alarm_handler); /* handler for check timeout */
    (void)signal(SIGINT, term_handler);
    (void)signal(SIGQUIT, term_handler);
    (void)signal(SIGTERM, term_handler);

    mk::closefrom(3);

    /* Protocol: read / line by line: */
    /* 1. host name */
    /* 2. command to execute */
    /* 3. timeout in seconds */
    /* Send as an answer: */
    /* 1. Exit status: "0\n", "1\n", "2\n" or "3\n" */
    /* 2. ... */

    while (true) {
        std::string host_name_string;
        if (!std::getline(std::cin, host_name_string)) {
            exit(0); /* CMC closed socket */
        }
        const auto *host_name = host_name_string.c_str();

        std::string command;
        if (!std::getline(std::cin, command)) {
            exit(0); /* CMC closed socket */
        }

        if (command == "*") { /* restart myself */
            (void)fflush(stdout);
            ::execvp(argv[0], argv);
        }

        std::string timeout_str;
        if (!std::getline(std::cin, timeout_str)) {
            exit(0); /* CMC closed socket. */
        }
        const int timeout = atoi(timeout_str.c_str());

        int return_code{0};
        int fd[2];
        if (pipe(fd) == -1) {
            return_code = 3;
            end_of_output =
                output + snprintf(output, sizeof(output),
                                  "%s\tcannot create pipe: %s", host_name,
                                  error_string().c_str());
        } else {
            errno = 0;
            const int pid = ::fork();
            switch (pid) {
                case -1: { /* fail */
                    const std::string err = error_string();
                    std::cerr << "Unable to fork for check execution: " << err
                              << " - Stopping CheckHelper process.\n";
                    exit(1);
                }
                case 0: { /* child process */
                    // Assign this child its own process group, so we can kill
                    // its entire process group when a timeout occurs
                    setpgid(getpid(), 0);

                    prepare_child_file_descriptors(fd);
                    adjust_stack_limit(rlim_t{32} * 1024 * 1024);
                    execute(command);
                }
                default: { /* parent process */
                    ::close(fd[1]);
                    g_pid = pid;
                    if (timeout != 0) {
                        alarm(timeout);
                    }

                    end_of_output = output + snprintf(output, sizeof(output),
                                                      "%s\t", host_name);
                    const char *ptr_end = output + sizeof(output) - 1;
                    ssize_t bytes_read{0};
                    while ((bytes_read = ::read(fd[0], end_of_output,
                                                ptr_end - end_of_output)) !=
                           0) {
                        end_of_output += bytes_read;
                        if (end_of_output == ptr_end) {
                            break;
                        }
                    }
                    *end_of_output = 0;

                    ::close(fd[0]);
                    int ret{0};
                    while (waitpid(pid, &ret, 0) == -1 && errno == EINTR) {
                    }
                    g_pid = 0;
                    alarm(0); /* cancel timeout */

                    if (WIFSIGNALED(ret)) {
                        const int signum = WTERMSIG(ret);
                        if (signum == SIGKILL) {
                            end_of_output =
                                output +
                                snprintf(
                                    output, sizeof(output),
                                    "%s\t(Check timed out after %d seconds)\n",
                                    host_name, timeout);
                            return_code = 3;
                        } else {
                            end_of_output =
                                output +
                                snprintf(
                                    output, sizeof(output),
                                    "%s\t(Check plugin died with signal %d)\n",
                                    host_name, signum);
                            return_code = 3;
                        }
                    } else {
                        return_code = WEXITSTATUS(ret);
                        if (*output == 0 || *output == '\n') {
                            end_of_output =
                                output +
                                snprintf(
                                    output, sizeof(output),
                                    "%s\t(No output returned from plugin)\n",
                                    host_name);
                        }
                    }
                }
            }
        } /* pipe created successfully */

        /* Send check result to the CMC */
        /* 1. status code (0: OK, etc.) */
        /* 2. size of check output in bytes */
        /* 3. host name */
        /* 4. checkoutput (without extra newline) */
        printf("%03d\n%08td\n", return_code, end_of_output - output);
        (void)fwrite(output, end_of_output - output, 1, stdout);
        (void)fflush(stdout);
    } /* end of main loop */
}
