// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include "cmc/daemonize.h"

#include <fcntl.h>
#include <features.h>
#include <sys/resource.h>  // IWYU pragma: keep
#include <unistd.h>

#include <cstdlib>
#include <stdexcept>
#include <string>
#include <utility>

#include "livestatus/Logger.h"

PIDFile::PIDFile(std::filesystem::path filename)
    : _filename{std::move(filename)} {}

PIDFile::~PIDFile() {
    if (_os.is_open()) {
        ::unlink(_filename.c_str());
    }
}

void PIDFile::write() {
    _os.open(_filename);
    _os << getpid() << "\n" << std::flush;
    if (!_os) {
        throw generic_error{"could not write pid into \"" + _filename.string() +
                            "\""};
    }
    // Note: We do *not* close _os yet, otherwise we would lose any lock on the
    // file. POSIX locks are associated with a process/inode pair, not with a
    // file descriptor.
}

void PIDFile::writeLocked() {
    // Make sure the file exists, but do not truncate it yet.
    if (!std::ofstream(_filename, std::ios::app)) {
        throw generic_error{"could not create pid file \"" +
                            _filename.string() + "\""};
    }
    file_lock flock(_filename.c_str());
    if (!flock.try_lock()) {
        throw std::runtime_error("already running");
    }
    using std::swap;
    swap(flock, _flock);
    write();
}

void daemonize(PIDFile &pid_file) {
    switch (::fork()) {
        case -1:
            throw generic_error("could not daemonize, 1st fork_failed");
        case 0:
            break;  // child continues, but is not a process group leader
        default:
            exit(EXIT_SUCCESS);  // parent terminates, caller can continue
    }

    // become the leader of a new session, no controlling terminal anymore
    if (setsid() == -1) {
        throw generic_error("could not daemonize, setsid failed");
    }

    // make sure we never reacquire the controlling terminal
    switch (::fork()) {
        case -1:
            throw generic_error("could not daemonize, 2nd fork_failed");
        case 0:
            break;  // child continues
        default:
            exit(EXIT_SUCCESS);  // parent terminates
    }
    // TODO(sp): Set umask to 0 and/or chdir to "/"?

    mk::closefrom(3);
    pid_file.writeLocked();
    ::close(0);
    ::close(1);
    ::close(2);

    // re-open stdin/stdout/stderr descriptors on /dev/null, so any I/O on them
    // won't fail and we don't accidentally get those filedescriptors again.
    const int fd0 = ::open("/dev/null", O_RDWR);
    const int fd1 = ::dup2(fd0, 1);
    const int fd2 = ::dup2(fd0, 2);
    if (fd0 != STDIN_FILENO || fd1 != STDOUT_FILENO || fd2 != STDERR_FILENO) {
        throw std::runtime_error(
            "unexpected file descriptors: " + std::to_string(fd0) + ", " +
            std::to_string(fd1) + ", " + std::to_string(fd2));
    }

    // Note: No point in logging anything here because of the above *until* we
    // open our logfile later...
}

namespace mk {
void closefrom(int fd) {
#if __GLIBC_PREREQ(2, 34)
    // BSDs, Solaris, and Glibc 2.34+ have this already.
    ::closefrom(fd);
#else
    // There are tons of more or less portable and sensible ways to retrieve the
    // maximum FD (sysconf, scanning /proc/self/fd, etc.), but we try only the
    // simplest one here (getrlimit), which seems to be good enough.
    rlimit limit;
    auto ret = getrlimit(RLIMIT_NOFILE, &limit);
    auto max_fd = ret == -1 ? 1024 : static_cast<int>(limit.rlim_cur);
    while (fd < max_fd) {
        ::close(fd++);
    }
#endif
}
}  // namespace mk
