// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <arpa/inet.h>
#include <netinet/if_ether.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <pcap.h>
#include <sys/types.h>
#include <unistd.h>

#include <cerrno>
#include <chrono>
#include <cstdint>
#include <cstdlib>
#include <iostream>
#include <string>
#include <system_error>
#include <thread>

using namespace std::chrono_literals;
using namespace std::string_literals;

/* Fix for older libpcap */
#ifndef PCAP_NETMASK_UNKNOWN
#define PCAP_NETMASK_UNKNOWN 0xffffffff
#endif

/* Fix for older distros with ancient <linux/if_ether.h> header */
#ifndef ETH_P_8021AD
#define ETH_P_8021AD 0x88A8
#endif

/*
   IP header according to RFC 791 (see struct ip in <netinet/ip.h>, too):

    0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |Version|  IHL  |Type of Service|          Total Length         |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |         Identification        |Flags|      Fragment Offset    |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |  Time to Live |    Protocol   |         Header Checksum       |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                       Source Address                          |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                    Destination Address                        |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                    Options                    |    Padding    |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
*/

constexpr int pcap_snaplen = 64;

int bail_out(int exitcode) {
    std::this_thread::sleep_for(3s);
    exit(exitcode);
}

pcap_t *setup_pcap(bool ignore_rst) {
    char error[PCAP_ERRBUF_SIZE];
    error[0] = 0;
    pcap_t *pcap = pcap_open_live(nullptr, pcap_snaplen, 0, 1000, error);
    if (pcap == nullptr) {
        std::cerr << "Error at pcap_open(): " << error << "\n";
        bail_out(1);
    }
    struct bpf_program bpf;
    // TODO(sp) Note that these filters accept both IPv4 and IPv6! Should we
    // change them to accept only IPv4? Or should we extend our protocol with
    // the Micro Core to handle both IPv4 and IPv6?
    const char *filter =
        ignore_rst
            ? "icmp or ( tcp and ( tcp[tcpflags] &  tcp-syn          != 0 ) )"
            : "icmp or ( tcp and ( tcp[tcpflags] & (tcp-syn|tcp-rst) != 0 ) )";
    // NOTE: Old pcap headers are broken (e.g. con CentOS 5), so we need the
    // const_cast below.
    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-const-cast)
    if (pcap_compile(pcap, &bpf, const_cast<char *>(filter), 1,
                     PCAP_NETMASK_UNKNOWN) != 0) {
        std::cerr << "Error at pcap_compile(): " << pcap_geterr(pcap) << "\n";
        bail_out(1);
    }
    if (pcap_setdirection(pcap, PCAP_D_IN) != 0) {
        std::cerr << "Error at pcap_setdirection(): " << pcap_geterr(pcap)
                  << "\n";
        bail_out(1);
    }
    if (pcap_setfilter(pcap, &bpf) != 0) {
        std::cerr << "Error at pcap_setfilter(): " << pcap_geterr(pcap) << "\n";
        bail_out(1);
    }

    // We capture packets from all interfaces (see pcap_open_live call above),
    // so we expect a link-layer header type DLT_LINUX_SLL (113), see the spec
    // at http://www.tcpdump.org/linktypes/LINKTYPE_LINUX_SLL.html.  For all
    // link-layer header types see: <pcap/dlt.h>
    const int dlt = pcap_datalink(pcap);
    if (dlt != DLT_LINUX_SLL) {
        const auto *name = pcap_datalink_val_to_name(dlt);
        std::cerr << "Unexpected link-layer header type "
                  << (name == nullptr ? "<unknown dlt>" : name)
                  << ", expected LINUX_SLL\n";
        bail_out(1);
    }
    return pcap;
}

// There doesn't seem to be a header which contains the SLL header structure, so
// we have to define one for ourselves. See tcpdump's print-sll.c sources and
// the spec at http://www.tcpdump.org/linktypes/LINKTYPE_LINUX_SLL.html.
struct sll_header {
    uint16_t sll_pkttype;   // packet type
    uint16_t sll_hatype;    // link-layer address type
    uint16_t sll_halen;     // link-layer address length
    uint8_t sll_addr[8];    // link-layer address
    uint16_t sll_protocol;  // protocol
};

void write_full(int fd, const void *buf, size_t count) {
    while (count > 0) {
        const ssize_t bytes_written = ::write(fd, buf, count);
        if (bytes_written == -1) {
            auto err =
                std::error_code{errno, std::generic_category()}.message();
            std::cerr << "Cannot write to FD " << fd << ": " << err << "\n";
            bail_out(1);
        }
        buf = reinterpret_cast<const char *>(buf) + bytes_written;
        count -= bytes_written;
    }
}

uint16_t get_protocol(const u_char *data) {
    // NOLINTNEXTLINE(readability-isolate-declaration)
    return ntohs(reinterpret_cast<const sll_header *>(data)->sll_protocol);
}

// Ugly casts caused by old-skool C callback interface. :-P
u_char *from_flag(bool *flag) { return reinterpret_cast<u_char *>(flag); }
bool *to_flag(u_char *args) { return reinterpret_cast<bool *>(args); }

void handle_packet(u_char *args, const struct pcap_pkthdr * /*unused*/,
                   const u_char *data) {
    // We handle *only* simple IPv4 packets below, so we silently drop other
    // packets here. In principle, we could handle other types, too, e.g. some
    // encapsulated types, but we would need some code to recursively unwrap
    // those packets, see e.g. tcpdump's sll_if_print function in print-sll.c
    auto ether_type = get_protocol(data);
    if (ether_type == ETH_P_8021AD) {  // skip outer 802.1ad Service VLAN header
        ether_type = get_protocol(data += 4);
    }
    if (ether_type == ETH_P_8021Q) {  // skip inner 802.1Q VLAN Extended Header
        ether_type = get_protocol(data += 4);
    }
    if (ether_type != ETH_P_IP) {
        // There are far too many esoteric packets out there in the wild, so we
        // should better not log by default.
        const bool log_unknown_packets = *to_flag(args);
        if (log_unknown_packets) {
            std::cerr << "ignoring packet with unknown EtherType " << ether_type
                      << "\n"
                      << std::flush;
        }
        return;
    }

    // Inform the core about the packet from this IPv4 address.
    const auto *ip =
        reinterpret_cast<const struct ip *>(data + sizeof(sll_header));
    write_full(1, &ip->ip_src.s_addr, sizeof(in_addr_t));
}
void main_loop(pcap_t *pcap, bool log_unknown_packets) {
    while (true) {
        if (pcap_loop(pcap, 1, handle_packet,
                      from_flag(&log_unknown_packets)) == -1) {
            std::cerr << "Error in pcap_loop(): " << pcap_geterr(pcap) << "\n"
                      << std::flush;
        }
    }
}

// Usage: icmpreceiver [--ignore-rst]
int main(int argc, char **argv) {
    const bool ignore_rst = argc == 2 && argv[1] == "--ignore-rst"s;
    // TODO(sp) This should be configurable via a commandline flag, and the CMC
    // should pass this when smart ping log level is "debug".
    const bool log_unknown_packets = false;
    pcap_t *pcap = setup_pcap(ignore_rst);
    main_loop(pcap, log_unknown_packets);
    pcap_close(pcap);
    return 0;
}
