// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/ip_icmp.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>

#include <algorithm>
#include <cerrno>
#include <chrono>
#include <compare>
#include <csignal>
#include <cstdint>
#include <cstdlib>
#include <iostream>
#include <queue>
#include <ratio>
#include <string>
#include <system_error>
#include <unordered_set>
#include <vector>

#include "livestatus/ChronoUtils.h"
#include "livestatus/Poller.h"

using namespace std::chrono_literals;
using namespace std::string_literals;

void sendReply(const std::string &msg) {
    std::cerr << msg << "\n" << std::flush;
}

void sendError(const std::string &msg) {
    sendReply(msg + ": " +
              std::error_code{errno, std::generic_category()}.message());
}

struct Options {
    // Usage: icmpsender [NUM_SOCKETS WAIT_TIME BULK_SIZE] [--omit-payload]
    // WAIT_TIME is in us
    Options(int argc, char **argv) {
        if (argc <= 1) {
            return;
        }

        if (argc >= 4) {
            num_sockets = atoi(argv[1]);
            wait_time = std::chrono::microseconds(atoi(argv[2]));
            bulk_size = atoi(argv[3]);
        }

        omit_payload = argc >= 2 && argv[argc - 1] == "--omit-payload"s;
    }

    bool omit_payload{false};  // if true do not put payload into ICMP packet
    int num_sockets{8};        // number of raw sockets to open
    int bulk_size{1000};       // number of packets to send at once
    std::chrono::microseconds wait_time{
        0};  // wait time between two bulks sends
};

class SocketPool {
public:
    // TODO(sp) Use RAII for a socket and remove the constructor/destructor
    // stuff below.
    SocketPool() = default;
    SocketPool(const SocketPool &) = delete;
    SocketPool(SocketPool &&) = delete;
    SocketPool &operator=(SocketPool) = delete;
    SocketPool &operator=(SocketPool &&) = delete;
    ~SocketPool() {
        for (const auto &sock : sockets_) {
            ::close(sock);
        }
    }

    void push_back(int fd) { sockets_.push_back(fd); }
    [[nodiscard]] auto size() const { return sockets_.size(); }
    auto operator[](size_t n) const { return sockets_[n]; }

private:
    std::vector<int> sockets_;
};

// .-Job handling----------------------------------------------------------.
// |         _       _       _                     _ _ _                   |
// |        | | ___ | |__   | |__   __ _ _ __   __| | (_)_ __   __ _       |
// |     _  | |/ _ \| '_ \  | '_ \ / _` | '_ \ / _` | | | '_ \ / _` |      |
// |    | |_| | (_) | |_) | | | | | (_| | | | | (_| | | | | | | (_| |      |
// |     \___/ \___/|_.__/  |_| |_|\__,_|_| |_|\__,_|_|_|_| |_|\__, |      |
// |                                                           |___/       |
// +-----------------------------------------------------------------------+
// |  Keep a queue of PINGs to be done.                                    |
// '-----------------------------------------------------------------------'

// Limit number of error messages to one every 10 seconds
constexpr auto overflow_warn_cooldown = 10s;

void overflow_warning() {
    static std::chrono::system_clock::time_point last_warning_at{};
    auto now = std::chrono::system_clock::now();
    if (now - last_warning_at >= overflow_warn_cooldown) {
        // TODO(mk): This does not work. Multiple hosts might have the same IP
        // address. Simplify the code again. Remove the _set. We might rather
        // add a timestamp to each job in the queue and issue a warning if the
        // latency (time between receiving the job and sending the PING) is
        // being too large.
        last_warning_at = now;
    }
}

class JobQueue {
public:
    JobQueue() = default;
    void read_input() {
        // TODO(sp) Use a standard C++ container instead of this highly
        // error-prone unreadable hand-made buffer fiddling.

        // buffer size for receiving destination addresses from stdin
        constexpr size_t in_buffer_size = 4096;
        char in_buffer[in_buffer_size];

        // Some bytes left over from last time?
        if (num_left_over_ > 0) {
            std::move(left_over_buffer_, left_over_buffer_ + num_left_over_,
                      in_buffer);
        }

        // Get new IP addresses from the core
        ssize_t bytes_left = ::read(0, in_buffer + num_left_over_,
                                    in_buffer_size - num_left_over_);
        if (bytes_left == 0) {
            sendReply("Master closed pipe. Exiting.");
            exit(0);
        } else if (bytes_left < 0) {
            sendError("Could not read from stdin");
            exit(1);
        }
        bytes_left += num_left_over_;

        char *read_pos = in_buffer;
        while (static_cast<size_t>(bytes_left) >= sizeof(in_addr_t)) {
            const in_addr_t dest_address =
                *reinterpret_cast<in_addr_t *>(read_pos);
            push(dest_address);
            read_pos += sizeof(in_addr_t);
            bytes_left -= sizeof(in_addr_t);
        }

        // Store bytes from incomplete addresses for next time
        if (bytes_left > 0) {
            std::move(read_pos, read_pos + bytes_left, left_over_buffer_);
        }

        num_left_over_ = bytes_left;
    }

    bool empty() { return queue_.empty(); }

    in_addr_t pop() {
        const in_addr_t addr = queue_.front();
        queue_.pop();
        set_.erase(addr);
        return addr;
    }

private:
    std::unordered_set<in_addr_t> set_;  // fast lookup for duplicate job
    std::queue<in_addr_t> queue_;        // keep order of jobs

    // remember incomplete IP addresses until next call
    char left_over_buffer_[16];

    // current number of bytes in left_over_buffer
    size_t num_left_over_{0};

    void push(in_addr_t addr) {
        if (set_.contains(addr)) {
            overflow_warning();
            return;
        }
        queue_.push(addr);
        set_.insert(addr);
    }
};

// .-PING------------------------------------------------------------------.
// |                        ____ ___ _   _  ____                           |
// |                       |  _ \_ _| \ | |/ ___|                          |
// |                       | |_) | ||  \| | |  _                           |
// |                       |  __/| || |\  | |_| |                          |
// |                       |_|  |___|_| \_|\____|                          |
// |                                                                       |
// +-----------------------------------------------------------------------+
// |  Code for sending ICMP echo requests                                  |
// '-----------------------------------------------------------------------'

/*
   IP header according to RFC 791:

    0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |Version|  IHL  |Type of Service|          Total Length         |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |         Identification        |Flags|      Fragment Offset    |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |  Time to Live |    Protocol   |         Header Checksum       |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                       Source Address                          |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                    Destination Address                        |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                    Options                    |    Padding    |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


   ICMP echo / echo-reply message according to RFC 792:

    0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |     Type      |     Code      |          Checksum             |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |           Identifier          |        Sequence Number        |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |     Data ...
   +-+-+-+-+-

*/

struct icmp_ping_data_t {
    struct timeval timestamp;
};

// NOLINTNEXTLINE(cppcoreguidelines-avoid-non-const-global-variables)
icmp_ping_data_t icmp_ping_data;

const char payload_numbers[40] = {
    0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19,
    0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20, 0x21, 0x22, 0x23,
    0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d,
    0x2e, 0x2f, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
};

// Each socket is allowed to have a send buffer of 2 MB.  This should be way too
// much, but just to be sure.
constexpr int so_sndbuf_size = 2 * 1024 * 1024;

constexpr size_t icmp_min_pkt_size = ICMP_MINLEN;

constexpr size_t icmp_max_pkt_size =
    ICMP_MINLEN + sizeof(icmp_ping_data) + sizeof(payload_numbers);

constexpr uint32_t icmp_min_id = 32769;
constexpr uint32_t icmp_max_id = 65535;

// NOLINTNEXTLINE(cppcoreguidelines-avoid-non-const-global-variables)
char ping_buffer[icmp_max_pkt_size];

void setup_ping(SocketPool &socket_pool, int num_sockets) {
    // Prepare ICMP echo request packet to be reused for all pings
    // Only the sequence number and target IP address will be set later.
    std::fill(ping_buffer, ping_buffer + sizeof(ping_buffer), 0);

    // Set ICMP header fields. Checksum is being filled later for each packet.
    auto *ping_packet = reinterpret_cast<struct icmp *>(&ping_buffer[0]);
    ping_packet->icmp_type = ICMP_ECHO;
    ping_packet->icmp_code = 0;  // echo request
    // We use always sequence number 0 (like a ping -c 1)
    // NOLINTNEXTLINE(readability-isolate-declaration)
    ping_packet->icmp_seq = htons(1);

    // The payload is not being used if --omit-payload is set, but anyway filled
    std::move(&payload_numbers[0], &payload_numbers[sizeof(payload_numbers)],
              reinterpret_cast<char *>(&ping_packet->icmp_data) +
                  sizeof(icmp_ping_data));

    // Open all needed sockets
    for (int i = 0; i < num_sockets; i++) {
        const int sock = ::socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
        if (sock < 0) {
            sendError("Cannot create raw socket (missing SUID root?)");
            exit(1);
        }

        unsigned char ttl = 64;
        if (0 != setsockopt(sock, SOL_IP, IP_TTL, &ttl, sizeof(ttl))) {
            sendError("Cannot set TTL to " + std::to_string(ttl));
            exit(1);
        }

        // Set send buffer to a size we can be sure that no "no buffer space
        // available" situation occurs. We try to set a buffer that exceeds the
        // /proc/sys/net/core/wmem_max limit on some systems.  To make that
        // possible, we need to set a socket option which makes the limitation
        // be ignored (since Linux 2.6.14).
        int opt = so_sndbuf_size;
        if (0 != setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &opt, sizeof(opt))) {
            sendError("Cannot set send buffer to " + std::to_string(opt));
            exit(1);
        }

        socket_pool.push_back(sock);
    }
}

uint16_t get_next_icmp_id() {
    static uint32_t next_id = icmp_min_id - 1;

    next_id++;

    // If the actual ICMP id corresponds to the PID of an existing process
    // we have to find the next one which does not
    while (kill(next_id, 0) == 0) {
        next_id++;
    }

    if (next_id > icmp_max_id) {
        next_id = icmp_min_id;

        // Since next_id was reset we have to check again that it does not
        // collide with
        // any existing PID
        while (kill(next_id, 0) == 0) {
            next_id++;
        }
    }
    // Reset errno: we do not care whether kill errored or not.
    errno = 0;

    return static_cast<uint16_t>(next_id);
}

// For the algorithm see: https://www.rfc-editor.org/rfc/rfc1071.txt
uint16_t icmp_checksum(void *buffer, size_t n) {
    auto *p = static_cast<uint16_t *>(buffer);

    // sum up 16 bit chunks
    uint64_t sum = 0;
    while (n > 1) {
        sum += *p++;
        n -= 2;
    }

    // Add left-over byte, if any
    if (n == 1) {
        sum += *reinterpret_cast<uint8_t *>(p);
    }

    // fold checksum into 16 bits
    while (sum > 0XffffU) {
        sum = (sum >> 16) + (sum & 0xffffU);
    }

    // one's complement
    return ~static_cast<uint16_t>(sum);
}

// TODO(sp) This should be a method of SocketPool
void send_ping(SocketPool &socket_pool, bool omit_payload, in_addr_t a) {
    sockaddr_in addr;
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = a;

    const int packet_size =
        omit_payload ? icmp_min_pkt_size : icmp_max_pkt_size;
    const uint16_t next_icmp_id = get_next_icmp_id();

    auto *ping_packet = reinterpret_cast<struct icmp *>(&ping_buffer[0]);
    ping_packet->icmp_cksum = 0;
    // NOLINTNEXTLINE(readability-isolate-declaration)
    ping_packet->icmp_id = htons(next_icmp_id);

    // Add timestamp, if payload is being used
    if (!omit_payload) {
        icmp_ping_data.timestamp =
            to_timeval(std::chrono::system_clock::now().time_since_epoch());
        const char *pd = reinterpret_cast<const char *>(&icmp_ping_data);
        std::move(pd, pd + sizeof(icmp_ping_data),
                  reinterpret_cast<char *>(&ping_packet->icmp_data));
    }

    // Compute checksum
    ping_packet->icmp_cksum = icmp_checksum(ping_packet, packet_size);

    const size_t socket_index = next_icmp_id % socket_pool.size();
    const ssize_t len = sendto(
        socket_pool[socket_index], ping_packet, packet_size, 0,
        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
        reinterpret_cast<struct sockaddr *>(&addr), sizeof(struct sockaddr));

    if (len != packet_size) {
        const in_addr addr_for_msg = {a};
        sendError("SOCK" + std::to_string(socket_index) + ": Could not send " +
                  std::to_string(packet_size) + " bytes of ICMP data to " +
                  inet_ntoa(addr_for_msg));
    }
}

// .-Main loop-------------------------------------------------------------.
// |             __  __       _         _                                  |
// |            |  \/  | __ _(_)_ __   | | ___   ___  _ __                 |
// |            | |\/| |/ _` | | '_ \  | |/ _ \ / _ \| '_ \                |
// |            | |  | | (_| | | | | | | | (_) | (_) | |_) |               |
// |            |_|  |_|\__,_|_|_| |_| |_|\___/ \___/| .__/                |
// |                                                 |_|                   |
// +-----------------------------------------------------------------------+
// |  Main code for doing the actual work                                  |
// '-----------------------------------------------------------------------'

/* Protocol: read 4 bytes from stdin. Send a ping to that address. */
void main_loop(SocketPool &socket_pool, const Options &options) {
    JobQueue jobs;
    while (true) {
        // No jobs queued, wait blocking for new jobs
        jobs.read_input();

        while (!jobs.empty()) {
            // Send one bulk of pings
            for (int i = 0; i < options.bulk_size && !jobs.empty(); i++) {
                send_ping(socket_pool, options.omit_payload, jobs.pop());
            }

            // In case the user configured a wait time we wait a few
            // microseconds in order not to flood clumsy firewalls that cannot
            // handle our high ping rate.
            if (!jobs.empty() && options.wait_time != 0s) {
                auto wait_time = options.wait_time;
                while (wait_time > 0s) {
                    auto start = std::chrono::steady_clock::now();
                    if (Poller{}.wait(wait_time, 0, PollEvents::in)) {
                        // Input pipe is readable, new jobs arrived (or EOF)
                        jobs.read_input();
                    } else if (errno != ETIMEDOUT) {
                        sendError("Polling failed");
                        exit(1);
                    }
                    wait_time -=
                        std::chrono::duration_cast<std::chrono::microseconds>(
                            std::chrono::steady_clock::now() - start);
                }
            }
        }
    }
}

// .-main------------------------------------------------------------------.
// |                                       _                               |
// |                       _ __ ___   __ _(_)_ __                          |
// |                      | '_ ` _ \ / _` | | '_ \                         |
// |                      | | | | | | (_| | | | | |                        |
// |                      |_| |_| |_|\__,_|_|_| |_|                        |
// |                                                                       |
// '-----------------------------------------------------------------------'

int main(int argc, char **argv) {
    const Options options{argc, argv};
    SocketPool socket_pool;  // list of FDs of ping sockets
    setup_ping(socket_pool, options.num_sockets);
    main_loop(socket_pool, options);
}
