// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <getopt.h>
#include <pwd.h>
#include <unistd.h>

#include <chrono>
#include <cstdlib>
#include <exception>
#include <filesystem>
#include <fstream>
#include <iostream>
#include <memory>
#include <numeric>
#include <stdexcept>
#include <string>
#include <system_error>
#include <utility>

#include "cmc/Core.h"
#include "cmc/GlobalConfig.h"
#include "cmc/GlobalState.h"
#include "cmc/PbConfigLoader.h"
#include "cmc/PbStateLoader.h"
#include "cmc/daemonize.h"
#include "config/v1/types.pb.h"
#include "google/protobuf/stubs/common.h"
#include "livestatus/Logger.h"
#include "livestatus/POSIXUtils.h"
#include "livestatus/TrialManager.h"
#include "state/v1/types.pb.h"

namespace fs = std::filesystem;

namespace {
fs::path getOMDRoot() {
    struct passwd *pw = getpwuid(getuid());
    if (pw == nullptr || pw->pw_dir == nullptr) {
        throw std::runtime_error(
            "cannot determine home directory of user with id " +
            std::to_string(getuid()));
    }
    return pw->pw_dir;
}

fs::path get_default_config_path() {
    return getOMDRoot() / "var/check_mk/core/config.pb";
}

fs::path get_pid_file_path() { return getOMDRoot() / "tmp/run/cmc.pid"; }

void print_usage(std::ostream &os, const std::string &prog_name) {
    os << "Usage: " << prog_name
       << "  [OPTIONS] [CONFIGFILE]\n"
          "\n"
          "  -V,  --version     print version an exit\n"
          "  -h,  --help        this cruft\n"
          "  -g,  --foreground  do not daemonize, run in the foreground\n"
          "\n"
          "CONFIGFILE is optional. The default value is "
       << get_default_config_path() << ".\n";
}

void print_version() {
    std::cout << "Check_MK Micro Core - Version " VERSION
                 "\nCopyright Mathias Kettner 2013 (mk@mathias-kettner.de)\n";
}

struct CmcOptions {
    CmcOptions(fs::path p, bool b)
        : path_to_config(std::move(p)), background(b) {}
    fs::path path_to_config;
    bool background;
};

const struct option long_options[] = {
    {"version", no_argument, nullptr, 'V'},
    {"help", no_argument, nullptr, 'h'},
    {"foreground", no_argument, nullptr, 'g'},
    {nullptr, 0, nullptr, 0},
};

CmcOptions getCmcOptions(int argc, char **argv) {
    const std::string prog_name(argv[0]);
    bool background = true;

    while (true) {
        int indexptr = 0;
        const int ret = getopt_long(argc, argv, "Vhg", long_options, &indexptr);
        if (ret == -1) {
            break;
        }
        switch (ret) {
            case 'V':
                print_version();
                exit(EXIT_SUCCESS);
            case 'h':
                print_usage(std::cout, prog_name);
                exit(EXIT_SUCCESS);
            case 'g':
                background = false;
                break;
            default:
                print_usage(std::cerr, prog_name);
                exit(EXIT_FAILURE);
        }
    }
    return {
        (optind < argc) ? fs::path{argv[optind]} : get_default_config_path(),
        background};
}

fs::path getAnticipatedLogFile() {
    auto log_file_dir = getOMDRoot() / "var/log";
    std::error_code ec;
    return fs::exists(log_file_dir, ec)
               ? log_file_dir / "cmc.log"
               : fs::path{"/tmp/cmc-" + std::to_string(::getpid()) + ".log"};
}

std::chrono::system_clock::time_point get_default_creation_time(
    const checkmk::cmc::state::v1::StateFile *state_file,
    std::chrono::system_clock::time_point now) {
    if (state_file == nullptr) {
        return now;  // first start
    }
    auto timestamp = PbStateLoader::loadGlobalState(state_file->global_state())
                         .getStateFileCreated();
    return timestamp == std::chrono::system_clock::time_point{}
               ? now         // >= 2.2 state files have an "empty" timestamp
               : timestamp;  // migrate from < 2.2 state file
}
}  // namespace

int main(int argc, char **argv) {
    GOOGLE_PROTOBUF_VERIFY_VERSION;

    setThreadName("main");

    Logger *logger_cmk = Logger::getLogger("cmk");
    logger_cmk->setUseParentHandlers(false);

    // TODO(sp) This is a cruel hack for a hen-and-egg problem: How to configure
    // logging before we have read our configuration? Do this better...
    Logger::getLogger("cmk.alert")->setHandler(nullptr);
    Logger::getLogger("cmk.carbon")->setHandler(nullptr);
    Logger::getLogger("cmk.core")->setHandler(nullptr);
    Logger::getLogger("cmk.downtime")->setHandler(nullptr);
    Logger::getLogger("cmk.helper")->setHandler(nullptr);
    Logger::getLogger("cmk.livestatus")->setHandler(nullptr);
    Logger::getLogger("cmk.notification")->setHandler(nullptr);
    Logger::getLogger("cmk.rrd")->setHandler(nullptr);
    Logger::getLogger("cmk.influxdb")->setHandler(nullptr);
    Logger::getLogger("cmk.smartping")->setHandler(nullptr);
    Logger::getLogger("cmk.fetcher")->setHandler(nullptr);

    Logger *logger_cmk_core = Logger::getLogger("cmk.core");
    logger_cmk_core->setLevel(LogLevel::notice);
    ContextLogger logger_core(logger_cmk_core, [](std::ostream &os) {
        os << "[core " << getpid() << "] ";
    });

    unsetenv("LANG");

    // This must live at least until the last logging has been done.
    std::unique_ptr<std::ofstream> log_file;
    try {
        const CmcOptions opts = getCmcOptions(argc, argv);

        // This must live until the end of the process to keep our lock.
        PIDFile pid_file(get_pid_file_path());

        auto now = std::chrono::system_clock::now();
        auto core_config = Core::loadConfigFromFile(opts.path_to_config);
        auto general = PbConfigLoader{*core_config}.makeGeneral(
            core_config->globals().general());
        auto state_file =
            Core::loadStateFromFile(Core::state_path(general.core_directory_));
        auto state_file_created = mk::state_file_created(
            general.state_file_created_file_,
            get_default_creation_time(state_file.get(), now));
        auto is_licensed = mk::is_licensed(general.licensed_state_file_);
        auto num_services = std::accumulate(
            core_config->hosts().cbegin(), core_config->hosts().cend(), 0,
            [](size_t s, const auto &host) {
                return s + host.services().size();
            });
        auto trial_manager = mk::validate_license(
            state_file_created, is_licensed, now, num_services);

        if (opts.background) {
            log_file = std::make_unique<std::ofstream>();
            daemonize(pid_file);

            // NOTE: we enable logging early into the default folder
            // This is not the best strategy, still reliable enough to have some
            // important info logged
            auto filename = getAnticipatedLogFile();
            log_file->open(filename, std::ofstream::app);
            if (log_file->good()) {
                auto *logger = Logger::getLogger("cmk");
                logger->setUseParentHandlers(false);
                logger->setHandler(std::make_unique<StreamHandler>(*log_file));
                Notice(logger) << "[core " << getpid() << "] "
                               << "daemon logging started";
            }
        } else {
            pid_file.writeLocked();
        }

        Core core{
            argv,           opts.path_to_config, std::move(core_config),
            log_file.get(), &logger_core,        std::move(trial_manager)};
        fs::current_path(general.core_directory_);
        core.createState(std::move(state_file));
        core.run();
    } catch (const std::exception &e) {
        Emergency(&logger_core) << e.what();
        exit(EXIT_FAILURE);
    }
    Notice(&logger_core) << "Check_MK Micro Core exiting. Good bye.";
}
