// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <algorithm>
#include <chrono>
#include <cstddef>
#include <functional>
#include <memory>
#include <ratio>
#include <string>
#include <unordered_set>
#include <vector>

#include "PbUtils.h"
#include "cmc/Comment.h"
#include "cmc/Comments.h"
#include "cmc/Core.h"
#include "cmc/Host.h"
#include "cmc/MonitoringLog.h"
#include "cmc/World.h"
#include "config/v1/types.pb.h"
#include "gtest/gtest.h"
#include "livestatus/Logger.h"
#include "livestatus/TrialManager.h"

using namespace std::chrono_literals;
using namespace std::string_literals;

namespace {
namespace pb = checkmk::cmc::config::v1;

std::unique_ptr<Core> make_core(std::vector<std::string> &log,
                                const std::vector<std::string> &hosts) {
    auto pbcore = makeConfig();
    // Avoid warning: "cannot make canonical path: ..."
    test::string_ref(pbcore, "/dev/null");
    for (auto &&name : hosts) {
        auto pbhost = addHost(pbcore, name);
    }
    auto *logger = Logger::getLogger("test");
    const auto now = std::chrono::system_clock::time_point{};
    auto core = std::make_unique<Core>(
        nullptr, "/dev/null"s, std::make_unique<pb::Core>(pbcore), nullptr,
        logger, TrialManager{now, false});
    for (auto &&name : hosts) {
        // *state == nullptr at this point
        core->world()->getHost(name)->createNewState(now, logger);
    }
    core->monitoring_log_.sink(
        [&log](const std::string &entry) { log.push_back(entry); });
    return core;
}
}  // namespace

TEST(TestComments, InsertUserComment) {
    const auto host = "foo_host"s;
    const auto id = 4711;
    auto log = std::vector<std::string>{};
    auto core = make_core(log, {host});
    auto &comments = core->_comments;

    comments.insert(std::make_unique<Comment>(
        Comment{._object = core->world()->getHost(host),
                ._id = id,
                ._entry_time = std::chrono::system_clock::now(),
                ._author = "user comment author"s,
                ._comment = "user comment"s,
                ._entry_type = Comment::Type::user,
                ._expire_time = {},
                ._persistent = true,
                ._sticky = false,
                ._source = Comment::Source::external}));

    EXPECT_EQ(size_t{1}, comments.size());
    EXPECT_TRUE(log.empty());
}

TEST(TestComments, InsertAcknowledgementComment) {
    const auto host = "foo_host"s;
    const auto now = std::chrono::system_clock::now();
    auto log = std::vector<std::string>{};
    auto core = make_core(log, {host});
    auto &comments = core->_comments;

    comments.insert(std::make_unique<Comment>(
        Comment{._object = core->world()->getHost(host),
                ._id = 42,
                ._entry_time = now,
                ._author = "ack author"s,
                ._comment = "ack comment"s,
                ._entry_type = Comment::Type::acknowledgement,
                ._expire_time = now + 1h,
                ._persistent = true,
                ._sticky = true,
                ._source = Comment::Source::internal}));

    EXPECT_EQ(size_t{1}, comments.size());
    ASSERT_EQ(size_t{1}, log.size());
    EXPECT_EQ(
        "HOST ACKNOWLEDGE ALERT: foo_host;STARTED;ack author;ack comment"s,
        log[0]);
}

TEST(TestComments, RemoveUserComment) {
    const auto host = "foo_host"s;
    const size_t id = 4711;
    auto log = std::vector<std::string>{};
    auto core = make_core(log, {host});
    auto &comments = core->_comments;

    comments.insert(std::make_unique<Comment>(
        Comment{._object = core->world()->getHost(host),
                ._id = id,
                ._entry_time = std::chrono::system_clock::now(),
                ._author = "user comment author"s,
                ._comment = "user comment"s,
                ._entry_type = Comment::Type::user,
                ._expire_time = {},
                ._persistent = true,
                ._sticky = false,
                ._source = Comment::Source::external}));
    comments.remove({id});

    EXPECT_TRUE(comments.empty());
    EXPECT_TRUE(log.empty());
}

TEST(TestComments, RemoveNonPersistentAcknowledgement) {
    const auto host = "foo_host"s;
    const auto persistent = true;
    auto log = std::vector<std::string>{};
    auto core = make_core(log, {host});
    const auto now = std::chrono::system_clock::now();
    auto &comments = core->_comments;

    comments.insert(std::make_unique<Comment>(
        Comment{._object = core->world()->getHost(host),
                ._id = 42,
                ._entry_time = now,
                ._author = "ack author"s,
                ._comment = "ack comment"s,
                ._entry_type = Comment::Type::acknowledgement,
                ._expire_time = now + 1h,
                ._persistent = !persistent,
                ._sticky = true,
                ._source = Comment::Source::internal}));
    comments.removeAcknowledgements({core->world()->getHost(host)},
                                    "CANCELLED");

    EXPECT_TRUE(comments.empty());
    ASSERT_EQ(size_t{2}, log.size());
    EXPECT_EQ(
        "HOST ACKNOWLEDGE ALERT: foo_host;STARTED;ack author;ack comment"s,
        log[0]);
    EXPECT_EQ(
        "HOST ACKNOWLEDGE ALERT: foo_host;CANCELLED;ack author;ack comment"s,
        log[1]);
}

TEST(TestComments, RemovePersistentAcknowledgement) {
    const auto host = "foo_host"s;
    const auto persistent = true;
    auto log = std::vector<std::string>{};
    auto core = make_core(log, {host});
    auto &comments = core->_comments;
    const auto now = std::chrono::system_clock::now();

    comments.insert(std::make_unique<Comment>(
        Comment{._object = core->world()->getHost(host),
                ._id = 42,
                ._entry_time = now,
                ._author = "ack author"s,
                ._comment = "ack comment"s,
                ._entry_type = Comment::Type::acknowledgement,
                ._expire_time = now + 1h,
                ._persistent = persistent,
                ._sticky = true,
                ._source = Comment::Source::internal}));
    comments.removeAcknowledgements({core->world()->getHost(host)},
                                    "CANCELLED");

    EXPECT_EQ(size_t{1}, comments.size());
    ASSERT_EQ(size_t{2}, log.size());
    EXPECT_EQ(
        "HOST ACKNOWLEDGE ALERT: foo_host;STARTED;ack author;ack comment"s,
        log[0]);
    EXPECT_EQ(
        "HOST ACKNOWLEDGE ALERT: foo_host;CANCELLED;ack author;ack comment"s,
        log[1]);
}
