// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <chrono>
#include <cstdlib>
#include <memory>
#include <ratio>

#include "cmc/CycleTime.h"
#include "gtest/gtest.h"
#include "livestatus/Logger.h"

using namespace std::chrono_literals;
using time_point = std::chrono::system_clock::time_point;

class TestCycleTimeFixture : public ::testing::Test {
protected:
    void SetUp() override {
        tz_ = std::getenv("TZ");
        ::setenv("TZ", "Europe/Berlin", 1);
    }
    void TearDown() override {
        tz_ == nullptr ? ::unsetenv("TZ") : ::setenv("TZ", tz_, 1);
    }
    static Logger *logger() { return Logger::getLogger("test"); }

    static constexpr auto dst_2020_start =
        time_point{1585443600s};  // 2020-03-29 1:00
    static constexpr auto dst_2020_end =
        time_point{1603587600s};  // 2020-10-25 1:00
    static constexpr auto dst_delta = 3600s;

private:
    char *tz_;
};

TEST_F(TestCycleTimeFixture, TestDSTShiftBetween) {
    EXPECT_EQ(0s,
              detail::dst_shift_between(dst_2020_start - std::chrono::days{2},
                                        dst_2020_start - std::chrono::days{1}));
    EXPECT_EQ(
        0s, detail::dst_shift_between(dst_2020_start + std::chrono::days{1},
                                      dst_2020_start + std::chrono::weeks{1}));
    EXPECT_EQ(dst_delta, detail::dst_shift_between(dst_2020_start - 1s,
                                                   dst_2020_start + 1s));
    EXPECT_EQ(-dst_delta,
              detail::dst_shift_between(dst_2020_end - 1s, dst_2020_end + 1s));
}

TEST_F(TestCycleTimeFixture, TestCycleTimeNextNone) {
    const CycleTime ct{CycleTime::Kind::none, logger()};

    EXPECT_EQ(1h, ct.next(time_point(0s)));
    EXPECT_EQ(1h, ct.next(dst_2020_start - 1min));
    EXPECT_EQ(1h, ct.next(dst_2020_start + 1min));
    EXPECT_EQ(1h + dst_delta, ct.next(dst_2020_end - 1min));
    EXPECT_EQ(1h, ct.next(dst_2020_end + 1min));
}

TEST_F(TestCycleTimeFixture, TestCycleTimeNextHourly) {
    const CycleTime ct{CycleTime::Kind::hourly, logger()};

    EXPECT_EQ(1h, ct.next(time_point(0s)));
    EXPECT_EQ(1h, ct.next(dst_2020_start - 1min));
    EXPECT_EQ(1h, ct.next(dst_2020_start + 1min));
    EXPECT_EQ(1h + dst_delta, ct.next(dst_2020_end - 1min));
    EXPECT_EQ(1h, ct.next(dst_2020_end + 1min));
}

TEST_F(TestCycleTimeFixture, TestCycleTimeNextDaily) {
    const CycleTime ct{CycleTime::Kind::daily, logger()};

    EXPECT_EQ(std::chrono::days{1}, ct.next(time_point(0s)));
    EXPECT_EQ(std::chrono::days{1} - dst_delta, ct.next(dst_2020_start - 1min));
    EXPECT_EQ(std::chrono::days{1}, ct.next(dst_2020_start + 1min));
    EXPECT_EQ(std::chrono::days{1} + dst_delta, ct.next(dst_2020_end - 1min));
    EXPECT_EQ(std::chrono::days{1}, ct.next(dst_2020_end + 1min));
}

TEST_F(TestCycleTimeFixture, TestCycleTimeNextWeekly) {
    const CycleTime ct{CycleTime::Kind::weekly, logger()};

    EXPECT_EQ(std::chrono::weeks{1}, ct.next(time_point(0s)));
    EXPECT_EQ(std::chrono::weeks{1} - dst_delta,
              ct.next(dst_2020_start - 1min));
    EXPECT_EQ(std::chrono::weeks{1}, ct.next(dst_2020_start + 1min));
    EXPECT_EQ(std::chrono::weeks{1} + dst_delta, ct.next(dst_2020_end - 1min));
    EXPECT_EQ(std::chrono::weeks{1}, ct.next(dst_2020_end + 1min));
}

TEST_F(TestCycleTimeFixture, TestCycleTimeNextBiweekly) {
    const CycleTime ct{CycleTime::Kind::biweekly, logger()};

    EXPECT_EQ(std::chrono::weeks{2}, ct.next(time_point(0s)));
    EXPECT_EQ(std::chrono::weeks{2} - dst_delta,
              ct.next(dst_2020_start - 1min));
    EXPECT_EQ(std::chrono::weeks{2}, ct.next(dst_2020_start + 1min));
    EXPECT_EQ(std::chrono::weeks{2} + dst_delta, ct.next(dst_2020_end - 1min));
    EXPECT_EQ(std::chrono::weeks{2}, ct.next(dst_2020_end + 1min));
}

TEST_F(TestCycleTimeFixture, TestCycleTimeNextEvery4Weeks) {
    const CycleTime ct{CycleTime::Kind::every_4weeks, logger()};

    EXPECT_EQ(std::chrono::weeks{4}, ct.next(time_point(0s)));
    EXPECT_EQ(std::chrono::weeks{4} - dst_delta,
              ct.next(dst_2020_start - 1min));
    EXPECT_EQ(std::chrono::weeks{4}, ct.next(dst_2020_start + 1min));
    EXPECT_EQ(std::chrono::weeks{4} + dst_delta, ct.next(dst_2020_end - 1min));
    EXPECT_EQ(std::chrono::weeks{4}, ct.next(dst_2020_end + 1min));
}

TEST_F(TestCycleTimeFixture, TestCycleTimeNextEvery5Min) {
    const CycleTime ct{CycleTime::Kind::every_5min, logger()};

    EXPECT_EQ(5min, ct.next(time_point(0s)));
    EXPECT_EQ(5min, ct.next(dst_2020_start - 1min));
    EXPECT_EQ(5min, ct.next(dst_2020_start + 1min));
    EXPECT_EQ(5min + dst_delta, ct.next(dst_2020_end - 1min));
    EXPECT_EQ(5min, ct.next(dst_2020_end + 1min));
}

TEST_F(TestCycleTimeFixture, TestCycleTimeNextNthWeekday) {
    const CycleTime ct{CycleTime::Kind::nth_weekday, logger()};

    EXPECT_EQ(std::chrono::weeks{5}, ct.next(time_point(0s)));
    EXPECT_EQ(std::chrono::weeks{5} - dst_delta,
              ct.next(dst_2020_start - 1min));
    EXPECT_EQ(std::chrono::weeks{5}, ct.next(dst_2020_start + 1min));
    EXPECT_EQ(std::chrono::weeks{4} + dst_delta, ct.next(dst_2020_end - 1min));
    EXPECT_EQ(std::chrono::weeks{4}, ct.next(dst_2020_end + 1min));
}

TEST_F(TestCycleTimeFixture, TestCycleTimeNextNthWeekdayFromEnd) {
    const CycleTime ct{CycleTime::Kind::nth_weekday_from_end, logger()};

    EXPECT_EQ(std::chrono::weeks{5}, ct.next(time_point(0s)));
    EXPECT_EQ(std::chrono::weeks{4} - dst_delta,
              ct.next(dst_2020_start - 1min));
    EXPECT_EQ(std::chrono::weeks{4}, ct.next(dst_2020_start + 1min));
    EXPECT_EQ(std::chrono::weeks{5} + dst_delta, ct.next(dst_2020_end - 1min));
    EXPECT_EQ(std::chrono::weeks{5}, ct.next(dst_2020_end + 1min));
}

TEST_F(TestCycleTimeFixture, TestCycleTimeNextDayOfMonth) {
    const CycleTime ct{CycleTime::Kind::nth_weekday_from_end, logger()};

    EXPECT_EQ(std::chrono::weeks{5}, ct.next(time_point(0s)));
    EXPECT_EQ(std::chrono::weeks{4} - dst_delta,
              ct.next(dst_2020_start - 1min));
    EXPECT_EQ(std::chrono::weeks{4}, ct.next(dst_2020_start + 1min));
    EXPECT_EQ(std::chrono::weeks{5} + dst_delta, ct.next(dst_2020_end - 1min));
    EXPECT_EQ(std::chrono::weeks{5}, ct.next(dst_2020_end + 1min));
}
