// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <unistd.h>

#include <chrono>
#include <future>
#include <initializer_list>
#include <iomanip>
#include <memory>
#include <optional>
#include <ratio>
#include <sstream>
#include <stdexcept>
#include <string>
#include <string_view>
#include <thread>
#include <variant>
#include <vector>

#include "cmc/FetcherHelperChannel.h"
#include "cmc/MessagePayload.h"
#include "gtest/gtest.h"
#include "livestatus/ChronoUtils.h"
#include "livestatus/Logger.h"
#include "livestatus/POSIXUtils.h"
#include "livestatus/overload.h"

using namespace std::chrono_literals;
using namespace std::string_literals;
using namespace std::string_view_literals;
namespace fp = fetcher::protocol;

TEST(FetcherMessageHeader, ParseResult) {
    std::istringstream is{"fetch:RESULT :        :123             :"};
    const FetcherMessageHeader h{is};
    EXPECT_EQ(FetcherMessageHeader::Kind::result, h.kind());
    EXPECT_FALSE(h.log_level());
    EXPECT_EQ(size_t{123}, h.payload_length());
}

TEST(FetcherMessageHeader, ParseLog) {
    std::istringstream is{"fetch:LOG    :info    :1               :"};
    const FetcherMessageHeader h{is};
    EXPECT_EQ(FetcherMessageHeader::Kind::log, h.kind());
    EXPECT_EQ(LogLevel::informational, h.log_level());
    EXPECT_EQ(size_t{1}, h.payload_length());
}

TEST(FetcherMessageHeader, ParseEndOfReply) {
    std::istringstream is{"fetch:ENDREPL:        :0               :"};
    const FetcherMessageHeader h{is};
    EXPECT_EQ(FetcherMessageHeader::Kind::end_of_reply, h.kind());
    EXPECT_FALSE(h.log_level());
    EXPECT_EQ(size_t{0}, h.payload_length());
}

TEST(FetcherMessageHeader, ParseFailures) {
    for (const auto &s : {// parseHaderSize
                          "fetch:"s,
                          // parseProtocolName
                          "hurz!:ENDREPL:        :0               :"s,
                          // parseSeparator
                          "fetch|RESULT :        :123             :"s,
                          // parseKind
                          "fetch:RESALT :        :123             :"s,
                          // parseSeparator
                          "fetch:RESULT |        :123             :"s,
                          // parseLogLevel
                          "fetch:RESULT :info    :0               :"s,
                          "fetch:ENDREPL:info    :0               :"s,
                          // parseSeparator
                          "fetch:RESULT :        |123             :"s,
                          // convertToLength
                          "fetch:RESULT :        :abc             :"s,
                          // parsePayloadLength
                          "fetch:ENDREPL:        :123             :"s,
                          // parseSeparator
                          "fetch:RESULT :        :123             |"s}) {
        std::istringstream is{s};
        // NOLINTNEXTLINE(cppcoreguidelines-avoid-goto,hicpp-avoid-goto)
        EXPECT_THROW(FetcherMessageHeader{is}, std::runtime_error);
    }
}

class FetcherHelperChannelFixture : public ::testing::Test {
public:
    void SetUp() override {
        // Build sockets, example for production AlertHelper:
        // CONTEXT: non-blocking is normal for CheckHelper
        auto sockets = SocketPair::make(SocketPair::Mode::local_non_blocking,
                                        SocketPair::Direction::bidirectional,
                                        Logger::getLogger("test"));
        if (!sockets) {
            FAIL() << "no socket pair";
        }
        local_io_ = sockets->local();
        remote_io_ = sockets->remote();

        auto *logger = Logger::getLogger("test");
        logger->setLevel(LogLevel::emergency);

        ch_ = std::make_unique<FetcherHelperChannel>(logger);
    }

    void TearDown() override {
        // NOTE: we kill sockets before killing channels which use those
        // sockets. Suitable only in tests.
        ::close(local_io_);
        ::close(remote_io_);
    }

    //
    //  All class data are public to simplify test coding: we do not want to
    //  write an accessor just to use member data. A trailing underscore is good
    //  enough to mark data as "PERSISTENT" and "MAY BE MUTABLE".
    //

    // CONTEXT: to test error subchannel, no-asio mode
    // This is example of production code in Alert/Check helpers.
    // I am not sure, that we need pipe, probably socket is better idea
    int local_io_{-1};
    int remote_io_{-1};

    std::vector<char> data_;  // storage to receive data from remote process

    static std::string createResultMessage(std::string_view payload) {
        return makeHeader("RESULT", "", payload.length()) +
               std::string(payload);
    }

    static std::string createLogMessage(const std::string &log_level,
                                        std::string_view message) {
        return makeHeader("LOG", log_level, message.length()) +
               std::string(message);
    }

    static std::string createEndOfReplyMessage() {
        return makeHeader("ENDREPL", "", 0);
    }

    std::unique_ptr<FetcherHelperChannel> ch_;  // usual channel
    static constexpr std::string_view payload_to_send{"And This Is Big"sv};
    Logger *logger_{Logger::getLogger("test")};

private:
    static std::string makeHeader(const std::string &kind,
                                  const std::string &log_level,
                                  ssize_t length) {
        std::ostringstream os;
        os << std::left << "fetch:"             //
           << std::setw(7) << kind << ":"       //
           << std::setw(8) << log_level << ":"  //
           << std::setw(16) << length << ":";
        return os.str();
    }
};

TEST_F(FetcherHelperChannelFixture, RecvResult) {
    auto reply_bytes = createResultMessage(payload_to_send);
    auto expected_length =
        FetcherMessageHeader::size + payload_to_send.length();
    ASSERT_EQ(expected_length, reply_bytes.size());
    // TODO(sp) There is no guarantee that we can send everything at once!
    auto written = ::write(remote_io_, reply_bytes.data(), reply_bytes.size());
    EXPECT_EQ(static_cast<ssize_t>(expected_length), written);

    while (std::visit(
        mk::overload{[](const fp::Incomplete & /*msg*/) { return true; },
                     [](const fp::Failure & /*msg*/) {
                         ADD_FAILURE();
                         return false;
                     },
                     [](const fp::Result &msg) {
                         EXPECT_EQ(payload_to_send,
                                   std::string_view{msg.payload});
                         return false;
                     },
                     [](const fp::Log & /*msg*/) {
                         ADD_FAILURE();
                         return false;
                     },
                     [](const fp::EndOfReply & /*msg*/) {
                         ADD_FAILURE();
                         return false;
                     }},
        ch_->receive(local_io_))) {
    }
}

TEST_F(FetcherHelperChannelFixture, RecvLog) {
    auto reply_bytes = createLogMessage("error", payload_to_send);
    auto expected_length =
        FetcherMessageHeader::size + payload_to_send.length();
    ASSERT_EQ(expected_length, reply_bytes.size());
    // TODO(sp) There is no guarantee that we can send everything at once!
    auto written = ::write(remote_io_, reply_bytes.data(), reply_bytes.size());
    EXPECT_EQ(static_cast<ssize_t>(expected_length), written);

    while (std::visit(
        mk::overload{[](const fp::Incomplete & /*msg*/) { return true; },
                     [](const fp::Failure & /*msg*/) {
                         ADD_FAILURE();
                         return false;
                     },
                     [](const fp::Result & /*msg*/) {
                         ADD_FAILURE();
                         return false;
                     },
                     [](const fp::Log &msg) {
                         EXPECT_EQ(LogLevel::error, msg.log_level);
                         EXPECT_EQ(payload_to_send,
                                   std::string_view{msg.message});
                         return false;
                     },
                     [](const fp::EndOfReply & /*msg*/) {
                         ADD_FAILURE();
                         return false;
                     }},
        ch_->receive(local_io_))) {
    }
}

TEST_F(FetcherHelperChannelFixture, RecvEndOfReply) {
    auto reply_bytes = createEndOfReplyMessage();
    auto expected_length = FetcherMessageHeader::size;
    ASSERT_EQ(expected_length, reply_bytes.size());
    // TODO(sp) There is no guarantee that we can send everything at once!
    auto written = ::write(remote_io_, reply_bytes.data(), reply_bytes.size());
    EXPECT_EQ(static_cast<ssize_t>(expected_length), written);

    while (std::visit(
        mk::overload{[](const fp::Incomplete & /*msg*/) { return true; },
                     [](const fp::Failure & /*msg*/) {
                         ADD_FAILURE();
                         return false;
                     },
                     [](const fp::Result & /*msg*/) {
                         ADD_FAILURE();
                         return false;
                     },
                     [](const fp::Log & /*msg*/) {
                         ADD_FAILURE();
                         return false;
                     },
                     [](const fp::EndOfReply & /*msg*/) { return false; }},
        ch_->receive(local_io_))) {
    }
}

namespace {
// TODO (sk): move in the test tools.
inline ::testing::AssertionResult ConfirmIo(int fd,
                                            const std::string &expected_string,
                                            std::chrono::milliseconds ms) {
    using namespace std::chrono_literals;

    constexpr auto interval = 5ms;
    auto attempts = (ms + interval - 1ms) / interval;
    const auto count = expected_string.size();
    std::vector<char> buf(count);
    size_t ready = 0;
    for (int i = 0; i < attempts; ++i) {
        auto sz = ::read(fd, buf.data() + ready, count - ready);
        ready += sz;
        if (ready == count) {
            const std::string result{buf.begin(), buf.end()};
            return expected_string == result
                       ? ::testing::AssertionSuccess()
                       : ::testing::AssertionFailure()
                             << "expected: " << expected_string
                             << " obtained: " << result;
        }
        std::this_thread::sleep_for(interval);
    }

    return ::testing::AssertionFailure()
           << "failed to read '" << expected_string << "'in "
           << mk::ticks<std::chrono::milliseconds>(ms) << " ms";
}
}  // namespace

TEST_F(FetcherHelperChannelFixture, TestRun) {
    const fp::Request request{4711, "b", "checking", 30s};
    const auto expected_string = "4711;b;checking;30\n"s;
    ch_->send(local_io_, request);

    const std::vector<char> buf(expected_string.size());
    ASSERT_TRUE(ConfirmIo(remote_io_, expected_string, 2000ms));
}

TEST_F(FetcherHelperChannelFixture, SimulateSlowFetcher) {
    auto result = createResultMessage(payload_to_send);
    auto expected_length =
        FetcherMessageHeader::size + payload_to_send.length();
    ASSERT_EQ(expected_length, result.size());

    // simulation of the slow fetcher, one byte per 1ms
    auto a = std::async(
        std::launch::async,                   // first parameter
        [this](const std::string &success) {  // lambda
            for (const auto &c : success) {
                if (::write(remote_io_, &c, 1) != 1) {
                    break;
                }
                std::this_thread::sleep_for(1ms);
            }
        },
        result  // lambda parameter
    );

    int sec_counter = 0;
    while (std::visit(
        mk::overload{[](const fp::Incomplete & /*msg*/) { return true; },
                     [](const fp::Failure & /*msg*/) {
                         ADD_FAILURE();
                         return false;
                     },
                     [](const fp::Result &msg) {
                         EXPECT_EQ(payload_to_send,
                                   std::string_view{msg.payload});
                         return false;
                     },
                     [](const fp::Log & /*msg*/) {
                         ADD_FAILURE();
                         return false;
                     },
                     [](const fp::EndOfReply & /*msg*/) {
                         ADD_FAILURE();
                         return false;
                     }},
        ch_->receive(local_io_))) {
        std::this_thread::sleep_for(5ms);
        if (++sec_counter > 40) {
            break;  // we don't want to hang in the test forever
        }
    }
    EXPECT_TRUE(sec_counter > 0);  // check, that we have really slow fetcher
}
