// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <chrono>
#include <cstdint>
#include <sstream>
#include <string>
#include <vector>

#include "cmc/InfluxDBLineProtocol.h"
#include "gtest/gtest.h"

using namespace std::string_literals;
using namespace std::chrono_literals;

class SerializeFixture : public ::testing::Test {
public:
    using LineProtocol = influxdb::LineProtocol;
    using Measurement = influxdb::Measurement;
    using Field = influxdb::Field;
    using Tag = influxdb::Tag;

    template <typename T>
    std::string serialize(const T &s) {
        stream_ << s;
        return stream_.str();
    }

private:
    std::ostringstream stream_{};
};

TEST_F(SerializeFixture, MeasurementEscape) {
    EXPECT_EQ(R"(hello\,\ beautiful\ world)"s,
              serialize(Measurement{"hello, beautiful world"}));
}

TEST_F(SerializeFixture, TagSerialize) {
    EXPECT_EQ(R"(tag=hello\ \=\ world)"s,
              serialize(Tag{"tag"s, "hello = world"s}));
}

TEST_F(SerializeFixture, FieldSerializeFloat) {
    EXPECT_EQ("field=1.5"s, serialize(Field{"field"s, 1.5}));
}

TEST_F(SerializeFixture, FieldSerializeInt) {
    EXPECT_EQ("field=1i"s, serialize(Field{"field"s, std::int64_t{1}}));
}

TEST_F(SerializeFixture, FieldSerializeUInt) {
    EXPECT_EQ("field=1u"s, serialize(Field{"field"s, std::uint64_t{1}}));
}

TEST_F(SerializeFixture, FieldSerializeString) {
    EXPECT_EQ(R"(field="hello \"world\"")"s,
              serialize(Field{"field"s, "hello \"world\""s}));
}

TEST_F(SerializeFixture, FieldSerializeBool) {
    EXPECT_EQ("field=true"s, serialize(Field{"field"s, true}));
}

TEST_F(SerializeFixture, LineProtocolSerializeNoTagOneField) {
    EXPECT_EQ(R"(meas fieldKey="a key" 0)"s + "\n",
              serialize(LineProtocol{Measurement{"meas"s},
                                     {},
                                     {Field{"fieldKey"s, "a key"s}},
                                     std::chrono::system_clock::time_point{}}));
}

TEST_F(SerializeFixture, LineProtocolSerializeNoTagTwoFields) {
    EXPECT_EQ(R"(meas key1="a field",key2="another field" 0)"s + "\n",
              serialize(LineProtocol{Measurement{"meas"s},
                                     {},
                                     {Field{"key1"s, "a field"s},
                                      Field{"key2"s, "another field"s}},
                                     std::chrono::system_clock::time_point{}}));
}

TEST_F(SerializeFixture, LineProtocolSerializeOneTagOneField) {
    EXPECT_EQ(R"(meas,tagKey=a\ tag fieldKey="a field" 0)"s + "\n",
              serialize(LineProtocol{Measurement{"meas"s},
                                     {Tag{"tagKey"s, "a tag"s}},
                                     {Field{"fieldKey"s, "a field"s}},
                                     std::chrono::system_clock::time_point{}}));
}

TEST_F(SerializeFixture, LineProtocolSerializeTwoTagsTwoFields) {
    EXPECT_EQ(
        R"(meas,tag1=a\ tag,tag2=another\=tag key1="a field",key2="another field" 0)"s +
            "\n",
        serialize(LineProtocol{
            Measurement{"meas"s},
            {Tag{"tag1"s, "a tag"s}, Tag{"tag2"s, "another=tag"s}},
            {Field{"key1"s, "a field"s}, Field{"key2"s, "another field"s}},
            std::chrono::system_clock::time_point{}}));
}

TEST_F(SerializeFixture, LineProtocolSerializeMoreTagsAndFields) {
    EXPECT_EQ(
        R"(meas,tag1=t1,tag2=t2,tag3=t3 field1="f1",field2="f2",field3="f3" 0)"s +
            "\n",
        serialize(LineProtocol{
            Measurement{"meas"s},
            {Tag{"tag1"s, "t1"s}, Tag{"tag2"s, "t2"s}, Tag{"tag3"s, "t3"s}},
            {Field{"field1"s, "f1"s}, Field{"field2"s, "f2"s},
             Field{"field3"s, "f3"s}},
            std::chrono::system_clock::time_point{}}));
}
