// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <limits>
#include <optional>
#include <string>
#include <variant>

#include "cmc/InfluxDBMetric.h"
#include "gtest/gtest.h"
#include "livestatus/Metric.h"

namespace {
template <typename T, typename V>
testing::AssertionResult equal_v(const T &value, const V &variant) {
    if (!std::holds_alternative<T>(variant)) {
        return testing::AssertionFailure() << "type mismatch [" << value << "]";
    }
    const auto other = std::get<T>(variant);
    return value == other
               ? testing::AssertionSuccess()
               : testing::AssertionFailure() << value << " != " << other;
}
}  // namespace

TEST(Value, MakeIntToDouble) {
    // The check API never sends integers but may as well silently
    // turn 1.000 into "1".
    const auto value = influxdb::Metric::Value::make("42");
    ASSERT_FALSE(value.empty());
    EXPECT_TRUE(equal_v(42.0, *value));
}

TEST(Value, MakeDouble) {
    const auto value = influxdb::Metric::Value::make("12.34");
    ASSERT_FALSE(value.empty());
    EXPECT_TRUE(equal_v(12.34, *value));
}

TEST(Value, MakeUnknown) {
    EXPECT_TRUE(influxdb::Metric::Value::make("").empty());
    EXPECT_TRUE(influxdb::Metric::Value::make("u").empty());
    EXPECT_TRUE(influxdb::Metric::Value::make("U").empty());
}

TEST(Value, MakeInvalidArgumentFails) {
    const auto value =
        influxdb::Metric::Value::make("we are not going to parse random text");
    EXPECT_TRUE(value.empty());
}

TEST(RealRange, ImplicitLowerBound) {
    const auto *const str = "10.0";
    const auto range = influxdb::Metric::Range::make(str);

    const auto bounds = range.bounds();
    if (!bounds) {
        FAIL() << "no bounds";
    }
    EXPECT_TRUE(equal_v(0.0, bounds->lower_bound));
    EXPECT_TRUE(equal_v(10.0, bounds->higher_bound));
    EXPECT_FALSE(range.inverted());

    EXPECT_FALSE(range.contains(-1.0));
    EXPECT_TRUE(range.contains(0.0));
    EXPECT_TRUE(range.contains(10.0));
    EXPECT_FALSE(range.contains(11.0));
}

TEST(RealRange, InfiniteHigherBound) {
    const auto *const str = "10.0:";
    const auto range = influxdb::Metric::Range::make(str);

    const auto bounds = range.bounds();
    if (!bounds) {
        FAIL() << "no bounds";
    }
    EXPECT_TRUE(equal_v(10.0, bounds->lower_bound));
    EXPECT_TRUE(
        equal_v(std::numeric_limits<double>::max(), bounds->higher_bound));
    EXPECT_FALSE(range.inverted());

    EXPECT_FALSE(range.contains(9.0));
    EXPECT_TRUE(range.contains(10.0));
}

TEST(RealRange, InfiniteLowerBound) {
    const auto *const str = "~:10.0";
    const auto range = influxdb::Metric::Range::make(str);

    const auto bounds = range.bounds();
    if (!bounds) {
        FAIL() << "no bounds";
    }
    EXPECT_TRUE(
        equal_v(std::numeric_limits<double>::lowest(), bounds->lower_bound));
    EXPECT_TRUE(equal_v(10.0, bounds->higher_bound));
    EXPECT_FALSE(range.inverted());

    EXPECT_TRUE(range.contains(-100.0));
    EXPECT_TRUE(range.contains(-1.0));
    EXPECT_TRUE(range.contains(0.0));
    EXPECT_TRUE(range.contains(10.0));
    EXPECT_FALSE(range.contains(11.0));
}

TEST(RealRange, LowerHigherBound) {
    const auto *const str = "10.0:20.0";
    const auto range = influxdb::Metric::Range::make(str);

    const auto bounds = range.bounds();
    if (!bounds) {
        FAIL() << "no bounds";
    }
    EXPECT_TRUE(equal_v(10.0, bounds->lower_bound));
    EXPECT_TRUE(equal_v(20.0, bounds->higher_bound));
    EXPECT_FALSE(range.inverted());

    EXPECT_FALSE(range.contains(9.0));
    EXPECT_TRUE(range.contains(10.0));
    EXPECT_TRUE(range.contains(20.0));
    EXPECT_FALSE(range.contains(21.0));
}

TEST(Range, Empty) {
    const auto *const str = "";
    const auto range = influxdb::Metric::Range::make(str);

    EXPECT_FALSE(range.bounds());
    EXPECT_FALSE(range.inverted());

    // No alarm on an empty range: the range
    // may thus contain anything.
    EXPECT_TRUE(range.contains(std::numeric_limits<double>::lowest()));
    EXPECT_TRUE(range.contains(0.0));
    EXPECT_TRUE(range.contains(std::numeric_limits<double>::max()));
}

TEST(Metric, MakeMetric) {
    auto raw = Metric{"label", "42.0", "B", "25:75.0", "10:100", "0", "69"};
    auto metric = influxdb::Metric::make(raw);

    EXPECT_EQ(raw.name().string(), metric.name().string());
    EXPECT_TRUE(equal_v(42.0, *metric.value()));
    EXPECT_EQ(raw.uom(), metric.uom());
    const auto warn = metric.warn().bounds();
    if (!warn) {
        FAIL() << "no warn bounds";
    }
    EXPECT_TRUE(equal_v(25.0, warn->lower_bound));
    EXPECT_TRUE(equal_v(75.0, warn->higher_bound));
    const auto crit = metric.crit().bounds();
    if (!crit) {
        FAIL() << "no crit bounds";
    }
    EXPECT_TRUE(equal_v(10.0, crit->lower_bound));
    EXPECT_TRUE(equal_v(100.0, crit->higher_bound));
    EXPECT_TRUE(equal_v(0.0, *metric.min()));
    EXPECT_TRUE(equal_v(69.0, *metric.max()));
}
