// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <algorithm>
#include <chrono>
#include <cstddef>
#include <iterator>
#include <string>
#include <vector>

#include "asio/io_context.hpp"
#include "cmc/LivestatusExternalCommand.h"
#include "gtest/gtest.h"
#include "livestatus/Logger.h"
#include "livestatus/Poller.h"
#include "livestatus/global_counters.h"

using namespace std::string_literals;
using namespace std::chrono_literals;

class LivestatusExternalCommandFixture : public ::testing::Test {
public:
    void SetUp() override {
        ls.logger()->setLevel(LogLevel::emergency);
        ls.start();
    }
    void TearDown() override {
        ls.stop();
        counterReset(Counter::commands);
    }
    asio::io_context io_context;
    Poller poller;
    std::vector<std::string> sink;
    LivestatusExternalCommand ls{
        io_context, [this](const std::string &cmd) { sink.emplace_back(cmd); },
        Logger::getLogger("test")};
};

TEST_F(LivestatusExternalCommandFixture, EnqueueAndDequeueOneCommand) {
    ASSERT_TRUE(ls.active());
    ASSERT_TRUE(sink.empty());

    ASSERT_EQ(size_t{0}, ls.queueSize());
    ASSERT_EQ(size_t{0}, ls.commandCount());
    ASSERT_EQ(0, counterValue(Counter::commands));

    // The class does not care about the actual format of the command.
    const auto cmd = "cmd"s;

    ls.setUpPoller(poller);
    ls.spoolCommand(cmd);
    EXPECT_EQ(size_t{1}, ls.queueSize());
    EXPECT_EQ(size_t{1}, ls.commandCount());
    EXPECT_EQ(1, counterValue(Counter::commands));

    poller.poll(10ms);
    ls.communicate(poller);
    EXPECT_EQ(cmd, !sink.empty() ? sink[0] : "failure"s);

    EXPECT_EQ(size_t{0}, ls.queueSize());
    EXPECT_EQ(size_t{1}, ls.commandCount());
    EXPECT_EQ(1, counterValue(Counter::commands));
}

TEST_F(LivestatusExternalCommandFixture, EnqueueAndDequeueManyCommands) {
    ASSERT_TRUE(ls.active());
    ASSERT_TRUE(sink.empty());

    ASSERT_EQ(size_t{0}, ls.queueSize());
    ASSERT_EQ(size_t{0}, ls.commandCount());
    ASSERT_EQ(0, counterValue(Counter::commands));

    const std::vector<std::string> cmds{"cmd0", "cmd1", "cmd2", "cmd3", "cmd4"};

    ls.setUpPoller(poller);
    std::for_each(std::begin(cmds), std::end(cmds),
                  [this](const auto &cmd) { ls.spoolCommand(cmd); });
    EXPECT_EQ(cmds.size(), ls.queueSize());
    EXPECT_EQ(cmds.size(), ls.commandCount());
    EXPECT_EQ(cmds.size(), size_t(counterValue(Counter::commands)));

    poller.poll(10ms);
    ls.communicate(poller);
    EXPECT_EQ(cmds, sink);

    EXPECT_EQ(size_t{0}, ls.queueSize());
    EXPECT_EQ(cmds.size(), ls.commandCount());
    EXPECT_EQ(cmds.size(), size_t(counterValue(Counter::commands)));
}
