// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <fcntl.h>

#include <algorithm>
#include <chrono>
#include <cstddef>
#include <filesystem>
#include <iomanip>
#include <iterator>
#include <random>
#include <sstream>
#include <string>
#include <vector>

#include "asio/buffer.hpp"
#include "asio/io_context.hpp"
#include "asio/posix/stream_descriptor.hpp"
#include "asio/write.hpp"
#include "cmc/LivestatusLocalCommand.h"
#include "gtest/gtest.h"
#include "livestatus/Logger.h"
#include "livestatus/Poller.h"

using namespace std::string_literals;
using namespace std::chrono_literals;

namespace {
std::string padded(int number, int width) {
    std::ostringstream out;
    out << std::internal << std::setfill('0') << std::setw(width) << number;
    return out.str();
}

std::string frame(const std::string &cmd) {
    // The helper roughly validates the format of the timestamp.
    return "[1234567890] "s + cmd + '\n';
}

// https://stackoverflow.com/questions/440133/how-do-i-create-a-random-alpha-numeric-string-in-c
std::string random_string(const std::string::size_type length) {
    static const auto &chrs =
        "0123456789"
        "abcdefghijklmnopqrstuvwxyz"
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

    thread_local static std::mt19937 rg{std::random_device{}()};
    thread_local static std::uniform_int_distribution<std::string::size_type>
        pick(0, sizeof(chrs) - 2);

    std::string str(length, 0);
    for (auto &c : str) {
        c = chrs[pick(rg)];
    };
    return str;
}
}  // namespace

class LivestatusLocalCommandFixture : public ::testing::Test {
public:
    LivestatusLocalCommandFixture()
        : basepath{std::filesystem::temp_directory_path() / "lslc_tests" /
                   random_string(12)}
        , fullpath{basepath / "sock"}
        , ls{[this](const std::string &cmd) { sink.emplace_back(cmd); },
             fullpath, Logger::getLogger("test")} {
        ls.logger()->setLevel(LogLevel::emergency);
    }

    ~LivestatusLocalCommandFixture() override {
        ls.stop();
        std::filesystem::remove_all(basepath);
    }

    std::filesystem::path basepath;
    std::filesystem::path fullpath;
    LivestatusLocalCommand ls;
    Poller poller;
    std::vector<std::string> sink;
};

TEST_F(LivestatusLocalCommandFixture, RepeatedStartStopIsNoop) {
    EXPECT_FALSE(std::filesystem::exists(fullpath));

    ls.stop();
    EXPECT_FALSE(std::filesystem::exists(fullpath));
    ls.stop();
    ls.stop();
    EXPECT_FALSE(std::filesystem::exists(fullpath));

    ls.start();
    EXPECT_TRUE(std::filesystem::exists(fullpath));
    EXPECT_TRUE(std::filesystem::is_fifo(fullpath));
    ls.start();
    ls.start();
    EXPECT_TRUE(std::filesystem::exists(fullpath));
    EXPECT_TRUE(std::filesystem::is_fifo(fullpath));

    ls.stop();
    EXPECT_FALSE(std::filesystem::exists(fullpath));
    ls.stop();
    ls.stop();
    EXPECT_FALSE(std::filesystem::exists(fullpath));
}

TEST_F(LivestatusLocalCommandFixture, EnqueueAndDequeueOneCommand) {
    ls.start();
    ASSERT_TRUE(std::filesystem::is_fifo(fullpath));
    ASSERT_TRUE(sink.empty());

    const auto cmd = "cmd"s;

    ls.setUpPoller(poller);

    const int fd = ::open(fullpath.c_str(), O_WRONLY);
    asio::io_context io_context;
    asio::posix::stream_descriptor fifo{io_context, fd};
    asio::write(fifo, asio::buffer(frame(cmd)));

    poller.poll(10ms);
    ls.communicate(poller);
    EXPECT_EQ(cmd, !sink.empty() ? sink[0] : "failure"s);
}

TEST_F(LivestatusLocalCommandFixture, EnqueueAndDequeueManyCommands) {
    ls.start();
    ASSERT_TRUE(std::filesystem::is_fifo(fullpath));
    ASSERT_TRUE(sink.empty());

    const std::vector<std::string> cmds{"cmd0", "cmd1", "cmd3", "cmd4"};

    ls.setUpPoller(poller);

    const int fd = ::open(fullpath.c_str(), O_WRONLY);
    asio::io_context io_context;
    asio::posix::stream_descriptor fifo{io_context, fd};

    std::for_each(std::begin(cmds), std::end(cmds), [&fifo](const auto &cmd) {
        asio::write(fifo, asio::buffer(frame(cmd)));
    });

    poller.poll(10ms);
    ls.communicate(poller);
    EXPECT_EQ(cmds, sink);
}

TEST_F(LivestatusLocalCommandFixture, EnqueueWrongFormatIsNoop) {
    ls.start();
    ASSERT_TRUE(std::filesystem::is_fifo(fullpath));
    ASSERT_TRUE(sink.empty());

    const std::vector<std::string> noframing{"cmd0", "cmd1", "cmd3", "cmd4"};
    ls.setUpPoller(poller);

    const int fd = ::open(fullpath.c_str(), O_WRONLY);
    asio::io_context io_context;
    asio::posix::stream_descriptor fifo{io_context, fd};

    std::for_each(
        std::begin(noframing), std::end(noframing),
        [&fifo](const auto &cmd) { asio::write(fifo, asio::buffer(cmd)); });

    poller.poll(10ms);
    ls.communicate(poller);
    EXPECT_TRUE(sink.empty());
}

TEST_F(LivestatusLocalCommandFixture, FullBufferDelaysCommands) {
    ls.start();
    ASSERT_TRUE(std::filesystem::is_fifo(fullpath));
    ASSERT_TRUE(sink.empty());

    ls.setUpPoller(poller);

    const int fd = ::open(fullpath.c_str(), O_WRONLY);
    asio::io_context io_context;
    asio::posix::stream_descriptor fifo{io_context, fd};

    for (std::size_t i = 0; i != 512; ++i) {
        // 512 = 8192 / 16 but with framing, we are sure to overflow.
        asio::write(fifo, asio::buffer(frame(padded(i, 16))));
    }

    poller.poll(10ms);
    ls.communicate(poller);
    ASSERT_FALSE(sink.empty());
    EXPECT_NE(512U, sink.size());
    EXPECT_EQ("0000000000000000"s, sink.front());
    EXPECT_NE("0000000000000512"s, sink.back());

    const auto accepted = sink.size() - 1;
    EXPECT_EQ(padded(accepted, 16), sink.back());

    // And now, the empty buffer gets the next commands from the pipe.
    sink.clear();
    ls.communicate(poller);
    EXPECT_FALSE(sink.empty());
    EXPECT_EQ(padded(accepted + 1, 16), sink.front());
}

TEST_F(LivestatusLocalCommandFixture,
       EnqueueWrongDataOnceLosesCommandsThenRecovers) {
    ls.start();
    ASSERT_TRUE(std::filesystem::is_fifo(fullpath));
    ASSERT_TRUE(sink.empty());

    const auto cmd = "cmd"s;

    ls.setUpPoller(poller);

    const int fd = ::open(fullpath.c_str(), O_WRONLY);
    asio::io_context io_context;
    asio::posix::stream_descriptor fifo{io_context, fd};
    asio::write(fifo, asio::buffer(cmd));         // unframed
    asio::write(fifo, asio::buffer(frame(cmd)));  // lost

    poller.poll(10ms);
    ls.communicate(poller);
    EXPECT_TRUE(sink.empty());

    asio::write(fifo, asio::buffer(frame(cmd)));  // recovered
    ls.communicate(poller);
    EXPECT_EQ(1U, sink.size());
}
