// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <algorithm>
#include <chrono>
#include <functional>
#include <map>
#include <memory>
#include <string>
#include <type_traits>
#include <utility>
#include <vector>

#include "PbUtils.h"
#include "cmc/Core.h"
#include "cmc/Host.h"
#include "cmc/MonitoringLog.h"
#include "cmc/NotifyHelper.h"
#include "cmc/Object.h"
#include "cmc/World.h"
#include "config/v1/types.pb.h"
#include "gtest/gtest.h"
#include "livestatus/Logger.h"
#include "livestatus/TrialManager.h"

using namespace std::string_literals;
using namespace std::chrono_literals;

namespace {
namespace pb = checkmk::cmc::config::v1;

std::unique_ptr<Core> make_core(std::vector<std::string> &log,
                                const std::vector<std::string> &hosts) {
    auto pbcore = makeConfig();
    for (auto &&name : hosts) {
        auto pbhost = addHost(pbcore, name);
    }
    auto *logger = Logger::getLogger("test");
    const auto now = std::chrono::system_clock::time_point{};
    auto core = std::make_unique<Core>(
        nullptr, "/dev/null"s, std::make_unique<pb::Core>(pbcore), nullptr,
        logger, TrialManager{now, false});
    for (auto &&name : hosts) {
        // *state == nullptr at this point
        core->world()->getHost(name)->createNewState(now, logger);
    }
    core->monitoring_log_.sink(
        [&log](const std::string &entry) { log.push_back(entry); });
    return core;
}

std::unique_ptr<NotifyHelper> make_helper(Core *core) {
    auto helper = std::make_unique<NotifyHelper>(
        "helper"s, Logger::getLogger("test"),
        [](const Object &object) {
            return std::make_pair(object.name(), "ack"s);
        },
        core->monitoring_log_, core->notification_spooler_);
    helper->updateConfig(core->world());
    return helper;
}

}  // namespace

TEST(NotifyHelper, TestEmitsProblemRecoveryHistoryEvent) {
    std::vector<std::string> log;
    const auto core = make_core(log, {"heute"s});
    const Object *object = core->world()->getHost("heute");

    const auto helper = make_helper(core.get());
    helper->broadcast(
        NotificationInfo{*object, "PROBLEM", "author", "comment", {}});
    helper->broadcast(
        NotificationInfo{*object, "RECOVERY", "author", "comment", {}});

    ASSERT_EQ(2U, log.size());
    EXPECT_EQ(
        "HOST NOTIFICATION: check-mk-notify;heute;PENDING;check-mk-notify;;author;comment",
        log[0]);
    EXPECT_EQ(
        "HOST NOTIFICATION: check-mk-notify;heute;PENDING;check-mk-notify;;author;comment",
        log[1]);
}

TEST(NotifyHelper, TestNotifyEmitsRepeatedAlertHistoryEvents) {
    // check that second broadcast is valid (CMK-5844 case)

    std::vector<std::string> log;
    const auto core = make_core(log, {"heute"s});
    const Object *object = core->world()->getHost("heute");
    const auto helper = make_helper(core.get());
    const NotificationInfo info{*object,
                                "ALERTHANDLER AAA",
                                "author",
                                "comment",
                                {std::make_pair("ALERTHANDLEROUTPUT", "Zzz")}};

    helper->broadcast(info);
    helper->broadcast(info);

    ASSERT_EQ(2U, log.size());
    EXPECT_EQ(
        "HOST NOTIFICATION: check-mk-notify;heute;ALERTHANDLER AAA;check-mk-notify;Zzz;author;comment",
        log[0]);
    EXPECT_EQ(
        "HOST NOTIFICATION: check-mk-notify;heute;ALERTHANDLER AAA;check-mk-notify;Zzz;author;comment",
        log[1]);
}
