// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <arpa/inet.h>

#include <chrono>
#include <cstddef>
#include <cstdint>
#include <cstdlib>
#include <functional>
#include <memory>
#include <optional>
#include <string>
#include <string_view>
#include <unordered_map>
#include <utility>
#include <vector>

#include "PbUtils.h"
#include "cmc/CarbonConfig.h"
#include "cmc/ContactGroup.h"
#include "cmc/CycleTime.h"
#include "cmc/DowntimeRule.h"
#include "cmc/Entity.h"
#include "cmc/GlobalConfig.h"
#include "cmc/InfluxDBConfig.h"
#include "cmc/Object.h"
#include "cmc/PbConfigLoader.h"
#include "cmc/Timeperiod.h"
#include "config/v1/types.pb.h"
#include "google/protobuf/text_format.h"
#include "gtest/gtest.h"
#include "livestatus/Logger.h"
#include "livestatus/User.h"

using namespace std::chrono_literals;
using namespace std::string_literals;
using namespace std::string_view_literals;
namespace cfg = checkmk::cmc::config::v1;

namespace {
std::string makeConfigText(cfg::Core &config, std::string_view text) {
    /// Perform string indexing.
    std::unordered_map<std::string, uint32_t> string_pool;
    auto string_ref = [&config, &string_pool](auto value) {
        auto [pos, inserted] = string_pool.insert({std::string{value}, {}});
        if (inserted) {
            pos->second = string_pool.size() - 1;
            config.mutable_string_pool()->add_entries(std::string{value});
        }
        return pos->second;
    };

    std::string config_text;
    size_t string_pos = 0;
    while (true) {
        auto start = text.find('$', string_pos);
        if (start == std::string::npos) {
            config_text += text.substr(string_pos);
            break;
        }
        config_text += text.substr(string_pos, start - string_pos);
        ++start;  // skip first marker
        auto end = text.find('$', start);
        if (end == std::string::npos) {
            abort();  // markers mismatch
        }
        auto index = string_ref(text.substr(start, end - start));
        config_text += std::to_string(index);
        string_pos = end + 1;  // skip marker
    }
    return config_text;
}

}  // namespace

TEST(PbConfigLoader, TestDowntimeRule) {
    Logger *logger = Logger::getLogger("test");
    const std::string downtime_data = R"(
            comment_ref: $comment$
            start_time: {
                seconds: 1
                nanos: 1
            }
            end_time: {
                seconds: 2
                nanos: 2
            }
            duration: {
                seconds: 3
                nanos: 3
            }
            flex_duration: {
                seconds: 4
                nanos: 4
            }
            cycle_time: CYCLE_TIME_HOURLY
        )";
    auto config = makeConfig();
    cfg::PersistentRecurringDowntime downtime_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
        makeConfigText(config, downtime_data), &downtime_config));
    auto pl = PbConfigLoader{config};
    const auto downtime =
        pl.makeDowntimeRule(downtime_config, PoolOffset{0}, logger);
    EXPECT_EQ(downtime._comment, "comment");
    EXPECT_EQ(downtime._start_time,
              std::chrono::system_clock::time_point(1s + 1ns));
    EXPECT_EQ(downtime._end_time,
              std::chrono::system_clock::time_point(2s + 2ns));
    EXPECT_EQ(downtime._duration, 3s + 3ns);
    EXPECT_EQ(downtime._flex_duration, 4s + 4ns);
    EXPECT_EQ(downtime._cycle_time, CycleTime(CycleTime::Kind::hourly, logger));
}

TEST(PbConfigLoader, TestDowntimeRuleCycleTime) {
    Logger *l = Logger::getLogger("test");
    const std::unordered_map<std::string, CycleTime> data = {
        {"CYCLE_TIME_NONE", CycleTime{CycleTime::Kind::none, l}},
        {"CYCLE_TIME_HOURLY", CycleTime{CycleTime::Kind::hourly, l}},
        {"CYCLE_TIME_DAILY", CycleTime{CycleTime::Kind::daily, l}},
        {"CYCLE_TIME_WEEKLY", CycleTime{CycleTime::Kind::weekly, l}},
        {"CYCLE_TIME_BIWEEKLY", CycleTime{CycleTime::Kind::biweekly, l}},
        {"CYCLE_TIME_EVERY_4WEEKS",
         CycleTime{CycleTime::Kind::every_4weeks, l}},
        {"CYCLE_TIME_NTH_WEEKDAY", CycleTime{CycleTime::Kind::nth_weekday, l}},
        {"CYCLE_TIME_NTH_WEEKDAY_FROM_END",
         CycleTime{CycleTime::Kind::nth_weekday_from_end, l}},
        {"CYCLE_TIME_DAY_OF_MONTH",
         CycleTime{CycleTime::Kind::day_of_month, l}},
        {"CYCLE_TIME_EVERY_5MIN ", CycleTime{CycleTime::Kind::every_5min, l}},
    };
    auto config = makeConfig();
    for (const auto &[name, expected] : data) {
        const auto text = makeConfigText(
            config, "comment_ref: $comment$\ncycle_time: " + name + "\n");
        cfg::PersistentRecurringDowntime downtime_config;
        ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
            text, &downtime_config));
        auto pl = PbConfigLoader{config};
        const auto downtime =
            pl.makeDowntimeRule(downtime_config, PoolOffset{0}, l);
        EXPECT_EQ(downtime._cycle_time, expected) << "at the value " << name;
    }
}

TEST(PbConfigLoader, TestGeneralConfig) {
    const std::string general_data = R"(
            serial: 15
            core_directory: { str_ref: $core_dir$ }
            log_file: { str_ref: $log_file$ }
            cmk_executable: { str_ref: $cmk_executable$ }
            crash_reports_directory: { str_ref: $crash_reports_directory$ }
            license_usage_history_file: { str_ref: $license_usage_history_file$ }
            inventory_directory: { str_ref: $inventory_directory$ }
            structured_status_directory : { str_ref: $structured_status_directory$ }
            robotmk_html_log_directory : { str_ref: $robotmk_html_log_directory$ }
            logwatch_directory : { str_ref: $logwatch_directory$ }
            event_console_status_socket: { str_ref: $event_console_status_socket$ }
            state_file_created_file: { str_ref: $state_file_created_blah$ }
            licensed_state_file: { str_ref: $licensed_state_hurz$ }
            log_level_core: LOG_LEVEL_ALERT
            log_level_downtime: LOG_LEVEL_CRITICAL
            log_microtime: true
            core_dumps_enabled: true
            housekeeping_interval: {
                seconds: 5
                nanos: 1000000
            }
            state_retention_interval: { seconds: 9 }
            event_console_enabled: true
            edition_ref: $cool_edition$
            max_long_output_size: 20000
        )";
    auto config = makeConfig();
    const auto config_text = makeConfigText(config, general_data);
    cfg::General general_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(config_text,
                                                              &general_config));
    const auto pl = PbConfigLoader{config};
    const auto general = pl.makeGeneral(general_config);

    EXPECT_EQ(general.serial_, 15U);
    EXPECT_EQ(general.core_directory_, "core_dir");
    EXPECT_EQ(general.log_file_, "log_file");
    EXPECT_EQ(general.cmk_executable_, "cmk_executable");
    EXPECT_EQ(general.crash_reports_directory_, "crash_reports_directory");
    EXPECT_EQ(general.license_usage_history_file_,
              "license_usage_history_file");
    EXPECT_EQ(general.inventory_directory_, "inventory_directory");
    EXPECT_EQ(general.structured_status_directory_,
              "structured_status_directory");
    EXPECT_EQ(general.robotmk_html_log_directory_,
              "robotmk_html_log_directory");
    EXPECT_EQ(general.logwatch_directory_, "logwatch_directory");
    EXPECT_EQ(general.event_console_status_socket_,
              "event_console_status_socket");
    EXPECT_EQ(general.state_file_created_file_, "state_file_created_blah");
    EXPECT_EQ(general.licensed_state_file_, "licensed_state_hurz");
    EXPECT_EQ(general.log_level_core_, LogLevel::alert);
    EXPECT_EQ(general.log_level_downtime_, LogLevel::critical);
    EXPECT_TRUE(general.log_microtime_);
    EXPECT_TRUE(general.core_dumps_enabled_);
    EXPECT_EQ(general.houskeeping_interval_, 5001ms);
    EXPECT_EQ(general.state_retention_interval_, 9s);
    EXPECT_TRUE(general.event_console_enabled_);
    EXPECT_EQ(general.edition_, "cool_edition");
    EXPECT_EQ(general.max_long_output_size_, 20000);
}

TEST(PbConfigLoader, TestAlertHandlerConfig) {
    const std::string alert_handler_data = R"(
            enabled: true
            event_types: 3
            log_level: LOG_LEVEL_ERROR
        )";
    auto config = makeConfig();
    const auto config_text = makeConfigText(config, alert_handler_data);
    cfg::AlertHandler alert_handler_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
        config_text, &alert_handler_config));
    const auto pl = PbConfigLoader{config};
    const auto alert_handler = pl.makeAlertHandlerConfig(alert_handler_config);
    EXPECT_TRUE(alert_handler.enabled_);
    EXPECT_EQ(alert_handler.event_types_, 3U);
    EXPECT_EQ(alert_handler.log_level_, LogLevel::error);
}

TEST(PbConfigLoader, TestSmartPingConfig) {
    const std::string smartping_data = R"(
            check_interval: {
                seconds: 3
                nanos: 1000
            }
            omit_payload: true
            num_sockets: 5
            ignore_rst: true
            wait_duration: {
                seconds: 7
                nanos: 1000
            }
            bulk_size: 7
            log_level: LOG_LEVEL_WARNING
            icmp_sender_executable: { str_ref: $icmp_sender_executable$ }
            icmp_receiver_executable: { str_ref: $icmp_receiver_executable$ }
        )";
    auto config = makeConfig();
    const auto config_text = makeConfigText(config, smartping_data);
    cfg::SmartPing smartping_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
        config_text, &smartping_config));
    const auto pl = PbConfigLoader{config};
    const auto smartping = pl.makeSmartPingConfig(smartping_config);
    EXPECT_EQ(smartping.check_interval_, 3'000'001us);
    EXPECT_TRUE(smartping.omit_payload_);
    EXPECT_EQ(smartping.num_sockets_, 5U);
    EXPECT_TRUE(smartping.ignore_rst_);
    EXPECT_EQ(smartping.wait_duration_, 7'000'001'000ns);
    EXPECT_EQ(smartping.bulk_size_, 7U);
    EXPECT_EQ(smartping.log_level_, LogLevel::warning);
    EXPECT_EQ(smartping.icmp_sender_executable_, "icmp_sender_executable");
    EXPECT_EQ(smartping.icmp_receiver_executable_, "icmp_receiver_executable");
}

TEST(PbConfigLoader, TestCheckingConfig) {
    const std::string checking_data = R"(
            generic_helper_executable: {str_ref: $generic_helper_executable$ }
            num_generic_helpers: 7
            num_real_time_helpers: 3
            real_time_port: {
                number:5432
            }
            fetcher_executable: {str_ref: $fetcher_executable$ }
            num_fetchers: 11
            fetchers_enabled: true
            num_checkers: 19
            log_level: LOG_LEVEL_NOTICE
            debug: true
            initial_scheduling_burst: 111
            initial_generic_scheduling_interval: {
                seconds: 1
                nanos: 1
            }
            initial_checkmk_scheduling_interval: {
                seconds: 2
                nanos: 2
            }
        )";
    auto config = makeConfig();
    const auto config_text = makeConfigText(config, checking_data);
    cfg::Checking checking_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
        config_text, &checking_config));
    const auto pl = PbConfigLoader{config};
    const auto checking = pl.makeCheckingConfig(checking_config);
    EXPECT_EQ(checking.generic_helper_executable_, "generic_helper_executable");
    EXPECT_EQ(checking.num_generic_helpers_, 7U);
    EXPECT_EQ(checking.num_real_time_helpers_, 3U);
    EXPECT_EQ(checking.real_time_port_, 5432U);
    EXPECT_EQ(checking.fetcher_executable_, "fetcher_executable");
    EXPECT_EQ(checking.num_fetchers_, 11U);
    EXPECT_TRUE(checking.fetchers_enabled_);
    EXPECT_EQ(checking.num_checkers_, 19U);
    EXPECT_EQ(checking.log_level_, LogLevel::notice);
    EXPECT_TRUE(checking.debug_);
    EXPECT_EQ(checking.initial_scheduling_burst_, 111U);
    EXPECT_EQ(checking.initial_generic_scheduling_interval_, 1'000'000'001ns);
    EXPECT_EQ(checking.initial_checkmk_scheduling_interval_, 2'000'000'002ns);
}

TEST(PbConfigLoader, TestLivestatusConfig) {
    const std::string livestatus_data = R"(
            socket: { str_ref: $socket$ }
            command_pipe: { str_ref: $command_pipe$ }
            stack_size: 13
            num_threads: 99
            log_level: LOG_LEVEL_INFORMATIONAL
            service_authorization: SERVICE_AUTHORIZATION_STRICT
            group_authorization: GROUP_AUTHORIZATION_LOOSE
        )";
    auto config = makeConfig();
    const auto config_text = makeConfigText(config, livestatus_data);
    cfg::Livestatus livestatus_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
        config_text, &livestatus_config));
    const auto pl = PbConfigLoader{config};
    const auto livestatus = pl.makeLivestatusConfig(livestatus_config);
    EXPECT_EQ(livestatus.socket_, "socket");
    EXPECT_EQ(livestatus.command_pipe_, "command_pipe");
    EXPECT_EQ(livestatus.stack_size_, 13U);
    EXPECT_EQ(livestatus.num_threads_, 99U);
    EXPECT_EQ(livestatus.log_level_, LogLevel::informational);
    EXPECT_EQ(livestatus.service_authorization_, ServiceAuthorization::strict);
    EXPECT_EQ(livestatus.group_authorization_, GroupAuthorization::loose);
}

TEST(PbConfigLoader, TestNotificationHandlerConfig) {
    const std::string notification_handler_data = R"(
            debug: true
            log_level: LOG_LEVEL_DEBUG
        )";
    auto config = makeConfig();
    const auto config_text = makeConfigText(config, notification_handler_data);
    cfg::NotificationHandler notification_handler_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
        config_text, &notification_handler_config));
    const auto pl = PbConfigLoader{config};
    const auto notification_handler =
        pl.makeNotificationHandlerConfig(notification_handler_config);
    EXPECT_TRUE(notification_handler.debug_);
    EXPECT_EQ(notification_handler.log_level_, LogLevel::debug);
}

TEST(PbConfigLoader, TestHistoryConfig) {
    const std::string history_data = R"(
            max_cached_messages: 95
            max_lines_per_file: 14
            cache_horizon: { seconds: 5 }
            max_core_downtime: { seconds: 9 }
            file: { str_ref: $file$ }
            archive_directory: { str_ref: $archive_directory$ }
            rotation_interval: LOG_ROTATION_INTERVAL_MINUTE
            rotation_size_limit: 100
        )";
    // TODO(sk): estimate parametrization of the test to check optional twice
    auto config = makeConfig();
    const auto config_text = makeConfigText(config, history_data);
    cfg::History history_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(config_text,
                                                              &history_config));
    const auto pl = PbConfigLoader{config};
    const auto history = pl.makeHistoryConfig(history_config);
    EXPECT_EQ(history.max_cached_messages_, 95U);
    EXPECT_EQ(history.max_lines_per_file_, 14U);
    EXPECT_EQ(history.cache_horizon_, 5s);
    EXPECT_EQ(history.max_core_downtime_, 9s);
    EXPECT_EQ(history.file_, "file");
    EXPECT_EQ(history.archive_directory_, "archive_directory");
    EXPECT_EQ(history.rotation_size_limit_, 100U);
}

TEST(PbConfigLoader, TestRRDConfig) {
    const std::string rrd_data = R"(
            single_directory: { str_ref: $single_directory$ }
            multiple_directory: { str_ref: $multiple_directory$ }
            rrdcached_socket: { str_ref: $rrdcached_socket$ }
            log_level: LOG_LEVEL_ALERT
            pnp4nagios_enabled: true
            pnp_update_delay: { seconds: 5 }
            pnp_update_on_restart: true
        )";
    auto config = makeConfig();
    const auto config_text = makeConfigText(config, rrd_data);
    cfg::RRD rrd_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(config_text,
                                                              &rrd_config));
    const auto pl = PbConfigLoader{config};
    const auto rrd = pl.makeRRDConfig(rrd_config);
    EXPECT_EQ(rrd.single_directory_, "single_directory");
    EXPECT_EQ(rrd.multiple_directory_, "multiple_directory");
    EXPECT_EQ(rrd.rrdcached_socket_, "rrdcached_socket");
    EXPECT_EQ(rrd.log_level_, LogLevel::alert);
    EXPECT_TRUE(rrd.pnp4nagios_enabled_);
    EXPECT_EQ(rrd.pnp_update_delay_, 5s);
    EXPECT_TRUE(rrd.pnp_update_on_restart_);
}

constexpr auto influx_db_connection = R"(
    id_ref: $id$
    enabled: true
    protocol_ref: $protocol$
    host_name_ref:  $host$
    port: { number: 6644}
    ca_cert_file: { str_ref: $ca_cert_file$ }
    use_certificate_check: true
    api_path_ref: $api_path$
    token_ref: $token$
    retry_interval {
        seconds: 1
        nanos: 1000
    }
)"sv;

TEST(PbConfigLoader, TestInfluxDbConnection) {
    const std::string connection_data{influx_db_connection};
    auto config = makeConfig();
    const auto config_text = makeConfigText(config, connection_data);
    cfg::InfluxDbConnection connection_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
        config_text, &connection_config));
    const auto pl = PbConfigLoader{config};
    const auto connection = pl.makeInfluxDbConnection(connection_config);
    EXPECT_EQ(connection.id, "id");
    EXPECT_TRUE(connection.enabled);
    EXPECT_EQ(connection.address.protocol, "protocol");
    EXPECT_EQ(connection.address.host, "host");
    EXPECT_EQ(connection.address.port, 6644U);
    EXPECT_EQ(connection.auth.ca_cert_file, "ca_cert_file");
    EXPECT_TRUE(connection.auth.use_certificate_check);
    EXPECT_EQ(std::string{connection.token}, "Token token");
    EXPECT_EQ(connection.retry_interval, 1000'001us);
}

TEST(PbConfigLoader, TestInfluxDbConfig) {
    const std::string connection_data =
        "{\n" + std::string{influx_db_connection} + "}\n";
    const std::string influx_db_data = "connections: [\n" + connection_data +
                                       "," + connection_data + "]\n" +
                                       "log_level: LOG_LEVEL_WARNING";
    auto config = makeConfig();
    const auto config_text = makeConfigText(config, influx_db_data);
    cfg::InfluxDb influx_db_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
        config_text, &influx_db_config));
    const auto pl = PbConfigLoader{config};
    const auto influx_db = pl.makeInfluxDbConfig(influx_db_config);
    EXPECT_EQ(influx_db.connections_.size(), 2U);
    EXPECT_EQ(influx_db.log_level_, LogLevel::warning);
}

// NOTE: should be in in sync with python code which should look
// connection.ip = int(IPv4Address("1.2.3.4")) # ip is 0x01020304
constexpr auto carbon_connection = R"(
    ip: { number: 0x01020304 }
    port: { number: 0xCCDD }
    connection_prefix_ref: $connection_prefix$
    mangling: CARBON_CONNECTION_MANGLING_C_STYLE
)"sv;

TEST(PbConfigLoader, TestCarbonConnection) {
    const std::string connection_data{carbon_connection};
    auto config = makeConfig();
    const auto config_text = makeConfigText(config, connection_data);
    cfg::CarbonConnection connection_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
        config_text, &connection_config));
    const auto pl = PbConfigLoader{config};
    const auto actual = pl.makeCarbonConnection(connection_config);
    const CarbonConnectionConfig expected{
        htonl(0x01020304), 0xCCDD, 15s, "connection_prefix",
        CarbonConnectionConfig::Mangling::c_style};
    EXPECT_EQ(actual, expected);
}

TEST(PbConfigLoader, TestCarbonConfig) {
    const std::string connection_data =
        "{\n" + std::string{carbon_connection} + "}\n";
    const std::string carbon_data = "connections: [\n" + connection_data + "," +
                                    connection_data + "]\n" +
                                    "log_level: LOG_LEVEL_NOTICE";
    auto config = makeConfig();
    const auto config_text = makeConfigText(config, carbon_data);
    cfg::Carbon carbon_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(config_text,
                                                              &carbon_config));
    const auto pl = PbConfigLoader{config};
    const auto carbon = pl.makeCarbonConfig(carbon_config);
    EXPECT_EQ(carbon.connections_.size(), 2U);
    EXPECT_EQ(carbon.log_level_, LogLevel::notice);
}

constexpr auto contact_group_data = R"(
        name_ref: $name$
        alias_ref: $alias$
    )"sv;

TEST(PbConfigLoader, TestContactGroup) {
    auto config = makeConfig();
    const auto config_text = makeConfigText(config, contact_group_data);
    cfg::ContactGroup contact_group_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
        config_text, &contact_group_config));
    const auto pl = PbConfigLoader{config};
    const auto contact_group = pl.makeContactGroup(contact_group_config);
    EXPECT_EQ(contact_group->name(), "name");
    EXPECT_EQ(contact_group->alias(), "alias");
}

TEST(PbConfigLoader, TestTimeperiod) {
    const auto timeperiod_data = R"(
    name_ref: $name$
    alias_ref: $alias$
    transitions: [
        {seconds:100, nanos:1},
        {seconds:200, nanos:2}
    ]
    )"sv;

    auto config = makeConfig();
    const auto config_text = makeConfigText(config, timeperiod_data);
    cfg::Timeperiod timeperiod_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
        config_text, &timeperiod_config));
    const auto pl = PbConfigLoader{config};
    const auto timeperiod = pl.makeTimeperiod(timeperiod_config);
    EXPECT_EQ(timeperiod.name(), "name");
    EXPECT_EQ(timeperiod.alias(), "alias");
    EXPECT_EQ(timeperiod.numTransitions(), 2);
    EXPECT_EQ(timeperiod.transition(0),
              std::chrono::system_clock::time_point(100s + 1ns));
    EXPECT_EQ(timeperiod.transition(1),
              std::chrono::system_clock::time_point(200s + 2ns));
}

TEST(PbConfigLoader, TestEntity) {
    const auto entity_data = R"(
    name_ref: $name$
    attributes: [
        { key_ref: $key1$, value_ref: $value1$ },
        { key_ref: $key2$, value_ref: $value2$ }
    ]
    )"sv;

    auto config = makeConfig();
    const auto config_text = makeConfigText(config, entity_data);
    cfg::Entity entity_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(config_text,
                                                              &entity_config));
    const auto pl = PbConfigLoader{config};
    const auto entity = pl.makeEntity(entity_config, PoolOffset{0U});
    EXPECT_EQ(entity.name(), "name");
    EXPECT_EQ(entity.lookup({}, "key1"), "value1");
    EXPECT_EQ(entity.lookup({}, "key2"), "value2");
}

TEST(PbConfigLoader, TestNotificationOptions) {
    const std::vector<std::pair<std::string, NotificationOptions>> data = {
        {"host_up", NotificationOptions::host_up},
        {"host_down", NotificationOptions::host_down},
        {"host_unreachable", NotificationOptions::host_unreachable},
        {"host_flapping", NotificationOptions::host_flapping},
        {"host_downtime", NotificationOptions::host_downtime},
        {"service_ok", NotificationOptions::service_ok},
        {"service_warning", NotificationOptions::service_warning},
        {"service_critical", NotificationOptions::service_critical},
        {"service_unknown", NotificationOptions::service_unknown},
        {"service_flapping", NotificationOptions::service_flapping},
        {"service_downtime", NotificationOptions::service_downtime}};
    auto config = makeConfig();
    for (const auto &[name, value] : data) {
        const auto config_text = makeConfigText(config, name + ": true");
        cfg::NotificationOptions options_config;
        ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
            config_text, &options_config));
        const auto pl = PbConfigLoader{config};
        const auto options = pl.makeNotificationOptions(options_config);
        EXPECT_EQ(options, value) << "at the value " << name;
    }
}

TEST(PbConfigLoader, TestCarbonSendMetricParts) {
    using Metrics = CarbonSendMetricParts;
    const std::vector<
        std::pair<std::string, std::function<bool(const Metrics &)>>>
        data = {
            {"uom", [](const auto &m) { return m.uom; }},
            {"warn", [](const auto &m) { return m.warn; }},
            {"crit", [](const auto &m) { return m.crit; }},
            {"min", [](const auto &m) { return m.min; }},
            {"max", [](const auto &m) { return m.max; }},
            {"state", [](const auto &m) { return m.state; }},
            {"scheduled_downtime_depth",
             [](const auto &m) { return m.scheduled_downtime_depth; }},
            {"acknowledged", [](const auto &m) { return m.acknowledged; }},
        };
    auto config = makeConfig();
    for (const auto &[name, checker] : data) {
        const auto config_text = makeConfigText(config, name + ": true");
        cfg::CarbonSendMetricParts metrics_config;
        ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
            config_text, &metrics_config));
        const auto pl = PbConfigLoader{config};
        const auto metrics = pl.makeCarbonSendMetricParts(metrics_config);
        EXPECT_TRUE(checker(metrics)) << "at the value " << name;
    }
}

TEST(PbConfigLoader, TestInfluxDbSenderConfigParts) {
    using Metrics = influxdb::SenderMetrics;
    const std::unordered_map<std::string, std::function<bool(const Metrics &)>>
        data = {
            {"metric_value", [](const auto &m) { return m.metric_value; }},
            {"metric_hst_state",
             [](const auto &m) { return m.metric_hst_state; }},
            {"metric_svc_state",
             [](const auto &m) { return m.metric_svc_state; }},
            {"add_hst_state", [](const auto &m) { return m.add_hst_state; }},
            {"add_svc_state", [](const auto &m) { return m.add_svc_state; }},
            {"add_downtime", [](const auto &m) { return m.add_downtime; }},
            {"add_acknowledgement",
             [](const auto &m) { return m.add_acknowledgement; }},
            {"add_uom", [](const auto &m) { return m.add_uom; }},
            {"add_thresholds", [](const auto &m) { return m.add_thresholds; }},
            {"tag_hst_name", [](const auto &m) { return m.tag_hst_name; }},
            {"tag_svc_name", [](const auto &m) { return m.tag_svc_name; }},
            {"tag_hst_alias", [](const auto &m) { return m.tag_hst_alias; }},
            {"tag_hst_label", [](const auto &m) { return m.tag_hst_label; }},
            {"tag_svc_label", [](const auto &m) { return m.tag_svc_label; }},
        };
    auto config = makeConfig();
    for (const auto &[name, checker] : data) {
        const auto config_text = makeConfigText(config,
                                                "connection_ref: $connection$\n"
                                                "org_ref: $org$\n"
                                                "bucket_ref: $bucket$\n"s +
                                                    name + ": true\n");
        cfg::InfluxDbSender sender_config;
        ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
            config_text, &sender_config));
        const auto pl = PbConfigLoader{config};
        const auto sender =
            pl.makeInfluxDbSenderConfig(sender_config, PoolOffset{0});
        EXPECT_TRUE(checker(sender.parts)) << "at the value " << name;
    }
}

TEST(PbConfigLoader, TestInfluxDbSenderConfig) {
    const std::string sender_data = R"(
            connection_ref: $connection$
            org_ref: $org$
            bucket_ref: $bucket$
        )";
    auto config = makeConfig();
    const auto config_text = makeConfigText(config, sender_data);
    cfg::InfluxDbSender sender_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(config_text,
                                                              &sender_config));
    const auto pl = PbConfigLoader{config};
    const auto sender =
        pl.makeInfluxDbSenderConfig(sender_config, PoolOffset{0});
    EXPECT_EQ(sender.connection, "connection");
    EXPECT_EQ(sender.org, "org");
    EXPECT_EQ(sender.bucket, "bucket");
    EXPECT_FALSE(sender.parts.add_uom);
}

TEST(PbConfigLoader, TestObjectGroup) {
    const std::string object_group_data = R"(
            name_ref: $name$
            alias_ref: $alias$
        )";
    auto config = makeConfig();
    const auto config_text = makeConfigText(config, object_group_data);
    cfg::ObjectGroup object_group_config;
    ASSERT_TRUE(google::protobuf::TextFormat::ParseFromString(
        config_text, &object_group_config));
    const auto pl = PbConfigLoader{config};
    const auto [name, alias] = pl.makeObjectGroup(object_group_config);
    EXPECT_EQ(name, "name");
    EXPECT_EQ(alias, "alias");
}
