// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <chrono>
#include <functional>
#include <memory>
#include <ratio>
#include <string>

#include "PbUtils.h"
#include "cmc/Comment.h"
#include "cmc/Core.h"
#include "cmc/CycleTime.h"
#include "cmc/Downtime.h"
#include "cmc/GlobalState.h"
#include "cmc/Host.h"
#include "cmc/ObjectResolver.h"
#include "cmc/PbStateLoader.h"
#include "cmc/PbStateSaver.h"
#include "cmc/Service.h"
#include "cmc/State.h"
#include "cmc/World.h"
#include "cmc/cmc.h"
#include "config/v1/types.pb.h"
#include "gtest/gtest.h"
#include "livestatus/Logger.h"
#include "livestatus/TrialManager.h"
#include "state/v1/types.pb.h"

using namespace std::chrono_literals;
using namespace std::string_literals;

namespace {
namespace pb = checkmk::cmc::config::v1;

std::unique_ptr<Core> make_core(const std::string &host) {
    auto pbcore = makeConfig();
    auto pbhost = addHost(pbcore, host);
    auto *logger = Logger::getLogger("test");
    const auto now = std::chrono::system_clock::time_point{};
    auto core = std::make_unique<Core>(
        nullptr, "/dev/null"s, std::make_unique<pb::Core>(pbcore), nullptr,
        logger, TrialManager{now, false});
    // *state == nullptr at this point
    core->world()->getHost(host)->createNewState(now, logger);
    return core;
}

std::unique_ptr<Core> make_core(const std::string &host,
                                const std::string &service) {
    auto pbcore = makeConfig();
    auto pbhost = addHost(pbcore, host);
    auto pbsvc = addService(pbcore, pbhost, service);
    auto *logger = Logger::getLogger("test");
    const auto now = std::chrono::system_clock::time_point{};
    auto core = std::make_unique<Core>(
        nullptr, "/dev/null"s, std::make_unique<pb::Core>(pbcore), nullptr,
        logger, TrialManager{now, false});
    // *state == nullptr at this point
    core->world()->getHost(host)->createNewState(now, logger);
    return core;
}
}  // namespace

TEST(TestStateConfigProtoParser, GlobalStateIdentity) {
    auto saver = PbStateSaver{};
    auto loader = PbStateLoader{};

    auto global_state = GlobalState{GlobalStateFields{
        ._enable_notifications = true,
        ._execute_service_checks = false,
        ._execute_host_checks = true,
        ._enable_flap_detection = false,
        ._process_performance_data = true,
        ._enable_alert_handlers = false,
        ._notification_next_problem_id = 1,
        ._state_file_created =
            std::chrono::time_point_cast<std::chrono::microseconds>(
                std::chrono::system_clock::now())}};
    // Change the value of a couple fields to make sure the test
    // is not fortuitously green.
    global_state.setExecuteServiceChecks(false);
    global_state.setProcessPerformanceData(false);

    EXPECT_EQ(loader.loadGlobalState(saver.saveGlobalState(global_state)),
              global_state);
}

TEST(TestStateConfigProtoParser, StateIdentity) {
    auto saver = PbStateSaver{};
    auto loader = PbStateLoader{};

    auto now = std::chrono::system_clock::now();

    auto state =
        State{.acknowledgement_type_ = AcknowledgementType::normal,
              .check_result_ = {.output_ = "hello output",
                                .long_output_ = "hello plugin",
                                .perf_data_ = "hello perf data",
                                .cached_at_ = now,
                                .cache_interval_ = 69s},
              .downtime_ = {.pending_flex_ = 42, .scheduled_depth_ = 42},
              .enabling_ = {.value_ = 1337, .modified_ = 1337},
              .flappiness_ = {.is_flapping_ = false, .value_ = 6.9},
              .hard_state_ =
                  {
                      .last_ = 42,
                      .previous_ = 69,
                      .changed_ = true,
                      .last_time_ = now,
                  },
              .notification_ = {.current_number_ = 42,
                                .last_time_ = now,
                                .next_time_ = now,
                                .pending_kind_ = NotificationKind::normal,
                                .problem_id_ = 42,
                                .problem_sent_ = true,
                                .postponement_reason_ = "hello reason"},
              .current_attempt_ = 42,
              .has_been_checked_ = true,
              .execution_time_ = std::chrono::duration<double>{1337},
              .latency_ = std::chrono::duration<double>{1337},
              .last_check_ = now,
              .last_check_started_ = now,
              .last_state_change_ = now,
              .last_time_state_ = {now, now, now, now},
              .next_check_ = now,
              .current_state_ = 69,
              .initial_state_ = 42,
              .last_state_ = 69,
              .state_type_ = StateType::hard};
    EXPECT_EQ(loader.loadState(saver.saveState(state)), state);
}

TEST(StateConfigProtoParser, DowntimeIdentity) {
    auto saver = PbStateSaver{};
    auto loader = PbStateLoader{};
    auto core = make_core({"host_name"s});
    auto resolver = ObjectResolver{core->world(), {}};
    auto *hst = core->world()->getHost("host_name"s);
    auto now = std::chrono::system_clock::now();
    auto downtime = Downtime{.core_ = *core,
                             .logger_downtime_ = core->_logger_downtime,
                             .logger_notification_ = core->_logger_notification,
                             ._object = hst,
                             ._id = 42,
                             ._entry_time = now,
                             ._author = "author",
                             ._comment = "text",
                             ._status = Downtime::Status::pending,
                             ._start_time = now,
                             ._end_time = now + 10min,
                             ._flex_downtime_start = now + 20min,
                             ._fixed = true,
                             ._cycle_time = CycleTime{999, nullptr},
                             ._duration = 1h,
                             ._hot = true,
                             ._origin_is_rule = false,
                             ._rule = nullptr};

    EXPECT_EQ(loader.loadDowntime(*hst, *core, saver.saveDowntime(downtime)),
              downtime);
}

TEST(StateConfigProtoParser, CommentIdentity) {
    auto saver = PbStateSaver{};
    auto loader = PbStateLoader{};
    auto core = make_core("host_name", "service name");
    auto resolver = ObjectResolver{core->world(), {}};
    auto *hst = core->world()->getHost("host_name");
    auto *svc = core->world()->getService("host_name", "service name");
    auto now = std::chrono::system_clock::now();
    auto hst_comment = Comment{hst,    42,
                               now,    "author",
                               "text", Comment::Type::user,
                               now,    true,
                               false,  Comment::Source::internal};
    auto svc_comment = Comment{svc,    42,
                               now,    "author",
                               "text", Comment::Type::user,
                               now,    true,
                               false,  Comment::Source::internal};

    EXPECT_EQ(loader.loadComment(*hst, saver.saveComment(hst_comment)),
              hst_comment);
    EXPECT_NE(loader.loadComment(*hst, saver.saveComment(hst_comment)),
              svc_comment);
    EXPECT_EQ(loader.loadComment(*svc, saver.saveComment(svc_comment)),
              svc_comment);
    EXPECT_NE(loader.loadComment(*svc, saver.saveComment(svc_comment)),
              hst_comment);
}
