// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <vector>

#include "gtest/gtest.h"
#include "livestatus/Metric.h"
#include "livestatus/PerformanceData.h"

TEST(PerformanceDataTest, Empty) {
    const PerformanceData pd{"", "default_command"};
    EXPECT_TRUE(pd.empty());
    for (const auto &metric : pd) {
        ADD_FAILURE() << "unexpected metric " << metric.name().string();
    }
    EXPECT_EQ("default_command", pd.checkCommandName());
}

TEST(PerformanceDataTest, OnlySpaces) {
    const PerformanceData pd{"    ", "default_command"};
    EXPECT_TRUE(pd.empty());
    for (const auto &metric : pd) {
        ADD_FAILURE() << "unexpected metric " << metric.name().string();
    }
    EXPECT_EQ("default_command", pd.checkCommandName());
}

TEST(PerformanceDataTest, CommandNameAtEnd) {
    const PerformanceData pd{
        "   time=0.073836s;;;0.000000; size=557B;;;0; [check_http]   ",
        "default_command"};
    auto it = pd.begin();

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("time", it->name().string());
    EXPECT_EQ("time", it->mangled_name().string());
    EXPECT_EQ("0.073836", it->value());
    EXPECT_EQ("s", it->uom());
    EXPECT_EQ("", it->warn());
    EXPECT_EQ("", it->crit());
    EXPECT_EQ("0.000000", it->min());
    EXPECT_EQ("", it->max());
    ++it;

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("size", it->name().string());
    EXPECT_EQ("size", it->mangled_name().string());
    EXPECT_EQ("557", it->value());
    EXPECT_EQ("B", it->uom());
    EXPECT_EQ("", it->warn());
    EXPECT_EQ("", it->crit());
    EXPECT_EQ("0", it->min());
    EXPECT_EQ("", it->max());
    ++it;

    EXPECT_EQ(pd.end(), it);
    EXPECT_EQ("check_http", pd.checkCommandName());
}

TEST(PerformanceDataTest, CommandNameLikeLabel) {
    const PerformanceData pd{
        "[abcd.abc:service#ABCMetrics,JobsCount]=0;15000;16000",
        "default_command"};
    auto it = pd.begin();

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("[abcd.abc:service#ABCMetrics,JobsCount]", it->name().string());
    EXPECT_EQ("[abcd.abc_service#ABCMetrics,JobsCount]",
              it->mangled_name().string());
    EXPECT_EQ("0", it->value());
    EXPECT_EQ("", it->uom());
    EXPECT_EQ("15000", it->warn());
    EXPECT_EQ("16000", it->crit());
    EXPECT_EQ("", it->min());
    EXPECT_EQ("", it->max());
    ++it;

    EXPECT_EQ(pd.end(), it);
    EXPECT_EQ("default_command", pd.checkCommandName());
}

TEST(PerformanceDataTest, Various) {
    const PerformanceData pd{
        "/=104611.519531;93214.590625;104866.414453;0;116518.238281 "
        "fs_size=116518.238281;;;; "
        "growth=-4913.655254;;;; "
        "trend=225.779229;;;0;4854.926595 "
        "inodes_used=1099557;6827212.8;7206502.4;0;7585792",
        "default_command"};
    auto it = pd.begin();

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("/", it->name().string());
    EXPECT_EQ("_", it->mangled_name().string());
    EXPECT_EQ("104611.519531", it->value());
    EXPECT_EQ("", it->uom());
    EXPECT_EQ("93214.590625", it->warn());
    EXPECT_EQ("104866.414453", it->crit());
    EXPECT_EQ("0", it->min());
    EXPECT_EQ("116518.238281", it->max());
    ++it;

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("fs_size", it->name().string());
    EXPECT_EQ("fs_size", it->mangled_name().string());
    EXPECT_EQ("116518.238281", it->value());
    EXPECT_EQ("", it->uom());
    EXPECT_EQ("", it->warn());
    EXPECT_EQ("", it->crit());
    EXPECT_EQ("", it->min());
    EXPECT_EQ("", it->max());
    ++it;

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("growth", it->name().string());
    EXPECT_EQ("growth", it->mangled_name().string());
    EXPECT_EQ("-4913.655254", it->value());
    EXPECT_EQ("", it->uom());
    EXPECT_EQ("", it->warn());
    EXPECT_EQ("", it->crit());
    EXPECT_EQ("", it->min());
    EXPECT_EQ("", it->max());
    ++it;

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("trend", it->name().string());
    EXPECT_EQ("trend", it->mangled_name().string());
    EXPECT_EQ("225.779229", it->value());
    EXPECT_EQ("", it->uom());
    EXPECT_EQ("", it->warn());
    EXPECT_EQ("", it->crit());
    EXPECT_EQ("0", it->min());
    EXPECT_EQ("4854.926595", it->max());
    ++it;

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("inodes_used", it->name().string());
    EXPECT_EQ("inodes_used", it->mangled_name().string());
    EXPECT_EQ("1099557", it->value());
    EXPECT_EQ("", it->uom());
    EXPECT_EQ("6827212.8", it->warn());
    EXPECT_EQ("7206502.4", it->crit());
    EXPECT_EQ("0", it->min());
    EXPECT_EQ("7585792", it->max());
    ++it;

    EXPECT_EQ(pd.end(), it);
    EXPECT_EQ("default_command", pd.checkCommandName());
}

TEST(PerformanceDataTest, QuotesSimple) {
    const PerformanceData pd{"'tbs-spotlight-alloc'=100MB;;;0;32767",
                             "default_command"};
    auto it = pd.begin();

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("tbs-spotlight-alloc", it->name().string());
    EXPECT_EQ("tbs-spotlight-alloc", it->mangled_name().string());
    EXPECT_EQ("100", it->value());
    EXPECT_EQ("MB", it->uom());
    EXPECT_EQ("", it->warn());
    EXPECT_EQ("", it->crit());
    EXPECT_EQ("0", it->min());
    EXPECT_EQ("32767", it->max());
    ++it;

    EXPECT_EQ(pd.end(), it);
    EXPECT_EQ("default_command", pd.checkCommandName());
}

TEST(PerformanceDataTest, SpecialCharsInQuotes) {
    const PerformanceData pd{
        "'Maker''s Mark 45% & Smørrebrød'=100MB;12;34;56;78",
        "default_command"};
    auto it = pd.begin();

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("Maker's Mark 45% & Smørrebrød", it->name().string());
    EXPECT_EQ("Maker's_Mark_45%_&_Smørrebrød", it->mangled_name().string());
    EXPECT_EQ("100", it->value());
    EXPECT_EQ("MB", it->uom());
    EXPECT_EQ("12", it->warn());
    EXPECT_EQ("34", it->crit());
    EXPECT_EQ("56", it->min());
    EXPECT_EQ("78", it->max());
    ++it;

    EXPECT_EQ(pd.end(), it);
    EXPECT_EQ("default_command", pd.checkCommandName());
}

TEST(PerformanceDataTest, EqualsCharInQuotes) {
    const PerformanceData pd{"'a² + b² = c²'=100MB;12;34;56;78",
                             "default_command"};
    auto it = pd.begin();

    // NOTE: Although "officially" allowed, we explicitly disallow the equals
    // char in labels, see PerformanceData::addMetric.
    if (false) {  // NOLINT
        ASSERT_NE(pd.end(), it);
        EXPECT_EQ("a² + b² = c²", it->name().string());
        EXPECT_EQ("a²_+_b²_=_c²", it->mangled_name().string());
        EXPECT_EQ("100", it->value());
        EXPECT_EQ("MB", it->uom());
        EXPECT_EQ("12", it->warn());
        EXPECT_EQ("34", it->crit());
        EXPECT_EQ("56", it->min());
        EXPECT_EQ("78", it->max());
        ++it;
    }

    EXPECT_EQ(pd.end(), it);
    EXPECT_EQ("default_command", pd.checkCommandName());
}

TEST(PerformanceDataTest, MangledName) {
    const PerformanceData pd{
        "'foo max/GNARK\\max:KENU'=12.34sheppey;56;78;89;90",
        "default_command"};
    auto it = pd.begin();

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("foo max/GNARK\\max:KENU", it->name().string());
    EXPECT_EQ("foo_max_GNARK_max_KENU", it->mangled_name().string());
    EXPECT_EQ("12.34", it->value());
    EXPECT_EQ("sheppey", it->uom());
    EXPECT_EQ("56", it->warn());
    EXPECT_EQ("78", it->crit());
    EXPECT_EQ("89", it->min());
    EXPECT_EQ("90", it->max());
    ++it;

    EXPECT_EQ(pd.end(), it);
    EXPECT_EQ("default_command", pd.checkCommandName());
}

TEST(PerformanceDataTest, DoubleQuotesDoNotQuote) {
    const PerformanceData pd{"\"foo bar space\"=1", "default_command"};
    auto it = pd.begin();

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("space\"", it->name().string());
    EXPECT_EQ("space\"", it->mangled_name().string());
    EXPECT_EQ("1", it->value());
    EXPECT_EQ("", it->uom());
    EXPECT_EQ("", it->warn());
    EXPECT_EQ("", it->crit());
    EXPECT_EQ("", it->min());
    EXPECT_EQ("", it->max());
    ++it;

    EXPECT_EQ(pd.end(), it);
    EXPECT_EQ("default_command", pd.checkCommandName());
}

TEST(PerformanceDataTest, ErrorEmptyUnquotedLabel) {
    const PerformanceData pd{"=1234 foo=5678", "default_command"};
    auto it = pd.begin();

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("foo", it->name().string());
    EXPECT_EQ("foo", it->mangled_name().string());
    EXPECT_EQ("5678", it->value());
    EXPECT_EQ("", it->uom());
    EXPECT_EQ("", it->warn());
    EXPECT_EQ("", it->crit());
    EXPECT_EQ("", it->min());
    EXPECT_EQ("", it->max());
    ++it;

    EXPECT_EQ(pd.end(), it);
    EXPECT_EQ("default_command", pd.checkCommandName());
}

TEST(PerformanceDataTest, ErrorEmptyQuotedLabel) {
    const PerformanceData pd{"''=1234 foo=5678", "default_command"};
    auto it = pd.begin();

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("foo", it->name().string());
    EXPECT_EQ("foo", it->mangled_name().string());
    EXPECT_EQ("5678", it->value());
    EXPECT_EQ("", it->uom());
    EXPECT_EQ("", it->warn());
    EXPECT_EQ("", it->crit());
    EXPECT_EQ("", it->min());
    EXPECT_EQ("", it->max());
    ++it;

    EXPECT_EQ(pd.end(), it);
    EXPECT_EQ("default_command", pd.checkCommandName());
}

TEST(PerformanceDataTest, ErrorFunnyChar) {
    const PerformanceData pd{"'hurz'maxKENU=1234 foo=5678", "default_command"};
    auto it = pd.begin();

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("foo", it->name().string());
    EXPECT_EQ("foo", it->mangled_name().string());
    EXPECT_EQ("5678", it->value());
    EXPECT_EQ("", it->uom());
    EXPECT_EQ("", it->warn());
    EXPECT_EQ("", it->crit());
    EXPECT_EQ("", it->min());
    EXPECT_EQ("", it->max());
    ++it;

    EXPECT_EQ(pd.end(), it);
    EXPECT_EQ("default_command", pd.checkCommandName());
}

TEST(PerformanceDataTest, ErrorQuoteInLabel) {
    const PerformanceData pd{"hurz'maxKENU=1234 foo=5678", "default_command"};
    auto it = pd.begin();

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("foo", it->name().string());
    EXPECT_EQ("foo", it->mangled_name().string());
    EXPECT_EQ("5678", it->value());
    EXPECT_EQ("", it->uom());
    EXPECT_EQ("", it->warn());
    EXPECT_EQ("", it->crit());
    EXPECT_EQ("", it->min());
    EXPECT_EQ("", it->max());
    ++it;

    EXPECT_EQ(pd.end(), it);
    EXPECT_EQ("default_command", pd.checkCommandName());
}

TEST(PerformanceDataTest, ErrorSpaceInLabel) {
    const PerformanceData pd{"hurz    foo=5678", "default_command"};
    auto it = pd.begin();

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("foo", it->name().string());
    EXPECT_EQ("foo", it->mangled_name().string());
    EXPECT_EQ("5678", it->value());
    EXPECT_EQ("", it->uom());
    EXPECT_EQ("", it->warn());
    EXPECT_EQ("", it->crit());
    EXPECT_EQ("", it->min());
    EXPECT_EQ("", it->max());
    ++it;

    EXPECT_EQ(pd.end(), it);
    EXPECT_EQ("default_command", pd.checkCommandName());
}

TEST(PerformanceDataTest, ErrorEmptyValue) {
    const PerformanceData pd{"hurz=MB;31415    foo=5678", "default_command"};
    auto it = pd.begin();

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("foo", it->name().string());
    EXPECT_EQ("foo", it->mangled_name().string());
    EXPECT_EQ("5678", it->value());
    EXPECT_EQ("", it->uom());
    EXPECT_EQ("", it->warn());
    EXPECT_EQ("", it->crit());
    EXPECT_EQ("", it->min());
    EXPECT_EQ("", it->max());
    ++it;

    EXPECT_EQ(pd.end(), it);
    EXPECT_EQ("default_command", pd.checkCommandName());
}

TEST(PerformanceDataTest, ErrorSemicolonAfterMax) {
    const PerformanceData pd{"hurz=12;34;56;78;90;    foo=5678",
                             "default_command"};
    auto it = pd.begin();

    ASSERT_NE(pd.end(), it);
    EXPECT_EQ("foo", it->name().string());
    EXPECT_EQ("foo", it->mangled_name().string());
    EXPECT_EQ("5678", it->value());
    EXPECT_EQ("", it->uom());
    EXPECT_EQ("", it->warn());
    EXPECT_EQ("", it->crit());
    EXPECT_EQ("", it->min());
    EXPECT_EQ("", it->max());
    ++it;

    EXPECT_EQ(pd.end(), it);
    EXPECT_EQ("default_command", pd.checkCommandName());
}
