// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <poll.h>
#include <unistd.h>

#include <array>
#include <cerrno>
#include <chrono>
#include <unordered_map>
#include <utility>
#include <vector>

#include "asio/buffer.hpp"
#include "asio/io_context.hpp"
#include "asio/local/connect_pair.hpp"
#include "asio/local/stream_protocol.hpp"
#include "gtest/gtest.h"
#include "livestatus/Poller.h"

using namespace std::chrono_literals;

using asio::local::connect_pair;
using asio::local::stream_protocol;

// No-op: We do not test logging here.
class PollerFixture : public ::testing::Test {
public:
    static constexpr ssize_t data_count = 2;

    static ssize_t writeFd(int fd) { return ::write(fd, "aa", data_count); }
    static ssize_t readFd(int fd) {
        char buf[data_count];
        return ::read(fd, buf, data_count);
    }
    asio::io_context io_context;
    // No asio: we are using simplest possible technic when testing
    void SetUp() override { ASSERT_NE(-1, ::pipe(fd_table.data())); }
    void TearDown() override {
        ::close(fd_table[0]);
        ::close(fd_table[1]);
        send_sock.close();
        recv_sock.close();
    }

    stream_protocol::socket send_sock{io_context};
    stream_protocol::socket recv_sock{io_context};

    Poller poller;
    std::array<int, 2> fd_table{-1, -1};
};

TEST_F(PollerFixture, Descriptors) {
    EXPECT_TRUE(poller._pollfds.empty());
    EXPECT_TRUE(poller._fd_to_pollfd.empty());
    poller.addFileDescriptor(1, PollEvents::in);
    EXPECT_FALSE(poller.isFileDescriptorSet(1, PollEvents::in));
    EXPECT_FALSE(poller.isFileDescriptorSet(2, PollEvents::in));
    EXPECT_EQ(1U, poller._pollfds.size());
    EXPECT_EQ(1U, poller._fd_to_pollfd.size());
}

TEST_F(PollerFixture, ToMask) {
    constexpr std::pair<PollEvents, short> e_pairs[] = {
        {PollEvents::in, POLLIN},
        {PollEvents::out, POLLOUT},
        {PollEvents::hup, POLLHUP},
    };
    for (const auto &p : e_pairs) {
        EXPECT_EQ(p.second, Poller::toMask(p.first));
    }
}

TEST_F(PollerFixture, WaitTimeout) {
    EXPECT_FALSE(poller.wait(2ms, fd_table[0], PollEvents::in));
    EXPECT_EQ(ETIMEDOUT, errno);
    EXPECT_FALSE(poller.wait(2ms, fd_table[1], PollEvents::in));
    EXPECT_EQ(ETIMEDOUT, errno);
}

TEST_F(PollerFixture, WaitSignaled) {
    // raise signal
    EXPECT_EQ(data_count, writeFd(fd_table[1])) << "errno = " << errno;

    // check output
    auto ret = poller.wait(10ms, fd_table[0], PollEvents::in);
    EXPECT_TRUE(ret);

    // check signal again INCORRECTLY, this is bad method for
    // the bad API 'wait' which makes two terrible things
    // - mutates(!) poller
    // - adds duplicated fd, thus making poller rather unusable
    // we check only situation with wrong file descriptor
    ret = poller.wait(10ms, fd_table[1], PollEvents::in);
    EXPECT_FALSE(ret);
    EXPECT_EQ(EBADF, errno);
}

TEST_F(PollerFixture, PollPipe) {
    ASSERT_EQ(0, poller.poll(10ms));
    poller.addFileDescriptor(fd_table[0], PollEvents::in);
    ASSERT_EQ(0, poller.poll(10ms));

    // raise signal
    EXPECT_EQ(data_count, writeFd(fd_table[1])) << "errno = " << errno;

    // check for signal 0->1
    EXPECT_EQ(1, poller.poll(10ms));
    EXPECT_TRUE(poller.isFileDescriptorSet(fd_table[0], PollEvents::in));

    // next signal 1->2
    poller.addFileDescriptor(fd_table[1], PollEvents::out);
    EXPECT_EQ(2, poller.poll(10ms));
    EXPECT_TRUE(poller.isFileDescriptorSet(fd_table[0], PollEvents::in));
    EXPECT_TRUE(poller.isFileDescriptorSet(fd_table[1], PollEvents::out));

    // remove signal
    EXPECT_EQ(data_count, readFd(fd_table[0])) << "errno = " << errno;

    // check for signal 2->1
    EXPECT_EQ(1, poller.poll(10ms));
    EXPECT_TRUE(poller.isFileDescriptorSet(fd_table[1], PollEvents::out));
    EXPECT_EQ(1, poller.poll(10ms));
}

TEST_F(PollerFixture, Asio) {
    // if you have to extend Asio testing with another foo you may move
    // the variable to fixture
    constexpr size_t asio_count = 3;

    connect_pair(send_sock, recv_sock);

    poller.addFileDescriptor(send_sock.native_handle(), PollEvents::in);
    ASSERT_EQ(0, poller.poll(10ms));

    // raise signal
    EXPECT_EQ(asio_count, recv_sock.write_some(asio::buffer("AaA", asio_count)))
        << "errno = " << errno;

    EXPECT_EQ(1, poller.poll(10ms));
    EXPECT_TRUE(
        poller.isFileDescriptorSet(send_sock.native_handle(), PollEvents::in));
}
