// Copyright (C) 2021 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <algorithm>
#include <chrono>
#include <compare>
#include <initializer_list>
#include <optional>
#include <string>
#include <string_view>
#include <thread>
#include <vector>

#include "cmc/Producer.h"
#include "gtest/gtest.h"
#include "livestatus/Logger.h"
#include "livestatus/POSIXUtils.h"

using namespace std::chrono_literals;

class LineProducerTest : public ::testing::Test {
protected:
    using lines = std::vector<std::string>;

    lines produced;

    LineProducer consume{
        [this](std::string_view line) { produced.emplace_back(line); }};
};

TEST_F(LineProducerTest, InitiallyEmpty) { EXPECT_EQ(lines{}, produced); }

TEST_F(LineProducerTest, NoNewline) {
    consume("foo");
    consume("");
    consume("bar");
    EXPECT_EQ(lines{}, produced);
}

TEST_F(LineProducerTest, OnlyNewline) {
    consume("\n");
    EXPECT_EQ(lines({""}), produced);
}

TEST_F(LineProducerTest, SeveralLines) {
    consume("foo\n");
    consume("bar\n");
    consume("blubb\n");
    EXPECT_EQ(lines({"foo", "bar", "blubb"}), produced);
}

TEST_F(LineProducerTest, GeneralCase) {
    consume("foo\nba");
    consume("r\n");
    consume("");
    consume("blu");
    consume("bb\ntail");
    EXPECT_EQ(lines({"foo", "bar", "blubb"}), produced);
}

TEST(ChunkProducerTest, SeveralChunks) {
    auto sp = SocketPair::make(SocketPair::Mode::local_non_blocking,
                               SocketPair::Direction::remote_to_local,
                               Logger::getLogger("test"));
    if (!sp) {
        FAIL() << "no socket pair";
    }

    // Emulate a sender sending the whole message in several chunks. Note that
    // due to buffering in the kernel there is no guarantee at all that the
    // receiver sees the chunks in the same way, but at least we try our best...
    std::thread writer{[&sp]() {
        for (const auto &chunk : {"bl", "ubb", "er"}) {
            writeWithTimeout(sp->remote(), chunk, 20s);
            std::this_thread::sleep_for(1ms);
        }
    }};

    std::string buffer;
    ChunkProducer consume{
        [&buffer](std::string_view chunk) { buffer += chunk; }};
    auto start = std::chrono::steady_clock::now();
    while (buffer != "blubber") {
        // Yet another ad hoc timeout, hopefully never needed... :-/
        if (std::chrono::steady_clock::now() - start > 20s) {
            writer.join();
            FAIL() << "timeout while waiting for reply";
        }
        consume(sp->local());
    }
    writer.join();
}
