// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <algorithm>
#include <chrono>
#include <cstddef>
#include <iterator>
#include <map>
#include <optional>
#include <ostream>
#include <ratio>
#include <string>
#include <vector>

#include "cmc/Schedule.h"
#include "gtest/gtest.h"
#include "livestatus/Logger.h"

using namespace std::chrono_literals;
using namespace std::string_literals;

// No-op: We do not test logging here.
std::ostream &operator<<(std::ostream &os, const Schedule<int>::Job & /*job*/) {
    return os;
}

namespace {
const auto t0 = std::chrono::system_clock::time_point{};
const auto t1 = t0 + 3001h;
const auto t2 = t0 + 3002h;
const auto t3 = t0 + 3003h;
const auto t4 = t0 + 3004h;
const auto t5 = t0 + 3005h;
}  // namespace

class ScheduleFixture : public ::testing::Test {
public:
    Schedule<int> s{Logger::getLogger("test"), "name"};
    void SetUp() override { s.logger()->setLevel(LogLevel::emergency); }
};

TEST_F(ScheduleFixture, Accessors) {
    EXPECT_EQ("name"s, s.name());
    EXPECT_TRUE(s.empty());
    EXPECT_EQ(size_t{0}, s.size());
    EXPECT_FALSE(s.extractLowestUnder(t0));
    EXPECT_FALSE(s.extractLowestUnder(t1));
    s.clear();
    EXPECT_TRUE(s.empty());
}

TEST_F(ScheduleFixture, Unload) {
    EXPECT_TRUE(s.empty());
    EXPECT_EQ(std::nullopt, s.extractFirst());

    EXPECT_EQ(t1, s.insert({0, t1}));
    EXPECT_EQ(t2, s.insert({1, t2}));
    EXPECT_EQ(t3, s.insert({2, t3}));
    EXPECT_EQ(t4, s.insert({3, t4}));
    EXPECT_EQ(4U, s.size());
    std::vector<Schedule<int>::Job> v;
    while (true) {
        if (const auto j = s.extractFirst()) {
            v.emplace_back(*j);
        } else {
            break;
        }
    }
    EXPECT_TRUE(s.empty());
    EXPECT_EQ(4U, v.size());
}

TEST_F(ScheduleFixture, Iterators) {
    EXPECT_TRUE(s.empty());

    EXPECT_EQ(t1, s.insert({0, t1}));
    EXPECT_EQ(t2, s.insert({1, t2}));
    EXPECT_EQ(t3, s.insert({2, t3}));
    EXPECT_EQ(t4, s.insert({3, t4}));
    EXPECT_EQ(4U, s.size());

    EXPECT_EQ(static_cast<long>(s.size()),
              std::distance(std::begin(s), std::end(s)));
    EXPECT_EQ(*std::next(std::begin(s), 0), 0);
    EXPECT_EQ(*std::next(std::begin(s), 1), 1);
    EXPECT_EQ(*std::next(std::begin(s), 2), 2);
    EXPECT_EQ(*std::next(std::begin(s), 3), 3);
    EXPECT_EQ(std::next(std::begin(s), 4), std::end(s));

    EXPECT_EQ(4U, s.size());
}

TEST_F(ScheduleFixture, ScheduleEpochIsNoop) {
    EXPECT_TRUE(s.empty());

    EXPECT_EQ(std::nullopt, s.insert({0, t0}));
    EXPECT_TRUE(s.empty());
}

TEST_F(ScheduleFixture, ScheduleOne) {
    EXPECT_TRUE(s.empty());

    EXPECT_EQ(t1, s.insert({0, t1}));
    EXPECT_EQ(1U, s.size());

    EXPECT_FALSE(s.extractLowestUnder(t1 - 1s));
    EXPECT_TRUE(s.extractLowestUnder(t1 + 1s));
}

TEST_F(ScheduleFixture, ScheduleAndClear) {
    EXPECT_TRUE(s.empty());

    EXPECT_EQ(t1, s.insert({0, t1}));
    EXPECT_EQ(t2, s.insert({1, t2}));
    EXPECT_EQ(t3, s.insert({2, t3}));
    EXPECT_EQ(t4, s.insert({3, t4}));
    EXPECT_EQ(4U, s.size());

    s.clear();
    EXPECT_TRUE(s.empty());
}

TEST_F(ScheduleFixture, RunnableCount) {
    EXPECT_TRUE(s.empty());

    s.insert({0, t1});
    s.insert({1, t2});
    s.insert({2, t3});
    s.insert({3, t4});

    EXPECT_EQ(4U, s.runnableCount(t5));
    EXPECT_EQ(4U, s.runnableCount(t4));
    EXPECT_EQ(3U, s.runnableCount(t3));
    EXPECT_EQ(0U, s.runnableCount(t0));

    s.clear();
}

TEST_F(ScheduleFixture, ScheduleMany) {
    EXPECT_TRUE(s.empty());

    EXPECT_EQ(t1, s.insert({1, t1}));
    EXPECT_EQ(t2, s.insert({2, t2}));
    EXPECT_EQ(t3, s.insert({3, t3}));
    EXPECT_EQ(t4, s.insert({4, t4}));
    EXPECT_EQ(4U, s.size());

    EXPECT_FALSE(s.extractLowestUnder(t0));

    const auto job1 = s.extractLowestUnder(t4 + 1s);
    EXPECT_TRUE(job1 && 1 == job1->value());
    const auto job2 = s.extractLowestUnder(t4 + 1s);
    EXPECT_TRUE(job2 && 2 == job2->value());
    const auto job3 = s.extractLowestUnder(t4 + 1s);
    EXPECT_TRUE(job3 && 3 == job3->value());
    const auto job4 = s.extractLowestUnder(t4 + 1s);
    EXPECT_TRUE(job4 && 4 == job4->value());

    EXPECT_TRUE(s.empty());

    ASSERT_FALSE(s.extractLowestUnder(t4 + 1s));
}

TEST_F(ScheduleFixture, TestLowestUnder) {
    EXPECT_TRUE(s.empty());

    EXPECT_EQ(t1, s.insert({1, t1}));
    EXPECT_EQ(t2, s.insert({2, t2}));
    EXPECT_EQ(t3, s.insert({3, t3}));
    EXPECT_EQ(t4, s.insert({4, t4}));
    EXPECT_EQ(4U, s.size());

    const auto job1 = s.lowestUnder(t4);
    EXPECT_TRUE(job1 && 1 == job1->value());
    EXPECT_EQ(4U, s.size());

    EXPECT_FALSE(s.lowestUnder(t0));
    EXPECT_EQ(4U, s.size());

    const auto job3 = s.lowestUnder(t3);
    EXPECT_TRUE(job3 && 1 == job3->value());
    EXPECT_EQ(4U, s.size());
}

TEST_F(ScheduleFixture, LowestUnderOnEmpty) {
    EXPECT_TRUE(s.empty());

    EXPECT_FALSE(s.lowestUnder(t0));
    EXPECT_TRUE(s.empty());

    EXPECT_FALSE(s.lowestUnder(t4));
    EXPECT_TRUE(s.empty());
}

TEST_F(ScheduleFixture, TestExtractLowestUnder) {
    EXPECT_TRUE(s.empty());

    EXPECT_EQ(t1, s.insert({1, t1}));
    EXPECT_EQ(t2, s.insert({2, t2}));
    EXPECT_EQ(t3, s.insert({3, t3}));
    EXPECT_EQ(t4, s.insert({4, t4}));
    EXPECT_EQ(4U, s.size());

    auto now = t2 + 5min;
    EXPECT_LT(t2, now);
    EXPECT_GT(t3, now);

    // Before now.
    const auto job1 = s.extractLowestUnder(now);
    EXPECT_TRUE(job1 && 1 == job1->value());
    EXPECT_EQ(3U, s.size());
    const auto job2 = s.extractLowestUnder(now);
    EXPECT_TRUE(job2 && 2 == job2->value());
    EXPECT_EQ(2U, s.size());

    // After now.
    EXPECT_FALSE(s.extractLowestUnder(now));
    EXPECT_EQ(2U, s.size());
    EXPECT_FALSE(s.extractLowestUnder(now));
    EXPECT_EQ(2U, s.size());
}

TEST_F(ScheduleFixture, RescheduleJobDoesNotDuplicate) {
    EXPECT_TRUE(s.empty());

    EXPECT_EQ(t1, s.insert({1, t1}));
    EXPECT_EQ(t2, s.insert({2, t2}));
    EXPECT_EQ(2U, s.size());

    EXPECT_EQ(t3, s.insert({1, t3}));
    EXPECT_EQ(t4, s.insert({2, t4}));
    EXPECT_EQ(2U, s.size());

    const auto job1 = s.extractLowestUnder(t4);
    EXPECT_TRUE(job1 && 1 == job1->value());
    const auto job2 = s.extractLowestUnder(t4);
    EXPECT_TRUE(job2 && 2 == job2->value());
    EXPECT_TRUE(s.empty());
}

TEST_F(ScheduleFixture, ExtractLowestUnderOnEmptyIsNoop) {
    EXPECT_TRUE(s.empty());

    EXPECT_FALSE(s.extractLowestUnder(t0));
    EXPECT_FALSE(s.extractLowestUnder(t1));
    EXPECT_FALSE(s.extractLowestUnder(t4));

    EXPECT_TRUE(s.empty());
}

TEST_F(ScheduleFixture, ExtractLowestUnderEarliestIsNoop) {
    EXPECT_TRUE(s.empty());

    EXPECT_EQ(t1, s.insert({1, t1}));
    EXPECT_EQ(t2, s.insert({2, t2}));
    EXPECT_EQ(t3, s.insert({3, t3}));
    EXPECT_EQ(t4, s.insert({4, t4}));
    EXPECT_EQ(4U, s.size());

    EXPECT_FALSE(s.extractLowestUnder(t0));
    EXPECT_EQ(4U, s.size());
}
