// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <cstdlib>
#include <string>
#include <unordered_set>
#include <utility>
#include <vector>

#include "cmc/SerialToken.h"
#include "gtest/gtest.h"
#include "livestatus/Logger.h"

class SerialTokenFixture : public ::testing::Test {
public:
    SerialTokenFactory stf_{Logger::getLogger("test")};

    SerialToken createToken(config_serial_t serial) {
        // Short strings are kept in the string object itself, but we want to
        // test the general case. In addition, test temporariy arguments, too.
        return stf_.create("a_long_string_to_trigger_dynamic_allocation" +
                               std::to_string(::random()),
                           serial);
    }

    bool isUsed(config_serial_t serial) const {
        return stf_.used().contains(serial);
    }

    size_t numSerialsInUse() const { return stf_.used().size(); }
};

TEST_F(SerialTokenFixture, SimpleCreate) {
    // arrange -----------------------------------------------------------------
    const config_serial_t serial{11};
    ASSERT_FALSE(isUsed(serial));
    ASSERT_EQ(size_t{0}, numSerialsInUse());
    {
        // act -----------------------------------------------------------------
        auto token{createToken(serial)};
        // assert --------------------------------------------------------------
        EXPECT_EQ(serial, token.value());
        EXPECT_TRUE(isUsed(serial));
        EXPECT_EQ(size_t{1}, numSerialsInUse());
    }
    EXPECT_FALSE(isUsed(serial));
    EXPECT_EQ(size_t{0}, numSerialsInUse());
}

TEST_F(SerialTokenFixture, NestedCreate) {
    // arrange -----------------------------------------------------------------
    const config_serial_t serial{11};
    ASSERT_FALSE(isUsed(serial));
    ASSERT_EQ(size_t{0}, numSerialsInUse());
    {
        auto token1{createToken(serial)};
        ASSERT_TRUE(isUsed(serial));
        ASSERT_EQ(size_t{1}, numSerialsInUse());
        {
            // act -------------------------------------------------------------
            auto token2{createToken(serial)};
            // assert ----------------------------------------------------------
            EXPECT_TRUE(isUsed(serial));
            EXPECT_EQ(size_t{1}, numSerialsInUse());
        }
        EXPECT_TRUE(isUsed(serial));
        EXPECT_EQ(size_t{1}, numSerialsInUse());
    }
    EXPECT_FALSE(isUsed(serial));
    EXPECT_EQ(size_t{0}, numSerialsInUse());
}

TEST_F(SerialTokenFixture, MultipleCreate) {
    // arrange -----------------------------------------------------------------
    const std::vector<config_serial_t> serials{11, 22, 33, 44, 55,
                                               66, 77, 88, 99};
    {
        std::vector<SerialToken> tokens;
        tokens.reserve(serials.size());
        // act -----------------------------------------------------------------
        for (auto serial : serials) {
            tokens.push_back(createToken(serial));
        }
        // assert --------------------------------------------------------------
        for (auto serial : serials) {
            EXPECT_TRUE(isUsed(serial));
        }
        EXPECT_EQ(tokens.size(), numSerialsInUse());
    }
    for (auto serial : serials) {
        EXPECT_FALSE(isUsed(serial));
    }
    EXPECT_EQ(size_t{0}, numSerialsInUse());
}

TEST_F(SerialTokenFixture, MoveConstruction) {
    // arrange -----------------------------------------------------------------
    const config_serial_t serial{11};
    {
        auto token{createToken(serial)};
        ASSERT_EQ(serial, token.value());
        ASSERT_TRUE(isUsed(serial));
        ASSERT_EQ(size_t{1}, numSerialsInUse());
        // act -----------------------------------------------------------------
        auto moved{std::move(token)};
        // assert --------------------------------------------------------------
        EXPECT_EQ(serial, moved.value());
        EXPECT_TRUE(isUsed(serial));
        EXPECT_EQ(size_t{1}, numSerialsInUse());
    }
    EXPECT_FALSE(isUsed(serial));
    EXPECT_EQ(size_t{0}, numSerialsInUse());
}

TEST_F(SerialTokenFixture, MoveAssignment) {
    // arrange -----------------------------------------------------------------
    const config_serial_t serial1{11};
    const config_serial_t serial2{22};
    {
        auto token1{createToken(serial1)};
        auto token2{createToken(serial2)};
        ASSERT_EQ(serial1, token1.value());
        ASSERT_TRUE(isUsed(serial1));
        ASSERT_EQ(serial2, token2.value());
        ASSERT_TRUE(isUsed(serial2));
        ASSERT_EQ(size_t{2}, numSerialsInUse());
        // act -----------------------------------------------------------------
        token2 = std::move(token1);
        // assert --------------------------------------------------------------
        EXPECT_TRUE(isUsed(serial1));
        EXPECT_FALSE(isUsed(serial2));
        EXPECT_EQ(size_t{1}, numSerialsInUse());
    }
    EXPECT_FALSE(isUsed(serial1));
    EXPECT_FALSE(isUsed(serial2));
    EXPECT_EQ(size_t{0}, numSerialsInUse());
}
