// Copyright (C) 2019 Checkmk GmbH - License: Check_MK Enterprise License
// This file is part of Checkmk (https://checkmk.com). It is subject to the
// terms and conditions defined in the file COPYING, which is part of this
// source code package.

#include <chrono>
#include <ctime>
#include <iterator>
#include <vector>

#include "cmc/Timeperiod.h"
#include "gtest/gtest.h"

namespace {
using system_clock = std::chrono::system_clock;

system_clock::time_point _(std::time_t t) {
    return system_clock::from_time_t(t);
}
}  // namespace

class TimePeriodFixture : public ::testing::Test {
public:
    Timeperiod tp{"tp_name",
                  "tp_alias",
                  {_(1500000000), _(1510000000), _(1520000000), _(1530000000),
                   _(1540000000), _(1550000000)}};
};

TEST_F(TimePeriodFixture, TestAccessors) {
    EXPECT_EQ("tp_name", tp.name());
    EXPECT_EQ("tp_alias", tp.alias());
    EXPECT_EQ(6, tp.numTransitions());
    EXPECT_EQ(_(1500000000), tp.transition(0));
    EXPECT_EQ(_(1540000000), tp.transition(4));
    EXPECT_EQ(_(1500000000), *tp.begin());
    EXPECT_EQ(_(1550000000), *std::prev(tp.end()));
    EXPECT_EQ(_(0), tp.lastStateChange());
    EXPECT_EQ(_(0), tp.previousLastStateChange());
}

TEST_F(TimePeriodFixture, TestFindTransition) {
    EXPECT_EQ(-1, tp.findTransition(_(0)));
    EXPECT_EQ(-1, tp.findTransition(_(1500000000 - 1)));
    EXPECT_EQ(0, tp.findTransition(_(1500000000)));
    EXPECT_EQ(3, tp.findTransition(_(1540000000 - 1)));
    EXPECT_EQ(5, tp.findTransition(_(1550000000)));
    EXPECT_EQ(5, tp.findTransition(_(1600000000)));
};

TEST_F(TimePeriodFixture, TestIsActiveTime) {
    EXPECT_FALSE(tp.isActiveTime(_(1500000000 - 1)));
    EXPECT_TRUE(tp.isActiveTime(_(1500000000)));
    EXPECT_FALSE(tp.isActiveTime(_(1510000000)));
    EXPECT_TRUE(tp.isActiveTime(_(1520000000)));
}

TEST_F(TimePeriodFixture, TestNextActiveTime) {
    EXPECT_EQ(_(1500000000), tp.nextActiveTime(_(0)));
    EXPECT_EQ(_(1520000000), tp.nextActiveTime(_(1510000000)));
    EXPECT_EQ(_(1520000000), tp.nextActiveTime(_(1520000000)));
    EXPECT_EQ(_(1540000000), tp.nextActiveTime(_(1530000000)));
    EXPECT_EQ(_(1540000000), tp.nextActiveTime(_(1540000000)));
    EXPECT_EQ(_(0), tp.nextActiveTime(_(1550000000)));
    EXPECT_EQ(_(0), tp.nextActiveTime(_(1600000000)));
}

TEST_F(TimePeriodFixture, TestUpdate) {
    EXPECT_EQ(0, tp.nextTransition());
    EXPECT_EQ(_(1500000000), tp.nextTransitionTime());

    tp.update(nullptr, _(1500000000 + 1000));
    EXPECT_EQ(1, tp.nextTransition());
    EXPECT_EQ(_(1510000000), *tp.begin());
    EXPECT_EQ(_(1510000000), tp.nextTransitionTime());
    EXPECT_EQ(_(1500000000), tp.lastStateChange());
    EXPECT_EQ(_(0), tp.previousLastStateChange());

    tp.update(nullptr, _(1500000000 + 2000));
    EXPECT_EQ(1, tp.nextTransition());
    EXPECT_EQ(_(1510000000), *tp.begin());
    EXPECT_EQ(_(1510000000), tp.nextTransitionTime());
    EXPECT_EQ(_(1500000000), tp.lastStateChange());
    EXPECT_EQ(_(0), tp.previousLastStateChange());

    tp.update(nullptr, _(1510000000));
    EXPECT_EQ(2, tp.nextTransition());
    EXPECT_EQ(_(1520000000), *tp.begin());
    EXPECT_EQ(_(1520000000), tp.nextTransitionTime());
    EXPECT_EQ(_(1510000000), tp.lastStateChange());
    EXPECT_EQ(_(1500000000), tp.previousLastStateChange());

    tp.update(nullptr, _(1550000000));
    EXPECT_EQ(6, tp.nextTransition());
    EXPECT_EQ(tp.end(), tp.begin());
    EXPECT_EQ(_(0), tp.nextTransitionTime());
    EXPECT_EQ(_(1550000000), tp.lastStateChange());
    EXPECT_EQ(_(1540000000), tp.previousLastStateChange());

    tp.update(nullptr, _(1560000000));
    EXPECT_EQ(6, tp.nextTransition());
    EXPECT_EQ(tp.end(), tp.begin());
    EXPECT_EQ(_(0), tp.nextTransitionTime());
    EXPECT_EQ(_(1550000000), tp.lastStateChange());
    EXPECT_EQ(_(1540000000), tp.previousLastStateChange());

    tp.update(nullptr, _(1600000000));
    EXPECT_EQ(6, tp.nextTransition());
    EXPECT_EQ(tp.end(), tp.begin());
    EXPECT_EQ(_(0), tp.nextTransitionTime());
    EXPECT_EQ(_(1550000000), tp.lastStateChange());
    EXPECT_EQ(_(1540000000), tp.previousLastStateChange());
}
