#!/usr/bin/env python3
# Copyright (C) 2023 Checkmk GmbH - License: GNU General Public License v2
# This file is part of Checkmk (https://checkmk.com). It is subject to the terms and
# conditions defined in the file COPYING, which is part of this source code package.

import argparse
import os
import re
from pathlib import Path

from . import load_werk
from .config import load_config


def main(
    werks_to_check: list[Path],
    werks_config: Path,
    defines_make: Path,
    version_regex: re.Pattern[str],
) -> None:
    if werks_to_check:
        pass
    elif changed_werk_files := os.environ.get("CHANGED_WERK_FILES"):
        werks_to_check = list(Path(line) for line in changed_werk_files.split(" ") if line)
    else:
        werks_to_check = list(
            path
            for path in Path(".werks").iterdir()
            if path.name.isdigit() or path.name.endswith(".md")
        )

    config = load_config(werks_config, defines_make)
    choices_component = {e[0] for e in config.all_components()}

    for werk_path in werks_to_check:
        werk_content = werk_path.read_text(encoding="utf-8")
        try:
            werk = load_werk(file_content=werk_content, file_name=werk_path.name)

            if werk.component not in choices_component:
                raise ValueError(
                    f"Component {werk.component!r} not defined in config.\n"
                    f"Available values: {choices_component!r}"
                )
            if not version_regex.match(werk.version):
                raise ValueError(f"Version {werk.version!r} is not valid")

        except Exception as e:
            raise RuntimeError(f"Error while loading werk {werk_path}\n\n{werk_content}") from e

    print(f"Successfully validated {len(werks_to_check)} werks")


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version-regex",
        help="regex used to validate the version in the werk",
        default=r"^\d.\d.\d([ipb]\d+)?$",
        type=lambda x: re.compile(x),  # pylint: disable=unnecessary-lambda
    )
    parser.add_argument(
        "--werk-config",
        help="path to werks config, typically .werks/config",
        default=".werks/config",
        type=Path,
    )
    parser.add_argument(
        "--defines-make",
        help="path to defines.make to load current version, typically defines.make",
        default="defines.make",
        type=Path,
    )
    parser.add_argument(
        "werk_paths",
        help="werk paths to validate, can also be passed via $CHANGED_WERK_FILES",
        nargs="*",
        type=Path,
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    main(args.werk_paths, args.werk_config, args.defines_make, args.version_regex)
