import logging
import pprint
import re
import subprocess
import time
from collections.abc import Iterator

import pytest

from tests.testlib.pytest_helpers.marks import skip_if_saas_edition
from tests.testlib.site import Site

from cmk.ec.config import (  # pylint: disable=cmk-module-layer-violation
    ECRulePackSpec,
    EventLimit,
    Rule,
    ServiceLevel,
    State,
)

from cmk.gui.watolib.site_changes import ChangeSpec  # pylint: disable=cmk-module-layer-violation

logger = logging.getLogger(__name__)


def _validate_process_return_code(process: subprocess.Popen, assert_msg: str) -> None:
    """Validate return code is 0.

    Raises `AssertionError`, if not. `AssertionError` is raised with
    + custom message
    + stdout generated by the process
    + stderr generated by the process.

    NOTE: `process` should use subprocess.PIPE for stdout and stderr.
    """
    if process.wait() != 0:
        p_out, p_err = process.communicate()
        msg = (assert_msg, f"STDOUT: {p_out}", f"STDERR: {p_err}")
        raise AssertionError("\n".join(msg))


def _get_ec_rule_packs(rule_id: str, title: str, state: State, match: str) -> list[ECRulePackSpec]:
    """EC rule to inject in the test-site"""
    rule = Rule(
        id=rule_id,
        description="",
        comment="",
        docu_url="",
        disabled=False,
        state=state,
        sl=ServiceLevel(value=0, precedence="message"),
        actions=[],
        actions_in_downtime=True,
        cancel_actions=[],
        cancel_action_phases="always",
        autodelete=False,
        event_limit=EventLimit(action="None", limit=0),
        match=match,
        invert_matching=False,
    )
    return [
        ECRulePackSpec(
            id=rule_id,
            title=title,
            disabled=False,
            rules=[rule],
        )
    ]


def _get_replication_change() -> ChangeSpec:
    """Replication change to inject in the test-site"""
    return ChangeSpec(
        id="",
        action_name="edit-rule-pack",
        text="Modified rule pack test",
        object=None,
        user_id="cmkadmin",
        domains=["ec"],
        time=0,
        need_sync=True,
        need_restart=True,
        domain_settings={},
        prevent_discard_changes=False,
        diff_text=None,
        has_been_activated=False,
    )


def _write_ec_rule(site: Site, rule: list | None) -> None:
    ec_rules_path = site.path("etc/check_mk/mkeventd.d/wato/rules.mk")
    site.write_text_file(str(ec_rules_path), f"rule_packs += {rule}" if rule else "")


def _activate_ec_changes(site: Site) -> None:
    replication_changes_path = site.path(f"var/check_mk/wato/replication_changes_{site.id}.mk")
    site.write_text_file(str(replication_changes_path), str(_get_replication_change()))
    site.openapi.activate_changes_and_wait_for_completion(force_foreign_changes=True, timeout=120)


def _generate_event_message(site: Site, message: str) -> None:
    """Generate EC message via Unix socket"""
    events_path = site.path("tmp/run/mkeventd/events")
    cmd = f"sudo su -l {site.id} -c 'echo {message} > {events_path}'"
    with subprocess.Popen(
        cmd,
        encoding="utf-8",
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as process:
        _validate_process_return_code(process, "Failed to generate EC message via Unix socket.")


def _wait_for_queried_column(
    site: Site, query: str, sleep_time: float = 1, max_count: int = 20
) -> list[str]:
    count = 0
    while not (queried_column := site.live.query_column(query)) and count < max_count:
        logger.info("Waiting for the following livestatus query: %s", repr(query))
        time.sleep(sleep_time)
        count += 1
    assert queried_column, f"Failed to retrieve livestatus query: {repr(query)}"
    return queried_column


def _get_snmp_trap_cmd(event_message: str) -> list:
    return [
        "snmptrap",
        "-v",
        "1",
        "-c",
        "public",
        "127.0.0.1",
        ".1.3.6.1",
        "192.168.178.30",
        "6",
        "17",
        '""',
        ".1.3.6.1",
        "s",
        f'"{event_message}"',
    ]


@pytest.fixture(name="setup_ec", scope="function")
def _setup_ec(site: Site) -> Iterator:
    match = "dummy"
    rule_id = f"test {match}"
    rule_state: State = 1

    _write_ec_rule(
        site,
        _get_ec_rule_packs(title="", rule_id=rule_id, state=rule_state, match=match),
    )

    # in order for the EC rule to take effect, we need to inject a change in the EC domain and
    # perform a changes' activation
    _activate_ec_changes(site)

    yield match, rule_id, rule_state

    # cleanup: remove EC rules and activate changes
    _write_ec_rule(site, rule=None)
    _activate_ec_changes(site)

    # cleanup: archive generated events
    for event_id in site.live.query_column("GET eventconsoleevents\nColumns: event_id\n"):
        resp = site.openapi.post(
            "domain-types/event_console/actions/delete/invoke",
            headers={"Content-Type": "application/json"},
            json={
                "site_id": site.id,
                "filter_type": "by_id",
                "event_id": event_id,
            },
        )

        assert resp.status_code == 204, pprint.pformat(resp.json())


@pytest.fixture(name="restart_site", scope="module")
def _restart_site(site: Site) -> Iterator[None]:
    site.stop()
    yield
    site.start()


@pytest.fixture(name="enable_snmp_trap_receiver", scope="module")
def _enable_snmp_trap_receiver(site: Site, restart_site: None) -> Iterator[None]:
    with site.execute(
        ["omd", "config", "show", "MKEVENTD_SNMPTRAP"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as process:
        _validate_process_return_code(process, "Failed to retrieve SNMP trap receiver status.")
        initial_config = process.communicate()[0].strip()

    logger.info("Setting SNMP trap receiver to 'on'...")
    with site.execute(
        ["omd", "config", "set", "MKEVENTD_SNMPTRAP", "on"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as process:
        _validate_process_return_code(process, "Failed to change SNMP trap receivers.")
    site.start()

    yield

    site.stop()
    logger.info("Setting SNMP trap receiver to '%s'...", initial_config)
    with site.execute(
        ["omd", "config", "set", "MKEVENTD_SNMPTRAP", f"{initial_config}"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as process:
        _validate_process_return_code(process, "Failed to change SNMP trap receivers.")


@pytest.fixture(name="enable_snmp_trap_translation", scope="function")
def _enable_snmp_trap_translation(site: Site) -> Iterator[None]:
    logger.info("Enabling SNMP trap translation...")
    ec_global_rules_path = site.path("etc/check_mk/mkeventd.d/wato/9999-test_ec.mk")
    site.write_text_file(str(ec_global_rules_path), "translate_snmptraps = (True, {})")
    _activate_ec_changes(site)

    yield

    logger.info("Disabling SNMP trap translation...")
    site.delete_file(str(ec_global_rules_path))
    _activate_ec_changes(site)


@skip_if_saas_edition(reason="EC is disabled in the SaaS edition")
def test_ec_rule_match(site: Site, setup_ec: Iterator) -> None:
    """Generate a message matching an EC rule and assert an event is created"""
    match, rule_id, rule_state = setup_ec

    event_message = f"some {match} status"
    _generate_event_message(site, event_message)

    # retrieve id of matching rule via livestatus query
    queried_rule_ids = _wait_for_queried_column(site, "GET eventconsolerules\nColumns: rule_id\n")
    assert rule_id in queried_rule_ids

    # retrieve matching event state via livestatus query
    queried_event_states = _wait_for_queried_column(
        site, "GET eventconsoleevents\nColumns: event_state\n"
    )
    assert len(queried_event_states) == 1
    assert queried_event_states[0] == rule_state

    # retrieve matching event message via livestatus query
    queried_event_messages = _wait_for_queried_column(
        site, "GET eventconsoleevents\nColumns: event_text"
    )
    assert len(queried_event_messages) == 1
    assert queried_event_messages[0] == event_message


@skip_if_saas_edition(reason="EC is disabled in the SaaS edition")
def test_ec_rule_no_match(site: Site, setup_ec: Iterator) -> None:
    """Generate a message not matching any EC rule and assert no event is created"""
    match, _, _ = setup_ec
    event_message = "some other status"
    assert match not in event_message

    _generate_event_message(site, event_message)

    queried_event_states = site.live.query_column("GET eventconsoleevents\nColumns: event_state\n")
    assert not queried_event_states

    queried_event_messages = site.live.query_column("GET eventconsoleevents\nColumns: event_text")
    assert not queried_event_messages


@skip_if_saas_edition(reason="EC is disabled in the SaaS edition")
def test_ec_rule_match_snmp_trap(
    site: Site, setup_ec: Iterator, enable_snmp_trap_receiver: None
) -> None:
    """Generate a message via SNMP trap matching an EC rule and assert an event is created"""
    match, rule_id, rule_state = setup_ec
    event_message = f"some {match} status"

    process = site.execute(
        _get_snmp_trap_cmd(event_message),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    _validate_process_return_code(process, "Failed to send message via SNMP trap.")

    assert event_message in site.read_file("var/log/mkeventd.log")

    # retrieve id of matching rule via livestatus query
    queried_rule_ids = _wait_for_queried_column(site, "GET eventconsolerules\nColumns: rule_id\n")
    assert rule_id in queried_rule_ids

    # retrieve matching event state via livestatus query
    queried_event_states = _wait_for_queried_column(
        site, "GET eventconsoleevents\nColumns: event_state\n"
    )
    assert len(queried_event_states) == 1
    assert queried_event_states[0] == rule_state

    # retrieve matching event message via livestatus query
    queried_event_messages = _wait_for_queried_column(
        site, "GET eventconsoleevents\nColumns: event_text"
    )
    assert len(queried_event_messages) == 1
    assert event_message in queried_event_messages[0]


@skip_if_saas_edition(reason="EC is disabled in the SaaS edition")
def test_ec_rule_no_match_snmp_trap(
    site: Site, setup_ec: Iterator, enable_snmp_trap_receiver: None
) -> None:
    """Generate a message via SNMP trap not matching any EC rule and assert no event is created"""
    match, _, _ = setup_ec
    event_message = "some other status"
    assert match not in event_message

    process = site.execute(
        _get_snmp_trap_cmd(event_message),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    _validate_process_return_code(process, "Failed to send message via SNMP trap.")

    queried_event_states = site.live.query_column("GET eventconsoleevents\nColumns: event_state\n")
    assert not queried_event_states

    queried_event_messages = site.live.query_column("GET eventconsoleevents\nColumns: event_text")
    assert not queried_event_messages


@skip_if_saas_edition(reason="EC is disabled in the SaaS edition")
def test_ec_global_settings(
    site: Site,
    setup_ec: Iterator,
    enable_snmp_trap_receiver: None,
    enable_snmp_trap_translation: None,
) -> None:
    """Assert that global settings of the EC are applied to the EC

    * Activate SNMP traps translation via EC global rules
    * Send message via SNMP trap
    * Assert SNMP-MIB is found in the event message
    """
    match, _, _ = setup_ec
    event_message = f"some {match} status"

    process = site.execute(
        _get_snmp_trap_cmd(event_message),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    _validate_process_return_code(process, "Failed to send message via SNMP trap.")

    queried_event_messages = _wait_for_queried_column(
        site, "GET eventconsoleevents\nColumns: event_text"
    )
    assert len(queried_event_messages) == 1

    pattern = "SNMP.*MIB"  # pattern expected after SNMP traps translation
    assert re.compile(pattern).search(
        queried_event_messages[0]
    ), f"{pattern} not found in the event message:\n {queried_event_messages[0]}"
