#!/usr/bin/env python3
# Copyright (C) 2019 Checkmk GmbH - License: GNU General Public License v2
# This file is part of Checkmk (https://checkmk.com). It is subject to the terms and
# conditions defined in the file COPYING, which is part of this source code package.

from collections.abc import Iterable

import pytest

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    Attributes,
    HostLabel,
    TableRow,
)
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    InventoryResult,
    StringTable,
)
from cmk.base.plugins.agent_based.inventory_snmp_extended_info import (
    inventory_snmp_extended_info,
    parse_snmp_extended_info,
)

from cmk.plugins.lib.device_types import get_device_type_label

from .utils_inventory import sort_inventory_result


@pytest.mark.parametrize(
    "string_table, expected_result",
    [
        (
            [
                ["_", "fibrechannel switch", "_", "_", "_", "_", "_", "_", "_"],
                ["_", "_", "_", "_", "_", "_", "_", "_", "_"],
            ],
            [HostLabel("cmk/device_type", "fcswitch")],
        ),
        (
            [
                ["_", "switch (fc1)", "_", "_", "_", "_", "_", "_", "_"],
            ],
            [HostLabel("cmk/device_type", "switch")],
        ),
        (
            [
                ["_", "fc switch", "_", "_", "_", "_", "_", "_", "_"],
            ],
            [HostLabel("cmk/device_type", "fcswitch")],
        ),
    ],
)
def test_inventory_snmp_extended_info_host_labels(
    string_table: StringTable, expected_result: Iterable[HostLabel]
) -> None:
    section = parse_snmp_extended_info(string_table)
    assert list(get_device_type_label(section)) == expected_result


@pytest.mark.parametrize(
    "string_table, expected_result",
    [
        ([], []),
        (
            # Only one parent
            [
                [
                    "child",
                    "description",
                    "0",
                    "1",
                    "name",
                    "software",
                    "serial",
                    "manufacturer",
                    "model",
                ],
            ],
            [
                Attributes(
                    path=["hardware", "system"],
                    inventory_attributes={
                        "serial": "serial",
                        "model": "model",
                    },
                    status_attributes={},
                ),
            ],
        ),
        (
            # Ignore ports ('10')
            [
                [
                    "child",
                    "description",
                    "parent",
                    "10",
                    "name",
                    "software",
                    "serial",
                    "manufacturer",
                    "model",
                ],
            ],
            [],
        ),
        (
            # Only children
            [
                [
                    "child1",
                    "description",
                    "1",
                    "1",
                    "name1",
                    "software",
                    "serial",
                    "manufacturer",
                    "model",
                ],
                [
                    "child2",
                    "description",
                    "2",
                    "1",
                    "name2",
                    "software",
                    "serial",
                    "manufacturer",
                    "model",
                ],
            ],
            [
                TableRow(
                    path=["hardware", "components", "others"],
                    key_columns={
                        "index": "child1",
                        "name": "name1",
                    },
                    inventory_columns={
                        "description": "description",
                        "software": "software",
                        "serial": "serial",
                        "manufacturer": "manufacturer",
                        "model": "model",
                        "location": "Missing in ENTITY table (1)",
                    },
                    status_columns={},
                ),
                TableRow(
                    path=["hardware", "components", "others"],
                    key_columns={
                        "index": "child2",
                        "name": "name2",
                    },
                    inventory_columns={
                        "description": "description",
                        "software": "software",
                        "serial": "serial",
                        "manufacturer": "manufacturer",
                        "model": "model",
                        "location": "Missing in ENTITY table (2)",
                    },
                    status_columns={},
                ),
            ],
        ),
    ],
)
def test_inventory_snmp_extended_info(
    string_table: StringTable, expected_result: InventoryResult
) -> None:
    assert sort_inventory_result(
        inventory_snmp_extended_info(parse_snmp_extended_info(string_table))
    ) == sort_inventory_result(expected_result)
